const path = require("path");
const fs = require("fs");
const Tool = require("./libs/Tool");
const LayaJsonParse = require("./libs/LayaJsonParse");
const DependencieObj = require("./libs/DependencieObj");
const FileObj = require("./libs/FileObj");
const PackageObj = require("./libs/packageObj");
const ZipObj = require("./libs/ZipObj");
const Laya = require("./libs/laya.core copy");
const UZIP = require("./libs/UZIP");

class AssetBundle {
    constructor() {
        this.Version = 'v1.0.1';
        this.debug = 0;
        this.LayaRoot = 'unity/';
        this.fileConfs = {
            laya: {},
            other: {}
        };
        this.dependencieObj = new DependencieObj(); //依赖实例
        this.Multifiles = [];
    }

    run() {
        this.getAssetConf(); //获取配置
        for (let name in this.fileConfs.laya) {
            let item = this.fileConfs.laya[name];
            if (item && item.length > 0) {
                //遍历目录所有文件
                let packageObj = this.getConventionaFiles(item[0]);
                this.dependencieObj.packages[name] = packageObj;
                //标记依赖
                this.packageWalk(item, packageObj);
            }
        }

        //提取package共用列表
        for (let name in this.dependencieObj.packages) {
            let item = this.dependencieObj.packages[name];
            let commonFiles = this.getPackageComnFiles(item);
            item.commonFiles = commonFiles;
            item.dlist = item.list;
        }

        //创建zip包
        for (let name in this.dependencieObj.zips) {
            let item = this.dependencieObj.zips[name];
            this.createZip(item);
        }
        //删除空目录
        this.delDirWalk(this.LayaRoot);
    }

    //例遍删除空文件夹
    delDirWalk(_path) {
        let self = this;
        let files = fs.readdirSync(_path);
        files.forEach(function (item) {
            let tmpPath = path.posix.join(_path, item);
            let stats = fs.statSync(tmpPath);
            if (stats.isDirectory()) {
                let _files = fs.readdirSync(tmpPath);
                if (_files.length == 0) {
                    //空文件夹
                    // console.log('空文件夹', tmpPath);
                    fs.rmdirSync(tmpPath);
                } else {
                    self.delDirWalk(tmpPath);
                }
            }
        });
    }

    //生成zip包
    createZip(zipObj) {
        if (zipObj.dependencies && Array.isArray(zipObj.dependencies)) {
            //排除依赖文件
            let packname = zipObj.info.packname;
            let _package = this.dependencieObj.packages[packname];
            let commonFiles = _package.commonFiles;
            let files = [];
            if (zipObj.info.isDir == 0) {
                for (let i = 0; i < zipObj.dependencies.length; i++) {
                    let _file = zipObj.dependencies[i];
                    let index_common = commonFiles.indexOf(_file);
                    //首包所有文件都包含
                    if (zipObj.info.main == 1) {
                        files.push(_file);
                        let index = _package.dlist.indexOf(_file);
                        if (index > -1) delete _package.dlist[index];
                    } else {
                        //不在公共
                        if (index_common == -1) {
                            files.push(_file);
                            let index = _package.dlist.indexOf(_file);
                            if (index > -1) delete _package.dlist[index];
                        }
                    }
                }
            } else { //是目录                
                files = _package.dlist.filter(d => d); //删除数组中空项
            }
            //获取所有文件的buffer
            let fileObjList = [];
            for (let i = 0, len = files.length; i < len; i++) {
                let _file = files[i];
                let fileObj = this.dependencieObj.allfiles[_file];
                if (fileObj) {
                    let _buffer = fileObj.getBuffer();
                    if (_buffer && _buffer != null) {
                        fileObjList.push({
                            filePath: _file,
                            buffer: _buffer
                        });
                    }
                } else {
                    //打包图片
                    let imgObj = new FileObj(_file);
                    let _buffer = imgObj.getBuffer();
                    if (_buffer && _buffer != null) {
                        fileObjList.push({
                            filePath: _file,
                            buffer: _buffer
                        });
                    }
                }
            }
            if (files.length > 0) this.mergeBinaryFile(fileObjList, zipObj);
        }
    }

    mergeBinaryFile(arrayBufferList, zipObj) {
        let len = arrayBufferList.length;
        //构造
        let byte = new Laya.Byte();
        byte.writeInt32(len); // 有多少个文件
        let arrayItem;
        let arrayBuffer;
        let filePath;
        let byteLength;
        let maptext = '';
        let bytezip = 0;
        for (let i = 0; i < len; i++) {
            arrayItem = arrayBufferList[i];
            arrayBuffer = Tool.toArrayBuffer(arrayItem.buffer);
            filePath = arrayItem.filePath;
            filePath = filePath.replace(/\\/g, "/");
            byteLength = arrayBuffer.byteLength;
            byte.writeUTFString(filePath); // 名字
            byte.writeInt32(byteLength); // 大小
            byte.writeArrayBuffer(arrayBuffer, 0, byteLength); // 内容
            maptext += "\t" + filePath + "\t" + Math.ceil(byteLength / 1024) + ' KB\n';
            bytezip += byteLength;
            if (this.debug == 0) {
                //删除文件
                fs.unlinkSync(filePath);
            }
        }
        let buffer = Buffer.from(byte.buffer);
        //压缩文件
        buffer = UZIP.deflate(buffer);
        let outPath = path.join("assets", zipObj.zipname + '.zip');
        fs.writeFileSync(outPath, buffer);
        console.log('打包 ' + zipObj.zipname + '.zip');
        let size = bytezip / (1024 * 1024);
        let zipsize = buffer.length / (1024 * 1024);
        this.writeMapSync("-----------------------------------\n" + zipObj.zipname + '.zip\t' + size.toFixed(2) + ' MB => ' + zipsize.toFixed(2) +
            ' MB\n----------------------------------\n' + maptext);
    }

    //写入调试map文件
    writeMapSync(maptext) {
        if (this.debug == 1) {
            fs.writeFileSync("assets/map.txt", maptext, {
                'flag': 'a'
            });
        }
    }

    //获取共用文件
    getPackageComnFiles(node) {
        let commonFiles = [];
        for (let i = 0, len = node.list.length; i < len; i++) {
            let _file = node.list[i];
            let fileObj = this.dependencieObj.allfiles[_file];
            if (fileObj) {
                if (fileObj.dependencieNum > 1) {
                    commonFiles.push(_file);
                }
            }
        }
        return commonFiles;
    }

    //递归标记依赖
    packageWalk(node, packageObj) {
        let self = this;
        for (let key in node) {
            let item = node[key];
            if (!item) continue;
            if (item.isDir == 0 && item.extname == 'ls') {
                this.jsonDependencie(item, packageObj);
            }
            if (item.isDir == 0 && item.extname == 'lh') {
                this.jsonDependencie(item, packageObj);
            }
            //文件夹
            if (item.isDir == 1) {
                let zipObj = new ZipObj();
                zipObj.info = item;
                zipObj.zipname = item.zipname;
                this.dependencieObj.zips[item.zipname] = zipObj;
            }
            delete node[key];
            self.packageWalk(node, packageObj);
        }
    }

    //通过ls lh标记依赖
    jsonDependencie(item, packageObj) {
        let LayaParse;
        let dependencieList = {};
        let isImg = false;
        if (item.main == 1) { //打包图片
            isImg = true;
        }
        //处理多文件       
        if (item.list) {
            for (let i = 0, len = item.list.length; i < len; i++) {
                let _path = item.list[i];
                LayaParse = new LayaJsonParse(this.dependencieObj);
                let _dependencieList = LayaParse.getdependencies(_path, isImg);
                Object.assign(dependencieList, _dependencieList);
                //自身ls&lh文件也属于依赖
                dependencieList[_path] = 1;
            }
        } else {
            //找到ls所有依赖文件
            LayaParse = new LayaJsonParse(this.dependencieObj);
            dependencieList = LayaParse.getdependencies(item.path, isImg);
            //自身ls&lh文件也属于依赖
            dependencieList[item.path] = 1;
        }
        //标记依赖
        this.signDependencies(item, dependencieList);
        //zip实例
        let zipObj = new ZipObj();
        zipObj.info = item;
        zipObj.zipname = item.zipname;
        zipObj.dependencies = Object.keys(dependencieList);
        this.dependencieObj.zips[item.zipname] = zipObj;
    }

    //标记依赖
    signDependencies(node, files) {
        if (node && files) {
            for (let _file in files) {
                let fileObj = this.dependencieObj.allfiles[_file];
                if (fileObj) {
                    fileObj.setDependencies(node);
                }
            }
        }
    }

    //获取Conventional目录的文件列表实例
    getConventionaFiles(node) {
        let Layadir = node.dir;
        if (Tool.fsExistsSync(Layadir)) {
            let packageObj = new PackageObj(node);
            this.readDirMapWalk(Layadir, node, packageObj);
            return packageObj;
        }
    }

    //例遍文件夹,返回文件实例
    readDirMapWalk(_path, node, packageObj, isLaya = true) {
        let self = this;
        let files = fs.readdirSync(_path);
        files.forEach(function (item) {
            let tmpPath = path.posix.join(_path, item);
            if (tmpPath.indexOf(".svn") > -1)
                return;
            if (tmpPath.indexOf(".git") > -1)
                return;
            if (tmpPath.indexOf(".laya") > -1)
                return;
            let stats = fs.statSync(tmpPath);
            if (stats.isDirectory()) {
                self.readDirMapWalk(tmpPath, node, packageObj, isLaya);
            } else {
                if (isLaya) {
                    let extname = path.extname(tmpPath).replace(/^\./, "");
                    // if (!['ls', 'lh'].includes(extname)) {
                    if (!['ls', 'lh', 'lmat', 'lm', 'lani'].includes(extname)) {
                        return;
                    }
                }
                tmpPath = tmpPath.replace(/\\/g, '/');
                if (!self.dependencieObj.allfiles[tmpPath]) {
                    //创建文件实例
                    let _fileObj = new FileObj(tmpPath);
                    _fileObj.zipname = node.zipname;
                    _fileObj.layaRoot = node.dir;
                    self.dependencieObj.allfiles[tmpPath] = _fileObj;
                }
                packageObj.list.push(tmpPath);
                // let _fileInfo = self.dependencieObj.allfiles[tmpPath];
                // packageObj.dlist.push(tmpPath);
                // if (!packageObj.map[_fileInfo.dir]) packageObj.map[_fileInfo.dir] = [];
                // packageObj.map[_fileInfo.dir].push(tmpPath);
            }
        });
    }

    //获取打包配置表
    getAssetConf() {
        let _data = fs.readFileSync('.assets.ini', 'utf-8');
        _data = _data.split("\r\n");
        let lsMultifile = 0;
        let asset = {};
        let ignoreMultifile = [];
        for (let i = 0, len = _data.length; i < len; i++) {
            let _path = _data[i].trim(); //去掉空行空格
            //判断是否为注释符
            if (_path.indexOf('#') === 0 || _path.indexOf(';') === 0) {
                continue;
            }
            //强行去除所有空格
            let _cmd = _path.replace(/\s/g, "");
            if (!_cmd) continue;
            //判断压缩总开关
            if ('ZIP=0' == _cmd) {
                this.zip = 0;
                continue;
            } else if ('ZIP=1' == _cmd) {
                this.zip = 1;
                continue;
            }
            if (_cmd.indexOf('debug') > -1) {
                let _debug = _cmd.split('=');
                _debug[1] && (this.debug = _debug[1]);
                continue;
            }
            //laya根目录路径
            if (_cmd.indexOf('laya_root') > -1) {
                let _rootArr = _cmd.split('=');
                if (_rootArr[1]) this.LayaRoot = _rootArr[1];
                continue;
            }
            //多文件开始
            if (_cmd.indexOf('{') > 0) {
                asset = {
                    zipname: "",
                    list: []
                };
                if (_cmd.indexOf('+') == 0) {
                    asset.main = 1;
                    _cmd = _cmd.substring(1);
                }
                asset.zipname = _cmd.substring(0, _cmd.length - 1);
                lsMultifile = 1;
                continue;
            }
            //多文件结束            
            if (_cmd.indexOf('}') === 0) {
                this.setLayaInfo(asset.list[0], asset);
                asset = {};
                lsMultifile = 0;
                continue;
            }
            if (lsMultifile == 1 && asset.zipname) {
                ignoreMultifile.push(_path);
                asset.list.push(_path);
                this.Multifiles.push(_path);
                continue;
            }
            asset = {};
            //ignore忽略文件
            if (_cmd.indexOf('!') === 0) {
                _path = _path.substring(1);
                this.ignoreConfs.push(_path);
                continue;
            }
            //单文件--------
            //取包名和路径
            if (_cmd.indexOf('+') == 0) {
                asset.main = 1;
                _path = _path.substring(1);
            }
            if (_path.indexOf('|') > -1) {
                let _ps = _path.split('|');
                asset.zipname = _ps[0].trim();
                _path = _path.substring(1);
                _path = _ps[1].trim();
            }
            _path = _path.trim();
            this.setLayaInfo(_path, asset);
            //删除map文件
            let _mapfile = "assets/map.txt";
            if (Tool.fsExistsSync(_mapfile)) fs.unlinkSync(_mapfile);
        }
    }

    //写入文件信息
    setLayaInfo(_path, asset = {}) {
        let _layaName = this.getLayaName(_path, true);
        let _infoArr = this.getFileInfo(_path);
        
        console.log(_path,   this.fileConfs, _layaName)
        console.log(_layaName)
        if (_layaName == null) {
            //非laya目录
            !asset.zipname && (asset.zipname = _path.split('/')[0]);
            asset = Object.assign(asset, _infoArr);
            this.fileConfs.other.push(asset);
        } else {
            !asset.zipname && (asset.zipname = _layaName + "_" + _infoArr.slug);
            if (!this.fileConfs.laya[_layaName]) this.fileConfs.laya[_layaName] = [];
            let Layadir = _infoArr.dir;
            if (Layadir) {
                let _pos = _infoArr.dir.indexOf("Conventional");
                if (_pos > -1) {
                    Layadir = Layadir.substring(0, _pos);
                }
                let _root = path.join(Layadir, 'Conventional');
                _root = _root.replace(/\\/g, '/');
                asset.root = _root;
            }
            //获取正则包含的文件
            if (_infoArr.fileName.indexOf('*') > -1) {
                let _files = this.getDirFileReg(_infoArr);
                for (let i = 0, len = _files.length; i < len; i++) {
                    //排除多文件包含
                    if (this.Multifiles.indexOf(_files[i]) == -1) {
                        let _info = this.getFileInfo(_files[i]);
                        _info.packname = _layaName;
                        _info.zipname = _layaName + "_" + _info.slug;
                        this.fileConfs.laya[_layaName].push(_info);
                    }
                }
            } else {
                asset = Object.assign(asset, _infoArr);
                asset.packname = _layaName;
                this.fileConfs.laya[_layaName].push(asset);
            }
        }
    }

    //获取laya目录名称
    getLayaName(_path, cut = false) {
        if (_path.indexOf(this.LayaRoot) == 0) {
            _path = _path.substring(this.LayaRoot.length);
            let _arr = _path.split('/');
            return _arr[0]
        }
        return null;
    }

    //通过path获取文件信息
    getFileInfo(_path) {
        let _arr = _path.split('/');
        let _name = _arr.pop();
        let _infoArr = {};
        _infoArr.path = _path;
        _infoArr.dir = _arr.join('/');
        _infoArr.fileName = _name;
        if (_name == '') { //文件夹
            _infoArr.isDir = 1;
        } else {
            //文件
            _infoArr.isDir = 0;
            _infoArr.extname = path.extname(_name).replace(/^\./, "");
            _infoArr.slug = _name.replace(/\./, "_");
        }
        return _infoArr;
    }

    //查找文件夹下符合的文件
    getDirFileReg(data) {
        let self = this;
        let files = fs.readdirSync(data.dir);
        let reg = data.fileName.replace('*', '(.*?)');
        reg = new RegExp(`^${reg}$`);
        let fileLists = [];
        files.forEach(function (item) {
            let tmpPath = path.posix.join(data.dir, item);
            let stats = fs.statSync(tmpPath);
            if (!stats.isDirectory()) {
                let extname = path.extname(tmpPath).replace(/^\./, "");
                if (extname == data.extname) {
                    if (data.fileName.indexOf('*') > -1 && reg.test(item)) {
                        fileLists.push(tmpPath);
                    } else if (item == data.fileName) {
                        fileLists.push(tmpPath);
                    }
                }
            }
        });
        return fileLists;
    }
}

new AssetBundle().run();