(function () {
    'use strict';

    class GameConfig {
        constructor() { }
        static init() {
            Laya.ClassUtils.regClass;
        }
    }
    GameConfig.width = 750;
    GameConfig.height = 1334;
    GameConfig.scaleMode = "full";
    GameConfig.screenMode = "vertical";
    GameConfig.alignV = "middle";
    GameConfig.alignH = "center";
    GameConfig.startScene = "";
    GameConfig.sceneRoot = "";
    GameConfig.debug = false;
    GameConfig.stat = false;
    GameConfig.physicsDebug = false;
    GameConfig.exportSceneToJson = true;
    GameConfig.init();

    class Singleton {
        static getInst() {
            if (!this.instance) {
                this.instance = new this();
            }
            return this.instance;
        }
    }

    class Fsm {
        constructor() {
            this.currentStateTime = 0;
            this.states = {};
            this.data = {};
        }
        static Create(name, owner, states, debug) {
            var fsm = new Fsm();
            fsm.name = name;
            fsm.owner = owner;
            fsm.debug = debug;
            states.forEach(state => {
                let name = state.constructor.name;
                fsm.states[name] = state;
                state._OnInit(fsm);
            });
            return fsm;
        }
        get Name() {
            return this.name;
        }
        get Owner() {
            return this.owner;
        }
        get CurrentState() {
            return this.currentState;
        }
        get CurrentStateName() {
            return this.currentState.constructor.name;
        }
        get CurrentStateTime() {
            return this.currentStateTime;
        }
        Start(startState) {
            var state = this.GetState(startState);
            if (state == null) {
                throw new Error(`${state.name} is not in Fsm{${this.name}}`);
            }
            this.currentState = state;
            this.currentStateTime = 0;
            state.OnEnter(this);
            if (this.debug) {
                console.log(`fsm<${this.name}> start %c${this.name}`, 'color: green;');
            }
        }
        HasState(stateType) {
            return this.states[stateType.name] != null;
        }
        GetState(stateType) {
            return this.states[stateType.name];
        }
        ChangeState(stateType, ...params) {
            var state = this.GetState(stateType);
            if (state == null) {
                throw new Error(`${state.constructor.name} is not in Fsm{${this.name}}`);
            }
            if (this.currentState != null) {
                this.currentState.OnLeave(this);
            }
            this.currentStateTime = 0;
            this.currentState = state;
            state.OnEnter(this, ...params);
            if (this.debug) {
                console.log(`fsm<${this.name}> enter %c${this.currentState.name}`, 'color: green;');
            }
        }
        ExiteCurrentState() {
            if (this.currentState != null) {
                this.currentState.OnLeave(this);
            }
        }
        GetAllStates() {
            return Object.values(this.states);
        }
        OnUpdate(elapseSeconds) {
            if (this.currentState == null) {
                return;
            }
            this.currentStateTime += elapseSeconds;
            this.currentState.OnUpdate(this, elapseSeconds);
        }
        HasData(name) {
            return this.data[name] != null;
        }
        GetData(name) {
            return this.data[name];
        }
        SetData(name, value) {
            this.data[name] = value;
        }
        RemoveData(name) {
            delete this.data[name];
        }
    }

    class FsmManager extends Singleton {
        constructor() {
            super(...arguments);
            this.fsms = {};
        }
        Update(elapseSeconds) {
            var keys = Object.keys(this.fsms);
            for (var i = 0; i < keys.length; i++) {
                this.fsms[keys[i]].OnUpdate(elapseSeconds);
            }
        }
        CreateFsm(name, owner, states, debug) {
            var key = owner.constructor.name + '_' + name;
            if (this.fsms[key] != null) {
                throw new Error(`Already exist FSM:${key}`);
            }
            var fsm = Fsm.Create(name, owner, states);
            fsm.debug = debug;
            this.fsms[key] = fsm;
            return fsm;
        }
    }

    class LayaUtils {
        static get playingTime() {
            return Laya.timer.currTimer - this._startTime;
        }
        static get frameRate() {
            return 1000 / this.deltaTime;
        }
        static get deltaTimeScale() {
            return Math.min(60, Laya.timer.delta) / 60;
        }
        static get timeScale() {
            return Laya.timer.scale;
        }
        static get deltaTime() {
            if (this._recordFrame != Laya.timer.currFrame) {
                this._deltaTime = Math.min(Laya.timer.delta, 100);
                this._deltaTimeSec = this._deltaTime * 0.001;
                this._recordFrame = Laya.timer.currFrame;
            }
            return this._deltaTime * this.timeScale;
        }
        static get deltaTimeSec() {
            if (this._recordFrame != Laya.timer.currFrame) {
                this._deltaTime = Math.min(Laya.timer.delta, 100);
                this._deltaTimeSec = this._deltaTime * 0.001;
                this._recordFrame = Laya.timer.currFrame;
            }
            return this._deltaTimeSec * this.timeScale;
        }
        static IsZeroVec2(a) {
            return Laya.MathUtils3D.isZero(a.x) && Laya.MathUtils3D.isZero(a.y);
        }
        static addVec2(a, b, out) {
            out.x = a.x + b.x;
            out.y = a.y + b.y;
        }
        static subtractVec2(a, b, out) {
            out.x = a.x - b.x;
            out.y = a.y - b.y;
        }
        static distanceSquaredVec2(a, b) {
            let dx = a.x - b.x;
            let dy = a.y - b.y;
            return dx * dx + dy * dy;
        }
        static distanceVec2(a, b) {
            let dx = a.x - b.x;
            let dy = a.y - b.y;
            return Math.sqrt(dx * dx + dy * dy);
        }
        static getLengthVec2(a) {
            return Math.sqrt(a.x * a.x + a.y * a.y);
        }
        static normalizeVec2(a, out) {
            var len = Math.sqrt(a.x * a.x + a.y * a.y);
            if (len < Laya.MathUtils3D.zeroTolerance) {
                return;
            }
            out.x = a.x / len;
            out.y = a.y / len;
        }
        static AngleVec2(a, b) {
            var square = a.x * a.x + b.y * b.y;
            if (square < Laya.MathUtils3D.zeroTolerance) {
                return 0;
            }
            var dot = (a.x * b.x + a.y * b.y) / Math.sqrt(square);
            return Math.acos(dot) / Laya.MathUtils3D.Deg2Rad;
        }
        static truncateVec2(a, maxLen, out) {
            var square = a.x * a.x + a.y * a.y;
            if (square < Laya.MathUtils3D.zeroTolerance) {
                return;
            }
            if (square <= maxLen * maxLen) {
                out.x = a.x;
                out.y = a.y;
            }
            else {
                var len = Math.sqrt(square);
                var ratio = maxLen / len;
                out.x = a.x * ratio;
                out.y = a.y * ratio;
            }
        }
        static scaleVec2(a, mulLen, out) {
            out.x = a.x * mulLen;
            out.y = a.y * mulLen;
        }
        static setLengthVec2(a, len, out) {
            if (this.IsZeroVec2(a)) {
                out.x = out.y = 0;
                return;
            }
            var _len = Math.sqrt(a.x * a.x + a.y * a.y);
            var ratio = len / _len;
            out.x = a.x * ratio;
            out.y = a.y * ratio;
        }
        static DotVec2(a, b) {
            return a.x * b.x + a.y * b.y;
        }
    }
    LayaUtils.UP = new Laya.Vector3(0, 1, 0);
    LayaUtils.Vector3Zero = new Laya.Vector3(0, 0, 0);
    LayaUtils.Vector2Zero = new Laya.Vector2();
    LayaUtils._startTime = Date.now();
    LayaUtils._recordFrame = 0;
    LayaUtils._deltaTime = 30;
    LayaUtils._deltaTimeSec = 0.016;

    const PI$1 = Math.PI;
    const _d2r$1 = PI$1 / 180.0;
    const _r2d$1 = 180.0 / PI$1;
    class Maths$1 {
        static formula2(value, start, length) {
            return (value + (start % length)) % length;
        }
        static formula(value, mul) {
            let res = mul * value;
            while (value--)
                res += value;
            return res;
        }
        static isEven(value) {
            return value % 2 == 0;
        }
        static toRadian(a) {
            return a * _d2r$1;
        }
        static converGold(value) {
            return "";
        }
        static equallyNum(value, count, randMin, randMax, isPositive = false) {
            let equally = Math.floor(value / count);
            if (isPositive)
                randMax = equally - 1;
            let res = [];
            let lastRand = 0;
            let is = true;
            for (let i = 0; i < count; i++) {
                if (is) {
                    is = false;
                    lastRand = this.minToMax(randMin, randMax);
                    res.push(equally + lastRand);
                }
                else {
                    is = true;
                    res.push(equally - lastRand);
                }
            }
            return res;
        }
        static decimal(value) {
            value = value - (Math.floor(value));
            return value;
        }
        static toDegree(a, is_ormalize = true) {
            if (a == undefined)
                return 0;
            if (is_ormalize)
                a = this.normalizeAngle(a);
            return a * _r2d$1;
        }
        static normalizeAngle(angle) {
            angle = angle % (2 * PI$1);
            if (angle < 0) {
                angle += (2 * PI$1);
            }
            return angle;
        }
        static checkRange(value, value2, range) {
            return Math.abs(value - value2) <= range;
        }
        static isTargetRange(value, target, range) {
            if (value >= target - range)
                return true;
            if (value <= target + range)
                return true;
            return false;
        }
        static zeroEqually(space, index, length) {
            return (index - (length - 1) / 2) * space;
        }
        static equally(count = 1, min, max) {
            count++;
            let equally = [];
            let space = ((max - min) / count);
            for (let i = 1; i < count; i++) {
                equally.push(~~(min + space * i));
            }
            return equally;
        }
        static abs_sub_length(value1, value2) {
            return Math.abs(value1 - value2);
        }
        static add_arrs(values) {
            let count = 0;
            for (let value of values)
                count += value;
            return count;
        }
        static lerp(a, b, r) {
            return a + (b - a) * r;
        }
        static clamp01(value) {
            return value < 0 ? 0 : value < 1 ? value : 1;
        }
        static clampf(value, min, max) {
            if (min > max) {
                var temp = min;
                min = max;
                max = temp;
            }
            return value < min ? min : value < max ? value : max;
        }
        static isClampf(value, min, max) {
            if (min > max) {
                var temp = min;
                min = max;
                max = temp;
            }
            if (value <= max && value >= min)
                return true;
            else
                return false;
        }
        static create_arr_num(max) {
            let values = [];
            for (let i = 0; i < max; i++)
                values.push(i);
            return values;
        }
        static isRandom(value) {
            if (!value)
                return false;
            if (value >= 1)
                return true;
            if (value <= 0)
                return false;
            return value > Math.random();
        }
        static zeroToMax(max) {
            if (max <= 1)
                return 0;
            let random = Math.random();
            return ~~((random == 1 ? 0.5 : random) * max);
        }
        static minToMax(min, max, is_int = true) {
            if (min == max)
                return min;
            if (min < 0) {
                if (is_int)
                    return Math.round(Math.random() * (Math.abs(min) + max)) + min;
                else
                    return Math.random() * (Math.abs(min) + max) + min;
            }
            if (is_int)
                return Math.round(Math.random() * (max - min)) + min;
            else
                return Math.random() * (max - min) + min;
        }
        static converNumber(value) {
            if (typeof value == "string")
                return Number(value);
        }
        static find_num_section(arr, num, out, is_contain_num = false) {
            if (!out)
                out = { i: 0, count: 0 };
            out.i = arr.length;
            out.count = 0;
            for (let i = 0; i < arr.length; i++) {
                out.count += arr[i];
                if (is_contain_num) {
                    if (num <= out.count) {
                        out.i = i;
                        return out;
                    }
                }
                else if (num < out.count) {
                    out.i = i;
                    return out;
                }
            }
            return out;
        }
        static find_num_section_random(arr) {
            let all = this.add_arrs(arr);
            let last = 0;
            let rand = this.zeroToMax(all - 1);
            for (let i = 0; i < arr.length; i++) {
                let value = arr[i];
                if (rand >= last && rand < last + value)
                    return i;
                last += value;
            }
            return arr.length;
        }
    }

    new Laya.Vector3();
    const v3_zeor = new Laya.Vector3();
    new Laya.Vector3(1, 1, 1);
    new Laya.Vector3(0, 1, 0);
    new Laya.Vector3(0, 0, 1);
    new Laya.Vector3(0, 180, 0);
    new Laya.Point();
    new Laya.Quaternion();

    const isValid$1 = function (key) {
        if (key == null || key == undefined || key === false)
            return false;
        else
            return true;
    };
    const minxi$1 = function (value) {
        if (value === null || value === undefined || value === 0 || value === "")
            return value;
        let _getPropertyDescriptor = function (obj, name) {
            while (obj) {
                let pd = Object.getOwnPropertyDescriptor(obj, name);
                if (pd) {
                    return pd;
                }
                obj = Object.getPrototypeOf(obj);
            }
            return null;
        };
        let _copyprop = function (name, source, target) {
            let pd = _getPropertyDescriptor(source, name);
            Object.defineProperty(target, name, pd);
        };
        let _mixin = function (obj, ...sourceObj) {
            obj = obj || {};
            for (let i = 1, length = arguments.length; i < length; i++) {
                let source = arguments[i];
                if (source) {
                    if (typeof source !== "object") {
                        continue;
                    }
                    for (let name in source) {
                        _copyprop(name, source, obj);
                    }
                }
            }
            return obj;
        };
        let target = null;
        if (typeof value === "object") {
            target = _mixin(Array.isArray(value) ? [] : {}, value);
        }
        else {
            target = value;
        }
        return target;
    };

    class Sets$1 {
        constructor(isRepeat = true, ...values) {
            this.sets = [];
            this.isRepeat = true;
            this.isRepeat = isRepeat;
            this.add(...values);
        }
        static delete(arr, value) {
            if (arr.length == 0)
                return false;
            let index = arr.indexOf(value);
            if (index == -1)
                return false;
            arr.splice(index, 1);
            return true;
        }
        static map(arr, cb) {
            let set = [];
            for (let i = 0; i < arr.length; i++) {
                let data = cb(arr[i], i, arr);
                set.push(data);
            }
            return set;
        }
        static has(arr, value, stringCompare = false) {
            if (arr.length == 0)
                return false;
            let index;
            if (stringCompare)
                index = this.map(arr, v => v.toString()).indexOf(value.toString());
            else
                index = arr.indexOf(value);
            return index != -1;
        }
        static add(arr, ...values) {
            for (let i = 0; i < values.length; i++) {
                let value = values[i];
                if (this.has(arr, value))
                    continue;
                arr.push(value);
            }
        }
        static random(value, exclude) {
            if (value.length == 0)
                return null;
            if (exclude) {
                let values = [];
                for (let i = 0, l = value.length; i < l; i++)
                    if (!exclude(value[i], i))
                        values.push(values[i]);
                return this.random(values);
            }
            return value[Maths$1.zeroToMax(value.length)];
        }
        static getAt(value, index) {
            index = Maths$1.clampf(index, 0, value.length - 1);
            return value[index];
        }
        static filterValue(values, cb) {
            let idx = this.filterOf(values, cb);
            if (idx == -1) {
                return null;
            }
            return values[idx];
        }
        static filterOf(values, cb) {
            let idx = -1;
            for (let i = 0; i < values.length; i++) {
                if (values[i] && cb(values[i], i)) {
                    return i;
                }
            }
            return idx;
        }
        static filter(values, cb, out) {
            if (!out)
                out = [];
            for (let i = 0; i < values.length; i++) {
                if (values[i] && cb(values[i], i)) {
                    out.push(values[i]);
                }
            }
            return out;
        }
        static push(values, push) {
            for (let value of push)
                values.push(value);
            return values;
        }
        static sort(values, cb) {
            return values.sort(cb);
        }
        static shuffle(arr, isChangeOriginal = true) {
            if (!isChangeOriginal) {
                let tmp = arr.concat();
                tmp.sort((a, b) => {
                    return Math.random() > .5 ? 1 : -1;
                });
                return tmp;
            }
            else {
                arr.sort((a, b) => {
                    return Math.random() > .5 ? 1 : -1;
                });
                return arr;
            }
        }
        static random_arr(vlaues, count = 1) {
            let res = this.shuffle(vlaues, false), arr = [];
            for (let i = 0; i < count; i++)
                if (isValid$1(res[i]))
                    arr.push(res[i]);
            return arr;
        }
        add(...values) {
            for (let i = 0; i < values.length; i++) {
                let value = values[i];
                if (this.has(value) && this.isRepeat)
                    continue;
                this.sets.push(value);
            }
            return this;
        }
        delete(value) {
            return Sets$1.delete(this.sets, value);
        }
        get size() {
            return this.sets.length;
        }
        indexOf(value) {
            return this.sets.indexOf(value);
        }
        has(value, stringCompare = false) {
            return Sets$1.has(this.sets, value, stringCompare);
        }
        hasOf(index) {
            return isValid$1(this.sets[index]);
        }
        clear() {
            this.sets.length = 0;
        }
        shift(isOrign = true) {
            if (isOrign) {
                return this.sets.shift();
            }
            else {
                return this.sets.concat()[0];
            }
        }
        toString() {
            return this.sets.toString();
        }
        forEach(cb) {
            this.sets.forEach(cb);
            return this;
        }
        map(cb) {
            return Sets$1.map(this.sets, cb);
        }
        findContinueIndexs(cb) {
            let values = this.map(cb);
            let indexs = [];
            for (let i = 0; i < values.length; i++) {
                let continueIndexs = [];
                for (let j = 0; j < values.length - i; j++) {
                    let cur = values[i + j];
                    let last = values[i] + j;
                    if (cur == last)
                        continueIndexs.push(values.indexOf(cur));
                    else
                        break;
                }
                if (continueIndexs.length > 1)
                    indexs.push(continueIndexs);
            }
            let res = [];
            for (let i = 0; i < indexs.length; i++) {
                let self = indexs[i];
                let id = this._findContinueIndexs(res, self);
                if (id == -1)
                    res.push(self);
                else if (res[id].length < self.length)
                    res[id] = self;
            }
            return res;
        }
        _findContinueIndexs(maxs, min) {
            for (let i = 0; i < maxs.length; i++) {
                for (let j = 0; j < maxs[i].length; j++) {
                    if (min.indexOf(maxs[i][j]) != -1) {
                        return i;
                    }
                }
            }
            return -1;
        }
        findEquallyIndexs() {
            let indexs = new Sets$1();
            for (let i = 0; i < this.sets.length; i++) {
                let self = this.sets[i];
                let equally = [];
                for (let j = 0; j < this.sets.length; j++) {
                    let all = this.sets[j];
                    if (self == all) {
                        equally.push(j);
                    }
                }
                if (equally.length > 1) {
                    if (!indexs.has(equally, true)) {
                        indexs.add(equally);
                    }
                }
            }
            return indexs;
        }
        sort(cb) {
            Sets$1.sort(this.sets, cb);
            return this;
        }
        deleteFilterOf(cb) {
            let idx = this.filterOf(cb);
            if (idx == -1) {
                return false;
            }
            this.splice(idx);
            return true;
        }
        filterOf(cb) {
            return Sets$1.filterOf(this.sets, cb);
        }
        filterValue(cb) {
            return Sets$1.filterValue(this.sets, cb);
        }
        filter(cb) {
            let tmp = new Sets$1(this.isRepeat);
            for (let i = 0; i < this.sets.length; i++) {
                let data = this.sets[i];
                if (cb(data, i, this.sets)) {
                    tmp.add(data);
                }
            }
            return tmp;
        }
        deletePops(count = 1, cb) {
            for (let i = 0; i < count; i++) {
                let pop = this.pop();
                pop && cb && cb(pop);
            }
        }
        splice(start, deleteCount = 1) {
            return this.sets.splice(start, deleteCount);
        }
        unShift() {
        }
        pop(remove = true) {
            return Sets$1.pop(this.sets, remove);
        }
        static pop(values, remove = true) {
            if (remove) {
                return values.pop();
            }
            else {
                return values[values.length - 1] || null;
            }
        }
        lastValueOf(index) {
            return this.sets[this.sets.length - index - 1] || null;
        }
        valuesOf(index) {
            return this.sets[index] || null;
        }
        randomOf() {
            if (this.size == 0)
                return null;
            return this.sets[Maths$1.zeroToMax(this.sets.length)];
        }
        addRandomSelf() {
            if (this.size == 0)
                return false;
            let data = this.sets[Maths$1.zeroToMax(this.sets.length)];
            this.add(data);
            return true;
        }
        shuffle(isChangeOriginal = true) {
            return Sets$1.shuffle(this.sets, isChangeOriginal);
        }
        deep() {
            let deep = JSON.parse(JSON.stringify(this.sets));
            return new Sets$1(this.isRepeat, ...deep);
        }
        concat(...value) {
            return new Sets$1(this.isRepeat, ...this.sets.concat(value));
        }
    }

    let value$1 = null;
    let valueType$1 = null;
    let index$1 = 0;
    class LocalCache$1 {
        static get setCatchKeyPrefix() {
            return this._setCatchKeyPrefix;
        }
        static converKey(key) {
            if (key.indexOf(this.setCatchKeyPrefix) == -1)
                return this.setCatchKeyPrefix + key;
            return key;
        }
        static set(key, value, isDelay = true) {
            key = this.converKey(key);
            if (!this.checkCacheValue(value))
                return;
            this.cache[key] = value;
            Sets$1.add(this.updateCache, key);
            if (isDelay)
                Laya.timer.once(this.delayCache, this, this.onSet, null, true);
            else
                this.onSet();
        }
        static get(key, defaultValue = null) {
            key = this.converKey(key);
            value$1 = this.cache[key];
            if (!this.checkCacheValue(value$1)) {
                value$1 = Laya.LocalStorage.getItem(key);
                if (!this.checkCacheValue(value$1)) {
                    value$1 = minxi$1(defaultValue);
                    this.cache[key] = value$1;
                    this.set(key, this.cache[key], false);
                }
                else {
                    index$1 = value$1.indexOf("-");
                    valueType$1 = value$1.substring(0, index$1);
                    value$1 = value$1.substring(index$1 + 1);
                    switch (valueType$1) {
                        case "boolean":
                            this.cache[key] = value$1 == "true";
                            break;
                        case "number":
                            this.cache[key] = Number(value$1);
                            break;
                        case "object":
                            this.cache[key] = JSON.parse(value$1);
                            break;
                        case "string":
                            this.cache[key] = value$1;
                            break;
                        default:
                            value$1 = minxi$1(defaultValue);
                            this.cache[key] = value$1;
                            break;
                    }
                }
            }
            return this.cache[key];
        }
        static checkCacheValue(value) {
            if (value == null || value == "null" || value == undefined)
                return false;
            return true;
        }
        static onSet() {
            for (let key of this.updateCache) {
                value$1 = this.cache[key];
                valueType$1 = typeof value$1;
                switch (valueType$1) {
                    case "boolean":
                        value$1 = value$1 + "";
                        break;
                    case "number":
                        value$1 = value$1 + "";
                        break;
                    case "object":
                        value$1 = JSON.stringify(value$1);
                        break;
                    case "string":
                        break;
                    default:
                        value$1 = null;
                        break;
                }
                if (value$1 != null)
                    Laya.LocalStorage.setItem(key, valueType$1 + "-" + value$1);
            }
            this.updateCache.length = 0;
        }
        static has(key) {
            key = this.converKey(key);
            return !!this.get(key);
        }
        static delete(key) {
            key = this.converKey(key);
            delete this.cache[key];
            Laya.LocalStorage.removeItem(key);
        }
        ;
        static clear() {
            this.cache = {};
            Laya.LocalStorage.clear();
        }
    }
    LocalCache$1.cache = {};
    LocalCache$1.updateCache = [];
    LocalCache$1._setCatchKeyPrefix = "";
    LocalCache$1.delayCache = 200;

    /*! *****************************************************************************
    Copyright (c) Microsoft Corporation.

    Permission to use, copy, modify, and/or distribute this software for any
    purpose with or without fee is hereby granted.

    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
    REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
    AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
    INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
    OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
    PERFORMANCE OF THIS SOFTWARE.
    ***************************************************************************** */

    function __decorate(decorators, target, key, desc) {
        var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
        if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
        else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
        return c > 3 && r && Object.defineProperty(target, key, r), r;
    }

    function __awaiter(thisArg, _arguments, P, generator) {
        function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
        return new (P || (P = Promise))(function (resolve, reject) {
            function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
            function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
            function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
            step((generator = generator.apply(thisArg, _arguments || [])).next());
        });
    }

    var ReportType;
    (function (ReportType) {
        ReportType["LOADING_START"] = "\u52A0\u8F7D\u5F00\u59CB";
        ReportType["LOADING_END"] = "\u52A0\u8F7D\u7ED3\u675F";
        ReportType["GUIDE_COLLECT_TAG"] = "\u65B0\u624B\u5F15\u5BFC\u6536\u96C6\u94ED\u724C";
        ReportType["GUIDE_BUILD_ENEMY"] = "\u65B0\u624B\u5F15\u5BFC\u5EFA\u9020\u5175\u8425";
        ReportType["GUIDE_LEVEL_UP"] = "\u65B0\u624B\u5F15\u5BFC\u5347\u7EA7\u519B\u8854";
        ReportType["GUIDE_BATTLE_START"] = "\u65B0\u624B\u5F15\u5BFC\u51FA\u5175";
        ReportType["WIN_NUKEALERT"] = "\u6838\u5F39\u8BF7\u6C42\u5F39\u7A97";
        ReportType["WIN_NUKEALERT_VIDEO"] = "\u6838\u5F39\u8BF7\u6C42\u5F39\u7A97\u89C6\u9891";
        ReportType["NUKE_YUYUE_WIN"] = "\u6838\u5F39\u9884\u7EA6\u5F39\u7A97";
        ReportType["NUKE_CLICK_YUYUE"] = "\u70B9\u51FB\u6838\u5F39\u9884\u7EA6";
        ReportType["NUKE_CLICK_YUYUE_N"] = "\u70B9\u51FB\u9884\u7EA6\u53D6\u6D88";
        ReportType["NUKE_FALL_FREE_START"] = "\u514D\u8D39\u6838\u5F39\u6389\u843D\u5F00\u59CB";
        ReportType["NUKE_FALL_FREE_END"] = "\u514D\u8D39\u6838\u5F39\u6389\u843D\u5B8C\u6BD5";
        ReportType["NUKE_GET_FREE"] = "\u62FE\u53D6\u514D\u8D39\u6838\u5F39";
        ReportType["NUKE_USE_FREE"] = "\u4F7F\u7528\u514D\u8D39\u6838\u5F39";
        ReportType["WIN_PLANE"] = "\u4E0B\u5C40\u8F70\u70B8\u673A\u5F39\u7A97";
        ReportType["WIN_PLANE_VIDEO"] = "\u4E0B\u5C40\u8F70\u70B8\u673A\u5F39\u7A97\u89C6\u9891";
        ReportType["GET_CAERRIER"] = "\u83B7\u5F97\u6C38\u4E45\u6218\u9A6C";
        ReportType["TRY_CAERRIER_VIDEO"] = "\u89C6\u9891\u8BD5\u7528\u6218\u9A6C";
        ReportType["LEVEL_UP"] = "\u5347\u7EA7\u519B\u8854";
        ReportType["ENTER_GAME"] = "\u8FDB\u5165\u5173\u5361";
        ReportType["GAME_END"] = "\u5173\u5361\u7ED3\u7B97";
        ReportType["COLLECT_TAG"] = "\u6536\u96C6\u94ED\u724C";
        ReportType["WIN_REBORN"] = "\u590D\u6D3B\u5F39\u7A97";
        ReportType["WIN_REBORN_VIDEO"] = "\u590D\u6D3B\u5F39\u7A97\u89C6\u9891";
        ReportType["WIN_TAG"] = "\u989D\u5916\u94ED\u724C\u5F39\u7A97";
        ReportType["WIN_TAG_VIDEO"] = "\u989D\u5916\u94ED\u724C\u5F39\u7A97\u89C6\u9891";
        ReportType["WIN_AIRTAG"] = "\u7A7A\u6295\u53CC\u500D\u94ED\u724C\u5F39\u7A97";
        ReportType["WIN_AIRTAG_VIDEO"] = "\u7A7A\u6295\u53CC\u500D\u94ED\u724C\u5F39\u7A97\u89C6\u9891";
        ReportType["WIN_SHARE"] = "\u89C6\u9891\u5931\u8D25\u5206\u4EAB\u5F39\u7A97";
        ReportType["WIN_SHARE_OK"] = "\u70B9\u51FB\u5206\u4EAB\u5F39\u7A97\u5206\u4EAB";
        ReportType["WIN_SHARE_CLOSE"] = "\u70B9\u51FB\u5206\u4EAB\u5F39\u7A97\u5173\u95ED";
        ReportType["BUILD_ENEMY1"] = "\u5EFA\u9020\u5175\u84251";
        ReportType["BUILD_ENEMY2"] = "\u5EFA\u9020\u5175\u84252";
        ReportType["BUILD_ENEMY3"] = "\u5EFA\u9020\u5175\u84253";
        ReportType["BUILD_ENEMY4"] = "\u5EFA\u9020\u5175\u84254";
        ReportType["BUILD_TANK"] = "\u5EFA\u9020\u5766\u514B\u8425";
        ReportType["BUILD_WEAPON"] = "\u5EFA\u9020\u6B66\u5668\u751F\u4EA7\u8BBE\u5907";
        ReportType["CLICK_HANGMU"] = "\u70B9\u51FB\u822A\u6BCD\u56FE\u6807";
        ReportType["CLICK_DUANZAO"] = "\u70B9\u51FB\u822A\u6BCD\u953B\u9020";
        ReportType["GET_HANGMU"] = "\u83B7\u5F97\u822A\u6BCD";
        ReportType["CLICK_ROBOT_BTN"] = "\u672A\u89E3\u9501\u70B9\u51FB\u673A\u7532";
        ReportType["CLICK_ROBOT_UNLOCK"] = "\u70B9\u51FB\u6309\u94AE\u89E3\u9501\u673A\u7532";
        ReportType["CLICK_ROBOT"] = "\u673A\u7532\u5F39\u7A97";
        ReportType["TRY_ROBOT"] = "\u673A\u7532\u8BD5\u7528";
        ReportType["SHARE_ROBOT"] = "\u673A\u7532\u5206\u4EAB\u6309\u94AE";
        ReportType["SHARE_OPENDATA_ROBOT"] = "\u673A\u7532\u5F00\u653E\u57DF\u5206\u4EAB\u6309\u94AE";
        ReportType["GET_ROBOT"] = "\u83B7\u5F97\u673A\u7532";
    })(ReportType || (ReportType = {}));

    var BattleEvent;
    (function (BattleEvent) {
        BattleEvent["RankLevelUp"] = "RankLevelUp";
        BattleEvent["SpwanPlayer"] = "SpwanPlayer";
        BattleEvent["SpawnInitBuilding"] = "SpawnInitBuilding";
        BattleEvent["SpwanEntity"] = "SpwanEntity";
        BattleEvent["ShowAttackUPBar"] = "ShowAttackUPBar";
        BattleEvent["HideAttackUPBar"] = "HideAttackUPBar";
        BattleEvent["SpwanEntityToView"] = "SpwanEntityNotifyView";
        BattleEvent["SpwanEntityFromView"] = "SpwanEntityFromView";
        BattleEvent["ResetFootSoliderSlot"] = "ResetFootSoliderSlot";
        BattleEvent["UnlockFirstSoilderTent"] = "UnlockFirstSoilderTent";
        BattleEvent["SpwanSoilderTent"] = "SpwanSoilderTent";
        BattleEvent["SpwanSoilderTentToView"] = "SpwanSoilderTentNotifyView";
        BattleEvent["DestroySoilderTentToView"] = "DestroySoilderTentNotifyView";
        BattleEvent["DeleteWeaponFromfactory"] = "GetWeaponFromfactory";
        BattleEvent["SpwanWeaponFactoryTent"] = "SpwanWeaponFactoryTent";
        BattleEvent["SpwanWeaponFactoryToView"] = "SpwanWeaponFactoryToView";
        BattleEvent["DestroyWeaponFactoryToView"] = "DestroyWeaponFactoryToView";
        BattleEvent["FireBomb"] = "FireBomb";
        BattleEvent["FireShip"] = "FireShip";
        BattleEvent["FireHangMu"] = "FireHangMu";
        BattleEvent["KillEnemy"] = "KillEnemy";
        BattleEvent["ChangeWeapon"] = "ChangeWeapon";
        BattleEvent["ChangeSoliderSkin"] = "ChangeSoliderSkin";
        BattleEvent["EnemyLoadEnd"] = "EnemyLoadEnd";
        BattleEvent["EnemyCountInfo"] = "EnemyCountInfo";
        BattleEvent["SpwanMechaBox"] = "SpwanMechaBox";
        BattleEvent["MountMecha"] = "MountJijia";
        BattleEvent["MountMechaEnd"] = "MountMechaEnd";
        BattleEvent["ClickMecha"] = "ClickMecha";
        BattleEvent["UpdateMechaFreeTime"] = "UpdateMechaFreeTime";
        BattleEvent["MechFreeEnd"] = "MechFreeEnd";
        BattleEvent["BattleWin"] = "BattleWin";
        BattleEvent["TakeEnemyFlag"] = "TakeEnemyFlag";
        BattleEvent["GoToNextScene"] = "GoToNextScene";
        BattleEvent["EntityDeath"] = "EntityDeath";
        BattleEvent["CreateDogTag"] = "CreateDogTag";
        BattleEvent["SpwanDogTag"] = "SpwanDogTag";
        BattleEvent["AddTagToPlayer"] = "AddTagToPlayer";
        BattleEvent["DeleteTagFromPlayer"] = "DeleteTagFromPlayer";
        BattleEvent["UserTagChange"] = "UserTagChange";
        BattleEvent["InitUserTag"] = "InitUserTag";
        BattleEvent["PlaneStart"] = "PlaneStart";
        BattleEvent["PlaneEnd"] = "PlaneEnd";
        BattleEvent["FireBombStart"] = "FireBombStart";
        BattleEvent["FireBombEnd"] = "FireBombEnd";
        BattleEvent["HideMainUI"] = "HideMainUI";
        BattleEvent["ShowMainUI"] = "ShowMainUI";
        BattleEvent["CallHelicopter"] = "SpawnHelicopter";
        BattleEvent["CallDropBox"] = "CallDropBox";
        BattleEvent["CallBomber"] = "CallBomber";
        BattleEvent["MountHorse"] = "MountHorse";
        BattleEvent["MountHorseEnd"] = "MountHorseEnd";
        BattleEvent["TriggerFeiJi"] = "TriggerFeiJi";
        BattleEvent["SpwanDropBoxTrigger"] = "SpwanDropBoxTrigger";
        BattleEvent["SpwanDogTagDouble"] = "SpwanDogTagDouble";
        BattleEvent["SpwanFreeNukeTrigger"] = "SpwanFreeNukeTrigger";
        BattleEvent["TriggerFreeNuke"] = "TriggerFreeNuke";
        BattleEvent["UpdatePlaneBullet"] = "UpdatePlaneBullet";
        BattleEvent["GameOver"] = "GameOver";
        BattleEvent["TurnOnBattle"] = "TurnOnBattle";
        BattleEvent["TurnOffBattle"] = "TurnOffBattle";
        BattleEvent["SetGuideStep"] = "SetGuideStep";
        BattleEvent["UpdatePlayerPos"] = "UpdatePlayerPos";
        BattleEvent["StartAimMode"] = "StartAimMode";
        BattleEvent["EndAimMode"] = "EndAimMode";
    })(BattleEvent || (BattleEvent = {}));

    class MsgManager extends Singleton {
        constructor() {
            super();
            this.ed = new Laya.EventDispatcher();
        }
        event(name, ...args) {
            this.ed.event(name, args);
        }
        on(name, caller, listener) {
            this.ed.on(name, caller, listener);
        }
        off(name, caller, listener) {
            this.ed.off(name, caller, listener);
        }
        offAll(name) {
            this.ed.offAll(name);
        }
        offAllCaller(_this) {
            this.ed.offAllCaller(_this);
        }
        HasListener(name) {
            return this.ed.hasListener(name);
        }
    }
    MsgManager.Setting = 'Setting';
    var EventType;
    (function (EventType) {
        EventType["Setting"] = "Setting";
        EventType["BuildNodeNotice"] = "BuildNodeNotice";
        EventType["BuildPrefabNodeNotice"] = "BuildPrefabNodeNotice";
        EventType["BuildNodeCompletedNotice"] = "BuildNodeCompletedNotice";
        EventType["BuildPrefabNodeCompletedNotice"] = "buildPrefabNodeCompletedNotice";
        EventType["TriggerItem"] = "TriggerItem";
        EventType["ZiChanUpdate"] = "ZiChanUpdate";
        EventType["StartGame"] = "StartGame";
        EventType["PauseGame"] = "PauseGame";
        EventType["ContinueGame"] = "ContinueGame";
        EventType["ResetObstcalFlag"] = "ResetObstcalFlag";
        EventType["SetMoveable"] = "SetMoveable";
        EventType["SetDragable"] = "SetDragable";
        EventType["SetDragAngle"] = "SetDragAngle";
        EventType["WxShareFriend"] = "WxShareFriend";
    })(EventType || (EventType = {}));

    class UI_Tips extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Loading", "Tips"));
        }
        onConstruct() {
            this.m_holder = (this.getChildAt(0));
        }
    }
    UI_Tips.URL = "ui://ibgwcr5zu27x3c";

    class UI_TipsItem extends fgui.GLabel {
        static createInstance() {
            return (fgui.UIPackage.createObject("Loading", "TipsItem"));
        }
        onConstruct() {
            this.m_bg = (this.getChildAt(0));
            this.m_fade = this.getTransitionAt(0);
        }
    }
    UI_TipsItem.URL = "ui://ibgwcr5zu27x3e";

    var LoaderAlign;
    (function (LoaderAlign) {
        LoaderAlign["left"] = "left";
        LoaderAlign["center"] = "center";
        LoaderAlign["right"] = "right";
    })(LoaderAlign || (LoaderAlign = {}));
    var LoaderVerticalAlign;
    (function (LoaderVerticalAlign) {
        LoaderVerticalAlign["top"] = "top";
        LoaderVerticalAlign["middle"] = "middle";
        LoaderVerticalAlign["bottom"] = "bottom";
    })(LoaderVerticalAlign || (LoaderVerticalAlign = {}));

    class IUI {
    }
    class BaseUI extends IUI {
        constructor(ui) {
            super();
            this.show = false;
            if (ui) {
                this.wrapCom = ui;
                this.ui = ui;
                ui['_cls'] = this;
                this.OnCreate();
            }
        }
        get FadeInOut() {
            return false;
        }
        get FadeInDuration() {
            return 0.2;
        }
        get FadeOutDuration() {
            return 0.1;
        }
        get FullScreenItems() {
            return null;
        }
        ;
        get Align() {
            return LoaderAlign.center;
        }
        get VerticalAlign() {
            return LoaderVerticalAlign.middle;
        }
        get LoaderFillType() {
            return fgui.LoaderFillType.Scale;
        }
        get UseLoaderWarp() {
            return true;
        }
        IsShow() {
            return this.show;
        }
        Create() {
            if (this.wrapCom != null)
                return;
            this.checkBind();
            if (this.UseLoaderWarp) {
                this.wrapCom = new fgui.GComponent();
                this.wrapCom.makeFullScreen();
                this.wrapCom.addRelation(fgui.GRoot.inst, fgui.RelationType.Size);
                var loader = new fgui.GLoader();
                loader.makeFullScreen();
                loader.displayObject.hitArea = null;
                loader.displayObject.mouseThrough = true;
                loader.url = this.URL;
                loader.fill = this.LoaderFillType;
                loader.align = this.Align;
                loader.verticalAlign = this.VerticalAlign;
                this.ui = loader.component;
                this.wrapCom.addChild(loader);
                loader.addRelation(this.wrapCom, fgui.RelationType.Size);
                this.wrapCom.on(fgui.Events.SIZE_CHANGED, this, this.updateFullItems);
                this.updateFullItems();
            }
            else {
                this.wrapCom = this.ui = fgui.UIPackage.createObjectFromURL(this.URL).asCom;
                this.wrapCom.makeFullScreen();
                this.wrapCom.addRelation(fgui.GRoot.inst, fgui.RelationType.Size);
            }
            this.OnCreate();
        }
        checkBind() {
            if (fgui.UIObjectFactory.extensions[this.URL] == null) {
                console.error(this.ClassName, 'is not bind');
            }
        }
        get ClassName() {
            return this.constructor.name;
        }
        get ComName() {
            return fgui.UIPackage.getItemByURL(this.URL).name;
        }
        updateFullItems() {
            var fullItems = this.FullScreenItems;
            var inverScaleX = 1 / this.ui.scaleX;
            var inverScaleY = 1 / this.ui.scaleY;
            if (fullItems != null) {
                for (var i = 0; i < fullItems.length; i++) {
                    var item = fullItems[i];
                    item.makeFullScreen();
                    item.scaleX = inverScaleX;
                    item.scaleY = inverScaleY;
                    item.x = -this.ui.x * inverScaleX;
                    item.y = -this.ui.y * inverScaleY;
                }
            }
        }
        Show(layer, ...args) {
            this.Create();
            if (this.wrapCom.parent == null) {
                fgui.GRoot.inst.addChild(this.wrapCom);
                this.wrapCom.visible = true;
            }
            if (layer != null) {
                this.layer = layer;
                this.wrapCom.sortingOrder = layer;
            }
            this.OnShow(...args);
            this.show = true;
            if (this.FadeInOut) {
                this.ui.alpha = 0;
                this.fadeTween = fgui.GTween.to(0, 1, this.FadeInDuration).onUpdate((t) => {
                    this.ui.alpha = t.value.x;
                });
            }
            else {
                this.ui.alpha = 1;
            }
        }
        Hide(waitFade, dispose) {
            if (waitFade || this.FadeInOut) {
                this.fadeTween = fgui.GTween.to(1, 0, this.FadeOutDuration).onUpdate((t) => {
                    this.ui.alpha = t.value.x;
                }).onComplete(() => { this._Hide(dispose); }, this);
            }
            else {
                this._Hide(dispose);
            }
        }
        _Hide(dispose) {
            if (this.fadeTween != null) {
                if (this.fadeTween._killed == false) {
                    this.fadeTween.kill(false);
                }
                this.fadeTween = null;
            }
            fgui.GRoot.inst.removeChild(this.wrapCom);
            this.OnHide();
            this.show = false;
        }
        Dispose() {
            this.OnDispose();
            this.wrapCom.dispose();
            this.wrapCom = null;
        }
    }

    var UILayer;
    (function (UILayer) {
        UILayer[UILayer["base"] = 1] = "base";
        UILayer[UILayer["lowMid"] = 2] = "lowMid";
        UILayer[UILayer["mid"] = 3] = "mid";
        UILayer[UILayer["top"] = 4] = "top";
        UILayer[UILayer["adTop"] = 5] = "adTop";
        UILayer[UILayer["overlay"] = 6] = "overlay";
        UILayer[UILayer["insertAd"] = 101] = "insertAd";
        UILayer[UILayer["cover"] = 102] = "cover";
        UILayer[UILayer["login"] = 103] = "login";
    })(UILayer || (UILayer = {}));

    class UIManager {
        constructor() {
            this.menuMap = {};
            this.uiLayarMap = {};
            this.uiStack = {};
        }
        static get inst() {
            if (this._inst == null) {
                this._inst = new UIManager();
            }
            return this._inst;
        }
        RegisterMenu(UI) {
            this.menuMap[UI.name] = UI;
        }
        ShowUI(UI, layer = UILayer.base, ...args) {
            var key = UI.name;
            console.log(key, "===================>UI.name名字");
            console.log("===========>1");
            var ui = this.GetUIInstance(layer, key);
            console.log("===========>2");
            let stack = this.GetStack(layer);
            console.log("===========>3");
            if (stack.length > 0) {
                console.log("===========>4");
                let oldUI = stack.pop();
                if (this.HasUIInstance(layer, oldUI)) {
                    console.log("===========>5");
                    this.GetUIInstance(layer, oldUI).Hide(ui.FadeInOut, false);
                }
            }
            console.log("===========>6");
            ui.Show(layer, ...args);
            console.log("===========>7");
            stack.push(ui.ClassName);
            console.log("===========>8");
            return ui;
        }
        PushUI(UI, layer = UILayer.base, ...args) {
            var key = UI.name;
            var ui = this.GetUIInstance(layer, key);
            var ui = this.uiLayarMap[layer][key];
            let stack = this.GetStack(layer);
            if (stack.length > 0) {
                let lastUI = stack[stack.length - 1];
                if (this.HasUIInstance(layer, lastUI)) {
                    this.GetUIInstance(layer, lastUI).Hide(ui.FadeInOut, false);
                }
            }
            ui.Show(layer, ...args);
            stack.push(ui.ClassName);
            return ui;
        }
        PopUI(layer, disposeUI, ...args) {
            let stack = this.GetStack(layer);
            var popedUI;
            if (stack.length > 0) {
                let popedUIName = stack.pop();
                if (this.HasUIInstance(layer, popedUIName)) {
                    popedUI = this.GetUIInstance(layer, popedUIName);
                    popedUI.Hide(false, disposeUI);
                }
            }
            if (stack.length > 0) {
                let uiName = stack[stack.length - 1];
                let ui = this.GetUIInstance(layer, uiName);
                ui.Show(layer, ...args);
            }
            return popedUI;
        }
        GetUIInstance(layer, name) {
            var uiMap = this.uiLayarMap[layer];
            if (uiMap == null) {
                uiMap = this.uiLayarMap[layer] = {};
            }
            var ui = uiMap[name];
            if (ui == null) {
                if (this.menuMap[name] == null) {
                    console.error(name + ' is not register!');
                    return null;
                }
                ui = uiMap[name] = this.createByClassSign(this.menuMap[name]);
            }
            return ui;
        }
        HasUIInstance(layer, name) {
            if (this.uiLayarMap[layer] == null) {
                return false;
            }
            if (this.uiLayarMap[layer][name] == null) {
                return false;
            }
            return true;
        }
        GetStack(layer) {
            this.uiStack[layer] = this.uiStack[layer] || [];
            return this.uiStack[layer];
        }
        createByClassSign(cls) {
            return new cls();
        }
    }
    function ui_register() {
        return function (_constructor) {
            UIManager.inst.RegisterMenu(_constructor);
        };
    }

    class TipsMgr {
        static get tips() {
            if (this._tips == null) {
                this._tips = UIManager.inst.ShowUI(TipsUI, UILayer.overlay);
            }
            return this._tips;
        }
        static Tips(msg) {
            this.tips.ShowTips(msg);
        }
    }
    let TipsUI = class TipsUI extends BaseUI {
        get URL() {
            return UI_Tips.URL;
        }
        OnCreate() {
        }
        OnShow(msg) {
        }
        OnHide() {
        }
        OnDispose() {
        }
        ShowTips(msg) {
            var l = UI_TipsItem.createInstance();
            l.title = msg;
            l.m_fade.play();
            this.ui.m_holder.addChild(l);
            l.sortingOrder = 1000;
            Laya.timer.once(1000, this, this.onRemoveTips, [l], false);
        }
        onRemoveTips(l) {
            l.dispose();
        }
    };
    TipsUI = __decorate([
        ui_register()
    ], TipsUI);

    class Handler {
        constructor(caller = null, method = null, args = null, once = false) {
            this.once = false;
            this._id = 0;
            this.setTo(caller, method, args, once);
        }
        setTo(caller, method, args, once = false) {
            this._id = Handler._gid++;
            this.caller = caller;
            this.method = method;
            this.args = args;
            this.once = once;
            return this;
        }
        run() {
            if (this.method == null)
                return null;
            var id = this._id;
            var result = this.method.apply(this.caller, this.args);
            this._id === id && this.once && this.recover();
            return result;
        }
        runWith(data) {
            if (this.method == null)
                return null;
            var id = this._id;
            if (data == null)
                var result = this.method.apply(this.caller, this.args);
            else if (!this.args && !data.unshift)
                result = this.method.call(this.caller, data);
            else if (this.args)
                result = this.method.apply(this.caller, this.args.concat(data));
            else
                result = this.method.apply(this.caller, data);
            this._id === id && this.once && this.recover();
            return result;
        }
        clear() {
            this.caller = null;
            this.method = null;
            this.args = null;
            return this;
        }
        recover() {
            if (this._id > 0) {
                this._id = 0;
                Handler._pool.push(this.clear());
            }
        }
        static create(caller, method, args = null, once = true) {
            if (Handler._pool.length)
                return Handler._pool.pop().setTo(caller, method, args, once);
            return new Handler(caller, method, args, once);
        }
    }
    Handler._pool = [];
    Handler._gid = 1;

    class SDKConfig {
    }
    SDKConfig.getLocation = false;
    SDKConfig.useOpenid = false;
    SDKConfig.GAMEID = "";
    SDKConfig.VIVO_VERSION = "";
    SDKConfig.OPPO_VERSION = "";
    SDKConfig.TOUTIAO_GAME_NAME = "";
    SDKConfig.TOUTIAO_VERSION = "";
    SDKConfig.SHARE_CONTENT = [];
    SDKConfig.TOUTIAO_SHARECONTENT = [];
    SDKConfig.TOUTIAO_SHARETEMPLATEID = [];
    SDKConfig.WECHAT_GAME_NAME = "";
    SDKConfig.WECHAT_VERSION = "";
    SDKConfig.WECHAT_SHARECONTENT = [];
    SDKConfig.WECHAT_SHARETEMPLATEID = [];
    SDKConfig.HBS_VERSION = "";
    SDKConfig.GAME233_APPKEY = "";
    SDKConfig.GAME233_POS = "";
    SDKConfig.APP_IOS_APPID = "";
    SDKConfig.MOMOYU_VERSION = "";

    class NavigateItem {
    }

    class ToolSDK {
        static GET_PLT() {
            let plat_type = PlatType$1.WEB;
            let bridge = this.getBridge("demo.JSBridge");
            let plt_str = "浏览器";
            if (bridge != undefined) {
                let platrom = bridge.call("platform");
                if (platrom == "Game233") {
                    plat_type = PlatType$1.GAME233;
                    plt_str = "233乐园";
                }
                else if (platrom == "momoyu") {
                    plat_type = PlatType$1.MOMOYU;
                    plt_str = "摸摸鱼";
                }
            }
            else {
                if (window['kwaigame']) {
                    plat_type = PlatType$1.KUAIGAME;
                    plt_str = "快手";
                }
                else if (window["tt"] && Laya.Browser["onTTMiniGame"]) {
                    plat_type = PlatType$1.TOUTIAO;
                    plt_str = "头条";
                }
                else if (window['uc']) {
                    plat_type = PlatType$1.GAMEUC;
                    plt_str = "uc浏览器";
                }
                else if (window["qg"]) {
                    if (Laya.Browser.onQGMiniGame) {
                        plat_type = PlatType$1.OPPO;
                        plt_str = "OPPO";
                    }
                    else if (Laya.Browser.onVVMiniGame) {
                        plat_type = PlatType$1.VIVO;
                        plt_str = "VIVO";
                    }
                    else
                        plat_type = PlatType$1.WEB;
                }
                else if (window["hbs"] && Laya.Browser["onHWMiniGame"]) {
                    plat_type = PlatType$1.HBS;
                    plt_str = "华为";
                }
                else if (window["wx"]) {
                    if (Laya.Browser.onWeiXin) {
                        plat_type = PlatType$1.WECHAT;
                        plt_str = "微信";
                    }
                    else
                        plat_type = PlatType$1.WEB;
                }
                else
                    plat_type = PlatType$1.WEB;
            }
            plat_type = PlatType$1.HBS;
            plt_str = "华为";
            console.log("自动获取平台:" + plt_str);
            return plat_type;
        }
        static getBridge(pathClass) {
            if (window['conchConfig'] == undefined) {
                return undefined;
            }
            let os = window['conchConfig'].getOS();
            if (os != "Conch-android") {
                return undefined;
            }
            return window['PlatformClass'].createClass(pathClass);
        }
        static randomItem(array) {
            if (array && array.length > 0) {
                return array[this.random(0, array.length - 1)];
            }
            else {
                return null;
            }
        }
        static random(min, max) {
            return Math.round(Math.random() * (max - min) + min);
        }
        static get_cache(key) {
            let data = Laya.LocalStorage.getItem(key);
            if (data == null || data == "null" || data == undefined || data == "" || data == "NaN") {
                return null;
            }
            return data;
        }
        ;
        static set_cache(key, value) {
            Laya.LocalStorage.setItem(key, value);
        }
        ;
        static has(key) {
            return !!this.get_cache(key);
        }
        static delete_cache(key) {
            Laya.LocalStorage.removeItem(key);
        }
        ;
        static clear_cache() {
            Laya.LocalStorage.clear();
        }
        static stopSound() {
            Laya.SoundManager.setMusicVolume(0);
        }
        static replaySound() {
            Laya.SoundManager.setMusicVolume(1);
        }
    }

    var PlayerFlag;
    (function (PlayerFlag) {
        PlayerFlag[PlayerFlag["NO_AD"] = 0] = "NO_AD";
        PlayerFlag[PlayerFlag["NOMAL_HAS_AD"] = 1] = "NOMAL_HAS_AD";
        PlayerFlag[PlayerFlag["CHANNEL0"] = 2] = "CHANNEL0";
    })(PlayerFlag || (PlayerFlag = {}));
    var ErrorCountType;
    (function (ErrorCountType) {
        ErrorCountType[ErrorCountType["VIDEO"] = 0] = "VIDEO";
        ErrorCountType[ErrorCountType["BANNER"] = 1] = "BANNER";
        ErrorCountType[ErrorCountType["NATIVE"] = 2] = "NATIVE";
    })(ErrorCountType || (ErrorCountType = {}));
    class BaseSDK {
        constructor() {
            this.bannerAd = "p90y1vr7zg";
            this.rewardAd = "q6v8u0thm1";
            this.insertAd = "g6jj6q89bv";
            this.nativeAd = "k0b74jp66v";
            this._sdkGameData = {};
            this._version = "1.0.10";
            this.onADViewComplete = null;
            this.configURL = "";
            this.data = {};
            this.isShortCutInstalled = false;
            this._isBannerShowing = false;
            this._navigateList = [];
            this._isInitNavigate = false;
            this.cityData = { "code": 200, "loc": ["中国"], "ip": "", "env": "", "nowtime": "" };
            this.defaultServerConfig = {
                "is_AddDesk": { state: "1" },
                "is_AvdFloat": { state: "1" },
                "is_AvdList": { state: "1" },
                "is_GameNative": { state: "1", interval: 5 },
                "is_MoreGameBtn": { state: "0" },
                "is_PropSceneUI": { state: "1", interval: 5 },
                "is_banner": { state: "1", interval: 5 },
                "ad_all_interval": 10,
            };
            this.playerFlag = PlayerFlag.NOMAL_HAS_AD;
            this.mousedowntime = 0;
            var data = ToolSDK.get_cache(("GameData" + SDKConfig.GAMEID));
            if (data) {
                this._sdkGameData = JSON.parse(data);
            }
            else {
                this._sdkGameData = {};
            }
            this.mousedowntime = Date.now();
            Laya.stage.on(Laya.Event.MOUSE_DOWN, this, this.onStageMouseDown);
            Laya.timer.loop(1000, this, () => {
                if (Date.now() - this.mousedowntime > 20 * 1000) {
                    SDKEventMgr.Int.emit(SDKEventName.OpenAutoFN);
                    this.mousedowntime = Date.now();
                    SDKEventMgr.Int.emit(SDKEventName.LONG_TIME_NO_OPERATION);
                }
            });
            SDKEventMgr.Int.on(SDKEventName.OpenAutoFN, this.OpenAutoFN, this);
        }
        onStageMouseDown() {
            this.mousedowntime = Date.now();
        }
        get sdkGameData() {
            return this._sdkGameData;
        }
        avdErrorLevel() {
            if (this.data.ErrorLevel == undefined) {
                return 0;
            }
            return parseInt(this.data.ErrorLevel);
        }
        errorMaxCount(type) {
            if (type == ErrorCountType.BANNER) {
                if (this.data.ErrorCountB == undefined) {
                    return 0;
                }
                return parseInt(this.data.ErrorCountB);
            }
            else if (type == ErrorCountType.VIDEO) {
                if (this.data.ErrorCountV == undefined) {
                    return 0;
                }
                return parseInt(this.data.ErrorCountV);
            }
            else if (type == ErrorCountType.NATIVE) {
                if (this.data.ErrorCount == undefined) {
                    return 0;
                }
                return parseInt(this.data.ErrorCount);
            }
        }
        checkIsTody(time) {
            var date = new Date(time);
            var dateNow = new Date();
            let bNewDay = false;
            if (date.getFullYear() == dateNow.getFullYear() &&
                date.getMonth() == dateNow.getMonth() &&
                date.getDate() == dateNow.getDate()) {
                bNewDay = true;
            }
            return bNewDay;
        }
        errorCount(type) {
            if (!this._sdkGameData.time)
                this._sdkGameData.time = Date.now();
            if (!this.checkIsTody(this._sdkGameData.time)) {
                this._sdkGameData.curErrorCountB = 0;
                this._sdkGameData.curErrorCountV = 0;
                this._sdkGameData.curErrorCount = 0;
                ToolSDK.set_cache(("GameData" + SDKConfig.GAMEID), JSON.stringify(this._sdkGameData));
            }
            if (type == ErrorCountType.BANNER) {
                if (this._sdkGameData.curErrorCountB == undefined) {
                    return 0;
                }
                return this._sdkGameData.curErrorCountB;
            }
            else if (type == ErrorCountType.VIDEO) {
                if (this._sdkGameData.curErrorCountV == undefined) {
                    return 0;
                }
                return this._sdkGameData.curErrorCountV;
            }
            else if (type == ErrorCountType.NATIVE) {
                if (this._sdkGameData.curErrorCount == undefined) {
                    return 0;
                }
                return this._sdkGameData.curErrorCount;
            }
        }
        updataAvdErrorCount(value, type) {
            this._sdkGameData.time = Date.now();
            if (type == ErrorCountType.BANNER) {
                if (this._sdkGameData.curErrorCountB == value) {
                    return false;
                }
                this._sdkGameData.curErrorCountB = value;
                ToolSDK.set_cache(("GameData" + SDKConfig.GAMEID), JSON.stringify(this._sdkGameData));
                return true;
            }
            else if (type == ErrorCountType.VIDEO) {
                if (this._sdkGameData.curErrorCountV == value) {
                    return false;
                }
                this._sdkGameData.curErrorCountV = value;
                ToolSDK.set_cache(("GameData" + SDKConfig.GAMEID), JSON.stringify(this._sdkGameData));
                return true;
            }
            else if (type == ErrorCountType.NATIVE) {
                if (this._sdkGameData.curErrorCount == value) {
                    return false;
                }
                this._sdkGameData.curErrorCount = value;
                ToolSDK.set_cache(("GameData" + SDKConfig.GAMEID), JSON.stringify(this._sdkGameData));
                return true;
            }
        }
        avdErrorProp() {
            if (this.data.ErrorProba == undefined) {
                return 0;
            }
            return parseFloat(this.data.ErrorProba);
        }
        playGameErrorProp() {
            if (this.data.PlayGameAdProba == undefined) {
                return 0;
            }
            return parseFloat(this.data.PlayGameAdProba);
        }
        enterGameVideoProp() {
            if (this.data.EnterAdProba == undefined) {
                return 0;
            }
            return parseFloat(this.data.EnterAdProba);
        }
        waterFallErrorProp() {
            if (this.data.WaterfallProba == undefined) {
                return 0;
            }
            return parseFloat(this.data.WaterfallProba);
        }
        playGameLevel() {
            if (this.data.PlayGameAd == undefined) {
                return 0;
            }
            return parseInt(this.data.PlayGameAd);
        }
        isCurrentVersion() {
            if (this._version == undefined) {
                return false;
            }
            if (!this.data.version)
                return false;
            return this.data.version == this._version;
        }
        get version() {
            return this._version;
        }
        byteToString(arr) {
            if (typeof arr === 'string') {
                return arr;
            }
            var str = '', _arr = arr;
            for (var i = 0; i < _arr.length; i++) {
                var one = _arr[i].toString(2), v = one.match(/^1+?(?=0)/);
                if (v && one.length == 8) {
                    var bytesLength = v[0].length;
                    var store = _arr[i].toString(2).slice(7 - bytesLength);
                    for (var st = 1; st < bytesLength; st++) {
                        store += _arr[st + i].toString(2).slice(2);
                    }
                    str += String.fromCharCode(parseInt(store, 2));
                    i += bytesLength - 1;
                }
                else {
                    str += String.fromCharCode(_arr[i]);
                }
            }
            return str;
        }
        getBytes(str) {
            var bytes = [];
            for (var i = 0; i < str.length; i++) {
                bytes.push(str.charCodeAt(i));
            }
            return bytes;
        }
        decode(data) {
            let bytes = this.getBytes(data);
            for (var i = 0; i < bytes.length; i++) {
                bytes[i] = bytes[i] ^ 359;
            }
            return this.byteToString(bytes);
        }
        initConfig(cb) {
            let count_call = 2;
            let cb_call = () => {
                count_call--;
                if (count_call <= 0) {
                    for (let key in this.defaultServerConfig) {
                        if (this.data[key] == undefined) {
                            this.data[key] = this.defaultServerConfig[key];
                        }
                    }
                    console.log("nnnnnnnnn");
                    this.onConfigLoad();
                    if (cb)
                        cb();
                }
            };
            Laya.timer.once(5000, this, () => {
                if (count_call > 0) {
                    console.log("ttttttttt");
                    this.onConfigLoad();
                    if (cb)
                        cb();
                }
            });
            if (GameSDK$1.isDecode)
                this.configURL = this.configURL.replace("api.json", "apijson_en");
            this.request({
                url: this.configURL + "?a=" + Math.random(),
                success: (res) => {
                    if (GameSDK$1.isDecode)
                        res = JSON.parse(this.decode(res.data));
                    else
                        res = JSON.parse(res.data);
                    console.log("-------------initConfig--------------", res);
                    for (let key in res) {
                        this.data[key] = res[key];
                    }
                    cb_call();
                }
            });
            this.request({
                url: "https://iploc.jpsdk.com/api/city",
                success: (res) => {
                    console.log("-------------initCity--------------", res.data);
                    this.cityData = JSON.parse(res.data);
                    cb_call();
                }
            });
        }
        hasShareRecorder() {
            return false;
        }
        getNavigateList() {
            if (this._isInitNavigate) {
                return this._navigateList;
            }
            let list = this.data.applist || this.data.navigateAppList;
            let len = list.length;
            let item;
            let info;
            for (let i = 0; i < len; i += 1) {
                info = list[i];
                item = new NavigateItem();
                item.appId = info.appid;
                item.name = info.name;
                item.iconPath = info.iconPath.split(",");
                item.videoId = info.videoId;
                item.weight = info.weight;
                this._navigateList.push(item);
            }
            this.shuffle(this._navigateList);
            this._isInitNavigate = true;
            return this._navigateList;
        }
        shuffle(arr, isChangeOriginal = true) {
            if (!isChangeOriginal) {
                let tmp = arr.concat();
                tmp.sort((a, b) => {
                    return Math.random() > .5 ? 1 : -1;
                });
                return tmp;
            }
            else {
                arr.sort((a, b) => {
                    return Math.random() > .5 ? 1 : -1;
                });
                return arr;
            }
        }
        OpenAutoFN() {
        }
        shortCutInstalled(onComplete) { }
        checkShortcut(onComplete) { }
        getLogicScene(onComplete) { }
        navigateToMiniGame(msg) {
        }
        createBanner() {
        }
        showBanner() {
            this._isBannerShowing = true;
        }
        huaWeiLogin(succedFun, FailFun, caller) {
        }
        hideBanner() {
            this._isBannerShowing = false;
        }
        isBannerShow() {
            return this._isBannerShowing;
        }
        showADVideo(caller, onComplete, reason, args) {
            this.noticeOpenVideo(reason);
            if (onComplete) {
                this.onADViewComplete = Handler.create(caller, onComplete, args);
                this.onADViewComplete.runWith(true);
            }
        }
        showADInsert() {
        }
        noticeOpenVideo(reason) {
            this.sendVideo({ 'point': reason, 'status': 0 });
        }
        noticeCloseVideo(reason) {
            this.sendVideo({ 'point': reason, 'status': 1 });
        }
        startVideoRecord(time, callback) {
        }
        pauseVideoRecord() {
        }
        resumeVideoRecord() {
        }
        stopVideoRecord() {
        }
        shareVideoRecord(callback) {
            console.log("ADBase.shareVideoRecord");
            if (callback)
                callback(true);
        }
        isSupportShareVideo() {
            return true;
        }
        isSupportErrorClick() {
            return false;
        }
        vibrateShort(type) {
        }
        vibrateLong() {
        }
        createNative(value) {
        }
        showToast(str, time) { }
        onConfigLoad() {
        }
        sendVideo(data) {
            SxyGame.sendVideo(data);
        }
        stageOnStart(data) {
            SxyGame.stage.onStart(data);
        }
        statgeOnRunning(data) {
            SxyGame.stage.onRunning(data);
        }
        stageOnEnd(data) {
            SxyGame.stage.onEnd(data);
        }
        sendEvent(name, params = "") {
            SxyGame.sendEvent(name, params);
        }
        request(value) {
            console.log("PlatformVirtual.request", value);
            var xmlhttp;
            if (window.XMLHttpRequest) {
                xmlhttp = new XMLHttpRequest();
            }
            else {
                xmlhttp = new ActiveXObject("Microsoft.XMLHTTP");
            }
            xmlhttp.onreadystatechange = function () {
                if (xmlhttp.readyState == 4 && xmlhttp.status == 200) {
                    if (value.success) {
                        var jsonData = null;
                        if (xmlhttp.responseText && xmlhttp.responseText != "")
                            jsonData = xmlhttp.responseText;
                        value.success({ code: xmlhttp.status, data: jsonData });
                    }
                }
            };
            var url = value.url;
            if (value.data) {
                let params = Object.keys(value.data).map(function (key) {
                    return encodeURIComponent(key) + "=" + encodeURIComponent(value.data[key]);
                }).join("&");
                url += "?" + params;
            }
            xmlhttp.open("GET", url, true);
            xmlhttp.send();
        }
        loadSubpackage(pack, zipFileName, topath, comlete, progress) {
            if (comlete)
                comlete(true);
        }
        shareMsg(caller, onComplete, reason, args) {
        }
        showGameBanner() {
        }
        hideGameBanner() {
        }
        showGamePortal() {
        }
        checkFollowState(callback) {
            if (callback)
                callback(true);
        }
        openUserProfile(callback) {
            if (callback)
                callback();
        }
    }

    const cityData = {
        "code": 200,
        "loc": ["中国", "四川", "成都", "武侯", "电信", "104.04303", "30.64235"],
        "ip": "171.221.145.138",
        "env": "",
        "nowtime": "2020-12-30 16:54:16"
    };
    const apiData = {
        "Offcitys": [],
        "aderrormaxcount": 0,
        "version": "1.0.1",
        "waterfall": "1",
        "textDelay": 0,
        "ErrorLevel": 5,
        "ErrorProba": 0.2,
        "NativeAd": 272469,
        "BannerAd": "272470",
        "RewardAd": "272471",
        "appid": "30312415",
        "NativeAvd": false,
        "IntervalAdTime": 3,
        "navigateAppList": [
            { "id": 5324, "appid": "wxe4821ebb667a1eae", "name": "线条赛车", "packageName": "", "weight": 0, "adPosition": "2", "iconPath": "https:\/\/wxoss.jpsdk.com\/ads\/202008\/80851598842427.png,https:\/\/wxoss.jpsdk.com\/ads\/202010\/63021603263382.png,https:\/\/wxoss.mgsdk.com\/ads\/202103\/35011615801214.png" },
            { "id": 5323, "appid": "wx5e18cf1ffa28bf54", "name": "枪王来了", "packageName": "", "weight": 0, "adPosition": "2", "iconPath": "https:\/\/wxoss.mgsdk.com\/ads\/202103\/32701615889762.png" },
            { "id": 5322, "appid": "wx82a6148ae9de31c8", "name": "热血火线", "packageName": "", "weight": 0, "adPosition": "2", "iconPath": "https:\/\/wxoss.jpsdk.com\/ads\/202012\/11931608284191.png,https:\/\/wxoss.jpsdk.com\/ads\/202012\/22941608284198.png,https:\/\/wxoss.jpsdk.com\/ads\/202012\/94211608284202.png" },
            { "id": 5321, "appid": "wx2dcb2f3168e585d3", "name": "恐龙大猎杀", "packageName": "", "weight": 0, "adPosition": "2", "iconPath": "https:\/\/wxoss.mgsdk.com\/ads\/202103\/48531615800760.png,https:\/\/wxoss.mgsdk.com\/ads\/202103\/15541615800788.png" },
            { "id": 5320, "appid": "wxa1ba781ad9773cd3", "name": "雷电射击", "packageName": "", "weight": 0, "adPosition": "2", "iconPath": "https:\/\/wxoss.mgsdk.com\/ads\/202103\/36211615801259.png,https:\/\/wxoss.mgsdk.com\/ads\/202103\/70421615801292.png" },
            { "id": 5319, "appid": "wx70d2d6827e6e7577", "name": "全民怼雪球", "packageName": "", "weight": 0, "adPosition": "2", "iconPath": "https:\/\/wxoss.mgsdk.com\/ads\/202103\/65801615801406.png,https:\/\/wxoss.mgsdk.com\/ads\/202103\/30081615801415.png,https:\/\/wxoss.mgsdk.com\/ads\/202103\/40901615801424.png" },
            { "id": 5318, "appid": "wxff6356c364b0e69c", "name": "街头轮滑", "packageName": "", "weight": 0, "adPosition": "2", "iconPath": "https:\/\/wxoss.mgsdk.com\/ads\/202103\/19641615801339.png,https:\/\/wxoss.mgsdk.com\/ads\/202103\/88911615801342.png" },
            { "id": 5317, "appid": "wx81128cd788129605", "name": "完美切菜", "packageName": "", "weight": 0, "adPosition": "2", "iconPath": "https:\/\/wxoss.mgsdk.com\/ads\/202103\/88621615804700.png" },
            { "id": 5316, "appid": "wx7c3aa40df71820bc", "name": "摩托车王者", "packageName": "", "weight": 0, "adPosition": "2", "iconPath": "https:\/\/wxoss.mgsdk.com\/ads\/202103\/51701615883743.jpg,https:\/\/wxoss.mgsdk.com\/ads\/202103\/22861615883748.png" },
            { "id": 5315, "appid": "wxe3cad07db86df29c", "name": "疯狂切瓜", "packageName": "", "weight": 0, "adPosition": "2", "iconPath": "https:\/\/wxoss.mgsdk.com\/ads\/202103\/49491615884409.png" },
            { "id": 5314, "appid": "wx9917af5857362fa7", "name": "狙击病毒", "packageName": "", "weight": 0, "adPosition": "2", "iconPath": "https:\/\/wxoss.mgsdk.com\/ads\/202103\/65561615884781.png" }
        ],
        "shareList": [],
        "is_AddDesk": { state: "0" },
        "is_AvdFloat": { state: "0" },
        "is_AvdList": { state: "0" },
        "is_GameNative": { state: "1", interval: 5 },
        "is_MoreGameBtn": { state: "0" },
        "is_PropSceneUI": { state: "1", interval: 5 },
        "is_banner": { state: "1", interval: 5 },
        "ad_all_interval": 10,
    };
    class Web extends BaseSDK {
        constructor() {
            super();
            this.configURL = `https://hallcq.jpsdk.com/static/${SDKConfig.GAMEID}/toutiao/api.json`;
            LocalCache$1._setCatchKeyPrefix = SDKConfig.GAMEID + "-";
        }
        initConfig(cb) {
            this.cityData = cityData;
            this.data = apiData;
            console.log("使用自定义城市配置 方便测试:", this.cityData);
            console.log("使用自定义后台数据配置 方便测试:", this.data);
            this.onConfigLoad();
            if (cb)
                cb();
        }
    }

    class DesMg {
        static encode(str) {
            var bytes = this.utf8Encode(str);
            var key = bytes.slice(0, 2);
            for (var n = 0; n < bytes.length; n++) {
                var a = bytes[n];
                a = a ^ key[n % 2];
                key[n + 2] = a;
            }
            return this.encode_base(key, true);
        }
        static decode(str) {
            var bytes = this.decode_base(str, true);
            var s = bytes.slice(0, 2);
            var t = bytes.slice(2, bytes.length);
            for (var n = 0, r = t.length; n < r; n++) {
                var a = t[n];
                a = a ^ s[n % 2];
                t[n] = a;
            }
            return this.utf8Decode(t);
        }
        static encode_base(str, isDes) {
            var out = "";
            var chr1, chr2, chr3, enc1, enc2, enc3, enc4;
            var i = 0;
            var bytes = isDes ? str : this.utf8Encode(str);
            while (i < bytes.length) {
                chr1 = bytes[i++];
                chr2 = bytes[i++];
                chr3 = bytes[i++];
                enc1 = chr1 >> 2;
                enc2 = ((chr1 & 3) << 4) | (chr2 >> 4);
                enc3 = ((chr2 & 15) << 2) | (chr3 >> 6);
                enc4 = chr3 & 63;
                if (isNaN(chr2)) {
                    enc3 = enc4 = 64;
                }
                else if (isNaN(chr3)) {
                    enc4 = 64;
                }
                out = out + this._keyStr.charAt(enc1) + this._keyStr.charAt(enc2) + this._keyStr.charAt(enc3) + this._keyStr.charAt(enc4);
            }
            return out;
        }
        static decode_base(str, isDes) {
            var chr1, chr2, chr3;
            var enc1, enc2, enc3, enc4;
            var i = 0;
            str = str.replace(/[^A-Za-z0-9\+\/\=]/g, "");
            var bytes = [];
            while (i < str.length) {
                enc1 = this._keyStr.indexOf(str.charAt(i++));
                enc2 = this._keyStr.indexOf(str.charAt(i++));
                enc3 = this._keyStr.indexOf(str.charAt(i++));
                enc4 = this._keyStr.indexOf(str.charAt(i++));
                chr1 = (enc1 << 2) | (enc2 >> 4);
                chr2 = ((enc2 & 15) << 4) | (enc3 >> 2);
                chr3 = ((enc3 & 3) << 6) | enc4;
                bytes.push(chr1);
                if (enc3 != 64) {
                    bytes.push(chr2);
                }
                if (enc4 != 64) {
                    bytes.push(chr3);
                }
            }
            if (isDes === true) {
                return bytes;
            }
            var out = this.utf8Decode(bytes);
            return out;
        }
        static utf8Encode(string) {
            string = string.replace(/\r\n/g, "\n");
            var bytes = [];
            for (var n = 0; n < string.length; n++) {
                var c = string.charCodeAt(n);
                if (c < 128) {
                    bytes.push(c);
                }
                else if (c > 127 && c < 2048) {
                    bytes.push(c(c >> 6) | 192);
                    bytes.push((c & 63) | 128);
                }
                else {
                    bytes.push((c >> 12) | 224);
                    bytes.push(((c >> 6) & 63) | 128);
                    bytes.push((c & 63) | 128);
                }
            }
            return bytes;
        }
        static utf8Decode(bytes) {
            var str = "";
            var i = 0;
            var c, c2, c3 = 0;
            while (i < bytes.length) {
                c = bytes[i];
                if (c < 128) {
                    str += String.fromCharCode(c);
                    i++;
                }
                else if (c > 191 && c < 224) {
                    c2 = bytes[i + 1];
                    str += String.fromCharCode(((c & 31) << 6) | (c2 & 63));
                    i += 2;
                }
                else {
                    c2 = bytes[i + 1];
                    c3 = bytes[i + 2];
                    str += String.fromCharCode(((c & 15) << 12) | ((c2 & 63) << 6) | (c3 & 63));
                    i += 3;
                }
            }
            return str;
        }
    }
    DesMg._keyStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

    const PI = Math.PI;
    const _d2r = PI / 180.0;
    const _r2d = 180.0 / PI;
    class Maths {
        static formula2(value, start, length) {
            return (value + (start % length)) % length;
        }
        static formula(value, mul) {
            let res = mul * value;
            while (value--)
                res += value;
            return res;
        }
        static isEven(value) {
            return value % 2 == 0;
        }
        static toRadian(a) {
            return a * _d2r;
        }
        static converGold(value) {
            return "";
        }
        static equallyNum(value, count, randMin, randMax, isPositive = false) {
            let equally = Math.floor(value / count);
            if (isPositive)
                randMax = equally - 1;
            let res = [];
            let lastRand = 0;
            let is = true;
            for (let i = 0; i < count; i++) {
                if (is) {
                    is = false;
                    lastRand = this.minToMax(randMin, randMax);
                    res.push(equally + lastRand);
                }
                else {
                    is = true;
                    res.push(equally - lastRand);
                }
            }
            return res;
        }
        static decimal(value) {
            value = value - (Math.floor(value));
            return value;
        }
        static toDegree(a, is_ormalize = true) {
            if (a == undefined)
                return 0;
            if (is_ormalize)
                a = this.normalizeAngle(a);
            return a * _r2d;
        }
        static normalizeAngle(angle) {
            angle = angle % (2 * PI);
            if (angle < 0) {
                angle += (2 * PI);
            }
            return angle;
        }
        static checkRange(value, value2, range) {
            return Math.abs(value - value2) <= range;
        }
        static isTargetRange(value, target, range) {
            if (value >= target - range)
                return true;
            if (value <= target + range)
                return true;
            return false;
        }
        static zeroEqually(space, index, length) {
            return (index - (length - 1) / 2) * space;
        }
        static equally(count = 1, min, max) {
            count++;
            let equally = [];
            let space = ((max - min) / count);
            for (let i = 1; i < count; i++) {
                equally.push(~~(min + space * i));
            }
            return equally;
        }
        static abs_sub_length(value1, value2) {
            return Math.abs(value1 - value2);
        }
        static add_arrs(values) {
            let count = 0;
            for (let value of values)
                count += value;
            return count;
        }
        static lerp(a, b, r) {
            return a + (b - a) * r;
        }
        static clamp01(value) {
            return value < 0 ? 0 : value < 1 ? value : 1;
        }
        static clampf(value, min, max) {
            if (min > max) {
                var temp = min;
                min = max;
                max = temp;
            }
            return value < min ? min : value < max ? value : max;
        }
        static isClampf(value, min, max) {
            if (min > max) {
                var temp = min;
                min = max;
                max = temp;
            }
            if (value <= max && value >= min)
                return true;
            else
                return false;
        }
        static create_arr_num(max) {
            let values = [];
            for (let i = 0; i < max; i++)
                values.push(i);
            return values;
        }
        static isRandom(value) {
            if (!value)
                return false;
            if (value >= 1)
                return true;
            if (value <= 0)
                return false;
            return value > Math.random();
        }
        static zeroToMax(max) {
            if (max <= 1)
                return 0;
            let random = Math.random();
            return ~~((random == 1 ? 0.5 : random) * max);
        }
        static minToMax(min, max, is_int = true) {
            if (min == max)
                return min;
            if (min < 0) {
                if (is_int)
                    return Math.round(Math.random() * (Math.abs(min) + max)) + min;
                else
                    return Math.random() * (Math.abs(min) + max) + min;
            }
            if (is_int)
                return Math.round(Math.random() * (max - min)) + min;
            else
                return Math.random() * (max - min) + min;
        }
        static converNumber(value) {
            if (typeof value == "string")
                return Number(value);
        }
        static find_num_section(arr, num, out, is_contain_num = false) {
            if (!out)
                out = { i: 0, count: 0 };
            out.i = arr.length;
            out.count = 0;
            for (let i = 0; i < arr.length; i++) {
                out.count += arr[i];
                if (is_contain_num) {
                    if (num <= out.count) {
                        out.i = i;
                        return out;
                    }
                }
                else if (num < out.count) {
                    out.i = i;
                    return out;
                }
            }
            return out;
        }
        static find_num_section_random(arr) {
            let all = this.add_arrs(arr);
            let last = 0;
            let rand = this.zeroToMax(all - 1);
            for (let i = 0; i < arr.length; i++) {
                let value = arr[i];
                if (rand >= last && rand < last + value)
                    return i;
                last += value;
            }
            return arr.length;
        }
    }

    const isValid = function (key) {
        if (key == null || key == undefined || key === false)
            return false;
        else
            return true;
    };
    const minxi = function (value) {
        if (value === null || value === undefined || value === 0 || value === "")
            return value;
        let _getPropertyDescriptor = function (obj, name) {
            while (obj) {
                let pd = Object.getOwnPropertyDescriptor(obj, name);
                if (pd) {
                    return pd;
                }
                obj = Object.getPrototypeOf(obj);
            }
            return null;
        };
        let _copyprop = function (name, source, target) {
            let pd = _getPropertyDescriptor(source, name);
            Object.defineProperty(target, name, pd);
        };
        let _mixin = function (obj, ...sourceObj) {
            obj = obj || {};
            for (let i = 1, length = arguments.length; i < length; i++) {
                let source = arguments[i];
                if (source) {
                    if (typeof source !== "object") {
                        continue;
                    }
                    for (let name in source) {
                        _copyprop(name, source, obj);
                    }
                }
            }
            return obj;
        };
        let target = null;
        if (typeof value === "object") {
            target = _mixin(Array.isArray(value) ? [] : {}, value);
        }
        else {
            target = value;
        }
        return target;
    };

    class Sets {
        constructor(isRepeat = true, ...values) {
            this.sets = [];
            this.isRepeat = true;
            this.isRepeat = isRepeat;
            this.add(...values);
        }
        static delete(arr, value) {
            if (arr.length == 0)
                return false;
            let index = arr.indexOf(value);
            if (index == -1)
                return false;
            arr.splice(index, 1);
            return true;
        }
        static map(arr, cb) {
            let set = [];
            for (let i = 0; i < arr.length; i++) {
                let data = cb(arr[i], i, arr);
                set.push(data);
            }
            return set;
        }
        static has(arr, value, stringCompare = false) {
            if (arr.length == 0)
                return false;
            let index;
            if (stringCompare)
                index = this.map(arr, v => v.toString()).indexOf(value.toString());
            else
                index = arr.indexOf(value);
            return index != -1;
        }
        static add(arr, ...values) {
            for (let i = 0; i < values.length; i++) {
                let value = values[i];
                if (this.has(arr, value))
                    continue;
                arr.push(value);
            }
        }
        static random(value, exclude) {
            if (value.length == 0)
                return null;
            if (exclude) {
                let values = [];
                for (let i = 0, l = value.length; i < l; i++)
                    if (!exclude(value[i], i))
                        values.push(values[i]);
                return this.random(values);
            }
            return value[Maths.zeroToMax(value.length)];
        }
        static getAt(value, index) {
            index = Maths.clampf(index, 0, value.length - 1);
            return value[index];
        }
        static filterValue(values, cb) {
            let idx = this.filterOf(values, cb);
            if (idx == -1) {
                return null;
            }
            return values[idx];
        }
        static filterOf(values, cb) {
            let idx = -1;
            for (let i = 0; i < values.length; i++) {
                if (values[i] && cb(values[i], i)) {
                    return i;
                }
            }
            return idx;
        }
        static filter(values, cb, out) {
            if (!out)
                out = [];
            for (let i = 0; i < values.length; i++) {
                if (values[i] && cb(values[i], i)) {
                    out.push(values[i]);
                }
            }
            return out;
        }
        static push(values, push) {
            for (let value of push)
                values.push(value);
            return values;
        }
        static sort(values, cb) {
            return values.sort(cb);
        }
        static shuffle(arr, isChangeOriginal = true) {
            if (!isChangeOriginal) {
                let tmp = arr.concat();
                tmp.sort((a, b) => {
                    return Math.random() > .5 ? 1 : -1;
                });
                return tmp;
            }
            else {
                arr.sort((a, b) => {
                    return Math.random() > .5 ? 1 : -1;
                });
                return arr;
            }
        }
        static random_arr(vlaues, count = 1) {
            let res = this.shuffle(vlaues, false), arr = [];
            for (let i = 0; i < count; i++)
                if (isValid(res[i]))
                    arr.push(res[i]);
            return arr;
        }
        add(...values) {
            for (let i = 0; i < values.length; i++) {
                let value = values[i];
                if (this.has(value) && this.isRepeat)
                    continue;
                this.sets.push(value);
            }
            return this;
        }
        delete(value) {
            return Sets.delete(this.sets, value);
        }
        get size() {
            return this.sets.length;
        }
        indexOf(value) {
            return this.sets.indexOf(value);
        }
        has(value, stringCompare = false) {
            return Sets.has(this.sets, value, stringCompare);
        }
        hasOf(index) {
            return isValid(this.sets[index]);
        }
        clear() {
            this.sets.length = 0;
        }
        shift(isOrign = true) {
            if (isOrign) {
                return this.sets.shift();
            }
            else {
                return this.sets.concat()[0];
            }
        }
        toString() {
            return this.sets.toString();
        }
        forEach(cb) {
            this.sets.forEach(cb);
            return this;
        }
        map(cb) {
            return Sets.map(this.sets, cb);
        }
        findContinueIndexs(cb) {
            let values = this.map(cb);
            let indexs = [];
            for (let i = 0; i < values.length; i++) {
                let continueIndexs = [];
                for (let j = 0; j < values.length - i; j++) {
                    let cur = values[i + j];
                    let last = values[i] + j;
                    if (cur == last)
                        continueIndexs.push(values.indexOf(cur));
                    else
                        break;
                }
                if (continueIndexs.length > 1)
                    indexs.push(continueIndexs);
            }
            let res = [];
            for (let i = 0; i < indexs.length; i++) {
                let self = indexs[i];
                let id = this._findContinueIndexs(res, self);
                if (id == -1)
                    res.push(self);
                else if (res[id].length < self.length)
                    res[id] = self;
            }
            return res;
        }
        _findContinueIndexs(maxs, min) {
            for (let i = 0; i < maxs.length; i++) {
                for (let j = 0; j < maxs[i].length; j++) {
                    if (min.indexOf(maxs[i][j]) != -1) {
                        return i;
                    }
                }
            }
            return -1;
        }
        findEquallyIndexs() {
            let indexs = new Sets();
            for (let i = 0; i < this.sets.length; i++) {
                let self = this.sets[i];
                let equally = [];
                for (let j = 0; j < this.sets.length; j++) {
                    let all = this.sets[j];
                    if (self == all) {
                        equally.push(j);
                    }
                }
                if (equally.length > 1) {
                    if (!indexs.has(equally, true)) {
                        indexs.add(equally);
                    }
                }
            }
            return indexs;
        }
        sort(cb) {
            Sets.sort(this.sets, cb);
            return this;
        }
        deleteFilterOf(cb) {
            let idx = this.filterOf(cb);
            if (idx == -1) {
                return false;
            }
            this.splice(idx);
            return true;
        }
        filterOf(cb) {
            return Sets.filterOf(this.sets, cb);
        }
        filterValue(cb) {
            return Sets.filterValue(this.sets, cb);
        }
        filter(cb) {
            let tmp = new Sets(this.isRepeat);
            for (let i = 0; i < this.sets.length; i++) {
                let data = this.sets[i];
                if (cb(data, i, this.sets)) {
                    tmp.add(data);
                }
            }
            return tmp;
        }
        deletePops(count = 1, cb) {
            for (let i = 0; i < count; i++) {
                let pop = this.pop();
                pop && cb && cb(pop);
            }
        }
        splice(start, deleteCount = 1) {
            return this.sets.splice(start, deleteCount);
        }
        unShift() {
        }
        pop(remove = true) {
            return Sets.pop(this.sets, remove);
        }
        static pop(values, remove = true) {
            if (remove) {
                return values.pop();
            }
            else {
                return values[values.length - 1] || null;
            }
        }
        lastValueOf(index) {
            return this.sets[this.sets.length - index - 1] || null;
        }
        valuesOf(index) {
            return this.sets[index] || null;
        }
        randomOf() {
            if (this.size == 0)
                return null;
            return this.sets[Maths.zeroToMax(this.sets.length)];
        }
        addRandomSelf() {
            if (this.size == 0)
                return false;
            let data = this.sets[Maths.zeroToMax(this.sets.length)];
            this.add(data);
            return true;
        }
        shuffle(isChangeOriginal = true) {
            return Sets.shuffle(this.sets, isChangeOriginal);
        }
        deep() {
            let deep = JSON.parse(JSON.stringify(this.sets));
            return new Sets(this.isRepeat, ...deep);
        }
        concat(...value) {
            return new Sets(this.isRepeat, ...this.sets.concat(value));
        }
    }

    let value = null;
    let valueType = null;
    let index = 0;
    class LocalCache {
        static get setCatchKeyPrefix() {
            return this._setCatchKeyPrefix;
        }
        static converKey(key) {
            if (key.indexOf(this.setCatchKeyPrefix) == -1)
                return this.setCatchKeyPrefix + key;
            return key;
        }
        static set(key, value, isDelay = true) {
            key = this.converKey(key);
            if (!this.checkCacheValue(value))
                return;
            this.cache[key] = value;
            Sets.add(this.updateCache, key);
            if (isDelay)
                Laya.timer.once(this.delayCache, this, this.onSet, null, true);
            else
                this.onSet();
        }
        static get(key, defaultValue = null) {
            key = this.converKey(key);
            value = this.cache[key];
            if (!this.checkCacheValue(value)) {
                value = Laya.LocalStorage.getItem(key);
                if (!this.checkCacheValue(value)) {
                    value = minxi(defaultValue);
                    this.cache[key] = value;
                    this.set(key, this.cache[key], false);
                }
                else {
                    index = value.indexOf("-");
                    valueType = value.substring(0, index);
                    value = value.substring(index + 1);
                    switch (valueType) {
                        case "boolean":
                            this.cache[key] = value == "true";
                            break;
                        case "number":
                            this.cache[key] = Number(value);
                            break;
                        case "object":
                            this.cache[key] = JSON.parse(value);
                            break;
                        case "string":
                            this.cache[key] = value;
                            break;
                        default:
                            value = minxi(defaultValue);
                            this.cache[key] = value;
                            break;
                    }
                }
            }
            return this.cache[key];
        }
        static checkCacheValue(value) {
            if (value == null || value == "null" || value == undefined)
                return false;
            return true;
        }
        static onSet() {
            for (let key of this.updateCache) {
                value = this.cache[key];
                valueType = typeof value;
                switch (valueType) {
                    case "boolean":
                        value = value + "";
                        break;
                    case "number":
                        value = value + "";
                        break;
                    case "object":
                        value = JSON.stringify(value);
                        break;
                    case "string":
                        break;
                    default:
                        value = null;
                        break;
                }
                if (value != null)
                    Laya.LocalStorage.setItem(key, valueType + "-" + value);
            }
            this.updateCache.length = 0;
        }
        static has(key) {
            key = this.converKey(key);
            return !!this.get(key);
        }
        static delete(key) {
            key = this.converKey(key);
            delete this.cache[key];
            Laya.LocalStorage.removeItem(key);
        }
        ;
        static clear() {
            this.cache = {};
            Laya.LocalStorage.clear();
        }
    }
    LocalCache.cache = {};
    LocalCache.updateCache = [];
    LocalCache._setCatchKeyPrefix = "";
    LocalCache.delayCache = 200;

    class Times {
        constructor(time) {
            this.time = 0;
            if (time) {
                this.time = time;
            }
            else {
                this.time = Times.now;
            }
        }
        static get now() {
            return new Date().getTime();
        }
        static getDate() {
            return new Date().getDate();
        }
        static console(key) {
            this._console[key] = Times.now;
        }
        static consoleEnd(key, ...args) {
            console.log("执行耗时 ", key, (Times.now - (this._console[key] || Times.now)) / 1000, ...args);
        }
        static get Date() {
            return new Date();
        }
        static day(time) {
            return Math.floor(this.hour(time) / 24);
        }
        static hour(time) {
            return Math.floor(this.minute(time) / 60);
        }
        static minute(time) {
            return Math.floor(this.second(time) / 60);
        }
        static second(time) {
            if (time == null || time == undefined)
                time = this.now;
            return Math.floor(time / 1000);
        }
        static numHour(time) {
            let res;
            res = this.converMinTen(this.second(time) % 60) + "";
            if (this.second() >= 60) {
                res = this.converMinTen(this.minute(time) % 60) + ":" + res;
            }
            else {
                res = "00:" + res;
            }
            if (this.minute() >= 60) {
                res = this.converMinTen(this.hour(time)) + ":" + res;
            }
            else {
                res = "00:" + res;
            }
            return res;
        }
        static numMinute(time) {
            let res;
            res = this.converMinTen(this.second(time) % 60) + "";
            if (this.second() >= 60) {
                res = this.converMinTen(this.minute(time)) + ":" + res;
            }
            else {
                res = "00:" + res;
            }
            return res;
        }
        static numSecond(time) {
            return this.converMinTen(this.second(time));
        }
        static chinaHour(time) {
            let res;
            res = this.converMinTen(this.second(time) % 60) + "秒";
            if (this.second() >= 60) {
                res = this.converMinTen(this.minute(time) % 60) + "分" + res;
            }
            if (this.minute() >= 60) {
                res = this.converMinTen(this.hour(time)) + "时" + res;
            }
            return res;
        }
        static chinaMinute(time) {
            let res;
            res = this.converMinTen(this.second(time) % 60) + "秒";
            if (this.second() >= 60) {
                res = this.converMinTen(this.minute(time)) + "分" + res;
            }
            return res;
        }
        static chinaSecond(time) {
            return this.converMinTen(this.second(time)) + "秒";
        }
        static converMinTen(num) {
            return num < 10 ? ("0" + num) : num.toString();
        }
        add(time) {
            this.time += time;
            return this;
        }
        sub(time) {
            this.time -= time;
            return this;
        }
        get now() { return this.time; }
        ;
        get day() {
            return Math.floor(this.hour / 24);
        }
        get hour() {
            return Math.floor(this.minute / 60);
        }
        get minute() {
            return Math.floor(this.second / 60);
        }
        ;
        get second() {
            return Math.floor(this.time / 1000);
        }
        get numHour() {
            let res;
            res = this.converMinTen(this.second % 60) + "";
            if (this.second >= 60) {
                res = this.converMinTen(this.minute % 60) + ":" + res;
            }
            else {
                res = "00:" + res;
            }
            if (this.minute >= 60) {
                res = this.converMinTen(this.hour) + ":" + res;
            }
            else {
                res = "00:" + res;
            }
            return res;
        }
        get numMinute() {
            let res;
            res = this.converMinTen(this.second % 60) + "";
            if (this.second >= 60) {
                res = this.converMinTen(this.minute) + ":" + res;
            }
            else {
                res = "00:" + res;
            }
            return res;
        }
        get numSecond() {
            return this.converMinTen(this.second);
        }
        get chinaHour() {
            let res;
            res = this.converMinTen(this.second % 60) + "秒";
            if (this.second >= 60) {
                res = this.converMinTen(this.minute % 60) + "分" + res;
            }
            if (this.minute >= 60) {
                res = this.converMinTen(this.hour) + "时" + res;
            }
            return res;
        }
        get chinaMinute() {
            let res;
            res = this.converMinTen(this.second % 60) + "秒";
            if (this.second >= 60) {
                res = this.converMinTen(this.minute) + "分" + res;
            }
            return res;
        }
        get chinaSecond() {
            return this.converMinTen(this.second) + "秒";
        }
        converMinTen(num) {
            return num < 10 ? ("0" + num) : num.toString();
        }
        static isSameDay(t) {
            return new Date(t).toLocaleDateString() === new Date().toLocaleDateString();
        }
        static compare_time(time) {
            var today_end = this.get_endtime();
            var time_old = time;
            if (today_end == time_old) {
                return 1;
            }
            else {
                return 0;
            }
        }
        static get_endtime() {
            var time_end = new Date(new Date(new Date().toLocaleDateString()).getTime() + 24 * 60 * 60 * 1000 - 1);
            var time_format = this.format_date(time_end);
            return time_format;
        }
        static format_date(now) {
            var year = now.getFullYear();
            var month = now.getMonth() + 1;
            var day = now.getDate();
            var hh = now.getHours();
            var mm = now.getMinutes();
            var ss = now.getSeconds();
            var clock = year + "";
            if (month < 10) {
                clock += "0";
            }
            clock += month + "";
            if (day < 10) {
                clock += "0";
            }
            clock += day + "";
            if (hh < 10) {
                clock += "0";
            }
            clock += hh + "";
            if (mm < 10) {
                clock += '0';
            }
            clock += mm;
            if (ss < 10) {
                clock += '0';
            }
            clock += ss;
            return clock;
        }
    }
    Times._console = {};

    class UI_ui_videofail extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "ui_videofail"));
        }
        onConstruct() {
            this.m_img_bg = (this.getChildAt(0));
            this.m_btn_fengxiang = (this.getChildAt(1));
            this.m_img_fenxang = (this.getChildAt(2));
            this.m_btn_quxioa = (this.getChildAt(3));
        }
    }
    UI_ui_videofail.URL = "ui://s88k2hl2gxcy48";

    class UI_HangMuNengLiang extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "HangMuNengLiang"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m_bg = (this.getChildAt(0));
            this.m_comp_fengxiang = (this.getChildAt(1));
            this.m_btn_fanhui = (this.getChildAt(2));
        }
    }
    UI_HangMuNengLiang.URL = "ui://s88k2hl2akvi5c";

    class UI_TouPiao extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "TouPiao"));
        }
        onConstruct() {
            this.m_bg = (this.getChildAt(0));
            this.m_view = (this.getChildAt(1));
            this.m_btn_fanhui = (this.getChildAt(2));
        }
    }
    UI_TouPiao.URL = "ui://s88k2hl2eper2u";

    class UI_Comp_hangmv2 extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "Comp_hangmv2"));
        }
        onConstruct() {
            this.m_loader = (this.getChildAt(0));
            this.m_zhezhao = (this.getChildAt(1));
            this.m_img_diban = (this.getChildAt(2));
            this.m_img_wenzi = (this.getChildAt(3));
            this.m_img_tishi = (this.getChildAt(4));
            this.m_img_shengyu = (this.getChildAt(5));
            this.m_img_nenglia = (this.getChildAt(6));
            this.m_power = (this.getChildAt(7));
            this.m_btn_duanzao = (this.getChildAt(8));
            this.m_progressBar = (this.getChildAt(9));
            this.m_progressV = (this.getChildAt(10));
            this.m_btn_fanhui = (this.getChildAt(11));
            this.m_add1 = (this.getChildAt(12));
            this.m_addOne = this.getTransitionAt(0);
        }
    }
    UI_Comp_hangmv2.URL = "ui://s88k2hl2gddx55";

    class UI_ui_jm extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "ui_jm"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m_c2 = this.getControllerAt(1);
            this.m_hedan_mask = (this.getChildAt(0));
            this.m_hongzha3D = (this.getChildAt(1));
            this.m_hedan = (this.getChildAt(2));
            this.m_img_bg = (this.getChildAt(3));
            this.m_img_buchong = (this.getChildAt(4));
            this.m_img_sidai = (this.getChildAt(5));
            this.m_img_mrjl = (this.getChildAt(6));
            this.m_img_wz1 = (this.getChildAt(7));
            this.m_btn_rtfq = (this.getChildAt(8));
            this.m_btn_yueyue = (this.getChildAt(9));
            this.m_icon = (this.getChildAt(10));
            this.m_img_tishi = (this.getChildAt(11));
            this.m_img_yycg = (this.getChildAt(13));
            this.m_img_wz3 = (this.getChildAt(14));
            this.m_btn_qifei = (this.getChildAt(15));
            this.m_btn_fashe = (this.getChildAt(16));
            this.m_img_diban2 = (this.getChildAt(17));
            this.m_img_fujiang = (this.getChildAt(18));
            this.m_btn_fh = (this.getChildAt(19));
            this.m_btn_2bei = (this.getChildAt(20));
            this.m_img_xisheng = (this.getChildAt(21));
            this.m_btn_close1 = (this.getChildAt(23));
            this.m_btn_ljhd = (this.getChildAt(24));
            this.m_btn_fanhui = (this.getChildAt(25));
            this.m_btn_quxioa = (this.getChildAt(26));
            this.m_img_shuangbei = (this.getChildAt(27));
            this.m_img_lingqujl = (this.getChildAt(28));
            this.m_btn_xhuanze = (this.getChildAt(29));
            this.m_btn_close3 = (this.getChildAt(31));
        }
    }
    UI_ui_jm.URL = "ui://s88k2hl2h6t30";

    class UI_btn_fh extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "btn_fh"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(1);
            this.m__mask = (this.getChildAt(1));
            this.m_t0 = this.getTransitionAt(0);
        }
    }
    UI_btn_fh.URL = "ui://s88k2hl2h6t310";

    class UI_btn_ljhd extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "btn_ljhd"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m_t0 = this.getTransitionAt(0);
        }
    }
    UI_btn_ljhd.URL = "ui://s88k2hl2h6t313";

    class UI_btn_yuyue extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "btn_yuyue"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m_label_time = (this.getChildAt(4));
            this.m_t0 = this.getTransitionAt(0);
        }
    }
    UI_btn_yuyue.URL = "ui://s88k2hl2h6t3u";

    class UI_btn_ljfs extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "btn_ljfs"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m__mask = (this.getChildAt(0));
            this.m_t0 = this.getTransitionAt(0);
        }
    }
    UI_btn_ljfs.URL = "ui://s88k2hl2h6t3v";

    class UI_comp_yinpai extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "comp_yinpai"));
        }
        onConstruct() {
            this.m_type = this.getControllerAt(0);
            this.m_img_diban2 = (this.getChildAt(0));
        }
    }
    UI_comp_yinpai.URL = "ui://s88k2hl2h6t3w";

    class UI_btn_2bei extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "btn_2bei"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m__maks = (this.getChildAt(1));
            this.m_t0 = this.getTransitionAt(0);
        }
    }
    UI_btn_2bei.URL = "ui://s88k2hl2h6t3x";

    class UI_ui_invate extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "ui_invate"));
        }
        onConstruct() {
            this.m_cpt_page = (this.getChildAt(0));
            this.m_btn_fanhui = (this.getChildAt(1));
        }
    }
    UI_ui_invate.URL = "ui://s88k2hl2i8dz4z";

    class UI_Comp_zaiju extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "Comp_zaiju"));
        }
        onConstruct() {
            this.m_btn_zaiju = (this.getChildAt(2));
            this.m_btn_zaij = (this.getChildAt(3));
            this.m_btn_shiypong = (this.getChildAt(4));
            this.m_btn_shiypong_2 = (this.getChildAt(5));
            this.m_img_wenzi = (this.getChildAt(6));
        }
    }
    UI_Comp_zaiju.URL = "ui://s88k2hl2l8nw3o";

    class UI_ZaiJu extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "ZaiJu"));
        }
        onConstruct() {
            this.m_bg = (this.getChildAt(0));
            this.m_view = (this.getChildAt(1));
            this.m_btn_fanhui = (this.getChildAt(2));
        }
    }
    UI_ZaiJu.URL = "ui://s88k2hl2omf03y";

    class UI_Comp_toupiao extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "Comp_toupiao"));
        }
        onConstruct() {
            this.m_haizhan1 = (this.getChildAt(2));
            this.m_hanzhan2 = (this.getChildAt(3));
            this.m_kongzhan1 = (this.getChildAt(4));
            this.m_kongzhan2 = (this.getChildAt(5));
            this.m_duizhan1 = (this.getChildAt(6));
            this.m_duizhan2 = (this.getChildAt(7));
        }
    }
    UI_Comp_toupiao.URL = "ui://s88k2hl2rmtm2k";

    class UI_btn_xunaze extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "btn_xunaze"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
        }
    }
    UI_btn_xunaze.URL = "ui://s88k2hl2t91k5l";

    class UI_jijia extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "jijia"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m_comp_fengxiang = (this.getChildAt(0));
            this.m_btn_fanhui = (this.getChildAt(1));
            this.m_comp_hangmv = (this.getChildAt(2));
        }
    }
    UI_jijia.URL = "ui://s88k2hl2vpo94q";

    class UI_Comp_fengxiang extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "Comp_fengxiang"));
        }
        onConstruct() {
            this.m_btn_get = (this.getChildAt(6));
        }
    }
    UI_Comp_fengxiang.URL = "ui://s88k2hl2vpo94r";

    class UI_Comp_fengxiangjm extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("wx", "Comp_fengxiangjm"));
        }
        onConstruct() {
            this.m_head1 = (this.getChildAt(4));
            this.m_head2 = (this.getChildAt(5));
            this.m_btn_share = (this.getChildAt(7));
            this.m_btn_huan = (this.getChildAt(8));
            this.m_btn_invate1 = (this.getChildAt(9));
            this.m_btn_invate2 = (this.getChildAt(10));
            this.m_btn_invate3 = (this.getChildAt(11));
        }
    }
    UI_Comp_fengxiangjm.URL = "ui://s88k2hl2vpo94t";

    class wxBinder {
        static bindAll() {
            fgui.UIObjectFactory.setExtension(UI_HangMuNengLiang.URL, UI_HangMuNengLiang);
            fgui.UIObjectFactory.setExtension(UI_TouPiao.URL, UI_TouPiao);
            fgui.UIObjectFactory.setExtension(UI_Comp_hangmv2.URL, UI_Comp_hangmv2);
            fgui.UIObjectFactory.setExtension(UI_ui_videofail.URL, UI_ui_videofail);
            fgui.UIObjectFactory.setExtension(UI_ui_jm.URL, UI_ui_jm);
            fgui.UIObjectFactory.setExtension(UI_btn_fh.URL, UI_btn_fh);
            fgui.UIObjectFactory.setExtension(UI_btn_ljhd.URL, UI_btn_ljhd);
            fgui.UIObjectFactory.setExtension(UI_btn_yuyue.URL, UI_btn_yuyue);
            fgui.UIObjectFactory.setExtension(UI_btn_ljfs.URL, UI_btn_ljfs);
            fgui.UIObjectFactory.setExtension(UI_comp_yinpai.URL, UI_comp_yinpai);
            fgui.UIObjectFactory.setExtension(UI_btn_2bei.URL, UI_btn_2bei);
            fgui.UIObjectFactory.setExtension(UI_ui_invate.URL, UI_ui_invate);
            fgui.UIObjectFactory.setExtension(UI_Comp_zaiju.URL, UI_Comp_zaiju);
            fgui.UIObjectFactory.setExtension(UI_ZaiJu.URL, UI_ZaiJu);
            fgui.UIObjectFactory.setExtension(UI_Comp_toupiao.URL, UI_Comp_toupiao);
            fgui.UIObjectFactory.setExtension(UI_btn_xunaze.URL, UI_btn_xunaze);
            fgui.UIObjectFactory.setExtension(UI_jijia.URL, UI_jijia);
            fgui.UIObjectFactory.setExtension(UI_Comp_fengxiang.URL, UI_Comp_fengxiang);
            fgui.UIObjectFactory.setExtension(UI_Comp_fengxiangjm.URL, UI_Comp_fengxiangjm);
        }
    }

    let VideoFailUI = class VideoFailUI extends BaseUI {
        constructor() {
            super(...arguments);
            this._showTips = false;
        }
        get URL() {
            wxBinder.bindAll();
            return UI_ui_videofail.URL;
        }
        OnCreate() {
            this.ui.m_btn_fengxiang.onClick(this, this._onClickShare);
            this.ui.m_btn_quxioa.onClick(this, this._onClickCancle);
        }
        OnShow(handler, reason) {
            ReportMgr.inst.sendReport(ReportType.WIN_SHARE);
            this._reason = reason + "-分享";
            this._callBack = (s) => {
                if (s) {
                    handler.runWith(s);
                    this._closeSelf();
                }
                else {
                    if (this._showTips) {
                        TipsMgr.Tips("分享失败");
                    }
                }
            };
        }
        OnHide() {
        }
        OnDispose() {
        }
        onClose() {
        }
        _onClickShare() {
            ReportMgr.inst.sendReport(ReportType.WIN_SHARE_OK);
            this._showTips = true;
            GameSDK.shareMsg(this, this._callBack, this._reason);
        }
        _onClickCancle() {
            ReportMgr.inst.sendReport(ReportType.WIN_SHARE_CLOSE);
            this._showTips = false;
            this._callBack.call(this, false);
            this._closeSelf();
        }
        _closeSelf() {
            UIManager.inst.PopUI(this.layer, false);
        }
    };
    VideoFailUI = __decorate([
        ui_register()
    ], VideoFailUI);

    class WeChat extends BaseSDK {
        constructor() {
            super();
            this._closeCallback = null;
            this.defaultServerConfig = {
                "is_AddDesk": { state: "1" },
                "is_AvdFloat": { state: "1" },
                "is_AvdList": { state: "1" },
                "is_GameNative": { state: "0", interval: 0 },
                "is_MoreGameBtn": { state: "1" },
                "is_PropSceneUI": { state: "1", interval: 0 },
                "is_banner": { state: "1", interval: 0 },
                "ad_all_interval": 0,
            };
            this._jumpOther = false;
            this._hasLogin = false;
            this._bannerIds = [];
            this.isShowBanner = false;
            this.bannerShowCount = 0;
            this._curBannerKey = "";
            this._bannerKeyMap = new Map();
            this._backCreating = false;
            this.interstitialAdIndex = 0;
            this.interstitialAdReady = false;
            this.isInterstitialShow = false;
            this.isGame = true;
            this.isNoMouseClick = false;
            this.clickTime = 0;
            this.showLastTime = 0;
            this.errorAdCount = 0;
            this.RewardedVideoAdIndex = 0;
            this.isAutoPlayVidoeAd = false;
            this._lastAlertShareTime = "LogicData._lastAlertShareTime";
            this._todayAlertShare = "LogicData._todayAlertShare";
            this.isLoading = false;
            this.shareTime = 0;
            this.isShare = false;
            this._customLog = true;
            this._customAdArr = [];
            this._customAdShow = [];
            this._customAdState = [];
            this._version = SDKConfig.WECHAT_VERSION;
            this.configURL = `https://game3.jpsdk.com/static/api/${SDKConfig.GAMEID}/wx/api.json`;
            this.wx = Laya.Browser.window.wx;
            let self = this;
            this.wx.onShow((res) => {
                self.initConfig(null);
                let newTime = Date.now();
                let str = JSON.stringify(res);
                try {
                    if (this.data.Channel0 && this.data.Channel0 != "") {
                        if (str.indexOf(this.data.Channel0) != -1) {
                            this.playerFlag = PlayerFlag.CHANNEL0;
                            console.log("标配为渠道0");
                        }
                    }
                }
                catch (error) { }
                let scene;
                if (self.isShare) {
                    this.sendEvent({ 'share': this.reason, 'status': 1 });
                    if (self.shareTime > 0 && newTime - self.shareTime >= 3000) {
                        self.onShareComplete && self.onShareComplete.runWith(true);
                    }
                    else {
                        self.onShareComplete && self.onShareComplete.runWith(false);
                    }
                    scene = PlatEventName.PLAT_SHARE;
                }
                if (self.isShowBanner) {
                    self.showBanner();
                }
                if (self._jumpOther) {
                    self._jumpOther = false;
                    scene = PlatEventName.PLAT_JUMP_GAME;
                }
                self.isShare = false;
                GameSDK.callShowFunc(scene);
            });
            this.wx.onShareAppMessage(() => {
                let info = ToolSDK.randomItem(SDKConfig.WECHAT_SHARETEMPLATEID);
                return {
                    title: info.title,
                    imageUrl: info.iconPath
                };
            });
            let systemInfo = wx.getSystemInfoSync();
            this.wx.onHide((res) => {
                self._jumpOther = false;
                self.destroyBanner();
                SDKEventMgr.Int.emit(PlatEventName.PLAT_HIDE, res);
                if (!res || !res.mode) {
                    return;
                }
                if ("close" === res.mode) {
                    this.onAppHide();
                    return SDKEventMgr.Int.emit(PlatEventName.PLAT_CLOSE, res);
                }
                let scene;
                let _path = res.targetPagePath;
                if (!_path) {
                    if ("android" === systemInfo.platform && "hide" === res.mode && 10 === res.targetAction) {
                        scene = PlatEventName.PLAT_TOUSU;
                        SDKEventMgr.Int.emit(PlatEventName.PLAT_TOUSU, res);
                        this.playerFlag = PlayerFlag.NO_AD;
                    }
                    return;
                }
                if (8 === res.targetAction) {
                    let pathArr = _path.split(".");
                    switch (pathArr.pop()) {
                        case "SelectConversationUI":
                        case "MMUINavigationController":
                            scene = PlatEventName.PLAT_SHARE;
                            SDKEventMgr.Int.emit(PlatEventName.PLAT_SHARE, res);
                            break;
                        case "SnsAdNativeLandingPagesPreviewUI":
                            res.adType = "app";
                            scene = PlatEventName.PLAT_TOUCH_BANNER_AD;
                            SDKEventMgr.Int.emit(PlatEventName.PLAT_TOUCH_BANNER_AD, res);
                            break;
                        case "SnsUploadUI":
                            scene = PlatEventName.PLAT_SHARE_SNS;
                            SDKEventMgr.Int.emit(PlatEventName.PLAT_SHARE_SNS, res);
                            break;
                        case "AppBrandProfileUI":
                        case "NewWAProfileViewController":
                            scene = PlatEventName.PLAT_ABOUT;
                            SDKEventMgr.Int.emit(PlatEventName.PLAT_ABOUT, res);
                            this.playerFlag = PlayerFlag.NO_AD;
                            break;
                        case "AppBrandAuthorizeUI":
                        case "WAAppAuthorizationViewController":
                            scene = PlatEventName.PLAT_CONFIG;
                            SDKEventMgr.Int.emit(PlatEventName.PLAT_CONFIG, res);
                            this.playerFlag = PlayerFlag.NO_AD;
                            break;
                        case "AppBrandProcessProxyUI":
                            scene = PlatEventName.PLAT_COLLECT;
                            SDKEventMgr.Int.emit(PlatEventName.PLAT_COLLECT, res);
                            break;
                    }
                }
                else if (9 === res.targetAction || 10 === res.targetAction) {
                    self._jumpOther = true;
                    if (-1 !== _path.indexOf("weixinadinfo") && -1 !== _path.indexOf("gdt_vid")) {
                        if (9 === res.targetAction) {
                            res.appId = _path.substring(0, 18);
                        }
                        res.adType = (10 === res.targetAction) ? "h5" : "program";
                        scene = PlatEventName.PLAT_TOUCH_BANNER_AD;
                        return SDKEventMgr.Int.emit(PlatEventName.PLAT_TOUCH_BANNER_AD, res);
                    }
                    if (9 === res.targetAction) {
                        let appid = _path.substring(0, 18);
                        res.appId = appid;
                        scene = PlatEventName.PLAT_JUMP_GAME;
                        SDKEventMgr.Int.emit(PlatEventName.PLAT_JUMP_GAME, res);
                        this.onAppHide();
                    }
                    else if (10 === res.targetAction) {
                        if (-1 !== _path.indexOf("wapreportwxadevlog")) {
                            scene = PlatEventName.PLAT_TOUSU;
                            SDKEventMgr.Int.emit(PlatEventName.PLAT_TOUSU, res);
                            this.playerFlag = PlayerFlag.NO_AD;
                        }
                    }
                }
                GameSDK.callHideFunc(scene);
            });
            this.wx.showShareMenu({
                withShareTicket: true,
                menus: ['shareAppMessage', 'shareTimeline']
            });
            Laya.stage.on(Laya.Event.MOUSE_DOWN, this, this.onMouseHandler);
            Laya.stage.on(Laya.Event.MOUSE_MOVE, this, this.onMouseHandler);
            this._initShareFriendListener();
        }
        onAppHide() {
            let needCount = Math.random();
            let count = this.data.HIDE_REWARDAD_RATE;
            let rewardAd = this.data.HIDE_REWARDAD_MAX;
            let hideRewardAdCount = LocalCache.get("HIDE_REWARDAD_COUNT", 0);
            let nowTime = Times.now;
            let hideRewardAdTime = LocalCache.get("HIDE_REWARDAD_TIME", nowTime);
            if (hideRewardAdTime != 0 && !Times.isSameDay(hideRewardAdTime)) {
                hideRewardAdCount = 0;
                LocalCache.set("HIDE_REWARDAD_TIME", nowTime);
            }
            if (GameSDK.isOpenCitys() && !GameSDK.isCurrentVersion && needCount < count && hideRewardAdCount < rewardAd) {
                LocalCache.set("HIDE_REWARDAD_COUNT", hideRewardAdCount += 1);
                this.showADInsert();
                this.showADVideo(this, () => { }, "");
                this.sendEvent("视频H");
            }
        }
        sendEvent(name, params) {
            if (!!this.wx && !!this.wx.aldSendEvent) {
                if (!!params) {
                    this.wx.aldSendEvent(name, { "params": JSON.stringify(params) });
                }
                else {
                    this.wx.aldSendEvent(name);
                }
            }
            super.sendEvent(name, params);
        }
        onConfigLoad() {
            if (this.data.shareList && this.data.shareList.length > 0) {
                SDKConfig.WECHAT_SHARETEMPLATEID = this.data.shareList;
            }
            if (this._insertAdNode) {
                this._insertAdNode.destroy();
                this._insertAdNode = null;
            }
            this.wx.setKeepScreenOn({ keepScreenOn: true });
            this.createBanner();
            this.startRefeshInsertAd();
            try {
                var obj = this.wx.getLaunchOptionsSync();
                let str = JSON.stringify(obj);
                if (this.data.Channel0 && this.data.Channel0 != "") {
                    if (str.indexOf(this.data.Channel0) != -1) {
                        this.playerFlag = PlayerFlag.CHANNEL0;
                        console.log("标配为渠道0");
                    }
                }
            }
            catch (error) { }
            let self = this;
            if (!this._hasLogin) {
                this._hasLogin = true;
                this.login().then((data) => {
                    self.onLoginSuccess(data);
                });
            }
        }
        onLoginSuccess(data) {
            let option = this.getLaunchOptionsSync();
            data.clue_token = (!!option && !!option.query && !!option.query.clue_token) ? option.query.clue_token : null;
            data.invite_uid = (!!option && !!option.query && option.query.invite_uid != undefined) ? option.query.invite_uid : null;
            WxMgr.inst.login(data);
        }
        createBanner() {
            var self = this;
            var system = this.wx.getSystemInfoSync();
            if (!!!this._bannerShow) {
                if (!!this._bannerIds && this._bannerIds.length > 1 && !!this._bannerBack) {
                    this._bannerShow = this._bannerBack;
                    this._curBannerId = this._curBannerId == this._bannerIds[0] ? this._bannerIds[1] : this._bannerIds[0];
                    this._bannerBack = null;
                    if (this.isShowBanner) {
                        this.showBanner();
                    }
                }
                else {
                    this._curBannerId = !!this._bannerIds && this._bannerIds.length > 1 ? (this._curBannerId == this._bannerIds[0] ? this._bannerIds[1] : this._bannerIds[0]) : this.data.BannerAd;
                    if (!!this._curBannerKey) {
                        this._bannerKeyMap.delete(this._curBannerKey);
                    }
                    if (!!!this._curBannerId) {
                        this._curBannerId = this.data.BannerAd;
                    }
                    this._curBannerKey = this._curBannerId + Date.now();
                    this._bannerShow = this._createBannerAd(this._curBannerId);
                    this._bannerShow.onResize(function (e) {
                        self._bannerShow.style.top = system.screenHeight - e.height;
                        self._bannerShow.style.left = (system.screenWidth - e.width) * 0.5;
                    });
                    this._bannerShow.onLoad(function () {
                        if (self.isShowBanner) {
                            self.showBanner();
                        }
                        else {
                            self.hideBanner();
                        }
                    });
                    this._bannerShow.onError(function (err) {
                        if (!!self._curBannerKey) {
                            self._bannerKeyMap.delete(self._curBannerKey);
                        }
                        self.isShowBanner = false;
                        self._destroyBannerAd(self._bannerShow);
                        self._bannerShow = null;
                    });
                }
            }
            if (!!this._bannerIds && this._bannerIds.length > 1 && !!!this._bannerBack) {
                if (this._backCreating) {
                    return;
                }
                this._backCreating = true;
                let banner = this._createBannerAd(this._curBannerId == this._bannerIds[0] ? this._bannerIds[1] : this._bannerIds[0]);
                banner.onResize(function (e) {
                    banner.style.top = system.screenHeight - e.height;
                    banner.style.left = (system.screenWidth - e.width) * 0.5;
                });
                banner.onLoad(function () {
                    self._bannerBack = banner;
                    self._backCreating = false;
                });
                banner.onError(function (err) {
                    self._destroyBannerAd(self._bannerBack);
                    self._bannerBack = null;
                    self._backCreating = false;
                });
            }
        }
        showBanner() {
            this.isShowBanner = true;
            if (!!!this._bannerShow) {
                this.createBanner();
            }
            else {
                this._bannerShow.show().then(() => {
                    if (!this.isShowBanner) {
                        this.hideBanner();
                    }
                    else {
                        if (!this._bannerKeyMap.get(this._curBannerKey)) {
                            this._bannerKeyMap.set(this._curBannerKey, true);
                            this.bannerShowCount++;
                        }
                    }
                });
            }
        }
        hideBanner() {
            this.isShowBanner = false;
            if (!!this._bannerShow) {
                this._bannerShow.hide().then(() => {
                    if (this.isShowBanner) {
                        this.showBanner();
                    }
                });
            }
        }
        destroyBanner() {
            this.isShowBanner = false;
            this._destroyBannerAd(this._bannerShow);
            this._bannerShow = null;
        }
        _destroyBannerAd(banner) {
            if (!!banner) {
                banner.offResize();
                banner.offLoad();
                banner.offError();
                banner.destroy();
            }
        }
        _createBannerAd(id) {
            var system = this.wx.getSystemInfoSync();
            let banner = this.wx.createBannerAd({
                adUnitId: id,
                adIntervals: 60,
                style: {
                    top: system.windowHeight - 100,
                    left: (system.windowWidth - 300) * 0.5,
                    width: 300,
                    height: 100
                }
            });
            return banner;
        }
        showADInsert() {
            let self = this;
            if (!this.interstitialAd || !this.interstitialAdReady) {
                let adlist = this.data.InterstAd.split(",");
                let adLength = adlist.length;
                this.interstitialAd = this.wx.createInterstitialAd({
                    adUnitId: adlist[this.interstitialAdIndex % adLength]
                });
                this.interstitialAd.onLoad(() => {
                    self.interstitialAdReady = true;
                    console.log('插屏 广告加载成功');
                });
                this.interstitialAd.onError(err => {
                    self.isNoMouseClick = false;
                    console.log("onError", err);
                });
                this.interstitialAd.onClose(res => {
                    self.isInterstitialShow = false;
                    self.isNoMouseClick = false;
                    console.log('插屏 广告关闭');
                    if (GameSDK.isOpenCitys() && !GameSDK.isCurrentVersion && +self.data.DOPEN == 1) {
                        setTimeout(function () {
                            self.interstitialAd.destroy();
                            self.interstitialAd = null;
                            console.log('插屏广告 销毁');
                        }, 480);
                    }
                });
            }
            this.isInterstitialShow = false;
            this.interstitialAd.show().then((res) => {
                self.interstitialAdIndex++;
                self.isInterstitialShow = true;
                self.autioCloseInterstitialAd();
            }).catch((err) => {
            });
        }
        autioCloseInterstitialAd() {
            if (this.isInterstitialShow && this.interstitialAd && this.isNoMouseClick && +this.data.INTERSTITIAL_AUTO_CLOSE == 1) {
                Laya.timer.once(15000, this, () => {
                    if (this.interstitialAd && this.interstitialAd.destroy)
                        this.interstitialAd.destroy();
                    this.interstitialAd = null;
                });
            }
        }
        pauseVideoRecord() {
            this.isGame = false;
        }
        resumeVideoRecord() {
            this.isGame = false;
        }
        stopVideoRecord() {
            this.isGame = false;
        }
        startVideoRecord(time, callback) {
            this.isGame = true;
        }
        startRefeshInsertAd() {
            this.stopRefeshInsertAd();
            Laya.timer.loop(1000, this, this.refeshInsertAd);
        }
        onMouseHandler() {
            this.clickTime = Times.now;
        }
        refeshInsertAd() {
            if (!GameSDK.isCurrentVersion && !this.isGame) {
                let nowTime = Times.now;
                if (!!!this.clickTime) {
                    this.clickTime = nowTime - 1000;
                    return;
                }
                if (nowTime - this.clickTime >= 20000) {
                    this.isNoMouseClick = true;
                    this.clickTime = nowTime;
                    this.showADInsert();
                }
            }
        }
        stopRefeshInsertAd() {
            Laya.timer.clear(this, this.refeshInsertAd);
        }
        showADVideo(caller, onComplete, reason, args) {
            var self = this;
            if (typeof (this.wx.createRewardedVideoAd) == 'undefined')
                return;
            Laya.MouseManager.enabled = false;
            this.noticeOpenVideo(reason);
            this.onADViewComplete = Handler.create(caller, onComplete, args);
            this.isAutoPlayVidoeAd = true;
            this._videoReason = reason;
            if (!this._videoAd) {
                console.log("创建视频");
                this.data.RewardedVideoAd.split(",");
                this._videoAd = this.wx.createRewardedVideoAd({ adUnitId: this.data.RewardedVideoAd, multiton: true });
                this._videoAd.onLoad(() => {
                    self.errorAdCount = 0;
                    Laya.MouseManager.enabled = true;
                    self.pauseVideoRecord();
                    if (self.isAutoPlayVidoeAd) {
                        self._videoAd.show().then(() => {
                            self.stopRefeshInsertAd();
                        }).catch(() => {
                            self._videoAd.load();
                        });
                    }
                });
                this._videoAd.onError((err) => {
                    Laya.MouseManager.enabled = true;
                    self.sendEvent("激励广告错误" + err.errCode);
                    if (err.errCode == 1004) ;
                    if (GameSDK.isLimit()) {
                        TipsMgr.Tips("暂无视频");
                        self.onADViewComplete.runWith(false);
                    }
                    else {
                        let shareTime = LocalCache.get(this._lastAlertShareTime, 0);
                        let now = Date.now();
                        let shareCount = LocalCache.get(this._todayAlertShare, 0);
                        if (!!!shareTime || now - shareTime >= 24 * 60 * 60 * 1000 || new Date(shareTime).getDay() != new Date(now).getDay()) {
                            shareCount = 0;
                            LocalCache.set(this._todayAlertShare, shareCount);
                        }
                        if (shareCount < 3) {
                            shareCount++;
                            LocalCache.set(this._lastAlertShareTime, now);
                            LocalCache.set(this._todayAlertShare, shareCount);
                            UIManager.inst.ShowUI(VideoFailUI, UILayer.top, self.onADViewComplete, this._videoReason);
                        }
                        else {
                            TipsMgr.Tips("暂无视频");
                            self.onADViewComplete.runWith(false);
                        }
                    }
                });
                this._videoAd.onClose((res) => {
                    self.RewardedVideoAdIndex++;
                    Laya.MouseManager.enabled = true;
                    self.isAutoPlayVidoeAd = false;
                    self.startRefeshInsertAd();
                    self.resumeVideoRecord();
                    if ((res && res.isEnded) || (res === undefined)) {
                        self.noticeCloseVideo(self._videoReason);
                        self.onADViewComplete.runWith(true);
                    }
                    else {
                        self.onADViewComplete.runWith(false);
                        self.showTips("需要看完视频才能获得奖励！", 1000);
                    }
                });
            }
            else {
                Laya.MouseManager.enabled = true;
                console.log(JSON.stringify(this._videoAd.show));
                this._videoAd.show().then(() => {
                    this.stopRefeshInsertAd();
                }).catch(() => {
                    this._videoAd.load();
                });
            }
            if (!Laya.MouseManager.enabled) {
                Laya.timer.once(2000, this, function () {
                    Laya.MouseManager.enabled = true;
                });
            }
        }
        showTips(message, time) {
            this.wx.showToast({
                title: message,
                icon: "none",
                duration: time
            });
        }
        showLoading() {
            this.isLoading = true;
            if (this.wx && this.wx.showLoading) {
                this.wx.showLoading({ title: "请求中，请稍后..." });
                if (this.showTime)
                    clearTimeout(this.showTime);
                this.showTime = setTimeout(this.hideLoading, 3000);
            }
        }
        hideLoading() {
            this.isLoading = false;
            if (this.wx && this.wx.hideLoading) {
                this.wx.hideLoading({});
            }
        }
        createNative(navigateId) {
            return null;
        }
        showToast(message, time = 2000) {
            console.log(message, time);
            this.wx.showToast({
                title: message,
                duration: time
            });
        }
        navigateToMiniGame(msg) {
            this.wx && this.wx.navigateToMiniProgram({
                appId: msg.appId,
                path: "",
                success: function () {
                    console.log("-----------跳转小游戏:" + msg.name + "成功-----------");
                },
                fail: function (res) {
                    console.log("-----------跳转小游戏:" + msg.name + "失败-----------");
                    console.log(JSON.stringify(res));
                }
            });
        }
        vibrateShort(type) {
            this.wx && this.wx.vibrateShort({
                type: type
            });
            console.log("短振动");
        }
        vibrateLong() {
            this.wx && this.wx.vibrateLong();
            console.log("长振动");
        }
        loadSubpackage(packName, zipFileName, topath, callbackSuccess, callbackProgress) {
            console.log("O_O 微信下载分包 开始");
            let wx = Laya.Browser.window.wx;
            if (wx.loadSubpackage) {
                var loadTask = wx.loadSubpackage({
                    name: packName,
                    success: (res) => {
                        console.log("微信分包加载成功！", packName);
                        if (callbackSuccess)
                            callbackSuccess();
                    },
                    fail: (res) => {
                        console.log("微信分包加载失败！", packName);
                        if (callbackProgress)
                            callbackProgress();
                    }
                });
                loadTask.onProgressUpdate(res => {
                    console.log('下载进度', res.progress);
                    console.log('已经下载的数据长度', res.totalBytesWritten);
                    console.log('预期需要下载的数据总长度', res.totalBytesExpectedToWrite);
                    if (callbackProgress) {
                        callbackProgress(res.progress);
                    }
                });
            }
            else {
                if (callbackSuccess)
                    callbackSuccess();
            }
        }
        unZip(zipFileName, path = '', callbackSuccess = undefined, callbackFailed = undefined) {
            let wx = Laya.Browser.window.wx;
            if (wx) {
                var fileManager = wx.getFileSystemManager();
                var targetPath = Laya.Browser.window.wx.env.USER_DATA_PATH;
                if (path != undefined && path.length > 0) {
                    targetPath += "/" + path;
                }
                fileManager.unzip({
                    zipFilePath: zipFileName,
                    targetPath: targetPath,
                    success: function (res) {
                        console.log("解压成功:" + zipFileName);
                        Laya.URL.basePath = Laya.Browser.window.wx.env.USER_DATA_PATH + "/res/";
                        if (callbackSuccess) {
                            callbackSuccess();
                        }
                    },
                    fail: function (res) {
                        console.log("解压失败:" + zipFileName);
                        if (callbackFailed) {
                            callbackFailed();
                        }
                    },
                });
            }
            else {
                return callbackSuccess();
            }
        }
        shareMsg(caller, onComplete, reason, args) {
            this.onShareComplete = Handler.create(caller, onComplete, args);
            this.shareTime = Date.now();
            this.reason = reason;
            this.sendEvent({ 'share': reason, 'status': 0 });
            this.isShare = true;
            console.log("分享时间：", this.shareTime);
            let info = ToolSDK.randomItem(SDKConfig.WECHAT_SHARETEMPLATEID);
            let uid;
            if (!!WxMgr.inst.token) {
                let data = DesMg.decode(WxMgr.inst.token);
                if (!!data) {
                    data = JSON.parse(data);
                    if (!!data && data.user_id != undefined) {
                        uid = data.user_id;
                    }
                }
            }
            this.wx.shareAppMessage({
                title: info.title,
                imageUrl: info.iconPath,
                query: uid == undefined ? "" : `invite_uid=${uid}`
            });
        }
        get _customCfg() {
            return this.data.customCfg;
        }
        showCustomAd(type, top) {
            return new Promise((resolve, reject) => {
                let self = this;
                top = this._checkCustomLimit(type, top);
                this._customAdShow[type] = true;
                let curAd = this._customAdArr[type];
                if (!!!curAd) {
                    this.createCustomAd(type, top).then(s => {
                        resolve(s);
                    });
                }
                else {
                    if (!!this._customAdState[type]) {
                        resolve(false);
                        return;
                    }
                    if (this._customLog) {
                        console.log("ready show custom", type);
                    }
                    curAd.show().then(() => {
                        let show = self._customAdShow[type];
                        if (this._customLog) {
                            console.log("show custom " + type + " success:", show);
                        }
                        if (!show) {
                            self.hideCustomAd(type);
                            resolve(false);
                        }
                        else {
                            resolve(true);
                        }
                    });
                }
            });
        }
        createCustomAd(type, top) {
            return new Promise((resolve, reject) => {
                let self = this;
                let systemInfo = wx.getSystemInfoSync();
                if (!!this._customAdArr[type]) {
                    this._destroyCustomAd(type);
                }
                if (!!this._customAdState[type]) {
                    resolve(false);
                    return;
                }
                this._customAdState[type] = 1;
                top = this._checkCustomLimit(type, top);
                let base = top;
                top = (top / 1334) * systemInfo.screenHeight;
                let curAd = this.wx.createCustomAd({
                    adUnitId: this._customCfg[type].id,
                    style: {
                        left: this._customCfg[type].side == 0 ? this._customCfg[type].left : (this._customCfg[type].side == 1 ? (systemInfo.screenWidth + this._customCfg[type].left) >> 1 : systemInfo.screenWidth + this._customCfg[type].left),
                        top: top,
                        width: 750,
                        fixed: true
                    }
                });
                curAd.onLoad(() => {
                    let show = self._customAdShow[type];
                    if (this._customLog) {
                        console.log("create custom " + type + " success:", show);
                    }
                    self._customAdState[type] = 0;
                    if (!show) {
                        self.hideCustomAd(type);
                        resolve(false);
                    }
                    else {
                        self.showCustomAd(type, base).then(s => {
                            resolve(s);
                        });
                    }
                });
                curAd.onError((e) => {
                    resolve(false);
                    console.error("custom " + type + " create fail", e);
                    let show = self._customAdShow[type];
                    self._destroyCustomAd(type);
                    self._customAdShow[type] = show;
                    Laya.timer.once(10000, self, self.createBanner, [type, base]);
                });
                curAd.onClose(() => {
                    self.hideCustomAd(type);
                });
                this._customAdArr[type] = curAd;
            });
        }
        _checkCustomLimit(type, top) {
            let destroy = false;
            top = (top == undefined || top == null) ? this._customCfg[type].top : top;
            if (type == CustomType.Bottom_Roll) {
                top = this.data.bannertop || top;
            }
            destroy = this._customCfg[type].top != top;
            this._customCfg[type].top = top;
            if (destroy) {
                this._destroyCustomAd(type);
            }
            return top;
        }
        _destroyCustomAd(type) {
            let curAd = this._customAdArr[type];
            if (!!curAd) {
                curAd.offLoad();
                curAd.offError();
                curAd.offClose();
                curAd.destroy();
            }
            this._customAdShow[type] = false;
            this._customAdArr[type] = null;
            this._customAdState[type] = 0;
        }
        hideCustomAd(type) {
            if (type == undefined) {
                for (let i = CustomType.Left_Grid; i <= CustomType.Main_Rect; ++i) {
                    this.hideCustomAd(i);
                }
            }
            else {
                this._customAdShow[type] = false;
                if (!!this._customAdState[type]) {
                    return;
                }
                let curAd = this._customAdArr[type];
                if (!!curAd) {
                    let self = this;
                    if (this._customLog) {
                        console.log("ready hide custom", type);
                    }
                    curAd.hide().then(() => {
                        let show = self._customAdShow[type];
                        if (this._customLog) {
                            console.log("hide custom " + type + " success:", show);
                        }
                        if (show) {
                            self.showCustomAd(type);
                        }
                    });
                }
            }
        }
        isShortScreen() {
            let info = wx.getSystemInfoSync();
            let rate = info.windowHeight / info.windowWidth;
            return rate <= 1.8;
        }
        getLaunchOptionsSync() {
            return this.wx.getLaunchOptionsSync();
        }
        getEnterOptionsSync() {
            return this.wx.getEnterOptionsSync();
        }
        navigateToMiniProgram() {
        }
        postMessage(type, target, args) {
            let context = this._getShareContext();
            if (!!context) {
                if (!!target) {
                    context.postMessage({
                        event: "viewport",
                        fromEngine: true,
                        x: target.x,
                        y: target.y,
                        width: target.width,
                        height: target.height
                    });
                }
                context.postMessage({
                    event: type,
                    args: args
                });
            }
        }
        _getShareContext() {
            const context = this.wx.getOpenDataContext();
            return context;
        }
        setShareFriendQuery(scene) {
            this.wx.setMessageToFriendQuery({ shareMessageToFriendScene: scene });
        }
        _initShareFriendListener() {
            this.wx.onShareMessageToFriend((data) => {
                if (!!data) {
                    if (data.success) {
                        MsgManager.getInst().event(EventType.WxShareFriend, true);
                        return;
                    }
                }
                MsgManager.getInst().event(EventType.WxShareFriend, false);
            });
        }
        getShareCanvas() {
            let context = this._getShareContext();
            if (!!context) {
                return context.canvas;
            }
            return null;
        }
        _getUserInteractiveStorage(key) {
            return new Promise((resolve, reject) => {
                this.wx.getUserInteractiveStorage({
                    keyList: [key],
                    success: (e) => {
                        resolve(e);
                    },
                    fail: (e) => {
                        reject(e);
                    }
                });
            });
        }
        getEncryptData() {
            return new Promise((resolve, reject) => {
                this._getUserInteractiveStorage(WxMgr.inst.SHARE_ROBOT + "").then((res) => {
                    this.login().then((data) => {
                        data.iv = res.iv;
                        data.encryptedData = res.encryptedData;
                        resolve(data);
                    }).catch((e) => {
                        reject(e);
                    });
                }).catch((e) => {
                    reject(e);
                });
            });
        }
        set sessionKey(val) {
            this._sessionKey = val;
        }
        login() {
            return new Promise((resolve, reject) => {
                this._checkSessionKey().then(() => {
                    resolve({ session_key: this._sessionKey });
                }).catch(() => {
                    console.error("微信登录");
                    this.wx.login({
                        success: (res) => {
                            resolve({ code: res.code });
                        },
                        fail: function (e) {
                            console.error("wx login fail==>", e);
                            reject(e);
                        },
                    });
                });
            });
        }
        _checkSessionKey() {
            return new Promise((resolve, reject) => {
                if (!!!this._sessionKey) {
                    reject();
                }
                else {
                    this.wx.checkSession({
                        success: () => {
                            resolve();
                        },
                        fail: () => {
                            reject();
                        }
                    });
                }
            });
        }
        setUserCloudStorage(k, v) {
            this.wx.setUserCloudStorage({
                KVDataList: [
                    { key: k, value: v }
                ]
            });
        }
        getSetting() {
            return new Promise((resolve, reject) => {
                this.wx.getSetting({
                    success: (e) => {
                        resolve(e);
                    }, fail: (e) => {
                        reject(e);
                    }
                });
            });
        }
        authorize(scope) {
            return new Promise((resolve, reject) => {
                this.wx.authorize({
                    scope: scope,
                    success: () => {
                        resolve(true);
                    },
                    fail: e => {
                        resolve(false);
                    }
                });
            });
        }
    }

    class SDKData {
        static get data() {
            return {
                PlatType: PlatType$1.AUTO,
                callBack: function () { },
                use_count_banner: !1,
                fgui_sdk_path: "mainpack/fgui/",
                GAMEID: "20301",
                VIVO_VERSION: "1.0.0",
                OPPO_VERSION: "1.0.0",
                TOUTIAO_VERSION: "1.0.0",
                TOUTIAO_GAME_NAME: "王牌指挥官",
                TOUTIAO_SHARECONTENT: ["带领你的小兵冲锋陷阵!"],
                SHARE_CONTENT: ["带领你的小兵冲锋陷阵!"],
                TOUTIAO_SHARETEMPLATEID: ["8bd9l7fl84i6ikckbd"],
                WECHAT_GAME_NAME: "王牌指挥官",
                WECHAT_VERSION: "1.0.2",
                WECHAT_SHARECONTENT: ["带领你的小兵冲锋陷阵!"],
                WECHAT_SHARETEMPLATEID: [{
                        title: "带领你的小兵冲锋陷阵! ",
                        iconPath: "https://wxoss.jpsdk.com/game/20280/wechat/share/shareIcon.jpg"
                    }],
                HBS_VERSION: "1.0.0",
                GAME233_APPKEY: "1606253762",
                GAME233_POS: "999000000",
                APP_IOS_APPID: "5171365",
                GAME233_POSVideo: "999000000",
                GAME233_POSVideoAll: "999000001",
                GAME233_POSBanner: "999000002",
                GAME233_POSInsert: "999000003",
                PltDefaultConfig: [{
                        key: PlatType$1.TOUTIAO,
                        value: {
                            BannerAd: "5hqtveq3juv1rho0ko",
                            InterstAd: "1t9oueht6caho55mi7",
                            InterstAds: ["1t9oueht6caho55mi7"],
                            IntervalAdTime: 3,
                            NativeAvd: !1,
                            RewardAd: "1a9pl5aaqqi539mjx2",
                            addelay: 0,
                            appid: "ttc9d0bf6d56e99ad402",
                            navigateAppList: [],
                            Offcitys: [],
                            OPLISTS: [],
                            REWARDAD_MAX: 0,
                            REWARDAD_RATE: 0,
                            REWARDAD_RESTART_GAME: 0,
                            HIDE_REWARDAD_RATE: 0,
                            HIDE_REWARDAD_MAX: 0,
                            shareList: [],
                            Shareid: "1huiaxi2py32o548h6",
                            Version: "1.0.0",
                            playerSkinShield: "2,3,4,5,6",
                            is_AddDesk: {
                                state: "0"
                            },
                            is_AvdFloat: {
                                state: "0"
                            },
                            is_AvdList: {
                                state: "0"
                            },
                            is_GameNative: {
                                state: "0",
                                interval: 0
                            },
                            is_MoreGameBtn: {
                                state: "0"
                            },
                            is_PropSceneUI: {
                                state: "0",
                                interval: 0
                            },
                            is_banner: {
                                state: "1",
                                interval: 0
                            },
                            ad_all_interval: 0
                        }
                    }, {
                        key: PlatType$1.OPPO,
                        value: {
                            BannerAd: "317030",
                            IntervalAdTime: 0,
                            NativeAd: 317032,
                            InterstAd: 317033,
                            RewardAd: "317031",
                            ad_all_interval: 10,
                            aderrormaxcount: 0,
                            ErrorProba: .2,
                            appid: "30521975",
                            GameBanner: "317033",
                            GamePortal: "304450",
                            is_AddDesk: {
                                state: "1"
                            },
                            is_AvdFloat: {
                                state: "0"
                            },
                            is_AvdList: {
                                state: "0"
                            },
                            is_GameNative: {
                                state: "1",
                                interval: 5
                            },
                            is_MoreGameBtn: {
                                state: "1"
                            },
                            is_PropSceneUI: {
                                state: "0",
                                interval: 5
                            },
                            is_banner: {
                                state: "1",
                                interval: "0"
                            },
                            isaggreation: 5,
                            navigateAppList: [],
                            Offcitys: [],
                            shareList: [],
                            textDelay: 0,
                            Version: "1.0.0",
                            waterfall: "0"
                        }
                    }, {
                        key: PlatType$1.KUAIGAME,
                        value: {
                            BannerAd: "",
                            IntervalAdTime: 0,
                            NativeAd: {
                                HomeNativeAd: 297428,
                                WinNativeAd: 297428,
                                FailNativeAd: 297428
                            },
                            NativeAvd: !1,
                            RewardedVideoAd: "2300000798_01",
                            ad_all_interval: 10,
                            aderrormaxcount: 0,
                            aderrorrate: .2,
                            appid: "ks709035469824743963",
                            is_AddDesk: {
                                state: "0"
                            },
                            is_AvdFloat: {
                                state: "0"
                            },
                            is_AvdList: {
                                state: "0"
                            },
                            is_GameNative: {
                                state: "1",
                                interval: 5
                            },
                            is_MoreGameBtn: {
                                state: "0"
                            },
                            is_PropSceneUI: {
                                state: "0",
                                interval: 5
                            },
                            is_banner: {
                                state: "0",
                                interval: "0"
                            },
                            isaggreation: 5,
                            navigateAppList: [],
                            Offcitys: [],
                            shareList: [],
                            textDelay: 0,
                            version: "1.0.0",
                            waterfall: "0"
                        }
                    }, {
                        key: PlatType$1.WECHAT,
                        value: {
                            BannerAd: "adunit-4a5ef10175e231a8",
                            IntervalAdTime: 0,
                            NativeAd: {
                                HomeNativeAd: 297428,
                                WinNativeAd: 297428,
                                FailNativeAd: 297428
                            },
                            NativeAvd: !1,
                            RewardedVideoAd: "adunit-0ea48866f608bd73",
                            InterstAd: "adunit-80d0929843c6db19",
                            ad_all_interval: 10,
                            aderrormaxcount: 0,
                            aderrorrate: .2,
                            appid: "wxe4821ebb667a1eae",
                            is_AddDesk: {
                                state: "0"
                            },
                            is_AvdFloat: {
                                state: "0"
                            },
                            is_AvdList: {
                                state: "0"
                            },
                            is_GameNative: {
                                state: "0",
                                interval: 5
                            },
                            is_MoreGameBtn: {
                                state: "0"
                            },
                            is_PropSceneUI: {
                                state: "0",
                                interval: 5
                            },
                            is_banner: {
                                state: "1",
                                interval: "0"
                            },
                            isaggreation: 5,
                            navigateAppList: [],
                            Offcitys: [],
                            shareList: [{
                                    title: "带领你的小兵冲锋陷阵!",
                                    iconPath: "https://wxoss.jpsdk.com/ads/202204/98791650796871.jpg"
                                }],
                            textDelay: 0,
                            version: "1.0.2",
                            waterfall: "0",
                            customCfg: [{
                                    id: 'adunit-3683c38faa57e686',
                                    top: 530,
                                    left: 0,
                                    side: 0
                                }, {
                                    id: 'adunit-adfcaa13cf1cb838',
                                    top: 530,
                                    left: -68,
                                    side: 2
                                }, {
                                    id: 'adunit-2c53c2ec61602a9f',
                                    top: 1100,
                                    left: -360,
                                    side: 1
                                }, {
                                    id: 'adunit-3099d0a6b786705d',
                                    top: 280,
                                    left: 0,
                                    side: 0
                                }]
                        }
                    }]
            };
        }
        ;
    }

    class UI_addDesk extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("gamesdk", "addDesk"));
        }
        onConstruct() {
            this.m_loader_icon = (this.getChildAt(0));
        }
    }
    UI_addDesk.URL = "ui://h0rf8xgkcfrget";

    class UI_com_clickgift extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("gamesdk", "com_clickgift"));
        }
        onConstruct() {
            this.m_tip1 = (this.getChildAt(1));
            this.m_tip2 = (this.getChildAt(2));
            this.m_progress = (this.getChildAt(6));
            this.m_chest1 = (this.getChildAt(8));
            this.m_chest2 = (this.getChildAt(9));
            this.m_chest3 = (this.getChildAt(10));
            this.m_chest4 = (this.getChildAt(11));
            this.m_group1 = (this.getChildAt(12));
            this.m_btn_click = (this.getChildAt(13));
        }
    }
    UI_com_clickgift.URL = "ui://h0rf8xgkdt27fr";

    class UI_NormalBanner extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("gamesdk", "NormalBanner"));
        }
        onConstruct() {
            this.m_normalBanner = (this.getChildAt(0));
        }
    }
    UI_NormalBanner.URL = "ui://h0rf8xgkf4q9hy";

    class UI_Native1 extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("gamesdk", "Native1"));
        }
        onConstruct() {
            this.m_RealNativeBanner = (this.getChildAt(0));
        }
    }
    UI_Native1.URL = "ui://h0rf8xgkf4q9hz";

    class UI_NativeAdBanner1 extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("gamesdk", "NativeAdBanner1"));
        }
        onConstruct() {
            this.m_micon = (this.getChildAt(1));
            this.m_mclose = (this.getChildAt(2));
            this.m_mtxt = (this.getChildAt(4));
        }
    }
    UI_NativeAdBanner1.URL = "ui://h0rf8xgkf4q9i0";

    class UI_NativeInsertUI extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("gamesdk", "NativeInsertUI"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m_cover = (this.getChildAt(0));
            this.m_NativeCom = (this.getChildAt(1));
            this.m_btnclose = (this.getChildAt(2));
        }
    }
    UI_NativeInsertUI.URL = "ui://h0rf8xgkf4q9ia";

    class UI_Native$2 extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("gamesdk", "Native"));
        }
        onConstruct() {
            this.m_NativeBannerCom = (this.getChildAt(0));
        }
    }
    UI_Native$2.URL = "ui://h0rf8xgkf4q9ib";

    class UI_NativeAdBanner$2 extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("gamesdk", "NativeAdBanner"));
        }
        onConstruct() {
            this.m_Aicon = (this.getChildAt(1));
            this.m_Bbiaoti = (this.getChildAt(4));
            this.m_qukankan = (this.getChildAt(5));
            this.m_closebtn = (this.getChildAt(6));
            this.m_resoure = (this.getChildAt(7));
        }
    }
    UI_NativeAdBanner$2.URL = "ui://h0rf8xgkf4q9ic";

    class UI_comp_douyinhudao extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("gamesdk", "comp_douyinhudao"));
        }
        onConstruct() {
            this.m_graph_bg = (this.getChildAt(0));
            this.m_btn_zuire = (this.getChildAt(1));
            this.m_btn_zuixin = (this.getChildAt(2));
            this.m_group_top = (this.getChildAt(3));
            this.m_btn_close = (this.getChildAt(5));
            this.m_list_icon = (this.getChildAt(6));
        }
    }
    UI_comp_douyinhudao.URL = "ui://h0rf8xgkghtlg6";

    class UI_btn_icon extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("gamesdk", "btn_icon"));
        }
        onConstruct() {
            this.m_loader_icon = (this.getChildAt(1));
        }
    }
    UI_btn_icon.URL = "ui://h0rf8xgkghtlhg";

    class UI_btn_small extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("gamesdk", "btn_small"));
        }
        onConstruct() {
            this.m_laoder_bg = (this.getChildAt(0));
            this.m_laoder_type = (this.getChildAt(1));
            this.m_text_name = (this.getChildAt(2));
            this.m_comp_icon = (this.getChildAt(3));
        }
    }
    UI_btn_small.URL = "ui://h0rf8xgkghtlhh";

    class UI_comp_iconzhanshi extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("gamesdk", "comp_iconzhanshi"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m_btn_icon = (this.getChildAt(0));
            this.m_text_name = (this.getChildAt(1));
        }
    }
    UI_comp_iconzhanshi.URL = "ui://h0rf8xgkghtlhi";

    class UI_FloatAd extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("gamesdk", "FloatAd"));
        }
        onConstruct() {
            this.m_name = (this.getChildAt(1));
            this.m_icon = (this.getChildAt(2));
            this.m_t0 = this.getTransitionAt(0);
        }
    }
    UI_FloatAd.URL = "ui://h0rf8xgks5ol2g";

    class UI_Icon extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("gamesdk", "Icon"));
        }
        onConstruct() {
            this.m_img = (this.getChildAt(1));
        }
    }
    UI_Icon.URL = "ui://h0rf8xgks5ol2h";

    class UI_moreBtn extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("gamesdk", "moreBtn"));
        }
        onConstruct() {
            this.m_loader_icon = (this.getChildAt(0));
        }
    }
    UI_moreBtn.URL = "ui://h0rf8xgks5ol33";

    class gamesdkBinder {
        static bindAll() {
            fgui.UIObjectFactory.setExtension(UI_addDesk.URL, UI_addDesk);
            fgui.UIObjectFactory.setExtension(UI_com_clickgift.URL, UI_com_clickgift);
            fgui.UIObjectFactory.setExtension(UI_NormalBanner.URL, UI_NormalBanner);
            fgui.UIObjectFactory.setExtension(UI_Native1.URL, UI_Native1);
            fgui.UIObjectFactory.setExtension(UI_NativeAdBanner1.URL, UI_NativeAdBanner1);
            fgui.UIObjectFactory.setExtension(UI_NativeInsertUI.URL, UI_NativeInsertUI);
            fgui.UIObjectFactory.setExtension(UI_Native$2.URL, UI_Native$2);
            fgui.UIObjectFactory.setExtension(UI_NativeAdBanner$2.URL, UI_NativeAdBanner$2);
            fgui.UIObjectFactory.setExtension(UI_comp_douyinhudao.URL, UI_comp_douyinhudao);
            fgui.UIObjectFactory.setExtension(UI_btn_icon.URL, UI_btn_icon);
            fgui.UIObjectFactory.setExtension(UI_btn_small.URL, UI_btn_small);
            fgui.UIObjectFactory.setExtension(UI_comp_iconzhanshi.URL, UI_comp_iconzhanshi);
            fgui.UIObjectFactory.setExtension(UI_FloatAd.URL, UI_FloatAd);
            fgui.UIObjectFactory.setExtension(UI_Icon.URL, UI_Icon);
            fgui.UIObjectFactory.setExtension(UI_moreBtn.URL, UI_moreBtn);
        }
    }

    let UINativeInsert = class UINativeInsert extends BaseUI {
        constructor() {
            super(...arguments);
            this.m_adid = "";
        }
        get URL() {
            gamesdkBinder.bindAll();
            return UI_NativeInsertUI.URL;
        }
        OnCreate() {
        }
        OnShow(...args) {
            this.ui.m_NativeCom.visible = true;
            this.ui.m_btnclose.visible = true;
            console.log("显示广告");
            let nativeInfo = args[0];
            let panelData = args[1];
            let rightPos = new Laya.Vector2(455, 25);
            let leftPos = new Laya.Vector2(22, 25);
            let tempPos = new Laya.Vector2();
            tempPos = panelData.closePos == 1 ? rightPos : leftPos;
            this.ui.m_c1.selectedIndex = panelData.panelPos;
            console.log(args, "参数数组");
            if (args[2]) {
                this.ui.m_cover.visible = true;
                console.log("存在第二个参数=======================》");
            }
            else {
                this.ui.m_cover.visible = false;
            }
            this.ui.m_NativeCom.m_NativeBannerCom.m_closebtn.x = tempPos.x;
            this.ui.m_NativeCom.m_NativeBannerCom.m_closebtn.y = tempPos.y;
            console.log("test----- nativeInfo", JSON.stringify(nativeInfo));
            console.log("test----- m_Aicon" + this.ui.m_NativeCom.m_NativeBannerCom.m_Aicon);
            console.log("test----- m_Bbiaoti" + this.ui.m_NativeCom.m_NativeBannerCom.m_Bbiaoti);
            let Rurl;
            if (nativeInfo.iconUrlList && nativeInfo.iconUrlList.length > 0) {
                Rurl = nativeInfo.iconUrlList[0];
            }
            else if (nativeInfo.imgUrlList && nativeInfo.imgUrlList.length > 0) {
                Rurl = nativeInfo.imgUrlList[0];
            }
            this.ui.m_NativeCom.m_NativeBannerCom.m_resoure.text = nativeInfo.source;
            this.ui.m_NativeCom.m_NativeBannerCom.m_Aicon.url = Rurl;
            this.ui.m_NativeCom.m_NativeBannerCom.m_Bbiaoti.text = nativeInfo.title;
            this.m_adid = nativeInfo.adId;
            GameSDK$1.m_adId = this.m_adid;
            GameSDK$1.GameNativeInsert.reportAdShow(this.m_adid);
            this.ui.m_NativeCom.m_NativeBannerCom.m_Aicon.onClick(this, () => {
                console.log("test-------m_adid--" + this.m_adid);
                UIManager.inst.PopUI(this.layer, false);
                GameSDK$1.GameNativeInsert.reportAdClick(this.m_adid);
            });
            this.ui.m_NativeCom.m_NativeBannerCom.m_closebtn.onClick(this, () => {
                console.log(" 真 延迟  关闭按钮 111");
                UIManager.inst.PopUI(this.layer, false);
            });
        }
        reportShow() {
            GameSDK$1.GameNativeInsert.reportAdShow(this.m_adid);
        }
        OnHide() {
            this.ui.m_NativeCom.visible = false;
            this.ui.m_btnclose.visible = false;
        }
        OnDispose() {
        }
        onSwitch() {
        }
        onAniEnd() {
        }
        updateScene(offset = 1) {
        }
    };
    UINativeInsert = __decorate([
        ui_register()
    ], UINativeInsert);

    class GameNativeInsert {
        constructor(nativeId) {
            this.showCover = false;
            console.log(" GameNativeData------------ ");
            this.createNative(nativeId);
        }
        setIsShow(show) {
            this.isShow = show;
            if (!this.isShow) {
                console.log("关闭界面成功");
                UIManager.inst.PopUI(UILayer.insertAd, false);
            }
        }
        load(panelData, showCover) {
            console.log("加载插屏");
            this.setIsShow(true);
            this.panelData = panelData;
            this.showCover = showCover;
            if (this._nativeNode)
                this._nativeNode.load();
        }
        createNative(nativeId) {
            console.log("GameNativeData createNative-@@---------- ");
            this.nativeId = nativeId;
            this._nativeNode = GameSDK$1.createNative(this.nativeId);
            if (!this._nativeNode)
                return;
            this._nativeNode.onError((e) => {
                console.log("=========>oppo 原生广告加载失败", e);
                this.setIsShow(false);
            });
            this._nativeNode.onLoad((res) => {
                console.log("=========>oppo 原生广告加载成功", res);
                if (res && res.adList) {
                    let nativeInfo = res.adList[res.adList.length - 1];
                    console.log(" GameNativeData 原生广告数据：", nativeInfo);
                    if (this.isShow) {
                        if (this.showCover) {
                            UIManager.inst.PushUI(UINativeInsert, UILayer.insertAd, nativeInfo, this.panelData, true);
                        }
                        else {
                            UIManager.inst.PushUI(UINativeInsert, UILayer.insertAd, nativeInfo, this.panelData);
                        }
                    }
                }
            });
        }
        reportAdClick(_useNtAdId) {
            if (_useNtAdId != null && this._nativeNode)
                this._nativeNode.reportAdClick({ adId: _useNtAdId.toString() });
        }
        reportAdShow(_useNtAdId) {
            if (_useNtAdId != null && this._nativeNode) ;
            this._nativeNode.reportAdShow({ adId: _useNtAdId.toString() });
        }
        reportAdShowNew() {
            if (this.isShow) {
                this.reportAdShow(GameSDK$1.m_adId);
            }
        }
        removeView() {
        }
    }

    class GameNativeBanner {
        constructor(nativeId) {
            console.log(" GameNativeData2------------ ");
            this.createNative(nativeId);
        }
        load() {
            if (this._nativeNode)
                this._nativeNode.load();
        }
        createNative(nativeId) {
            console.log("GameNativeData2 createNative-@@---------- ");
            this.nativeId = nativeId;
            this._nativeNode = GameSDK$1.createNative(this.nativeId);
            if (!this._nativeNode)
                return;
            this._nativeNode.onError(function (e) {
                console.log("GameNativeData2=========>oppo 原生广告加载失败", e);
            });
            this._nativeNode.onLoad(function (res) {
                console.log("=========>oppo 原生广告加载成功", res);
                if (res && res.adList) {
                    let nativeInfo = res.adList[res.adList.length - 1];
                    console.log("GameNativeData2 原生广告数据：", nativeInfo);
                    if (GameSDK$1.NativeDataArr2.length > 3) {
                        GameSDK$1.NativeDataArr2.shift();
                        GameSDK$1.NativeDataArr2.push(nativeInfo);
                    }
                    else {
                        GameSDK$1.NativeDataArr2.push(nativeInfo);
                    }
                }
            });
        }
        reportAdClick(_useNtAdId) {
            if (_useNtAdId != null && this._nativeNode)
                this._nativeNode.reportAdClick({ adId: _useNtAdId.toString() });
        }
        reportAdShow(_useNtAdId) {
            if (_useNtAdId != null && this._nativeNode)
                this._nativeNode.reportAdShow({ adId: _useNtAdId.toString() });
        }
        removeView() {
        }
    }

    class UI_ui_login extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "ui_login"));
        }
        onConstruct() {
            this.m_bg = (this.getChildAt(0));
            this.m_content = (this.getChildAt(1));
        }
    }
    UI_ui_login.URL = "ui://ixnra5xgjycsxu7";

    let LoginUI = class LoginUI extends BaseUI {
        get URL() {
            return UI_ui_login.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_bg];
        }
        OnCreate() {
            this.ui.m_content.m_btn_tuichu.onClick(this, this.onClickClose);
            this.ui.m_content.m_btn_login.onClick(this, this.onLogin);
        }
        OnShow(...args) {
            this.ui.m_content.m_descripe.text = args[0];
        }
        OnHide() {
        }
        OnDispose() {
        }
        loginHuaWei() {
        }
        onLogin() {
            GameSDK$1.huaWeiLogin(() => {
            }, () => {
            });
        }
        onClickClose() {
            qg.exitApplication({});
            UIManager.inst.PopUI(this.layer, false);
        }
    };
    LoginUI = __decorate([
        ui_register()
    ], LoginUI);

    class UI_ui_yinsi1 extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("yinsizhengce", "ui_yinsi1"));
        }
        onConstruct() {
            this.m_com_mengban = (this.getChildAt(0));
            this.m_comp_yinsi1 = (this.getChildAt(1));
        }
    }
    UI_ui_yinsi1.URL = "ui://jyvm79orfkj20";

    class UI_ui_yinsi2 extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("yinsizhengce", "ui_yinsi2"));
        }
        onConstruct() {
            this.m_com_mengban = (this.getChildAt(0));
            this.m_comp_yinsi2 = (this.getChildAt(1));
        }
    }
    UI_ui_yinsi2.URL = "ui://jyvm79orfkj21";

    class UI_comp_yinsi1 extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("yinsizhengce", "comp_yinsi1"));
        }
        onConstruct() {
            this.m_btn_close = (this.getChildAt(3));
            this.m_btn_agree = (this.getChildAt(5));
            this.m_btn_reject = (this.getChildAt(6));
            this.m_btn_yonghuxieyi = (this.getChildAt(7));
            this.m_btn_chanpinyinsishuoming = (this.getChildAt(8));
        }
    }
    UI_comp_yinsi1.URL = "ui://jyvm79orfkj211";

    class UI_comp_yinsi2 extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("yinsizhengce", "comp_yinsi2"));
        }
        onConstruct() {
            this.m_btn_agree = (this.getChildAt(4));
            this.m_btn_reject = (this.getChildAt(5));
            this.m_btn_fuwuxieyi = (this.getChildAt(6));
            this.m_btn_yinsizhengce = (this.getChildAt(7));
        }
    }
    UI_comp_yinsi2.URL = "ui://jyvm79orfkj212";

    class UI_ui_wendang extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("yinsizhengce", "ui_wendang"));
        }
        onConstruct() {
            this.m_bg = (this.getChildAt(0));
            this.m_loader_biaoti = (this.getChildAt(1));
            this.m_btn_fanhui = (this.getChildAt(2));
            this.m_wenAn = (this.getChildAt(4));
        }
    }
    UI_ui_wendang.URL = "ui://jyvm79orfkj2z";

    class UI_Component1 extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("yinsizhengce", "Component1"));
        }
        onConstruct() {
            this.m_text_wenan = (this.getChildAt(0));
        }
    }
    UI_Component1.URL = "ui://jyvm79orumxf14";

    class yinsizhengceBinder {
        static bindAll() {
            fgui.UIObjectFactory.setExtension(UI_ui_yinsi1.URL, UI_ui_yinsi1);
            fgui.UIObjectFactory.setExtension(UI_ui_yinsi2.URL, UI_ui_yinsi2);
            fgui.UIObjectFactory.setExtension(UI_comp_yinsi1.URL, UI_comp_yinsi1);
            fgui.UIObjectFactory.setExtension(UI_comp_yinsi2.URL, UI_comp_yinsi2);
            fgui.UIObjectFactory.setExtension(UI_ui_wendang.URL, UI_ui_wendang);
            fgui.UIObjectFactory.setExtension(UI_Component1.URL, UI_Component1);
        }
    }

    let UIwendang = class UIwendang extends BaseUI {
        constructor() {
            super(...arguments);
            this.text_yinsi = "我司隐私政策\n前言\n生效时间：2021年11月16日\n更新时间：2022年1月10日\n我司产品与/或服务是由我司（以下简称\“我们\”）为用户（以下简称\“您\”）提供的休闲娱乐产品与/或服务，包括但不限于\"我司\"应用，详见关键词定义），以及我们不时为您提供的其他产品与/或服务。我们根据《中华人民共和国网络安全法》相关法律法规并参考《信息安全技术个人信息安全规范》（GB/T-35273），以及其他相关法律法规制定《我司隐私政策》（以下简称\“本政策\”）并保护您的个人信息安全。所有我司产品与或服务均适用本政策。我司产品与/或服务的适用对象为\“通过平台验证/授权的用户\”。对于儿童，必须由其监护人阅读本政策并同意本政策的全部内容后，我们方才向该儿童提供我司产品与/或服务。本政策介绍了我们向您提供我司产品与/或服务过程中收集、使用、存储、委托处理、共享、转让、公开批露您个人信息的目的、方式、范围和信息安全保护措施，以及我们为您提供的访问、更新、删除等控制个人信息的权利和方法。请您结合我们的《我司用户服务协议》阅读本政策。以下是我们的特别提示，请您重点关注：\n1.对所有重点内容，我们采用\“加粗\”的书写方式进行提示。\n2.对个人敏感信息，我们用\“加粗斜体\” 的书写方式进行特别提醒，希望您在阅读时特别关注。在您向我们提供您的任何个人敏感信息前，您已慎重考虑并在此确认该等提供是适当的，您在此同意您个人敏感信息可按照本政策所述的目的和方式进行处理。\n3.我们尤其重视对个人信息的保护。我们将遵循正当必要、知情同意、目的明确、安全保障、依法利用的原则，收集、使用、存储、委托处理、共享、转让、公开披露您的个人信息。根据中国法律法规，您确认并承诺您是使用我司产品与/ 或服务的儿童的监护人，或者已满 18 周岁的成年用户。\n4.我们会严格依据本政策以及您的同意处理您的个人信息。如您不同意本政策，您应当停止使用我司产品与 / 或服务。\n目录\n1.我们会获取哪些信息，作用用途说明\n2.如何使用 Cookie\n3.如何使用第三方 SDK 数据\n4.关于数据的跨境传输\n5.如何委托处理、共享、转让、公开披露您的个人信息\n6.您如何管理您的个人信息\n7.如何保护您的个人信息\n8.如何存储您的个人信息\n9.如何更新本政策\n10.如何联系我们\n11.附则\n12.关键词定义\n1. 我们会获取哪些信息，作用用途说明\n我们会根据合法正当、必要、透明的原则，基于本政策所述的目的，收集和使用您的个人信息。如果我们将您的个人信息用于本政策未载明的其它用途，或基于其他特定目的而收集您的个人信息，我们将以合理的方式告知您，并在使用前再次征得您的同意。\n1.1向您提供产品与/或服务过程中我们主动收集的个人信息\n在符合法律法规前提下，为了更好地为您提供优质的服务，我们会收集和使用您在使用我司产品与/或服务过程中产生的如下信息：\n1.1.1设备信息\n当您使用\"我司\"应用或应用切换到后台时，我们会收集您的设备信息，包括浏览器类型、运营商信息和使用情况、设备的相关信息（IP地址）、设备型号、操作系统版本号、AndoridID（仅限安卓）。这些信息是设备的唯一标识，我们可能会将这些信息进行关联，以便我们能够为您提供一致的服务。为免疑义，我们可能采集您设备的Mac地址、IMEI号上传到我们的服务器，仅用于产品BUG的实时排查和定位，监控产品稳定性。\n当您使用我们的软件时，我们可能会根据您在软件安装及使用中的具体权限，接收并记录您的相关设备信息，包括设备型号、唯一设备标识符、操作系统信息，以便我们为您提供一致的服务。\n1.1.2位置信息\n当您使用\"我司\"应用时，我们使用的友盟sdk会自动收集您在使用\"我司\"应用时的位置信息，友盟sdk收集位置信息，为开发者提供反舞弊功能，剔除作弊设备；同时校正用户的地域分布数据，使报表数据更加准确。\n1.1.3日志信息\n当您使用\"我司\"应用时，我们会自动收集您在使用\"我司\"应用时的详细情况并作为网络日志进行保存，包括浏览器类型、浏览记录、网络诊断、卡顿信息、点击事件、关卡成功或失败情况，以便我们统计\"我司\"应用的使用情况，优化\"我司\"应用，为您提供更好的服务体验。\n当您使用我们的软件时，我们会自动收集您使用的详细情况，并作为网络日志进行保存，包括您IP地址、浏览器类型、浏览记录，以便我们统计官网的使用情况，为您提供更好的体验和服务。\n1.1.4设备权限调用汇总\n我们在提供服务过程中，会调用您的主要设备权限汇总如下：\n权限名\n权限功能说明\n使用场景或目的\nINTERNET\n访问网络的权限\n请求广告,上报应用错误log时需要连接网络\nACCESS_NETWORK_STATE\n获取网络信息状态\n请求配置和广告时需要判断网络链接是否有效\nREAD_LOGS\n读取系统日志文件\nAPP出现错误或异常退出时需要上传日志\nACCESS_WIFI_STATE\n获取 WIFI 状态\n在下载应用、播放音频时，我们需要访问您的WIFI状态，我们将根据是否连接WiFi，来确定本应用、音频等资源加载方式，以优化您的使用体验。\nREAD_PHONE_STATE\n读取手机状态和身份：提供读取手机设备标识等信息\n读取设备通话状态和识别码，识别手机设备ID，用于信息展示、账号登录、安全保障等主要功能\nREAD_EXTERNAL_STORAGE\n读取SD上的内容：提供读取手机储存空间内数据的功能\n读取已下载的本应用文件,读取应用储存数据\nWRITE_EXTERNAL_STORAGE\n修改/删除SD卡中的内容：提供写入外部储存功能\n保存存储下载的本应用文件,保存应用使用数据\nREQUEST_INSTALL_PACKAGES\n允许安装第三方应用\n广告sdk会在用户同意后安装第三方app\nWAKE_LOCK\n屏幕常亮\n在播放激励视频广告时需要保持屏幕常亮，以免黑屏\nQUERY_ALL_PACKAGES\n读取已安装应用列表\n当用户点击广告后，sdk需要读取已安装应用列表判断是安装应用还是打开应用。\n详细内容请访问《权限使用说明》。您可以在设备的设置功能中选择关闭部分或全部权限。在不同设备中，权限显示方式及关闭方式可能有所不同，若您是安卓机型，您需要在手机【设置】-【应用设置】-【应用权限管理】中，选择本应用，并关闭您需要撤回的权限/信息。具体请参考设备及系统开发方说明或指引。\n1.2我们从第三方获取的您的个人信息\n1.2.1我们可能会在您的同意范围内从第三方（我们的合作方）处收集并使用您的个人信息。我们保证依照与第三方签订的合同以及相关法律规定处理您的个人信息，同时请您详细阅读该第三方的隐私政策/个人信息保护政策及用户协议。如您拒绝第三方在提供服务时收集、使用或者传递您的个人信息，将可能导致您无法使用我司产品与/或服务。\n1.2.2第三方登录：当您使用第三方账号（包括微信、QQ、华为、OPPO账号、vivo账号等其他渠道）登录时，可选择授权我们读取并获得您在该第三方平台上公开的信息。当您使用微信账号等第三方账号登录时，第三方账号将与我们共享您的信息详情见此处。获取上述信息是为了记住您作为我司用户的登录身份，以便向您提供更优质的产品与/或服务，并且我们仅会在您同意的范围内使用此类信息。当您使用第三方账号注册或登录后，您的第三方账号会与我司账号进行绑定。\n1.3征得同意的例外\n根据法律法规，以下情形中收集和使用您的个人信息无需征得您的同意：\n1.3.1与履行法律法规规定的义务相关的；\n1.3.2与国家安全、国防安全直接相关的；\n1.3.3与公共安全、公共卫生、重大公共利益直接相关的；\n1.3.4与犯罪侦查、起诉、审判和判决执行等直接相关的；\n1.3.5出于维护您或其他用户的生命、财产等重大合法权益相关但又很难得到本人同意的；\n1.3.6所涉及的个人信息是您自行向社会公众公开的；\n1.3.7根据您要求签订和履行合同所必需的；\n1.3.8从合法公开披露的信息中收集个人信息的，如合法的新闻报道、政府信息公开等渠道；\n1.3.9维护所提供的我司产品与/或服务的安全稳定运行所必需的，例如发现、处置我司产品与/或服务的故障；\n1.3.10法律法规规定的其他情形。\n2. 如何使用 Cookie\n2.1Cookie\n当您使用我司产品与/或服务时，我们会在您的计算机或移动设备上存储名为Cookie的小型数据文件，通常包含标识符、站点名称以及一些号码和字符。当您登录账号时，借助于Cookie，我们可以在您和使用我司产品与/或服务时识别您是否属于我们的用户；还可以实现直接从一个页面跳转到另一个页面，而不必在每个页面上重新登录。\n2.2如何使用Cookie\n我们使用该等信息判断用户是否已经登录、提升服务/产品质量及优化用户体验。我们不会将Cookie用于本政策所述目的之外的任何其他用途。\n2.3如何管理追踪技术\n在网页端，我们不会将上述追踪技术用于本政策所述目的之外的任何用途。您可根据自己的偏好管理或删除某些类别的追踪技术。有关详情，请参见https://aboutcookies.org/。另外，很多网络浏览器均设有\“禁止追踪\”（DoNotTrack）功能，该功能可向网站发布\“禁止追踪\”的请求。如果您的浏览器启用了\“禁止追踪\”功能，我们的所有网站都会尊重您的选择。如需详细了解如何更改浏览器设置，请访问相关链接。\n当您在非登陆状态下浏览页面时，为向您提供相应服务，我们会收集实现浏览功能所必需收集Cookie信息，但会在一个月内删除这些Cookie信息。\n请您知悉，倘若您拒绝使用或清除已有的追踪技术，则需要在每一次访问时亲自更改用户设置，并且，我们可能无法向您提供更加优质的使用体验，部分服务亦可能会无法正常运作。\n3. 如何使用第三方 SDK 数据\n3.1 我们的应用集成BuglySDK，为了更真实地为开发者还原Crash场景服务的，并不涉及用户隐私信息：\nCrash环境：Crash信息及线程堆栈，ROM/RAM/SD卡容量、网络/语言等状态。\nApp信息：包名、版本、所属进程名。\n设备信息：IMEI等设备识别，用于判断Crash设备统计。\n3.2 我们的应用集成第三方广告平台SDK，需要收集您的设备Mac地址、唯一设备识别码（IMEI/androidID/IDFA/OPENUDID/GUID、SIM卡IMSI信息）以提供统计分析服务。我们会对合作方获取信息的软件工具开发包（SDK）进行严格的安全检测，以保护数据安全。\n3.3 为方便阅读，其他更多SDK数据请查看接入的相关第三方SDK。\n4. 关于数据的跨境传输\n我们的产品不存在跨境传输的数据，且所有数据在传输前，均会事先获得您的明示授权或同意。\n5. 如何委托处理、共享、转让、公开披露您的个人信息\n5.1委托处理\n为了向您提供更完善、优质的产品和服务，某些功能可能由我们的技术服务合作方提供，我们可能会委托合作方处理您的某些个人信息。对我们委托处理您个人信息的公司、组织和个人，我们会与其签署严格的保密协定，明确双方责任、处理事项、处理期限、处理性质和处理目的等，要求他们仅按照我们的要求、本政策以及其他任何相关的保密和安全措施来处理个人信息。如果您拒绝我们的合作方在提供服务时收集为提供服务所必需的个人信息，将可能导致您无法使用我司产品与/或服务来享用该合作方提供的服务。我们的合作伙伴包括：\n5.1.1分析与/或统计服务SDK提供商。为了更好地分析与/或统计我司用户的使用情况，我们可能委托数据分析与/或统计的服务提供商处理您的设备信息、用户点击记录，用于维护或改进我司产品与/或服务，并提升用户体验。\n5.1.2技术服务SDK提供商。我们可能会委托该服务的合作伙伴（如语音技术服务提供商、云服务提供商）处理您的语音，以便提供发音准确性测评，更好地提升用户体验。我们亦可能委托该服务的合作伙伴（如应用异常情况技术服务提供商）处理您系统版本、生成ID、手机样式，以便提供\"我司\"应用的异常情况分析处理。\n5.2共享\n对于您的个人信息，我们仅在本政策所述的目的和范围内或根据法律法规的要求收集和使用，并严格保密。除非存在以下一种或多种情形（相关法域另有规定的除外），我们不会与第三方公司、组织或个人共享您的个人信息：\n5.2.1基于您自行提出的要求、或事先获得您的明示授权或同意；\n5.2.2与我们履行法律法规规定的义务相关的；\n5.2.3与国家安全、国防安全直接相关的；\n5.2.4与公共安全、公共卫生、重大公共利益直接相关的；\n5.2.5与犯罪侦查、起诉、审判和判决执行等直接相关的；\n5.2.6出于维护您和/或您孩子或其他个人的生命、财产等重大合法权益但又很难得到本人同意的；\n5.2.7您自行向社会公众公开的您和/或您孩子个人信息；\n5.2.8从合法公开披露的信息中收集个人信息的，如合法的新闻报道、政府信息公开等渠道。\n5.2.9与我们的业务合作伙伴进行必要的信息共享；\n为了向您提供更完善、优质的产品和服务，某些我司服务将由合作方提供，因此，我们可能会与合作方共享您的某些个人信息。我们仅会出于合法、正当、必要、特定、明确的目的共享的个人信息，并且只会共享为您提供服务所必要的个人信息。我们不会共享可以识别您身份的个人信息，除非法律法规另有规定。该等共享的合作方为数据控制者，以其自己的名义获得您的同意以处理您的个人信息。合作方可能有其独立的隐私政策和用户协议，我们建议您认真阅读并遵守第三方的用户协议及隐私政策。如果您拒绝我们的合作方在提供服务时收集为提供服务所必需的个人信息，将可能导致您无法使用我司产品与/或服务来享用该合作方提供的服务。在您购买我司产品与/或服务时，我们可能会与我们的合作伙伴（含渠道合作方）共享您的信息，例如，我们向支付宝提供交易金额、交易商品、订单号、交易时间，以便您能够完成支付并顺利购买。\n请您知悉，随着服务的更新和升级，您可能通过使用我们的服务或第三方提供的功能，主动公开、共享您的相关信息（如微信好友/朋友圈/QQ等）。在这种情况下，共享信息将被视为即时并广泛的传播。一旦您进行了此类信息共享，很难确保当您需要删除信息时，您所共享的信息在所有接受转发的互联网平台中被全部删除。在一些情况下，即使您从我们的服务中删除了您共享的信息，该类信息仍然可能会留存在公众可获知的领域，仍可由其他用户或与我们没有关联且不受我们控制的第三方独立地缓存、复制或存储，以及通过其他用户或该等第三方在公众可获知的领域保存。\n5.2.10其他法律法规规定的情形。\n请您知悉：\n（1）我们共享您的个人信息仅为了更精准地向您提供服务。除此以外，我们不会共享可以识别您身份的个人信息，除非法律法规另有规定。\n（2）通过我司产品与/或服务或第三方提供的功能，您可以主动公开、共享您的相关状态（如微信转发）。在这种情况下，共享信息将被视为即实并广泛的传播。一旦您进行了此等共享，很难确保当您需要删除信息时，您所共享的信息在所有接受转发的互联网平台中被全部删除。在一些情况下，即使您从我司产品与/或服务中删除了被共享的信息，该等信息仍然可能会留存在公众可获知的领域，仍可由其他用户或与我们没有关联及不受我们控制的第三方独立地缓存、复制或存储，或通过其他用户或该等第三方在公众可获知的领域保存。\n5.3转让\n原则上我们不会将您的个人信息控制权向其它公司、组织或个人进行转让，但在以下情况除外：\n5.3.1基于您自行提出的要求、或事先获得您的明示授权或同意；\n5.3.2与我们履行法律法规规定的义务相关的；\n5.3.3与国家安全、国防安全直接相关的；\n5.3.4与公共安全、公共卫生、重大公共利益直接相关的；\n5.3.5与犯罪侦查、起诉、审判和判决执行等直接相关的；\n5.3.6出于维护您和/或您孩子或其他个人的生命、财产等重大合法权益但又很难得到本人同意的；\n5.3.7您自行向社会公众公开的个人信息；\n5.3.8从合法公开披露的信息中收集个人信息的，如合法的新闻报道、政府信息公开等渠道；\n5.3.9发生公司合并、收购或破产清算，运营主体的变更时的转让。\n请您知悉：如具备上述事由确需转让，我们将在转让前向您告知转让信息的目的、类型、信息受让方（如涉及敏感信息，我们还将告知涉及的敏感信息的内容），并在征得您同意后再转让，但法律法规另有规定的除外。\n5.4公开披露\n原则上我们不会向社会或不特定人群发布您的个人信息，但以下情况除外：\n5.4.1基于与您的约定的；\n5.4.2根据法律法规应当披露的。\n请您知悉：\n（1）如具备上述事由确需公开披露，我们会在披露前向您告知公开披露的目的、信息的类型（如涉及敏感信息，我们还将告知涉及的敏感信息的内容）并在征得您同意后再公开披露，但法律法规另有规定的除外。\n（2）即使已经取得您的同意，我们也仅会处于合法、正当、必要、特定、明确的目的公开披露您的个人信息，并尽量对公开披露内容中的个人信息进行去标识化处理。\n6. 您如何管理您的个人信息\n6.1访问或更正您的个人信息\n6.1.1您有权随时登录并访问您的账户，查看您个人信息，包括您的账号信息和您的个人档案。如您发现我们收集、存储、使用、披露的您的个人信息有错误的，或者在法律法规规定的其他情况下，您可以更正您的个人信息。\n6.1.2如您希望访问或更正您其他个人信息，或在行使上述权利过程中遇到困难，您可通过本政策第10条提供的联系方式联系我们，我们将在核实您的身份后15日内及时响应您的请求，但法律法规另有规定的或本政策另有约定的除外。\n6.2删除您的个人信息\n6.2.1在以下情形中，您可以向我们提出删除您个人信息的请求，但已做数据匿名化处理或法律法规另有规定的除外：\n6.2.1.1我们违反法律、行政法规的规定或与您的约定，收集、存储、使用、转移、披露您的个人信息的；\n6.2.1.2我们超出目的范围或者必要期限收集、存储、使用、转移、披露您个人信息的；\n6.2.1.3您撤回同意的；\n6.2.1.4您通过注销等方式终止使用我司产品与/或服务的。\n6.2.1.5其他法律法规等规定的情形。\n6.2.2如您希望删除您的个人信息，您可以您可通过本政策第10条提供的联系方式联系我们。对于您的请求，我们原则上将于收到您的请求并在验证您用户身份后的15日内做出答复。\n6.2.3您在此同意，一旦删除请求被响应，您的个人信息将被删除。\n6.2.4请您知悉，我们可能不会将响应的信息立即从备份系统中删除，但会在备份系统更新时进行删除。\n6.3更改您授权同意的范围\n6.3.1请您知悉，为了您能正常使用我司产品与/或服务，您需要向我们提供一些基本的个人信息。如果您希望更改或撤回您授权同意，您可以在设备的设置功能中选择关闭部分或全部权限。在不同设备中，权限显示方式及关闭方式可能有所不同，具体请参考设备及系统开发方说明或指引。您也可以通过本政策第10条提供的联系方式联系我们，对于您的请求，我们原则上将于收到您的请求并在验证您用户身份后的15日内做出答复。\n6.3.2当您更改您的同意范围后，我们将不再收集和使用您相应的个人信息，但请您理解，在您更改您的同意范围后，我们无法再为您提供与之对应的我司产品与/或服务。您在此同意，您更改同意范围或撤回同意的行为不会影响我们基于您之前授权对您个人信息的处理和存储。\n6.4注销您的账号\n6.4.1除非法律法规另有规定或本政策另有约定，您也可以通过本政策第10条提供的联系方式联系我们来注销账号，对于您的请求，我们原则上将于收到您的请求并在验证您用户身份后的15日内做出答复。\n6.4.2请您知悉，在注销账号后，除非根据法律法规或根据监管部门的要求保留或存储您的个人信息，否则我们将即时删除您的个人信息或作匿名化处理。请您知悉，我们可能不会将该等信息立即从备份系统中删除，但会在备份系统更新时进行删除。\n6.4.3您注销账号的行为是不可逆的，一旦您的账号被注销，我们将不再收集您的个人信息，也不再为您提供我司产品与/或服务。因此我们建议您谨慎选择注销账号。\n6.5获取您的个人信息副本\n6.5.1在技术措施允许的情况下，我们也可以根据您的要求，将您的个人信息提供给第三方。\n6.6响应您的上述请求\n6.6.1为实现本政策第6.1-6.6条所述的用户权利，请您通过Email: tangyongqin@newto.cn与我们联系。为确保安全，我们可能会先要求您通过提交相应证明以验证自己的身份，然后再处理您的请求。对于您的请求，我们原则上将于收到您的请求并在验证您用户身份后的15日内做出答复。\n6.6.2对于您合理的请求，我们原则上不收取费用。对于与您的身份不直接关联的信息、无端重复信息，或者需要过多技术手段（例如，需要开发新系统或从根本上改变现行惯例）、给他人合法权益带来风险或者不切实际的请求，我们可能会予以拒绝。\n6.6.3如您对您的数据主体权利存在任何疑问、意见或建议的，也可通过Email: tangyongqin@newto.cn联系我们。\n7. 如何保护您的个人信息\n我们十分重视您的个人信息安全，为此我们采用了符合行业标准的安全技术措施及配套的组织架构和管理体系等多层面保护措施，以最大程度防止您的个人信息被泄露、毁损、误用、未授权访问、未授权披露或更改，具体包括：\n7.1数据安全技术措施\n7.1.1在数据安全传输方面，我们采用了传输层安全协议等密码技术，通过Https等方式防止传输链路被窃听、截取，建立安全的隐私数据环境，保证数据的私密性和完整性。我们还对您的信息采取加密存储等额外的安全保护措施。\n7.1.2在数据访问和使用的安全控制方面，我们实施了严格的数据权限控制机制，采取多重身份认证技术，并对能够处理您的个人信息的行为进行监控，避免数据被违规访问和未授权使用。我们还对数据生命周期的全流程进行监控，防止您的个人信息遭遇未授权的访问、公开披露、使用、修改、人为或意外的损坏或丢失等。\n7.2数据安全组织和管理措施\n7.2.1根据有关法规要求，我们成立了专门负责个人信息保护的部门。我们还建立了相关的内控管理流程，对可能接触到您个人信息的工作人员以最小授权为原则，严格设定信息访问权限，控制个人信息的知悉范围。\n7.2.2我们建立了数据分级分类制度、业务数据安全使用规范、数据合作规范等障对您个人信息的处置在收集、传输、使用、存储、转移、销毁等环节均满足法律法规的安全要求。根据有关法规要求，对需要访问个人信息的工作人员，我们会要求其经过个人信息保护负责人或者其授权的管理人员审批，记录访问情况，并采取技术措施，避免违法复制、下载个人信息。\n7.2.3我们组织员工参加安全与隐私保护相关培训并要求其完成规定的考核，加强员工对于保护个人信息重要性的认知。\n7.2.4我们还采取了其他可行的安全组织和管理措施。\n7.3合作协议条款保证\n7.3.1在我们从第三方间接收集您的个人信息前，我们会明确以书面形式（如合作协议、承诺书）要求该第三方在已经取得您明示同意后收集以及处理（如共享等）您的个人信息，在书面协议层面要求第三方对个人信息来源的合法性和合规性作出承诺，如第三方有违反行为的，我们会明确要求对方承担相应的法律责任。\n7.3.2在我们向业务合作伙伴共享您的个人信息前，我们会严格要求合作伙伴承担信息保护义务与责任，为此，我们还会要求业务合作伙伴签署数据处理协议或在合作协议中列明有关数据保护的条款，并要求业务合作伙伴在合作前签署。协议严格约定了业务合作伙伴的用户信息保密义务，包括信息的保管、使用和流转等均应满足我们的管控要求，接受我们的审核、监督和审计，一旦有任何违反，我们会要求对方承担相应的法律责任。\n7.4安全事件的处理\n7.4.1如不幸发生个人信息安全事件，我们将立即启动应急预案，采取补救措施，如造成或可能造成严重后果，我们将立即向有关部门报告，并向您告知安全事件的基本情况和可能的影响、我们已采取或将要采取的处置措施、您可自主防范和降低风险的建议、对您的补救措施等。我们将及时将事件相关情况以邮件、电话等方式告知您，难以逐一告知时，我们会采取合理、有效的方式发布公告。\n7.4.2请您知悉并理解，互联网并非绝对安全的环境，我们强烈建议您通过安全方式、使用复杂密码，协助我们保证您的账号安全。如您发现自己的个人信息泄密，尤其是您的账号或密码发生泄露，请您立即根据本政策第10条提供的联系方式联络我们，以便我们采取相应措施来保护您的信息安全。\n8. 如何存储您的个人信息\n8.1 我们会将在中国境内收集和产生的个人信息存储在中国境内。\n8.2 如需发生数据跨境传输，我们会向您告知您数据出境的目的、接收方，并征得您的同意，我们会确保数据接收方有充足的数据保护能力来保护您的个人信息。\n8.3 我们承诺，除非法律法规另有要求，您个人信息的存储时间将处于实现您授权使用的目的所必需的最短时间。\n8.4 如我们因经营不善或其他原因出现停止运营的情况，除非法律法规另有要求，我们会将此类情况向您告知，停止对您个人信息的收集，并删除已收集的个人信息或做匿名化处理。\n9. 如何更新本政策\n9.1 我们保留不时更新或修改本政策的权利。我们会在本页面上发布对本政策所做的任何变更。但是未经您明确同意，我们不会削减您按照本政策所应享有的权利。该等变更构成本政策的一部分。历史版本可参见这里。\n9.2 请您仔细阅读变更后的本政策内容。如果您不同意本政策，或对本政策修改、更新的内容有异议，您可以选择注销账号并不再使用我司产品与/或服务。但请您知悉，您账号注销之前、停止使用本服务之前的行为和活动仍受本政策的约束。如有资费相关问题，请阅读《我司用户服务协议》。\n9.3 若涉及重大、实质性变更，我们会依据具体情况，以显著的方式通知您。\n重大、实质性变更的情形包括但不限于以下情形：\n9.3.1我们的服务模式发生重大变化。如处理个人信息的目的、处理的个人信息类型、个人信息的使用方式等；\n9.3.2您参与个人信息处理方面的权利及其行使方式发生重大变化。\n10. 如何联系我们\n10.1我们设立了专门负责个人信息保护的部门，如您在执行或使用我司产品与/或服务时遇到与隐私保护相关的任何问题（包括问题咨询、投诉等），您可以通过以下方式与我们取得联系：（i）Email: tangyongqin@newto.cn；（ii）点击本应用APP中的【我的】-【联系我们】；（iii）我们提供的其他联系方式。\n11.2我们会在收到您的意见及建议并在验证您用户身份后的15日内尽快向您回复。如无法响应您的请求，我们会在法律规定的最大时限要求内，向您发送通知并解释原因。\n11.3如您对我们的回复不满意，或我们对您个人信息的处理方式侵害了您的合法权益，且在您向我们反馈后我们无法提供让您满意的回复且拒绝改进的，您可以向相关监管机构进行投诉和举报。\n11. 附则\n11.1本政策的解释及争议解决均应适用中国法律。与本政策相关的任何纠纷，双方应协商友好解决；若不能协商解决，您在此同意将争议提交至成都市高新区人民法院。\n11.2如果有管辖权的任何法院裁定或判决本政策的任何条款无效，则该条款将从本政策中移除，但该条款的无效不影响本政策其余条款的效力。本政策的其余条款将继续执行。\n11.3本政策的标题仅为方便阅读而设计，不影响本政策任何条款的含义或解释。\n12. 关键词定义\n12.1 用户或您：指使用我司产品与/或服务的儿童的监护人、已满18周岁的成年用户。\n13.2 儿童或您孩子：指您所监护的年龄不超过14周岁的未成年人。\n13.3 个人信息：指以电子或者其他方式记录的能够单独或者与其他信息结合识别特定自然人身份或者反映特定自然人活动情况的各种信息。个人信息包括个人基本信息、个人身份信息、个人生物识别信息、网络身份标识信息、个人健康生理信息、个人教育工作信息、个人财产信息、个人通信信息、联系人信息、个人上网记录、个人常用设备信息等。为免疑义，个人信息包括但不限于个人敏感信息。\n12.4 个人敏感信息：指一旦泄露、非法提供或滥用可能危害人身和财产安全，极易导致个人名誉、身心健康受到损害或歧视性待遇等的个人信息。个人敏感信息包括个人财产信息、个人健康生理信息、个人生物识别信息、个人身份信息、网络身份标识信息等。\n12.5 去标识化：指通过对个人信息的技术处理，使其在不借助额外信息的情况下，无法识别您的过程。\n12.6 匿名化：指通过对个人信息的技术处理，使得您无法被识别，且处理后的信息不能被复原的过程。\n12.7 中国或中国境内：指中华人民共和国（仅为本政策之目的，不包含香港特别行政区、澳门特别行政区和台湾地区）。\n12.8 设备：设备是指可用于访问我司产品和或或服务的装置，例如台式计算机、笔记本电脑、平板电脑、智能手机、智能大屏、智能手表、智能音箱等。\n成都上新游科技有限公司\n发布时间:2021年11月16日\n更新时间：2022年1月10日";
            this.text_canping = "用户服务协议\n生效时间：2021年11月16日\n更新时间：2022年1月10日\n本合作协议是甲方（指我司及/或其指定的关联公司，以下简称\“公司\”）与乙方（指以个人身份在我司关联的玩家，以下简称\“用户\”）之间就我司的应用（包括、APP、小程序等）使用事宜达成的具有法律效力的协议。\n为使用本软件及服务，您须仔细阅读并同意遵守《用户服务协议》（以下简称\“本协议\”）。您的下载、安装、使用、申请及获取账号、登录等行为即视为您已阅读并同意上述协议的约束。本公司有权不时地对本协议条款做任何修改和补充，并在本软件及服务中公布。对本协议条款的修改和补充将在公布时立即生效。您继续登录和使用本软件及服务即视为接受修订后的本协议。否则，您有权通过停止登录且拒绝使用本软件及服务的方式来终止本协议。\n您承诺遵守适用的法律、法规、规章及其他政府规定。用户发生违法或犯罪行为，或者本公司有相关证据合理怀疑用户可能存在违法、犯罪或违反本协议的行为的，本公司有权中止或终止对该用户的服务，有权在本公司及互联网络上公布该用户的违法、违约行为，有权向有关机关报案并配合有关机关的调查活动，包括但不限于披露该用户的所有用户信息、交易记录及其他信息，本公司实施上述任何及全部行为所造成的后果、风险及责任由用户自行承担。用户签署本协议即视为理解并接受上述条款，永久性不可撤销地授权本公司公开相关信息及采取相应措施。\n本协议内容包括本协议正文以及本公司已经发布的或将来可能发布的各类规则、声明、说明。所有规则、声明、说明为本协议不可分割的一部分，与本协议正文具有同等法律效力。\n如果您代表某个机构而非您个人注册、登录和使用本软件及服务，则您将被认为获得充分授权代表该机构同意本协议条款以及本协议条款不时的修改和补充。如果您未满18周岁，请在法定监护人的陪同下阅读本协议及其他上述协议，并在征得监护人同意的前提下使用本应用并特别注意本协议第十条约定的未成年人使用条款。未成年人行使和履行本协议项下的权利和义务即视为已获得了监护人的认可。\n乙方应仔细阅读并充分理解本协议后点击【\“同意\”】按钮。\n乙方应当清楚，当乙方点击【\“同意\”】按钮后，即表示乙方与\"我司\"科技签订本协议并生效，乙方在使用\"我司\"科技的产品及服务过程中不得违反本协议的约定。\n.总则\n签署本协议前，乙方已充分了解甲方软件的各项规则及要求，且有条件及有能力履行本协议约定的义务。乙方对本协议下所有条款及定义等内容均已明确知悉并有能力承担因此所产生的相应责任，并无疑义。\n1.本软件及服务可能不时开放不同功能，每一功能的使用及其限制将受制于针对该功能的特定业务规则。本公司有权制定及不时修订该等特定业务规则，并予以公示。您使用或继续使用该等功能即视为您同意并接受该等规则及其修订。\n2.账号的所有权归属本公司，用户完成注册手续后，仅获得账号的使用权，且该使用权仅属于申请注册人本人。账号申请注册人不得以任何方式向第三方有偿或无偿提供账号使用权，包括但不限于转让、赠与、租用、借用、许可使用。账号申请注册人违反上述约定的，因此导致的后果、责任均由其自行承担，给本公司造成损失的，应予全额赔偿，本公司并有权收回账户使用权。\n3.您应妥善保管账号及/或与账号绑定的微信账号，准确输入该等信息将作为您访问登录及使用本软件及服务的关键环节。您知悉并同意，任何在您账号项下发生的行为均将视为您或经您合法授权的行为，并将由您承担一切法律后果。如果您发现有人未经授权使用您的账号，您应立即向本公司反馈，否则未经授权的使用行为均视为用户本人行为。尽管有前述约定，任何情况下，本公司均不对您因账号、登录信息错误、遗失、被盗，或因您未能确保账号信息安全的其他情形而造成的任何损失或损害，承担任何责任。\n4.您在此授予本公司保护您账号安全的相关权利，本公司可定期或不定期采用不同的方式对您账号的使用安全进行检查、验证，包括但不限于主动联系您进行【身份验证、短消息验证、邮箱认证】等。如您无法完成验证或无正当理由拒绝验证的，本公司可合理怀疑您的账号出现异常或被盗，因而中止向该账号提供服务或采取进一步措施。\n.乙方\n为避免疑义，乙方特此确认其向甲方提供如下信息，并保证该信息真实、有效、完整：姓名、订单号、手机号码、商品信息、userID。\n1.甲方提供的应用分为免费和付费。对于现阶段免费提供的服务内容，甲方有权随时根据甲方政策的调整将其变更为收费服务，如果发生该等变更的，甲方将提前告知用户，如果用户不同意的，可以停止使用该服务。对于收费的服务内容，甲方有权自行确定该等服务的价格并不时调整，调整后的价格将在软件中进行公示，如果用户不同意调整后的价格的，则用户可以在其原购买的收费服务到期后停止付费购买收费服务。如果用户在甲方调整价格并公示后，仍继续购买收费服务的，则视为用户同意该调整后的价格。\n2.甲方不时将对付费服务提供优惠政策或开展优惠活动，该等优惠政策或优惠活动以软件上的实时信息为准，用户知悉并认可，甲方有权自行决定是否提供优惠政策或开展优惠互动，以及优惠政策或优惠活动的内容、条件和有效期，并不时进行调整或终止。\n3.如果用户通过合法方式获得了相关付费服务的优惠券或折扣等其他优惠方式的，则用户知悉并认可，该等优惠券或折扣必须满足一定条件才能使用，以优惠券上的信息提示和优惠券使用说明、或甲方公示的折扣使用说明为准。优惠券或折扣不兑现（即不直接兑换现金）、不找零（即订单金额不足抵用券金额的，抵用券剩余金额无效）、过期作废；优惠券或折扣仅限同一用户账号使用，严禁出售、转让或共享，如经甲方发现存在出售、转让或共享优惠券或折扣等其他优惠方式的，该优惠券或折扣将予以作废处理，且甲方有权视情节严重程度予以封号处理，并保留追究相关责任人法律责任的权利。\n.用户注意事项\n1.您理解并同意，本软件及服务的某些功能可能会让第三方知晓用户的信息，包括但不限于：用户的软件好友可以查询用户头像、名字等可公开的个人资料，您同意本公司无须就此类事项向您承担任何责任。\n2.基于互联网的开放性属性，您知悉用户将图片等资料上传到互联网上，有可能会被其他组织或个人复制、转载、擅改或做其它非法用途，您必须充分意识到此类风险的存在。您明确同意使用本软件及服务所存在的风险将完全由您自行承担；因您使用本软件及服务而产生的一切后果也由您自行承担。\n3.用户使用本软件及服务的特定功能时，本公司可能会调用第三方系统或者通过第三方支持用户的使用或访问，使用或访问的结果由该第三方提供（包括但不限于第三方通过本公司公众账号提供的服务，或通过开放平台接入的内容等），本公司不对第三方提供服务及内容的安全性、准确性、有效性及其他事项作出任何承诺或承担任何责任，相关风险及责任由用户自行承担。您在使用该等第三方产品或服务时，除遵守本协议约定外，还应遵守第三方的用户协议。如因本软件使用的第三方软件或技术引发的任何纠纷，应由该第三方负责解决，本公司不承担任何责任。本公司不对第三方软件或技术提供客服支持，若您需要获取支持，请与第三方联系。\n4.您理解并同意，不论在何种情况下，本公司均不对由于互联网连接故障，电脑、通讯或其他系统的故障，电力故障，罢工，劳动争议，暴乱，起义，骚乱，火灾，洪水，风暴，爆炸，不可抗力，战争，政府行为，国际、国内法院的命令，第三方的不作为或任何本公司不能合理控制的原因而造成的本软件及服务不能访问、信息及数据的延误、停滞或错误，不能提供或延迟提供服务，承担任何责任。\n5.您理解并同意，本公司有权决定将本软件及服务作商业用途，本公司承诺在推广过程中严格按照本协议约定保护您的个人信息，同时您亦可以根据系统设置选择屏蔽、拒绝接收相关推广信息。\n6.您理解并同意，本公司会将会尽其商业上的合理努力保障您在本软件及服务中的数据存储安全，但是，本公司并不能就此提供完全保证，包括但不限于以下情形：\n（1）本公司不对您在本软件及服务中相关数据的删除或储存失败负责；\n（2）本公司有权根据实际情况自行决定单个用户在本软件及服务中数据的最长储存期限，并在服务器上为其分配数据最大存储空间等。您可根据自己的需要自行备份本软件及服务中的相关数据；\n（3）如果您停止使用本软件及服务或服务被终止或取消，本公司可以从服务器上永久地删除您的数据。服务停止、终止或取消后，本公司没有义务向您返还任何数据。\n.软件安装、卸载及更新\n1.甲方可能为不同的终端设备开发了不同的软件版本，您应当根据实际情况选择下载合适的版本进行安装。\n2.下载安装程序后，您需要按照该程序提示的步骤正确安装。为提供更加优质、安全的服务，在本软件安装时甲方可能推荐您安装其他软件，您可以选择安装或不安装。如果您不再需要使用本软件或者需要安装新版软件，可以自行卸载。如果您愿意帮助甲方改进产品服务，请告知卸载的原因。\n3.为了改善用户体验、完善服务内容，本公司将不断努力开发新的功能或版本，并为您不时提供软件更新（这些更新可能会采取软件替换、修改、功能强化、版本升级等形式）。\n4.为了保证本软件及服务的安全性和功能的一致性，本公司有权不经向您特别通知而对本软件进行更新，或者对本软件的部分功能效果进行改变或限制。本软件新版本发布后，旧版本的软件可能无法使用，请您随时核对并下载最新版本，因未能及时下载及使用最新版本导致的使用不能及相关风险将由您承担。如果您继续使用本软件及服务，则视为您同意接受更新后的本软件及服务。如果您拒绝接受更新后的本软件及服务，您应立即停止登录及使用本软件及服务；该等情形下，您仍需就您在停止使用前已发生的行为承担全部责任。\n5.您理解并同意，本公司有权因本软件及服务维护、升级或其他原因而暂停服务、调整服务功能或终止服务，本公司将事先在软件上进行公示或通知。您同意本公司无须就此向您承担任何责任。\n.用户行为规范\n（一）信息内容合法性规范\n用户使用本软件及服务过程中所制作、复制、发布、传播的任何内容（包括但不限于应用账号头像、名字、用户说明等信息，或文字、语音、图片等发送、回复，以及使用应用账号或本软件及服务所产生的其他内容）的合法性。用户承诺不利用应用账号或本软件及服务制作、复制、发布、传播下干扰软件正常运营，以及侵犯其他用户或第三方合法权益的内容，包括但不限于：\n1.发布、传送、传播、储存法律法规禁止的内容。\n2.发布、传送、传播、储存侵害他人名誉权、肖像权、知识产权、商业秘密等合法权利的内容。\n3.涉及他人隐私、个人信息或资料的。\n4.发表、传送、传播骚扰、广告信息、过度营销信息及垃圾信息或含有任何性或性暗示的信息。\n5.违反中华人民共和国宪法确定的基本原则的，危害国家安全的、泄露国家秘密、颠覆国家政权、破坏国家统一的。\n6.煽动地域歧视、地域仇恨、民族仇恨、民族歧视、破坏民族团结的。\n7.破坏国家宗教政策，宣扬邪教和迷信的，以非法民间组织名义活动的。\n8.包含恐怖、淫秽、色情、赌博、暴力血腥、高危险性、凶杀恐怖、教唆犯罪、危害表演者自身或他人身心健康内容的。\n9.危害网络安全、利用网络从事危害国家安全、荣誉和利益的，不符合应当遵守法律法规、社会主义制度、国家利益、公民合法利益、公众秩序、社会道德风尚和信息真实性等\“七条底线\”要求的。\n10.含有虚假、有害、胁迫、侵害他人隐私、骚扰、侵害、中伤、粗俗、猥亵、或其它道德上令人反感的内容。\n11.其他违反法律法规、政策及公序良俗、社会公德或干扰软件正常运营和侵犯其他用户或第三方合法权益内容的信息。\n（二）软件使用合法性规范\n1.用户使用本软件及服务的行为须尊重及维护本公司对软件的知识产权及软件的正常稳定运营，用户承诺不实施任何可能侵犯软件知识产权或影响软件正常运营的行为，包括但不限于：\n（a）删除或修改本软件上关于著作权的信息；\n（b）对本软件进行反向工程、反向汇编、反向编译，或者以其他方式尝试发现本软件的源代码；\n（c）对本公司拥有知识产权的内容进行使用、出租、出借、复制、模仿修改、转载、汇编、发表、出版、建立镜像站点等；\n（d）对本软件或者本软件运行过程中释放到任何终端内存中的数据、软件运行过程中客户端与服务器端的交互数据，以及本软件运行所必需的系统数据，进行复制、修改、增加、删除、挂接运行或创作任何衍生作品，形式包括但不限于使用插件、外挂或非经授权的第三方工具/服务接入本软件和相关系统；\n（e）通过修改或伪造软件运行中的指令、数据，增加、删减、变动软件的功能或运行效果，或者将用于上述用途的软件、方法进行运营或向公众传播，无论这些行为是否为商业目的；\n（f）通过非本公司开发、授权的第三方软件、插件、外挂、系统，登录或使用本软件及服务，或制作、发布、传播上述工具；\n（g）自行或者授权他人、第三方软件对本软件及其组件、模块、数据进行干扰；\n（h）故意制作、传播计算机病毒等破坏性程序的；\n（i）其他违反法律法规规定、侵犯其他用户合法权益、干扰本软件及服务正常运营的行为。\n1.用户使用本软件及服务过程中不得发生下列行为：\n（a）提交、发布虚假信息，或冒充、利用他人名义的；\n（b）强制、诱导其他用户点击链接页面或分享信息的；\n（c）虚构事实、隐瞒真相以误导、欺骗他人的；\n（d）侵害他人名誉权、肖像权、知识产权、商业秘密等合法权利的；\n（e）将应用账号或其任何功能以任何方式提供给第三方，或用于任何商业性用途；\n（f）利用应用账号或本软件及服务从事任何违法犯罪活动的；\n（g）制作、发布与以上行为相关的方法、工具，或对此类方法、工具进行运营或传播，无论这些行为是否为商业目的；\n（h）利用技术手段批量建立虚假账号；\n（i）其他违反法律法规规定、侵犯其他用户合法权益、干扰本软件及服务正常运营的行为。\n（三）违约处理\n1.如果本公司发现或收到他人举报或投诉用户违反本协议约定，本公司有权不经通知随时对相关内容进行删除、屏蔽，并视行为情节对违规账号处以包括但不限于警告、限制、中断、终止或禁止使用部分或全部功能、账号封禁直至注销的处罚，并公告处理结果。\n2.您理解并同意，本公司有权依合理判断对违反有关法律法规或本协议规定的行为进行处罚，对违法违规的任何用户采取适当的法律行动，并依据法律法规保存有关信息向有关部门报告等，用户应独自承担由此而产生的一切法律责任。\n3.您理解并同意，因您违反本协议或相关服务条款的规定，导致或产生第三方主张的任何索赔、要求或损失，您应当独立承担责任；本公司因此遭受损失的，您也应当一并赔偿。\n4.您理解并同意，如果您注册的账号在任何连续【180】日内未实际使用，则本公司有权删除该账号并停止为该用户提供相关服务。\n.未成年人使用条款\n1.若用户是未满18周岁的未成年人，用户应在监护人的指导并获得监护人同意的情况下，认真阅读并同意本协议后，方可使用本软件及服务。\n2.本公司重视对未成年人个人信息的保护，未成年用户在填写个人信息时，请加强个人保护意识并谨慎对待，并应在取得监护人的同意以及在监护人指导下正确使用本软件及服务。\n3.为更好地保护未成年人隐私权益，本公司司特别提醒您慎重发布包含未成年人素材的内容，一经发布，即视为您已获得权利人同意在本软件及服务展示未成年人的肖像、声音等信息，且允许本公司司依据本协议使用、处理该等与未成年人相关的内容。\n4.监护人特别提示：如您的被监护人使用本软件及服务的，您作为监护人应指导并监督被监护人的注册和使用行为，如您的被监护人申请注册账号，公司将视为其已取得您的同意。\n5.未成年人用户及其监护人理解并确认，如您违反法律法规、本协议内容，则您及您的监护人应依照法律规定承担因此而可能导致的全部法律责任。\n.支付与资费政策\n1.我们在提供本应用应用产品与／或服务时，会对您（注册用户）需要提供付费服务，向您收取一定的费用，以完成付费充值。目前，充值服务类型主要有按特定时长的订阅服务（包括但不限于月卡、季卡、年卡）。如您不同意支付该等费用，则无法使用收费服务部分的内容。\n2.我们有权决定付费本应用应用产品与／或服务的资费标准。我们可能会因铁皮人产品与／或服务的服务类型和服务周期不同，制定不同的资费标准，详细约定以各款应用公布、实际展示的资费标准为准。\n此外，需要特别提示您，根据App Store、华为应用商店的政策，其可能会向您提供自动扣费服务，请您在选择使用前仔细阅读App Store、华为应用商店的相关政策。若您希望取消自动扣费功能的，请您根据App Store、华为应用商店的相关政策及退订规则操作或与苹果公司、华为公司取得联系。\n1.您理解并同意，您支付前述服务费用的方式包括通过我们接入的第三方支付工具（如微信支付、支付宝等）向我们进行付款，当然您也可以向App Store、华为应用商店等第三方应用商店进行充值以完成您的支付。如支付工具有更新，我们会在支付页面上及时更新支付规则并通知您，详细约定以本应用公布、实际展示的支付规则为准。\n2.您理解并同意，因使用第三方支付工具进行支付而可能产生的任何商业风险（包括但不限于交易信息输入错误、不法分子利用您账户或银行卡等有价卡等进行违法活动等）均有可能给您造成相应的经济损失。我们在充分履行本协议条款和规则的义务和符合法律规定的前提下，不对您的前述风险和损失承担任何责任。我们善意提醒您，在与第三方建立法律关系前，您应审慎阅读第三方的用户服务协议、隐私政策等法律文件。\n3.如您需取消当前订阅，请您联系Email: tangyongqin@newto.cn。但是，如您使用App Store、华为应用商店自动扣费服务的，可以根据App Store、华为应用商店的退款政策向苹果公司、华为公司主张取消订阅服务及／或处理退款事宜。\n.保密和知识产权保护\n1.受国际版权公约、中华人民共和国著作权法、专利法、及其他知识产权方面的法律法规的保护，甲方平台的产品与/或服务及所使用的软件、界面、交互等元素、LOGO、商标、标识、图形、文字、文本、声音、音频、视频、肖像、图片、技术、材料等的所有知识产权归甲方所有和享有。\“知识产权\”包括在专利法、版权法、商标法、反不正当竞争法中等法律规定的任何和所有权利、任何和所有其它所有权以及其中的任何和所有应用、更新、扩展和恢复。\n2.对于乙方发表到甲方平台上的任何内容，用户同意甲方在全世界范围内具有免费的、永久性的、不可撤销的、非独家的和完全再许可的权利和许可，以使用、复制、修改、改编、出版、翻译、据以创作衍生作品、传播、表演和展示此等内容（整体或部分），和/或将此等内容编入当前已知的或以后开发的其他任何形式的作品、媒体或技术中。并且，用户许可本公司有权利就任何主体侵权而单独采取法律行动（包括但不限于诉讼、举报、律师函等形式）。\n3.乙方对本协议的内容以及在本合同签订、履行过程中知悉、获取的甲方资料和信息，除甲方做出相反的书面说明外，均应视为甲方的商业秘密。未经甲方书面同意不得向任何第三方泄漏。乙方因违反保密义务而给甲方造成损失的，应负相应的赔偿责任。\n4.尽管有上述规定，但本条款不适用以下信息：\n.依据法律法规的规定应当提供，行政、司法等职权部门要求提供的信息；\n.披露方同意接受方向第三方提供的信息；\n.不违反本协议的情况下，接受方从第三方合法正当接受的，不受保密限制的信息；\n.接受方未违反本协议的情况下即已公开或可以公开获得的信息；\n.法律法规所要求披露的信息。\n1.无论本协议是否变更、解除或终止，本协议保密条款不受其限制而继续有效，双方都应继续承担保密条款约定的保密义务。\n2.甲方为培训课程所制作的全部资料，在教学过程中产生的全部资料，甲方网站上所显示、使用或提供的软件、程序及内容，未经甲方同意，乙方在任何时候均不得以商业目的（包括但不限于将资料上传到乙方或其他方自媒体平台等）对上述资料进行利用；不得以任何方式翻译、复制、传播、编辑上述资料；不得删除、掩盖或更改甲方的版权声明、商标或其他权利声明/标签。\n.违约责任\n1.用户发生如下情形之一的，视为违约：\n（1）使用平台服务时违反有关法律法规规定的；\n（2）违反本协议或平台发布的规则的。\n1.如果本公司发现或收到他人举报您有违约行为的，本公司有权根据自身的判断，不经通知直接采取以下一项或多项处理措施：\n（1）对相关内容进行删除、屏蔽；\n（2）书面警告，以短信、邮件、站内信等一种或多种形式通知您要求您限期纠正该等行为；\n（3）限制、暂停、终止您使用部分或全部服务；\n（4）不退还任何的服务费用，且不进行任何的赔偿或补偿；\n（5）其他根据平台规则可采取的措施。\n1.如您的行为使本公司遭受损失的（包括但不限于被相关行政机关进行行政处罚或被第三人主张赔偿的），您应赔偿本公司的全部损失（包括但不限于直接经济损失、商誉损失及对外支付的赔偿金、和解款、律师费、诉讼费等经济损失）。\n.不允诺担保、责任有限及免责\n1.不允诺担保：甲方软件的用户提问、回答、发言仅代表用户个人观点，与甲方无关，且甲方不对用户发表的回答或评论的正确性、准确性、真实性进行保证。甲方为按现状提供，包含瑕疵及缺陷，甲方不提供任何其他担保。甲方不允诺其他无论是明示的还是默示的保证和担保，包括但不限于对特定目的的适销性和适应性，不侵权，以及任何出自交易过程或商业惯例的保证和条件。从甲方或他处获取的任何建议或信息，不论是口头的还是书面的，都不会对甲方产生本协议明确约定之外的其他保证和担保。\n2.责任有限：适用法律所允许的最大范围内，不论出于何种原因，甲方对您的全部责任为您为使用甲方的产品和服务而向甲方支付的费用总额。如果甲方免费向您提供了甲方服务，则甲方无义务对您承担任何赔偿责任。任何情况下，甲方不对出自本协议或与本协议有关的或与甲方的运行有关的任何特殊的、偶然的、惩罚性或间接的损害（包括但不限于数据损失、利润损失或营业损失）或采购替代产品承担责任，不论该等责任是基于合同，担保，侵权（包括过失），严格责任或其他，也不论甲方是否被事先告知了损失或损害发生的可能性。即使本协议约定的有限补救措施未能实现其根本目的，前述的责任限制依然适用。\n3.免责：如因下所列任何原因而造成的用户资料泄露、丢失、被盗用或被篡改等，甲方均有权获得免责：\n（1）任何因政府管制而造成的暂时性关闭等影响网络正常浏览的不可抗力原因；\n（2）任何计算机系统问题、黑客攻击、计算机病毒侵入或发作的原因；\n（3）任何因网络安全问题导致的账户或银行卡信息泄露，该风险可能会给用户造成经济损失，用户承诺自行承担向侵权方追究责任及追究责任不能的后果；\n（4）任何因用户个人原因导致的用户资料泄露、丢失、被盗用或被篡改等；\n（5）其他不能归责于甲方的原因。\n.合同终止与解除\n1.存在以下情况之一的，本公司将有权终止协议：\n（1）您主动中止或终止，包括但不限于课程到期未进行续费等；\n（2）因为您的违约行为，本公司主动中止或终止提供服务的；\n（3）因国家或相关政府监管部门要求或发生不可抗力事件时，本公司中止或终止服务的；\n（4）其他根据法律法规应中止或终止服务的。\n1.中止或终止服务后，本公司有权利但无需确保您收到特别提示或通知。当您发现无法登录或享受服务时，可以咨询客服人员。\n2.中止或终止后的处理\n（1）除法律规定的责任外，本公司对本服务的中止或终止对您和第三人不承担任何责任；\n（2）已收取的费用按照退款制度予以退还；\n（3）在本公司正在进行的交易，本公司将根据情况合理判断，是否继续进行或终止相关交易，并视情况决定是否进行退费或扣除相关费用作为违约金。\n1.除本协议另有约定情况外，本协议自您选择接受或使用本服务后生效，直至您终止本服务/注销账号时终止。\n.争议解决\n1.在本协议执行期间如果双方发生争议，双方应友好协商解决。如果协商不成，任何一方可向甲方所在地有管辖权的人民法院提请诉讼。\n2.诉讼期间，除争议内容以外，双方均应继续履行本协议约定的其他内容。\n.其他\n1.本协议自乙方注册提交信息之日起生效，本协议最终解释权归本公司所有，并保留一切解释和修改的权力。本协议其他条款未明示授权的其他一切权利仍由甲方保留，您在行使这些权利时须另外取得甲方的书面许可。甲方如果未行使前述任何权利，并不构成对该权利的放弃。\n2.若本协议的任何条款被享有合法管辖权的法院裁定为无效或不可执行，则该等条款应在最低必要范围内予以限制或排除，并以最能体现本用户协议意图的新条款予以替换，以使本用户协议继续完全有效。\n3.本用户协议构成甲方及乙方之间使用本软件相关的整体协议，并在此取代及取消甲方及乙方之间先前存在的与该等使用相关的任何书面或口头协议或理解（如有）。\n4.本协议的任何条款无论因何种原因无效或不具可执行性，其余条款仍有效，对双方具有约束力。\n5.乙方注册的账号和/或与账号绑定的微信账号是乙方接受平台服务的唯一身份凭证，由乙方负责保管。如因乙方原因造成损失，由乙方承担。\n6.乙方在使用软件时不可将个人隐私、商业秘密、国家机密等不得公开的信息发表、上传至甲方软件，也不可将该等信息通过甲方软件的服务传播给其他人。如因乙方原因造成损失，由乙方承担。\n7.乙方注册的账号/或与账号绑定的微信账号是乙方接受平台服务的唯一身份凭证，由乙方自行负责保管，禁止赠与、借用、租用、转让或者售卖。如因此给甲方造成损害，由乙方承担。\n8.如果您对本协议有任何疑问、意见或建议，通过以下方式与甲方联系：Email: games@sxybyte.com。";
        }
        get URL() {
            yinsizhengceBinder.bindAll();
            return UI_ui_wendang.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_bg];
        }
        OnCreate() {
            this.ui.m_btn_fanhui.onClick(this, this.fanhuiFun);
        }
        fanhuiFun() {
            UIManager.inst.PopUI(this.layer, false);
        }
        OnShow(...args) {
            GameSDK.hideBanner();
            console.log(" UIwendang OnShow==>>", args[0]);
            if (args[0] == 0) {
                this.ui.m_loader_biaoti.url = "ui://jyvm79orfkj2s";
                this.ui.m_wenAn.m_text_wenan.text = this.text_yinsi;
            }
            else {
                this.ui.m_loader_biaoti.url = "ui://jyvm79orfkj2t";
                this.ui.m_wenAn.m_text_wenan.text = this.text_canping;
            }
        }
        OnHide() {
        }
        OnDispose() {
        }
        onSwitch() {
        }
        onAniEnd() {
        }
        updateScene(offset = 1) {
        }
    };
    UIwendang = __decorate([
        ui_register()
    ], UIwendang);

    let UIYinSi2 = class UIYinSi2 extends BaseUI {
        get URL() {
            yinsizhengceBinder.bindAll();
            return UI_ui_yinsi2.URL;
        }
        OnCreate() {
        }
        OnShow(...args) {
            if (args[0] == 0) {
                this.ui.m_comp_yinsi2.m_btn_agree.visible = false;
                this.ui.m_comp_yinsi2.m_btn_reject.visible = false;
            }
            else {
                this.ui.m_comp_yinsi2.m_btn_agree.visible = true;
                this.ui.m_comp_yinsi2.m_btn_reject.visible = true;
            }
            this.ui.m_comp_yinsi2.m_btn_agree.onClick(this, this.AgreeFun);
            this.ui.m_comp_yinsi2.m_btn_reject.onClick(this, this.RejectFun);
            this.ui.m_comp_yinsi2.m_btn_fuwuxieyi.onClick(this, this.yonghuxieyi);
            this.ui.m_comp_yinsi2.m_btn_yinsizhengce.onClick(this, this.chanpinyinsishuoming);
        }
        AgreeFun() {
            UIManager.inst.PopUI(this.layer, false);
            var jsonData = Laya.LocalStorage.getItem("HadAgreeFirst");
            if (!jsonData) {
                Laya.LocalStorage.setItem("HadAgreeFirst", "1");
            }
            GameSDK$1.hasAgreeXieYi = true;
        }
        RejectFun() {
            qg.exitApplication({});
            GameSDK$1.hasAgreeXieYi = false;
        }
        chanpinyinsishuoming() {
            UIManager.inst.PushUI(UIwendang, UILayer.top, 0);
        }
        yonghuxieyi() {
            UIManager.inst.PushUI(UIwendang, UILayer.top, 1);
        }
        OnHide() {
        }
        OnDispose() {
        }
        onSwitch() {
        }
        onAniEnd() {
        }
        updateScene(offset = 1) {
        }
    };
    UIYinSi2 = __decorate([
        ui_register()
    ], UIYinSi2);

    let UIYinSi = class UIYinSi extends BaseUI {
        get URL() {
            yinsizhengceBinder.bindAll();
            return UI_ui_yinsi1.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_com_mengban];
        }
        OnCreate() {
        }
        OnShow(...args) {
            if (args[0] == 0) {
                this.ui.m_comp_yinsi1.m_btn_agree.visible = false;
                this.ui.m_comp_yinsi1.m_btn_reject.visible = false;
                this.ui.m_comp_yinsi1.m_btn_close.visible = true;
            }
            else {
                this.ui.m_comp_yinsi1.m_btn_agree.visible = true;
                this.ui.m_comp_yinsi1.m_btn_reject.visible = true;
            }
            this.ui.m_comp_yinsi1.m_btn_agree.onClick(this, this.AgreeFun);
            this.ui.m_comp_yinsi1.m_btn_reject.onClick(this, this.RejectFun);
            this.ui.m_comp_yinsi1.m_btn_yonghuxieyi.onClick(this, this.yonghuxieyi);
            this.ui.m_comp_yinsi1.m_btn_chanpinyinsishuoming.onClick(this, this.chanpinyinsishuoming);
            this.ui.m_comp_yinsi1.m_btn_close.onClick(this, () => {
                UIManager.inst.PopUI(this.layer, true);
            });
        }
        AgreeFun() {
            UIManager.inst.PopUI(this.layer, false);
            var jsonData = Laya.LocalStorage.getItem("HadAgreeFirst");
            if (!jsonData) {
                Laya.LocalStorage.setItem("HadAgreeFirst", "1");
            }
            GameSDK$1.hasAgreeXieYi = true;
        }
        RejectFun() {
            UIManager.inst.PopUI(this.layer, true);
            UIManager.inst.PushUI(UIYinSi2, UILayer.top, 2);
            GameSDK$1.hasAgreeXieYi = false;
        }
        chanpinyinsishuoming() {
            UIManager.inst.PushUI(UIwendang, UILayer.top, 0);
        }
        yonghuxieyi() {
            UIManager.inst.PushUI(UIwendang, UILayer.top, 1);
        }
        OnHide() {
        }
        OnDispose() {
        }
        onSwitch() {
        }
        onAniEnd() {
        }
        updateScene(offset = 1) {
        }
    };
    UIYinSi = __decorate([
        ui_register()
    ], UIYinSi);

    class Hbs extends BaseSDK {
        constructor() {
            super();
            this._closeCallback = null;
            this.appid = "106267597";
            this.bannerAd = "p90y1vr7zg";
            this.rewardAd = "q6v8u0thm1";
            this.insertAd = "g6jj6q89bv";
            this.nativeAd = "k0b74jp66v";
            this.defaultServerConfig = {
                "is_AddDesk": { state: "0" },
                "is_AvdFloat": { state: "0" },
                "is_AvdList": { state: "0" },
                "is_GameNative": { state: "1", interval: 5 },
                "is_MoreGameBtn": { state: "0" },
                "is_PropSceneUI": { state: "0", interval: 5 },
                "is_banner": { state: "0", interval: 5 },
                "ad_all_interval": 10,
            };
            this.isShowBanner = false;
            this.isLoading = false;
            this.hasLogin = false;
            this.bannerShowSate = 0;
            this.count = 0;
            this._version = "1.0.0";
            SDKConfig.GAMEID = "20283";
            this.configURL = `https://hallcq.jpsdk.com/static/${SDKConfig.GAMEID}/huawei/api.json`;
            this.hbs = Laya.Browser.window.hbs;
            if (this.hbs) {
                this.hbs.hasShortcutInstalled({
                    success: function (res) {
                        this.isShortCutInstalled = res;
                    },
                    fail: function () {
                        this.isShortCutInstalled = false;
                    }
                });
                this.hbs.onShow(() => {
                    ToolSDK.replaySound();
                    this.initConfig(null);
                    GameSDK$1.GameNativeInsert.reportAdShowNew();
                });
                this.hbs.onHide(() => {
                });
            }
            this.createBanner();
        }
        onConfigLoad() {
            this.count++;
            if (this.count > 1)
                return;
            console.log("配置加载完毕，开始登陆..................." + "次数" + this.count);
        }
        huaWeiLogin(succedFun, FailFun, caller) {
            console.log("华为登录");
            if (!Laya.Browser.window.hbs) {
                GameSDK$1.curentState.ChangeLoadingState();
                UIManager.inst.PopUI(UILayer.cover, true);
                UIManager.inst.PopUI(UILayer.login, true);
                return;
            }
            this.showToast("正在登录中...", 1000);
            this.hbs && this.hbs.gameLoginWithReal({
                forceLogin: 1,
                appid: this.appid,
                success: (data) => {
                    console.log("Game login success:" + data);
                    if (succedFun && caller) {
                        succedFun.call(caller);
                    }
                    GameSDK$1.GameNativeInsert = new GameNativeInsert(this.nativeAd);
                    GameSDK$1.GameNativeBanner = new GameNativeBanner(this.nativeAd);
                    GameSDK$1.curentState.ChangeLoadingState();
                    UIManager.inst.PopUI(UILayer.cover, true);
                    UIManager.inst.PopUI(UILayer.login, true);
                    this.showToast("登录成功", 1000);
                    this.hasLogin = true;
                    this.showBanner();
                },
                fail: (data, code) => {
                    console.log("game login with real fail:" + data + ", code:" + code);
                    let str = "";
                    if (code == 7013) {
                        this.showToast("登录失败,请登录华为账号!", 1000);
                        str = "登录失败,请登录华为账号!";
                    }
                    else if (code == -1) {
                        this.showToast("登录失败,请检查网络状态!", 1000);
                        str = "登录失败,请检查网络状态!";
                    }
                    if (FailFun && caller) {
                        FailFun.call(caller);
                    }
                    this.hasLogin = false;
                    UIManager.inst.ShowUI(LoginUI, UILayer.login, str);
                }
            });
        }
        static isFastClick() {
            if (this.lastClickTime !== 0 && ((new Date().getTime() - this.lastClickTime) < this.CLICK_INTERVAL)) {
                this.lastClickTime = new Date().getTime();
                return true;
            }
            this.lastClickTime = new Date().getTime();
            return false;
        }
        navigateToMiniGame(msg) {
            this.hbs && this.hbs.navigateToQuickApp({
                packageName: msg.packageName,
                extraData: msg.data,
                success: function () {
                    console.log("-----------跳转小游戏:" + msg.name + "成功-----------");
                },
                fail: function (res) {
                    console.log("-----------跳转小游戏:" + msg.name + "失败-----------");
                    console.log(JSON.stringify(res));
                }
            });
        }
        shortCutInstalled(onComplete) {
            var self = this;
            this.hbs.installShortcut({
                message: 'Adds the shortcut to the desktop for convenient use next time.',
                success: function (ret) {
                    console.log('handling install success');
                    onComplete && onComplete(true);
                },
                fail: function (erromsg, errocode) {
                    console.log('handling install fail: ' + erromsg + ", errocode: " + errocode);
                    onComplete && onComplete(false);
                },
                complete: function () {
                    console.log("handling install  complete");
                    self.showToast("桌面图标已创建", 1000);
                }
            });
        }
        createBanner() {
            if (!this.hbs)
                return;
            var sysInfo = this.hbs.getSystemInfoSync();
            console.log("on getSystemInfoSync: success =" + JSON.stringify(sysInfo));
            let bannerTop = sysInfo.safeArea.height;
            this._bannerNode = this.hbs.createBannerAd({
                adUnitId: this.bannerAd,
                style: {
                    top: bannerTop - 57,
                    left: 0,
                    height: 57,
                    width: 360
                }
            });
            this._bannerNode.onLoad(() => {
                this.bannerShowSate = 1;
            });
            this._bannerNode.onError(e => {
                this.bannerShowSate = 0;
            });
            this._bannerNode.onClose(() => {
                this.bannerShowSate = 0;
            });
        }
        showBanner() {
            if (!this.hbs) {
                return;
            }
            if (!GameSDK$1.hasAgreeXieYi)
                return;
            if (!this.hasLogin) {
                return;
            }
            console.log("ad-展示banner");
            if (this._bannerNode) {
                this.isShowBanner = true;
                if (this.bannerShowSate == 0) {
                    this._bannerNode.show();
                }
            }
        }
        hideBanner() {
            console.log("ad-隐藏banner");
            if (this._bannerNode)
                this._bannerNode.hide();
            this.isShowBanner = false;
        }
        showADVideo(caller, onComplete, reason, args) {
            if (!this.hasLogin)
                return;
            var self = this;
            console.log("ad-展示激励视频");
            this.onADViewComplete = Handler.create(caller, onComplete, args);
            if (!this.hbs) {
                self.onADViewComplete.runWith(true);
                return;
            }
            if (typeof (this.hbs.createRewardedVideoAd) == 'undefined')
                return;
            if (Hbs.isFastClick()) {
                self.onADViewComplete.runWith(false);
                return;
            }
            if (this.isLoading)
                return;
            this.showLoading();
            this.noticeOpenVideo(reason);
            if (!this._videoNode) {
                console.log("激励视频广告id:" + this.rewardAd);
                this._videoNode = this.hbs.createRewardedVideoAd({ adUnitId: this.rewardAd });
                this._videoNode.onLoad((e) => {
                    console.log("视频加载成功:", JSON.stringify(e));
                    if (self.showTime)
                        clearTimeout(self.showTime);
                    self.hideLoading();
                    self.showVidwo(reason);
                    ToolSDK.stopSound();
                });
                this._videoNode.onError((err) => {
                    console.log("视频加载失败:", JSON.stringify(err));
                    if (self.showTime)
                        clearTimeout(self.showTime);
                    self.hideLoading();
                    self.onADViewComplete.runWith(false);
                    self.showToast("暂无视频", 2000);
                    ToolSDK.replaySound();
                });
            }
            if (this.showTime)
                clearTimeout(this.showTime);
            this._videoNode.load();
            this.showTime = setTimeout(this.hideLoading, 3000);
        }
        showVidwo(reason) {
            var self = this;
            if (this.isShowBanner)
                this._bannerNode.hide();
            this._videoNode.show();
            this._closeCallback && this._videoNode.offClose(this._closeCallback);
            this._closeCallback = function (res) {
                console.log("激励视频回调====================>");
                console.log(JSON.stringify(res));
                if (self.isShowBanner)
                    self._bannerNode.show();
                if (res && res.isEnded || res === undefined) {
                    self.noticeCloseVideo(reason);
                    self.onADViewComplete.runWith(true);
                }
                else {
                    self.onADViewComplete.runWith(false);
                }
                ToolSDK.replaySound();
            };
            this._videoNode.onClose(this._closeCallback);
        }
        showLoading() {
            this.isLoading = true;
            if (this.hbs && this.hbs.showLoading) {
                this.hbs.showLoading({ title: "请求中，请稍后..." });
            }
        }
        hideLoading() {
            this.isLoading = false;
            if (this.hbs && this.hbs.hideLoading) {
                this.hbs.hideLoading({});
            }
        }
        showToast(message, time) {
            this.hbs && this.hbs.showToast && this.hbs.showToast({
                title: message,
                icon: "none",
                duration: time
            });
        }
        createNative(navigateId) {
            if (!this.hbs)
                return null;
            console.log("广告id:" + navigateId);
            let nativeNode = this.hbs.createNativeAd({
                adUnitId: this.nativeAd
            });
            return nativeNode;
        }
        get addelay() {
            if (!this.data)
                return 0;
            return this.data.textDelay * 1000;
        }
        vibrateShort() {
            this.hbs.vibrateShort({
                success: function () {
                    console.log("log----vibrateShort-----");
                },
                fail: function () { },
                complete: function () { },
            });
        }
        vibrateLong() {
            this.hbs.vibrateLong({
                success: function () {
                    console.log("log----vibrateLong-----");
                },
                fail: function () { },
                complete: function () { },
            });
        }
        loadSubpackage(pack, zipFileName, topath, comlete, progress) {
            console.log("开始加载分包=========================》");
            this.hbs.getSystemInfoSync();
            let loadTask = this.hbs.loadSubpackage({
                subpackage: pack,
                success: function (info) {
                    console.log("分包加载成功=========================》");
                    console.log(pack);
                    console.log(info);
                    if (comlete) {
                        comlete(info);
                    }
                },
                fail: function (info) {
                    console.log("开始加载分包=========================》");
                    console.log(pack);
                    console.log(info);
                },
                complete: function () {
                }
            });
            loadTask.onprogress(res => {
                console.log('下载进度', res.progress);
                console.log('已经下载的数据长度', res.loadSize);
                console.log('预期需要下载的数据总长度', res.totalSize);
                if (progress) {
                    progress(res);
                }
            });
        }
        showADInsert() {
            if (!this.hbs) {
                return;
            }
            if (!this.hasLogin) {
                return;
            }
            console.log("ad-展示插屏");
            let insertAdNode = this.hbs.createInterstitialAd({
                posId: this.insertAd,
            });
            insertAdNode.onError(err => {
                console.log("插屏广告加载失败", err);
            });
            insertAdNode.onLoad(function (data) {
                console.log('onLoad data ' + JSON.stringify(data));
                insertAdNode.show();
            });
            insertAdNode.load();
        }
        openYinsi() {
            console.log("打开隐私协议");
            UIManager.inst.ShowUI(UIYinSi, UILayer.base);
        }
    }
    Hbs.CLICK_INTERVAL = 1000;
    Hbs.lastClickTime = 0;

    class ViewBase extends fgui.GComponent {
        constructor() {
            super(...arguments);
            this.ServerConfigKey = "";
        }
        static find_parent_component_once_node(node, comp) {
            if (!(node instanceof fairygui.GComponent))
                return null;
            let parent = node.parent;
            while (parent) {
                if (parent instanceof comp) {
                    break;
                }
                parent = parent.parent;
            }
            return parent;
        }
        constructFromXML() {
            super.constructFromXML(undefined);
            this.displayObject.onAwake = this._onAwake.bind(this);
            this.displayObject.onDestroy = this._onDestroy.bind(this);
            if (this.onClickByView)
                this.onClick(this, this.onClickByView);
            if (this.onCreate)
                this.onCreate();
        }
        _onAwake() {
            if (this.ServerConfigKey && !ViewBase.find_parent_component_once_node(this, ViewBase)) {
                if (!GameSDK.isConfigUIModel(this.ServerConfigKey))
                    this.visible = false;
            }
            if (this.onLoaded)
                this.onLoaded();
        }
        _onDestroy() {
            Laya.timer.clearAll(this);
            if (this.onDestroyed)
                this.onDestroyed();
        }
        random(min, max) {
            return Math.round(Math.random() * (max - min) + min);
        }
        randomItem(array) {
            if (array && array.length > 0) {
                return array[this.random(0, array.length - 1)];
            }
            else {
                return null;
            }
        }
    }

    class AvdCell extends ViewBase {
        onCreate() {
            this._icon = this.getChildByPath("img").asLoader;
            this._nameLab = this.getChildByPath("name").asTextField;
        }
        onClickByView() {
            if (this._data) {
                GameSDK.navigateToMiniGame(this._data);
            }
        }
        updataData(data) {
            if (this._data == data) {
                return;
            }
            this._data = data;
            if (data) {
                let bgs = [];
                for (var i = 1; i <= 6; ++i) {
                    let bg = this.getChild("bg" + i);
                    if (bg)
                        bgs.push(bg.asImage);
                }
                let index = this.random(0, bgs.length - 1);
                bgs.map((a, i) => {
                    a.visible = i == index;
                });
                this._icon.url = this.randomItem(data.iconPath);
                this._nameLab.text = data.name;
            }
        }
    }

    class AvdFloat extends ViewBase {
        constructor() {
            super(...arguments);
            this.ServerConfigKey = "is_AvdFloat";
        }
        onCreate() {
            this._icon = this.getChildByPath("icon.img").asLoader;
            this._nameLab = this.getChildByPath("name").asTextField;
            let timeOut = 3;
            if (this.data) {
                parseFloat(this.data.toString());
            }
            this._Listata = GameSDK.getNavigateList();
            if (this._Listata.length <= 0)
                return;
            Laya.timer.clear(this, this._freamtimeHandler);
            Laya.timer.loop(timeOut * 1000, this, this._freamtimeHandler);
            this._freamtimeHandler();
        }
        onClickByView(evet) {
            evet._stoped = true;
            if (this._data) {
                GameSDK.navigateToMiniGame(this._data);
            }
        }
        _freamtimeHandler() {
            let index = Math.floor(Math.random() * this._Listata.length);
            let info = this._Listata[index];
            this._icon.url = this.randomItem(info.iconPath);
            this._nameLab.text = info.name;
            this._data = info;
        }
    }

    class AvdList extends ViewBase {
        constructor() {
            super(...arguments);
            this.ServerConfigKey = "is_AvdList";
            this._data = [];
        }
        onCreate() {
            this._list = this.getChildByPath("leftBoxList").asList;
            this._list.itemRenderer = Laya.Handler.create(this, this.updataCellHandler, null, false);
            this.changeAll();
            Laya.timer.loop(3 * 1000, this, this.changeAll);
        }
        changeAll() {
            this._data = [];
            let list = GameSDK.getNavigateList();
            let _index = this.random(0, list.length - 5);
            if (_index < 0)
                return;
            for (var i = _index; i < _index + 4; ++i) {
                if (list[i])
                    this._data.push(list[i]);
            }
            this._list.numItems = this._data.length;
            var children = this._list._children;
            var index = Math.floor(Math.random() * children.length);
            let tmp = children[index];
            this.aniro(tmp, 10, () => {
                this.aniro(tmp, -10, () => {
                    this.aniro(tmp, 10, () => {
                        this.aniro(tmp, -10, () => {
                            this.aniro(tmp, 0, () => {
                            });
                        });
                    });
                });
            });
        }
        aniro(node, angle, callback) {
            Laya.Tween.to(node, { rotation: angle }, 300, null, Laya.Handler.create(this, () => {
                callback();
            }, null, false));
        }
        updataCellHandler(index, view) {
            view.updataData(this._data[index]);
        }
    }

    class IntervalDisplayAD {
        constructor(ServerConfigKey) {
            this.ServerConfigKey = "";
            this.last_display_time = 0;
            this.is_check = true;
            this.ServerConfigKey = ServerConfigKey;
            this.last_display_time = 0;
        }
        has() {
            if (!this.is_check)
                return true;
            let is_visible = GameSDK.isConfigUIModel(this.ServerConfigKey);
            if (!is_visible)
                return false;
            let interval = GameSDK.getConfigUIModelInterval(this.ServerConfigKey);
            if (this.ServerConfigKey == "" || interval == 0) {
                return true;
            }
            let now = Math.floor(Date.now() / 1000);
            let sub = 0;
            if (is_visible) {
                sub = now - this.last_display_time;
                if (interval != 0) {
                    if (sub >= interval) {
                        sub = now - IntervalDisplayAD.last_display_time;
                        interval = GameSDK.sdk.data[IntervalDisplayAD.ServerConfigKey];
                        if (typeof interval == "string")
                            interval = Number(interval);
                        else if (interval == undefined)
                            interval = 0;
                        if (sub >= interval) {
                            is_visible = true;
                        }
                        else {
                            is_visible = false;
                            console.log("后台控制广告显隐-", this.ServerConfigKey, "-全局间隔时间未达到", sub, interval);
                        }
                    }
                    else
                        is_visible = false;
                }
                this.last_display_time = now;
            }
            if (is_visible) {
                IntervalDisplayAD.last_display_time = now;
                console.log("后台控制广告显隐-", this.ServerConfigKey, "-: visible", is_visible, interval, sub);
            }
            return is_visible;
        }
    }
    IntervalDisplayAD.last_display_time = 0;
    IntervalDisplayAD.ServerConfigKey = "ad_all_interval";

    class MoreGameBtn extends ViewBase {
        onCreate() {
            if (GameSDK.type == PlatType.TOUTIAO) {
                let systemInfo = window["tt"].getSystemInfoSync();
                var isDouyin = systemInfo.appName === "Douyin";
                this.visible = isDouyin && !GameSDK.isCurrentVersion && MoreGameBtn.IntervalDisplayAD.has();
            }
        }
        onClickByView() {
            if (GameSDK.type == PlatType.OPPO) {
                GameSDK.showGamePortal();
            }
            else if (GameSDK.type == PlatType.WECHAT) {
                SDKViewMgr.popUpView(PropType.SmallItem);
            }
            else {
                SDKViewMgr.moreView();
            }
        }
    }
    MoreGameBtn.ServerConfigKey = "is_MoreGameBtn";
    MoreGameBtn.IntervalDisplayAD = new IntervalDisplayAD(MoreGameBtn.ServerConfigKey);

    class PropSceneUI extends ViewBase {
        constructor() {
            super(...arguments);
            this.ServerConfigKey = PropSceneUI.ServerConfigKey;
            this.navigatedata = [];
            this.pasue_end_time = 0;
            this.index = 0;
            this._dirction = 1;
            this.maxPosY = 10;
            this.posY = 0;
            this._list_speed = 3;
            this.level = Number.MAX_SAFE_INTEGER;
            this.bclick = false;
        }
        moveDownHandler(evet) {
            this.pasue_end_time = Date.now() + 300;
            Laya.stage.once(Laya.Event.MOUSE_UP, this, this.moveUpHandler);
        }
        moveUpHandler(evet) {
            Laya.stage.off(Laya.Event.MOUSE_UP, this, this.moveUpHandler);
            this.posY = this._list.scrollPane.posY;
        }
        randomOpenView() {
            if (GameSDK.isOffCitys()) {
                console.log("城市屏蔽");
                return;
            }
            if (GameSDK.isCurrentVersion) {
                console.log("版本屏蔽");
                return;
            }
            var children = this._list._children;
            var index = Math.floor(Math.random() * children.length);
            children[index].onClickByView();
        }
        enterframe() {
            if (Date.now() < this.pasue_end_time)
                return;
            this.posY = this._list.scrollPane.posY;
            this.posY += this._list_speed;
            if (this.posY >= this.maxPosY) {
                this.posY = this.maxPosY;
                this._list_speed *= -1;
            }
            else if (this.posY <= 0) {
                this.posY = 0;
                this._list_speed *= -1;
            }
            this._list.scrollPane.setPosY(this.posY);
        }
        onCreate() {
            SDKEventMgr.Int.on(PlatEventName.PLAT_TOUCH_BANNER_AD, this.touch_banner_close, this);
            GameSDK.hideBanner();
            this.on(Laya.Event.MOUSE_DOWN, this, this.moveDownHandler);
            this.getChild("btn_continue").onClick(this, this.click_continue);
            this.navigatedata = GameSDK.getNavigateList();
            this._list = this.getChild("list").asList;
            this._list.itemRenderer = Laya.Handler.create(this, this.updataCellHandler, null, false);
            this._list.numItems = this.navigatedata.length;
            Laya.timer.frameLoop(1, this, this.enterframe);
            Laya.timer.frameOnce(1, this, () => {
                this.maxPosY = this._list.scrollPane["_overlapSize"].y;
            });
            this.randomOpenView();
        }
        touch_banner_close() {
            PropSceneUI.lastbannertime = Date.now();
            SDKEventMgr.Int.off(PlatEventName.PLAT_TOUCH_BANNER_AD, this.touch_banner_close, this);
            GameSDK.hideBanner();
            GameSDK.updateErrorCount(ErrorCountType.BANNER);
            if (this._completeHandler) {
                this._completeHandler();
                this._completeHandler = null;
            }
            console.log("PropSceneUI----销毁");
            this.dispose();
        }
        onDestroyed() {
            console.log("-------------onDestroyed");
            SDKEventMgr.Int.off(PlatEventName.PLAT_TOUCH_BANNER_AD, this.touch_banner_close, this);
        }
        updataCellHandler(index, view) {
            view.updataData(this.navigatedata[index]);
        }
        setCallback(level, callBack) {
            GameSDK.hideBanner();
            this.level = level;
            this._completeHandler = callBack;
        }
        click_continue() {
            if (!this.bclick && Date.now() - PropSceneUI.lastbannertime > 10000 && GameSDK.canClick({ level: this.level }, ClickType.WaterFall)) {
                this.getChild("btn_continue").offClick(this, this.click_continue);
                this.bclick = true;
                console.log("banner");
                Laya.timer.once(800, this, () => {
                    GameSDK.showBanner();
                    Laya.timer.once(1800, this, () => {
                        GameSDK.hideBanner();
                        this.getChild("btn_continue").onClick(this, this.click_continue);
                    });
                });
            }
            else {
                GameSDK.hideBanner();
                if (this._completeHandler) {
                    this._completeHandler();
                    this._completeHandler = null;
                }
                SDKEventMgr.Int.off(PlatEventName.PLAT_TOUCH_BANNER_AD, this.touch_banner_close.bind(this), this);
                this.dispose();
            }
        }
    }
    PropSceneUI.ServerConfigKey = "is_PropSceneUI";
    PropSceneUI.IntervalDisplayAD = new IntervalDisplayAD(PropSceneUI.ServerConfigKey);
    PropSceneUI.lastbannertime = 0;

    class NativeBanner extends ViewBase {
        constructor() {
            super(...arguments);
            this._info = { NativeType: 0 };
        }
        onCreate() {
            this._iconLoader = this.getChildByPath("icon").asLoader;
            this.getChildByPath("closebtn").asCom.onClick(this, this.clickViewHandler);
            if (this.data) {
                try {
                    this._info = JSON.parse(this.data.toString());
                }
                catch (error) {
                    this._info = { NativeType: 0 };
                }
            }
            if (!this._info)
                this._info = { NativeType: 0 };
            this.createNative();
        }
        showView() {
            if (!GameSDK.isShowTime()) {
                Laya.timer.callLater(this, this.removeView);
            }
            else {
                if (this._nativeNode)
                    this._nativeNode.load();
            }
        }
        clickNativeBannerView() {
            this.removeView();
            this.reportAdClick();
        }
        clickViewHandler(evet) {
            console.log("关闭界面");
            evet._stoped = true;
            this.removeView();
        }
        createNative() {
            let self = this;
            let data = GameSDK.data;
            let nativeId = data.NativeAd;
            this._nativeNode = GameSDK.createNative(nativeId);
            if (!this._nativeNode)
                return;
            this._nativeNode.onError(function (e) {
                console.log("=========>oppo 原生广告加载失败", e);
                self.displayObject.event("LoadFailNativeBanner");
            });
            this._nativeNode.onLoad(function (res) {
                console.log("=========>oppo 原生广告加载成功", res);
                if (res && res.adList) {
                    let nativeInfo = res.adList[res.adList.length - 1];
                    console.log("原生广告数据：", nativeInfo);
                    self.initNatveView(nativeInfo);
                    self.displayObject.event("LoadSuccessNativeBanner");
                }
            });
        }
        initNatveView(nativeInfo) {
            if (nativeInfo.imgUrlList && nativeInfo.imgUrlList.length > 0) {
                this._iconLoader.url = nativeInfo.imgUrlList[0];
            }
            else if (nativeInfo.iconUrlList && nativeInfo.iconUrlList.length > 0) {
                this._iconLoader.url = nativeInfo.iconUrlList[0];
            }
            else {
                this.removeView();
                return;
            }
            this._useNtAdId = nativeInfo.adId;
            this.reportAdShow();
        }
        reportAdClick() {
            if (this._useNtAdId != null && this._nativeNode)
                this._nativeNode.reportAdClick({ adId: this._useNtAdId.toString() });
        }
        reportAdShow() {
            if (this._useNtAdId != null && this._nativeNode)
                this._nativeNode.reportAdShow({ adId: this._useNtAdId.toString() });
        }
        removeView() {
            this.displayObject.event("CloseNativeBanner");
            GameSDK.GameNativeInsert && GameSDK.GameNativeInsert.setIsShow(false);
        }
    }

    class AddDesk extends ViewBase {
        constructor() {
            super(...arguments);
            this.ServerConfigKey = "is_AddDesk";
        }
        onCreate() {
            if (GameSDK.isShortCutInstalled) {
                this.visible = false;
            }
        }
        onClickByView() {
            GameSDK.shortCutInstalled((is_succes) => {
                console.log("添加到桌面:", is_succes);
            });
        }
    }

    class GameNative extends ViewBase {
        constructor() {
            super(...arguments);
            this.ServerConfigKey = GameNative.ServerConfigKey;
        }
        onCreate() {
            this._showNative = this.getChild("n0");
        }
        showNativeView(avdBtn, continueBtn, compete, loadSucces, loadFail, plevel) {
            this._showNativeBtn = avdBtn;
            this._continueBtn = continueBtn;
            this._showNativeBtn.onClick(this, this.onShowNativeBannerHandler);
            if (this._continueBtn)
                this._continueBtn.onClick(this, this.onContinueHandler);
            this._cometeHandler = compete;
            this._loadFail = loadFail;
            this._loadSucces = loadSucces;
            this._showNative.displayObject.on("CloseNativeBanner", this, this.onCloseHandler);
            this._showNative.displayObject.on("LoadFailNativeBanner", this, this.onLoadFailHandler);
            this._showNative.displayObject.on("LoadSuccessNativeBanner", this, this.onLoadSuccessHandler);
            this._level = plevel;
            this._showNative.showView();
        }
        onShowNativeBannerHandler() {
            this._showNative.clickNativeBannerView();
        }
        onContinueHandler() {
            if (GameSDK.canClick({ "level": this._level }, ClickType.NativeAd)) {
                this._showNative.clickNativeBannerView();
            }
            else {
                this.onCloseHandler();
            }
        }
        onCloseHandler() {
            this._showNative.displayObject.off("LoadFailNativeBanner", this, this.onLoadFailHandler);
            this._showNative.displayObject.off("CloseNativeBanner", this, this.onCloseHandler);
            this._showNative.displayObject.off("LoadSuccessNativeBanner", this, this.onCloseHandler);
            if (this._cometeHandler) {
                this._cometeHandler();
            }
        }
        onLoadFailHandler() {
            console.log("原生广告成功回调===> 失败 : ", this._loadFail);
            this._showNative.displayObject.off("LoadFailNativeBanner", this, this.onLoadFailHandler);
            this._showNative.displayObject.off("CloseNativeBanner", this, this.onCloseHandler);
            this._showNative.displayObject.off("LoadSuccessNativeBanner", this, this.onCloseHandler);
            if (this._loadFail) {
                this._loadFail();
            }
        }
        onLoadSuccessHandler() {
            console.log("原生广告成功回调：", this._loadSucces);
            this._showNative.displayObject.off("LoadFailNativeBanner", this, this.onLoadFailHandler);
            this._showNative.displayObject.off("LoadSuccessNativeBanner", this, this.onCloseHandler);
            if (this._loadSucces) {
                this._loadSucces();
            }
        }
    }
    GameNative.ServerConfigKey = "is_GameNative";
    GameNative.IntervalDisplayAD = new IntervalDisplayAD(GameNative.ServerConfigKey);

    class ButtonEffect extends Laya.Script {
        onAwake() {
            this.node = fgui.GObject.cast(this.owner);
        }
        onEnable() {
            if (this._tween)
                this._tween.clear();
            if (GameSDK.isShow) {
                Laya.timer.once(1300, this, () => {
                    this.showBtnMaxEffect();
                });
            }
        }
        showBtnMaxEffect() {
            if (this._tween)
                this._tween.clear();
            this._tween = Laya.Tween.to(this.node, { scaleX: 1.2, scaleY: 1.1 }, 600, null, Laya.Handler.create(this, () => {
                this._tween.clear();
                this.showBtnMinEffect();
            }));
        }
        showBtnMinEffect() {
            if (this._tween)
                this._tween.clear();
            this._tween = Laya.Tween.to(this.node, { scaleX: 1, scaleY: 1 }, 600, null, Laya.Handler.create(this, () => {
                this._tween.clear();
                this.showBtnMaxEffect();
            }));
        }
        onDestroy() {
            if (this._tween)
                this._tween.clear();
        }
    }
    class YButtonEffect extends Laya.Script {
        onAwake() {
            this.node = fgui.GObject.cast(this.owner);
        }
        onEnable() {
            if (this._tween)
                this._tween.clear();
            if (GameSDK.isShow) {
                Laya.timer.once(300, this, () => {
                    this.showBtnMaxEffect();
                });
            }
        }
        showBtnMaxEffect() {
            if (this._tween)
                this._tween.clear();
            this._tween = Laya.Tween.to(this.node, { scaleX: 1, scaleY: 0.9 }, 600, null, Laya.Handler.create(this, () => {
                this._tween.clear();
                this.showBtnMinEffect();
            }));
        }
        showBtnMinEffect() {
            if (this._tween)
                this._tween.clear();
            this._tween = Laya.Tween.to(this.node, { scaleX: 0.8, scaleY: 0.8 }, 600, null, Laya.Handler.create(this, () => {
                this._tween.clear();
                this.showBtnMaxEffect();
            }));
        }
        onDestroy() {
            if (this._tween)
                this._tween.clear();
        }
    }

    class ClickGift extends ViewBase {
        constructor() {
            super(...arguments);
            this.progress = null;
            this.tip1 = null;
            this.tip2 = null;
            this.callback = null;
            this.n9 = null;
            this.n3 = null;
            this.showBannerRatio = 50;
            this.pause = false;
            this.target_value = 0;
        }
        onCreate() {
            SDKEventMgr.Int.on(PlatEventName.PLAT_TOUCH_BANNER_AD, this.touch_banner_close, this);
            this.progress = this.getChildInGroup("progress", this.getChild("group1").asGroup).asProgress;
            this.n3 = this.getChildInGroup("n3", this.getChild("group1").asGroup).asImage;
            this.n9 = this.getChildInGroup("n9", this.getChild("group1").asGroup).asImage;
            let gift = this.random(0, 3);
            [this.getChildInGroup("chest1", this.getChild("group1").asGroup),
                this.getChildInGroup("chest2", this.getChild("group1").asGroup),
            ].map((a, i) => {
                a.visible = i == gift;
            });
            this.progress.value = 0;
            this.getChild("btn_click").onClick(this, this.btn_click);
            this.tip1 = this.getChild("tip1").asImage;
            this.tip2 = this.getChild("tip2").asImage;
            this.tip2.visible = false;
            this.getChild("btn_click").displayObject.addComponent(ButtonEffect);
            Laya.timer.frameLoop(1, this, this.enterframe);
            Laya.timer.loop(500, this, () => {
                if (this.pause) {
                    this.tip1.visible = false;
                    this.tip2.visible = false;
                }
                else {
                    this.tip1.visible = !this.tip1.visible;
                    this.tip2.visible = !this.tip1.visible;
                }
            });
        }
        enterframe() {
            this.n3.rotation++;
            this.n9.rotation += 0.8;
            if (this.pause)
                return;
            this.target_value -= 0.8;
            if (this.target_value < 0)
                this.target_value = 0;
            this.progress.value += (this.target_value - this.progress.value) * 0.2;
        }
        touch_banner_close() {
            SDKEventMgr.Int.off(PlatEventName.PLAT_TOUCH_BANNER_AD, this.touch_banner_close, this);
            GameSDK.updateErrorCount(ErrorCountType.BANNER);
            this.dispose();
            if (this.callback) {
                this.callback();
                this.callback = null;
            }
        }
        btn_click() {
            if (this.pause)
                return;
            this.target_value += 20;
            if (this.target_value >= this.showBannerRatio) {
                GameSDK.showBanner();
                Laya.timer.once(300, this, () => {
                    GameSDK.hideBanner();
                });
                if (this.showBannerRatio == 50) {
                    this.showBannerRatio = 75;
                }
                else if (this.showBannerRatio == 75) {
                    this.showBannerRatio = 90;
                }
                else if (this.target_value >= 100) {
                    this.pause = true;
                    this.progress.value = 100;
                    GameSDK.showADVideo(this, (state) => {
                        this.touch_banner_close();
                    }, "点击开宝箱");
                }
            }
            console.log("btn_click");
        }
    }

    class MoreGameTT extends ViewBase {
        constructor() {
            super(...arguments);
            this.callback = null;
            this._list = null;
            this.navigatedata = [];
        }
        onCreate() {
            GameSDK.banner.hide();
            this.getChild("btn_close").onClick(this, this.onClickCloseHandler);
            this._hotItem = this.getChild("btn_zuire");
            this._newItem = this.getChild("btn_zuixin");
            this.navigatedata = GameSDK.getNavigateList();
            this._list = this.getChild("list_icon").asList;
            this._list.itemRenderer = Laya.Handler.create(this, this.updataCellHandler, null, false);
            this._list.numItems = this.navigatedata.length;
            this.initHotItem();
            this.initNewItem();
        }
        initHotItem() {
            for (let i = 0; i < this.navigatedata.length; i += 1) {
                let temp = this.navigatedata[i];
                if (temp.weight == 99) {
                    this._hotItem.updataData(temp);
                    break;
                }
                else if (i == this.navigatedata.length - 1) {
                    this._hotItem.updataData(this.navigatedata[0]);
                }
            }
        }
        initNewItem() {
            for (let i = 0; i < this.navigatedata.length; i += 1) {
                let temp = this.navigatedata[i];
                if (temp.weight == 88) {
                    this._newItem.updataData(temp);
                    break;
                }
                else if (i == this.navigatedata.length - 1) {
                    this._newItem.updataData(this.navigatedata[0]);
                }
            }
        }
        updataCellHandler(index, view) {
            view.updataData(this.navigatedata[index]);
        }
        onClickCloseHandler() {
            GameSDK.banner.show();
            this.dispose();
            if (this.callback) {
                this.callback();
                this.callback = null;
            }
        }
    }
    MoreGameTT.ServerConfigKey = "is_MoreGameBtn";
    MoreGameTT.IntervalDisplayAD = new IntervalDisplayAD(MoreGameTT.ServerConfigKey);

    class AvdCellTTSmall extends ViewBase {
        onCreate() {
            this._icon = this.getChildByPath("comp_icon.loader_icon").asLoader;
            this._nameLab = this.getChildByPath("text_name").asTextField;
        }
        onClickByView() {
            if (this._data) {
                GameSDK.navigateToMiniGame(this._data);
                this.touchable = false;
                Laya.timer.once(100, this, () => {
                    this.touchable = true;
                });
            }
        }
        updataData(data) {
            if (this._data == data) {
                return;
            }
            this._data = data;
            if (data) {
                this._icon.url = this.randomItem(data.iconPath);
                this._nameLab.text = data.name;
            }
        }
    }

    class AvdCellTTItem extends ViewBase {
        onCreate() {
            this._icon = this.getChildByPath("btn_icon.loader_icon").asLoader;
            this._nameLab = this.getChildByPath("text_name").asTextField;
            this._controller = this.getController("c1");
        }
        onClickByView() {
            if (this._data) {
                GameSDK.navigateToMiniGame(this._data);
                this.touchable = false;
                Laya.timer.once(100, this, () => {
                    this.touchable = true;
                });
            }
        }
        updataData(data) {
            if (this._data == data) {
                return;
            }
            this._data = data;
            if (data) {
                this._icon.url = this.randomItem(data.iconPath);
                this._nameLab.text = data.name;
                if (this._data.weight == 99) {
                    this._controller.setSelectedIndex(1);
                }
                else if (this._data.weight == 88) {
                    this._controller.setSelectedIndex(2);
                }
                else {
                    this._controller.setSelectedIndex(0);
                }
            }
        }
    }

    class SDKViewMgr {
        static init(path = "res/fui/", cb = null) {
            this._registfguiClass("moreBtn", MoreGameBtn);
            this._registfguiClass("FloatAd", AvdFloat);
            this._registfguiClass("EndListAd", AvdList);
            this._registfguiClass("boxItemBig", AvdCell);
            this._registfguiClass("boxItemSmall", AvdCell);
            this._registfguiClass("btn_small", AvdCellTTSmall);
            this._registfguiClass("comp_iconzhanshi", AvdCellTTItem);
            this._registfguiClass("com_screen_big", PropSceneUI);
            this._registfguiClass("com_screen_small", PropSceneUI);
            this._registfguiClass("com_clickgift", ClickGift);
            this._registfguiClass("comp_douyinhudao", MoreGameTT);
            this._registfguiClass("NativeAdBanner", NativeBanner);
            this._registfguiClass("Native", GameNative);
            this._registfguiClass("addDesk", AddDesk);
            fgui.UIPackage.loadPackage(path + this.packageName, Laya.Handler.create(this, cb));
        }
        static _registfguiClass(resName, type) {
            fgui.UIObjectFactory.setExtension("ui://" + this.packageName + "/" + resName, type);
        }
        static createObject(resName) {
            return fairygui.UIPackage.createObject(this.packageName, resName);
        }
        static popUpView(type, back = undefined, level = Number.MAX_SAFE_INTEGER) {
            if (type == PropType.BigItem && !PropSceneUI.IntervalDisplayAD.has()) {
                if (back)
                    back();
                return;
            }
            let view = this.createObject(type == PropType.BigItem ? "com_screen_big" : "com_screen_small");
            view.setCallback(level, back);
            view.makeFullScreen();
            fairygui.GRoot.inst.addChild(view);
        }
        static clickgift(callback = null) {
            let view = this.createObject("com_clickgift");
            view.makeFullScreen();
            view.callback = callback;
            fairygui.GRoot.inst.addChild(view);
        }
        static moreView(back = undefined) {
            if (!MoreGameTT.IntervalDisplayAD.has()) {
                if (back)
                    back();
                return;
            }
            if (GameSDK.type == PlatType.TOUTIAO) {
                let systemInfo = window["tt"].getSystemInfoSync();
                var isDouyin = systemInfo.appName === "Douyin";
                if (!isDouyin)
                    return;
            }
            let view = this.createObject("comp_douyinhudao");
            view.callback = back;
            view.makeFullScreen();
            fairygui.GRoot.inst.addChild(view);
        }
        static showNativeUI(node, ad_comp_name, btn_look, btn_continue, level, cb, su, fa) {
            let native = node.getChild(ad_comp_name);
            console.log("注册原生广告事件成功 level:", level);
            if (native.showNativeView) {
                native.showNativeView(node.getChild(btn_look).asButton, node.getChild(btn_continue).asButton, () => {
                    console.log("广告-点击继续");
                    if (cb)
                        cb();
                }, () => {
                    if (su)
                        su();
                }, () => {
                    if (fa)
                        fa();
                }, level);
            }
            else {
                console.log("没有注册原生广告 直接回调");
                if (cb)
                    cb();
            }
        }
    }
    SDKViewMgr.packageName = "gamesdk";

    let UINormalBanner = class UINormalBanner extends BaseUI {
        constructor() {
            super(...arguments);
            this.m_adid = "";
        }
        get URL() {
            gamesdkBinder.bindAll();
            return UI_NormalBanner.URL;
        }
        OnCreate() {
        }
        ReFresh() {
            try {
                console.log("test GameSDK.NativeDataArr.length----===>> " + GameSDK$1.NativeDataArr.length);
                if (GameSDK$1.NativeDataArr.length > 0) {
                    let index = this.GetIndex();
                    let nativeInfo = GameSDK$1.NativeDataArr[index];
                    let Rurl;
                    if (nativeInfo.iconUrlList && nativeInfo.iconUrlList.length > 0) {
                        Rurl = nativeInfo.iconUrlList[0];
                    }
                    else if (nativeInfo.imgUrlList && nativeInfo.imgUrlList.length > 0) {
                        Rurl = nativeInfo.imgUrlList[0];
                    }
                    this.ui.m_normalBanner.m_RealNativeBanner.m_micon.url = Rurl;
                    this.ui.m_normalBanner.m_RealNativeBanner.m_mtxt.text = nativeInfo.title;
                    this.m_adid = nativeInfo.adId;
                    GameSDK$1.GameNativeInsert.reportAdShow(this.m_adid);
                }
            }
            catch (error) {
                console.log(" err===>>", error);
            }
        }
        GetIndex() {
            GameSDK$1.NativeIndex++;
            let index = GameSDK$1.NativeIndex % GameSDK$1.NativeDataArr.length;
            return index;
        }
        OnShow(...args) {
            this.ui.m_normalBanner.visible = true;
            GameSDK$1.BannerPanelIsShow = true;
            console.log("----------UINormalBanner-@@@------------  ");
            this.ReFresh();
            GameSDK$1.GameNativeInsert.reportAdShow(this.m_adid);
            this.ui.m_normalBanner.m_RealNativeBanner.m_micon.onClick(this, () => {
                console.log("test-------m_adid--" + this.m_adid);
                GameSDK$1.GameNativeInsert.reportAdClick(this.m_adid);
            });
            this.ui.m_normalBanner.m_RealNativeBanner.m_mclose.onClick(this, () => {
                UIManager.inst.PopUI(this.layer, false);
            });
        }
        OnHide() {
            GameSDK$1.BannerPanelIsShow = false;
            this.ui.m_normalBanner.visible = false;
        }
        OnDispose() {
        }
        onSwitch() {
        }
        onAniEnd() {
        }
        updateScene(offset = 1) {
        }
    };
    UINormalBanner = __decorate([
        ui_register()
    ], UINormalBanner);

    var PlatType$1;
    (function (PlatType) {
        PlatType[PlatType["AUTO"] = -2] = "AUTO";
        PlatType[PlatType["TEST"] = -1] = "TEST";
        PlatType[PlatType["WEB"] = 0] = "WEB";
        PlatType[PlatType["TOUTIAO"] = 1] = "TOUTIAO";
        PlatType[PlatType["VIVO"] = 2] = "VIVO";
        PlatType[PlatType["OPPO"] = 3] = "OPPO";
        PlatType[PlatType["WECHAT"] = 5] = "WECHAT";
        PlatType[PlatType["HBS"] = 8] = "HBS";
        PlatType[PlatType["GAME233"] = 9] = "GAME233";
        PlatType[PlatType["GAMEUC"] = 10] = "GAMEUC";
        PlatType[PlatType["KUAIGAME"] = 11] = "KUAIGAME";
        PlatType[PlatType["MOMOYU"] = 13] = "MOMOYU";
        PlatType[PlatType["OPOPOAPP"] = 14] = "OPOPOAPP";
    })(PlatType$1 || (PlatType$1 = {}));
    var PropType$1;
    (function (PropType) {
        PropType[PropType["BigItem"] = 0] = "BigItem";
        PropType[PropType["SmallItem"] = 1] = "SmallItem";
    })(PropType$1 || (PropType$1 = {}));
    var ClickType;
    (function (ClickType) {
        ClickType[ClickType["NativeAd"] = 0] = "NativeAd";
        ClickType[ClickType["PlayGame"] = 1] = "PlayGame";
        ClickType[ClickType["WaterFall"] = 2] = "WaterFall";
        ClickType[ClickType["EnterGame"] = 3] = "EnterGame";
    })(ClickType || (ClickType = {}));
    var PlatEventName;
    (function (PlatEventName) {
        PlatEventName["PLAT_HIDE"] = "PLAT_HIDE";
        PlatEventName["PLAT_CLOSE"] = "PLAT_CLOSE";
        PlatEventName["PLAT_TOUCH_BANNER_AD"] = "PLAT_TOUCH_BANNER_AD";
        PlatEventName["PLAT_HOME"] = "PLAT_HOME";
        PlatEventName["PLAT_TOUSU"] = "PLAT_TOUSU";
        PlatEventName["PLAT_JUMP_GAME"] = "PLAT_JUMP_GAME";
        PlatEventName["PLAT_SHARE"] = "PLAT_SHARE";
        PlatEventName["PLAT_SHARE_SNS"] = "PLAT_SHARE_SNS";
        PlatEventName["PLAT_ABOUT"] = "PLAT_ABOUT";
        PlatEventName["PLAT_CONFIG"] = "PLAT_CONFIG";
        PlatEventName["PLAT_COLLECT"] = "PLAT_COLLECT";
    })(PlatEventName || (PlatEventName = {}));
    class GameSDK$1 {
        static _checkShowClickGift() {
            let time = this.data.bannerInterval || 7;
            if (this.bxCount > 0 && this.bxCount % time == 0) {
                this.bxCount = 0;
                return true;
            }
            return false;
        }
        static CommonShowNative() {
            if (GameSDK$1.NativeDataArr.length > 0) {
                console.log("setting@@--11------------");
                if (GameSDK$1.Can_Ys_show) {
                    UIManager.inst.PushUI(UINormalBanner, UILayer.mid);
                }
            }
        }
        static OpenNativeInsert() {
            GameSDK$1.GameNativeBanner.load();
            console.log(" GameSDK.NativeDataArr2.length==>>" + GameSDK$1.NativeDataArr2.length);
            console.log(" GameSDK.NativeDataArr2==>>" + JSON.stringify(GameSDK$1.NativeDataArr2));
            if (GameSDK$1.NativeDataArr2.length > 0) {
                console.log("GameSDK.NativeDataArr.length > 0 ");
                if (!this.isVerify()) {
                    console.log("test---isVerify ");
                    if (GameSDK$1.data["YS_CP"][0] != 0) {
                        console.log(" YS_CP==1");
                        if (this.JudgeSHowInsert()) {
                            console.log("   JudgeSHowInsert===true");
                            Laya.timer.once(Number(GameSDK$1.data["YS_CP"][1]) + 10, this, () => {
                                console.log("  延迟  打开按钮 ");
                                UIManager.inst.PushUI(UINativeInsert, UILayer.adTop);
                            });
                        }
                    }
                    else {
                        console.log(" YS_CP==0");
                    }
                }
            }
        }
        static JudgeSHowInsert() {
            let curTime = new Date().getTime();
            let span = (curTime - GameSDK$1.InsertTime) / 1000;
            if (span > GameSDK$1.data["YS_CP"][2]) {
                GameSDK$1.InsertTime = curTime;
                return true;
            }
            return false;
        }
        static isVerify() {
            if (GameSDK$1.isCurrentVersion) {
                return true;
            }
            if (GameSDK$1.isOffCitys()) {
                return true;
            }
            return false;
        }
        static ctor() {
            let value = SDKData.data;
            this.ctorData = value;
            for (let key in value)
                if (value[key] != undefined && SDKConfig[key] != undefined)
                    SDKConfig[key] = value[key];
            if (value.PlatType == PlatType$1.AUTO)
                value.PlatType = ToolSDK.GET_PLT();
            GameSDK$1.type = value.PlatType;
            SxyGame.instance.init();
            this.startGameTime = Date.now();
            let js = Web;
            switch (GameSDK$1.type) {
                case PlatType$1.TEST:
                    js = Web;
                    break;
                case PlatType$1.WEB:
                    js = Web;
                    break;
                case PlatType$1.WECHAT:
                    js = WeChat;
                    break;
                case PlatType$1.HBS:
                    js = Hbs;
            }
            this.sdk = new js();
            if (value.PltDefaultConfig) {
                let defaultConfig = value.PltDefaultConfig.filter(v => v.key == this.type);
                if (defaultConfig.length > 0)
                    this.sdk.data = defaultConfig[0].value;
            }
        }
        static initConfig(cb) {
            this.sdk.initConfig(() => {
                SDKViewMgr.init(this.ctorData.fgui_sdk_path || this.ctorData.sdk_res_path, cb);
                cb();
            });
        }
        static get sdkRes() {
            return this.ctorData.fgui_sdk_path || this.ctorData.sdk_res_path;
        }
        static huaWeiLogin(succedFun, FailFun, caller) {
            this.sdk.huaWeiLogin(succedFun, FailFun, caller);
        }
        static init(type = PlatType$1.TEST, callBack = null, fgui_sdk_path = "res/fgui/") {
            this.ctor({
                PlatType: type,
                callBack,
                fgui_sdk_path,
            });
        }
        static get data() {
            return this.sdk.data;
        }
        static playerLevel(level) {
            this.level = level;
            console.log("sdk玩家等级：" + level);
        }
        static get isShow() {
            if (this.isCurrentVersion)
                return false;
            return true;
        }
        static get isCurrentVersion() {
            return this.sdk.isCurrentVersion();
        }
        static sendVideo(data) {
            this.sdk.sendVideo(data);
        }
        static stageOnStart(data) {
            this.sdk.stageOnStart(data);
        }
        static statgeOnRunning(data) {
            this.sdk.statgeOnRunning(data);
        }
        static stageOnEnd(data) {
            this.sdk.stageOnEnd(data);
        }
        static sendReport(name, params = "") {
            SxyGame.sendEvent(name, params);
        }
        static get Nativecache() {
            if (this.sdk.data.Nativecache) {
                return this.sdk.data.Nativecache;
            }
            return 2;
        }
        static canClick(data, type) {
            if (this.sdk.playerFlag == PlayerFlag.CHANNEL0)
                return true;
            if (!this.isSupportErrorClick(data, type)) {
                return false;
            }
            if (type == ClickType.NativeAd) {
                var prop = Math.random();
                if (prop <= this.sdk.avdErrorProp()) {
                    this.updateErrorCount(ErrorCountType.NATIVE);
                    return true;
                }
                else {
                    return false;
                }
            }
            else if (type == ClickType.PlayGame) {
                var prop = Math.random();
                if (prop <= this.sdk.playGameErrorProp()) {
                    this.updateErrorCount(ErrorCountType.VIDEO);
                    return true;
                }
                else {
                    return false;
                }
            }
            else if (type == ClickType.WaterFall) {
                var prop = Math.random();
                if (prop <= this.sdk.waterFallErrorProp()) {
                    return true;
                }
                else {
                    return false;
                }
            }
            else if (type == ClickType.EnterGame) {
                var prop = Math.random();
                if (prop <= this.sdk.enterGameVideoProp()) {
                    this.updateErrorCount(ErrorCountType.VIDEO);
                    return true;
                }
                else {
                    return false;
                }
            }
        }
        static updateErrorCount(type) {
            let cCount = this.sdk.errorCount(type);
            this.sdk.updataAvdErrorCount(cCount + 1, type);
            console.log("更新最大次数", cCount + 1);
        }
        static showNativeUI(node, ad_comp_name, btn_look, btn_continue, level, cb, su, fa) {
        }
        static baseChek() {
            if (this.sdk.playerFlag == PlayerFlag.CHANNEL0)
                return true;
            if (this.sdk.playerFlag == PlayerFlag.NO_AD) {
                console.log("标记未无广告用户-屏蔽");
                return false;
            }
            if (this.isOffCitys()) {
                return false;
            }
            if (this.isCurrentVersion) {
                return false;
            }
            return true;
        }
        static isSupportErrorClick(data, type) {
            if (this.sdk.playerFlag == PlayerFlag.CHANNEL0)
                return true;
            if (!this.baseChek()) {
                return false;
            }
            if (type == ClickType.NativeAd) {
                if (this.sdk.errorMaxCount(ErrorCountType.NATIVE) == 0) {
                    console.warn("isSupportErrorClick-误点未配置-屏蔽");
                    return false;
                }
                let cCount = this.sdk.errorCount(ErrorCountType.NATIVE);
                if (this.sdk.errorMaxCount(ErrorCountType.NATIVE) <= cCount) {
                    console.warn("isSupportErrorClick-误点上限-屏蔽");
                    return false;
                }
                if (this.sdk.avdErrorLevel() == 0) {
                    console.warn("isSupportErrorClick-avdErrorLevel误点未配置-屏蔽");
                    return false;
                }
                if (this.sdk.avdErrorProp() == 0)
                    return false;
                if (data.level < this.sdk.avdErrorLevel()) {
                    console.warn("isSupportErrorClick-误点等级-屏蔽");
                    return false;
                }
            }
            else if (type == ClickType.PlayGame) {
                if (this.sdk.errorMaxCount(ErrorCountType.VIDEO) == 0) {
                    console.warn("isSupportErrorClick-误点未配置-屏蔽");
                    return false;
                }
                let cCount = this.sdk.errorCount(ErrorCountType.VIDEO);
                if (this.sdk.errorMaxCount(ErrorCountType.VIDEO) <= cCount) {
                    console.warn("isSupportErrorClick-误点上限-屏蔽");
                    return false;
                }
                if (this.sdk.playGameLevel() == 0) {
                    console.warn("isSupportErrorClick-playGameLevel误点未配置-屏蔽");
                    return false;
                }
                if (this.sdk.playGameErrorProp() == 0)
                    return false;
                if (data.level < this.sdk.playGameLevel()) {
                    console.warn("isSupportErrorClick-误点等级-屏蔽");
                    return false;
                }
            }
            else if (type == ClickType.WaterFall) {
                if (this.sdk.errorMaxCount(ErrorCountType.BANNER) == 0) {
                    console.warn("isSupportErrorClick-误点未配置-屏蔽");
                    return false;
                }
                let cCount = this.sdk.errorCount(ErrorCountType.BANNER);
                if (this.sdk.errorMaxCount(ErrorCountType.BANNER) <= cCount) {
                    console.warn("isSupportErrorClick-误点上限-屏蔽");
                    return false;
                }
                if (this.sdk.waterFallErrorProp() == 0)
                    return false;
            }
            else if (type == ClickType.EnterGame) {
                if (this.sdk.errorMaxCount(ErrorCountType.VIDEO) == 0) {
                    console.warn("isSupportErrorClick-误点未配置-屏蔽");
                    return false;
                }
                let cCount = this.sdk.errorCount(ErrorCountType.VIDEO);
                if (this.sdk.errorMaxCount(ErrorCountType.VIDEO) <= cCount) {
                    console.warn("isSupportErrorClick-误点上限-屏蔽");
                    return false;
                }
                if (this.sdk.enterGameVideoProp() == 0)
                    return false;
            }
            return true;
        }
        static openMoreGameView(level) {
            if (this.isCurrentVersion)
                return false;
            if (!this.sdk.data)
                return false;
            if (!this.sdk.data.VideoJumpStartLevel)
                return false;
            if (!this.sdk.data.VideoJumpInterval)
                return false;
            if (level < parseInt(this.sdk.data.VideoJumpStartLevel))
                return false;
            if ((level % parseInt(this.sdk.data.VideoJumpInterval)) != 0)
                return false;
            return true;
        }
        static EnterGame(level, back) {
            if (back)
                back();
        }
        static PlayGameAd(level, back) {
            if (!this.baseChek()) {
                if (back)
                    back();
                return false;
            }
            if (this.sdk.data.PlayGameAd == undefined) {
                console.warn("PlayGameAd-PlayGameAd-屏蔽");
                if (back)
                    back();
                return false;
            }
            var limitLevel = parseInt(this.sdk.data.PlayGameAd);
            if (limitLevel <= 0) {
                console.warn("PlayGameAd-limitLevel-屏蔽");
                if (back)
                    back();
                return false;
            }
            if (level < limitLevel) {
                console.warn("PlayGameAd-limitLevel-屏蔽");
                if (back)
                    back();
                return false;
            }
            if (!this.canClick({ level }, ClickType.PlayGame)) {
                if (back)
                    back();
                return false;
            }
            console.warn("EnterAd show");
            GameSDK$1.showADVideo(this, (state) => {
                if (back)
                    back(state);
            }, "点击开始游戏");
            return true;
        }
        static popUPFullScreen(fullData, back) {
            return true;
        }
        static addelay() {
            let time = 0;
            if (this.sdk.data && this.sdk.data.textDelay)
                time = this.sdk.data.textDelay * 1000;
            if (this.isOffCitys())
                time = 0;
            if (this.isCurrentVersion)
                time = 0;
            return time;
        }
        static get isShortCutInstalled() {
            return this.sdk.isShortCutInstalled;
        }
        static isOffCitys() {
            if (!this.sdk.cityData.loc)
                return false;
            var len = this.sdk.cityData.loc.length;
            var list = this.data.Offcitys || [];
            for (var i = 0; i < len; i += 1) {
                if (list.indexOf(this.sdk.cityData.loc[i]) >= 0) {
                    return true;
                }
            }
            return false;
        }
        static isBlackScene(scene) {
            if (!!this.sdk.data && this.sdk.data.blackScene != undefined) {
                let strs = this.sdk.data.blackScene.split(",");
                for (let i of strs) {
                    if (+i == scene) {
                        return true;
                    }
                }
                return false;
            }
            return true;
        }
        static isOpenCitys() {
            if (!this.sdk.cityData.loc)
                return false;
            var len = this.sdk.cityData.loc.length;
            var list = this.data.OPLISTS || [];
            for (var i = 0; i < len; i += 1) {
                if (list.indexOf(this.sdk.cityData.loc[i]) >= 0) {
                    return true;
                }
            }
            return false;
        }
        static getHours() {
            let nowtime = this.sdk.cityData.nowtime;
            var date = new Date(nowtime);
            return date.getHours();
        }
        static getNavigateList() {
            let list = this.sdk.getNavigateList();
            return list;
        }
        static isAderrormaxcount() {
            if (this.sdk.data.aderrormaxcount == undefined) {
                return false;
            }
            if (this.sdk.data.aderrormaxcount <= 0 || this.level <= 0) {
                return false;
            }
            return this.sdk.data.aderrormaxcount <= this.level;
        }
        static OpenAutoFN() {
            this.sdk.OpenAutoFN();
        }
        static navigateToMiniGame(msg) {
            this.sdk.navigateToMiniGame(msg);
            console.log(msg);
        }
        static showBanner() {
            this.sdk.showBanner();
        }
        static hideBanner() {
            this.sdk.hideBanner();
        }
        static isBannerShow() {
            return this.sdk._isBannerShowing;
        }
        static showADVideo(caller, onComplete, reason, args) {
            if (this.sdk instanceof Web) {
                onComplete.call(caller, true);
                MsgManager.getInst().event('AD_Showed');
                return;
            }
            this.callHideFunc();
            let func = (s) => {
                this.callShowFunc();
                onComplete.call(caller, s, args);
                if (s) {
                    MsgManager.getInst().event('AD_Showed');
                }
            };
            func.bind(this);
            this.sdk.showADVideo(this, func, reason, args);
        }
        static clickADVideo(btn, caller, onComplete, reason, args) {
            if (!!btn) {
                btn.touchable = false;
            }
            if (this.sdk instanceof Web) {
                if (!!btn) {
                    btn.touchable = true;
                }
                onComplete.call(caller, true);
                MsgManager.getInst().event('AD_Showed');
                return;
            }
            ReportMgr.inst.sendReport(this._videoReport + this._videoStartReport + reason, false);
            this.callHideFunc();
            let handler = Handler.create(caller, onComplete, args);
            this.sdk.showADVideo(this, (s) => {
                if (!!btn) {
                    btn.touchable = true;
                }
                this.callShowFunc();
                if (s) {
                    ReportMgr.inst.sendReport(this._videoReport + this._videoSuccessReport + reason, false);
                    if (s) {
                        MsgManager.getInst().event('AD_Showed');
                    }
                }
                else {
                    ReportMgr.inst.sendReport(this._videoReport + this._videoFailReport + reason, false);
                }
                handler.runWith(s);
            }, reason, args);
        }
        static startVideoRecord(time, callback) {
            this.sdk.startVideoRecord(time, callback);
        }
        static pauseVideoRecord() {
            this.sdk.pauseVideoRecord();
        }
        static resumeVideoRecord() {
            this.sdk.resumeVideoRecord();
        }
        static stopVideoRecord() {
            this.sdk.stopVideoRecord();
        }
        static shareVideoRecord(callback) {
            this.sdk.shareVideoRecord(callback);
        }
        static isSupportShareVideo() {
            return this.sdk.isSupportShareVideo();
        }
        static vibrateShort(type) {
            this.sdk.vibrateShort(type);
        }
        static vibrateLong() {
            this.sdk.vibrateLong();
        }
        static checkFollowState(callback) {
            this.sdk.checkFollowState(callback);
        }
        static openUserProfile(callback) {
            this.sdk.openUserProfile(callback);
        }
        static isConfigUIModel(key) {
            if (!key)
                return true;
            if (!GameSDK$1.sdk.data)
                return true;
            let value = GameSDK$1.sdk.data[key];
            if (value && value.state) {
                if (value.state == "1")
                    return true;
                else if (value.state == "0")
                    return false;
                else
                    return true;
            }
            return true;
        }
        static getConfigUIModelInterval(key) {
            if (!key)
                return 0;
            if (!GameSDK$1.sdk.data)
                return 0;
            let value = GameSDK$1.sdk.data[key];
            if (value && value.interval) {
                if (typeof value.interval == "string")
                    return Number(value.interval);
                else
                    return value.interval;
            }
            return 0;
        }
        static shortCutInstalled(callback) {
            this.sdk.shortCutInstalled(callback);
        }
        static checkShortcut(callback) {
            this.sdk.checkShortcut(callback);
        }
        static getLogicScene(callback) {
            this.sdk.getLogicScene(callback);
        }
        static VIDEOID() {
            console.log("----------------VIDEOID");
            if (this.sdk.data.VIDEOID == undefined) {
                return "";
            }
            console.log("----------------VIDEOID : " + this.sdk.data.VIDEOID);
            return this.sdk.data.VIDEOID;
        }
        static JUMP_OPS() {
            console.log("----------------JUMP_OPS");
            if (this.sdk.data.JUMP_OPS == undefined) {
                return 3;
            }
            console.log("----------------JUMP_OPS : " + this.sdk.data.JUMP_OPS);
            return this.sdk.data.JUMP_OPS;
        }
        static NingRongRongTime() {
            console.log("----------------NingRongRongTime");
            if (this.sdk.data.NingRongRongTime == undefined) {
                return 0;
            }
            console.log("----------------NingRongRongTime : " + this.sdk.data.NingRongRongTime);
            return Number(this.sdk.data.NingRongRongTime);
        }
        static ChangeCode34() {
            console.log("----------------ChangeCode34");
            if (this.sdk.data.ChangeCode34 == undefined) {
                return 0;
            }
            console.log("----------------ChangeCode34 : " + this.sdk.data.ChangeCode34);
            return Number(this.sdk.data.ChangeCode34);
        }
        static createNative(value) {
            return this.sdk.createNative(value);
        }
        static isOutAdSwitch() {
            if (this.isOffCitys()) {
                return false;
            }
            else {
                return this.sdk.data.isAdvert == 1;
            }
        }
        static openDanger() {
            if (this.isCurrentVersion) {
                return false;
            }
            if (this.isOffCitys()) {
                return false;
            }
            else {
                return this.sdk.data.openDanger == 1;
            }
        }
        static showToast(str, time = 1500) {
            this.sdk.showToast(str, time);
        }
        static needShowNative() {
            return this.type == PlatType$1.TEST || this.type == PlatType$1.OPPO || this.type == PlatType$1.VIVO;
        }
        static loadSubpackage(pack, zipFileName, topath, comlete, progress) {
            this.sdk.loadSubpackage(pack, zipFileName, topath, comlete, progress);
        }
        static isShowTime() {
            let newTime = Date.now();
            if (this.sdk.data.IntervalAdTime == null || this.sdk.data.IntervalAdTime == undefined)
                return false;
            if (newTime - this.startGameTime > this.sdk.data.IntervalAdTime * 1000) {
                return true;
            }
            return false;
        }
        static isOpenBanner() {
            if (!this.isShowTime())
                return false;
            return true;
        }
        static showADInsert() {
            if (this.isOffCitys()) {
                return;
            }
            if (this.isCurrentVersion) {
                return;
            }
            this.sdk.showADInsert();
        }
        static shareMsg(caller, onComplete, reason, args) {
            this.sdk.shareMsg(caller, onComplete, reason, args);
        }
        static showGamePortal() {
            this.sdk.showGamePortal();
        }
        static showGameBanner() {
            this.sdk.showGameBanner();
        }
        static hideGameBanner() {
            this.sdk.hideGameBanner();
        }
        static compareVersion(curVersion, compareVersion) {
            let v1 = curVersion;
            if (!v1 || !v1.split)
                return 0;
            let v2 = compareVersion;
            if (!v2 || !v2.split)
                return 0;
            let v1Arr = v1.split('.');
            let v2Arr = v2.split('.');
            const len = Math.max(v1Arr.length, v2Arr.length);
            while (v1Arr.length < len) {
                v1Arr.push('0');
            }
            while (v2Arr.length < len) {
                v2Arr.push('0');
            }
            for (let i = 0; i < len; i++) {
                const num1 = parseInt(v1Arr[i]);
                const num2 = parseInt(v2Arr[i]);
                if (num1 > num2) {
                    return 1;
                }
                else if (num1 < num2) {
                    return -1;
                }
            }
            return 0;
        }
        static onClickButton() {
            return new Promise((resolve, reject) => {
                if (!this.isSpecialUser()) {
                    resolve();
                    return;
                }
                if (!!!this._lastVideoTime) {
                    this._lastVideoTime = LocalCache$1.get(this.VIDEO_TIME_KEY, 0);
                }
                let now = Date.now();
                if (Math.floor(now / this.TIME_DAY) - Math.floor(this._lastVideoTime / this.TIME_DAY) >= 1) {
                    this._videoCount = 0;
                    LocalCache$1.set(this.VIDEO_COUNT_KEY, this._videoCount);
                }
                else {
                    this._videoCount = LocalCache$1.get(this.VIDEO_COUNT_KEY, 0);
                }
                if (this._videoCount < (this.sdk.data[this.DATA_VIDEO_MAX] || 0)) {
                    let limit = Math.max(this.sdk.data[this.DATA_VIDEO_LIMIT] || 2, 2);
                    this._touchCount++;
                    if (this._touchCount % limit == 0) {
                        this._videoCount++;
                        LocalCache$1.set(this.VIDEO_COUNT_KEY, this._videoCount);
                        this.showADVideo(this, () => {
                            resolve();
                        }, "");
                    }
                    else {
                        resolve();
                    }
                }
                else {
                    resolve();
                }
            });
        }
        static showCustomAd(type, top) {
            return new Promise((resolve, reject) => {
                if (this.sdk instanceof WeChat) {
                    this.sdk.showCustomAd(type, top).then(s => {
                        resolve(s);
                    });
                }
                else {
                    resolve(true);
                }
            });
        }
        static hideCustomAd(type) {
            if (this.sdk instanceof WeChat) {
                this.sdk.hideCustomAd(type);
            }
        }
        static isShortScreen() {
            if (this.sdk instanceof WeChat) {
                return this.sdk.isShortScreen();
            }
            return true;
        }
        static getLaterTime() {
            if (Math.random() * 100 < (this.sdk.data.delaypr || 0)) {
                return this.sdk.data.delaytime || 0;
            }
            return 0;
        }
        static replaceStr(val, ...strs) {
            let i = 0;
            while (val.match(/\{[0-9]+\}/)) {
                val = val.replace(/\{[0-9]+\}/, strs[i]);
                i++;
            }
            return val;
        }
        static showButtomAd() {
            let random = Math.random() * 100;
            if (random < (this.data.BannerOrGrid || 100)) {
                console.log("random show custom");
                this.hideBanner();
                this.showCustomAd(CustomType.Bottom_Roll);
            }
            else {
                console.log("random show banner");
                this.hideCustomAd(CustomType.Bottom_Roll);
                this.showBanner();
            }
        }
        static hideButtomAd() {
            this.hideBanner();
            this.hideCustomAd(CustomType.Bottom_Roll);
        }
        static isLimit() {
            return this.isCurrentVersion || this.isOffCitys();
        }
        static isSpecialUser() {
            if (this.sdk instanceof WeChat) {
                if (this.isCurrentVersion) {
                    return false;
                }
                let id = LocalCache$1.get("LogickData._getLoginId", 0);
                if (!!!id) {
                    let info = this.sdk.getLaunchOptionsSync();
                    if (!!info) {
                        if (!!info.scene) {
                            LocalCache$1.set("LogickData._getLoginId", +info.scene);
                            id = +info.scene;
                        }
                    }
                }
                this.sendReport("进入场景", id + "");
                return !!id && !this.isBlackScene(id) && !this.isOffCitys();
            }
            return false;
        }
        static get isFirstPlay() {
            if (LocalCache$1.get("LogickData._isFirstPlay", true)) {
                let now = Date.now();
                if (!!!this._createTime) {
                    this._createTime = LocalCache$1.get("LogicData._createTime" + SDKConfig.GAMEID, Date.now());
                }
                return now <= new Date(new Date(this._createTime).toLocaleDateString()).getTime() + 24 * 60 * 60 * 1000 - 1;
            }
            return false;
        }
        static registShowCaller(func, caller, args) {
            for (let i = 0; i < this._showFuncArr.length; ++i) {
                if (this._showFuncArr[i].caller == caller && this._showFuncArr[i].func == func) {
                    return;
                }
            }
            this._showFuncArr.push({ func: func, caller: caller, args: args });
        }
        static registHideCaller(func, caller, args) {
            for (let i = 0; i < this._hideFuncArr.length; ++i) {
                if (this._hideFuncArr[i].caller == caller && this._hideFuncArr[i].func == func) {
                    return;
                }
            }
            this._hideFuncArr.push({ func: func, caller: caller, args: args });
        }
        static callShowFunc(data) {
            if (!!this._showFuncArr && !!this._showFuncArr.length) {
                for (let i of this._showFuncArr) {
                    if (!!data) {
                        i.func.call(i.caller, data, i.args);
                    }
                    else {
                        i.func.call(i.caller, i.args);
                    }
                }
            }
        }
        static callHideFunc(data) {
            if (!!this._hideFuncArr && !!this._hideFuncArr.length) {
                for (let i of this._hideFuncArr) {
                    if (!!data) {
                        i.func.call(i.caller, data, i.args);
                    }
                    else {
                        i.func.call(i.caller, i.args);
                    }
                }
            }
        }
        static postMessage(type, target, args) {
            if (this.sdk instanceof WeChat) {
                this.sdk.postMessage(type, target, args);
            }
        }
        static setShareFriendQuery(scene) {
            if (this.sdk instanceof WeChat) {
                this.sdk.setShareFriendQuery(scene);
            }
        }
        static getShareCanvas() {
            if (this.sdk instanceof WeChat) {
                return this.sdk.getShareCanvas();
            }
            return null;
        }
        static getLaunchInfo() {
            if (this.sdk instanceof WeChat) {
                return this.sdk.getLaunchOptionsSync();
            }
            return null;
        }
        static login() {
            if (this.sdk instanceof WeChat) {
                return this.sdk.login();
            }
            return Promise.resolve(null);
        }
        static getEncryptData() {
            if (this.sdk instanceof WeChat) {
                return this.sdk.getEncryptData();
            }
            return Promise.resolve(null);
        }
        static setUserCloudStorage(k, v) {
            if (this.sdk instanceof WeChat) {
                this.sdk.setUserCloudStorage(k, v);
            }
        }
        static set sessionKey(val) {
            if (this.sdk instanceof WeChat) {
                this.sdk.sessionKey = val;
            }
        }
        static getSetting() {
            if (this.sdk instanceof WeChat) {
                return this.sdk.getSetting();
            }
            return Promise.resolve(null);
        }
        static authorize(scope) {
            if (this.sdk instanceof WeChat) {
                return this.sdk.authorize(scope);
            }
            return Promise.resolve(true);
        }
    }
    GameSDK$1.bxCount = 0;
    GameSDK$1.type = PlatType$1.TEST;
    GameSDK$1.startGameTime = 0;
    GameSDK$1.level = 1;
    GameSDK$1.ctorData = null;
    GameSDK$1.isDebug = false;
    GameSDK$1.Can_Ys_show = false;
    GameSDK$1.NativeDataArr = [];
    GameSDK$1.NativeIndex = 1;
    GameSDK$1.NativeDataArr2 = [];
    GameSDK$1.NativeIndex2 = 1;
    GameSDK$1.BannerPanelIsShow = false;
    GameSDK$1.InsertTime = 0;
    GameSDK$1.isDecode = false;
    GameSDK$1.clickCount = 0;
    GameSDK$1.isRight = false;
    GameSDK$1.hasAgreeXieYi = false;
    GameSDK$1.newenter = true;
    GameSDK$1._videoReport = "玩家看视频";
    GameSDK$1._videoStartReport = "-开始-";
    GameSDK$1._videoSuccessReport = "-成功-";
    GameSDK$1._videoFailReport = "-失败-";
    GameSDK$1._lastVideoTime = 0;
    GameSDK$1._videoCount = 0;
    GameSDK$1._touchCount = 0;
    GameSDK$1.VIDEO_TIME_KEY = "VIDEO_TIME_KEY";
    GameSDK$1.VIDEO_COUNT_KEY = "VIDEO_COUNT_KEY";
    GameSDK$1.TIME_DAY = 24 * 60 * 60 * 1000;
    GameSDK$1.DATA_VIDEO_MAX = "maxnumber";
    GameSDK$1.DATA_VIDEO_LIMIT = "count";
    GameSDK$1._showFuncArr = [];
    GameSDK$1._hideFuncArr = [];
    var CustomType;
    (function (CustomType) {
        CustomType[CustomType["Left_Grid"] = 0] = "Left_Grid";
        CustomType[CustomType["Right_Grid"] = 1] = "Right_Grid";
        CustomType[CustomType["Bottom_Roll"] = 2] = "Bottom_Roll";
        CustomType[CustomType["Main_Rect"] = 3] = "Main_Rect";
    })(CustomType || (CustomType = {}));
    const SDKEventName = {
        LONG_TIME_NO_OPERATION: "LONG_TIME_NO_OPERATION",
        OpenAutoFN: "OpenAutoFN",
    };
    let golab = window;
    golab.GameSDK = GameSDK$1;
    golab.PlatType = PlatType$1;
    golab.PropType = PropType$1;
    class SDKEventMgr {
        constructor() {
            this._eventMap = [];
            this.m_events = new Map();
        }
        static get Int() {
            if (!this._Int) {
                this._Int = new SDKEventMgr;
            }
            return this._Int;
        }
        on(name, callBack, caller) {
            if (!callBack || !caller)
                return;
            let event = this.m_events.get(name);
            if (!event) {
                event = [];
                this.m_events.set(name, event);
            }
            event.push({ callBack, caller });
        }
        emit(name, ...param) {
            let events = this.m_events.get(name);
            if (!events) {
                return;
            }
            for (let i = 0; i < events.length; i++) {
                let value = events[i];
                if (value)
                    value.callBack.call(value.caller, ...param);
            }
        }
        off(name, callBack, caller) {
            let events = this.m_events.get(name);
            if (!events)
                return;
            for (let i = 0; i < events.length; i++) {
                let value = events[i];
                if (value)
                    if (value.callBack == callBack && value.caller == caller) {
                        events.splice(i, 1);
                        return;
                    }
            }
        }
        offAll(name) {
            let events = this.m_events.get(name);
            if (!events) {
                return;
            }
            events.length = 0;
        }
    }

    class SxyGame {
        static get instance() {
            return SxyGame._instance || (SxyGame._instance = new SxyGame);
        }
        init() {
            this.getPlatform();
            this.checkSDK();
            this.getUUID();
            this.setNewPlayer();
            SxyGame.openid = this.getOpenid();
        }
        getPlatform() {
            if (typeof window["tt"] != "undefined") {
                SxyGame.platform = window["tt"];
                SxyGame.platformFlag = true;
                console.log("当前平台：头条");
            }
            else if (typeof wx != "undefined") {
                SxyGame.platform = wx;
                SxyGame.platformFlag = true;
                console.log("当前平台：微信");
            }
            else {
                window.getStorageSync = function (key) {
                    return window.localStorage.getItem(key);
                };
                window.setStorageSync = function (key, value) {
                    window.localStorage.setItem(key, value);
                };
                window.request = function (data) {
                    var xmlhttp;
                    if (window.XMLHttpRequest) {
                        xmlhttp = new XMLHttpRequest();
                    }
                    else {
                        xmlhttp = new ActiveXObject("Microsoft.XMLHTTP");
                    }
                    xmlhttp.onreadystatechange = function () {
                        if (xmlhttp.readyState == 4 && xmlhttp.status == 200) ;
                    };
                    var url = data.url;
                    if (data.data) {
                        let params = Object.keys(data.data).map(function (key) {
                            return encodeURIComponent(key) + "=" + encodeURIComponent(data.data[key]);
                        }).join("&");
                        url += "?" + params;
                    }
                    xmlhttp.open("GET", url, true);
                    xmlhttp.send();
                };
                SxyGame.platform = window;
            }
        }
        static sendVideo(data) {
            SxyGame.instance.send(data, "video");
        }
        static sendEvent(name, params = "") {
            let data = {
                'eventName': name,
                'params': params
            };
            SxyGame.instance.send(data, "event");
        }
        checkSDK() {
            SxyGame.gameId = SDKConfig.GAMEID;
            SxyGame.getLocation = SDKConfig.getLocation;
            SxyGame.useOpenid = SDKConfig.useOpenid;
            if (!SxyGame.gameId)
                console.error('缺少gameId配置');
            SxyGame.platform.request({
                url: SxyGame.logerUrl + "/static/app.json",
                method: "GET",
                success: function (res) {
                    if (200 === res.data.code) {
                        if (res.data.data.version > SxyGame.localVersion) {
                            console.error("您的上报SDK不是最新版本，请尽快升级！");
                        }
                    }
                }
            });
        }
        getUUID() {
            var uuid = '';
            try {
                uuid = SxyGame.platform.getStorageSync('sxy_uuid');
            }
            catch (e) {
                uuid = "uuid_default";
            }
            if (!uuid) {
                uuid = this.create_uuid();
                try {
                    SxyGame.platform.setStorageSync("sxy_uuid", uuid);
                }
                catch (e) {
                    SxyGame.platform.setStorageSync("sxy_uuid", 'uuid_default');
                }
            }
            return uuid;
        }
        setNewPlayer() {
            var reg_date = SxyGame.platform.getStorageSync('sxy_reg_date');
            var today = SxyGame.formatDate(new Date());
            if (!reg_date) {
                SxyGame.platform.setStorageSync("sxy_new_player", true);
                SxyGame.platform.setStorageSync("sxy_reg_date", today);
            }
            else {
                if (today > reg_date) {
                    SxyGame.platform.setStorageSync("sxy_new_player", false);
                }
            }
        }
        static formatDate(date) {
            var myyear = date.getFullYear();
            var mymonth = date.getMonth() + 1;
            var myweekday = date.getDate();
            if (mymonth < 10) {
                mymonth = "0" + mymonth;
            }
            if (myweekday < 10) {
                myweekday = "0" + myweekday;
            }
            return (myyear + "-" + mymonth + "-" + myweekday);
        }
        getOpenid() {
            return SxyGame.platform.getStorageSync("sxy_openid") || "";
        }
        create_uuid() {
            function e() {
                return Math.floor(65536 * (1 + Math.random())).toString(16).substring(1);
            }
            return e() + e() + e() + e() + e() + e() + e() + e();
        }
        send(data, action) {
            if (SxyGame.useOpenid) {
                this.getOpenid();
                if (!SxyGame.openid) {
                    console.error('没有openid，此次上报失败');
                    return;
                }
            }
            var that = this;
            new Promise(function (resolve, reject) {
                SxyGame.paramJoin(data);
                data.gameId = SxyGame.gameId;
                data.platform = GameSDK$1.type;
                data.session_key = SxyGame.session_key || "";
                data.openid = SxyGame.openid || "";
                data.uuid = that.getUUID();
                data.avatar = SxyGame.avatar || "";
                data.action = action;
                SxyGame.platform.request({
                    url: SxyGame.logerUrl + "/index/api.do",
                    data: data,
                    header: {},
                    method: "GET",
                    fail: function () {
                        resolve("");
                    },
                    success: function (res) {
                        resolve(200 == res.statusCode ? "" : "status error");
                    }
                });
            });
        }
        sendOpenid(openid) {
            if ("" === openid || !openid) {
                return console.error("openID不能为空");
            }
            SxyGame.platform.setStorageSync("sxy_openid", openid);
        }
        static paramJoin(data) {
            data.newPlayer = SxyGame.platform.getStorageSync('sxy_new_player');
        }
    }
    SxyGame.localVersion = "1.2.0";
    SxyGame.logerUrl = "https://loger.jpsdk.com";
    SxyGame.gameId = '';
    SxyGame.openid = '';
    SxyGame.getLocation = false;
    SxyGame.useOpenid = false;
    SxyGame.platformFlag = false;
    SxyGame.stage = {
        "onStart": function (data) {
            data.method = 'start';
            SxyGame.instance.send(data, "stage");
        },
        "onRunning": function (data) {
            data.method = 'running';
            SxyGame.instance.send(data, "stage");
        },
        "onEnd": function (data) {
            data.method = 'end';
            SxyGame.instance.send(data, "stage");
        }
    };

    class WxMgr {
        constructor() {
            this._loginEnd = false;
            this.SHARE_ROBOT = 11;
            this._openid = "";
            this._token = "";
            this.PLAY_GAME_INFO = "playGameInfo";
            this.FRIEND_SCOPE = "scope.WxFriendInteraction";
            this.battleStart = false;
            MsgManager.getInst().on(BattleEvent.MechFreeEnd, this, this._onMechFreeEnd);
            MsgManager.getInst().on(EventType.WxShareFriend, this, this._onShareFriendBack);
        }
        static get inst() {
            if (!!!WxMgr._inst) {
                WxMgr._inst = new WxMgr();
            }
            return WxMgr._inst;
        }
        login(data) {
            this.setSelfPlay();
            this._sendLogin(data).then(() => {
                this.init();
            }).catch(e => {
                this.init();
            });
        }
        shareMsg(caller, onComplete, reason, args) {
            if (!!this._token) {
                GameSDK.shareMsg(caller, onComplete, reason, args);
            }
            else {
                let handler = Handler.create(caller, onComplete, args);
                TipsMgr.Tips("分享失败，请检查您的网络状态或重启游戏");
                handler.runWith(false);
            }
        }
        sendEncryptData() {
            return new Promise((resolve, reject) => {
                this._showFriendScrope().then(s => {
                    if (s) {
                        GameSDK.getEncryptData().then(data => {
                            if (!!this._token) {
                                this._sendEncryptedData(data).then(e => {
                                    resolve(e);
                                }).catch(e => {
                                    reject(e);
                                });
                            }
                            else {
                                reject();
                            }
                        }).catch(e => {
                            console.error(e);
                            reject(e);
                        });
                    }
                    else {
                        resolve(null);
                    }
                });
            });
        }
        sendGetInvate() {
            return new Promise((resolve, reject) => {
                if (!!this._token) {
                    this._sendGetInvate().then(e => {
                        resolve(e);
                    }).catch(e => {
                        reject(e);
                    });
                }
                else {
                    TipsMgr.Tips("获取数据失败，请检查您的网络或重新打开游戏");
                    reject();
                }
            });
        }
        get token() {
            return this._token;
        }
        get openid() {
            return this._openid;
        }
        get hasOpenShareFriend() {
            return LocalCache$1.get("LogicData._openShareFriend", false);
        }
        set hasOpenShareFriend(val) {
            LocalCache$1.set("LogicData._openShareFriend", val);
        }
        _sendLogin(data) {
            return new Promise((resolve, reject) => {
                data.module = "login";
                data.platform = GameSDK.type;
                data.gameId = SDKConfig.GAMEID;
                data.uuid = SxyGame.instance.getUUID();
                console.log("登录code==>", data.code);
                this._sendHttpByPost(WxMgr._serverUrl, data).then(e => {
                    if (!!e && !!e.data && !!e.data.session_key) {
                        GameSDK.sessionKey = e.data.session_key;
                    }
                    resolve(e);
                }).catch(e => {
                    reject(e);
                });
            });
        }
        _sendEncryptedData(data) {
            return new Promise((resolve, reject) => {
                data.module = "game";
                data.action = "getEncryptedData";
                data.token = this._token;
                this._sendHttpByPost(WxMgr._serverUrl, data).then(e => {
                    resolve(e);
                }).catch(e => {
                    reject(e);
                });
            });
        }
        _sendGetInvate() {
            return new Promise((resolve, reject) => {
                let data = {};
                data.module = "game";
                data.action = "getInviteUsers";
                data.token = this._token;
                this._sendHttpByPost(WxMgr._serverUrl, data).then(e => {
                    resolve(e);
                }).catch(e => {
                    reject(e);
                });
            });
        }
        _sendHttpByPost(url, data) {
            return new Promise((resolve, reject) => {
                var xhr = new Laya.HttpRequest();
                xhr.http.timeout = 10000;
                xhr.once(Laya.Event.COMPLETE, this, (e) => {
                    if (!!e && e.code == 0) {
                        if (!!!this._token && !!e.data && !!e.data.token) {
                            this._token = e.data.token;
                        }
                        if (!!!this._openid && !!e.data && !!e.data.openid) {
                            this._openid = e.data.openid;
                        }
                        resolve(e);
                    }
                    else {
                        reject(e);
                    }
                });
                xhr.once(Laya.Event.ERROR, this, (e) => {
                    reject(e);
                });
                xhr.send(url, JSON.stringify(data), "POST", "json", ["Content-type", "application/json"]);
            });
        }
        init() {
            this._loginEnd = true;
            let launchInfo = GameSDK.getLaunchInfo();
            if (!!launchInfo && !!launchInfo.query && launchInfo.query.shareMessageToFriendScene == this.SHARE_ROBOT) {
                this.modifyData(WxMgr.inst.SHARE_ROBOT);
            }
        }
        _onMechFreeEnd() {
            TipsMgr.Tips("机甲试用完毕");
            PlayerData.UnlockMecha = true;
        }
        checkRobotUnlock() {
            return new Promise((resolve, reject) => {
                if (PlayerData.UnlockMecha) {
                    resolve(true);
                }
                else {
                    if (this._loginEnd) {
                        let len = 0;
                        this.sendGetInvate().then((e) => {
                            let arr = e.data;
                            len = arr.length | 0;
                            if (len < 2) {
                                WxMgr.inst.sendEncryptData().then(e => {
                                    if (!!e && !!e.data && !!e.data.kv_list) {
                                        let list = e.data.kv_list;
                                        let find = list.find(item => {
                                            return item.key == WxMgr.inst.SHARE_ROBOT;
                                        });
                                        if (!!find) {
                                            len += +find.value;
                                        }
                                    }
                                    resolve(len >= 2);
                                }).catch(e => {
                                    resolve(len >= 2);
                                });
                            }
                            else {
                                resolve(len >= 2);
                            }
                        }).catch(e => {
                            resolve(len >= 2);
                        });
                    }
                    else {
                        resolve(false);
                    }
                }
            });
        }
        _onShareFriendBack(success) {
            ReportMgr.inst.sendReport(ReportType.SHARE_OPENDATA_ROBOT + (success ? "-成功" : "-失败"));
        }
        postMessage(type, target, args) {
            GameSDK.postMessage(type, target, args);
        }
        clickInvate(index) {
            this.postMessage(WxMsgType.CLICK_INVATE, null, { index: index });
        }
        modifyData(key, quiet = true) {
            this.postMessage(WxMsgType.MODIFY_DATA, null, { key: key, quiet: quiet });
        }
        showFriendShare(target, server) {
            this._showFriendScrope().then(s => {
                if (s) {
                    this.postMessage(WxMsgType.SET_ID, null, this._openid);
                    this._drawShareCanvas(target);
                    GameSDK.setShareFriendQuery(this.SHARE_ROBOT);
                    this.postMessage(WxMsgType.FRIEND_SHARE, target, server);
                }
            });
        }
        refreshFriendShare() {
            this._showFriendScrope().then(s => {
                if (s) {
                    this.postMessage(WxMsgType.FRIEND_SHARE_CHANGE);
                }
            });
        }
        close() {
            if (!!this._view) {
                this._view.visible = false;
            }
            this.postMessage(WxMsgType.CLOSE);
        }
        _drawShareCanvas(target) {
            let canvas = GameSDK.getShareCanvas();
            if (!!canvas) {
                if (!!!this._view) {
                    this._view = new WXOpenDataViewer();
                    Laya.stage.addChild(this._view);
                }
                this._view.zOrder = fgui.GRoot.inst.displayObject.zOrder + 1;
                this._view.x = target.x;
                this._view.y = target.y;
                this._view.width = target.width;
                this._view.height = target.height;
                this._view.visible = true;
            }
        }
        setSelfPlay() {
            this._setUserCloudStorage(this.PLAY_GAME_INFO, "1");
        }
        _setUserCloudStorage(key, value) {
            GameSDK.setUserCloudStorage(key, value);
        }
        _showFriendScrope() {
            return new Promise((resolve, reject) => {
                if (this.hasOpenShareFriend) {
                    this._getFriendScope().then(s => {
                        if (s) {
                            resolve(true);
                        }
                        else {
                            GameSDK.authorize(this.FRIEND_SCOPE).then(s => {
                                resolve(s);
                            });
                        }
                    });
                }
                else {
                    resolve(false);
                }
            });
        }
        _getFriendScope() {
            return new Promise((resolve, reject) => {
                GameSDK.getSetting().then(e => {
                    if (!!e && !!e.authSetting) {
                        resolve(!!e.authSetting[this.FRIEND_SCOPE]);
                    }
                    else {
                        resolve(false);
                    }
                }).catch(e => {
                    resolve(false);
                });
            });
        }
    }
    WxMgr._serverUrl = "https://api3.jpsdk.com/api";
    var WxMsgType;
    (function (WxMsgType) {
        WxMsgType["REFRESH"] = "viewport";
        WxMsgType["FRIEND_SHARE"] = "ResultScene";
        WxMsgType["FRIEND_SHARE_CHANGE"] = "FRIEND_SHARE_CHANGE";
        WxMsgType["CLOSE"] = "DestroyScene";
        WxMsgType["SET_ID"] = "setSelfOpenId";
        WxMsgType["CLICK_INVATE"] = "CLICK_INVATE";
        WxMsgType["MODIFY_DATA"] = "MODIFY_DATA";
    })(WxMsgType || (WxMsgType = {}));
    class WXOpenDataViewer extends Laya.Sprite {
        constructor() {
            super();
            this.width = this.height = 200;
            var tex = new Laya.Texture();
            tex.bitmap = new Laya.Texture2D();
            this.texture = tex;
        }
        onEnable() {
            if (window.wx && window["sharedCanvas"]) {
                Laya.timer.frameLoop(1, this, this._onLoop);
            }
        }
        onDisable() {
            Laya.timer.clear(this, this._onLoop);
        }
        _onLoop() {
            let _canvas = window["sharedCanvas"];
            this.texture.sourceWidth = _canvas.width;
            this.texture.sourceHeight = _canvas.height;
            this.texture.bitmap.loadImageSource(_canvas, true);
        }
        set width(value) {
            super.width = value;
            if (window["sharedCanvas"])
                window["sharedCanvas"].width = value;
        }
        get width() {
            return super.width;
        }
        set height(value) {
            super.height = value;
            if (window["sharedCanvas"]) {
                window["sharedCanvas"].height = value;
            }
        }
        get height() {
            return super.height;
        }
        set x(value) {
            super.x = value;
        }
        get x() {
            return super.x;
        }
        set y(value) {
            super.y = value;
        }
        get y() {
            return super.y;
        }
    }

    var PlatformType;
    (function (PlatformType) {
        PlatformType[PlatformType["Web"] = 0] = "Web";
        PlatformType[PlatformType["WX"] = 1] = "WX";
        PlatformType[PlatformType["QQ"] = 2] = "QQ";
        PlatformType[PlatformType["OPPO"] = 3] = "OPPO";
        PlatformType[PlatformType["VIVO"] = 4] = "VIVO";
        PlatformType[PlatformType["Ali"] = 5] = "Ali";
        PlatformType[PlatformType["XM"] = 6] = "XM";
        PlatformType[PlatformType["TT"] = 7] = "TT";
    })(PlatformType || (PlatformType = {}));
    class Global {
        static get Platform() {
            var u = window.navigator.userAgent;
            if (u.indexOf('AlipayMiniGame') > -1 && "my" in window) {
                this._platform = PlatformType.Ali;
            }
            if (u.indexOf('OPPO') == -1 && u.indexOf("MiniGame") > -1 && "wx" in window) {
                if ("qq" in window) {
                    this._platform = PlatformType.QQ;
                }
                else {
                    this._platform = PlatformType.WX;
                }
            }
            if (u.indexOf("MiniGame") > -1 && "qq" in window) {
                this._platform = PlatformType.QQ;
            }
            if (u.indexOf('QuickGame') > -1) {
                this._platform = PlatformType.XM;
            }
            if (u.indexOf('OPPO') > -1 && u.indexOf('MiniGame') > -1) {
                this._platform = PlatformType.OPPO;
            }
            if (u.indexOf('VVGame') > -1) {
                this._platform = PlatformType.VIVO;
            }
            if (window['tt'] != null) {
                this._platform = PlatformType.TT;
            }
            return this._platform;
        }
        static get PlatformInst() {
            switch (this.Platform) {
                case PlatformType.WX:
                    return window['wx'];
                case PlatformType.QQ:
                    return window['qq'];
                case PlatformType.TT:
                    return window['tt'];
                case PlatformType.OPPO:
                    return window['qg'];
                case PlatformType.VIVO:
                    return window['qg'];
                default:
                    return window['wx'];
            }
        }
        static AddMapItem(node) {
            this.MapNode.addChild(node);
        }
        static ClearNode() {
            for (var i = this.MapNode.numChildren - 1; i >= 0; i--) {
                var c = this.MapNode.getChildAt(i);
                this.MapNode.removeChildAt(i);
                c.destroy(true);
            }
        }
        static get deltaTime() {
            if (this._recordFrame != Laya.timer.currFrame) {
                this._deltaTime = Math.min(Laya.timer.delta, 100);
            }
            return this._deltaTime;
        }
        static get speed() {
            return this._speed;
        }
        static set speed(v) {
            this._speed = v;
        }
        static setAddNum(node, num) {
            var numNode = node.getChildByName('number');
            for (var i = 0; i < numNode.numChildren; i++) {
                numNode.getChildAt(i).active = (i == num);
            }
        }
    }
    Global.openLevelLock = true;
    Global._platform = PlatformType.Web;
    Global.IOS = false;
    Global.soundDir = 'audio/';
    Global.texturePathDir = 'texture/';
    Global.prefabPathDir = 'prefab/LayaScene_Prefab/Conventional/';
    Global.DominoCarDistance = 1.4;
    Global.up = new Laya.Vector3(0, 1, 0);
    Global._recordFrame = 0;
    Global._deltaTime = 30;
    Global.Playing = false;
    Global.CatchSlaveSqrDis = 0.9;
    Global.initSpeed = 6;
    Global._speed = Global.initSpeed;
    Global.CollectCoin = 0;
    Global.PlugGroup = Laya.Physics3DUtils.COLLISIONFILTERGROUP_CUSTOMFILTER1;
    Global.PlaneGroup = Laya.Physics3DUtils.COLLISIONFILTERGROUP_CUSTOMFILTER2;
    Global.SocketHoleRes = 'dogface_socket';
    Global.PlugModelRes = 'dogface_plug';
    Global.LevelEndSlaveCount = 0;
    Global.FakeLevel = null;
    Global.CollectProgress = 25;
    Global.playerPos = new Laya.Vector3();

    class ProgressPromise extends Promise {
        constructor(fn) {
            let self;
            super((resolve, reject) => {
                fn(resolve, value => self && self._progress(value), reject);
            });
            self = this;
        }
        _progress(v) {
            if (this._progressCB) {
                this._progressCB(v);
            }
        }
        progress(fn) {
            this._progressCB = fn;
            return this;
        }
    }

    var chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=';
    class InvalidCharacterError extends Error {
        constructor() {
            super(...arguments);
            this.name = 'InvalidCharacterError';
        }
    }
    class Utils {
        static Shuffle(a) {
            for (let i = a.length; i; i--) {
                let j = Math.floor(Math.random() * i);
                [a[i - 1], a[j]] = [a[j], a[i - 1]];
            }
            return a;
        }
        static GetDate() {
            var d = new Date();
            return d.getFullYear() + '年' + (d.getMonth() + 1) + '月' + d.getDate() + '日';
        }
        static TimeToString(mil, type = 0) {
            var seconds = Math.floor(mil / 1000);
            var s = seconds % 60;
            var min = Math.floor(seconds / 60) % 60;
            var hour = Math.floor(seconds / 3600) % 24;
            if (type == 0) {
                return Utils.formatNumber(min) + ':' + Utils.formatNumber(s);
            }
            else {
                return Utils.formatNumber(hour) + '小时' + Utils.formatNumber(min) + '分钟' + Utils.formatNumber(s) + '秒';
            }
        }
        static formatNumber(n) {
            var s = n.toString();
            return s[1] ? s : '0' + s;
        }
        static rand(a) {
            var i = Math.floor(Math.random() * a.length);
            return a[i];
        }
        static randByWeight(...weights) {
            var total = 0;
            for (var i = 0; i < weights.length; i++) {
                total += Number(weights[i]);
            }
            var rand = Math.random() * total;
            var count = 0;
            for (var i = 0; i < weights.length; i++) {
                var pre = count;
                count += Number(weights[i]);
                if (rand >= pre && rand <= count) {
                    return i;
                }
            }
            return 0;
        }
        static randIntRange(start, end) {
            return Math.round(start + Math.random() * (end - start));
        }
        static hexToRgba01(hex) {
            return new Laya.Vector4(parseInt("0x" + hex.slice(1, 3)) / 255, parseInt("0x" + hex.slice(3, 5)) / 255, parseInt("0x" + hex.slice(5, 7)) / 255, 1);
        }
        static rgb01ToHex(r, g, b) {
            r = Math.min(255, Math.round(r * 255));
            g = Math.min(255, Math.round(g * 255));
            b = Math.min(255, Math.round(b * 255));
            return '#' + ((r << 16) | (g << 8) | b).toString(16);
        }
        static rgb255ToHex(r, g, b) {
            return '#' + ((r << 16) | (g << 8) | b).toString(16);
        }
        static hexToRgb01(hex) {
            return new Laya.Vector3(parseInt("0x" + hex.slice(1, 3)) / 255, parseInt("0x" + hex.slice(3, 5)) / 255, parseInt("0x" + hex.slice(5, 7)) / 255);
        }
        static rgb255Torgb01(rgb) {
            rgb.x /= 255;
            rgb.y /= 255;
            rgb.z /= 255;
            return rgb;
        }
        static Angle2(ma, mb) {
            var v1 = (ma.x * mb.x) + (ma.y * mb.y) + (ma.z * mb.z);
            var ma_val = Math.sqrt(ma.x * ma.x + ma.y * ma.y + ma.z * ma.z);
            var mb_val = Math.sqrt(mb.x * mb.x + mb.y * mb.y + mb.z * mb.z);
            var cosM = v1 / (ma_val * mb_val);
            if (cosM < -1)
                cosM = -1;
            if (cosM > 1)
                cosM = 1;
            var angleAMB = Math.acos(cosM) * 180 / Math.PI;
            return angleAMB;
        }
        static Save(json, name) {
            var eleLink = Laya.Browser.document.createElement('a');
            eleLink.download = name;
            eleLink.style.display = 'none';
            var blob = new Blob([JSON.stringify(json)]);
            eleLink.href = URL.createObjectURL(blob);
            document.body.appendChild(eleLink);
            eleLink.click();
            document.body.removeChild(eleLink);
        }
        static SaveArray(array, name) {
            var eleLink = Laya.Browser.document.createElement('a');
            eleLink.download = name;
            eleLink.style.display = 'none';
            new ArrayBuffer(array.length * 4);
            var blob = new Blob([array]);
            eleLink.href = URL.createObjectURL(blob);
            document.body.appendChild(eleLink);
            eleLink.click();
            document.body.removeChild(eleLink);
        }
        static vibrateShort() {
            var inst = window['wx'];
            inst = inst == null ? inst = window['qq'] : inst;
            if (window['qg']) {
                inst = window['qg'];
            }
            if (inst && inst.vibrateShort != null) {
                Laya.timer.callLater(inst, inst.vibrateShort);
            }
        }
        static vibrateLong() {
            var inst = window['wx'];
            inst = inst == null ? inst = window['qq'] : inst;
            if (window['qg']) {
                inst = window['qg'];
            }
            if (inst && inst.vibrateLong != null) {
                Laya.timer.callLater(inst, inst.vibrateLong);
            }
        }
        static LoadPrefab(name) {
            return new Promise(r => {
                Laya.Sprite3D.load(`${Global.prefabPathDir}${name}.lh`, Laya.Handler.create(this, (s) => {
                    var s = Laya.Sprite3D.instantiate(s);
                    r(s);
                }));
            });
        }
        static get TodayMS() {
            var now = new Date();
            now.setHours(0, 0, 0, 0);
            return now.getTime();
        }
        static getTodayMS(h = 0, min = 0, s = 0) {
            var now = new Date();
            now.setHours(h, min, s, 0);
            return now.getTime();
        }
        static get CurrentTime() {
            return new Date().getTime();
        }
        static WaitFrame(frameCount) {
            return new Promise(r => {
                Laya.timer.frameOnce(frameCount, null, r);
            });
        }
        static WaitTime(duration) {
            return new Promise(r => {
                if (duration <= 0) {
                    r();
                }
                else {
                    setTimeout(r, duration);
                }
            });
        }
        static WaitWithBlock(checkFunc) {
            return new Promise(r => {
                var time = 0;
                var frameLoop = () => {
                    time += Laya.timer.delta;
                    if (checkFunc()) {
                        Laya.timer.clear(null, frameLoop);
                        r();
                    }
                };
                Laya.timer.frameLoop(1, null, frameLoop);
            });
        }
        static WaitTimeWithBlock(duration, checkFuc) {
            return new Promise(r => {
                var time = 0;
                var frameLoop = () => {
                    var d = Math.min(100, Laya.timer.delta);
                    time += d;
                    if (time >= duration || checkFuc()) {
                        Laya.timer.clear(null, frameLoop);
                        r();
                    }
                };
                Laya.timer.frameLoop(1, null, frameLoop);
            });
        }
        static WaitTimeWithDuartionCall(totalDuration, callDuration, cb) {
            return new Promise(r => {
                var time = 0;
                var cTime = 0;
                cb();
                var frameLoop = () => {
                    var d = Math.min(100, Laya.timer.delta);
                    time += d;
                    cTime += d;
                    if (cTime >= callDuration) {
                        cTime -= callDuration;
                        var cr = cb();
                        if (cr == false) {
                            Laya.timer.clear(null, frameLoop);
                            r();
                        }
                    }
                    if (totalDuration > 0 && time >= totalDuration) {
                        Laya.timer.clear(null, frameLoop);
                        r();
                    }
                };
                Laya.timer.frameLoop(1, null, frameLoop);
            });
        }
        static WaitAnimation(ani, times = 1) {
            return new Promise(r => {
                if (ani != null) {
                    ani.play(Laya.Handler.create(null, r), times);
                }
                else {
                    r();
                }
            });
        }
        static LoadSubPKG(name, progressCB) {
            return new ProgressPromise((r, p, e) => {
                var inst = window['wx'];
                inst = (inst == null) ? inst = window['qq'] : inst;
                if (window['qg']) {
                    inst = window['qg'];
                }
                if (inst == null || inst.loadSubpackage == null) {
                    r();
                    return;
                }
                var task = inst.loadSubpackage({
                    name: name,
                    success: function (res) {
                        console.log(name, '加载成功' + res);
                        r();
                    },
                    fail: function (res) {
                        console.log(name, '加载失败' + res);
                        e(res);
                    }
                });
                task.onProgressUpdate(res => {
                    p(res.totalBytesWritten / res.totalBytesExpectedToWrite);
                });
            });
        }
        static LoadSubPKGs(names) {
            return new ProgressPromise((r, p, e) => {
            });
        }
        static PromiseQueue(arg, ...promises) {
            var p = Promise.resolve(arg);
            if (promises) {
                promises.forEach(promise => {
                    p = p.then(promise);
                });
            }
            return p;
        }
        static SaveImgToTempPath(canvas, imgSrc = null) {
            return new Promise(r => {
                var inst = window['wx'];
                if (inst == null) {
                    r(null);
                    return;
                }
                imgSrc = imgSrc == null ? canvas.toBase64('image/png', 1) : imgSrc;
                var startIdx = imgSrc.indexOf('base64,') + 7;
                var save = inst.getFileSystemManager();
                var number = Utils.CurrentTime;
                var filePath = inst.env.USER_DATA_PATH + '/pic' + number + '.png';
                save.writeFile({
                    filePath: filePath,
                    data: imgSrc.slice(startIdx),
                    encoding: 'base64',
                    success: res => {
                        r(filePath);
                    },
                    fail: err => {
                        console.log(err);
                        r(null);
                    }
                });
            });
        }
        static OppoSaveImgToTempPath(src) {
            return new Promise(r => {
                var inst = window['wx'];
                if (inst == null) {
                    r(null);
                    return;
                }
                var save = inst.getFileSystemManager();
                var number = Utils.CurrentTime;
                var filePath = inst.env.USER_DATA_PATH + '/pic' + number + '.png';
                save.writeFile({
                    filePath: filePath,
                    data: src,
                    encoding: 'binary',
                    success: res => {
                        r(filePath);
                    },
                    fail: err => {
                        console.log(err);
                        r(null);
                    }
                });
            });
        }
        static SaveImgToPhotosAlbum(tempPath) {
            return new Promise(r => {
                var inst = Global.PlatformInst;
                if (inst == null) {
                    r(false);
                    return;
                }
                inst.saveImageToPhotosAlbum({
                    filePath: tempPath,
                    success: function (res) {
                        inst.showToast({
                            title: '保存成功',
                        });
                        inst.showToast({
                            title: '保存成功',
                        });
                        r(true);
                    },
                    fail: function (err) {
                        r(false);
                        console.log(err);
                    }
                });
            });
        }
        static btoa(input) {
            var str = String(input);
            for (var block, charCode, idx = 0, map = chars, output = ''; str.charAt(idx | 0) || (map = '=', idx % 1); output += map.charAt(63 & block >> 8 - idx % 1 * 8)) {
                charCode = str.charCodeAt(idx += 3 / 4);
                if (charCode > 0xFF) {
                    throw new InvalidCharacterError("'btoa' failed: The string to be encoded contains characters outside of the Latin1 range.");
                }
                block = block << 8 | charCode;
            }
            return output;
        }
        static atob(input) {
            var str = (String(input)).replace(/[=]+$/, '');
            if (str.length % 4 === 1) {
                throw new InvalidCharacterError("'atob' failed: The string to be decoded is not correctly encoded.");
            }
            for (var bc = 0, bs, buffer, idx = 0, output = ''; buffer = str.charAt(idx++); ~buffer && (bs = bc % 4 ? bs * 64 + buffer : buffer,
                bc++ % 4) ? output += String.fromCharCode(255 & bs >> (-2 * bc & 6)) : 0) {
                buffer = chars.indexOf(buffer);
            }
            return output;
        }
        static DistancePointWithLine(pos, linePos, lineDirection) {
            return Math.abs(this.SignDistancePointWithLine(pos, linePos, lineDirection));
        }
        static PointToSegmentDist(point, segPont1, segPont2) {
            var cross = (segPont2.x - segPont1.x) * (point.x - segPont1.x) + (segPont2.y - segPont1.y) * (point.y - segPont1.y);
            if (cross <= 0)
                return Math.sqrt((point.x - segPont1.x) * (point.x - segPont1.x) + (point.y - segPont1.y) * (point.y - segPont1.y));
            var d2 = (segPont2.x - segPont1.x) * (segPont2.x - segPont1.x) + (segPont2.y - segPont1.y) * (segPont2.y - segPont1.y);
            if (cross >= d2)
                return Math.sqrt((point.x - segPont2.x) * (point.x - segPont2.x) + (point.y - segPont2.y) * (point.y - segPont2.y));
            var r = cross / d2;
            var px = segPont1.x + (segPont2.x - segPont1.x) * r;
            var py = segPont1.y + (segPont2.y - segPont1.y) * r;
            return Math.sqrt((point.x - px) * (point.x - px) + (py - point.y) * (py - point.y));
        }
        static PointToSegmentDist2(x, y, x1, y1, x2, y2) {
            var cross = (x2 - x1) * (x - x1) + (y2 - y1) * (y - y1);
            if (cross <= 0)
                return Math.sqrt((x - x1) * (x - x1) + (y - y1) * (y - y1));
            var d2 = (x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1);
            if (cross >= d2)
                return Math.sqrt((x - x2) * (x - x2) + (y - y2) * (y - y2));
            var r = cross / d2;
            var px = x1 + (x2 - x1) * r;
            var py = y1 + (y2 - y1) * r;
            return Math.sqrt((x - px) * (x - px) + (py - y) * (py - y));
        }
        static SignDistancePointWithLine(pos, linePos, lineDirection) {
            var proVec = new Laya.Vector3();
            Laya.Vector3.subtract(pos, linePos, proVec);
            var left = new Laya.Vector3(-lineDirection.z, 0, lineDirection.x);
            Laya.Vector3.cross(new Laya.Vector3(0, 0, 1), lineDirection, left);
            Laya.Vector3.normalize(left, left);
            return Laya.Vector3.dot(proVec, left);
        }
        static CheckSegmentLineIntersect(segPos1, segPos2, linePos, lineDirection) {
            var d1 = this.SignDistancePointWithLine(segPos1, linePos, lineDirection);
            var d2 = this.SignDistancePointWithLine(segPos2, linePos, lineDirection);
            if (Laya.MathUtils3D.isZero(d1) || Laya.MathUtils3D.isZero(d2)) {
                return true;
            }
            return Math.sign(d1) != Math.sign(d2);
        }
        static SegmentLineIntersectPos(p0, d0, p1, d1) {
            var d11 = this.DistancePointWithLine(p0, p1, d1);
            var p11 = new Laya.Vector3();
            Laya.Vector3.add(p0, d0, p11);
            var d22 = this.DistancePointWithLine(p11, p1, d1);
            var k = d11 / (d11 + d22);
            var r = new Laya.Vector3();
            Laya.Vector3.scale(d0, k, r);
            Laya.Vector3.add(p0, r, r);
            return r;
        }
        static LineIntersectPos(p0, d0, p1, d1) {
            var e = new Laya.Vector3();
            Laya.Vector3.subtract(p1, p0, e);
            var kross = d0.x * d1.y - d0.y * d1.x;
            var s = (e.x * d1.y - e.y * d1.x) / kross;
            Laya.Vector3.scale(d0, s, d0);
            Laya.Vector3.add(p0, d0, e);
            return e;
        }
        static AngleLerpWithSpeed(src, dis, speed, t) {
            var delta = dis - src;
            if (delta < -180) {
                delta += 360;
            }
            else if (delta > 180) {
                delta -= 360;
            }
            var _speed = speed;
            if (delta < 0) {
                _speed = -speed;
            }
            src = src + _speed * t;
            if (src > 360) {
                src -= 360;
            }
            if (src < 0) {
                src += 360;
            }
            return src;
        }
        static AngleLerp(src, dis, t) {
            var delta = dis - src;
            if (delta < -180) {
                delta += 360;
            }
            else if (delta > 180) {
                delta -= 360;
            }
            src = src + delta * t;
            if (src > 360) {
                src -= 360;
            }
            if (src < 0) {
                src += 360;
            }
            return src;
        }
        static GetAllSKinMeshRender(node, mr) {
            if (node instanceof Laya.SkinnedMeshSprite3D) {
                var m = node.skinnedMeshRenderer;
                mr.push(m);
            }
            if (node instanceof Laya.SimpleSkinnedMeshSprite3D) {
                var sm = node.simpleSkinnedMeshRenderer;
                mr.push(sm);
            }
            for (let i = 0; i < node.numChildren; i++) {
                var child = node.getChildAt(i);
                this.GetAllSKinMeshRender(child, mr);
            }
        }
        static GetAllMeshRenderer(node, mr) {
            if (node instanceof Laya.MeshSprite3D) {
                var m = node.meshRenderer;
                mr.push(m);
            }
            for (let i = 0; i < node.numChildren; i++) {
                var child = node.getChildAt(i);
                this.GetAllMeshRenderer(child, mr);
            }
        }
        static SetAllMeshRendererMat(node, mat) {
            if (node instanceof Laya.MeshSprite3D) {
                var m = node.meshRenderer;
                m.sharedMaterial = mat;
            }
            for (let i = 0; i < node.numChildren; i++) {
                var child = node.getChildAt(i);
                this.SetAllMeshRendererMat(child, mat);
            }
        }
        static SetUnlitColor(node, color) {
            if (node instanceof Laya.MeshSprite3D) {
                let m = node.meshRenderer;
                m.material.albedoColor = color;
            }
            else if (node instanceof Laya.SkinnedMeshSprite3D) {
                let m = node.skinnedMeshRenderer;
                m.material.albedoColor = color;
            }
            for (let i = 0; i < node.numChildren; i++) {
                var child = node.getChildAt(i);
                this.SetUnlitColor(child, color);
            }
        }
        static GetAllPlugMeshRenderer(node, mr) {
            if (node.name != 'model' && node instanceof Laya.MeshSprite3D) {
                var m = node.meshRenderer;
                mr.push(m);
            }
            for (let i = 0; i < node.numChildren; i++) {
                var child = node.getChildAt(i);
                if (child.name != 'model') {
                    this.GetAllPlugMeshRenderer(child, mr);
                }
            }
        }
        static GetAllModelMeshRenderer(node, mr) {
            if (node.name == 'model' && node instanceof Laya.MeshSprite3D) {
                var m = node.meshRenderer;
                mr.push(m);
            }
            for (let i = 0; i < node.numChildren; i++) {
                var child = node.getChildAt(i);
                if (child.name == 'model') {
                    this.GetAllPlugMeshRenderer(child, mr);
                }
            }
        }
        static SetPhysicEnabled(node, enabled = false) {
            var pc = node.getComponent(Laya.PhysicsCollider);
            if (pc != null) {
                pc.enabled = enabled;
            }
            var r = node.getComponent(Laya.Rigidbody3D);
            if (r != null) {
                r.enabled = enabled;
            }
            for (var i = 0; i < node.numChildren; i++) {
                this.SetPhysicEnabled(node.getChildAt(i), enabled);
            }
        }
        static SetPhysicGroup(node, group) {
            var pc = node.getComponent(Laya.PhysicsCollider);
            if (pc != null) {
                pc.collisionGroup = group;
            }
            var r = node.getComponent(Laya.Rigidbody3D);
            if (r != null) {
                r.collisionGroup = group;
            }
            for (var i = 0; i < node.numChildren; i++) {
                this.SetPhysicGroup(node.getChildAt(i), group);
            }
        }
        static SetPhysicCanWithGroup(node, group) {
            var pc = node.getComponent(Laya.PhysicsCollider);
            if (pc != null) {
                pc.canCollideWith = group;
            }
            var r = node.getComponent(Laya.Rigidbody3D);
            if (r != null) {
                r.canCollideWith = group;
            }
            for (var i = 0; i < node.numChildren; i++) {
                this.SetPhysicCanWithGroup(node.getChildAt(i), group);
            }
        }
        static ParseVector3(str) {
            var strs = str.split(',');
            return new Laya.Vector3(Number(strs[0]), Number(strs[1]), Number(strs[2]));
        }
        static GetNode(node, name) {
            if (node.name == name) {
                return node;
            }
            if (node.numChildren == 0) {
                return null;
            }
            for (var i = 0; i < node.numChildren; i++) {
                var n = this.GetNode(node.getChildAt(i), name);
                if (n != null) {
                    return n;
                }
            }
        }
        static TimeToStringMS(mil, type = 0) {
            var seconds = Math.floor(mil / 1000);
            var s = seconds % 60;
            var ms = Math.round((mil % 1000) / 1000 * 100);
            var min = Math.floor(seconds / 60) % 60;
            if (type == 0) {
                return Utils.formatNumber(min) + ':' + Utils.formatNumber(s) + ':' + Utils.formatNumber(ms);
            }
            else {
                return Utils.formatNumber(min) + '小时' + Utils.formatNumber(s) + '分钟' + Utils.formatNumber(ms) + '秒';
            }
        }
        static TimeToStringMS2(mil, type = 0) {
            var seconds = Math.floor(mil / 1000);
            var s = seconds % 60;
            var ms = Math.round((mil % 1000) / 1000 * 100);
            var min = Math.floor(seconds / 60) % 60;
            var h = Math.floor(seconds / 3600) % 24;
            if (type == 0) {
                return Utils.formatNumber(h) + ':' + Utils.formatNumber(min) + ':' + Utils.formatNumber(s);
            }
            else {
                return Utils.formatNumber(min) + '小时' + Utils.formatNumber(s) + '分钟' + Utils.formatNumber(ms) + '秒';
            }
        }
        static SetFXColor(sp, color) {
            if (sp instanceof Laya.ShuriKenParticle3D) {
                sp.particleRenderer.material.color = color;
            }
            for (var i = 0; i < sp.numChildren; i++) {
                this.SetFXColor(sp.getChildAt(i), color);
            }
        }
        static AddComponent(node, Tt) {
            return node.addComponent(Tt);
        }
        static truncate(a, maxLen, out) {
            var disSq = Laya.Vector3.scalarLengthSquared(a);
            if (disSq > maxLen * maxLen) {
                Laya.Vector3.normalize(a, out);
                Laya.Vector3.scale(out, maxLen, out);
            }
            else {
                a.cloneTo(out);
            }
        }
        static setScalarLen(a, len, out) {
            Laya.Vector3.normalize(a, out);
            Laya.Vector3.scale(out, len, out);
        }
        static get safeArea() {
            if (this._safeArea == null) {
                var inst = window['wx'] || window['tt'] || window['qg'] || window['qq'];
                if (inst != null && inst.getSystemInfoSync != null) {
                    this._safeArea = inst.getSystemInfoSync().safeArea;
                }
                else {
                    this._safeArea = new SafeArea();
                    console.log(this._safeArea);
                }
            }
            return this._safeArea;
        }
    }
    Utils.isNextLevel = false;
    class SafeArea {
        constructor() {
            this.bottom = this.height = Laya.Browser.clientHeight;
            this.top = 0;
            this.left = 0;
            this.right = this.width = Laya.Browser.clientWidth;
        }
    }
    function throttle(time = 300) {
        return function (target, propertyKey, descriptor) {
            let oldFun = descriptor.value;
            let isLock = false;
            descriptor.value = function (...args) {
                if (isLock) {
                    return;
                }
                isLock = true;
                setTimeout(() => {
                    isLock = false;
                }, time);
                oldFun.apply(this, args);
            };
            return descriptor;
        };
    }

    class BaseScript3D extends Laya.Script3D {
        get sp() {
            return this.owner;
        }
        get transform() {
            return this.sp.transform;
        }
        get ani() {
            if (this._ani == null) {
                this._ani = this.sp.getComponent(Laya.Animator);
                if (this._ani == null) {
                    var node = this.Find("Animator");
                    if (node != null) {
                        this._ani = node.getComponent(Laya.Animator);
                    }
                }
            }
            return this._ani;
        }
        PlayAni(clip, fadeTime) {
            if (this._ani) {
                if (fadeTime == null) {
                    this.ani.play(clip);
                }
                else {
                    this.ani.crossFade(clip, fadeTime);
                }
            }
        }
        SwitchParent(parent) {
            let wm = this.sp.transform.worldMatrix.clone();
            parent.addChild(this.sp);
            this.sp.transform.worldMatrix = wm;
        }
        Find(childName) {
            return Utils.GetNode(this.sp, childName);
        }
        UpdatePosition() {
            this.sp.transform.position = this.sp.transform.position;
        }
    }

    class ArrowTips extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.disVec = new Laya.Vector3();
            this.show = true;
        }
        onAwake() {
            ArrowTips.inst = this;
            var ms = this.view = this.sp.getChildAt(0);
            this.mat = ms.meshRenderer.sharedMaterial;
        }
        onDestroy() {
            ArrowTips.inst = null;
        }
        Show() {
            this.show = true;
            this.view.active = true;
        }
        Hide() {
            this.show = false;
            this.view.active = false;
        }
        onUpdate() {
            if (this.hostEntity.isDeath) {
                this.sp.destroy(true);
                return;
            }
            if (!this.show || this.target == null)
                return;
            var hostNode = this.hostNode;
            if (hostNode == null || hostNode.destroyed) {
                this.hostNode = null;
                hostNode = this.hostEntity.aniView.sp;
            }
            if (hostNode == null)
                return;
            this.transform.position = hostNode.transform.position;
            this.transform.position.y = 0.3;
            this.UpdatePosition();
            Laya.Vector3.subtract(this.target.transform.position, this.transform.position, this.disVec);
            var dis = Laya.Vector3.scalarLength(this.disVec);
            if (dis < 1) {
                this.view.active = false;
                return;
            }
            this.view.active = true;
            this.transform.localScaleZ = dis;
            this.mat.tilingOffsetY = dis;
            this.transform.lookAt(this.target.transform.position, LayaUtils.UP);
            this.transform.localRotationEulerY += 180;
            this.transform.localRotationEulerX = this.transform.localRotationEulerZ = 0;
            this.mat.tilingOffsetW -= 5 * LayaUtils.deltaTimeSec;
        }
    }

    class ResLoadManager {
        static PushUrls(isCreate, ...urls) {
            if (isCreate) {
                for (var i = 0; i < urls.length; i++) {
                    var has = ResLoadManager.urlsCreate.indexOf(urls[i]) >= 0;
                    if (!has) {
                        ResLoadManager.urlsCreate.push(urls[i]);
                    }
                }
            }
            else {
                for (var i = 0; i < urls.length; i++) {
                    if (urls[i] != null) {
                        var has = ResLoadManager.urlsLoad.indexOf(urls[i]) >= 0;
                        if (!has) {
                            ResLoadManager.urlsLoad.push(urls[i]);
                        }
                    }
                }
            }
        }
        static LoadCreateRes(completed, progress) {
            console.log('load create urls:', ResLoadManager.urlsCreate);
            if (ResLoadManager.urlsCreate.length == 0) {
                completed(true);
            }
            else {
                Laya.loader.create(ResLoadManager.urlsCreate, Laya.Handler.create(null, completed), progress ? Laya.Handler.create(null, progress, null, false) : null);
            }
        }
        static LoadCreateResP() {
            return new ProgressPromise((r, p, e) => {
                if (ResLoadManager.urlsCreate == null || ResLoadManager.urlsCreate.length == 0) {
                    p(1);
                    r();
                    return;
                }
                Laya.loader.create(ResLoadManager.urlsCreate, Laya.Handler.create(null, r), Laya.Handler.create(null, p, null, false));
            });
        }
        static LoadCacheRes(completed, progress) {
            console.log('load urls:', ResLoadManager.urlsLoad);
            if (ResLoadManager.urlsLoad.length == 0) {
                completed(true);
            }
            else {
                var item = [];
                ResLoadManager.urlsLoad.forEach(url => {
                    if (url.indexOf('.bin') != -1) {
                        item.push({ url: url, type: Laya.Loader.BUFFER });
                    }
                    else if (url.indexOf('.png') != -1 || url.indexOf('.jpg') != -1) {
                        item.push({ url: url, type: Laya.Loader.IMAGE });
                    }
                });
                Laya.loader.create(item, Laya.Handler.create(null, completed), progress ? Laya.Handler.create(null, progress, null, false) : null, null, null, null, 1, true);
            }
        }
        static LoadPrefabRes(urls, completed, progress, caller) {
            if (urls.length == 0) {
                completed(true);
                return;
            }
            Laya.loader.create(urls, Laya.Handler.create(caller, completed), progress ? Laya.Handler.create(null, progress, null, false) : null);
        }
        static ClearRes(urls) {
            urls.forEach(url => {
                Laya.loader.clearRes(url);
            });
        }
    }
    ResLoadManager.urlsCreate = [];
    ResLoadManager.urlsLoad = [];

    function res_auto_bind() {
        return function (_constructor) {
            var newConstructor = class T extends _constructor {
                constructor(...args) {
                    super(...args);
                    var resPrefabClone = this['_res_prefab_clone'];
                    if (resPrefabClone) {
                        for (var i = 0; i < resPrefabClone.length; i++) {
                            var resBind = resPrefabClone[i];
                            this[resBind.key] = Laya.Sprite3D.instantiate(Laya.loader.getRes(resBind.url));
                        }
                        this['_res_prefab_clone'] = null;
                    }
                    var resAssigns = this['_res_assign'];
                    if (resAssigns) {
                        for (var i = 0; i < resAssigns.length; i++) {
                            var resAssign = resAssigns[i];
                            this[resAssign.key] = Laya.loader.getRes(resAssign.url);
                        }
                        this['_res_assign'] = null;
                    }
                    var resArrayAssigns = this['_res_array_assign'];
                    if (resArrayAssigns) {
                        for (var i = 0; i < resArrayAssigns.length; i++) {
                            var resArrayAssign = resArrayAssigns[i];
                            var arry = this[resArrayAssign.key] = [];
                            for (var j = 0; j < resArrayAssign.urls.length; j++) {
                                var url = resArrayAssign.urls[j];
                                arry.push(Laya.loader.getRes(url));
                            }
                        }
                        this['_res_array_assign'] = null;
                    }
                    if (this['_init_']) {
                        this['_init_']();
                    }
                }
            };
            return newConstructor;
        };
    }
    function res_assign(url) {
        ResLoadManager.PushUrls(true, url);
        return function (target, key) {
            target['_res_assign'] = target['_res_assign'] || [];
            var v = target['_res_assign'];
            v.push({ key: key, url: url });
        };
    }

    var BuildingConfigPoxy_1;
    let BuildingConfigPoxy = BuildingConfigPoxy_1 = class BuildingConfigPoxy {
        static get Inst() {
            if (BuildingConfigPoxy_1._inst == null) {
                BuildingConfigPoxy_1._inst = new BuildingConfigPoxy_1();
            }
            return BuildingConfigPoxy_1._inst;
        }
    };
    __decorate([
        res_assign("res/ConfigJson/BuildingConfig.json")
    ], BuildingConfigPoxy.prototype, "datas", void 0);
    BuildingConfigPoxy = BuildingConfigPoxy_1 = __decorate([
        res_auto_bind()
    ], BuildingConfigPoxy);

    var CommanderConfigPoxy_1;
    let CommanderConfigPoxy = CommanderConfigPoxy_1 = class CommanderConfigPoxy {
        static get Inst() {
            if (CommanderConfigPoxy_1._inst == null) {
                CommanderConfigPoxy_1._inst = new CommanderConfigPoxy_1();
            }
            return CommanderConfigPoxy_1._inst;
        }
    };
    __decorate([
        res_assign("res/ConfigJson/CommanderConfig.json")
    ], CommanderConfigPoxy.prototype, "datas", void 0);
    CommanderConfigPoxy = CommanderConfigPoxy_1 = __decorate([
        res_auto_bind()
    ], CommanderConfigPoxy);

    var GlobalConstPoxy_1;
    let GlobalConstPoxy = GlobalConstPoxy_1 = class GlobalConstPoxy {
        static get Inst() {
            if (GlobalConstPoxy_1._inst == null) {
                GlobalConstPoxy_1._inst = new GlobalConstPoxy_1();
            }
            return GlobalConstPoxy_1._inst;
        }
    };
    __decorate([
        res_assign("res/ConfigJson/GlobalConst.json")
    ], GlobalConstPoxy.prototype, "data", void 0);
    GlobalConstPoxy = GlobalConstPoxy_1 = __decorate([
        res_auto_bind()
    ], GlobalConstPoxy);

    var LevelConfigPoxy_1;
    let LevelConfigPoxy = LevelConfigPoxy_1 = class LevelConfigPoxy {
        static get Inst() {
            if (LevelConfigPoxy_1._inst == null) {
                LevelConfigPoxy_1._inst = new LevelConfigPoxy_1();
            }
            return LevelConfigPoxy_1._inst;
        }
    };
    __decorate([
        res_assign("res/ConfigJson/LevelConfig.json")
    ], LevelConfigPoxy.prototype, "datas", void 0);
    LevelConfigPoxy = LevelConfigPoxy_1 = __decorate([
        res_auto_bind()
    ], LevelConfigPoxy);

    var SoliderConfigPoxy_1;
    let SoliderConfigPoxy = SoliderConfigPoxy_1 = class SoliderConfigPoxy {
        static get Inst() {
            if (SoliderConfigPoxy_1._inst == null) {
                SoliderConfigPoxy_1._inst = new SoliderConfigPoxy_1();
            }
            return SoliderConfigPoxy_1._inst;
        }
    };
    __decorate([
        res_assign("res/ConfigJson/SoliderConfig.json")
    ], SoliderConfigPoxy.prototype, "datas", void 0);
    SoliderConfigPoxy = SoliderConfigPoxy_1 = __decorate([
        res_auto_bind()
    ], SoliderConfigPoxy);

    var WeaponConfigPoxy_1;
    let WeaponConfigPoxy = WeaponConfigPoxy_1 = class WeaponConfigPoxy {
        static get Inst() {
            if (WeaponConfigPoxy_1._inst == null) {
                WeaponConfigPoxy_1._inst = new WeaponConfigPoxy_1();
            }
            return WeaponConfigPoxy_1._inst;
        }
    };
    __decorate([
        res_assign("res/ConfigJson/WeaponConfig.json")
    ], WeaponConfigPoxy.prototype, "datas", void 0);
    WeaponConfigPoxy = WeaponConfigPoxy_1 = __decorate([
        res_auto_bind()
    ], WeaponConfigPoxy);

    class ConfigHelper {
        static get WeaponConfigMap() {
            if (this._weaponConfigMap == null) {
                this._weaponConfigMap = {};
                WeaponConfigPoxy.Inst.datas.forEach(config => {
                    this._weaponConfigMap[config.id] = config;
                });
            }
            return this._weaponConfigMap;
        }
        static get InitWeaponConfig() {
            return this.WeaponConfigMap[GlobalConstPoxy.Inst.data.initWeapon];
        }
        static GetWeaponByPrefab(prefab) {
            var weapons = WeaponConfigPoxy.Inst.datas;
            for (var i = 0; i < weapons.length; i++) {
                if (weapons[i].prefab === prefab) {
                    return weapons[i];
                }
            }
        }
        static get SoliderConfigMap() {
            if (this._soliderConfigMap == null) {
                this._soliderConfigMap = {};
                SoliderConfigPoxy.Inst.datas.forEach(config => {
                    this._soliderConfigMap[config.id] = config;
                });
            }
            return this._soliderConfigMap;
        }
        static GetSoliderByPrefab(prefab) {
            let soliders = SoliderConfigPoxy.Inst.datas;
            for (var i = 0; i < soliders.length; i++) {
                if (soliders[i].prefab === prefab) {
                    return soliders[i];
                }
            }
        }
        static get CommanderConfigLevelMap() {
            if (this._commanderConfigMap == null) {
                this._commanderConfigMap = {};
                CommanderConfigPoxy.Inst.datas.forEach(config => {
                    this._commanderConfigMap[config.id] = config;
                });
            }
            return this._commanderConfigMap;
        }
        static get FirstLevel() {
            return LevelConfigPoxy.Inst.datas[0];
        }
        static get MaxrankLevel() {
            return CommanderConfigPoxy.Inst.datas[CommanderConfigPoxy.Inst.datas.length - 1];
        }
        static get LevelMap() {
            if (this.levelMap == null) {
                this.levelMap = {};
                LevelConfigPoxy.Inst.datas.forEach(level => {
                    this.levelMap[level.id] = level;
                });
            }
            return this.levelMap;
        }
        static get BuildingMap() {
            if (this._buildingMap == null) {
                this._buildingMap = {};
                BuildingConfigPoxy.Inst.datas.forEach(config => {
                    this._buildingMap[config.id] = config;
                });
            }
            return this._buildingMap;
        }
        static get BuildingInRankMap() {
            if (this._buildingInRankMap == null) {
                this._buildingInRankMap = {};
                CommanderConfigPoxy.Inst.datas.forEach(config => {
                    var buildingGroup = this._buildingInRankMap[config.level] = this._buildingInRankMap[config.level] || {};
                    var buildingsInRank = this.GetBuildingInRank(config.level);
                    buildingsInRank.forEach(building => {
                        var buildingType = building.type;
                        var group = buildingGroup[buildingType] = buildingGroup[buildingType] || {};
                        group[building.slot_name] = group[building.slot_name] || [];
                        group[building.slot_name].push(building);
                    });
                });
            }
            return this._buildingInRankMap;
        }
        static get CurrentLevelBuildingMap() {
            return this.BuildingInRankMap[ConfigHelper.LevelMap[PlayerData.CurrentLevelID].max_rank];
        }
        static get CurrentLevelSoliderTents() {
            return this.CurrentLevelBuildingMap[BuilidngType.SoliderTent];
        }
        static get CurrentLevelTankTents() {
            return this.CurrentLevelBuildingMap[BuilidngType.Tankhangat];
        }
        static get WeaponFactorys() {
            return this.CurrentLevelBuildingMap[BuilidngType.AmmoFactory]['AmmoFactory'];
        }
        static get Airport() {
            return this.CurrentLevelBuildingMap[BuilidngType.AirportBuilding]['AirportBuilding'];
        }
        static GetBuildingInRank(rank) {
            return BuildingConfigPoxy.Inst.datas;
        }
        static get BattleSceneList() {
            if (this.battleSceneList == null) {
                this.battleSceneList = [];
                LevelConfigPoxy.Inst.datas.forEach(config => {
                    for (var i = 0; i < config.scenes.length; i++) {
                        var scene = config.scenes[i];
                        var isBoss = (i == config.scenes.length - 1);
                        this.battleSceneList.push({ scene, isBoss });
                    }
                });
            }
            return this.battleSceneList;
        }
    }
    var SoilderType;
    (function (SoilderType) {
        SoilderType["Soilder"] = "solider";
        SoilderType["Tank"] = "tank";
        SoilderType["Builidng"] = "building";
        SoilderType["Cammander"] = "cammander";
        SoilderType["Car"] = "car";
        SoilderType["Medic"] = "medic";
    })(SoilderType || (SoilderType = {}));
    var BuilidngType;
    (function (BuilidngType) {
        BuilidngType["SoliderTent"] = "1";
        BuilidngType["AmmoFactory"] = "2";
        BuilidngType["Tankhangat"] = "3";
        BuilidngType["MountBuilding"] = "4";
        BuilidngType["MedicaTent"] = "5";
        BuilidngType["MortarBuilding"] = "6";
        BuilidngType["AirportBuilding"] = "7";
        BuilidngType["StorageBox"] = "8";
        BuilidngType["chuan1"] = "9";
        BuilidngType["NukeBuilding"] = "10";
    })(BuilidngType || (BuilidngType = {}));

    class LocalStorage {
        static setItem(key, value, force = false) {
            if (this.cache[key] != value || force) {
                if (value == null || value == '') {
                    Laya.LocalStorage.removeItem(key);
                    delete this.cache[key];
                }
                else {
                    Laya.LocalStorage.setItem(key, JSON.stringify(value));
                }
            }
            this.cache[key] = value;
            if (value == null || value == '') {
                delete this.cache[key];
            }
        }
        static getItem(key) {
            if (this.cache[key] == undefined) {
                var str = Laya.LocalStorage.getItem(key);
                if (str != null && str != '' && this.isJSON(str)) {
                    this.cache[key] = JSON.parse(str);
                }
                else {
                    this.cache[key] = str;
                }
            }
            return this.cache[key];
        }
        static isJSON(str) {
            if (typeof str == 'string') {
                try {
                    var obj = JSON.parse(str);
                    return true;
                }
                catch (e) {
                    return false;
                }
            }
        }
        static clear() {
            Laya.LocalStorage.clear();
        }
        static removeItem(key) {
            delete this.cache[key];
            Laya.LocalStorage.removeItem(key);
        }
        static get Cache() {
            return this.cache;
        }
        static set Cache(v) {
            Object.keys(this.cache).forEach(key => {
                v[key] = this.cache[key];
            });
            this.cache = v;
            Object.keys(v).forEach(key => {
                LocalStorage.setItem(key, this.cache[key], true);
            });
        }
    }
    LocalStorage.cache = {};

    class SoundManager {
        constructor() {
            this.bgm = null;
            this.stopState = true;
            this.isPlayingBGM = false;
            this.times = 1;
            this.pauseCount = 0;
            MsgManager.getInst().on(EventType.Setting, this, this.onSetting);
            Laya.SoundManager.musicVolume = 0.5;
            Laya.SoundManager.soundVolume = 0.7;
            Laya.SoundManager.soundMuted = true;
            Laya.SoundManager.musicMuted = true;
            fgui.GRoot.inst.playOneShotSound = (url, volumeScale = 1) => {
                if (fgui.ToolSet.startsWith(url, "ui://"))
                    return;
                this.playSound(url);
            };
            this.endCB = this.onMusicEnded.bind(this);
            this.Refresh();
            GameSDK.registShowCaller(this.PlayBGM, this);
            GameSDK.registShowCaller(() => {
                Laya.timer.scale = 1;
                Laya.SoundManager.muted = false;
            }, this);
            GameSDK.registHideCaller(this.StopBGM, this);
            GameSDK.registHideCaller(() => {
                Laya.timer.scale = 0;
                Laya.SoundManager.muted = true;
            }, this);
        }
        static get inst() {
            if (!SoundManager._inst) {
                SoundManager._inst = new SoundManager;
                Laya.SoundManager.autoReleaseSound = false;
            }
            return SoundManager._inst;
        }
        Faker() {
        }
        get vibrate() {
            var s = LocalStorage.getItem('vibrate');
            if (s == null || s == '') {
                this.vibrate = true;
                return true;
            }
            return Number(s) == 1;
        }
        ;
        set vibrate(v) {
            var s = v ? 1 : 0;
            LocalStorage.setItem('vibrate', s.toString());
        }
        ;
        get music() {
            var s = LocalStorage.getItem('music');
            if (s == null || s == '') {
                this.music = true;
                return true;
            }
            return Number(s) == 1;
        }
        ;
        set music(v) {
            var s = v ? 1 : 0;
            LocalStorage.setItem('music', s.toString());
        }
        ;
        get sound() {
            var s = LocalStorage.getItem('sound');
            if (s == null || s == '') {
                this.sound = true;
                return true;
            }
            return Number(s) == 1;
        }
        ;
        set sound(v) {
            var s = v ? 1 : 0;
            LocalStorage.setItem('sound', s.toString());
        }
        ;
        onSetting(music, sound, vibrate) {
            this.Refresh();
        }
        Refresh() {
            Laya.SoundManager.soundMuted = !this.sound;
            Laya.SoundManager.musicMuted = !this.music;
            console.log('playr music', this.music);
            if (this.music) {
                this.PlayBGM('Bgm');
            }
        }
        PlayBGM(url, volume = 1) {
            url = url || this.bgm;
            this.bgm = url;
            if (url.indexOf(Global.soundDir) == -1) {
                url = Global.soundDir + url;
            }
            if (url.indexOf('.mp3') == -1) {
                url = url + '.mp3';
            }
            this.stopState = true;
            this.isPlayingBGM = true;
            this.stopState = false;
            var au = Laya.SoundManager.playMusic(url, 0);
            if (au) {
                au.volume = volume;
            }
        }
        StopBGM() {
            this.stopState = true;
            Laya.SoundManager.stopMusic();
        }
        onBGMEnd() {
            if (!this.stopState) {
                this.isPlayingBGM = false;
                this.PlayBGM(this.bgm);
            }
        }
        EnemyDeathVibrate() {
            this.playVibrate(false);
        }
        playVibrate(long) {
            if (!this.vibrate)
                return;
            navigator.vibrate = navigator.vibrate || navigator['webkitVibrate'] || navigator['mozVibrate'] || navigator['msVibrate'];
            if (navigator.vibrate) {
                navigator.vibrate(long ? 1000 : 200);
            }
            if (long) {
                Utils.vibrateLong();
            }
            else {
                Utils.vibrateShort();
            }
        }
        playCreateBuilding() {
            this.playSound('TentBuild');
        }
        playSound(url, volume = 1) {
            if (!this.sound)
                return;
            if (url.indexOf(Global.soundDir) == -1) {
                url = Global.soundDir + url;
            }
            if (url.indexOf('.mp3') == -1) {
                url = url + '.mp3';
            }
            var ac = Laya.SoundManager.playSound(url);
            if (ac) {
                ac.volume = volume;
            }
            return ac;
            if (Global.Platform == PlatformType.Web) {
                var ac = Laya.SoundManager.playSound(url);
            }
        }
        playShoot(name) {
            if (Laya.Browser.onTTMiniGame) {
                this.playSound(name);
            }
            else {
                Laya.timer.callLater(this, this.playSound, [name, 1]);
            }
        }
        PlayMusic(url, times = 1) {
            if (!this.music) {
                return;
            }
            if (Global.Platform == PlatformType.Web) {
                SoundManager.inst.PlayBGM(url);
            }
            else {
                if (this.curAC != null) {
                    this._stopMusic(this.curAC);
                }
                this.curMusic = url;
                this.curAC = this.createAC(url);
                if (this.pauseCount <= 0) {
                    this.curAC.play();
                }
                this.curAC.onEnded(this.endCB);
                this.times = times;
            }
        }
        createAC(url) {
            var ac = Global.PlatformInst.createInnerAudioContext();
            ac.src = url;
            return ac;
        }
        PauseMusic() {
            this.pauseCount++;
            if (this.curAC != null) {
                this.curAC.volume = 0;
                this.curAC.pause();
            }
        }
        ResumeMusic() {
            this.pauseCount--;
            this.pauseCount = Math.max(this.pauseCount, 0);
            if (this.pauseCount <= 0) {
                if (this.curAC != null && this.music) {
                    this.curAC.volume = 1;
                    this.curAC.play();
                }
            }
        }
        _stopMusic(ac) {
            ac.offEnded(this.endCB);
            ac.stop();
            ac.destroy();
        }
        onMusicEnded() {
            if (this.times == -1) {
                this.curAC.play();
            }
            else {
                this.times--;
                if (this.times > 0) {
                    this.curAC.play();
                }
            }
        }
    }
    __decorate([
        throttle(1000)
    ], SoundManager.prototype, "EnemyDeathVibrate", null);
    __decorate([
        throttle(100)
    ], SoundManager.prototype, "playCreateBuilding", null);

    class DataManager {
        constructor() {
            this.dataMap = {};
            this.recordTime = false;
            Laya.timer.frameLoop(5, this, this.Save);
        }
        static get inst() {
            if (DataManager._inst == null)
                DataManager._inst = new DataManager();
            return DataManager._inst;
        }
        static RegisterCtor(key, ctor) {
            this.storageName2ctorMap[key] = ctor;
            this.ctorName2storageNameMap[ctor.name] = key;
        }
        GetDataRecorder(clas) {
            var storageName = DataManager.ctorName2storageNameMap[clas.name];
            if (typeof clas == 'string') {
                storageName = clas;
            }
            if (this.dataMap[storageName] != null) {
                return this.dataMap[storageName];
            }
            if (DataManager.storageName2ctorMap[storageName] == null) {
                console.error('未注册的', storageName);
            }
            var data = null;
            if (this.storageGet) {
                data = this.storageGet(storageName);
            }
            else {
                var r = LocalStorage.getItem(storageName);
                if (r == null || r == '') {
                    data = {};
                }
                else {
                    data = r;
                    if (typeof r == 'string') {
                        data = JSON.parse(r);
                    }
                }
            }
            var t = new DataManager.storageName2ctorMap[storageName](storageName, data);
            this.dataMap[storageName] = t.proxy;
            return t.proxy;
        }
        Clear(name) {
            delete this.dataMap[name];
            LocalStorage.removeItem(name);
        }
        Save() {
            var keys = Object.keys(this.dataMap);
            for (var i = 0; i < keys.length; i++) {
                var key = keys[i];
                var dp = this.dataMap[key];
                if (dp.dirty) {
                    if (this.storageSet) {
                        this.storageSet(dp.storageName, dp);
                    }
                    else {
                        LocalStorage.setItem(dp.storageName, dp, true);
                    }
                    dp.ClearDirty();
                }
            }
        }
        setTime() {
            LocalStorage.setItem('record_time', Utils.CurrentTime);
        }
        get GetLastTime() {
            if (this.recordTime == false) {
                this.recordTime = true;
                Laya.timer.loop(1000, this, this.setTime);
            }
            var r = LocalStorage.getItem('record_time');
            if (r == null || r == '') {
                return -1;
            }
            return Number(r);
        }
    }
    DataManager.storageName2ctorMap = {};
    DataManager.ctorName2storageNameMap = {};

    class DataProxy {
        constructor(name, data) {
            this.storageName = name;
            this._data = data;
            this._proxy = new Proxy(this, {
                set: (target, p, v) => {
                    switch (p) {
                        case '_dirty':
                        case '_proxy':
                        case '_data':
                            target[p] = v;
                            return true;
                    }
                    if (target._data[p] != v) {
                        if (target._data[p] == null && v instanceof Object) {
                            var baseData = new BaseData(p, v);
                            target._data[p] = baseData.proxy;
                            baseData.__dirtyInvoker = this.dirtyInvoker.bind(this);
                            this._dirty = true;
                            try {
                                this.dirtyInvoker(p);
                            }
                            catch (_error) {
                                console.error(_error);
                                console.error(name, data);
                            }
                            return true;
                        }
                        target._data[p] = v;
                        if (target._data[p] == null) {
                            delete target._data[p];
                        }
                        this._dirty = true;
                        this.dirtyInvoker(p);
                    }
                    return true;
                },
                get: (target, p) => {
                    if (target[p] != null) {
                        return target[p];
                    }
                    var pv = target._data.p;
                    if (pv != null) {
                        return pv;
                    }
                    if (target._data[p] != null) {
                        return target._data[p];
                    }
                    if (target._data.prototype != null) {
                        let v = target._data.prototype.p;
                        if (v instanceof Object) {
                            var baseData = new BaseData(p, v);
                            target._data[p] = baseData.proxy;
                            baseData.__dirtyInvoker = this.dirtyInvoker.bind(this);
                        }
                        return target._data[p];
                    }
                    return null;
                },
                ownKeys: (target) => {
                    var keys = Object.keys(target._data);
                    for (var i = 0; i < keys.length; i++) {
                        target[keys[i]] = null;
                    }
                    return keys;
                },
                has: (target, p) => {
                    return p in target._data;
                }
            });
            var keys = Object.keys(this._data);
            for (var i = 0; i < keys.length; i++) {
                var v = this._data[keys[i]];
                if (v instanceof Object) {
                    var baseData = new BaseData(keys[i], v);
                    baseData.__dirtyInvoker = this.dirtyInvoker.bind(this);
                    this._data[keys[i]] = baseData.proxy;
                }
            }
        }
        get proxy() {
            return this._proxy;
        }
        get data() {
            return this._data;
        }
        get dirty() {
            return this._dirty == null ? false : this._dirty;
        }
        ClearDirty() {
            this._dirty = false;
        }
        toJSON() {
            return this._data;
        }
        dirtyInvoker(path) {
            this._dirty = true;
            if (this.__dirtyInvoker) {
                this.__dirtyInvoker(this.storageName + '/' + path);
            }
            else {
                DataManager.ctorName2storageNameMap[this._proxy.constructor.name];
            }
        }
    }
    class BaseData extends DataProxy {
    }

    function data_bind(storageName) {
        return function (_constructor) {
            DataManager.RegisterCtor(storageName, _constructor);
            return _constructor;
        };
    }

    class BuildingState {
    }
    class PosState {
    }
    let PlayerRecorder = class PlayerRecorder extends DataProxy {
    };
    PlayerRecorder = __decorate([
        data_bind('playerInfo')
    ], PlayerRecorder);
    let BattleRecorder = class BattleRecorder extends DataProxy {
    };
    BattleRecorder = __decorate([
        data_bind('battleInfo')
    ], BattleRecorder);
    class PlayerData {
        static get recoderPlayer() {
            if (this._recorderPlayer == null) {
                this._recorderPlayer = DataManager.inst.GetDataRecorder('playerInfo');
            }
            return this._recorderPlayer;
        }
        static get UnlockNuke() {
            if (this.recoderPlayer.unlockNuke == null) {
                this.recoderPlayer.unlockNuke = false;
            }
            return this.recoderPlayer.unlockNuke;
        }
        static set UnlockNuke(v) {
            this.recoderPlayer.unlockNuke = v;
        }
        static get YuyueNuke() {
            if (this.recoderPlayer.yuyueNuke == null) {
                this.recoderPlayer.yuyueNuke = false;
            }
            return this.recoderPlayer.yuyueNuke;
        }
        static set YuyueNuke(v) {
            if (v && !!!this.recoderPlayer.yuyueNuke && !!!this.recoderPlayer.yuyueDate) {
                var day_date = new Date(new Date(new Date().getTime() + 24 * 60 * 60 * 1000).setHours(1, 0, 0, 0)).getTime();
                PlayerData.YuyueDate = day_date;
            }
            this.recoderPlayer.yuyueNuke = v;
        }
        static get YuyueDate() {
            if (this.recoderPlayer.yuyueDate == null) {
                return Number.MAX_SAFE_INTEGER;
            }
            return this.recoderPlayer.yuyueDate;
        }
        static set YuyueDate(v) {
            this.recoderPlayer.yuyueDate = v;
        }
        static get NormalNuke() {
            if (this.recoderPlayer.normalNuke == null) {
                this.recoderPlayer.normalNuke = false;
            }
            return this.recoderPlayer.normalNuke;
        }
        static set FreeNuke(v) {
            this.recoderPlayer.freeNuke = v;
        }
        static get FreeNuke() {
            if (this.recoderPlayer.freeNuke == null) {
                this.recoderPlayer.freeNuke = 0;
            }
            return this.recoderPlayer.freeNuke;
        }
        static set NormalNuke(v) {
            this.recoderPlayer.normalNuke = v;
        }
        static get CurrentRankLevel() {
            if (this.recoderPlayer.currentRankLevel == null) {
                this.recoderPlayer.currentRankLevel = 1;
            }
            return this.recoderPlayer.currentRankLevel;
        }
        static set CurrentRankLevel(lv) {
            var oldLv = this.recoderPlayer.currentLevelID;
            this.recoderPlayer.currentRankLevel = lv;
            if (lv > oldLv) {
                SoundManager.inst.playSound('RankUp');
                MsgManager.getInst().event(BattleEvent.RankLevelUp);
            }
        }
        static get CurrentCommanderConfig() {
            return ConfigHelper.CommanderConfigLevelMap[this.CurrentRankLevel];
        }
        static get CurrentLevelID() {
            if (this.recoderPlayer.currentLevelID == null) {
                this.recoderPlayer.currentLevelID = ConfigHelper.FirstLevel.id;
            }
            return this.recoderPlayer.currentLevelID;
        }
        static set CurrentLevelID(v) {
            this.recoderPlayer.currentLevelID = v;
        }
        static get CurrentLevel() {
            return ConfigHelper.LevelMap[this.CurrentLevelID];
        }
        static get CurrentLevelScene() {
            if (this.recoderPlayer.currentLevelScene == null) {
                this.recoderPlayer.currentLevelScene = ConfigHelper.LevelMap[this.CurrentLevelID].scenes[0];
                GameSDK.stageOnStart({ level: this.recoderPlayer.currentLevelScene });
                ReportMgr.inst.sendReport(ReportType.ENTER_GAME);
                WxMgr.inst.battleStart = false;
            }
            return this.recoderPlayer.currentLevelScene;
        }
        static set CurrentLevelScene(v) {
            this.recoderPlayer.currentLevelScene = v;
        }
        static IsLastLevelScene() {
            var level = this.CurrentLevel;
            var scene = this.CurrentLevelScene;
            var i = level.scenes.indexOf(scene);
            return (i == level.scenes.length - 1);
        }
        static GoToNextLevelScene() {
            var levelIndex = this.CurrentLevelBattleIndex + 1;
            if (levelIndex >= this.CurrentLevelBattleCount) {
                this.CurrentLevelID++;
                this.CurrentLevelScene = this.CurrentLevel.scenes[0];
                GameSDK.stageOnStart({ level: this.recoderPlayer.currentLevelScene });
                BattleData.CleasrBigSceneData();
            }
            else {
                this.CurrentLevelScene = this.CurrentLevel.scenes[levelIndex];
                BattleData.ResetLittleBattleData();
            }
            ReportMgr.inst.sendReport(ReportType.ENTER_GAME);
            WxMgr.inst.battleStart = false;
        }
        static get CurrentLevelBattleCount() {
            return this.CurrentLevel.scenes.length;
        }
        static get CurrentLevelBattleIndex() {
            return this.CurrentLevel.scenes.indexOf(this.CurrentLevelScene);
        }
        static get StorageTag() {
            if (this._recorderPlayer.storageTag == null) {
                this._recorderPlayer.storageTag = 0;
            }
            return this._recorderPlayer.storageTag;
        }
        static set StorageTag(v) {
            this._recorderPlayer.storageTag = v;
        }
        static get UnlockStorage() {
            return this._recorderPlayer.unlockStorage || false;
        }
        static set UnlockStorage(v) {
            this._recorderPlayer.unlockStorage = v;
        }
        static get NewScene() {
            if (this._recorderPlayer.newScene == null) {
                return true;
            }
            return this._recorderPlayer.newScene;
        }
        static set NewScene(v) {
            this._recorderPlayer.newScene = v;
        }
        static get HorseAdCount() {
            if (this._recorderPlayer.horseAdCount == null) {
                return 0;
            }
            return this._recorderPlayer.horseAdCount;
        }
        static set HorseAdCount(v) {
            this._recorderPlayer.horseAdCount = v;
        }
        static get UnlockMecha() {
            if (this._recorderPlayer.unlockMecha == null) {
                this._recorderPlayer.unlockMecha = false;
            }
            return this._recorderPlayer.unlockMecha;
        }
        static set UnlockMecha(v) {
            this._recorderPlayer.unlockMecha = v;
        }
        static get FreeMecha() {
            if (this._recorderPlayer.freeMecha == null) {
                this._recorderPlayer.freeMecha = 1;
            }
            return this._recorderPlayer.freeMecha;
        }
        static set FreeMecha(v) {
            this._recorderPlayer.freeMecha = v;
        }
        static get UnlockHangMu() {
            if (this._recorderPlayer.unlockHangMu == null) {
                this._recorderPlayer.unlockHangMu = false;
            }
            return this._recorderPlayer.unlockHangMu;
        }
        static set UnlockHangMu(v) {
            this._recorderPlayer.unlockHangMu = v;
        }
        static get HangMuNengLiang() {
            if (this._recorderPlayer.HangMuNL == null) {
                this._recorderPlayer.HangMuNL = 0;
            }
            return this._recorderPlayer.HangMuNL;
        }
        static set HangMuNengLiang(v) {
            this._recorderPlayer.HangMuNL = v;
        }
        static get HangMuNengLiangToday() {
            if (this._recorderPlayer.HangMuNLs == null) {
                this._recorderPlayer.HangMuNLs = 0;
            }
            if (this._recorderPlayer.HangMuNLt != Utils.TodayMS) {
                this._recorderPlayer.HangMuNLt = Utils.TodayMS;
                this._recorderPlayer.HangMuNLs = 0;
            }
            return this._recorderPlayer.HangMuNLs;
        }
        static set HangMuNengLiangToday(v) {
            this._recorderPlayer.HangMuNLs = v;
        }
        static get HangMuNengLiangNotUse() {
            if (this._recorderPlayer.HangMuNLm == null) {
                this._recorderPlayer.HangMuNLm = 0;
            }
            return this._recorderPlayer.HangMuNLm;
        }
        static set HangMuNengLiangNotUse(v) {
            this._recorderPlayer.HangMuNLm = v;
        }
        static get mountsId() {
            if (!!!this._recorderPlayer.mountsId) {
                this.recoderPlayer.mountsId = 0;
            }
            return this._recorderPlayer.mountsId;
        }
        static set mountsId(val) {
            this._recorderPlayer.mountsId = val;
        }
    }
    class BattleData {
        static get recoderBattle() {
            if (this._recorderBattle == null) {
                this._recorderBattle = DataManager.inst.GetDataRecorder('battleInfo');
            }
            return this._recorderBattle;
        }
        static get PlayerNode() {
            if (this.recoderBattle.playerNode == null) {
                var b = new BuildingState();
                b.level = PlayerData.CurrentRankLevel;
                b.slotName = 'PlayerLocation';
                b.tag = 0;
                b.unlocked = true;
                this.recoderBattle.playerNode = b;
            }
            return this.recoderBattle.playerNode;
        }
        static get SoliderTents() {
            if (this.recoderBattle.soliderTents == null) {
                this.initSoliderTent();
            }
            return this.recoderBattle.soliderTents;
        }
        static get TankTent() {
            if (this.recoderBattle.tankTent == null) {
                var tents = ConfigHelper.CurrentLevelTankTents;
                var slot = Object.keys(tents)[0];
                var config = tents[slot][0];
                var state = new BuildingState();
                state.slotName = slot;
                state.id = config.id;
                state.tag = 0;
                state.type = BuilidngType.Tankhangat;
                this.recoderBattle.tankTent = state;
            }
            return this.recoderBattle.tankTent;
        }
        static initSoliderTent() {
            var tentsInSlot = ConfigHelper.CurrentLevelSoliderTents;
            this.recoderBattle.soliderTents = [];
            var slots = Object.keys(tentsInSlot);
            for (var i = 0; i < slots.length; i++) {
                var slot = slots[i];
                var config = tentsInSlot[slot][0];
                var state = new BuildingState();
                state.id = config.id;
                state.level = config.level;
                state.unlocked = (i == 0);
                state.slotName = slot;
                state.tag = 0;
                state.type = BuilidngType.SoliderTent;
                this.recoderBattle.soliderTents.push(state);
            }
        }
        static UpdateSoliderTent() {
            var tentsInSlot = ConfigHelper.CurrentLevelSoliderTents;
            var slots = Object.keys(tentsInSlot);
            for (var i = 0; i < slots.length; i++) {
                var slot = slots[i];
                var config = tentsInSlot[slot][0];
                var state = this.GetTent(slot);
                if (state == null) {
                    state = new BuildingState();
                    this.recoderBattle.soliderTents.push(state);
                }
                else {
                    config = ConfigHelper.BuildingMap[state.id];
                }
                state.id = config.id;
                state.level = config.level;
                if (i == 0) {
                    state.unlocked = true;
                }
                else {
                    state.unlocked = config.show_unlock <= PlayerData.CurrentRankLevel;
                }
                state.slotName = slot;
                state.tag = 0;
                state.type = BuilidngType.SoliderTent;
            }
        }
        static GetTent(slotName) {
            for (var i = 0; i < this.SoliderTents.length; i++) {
                var tent = this.SoliderTents[i];
                if (tent.slotName == slotName) {
                    return tent;
                }
            }
        }
        static get SilverTag() {
            if (this.recoderBattle.mountSilverTag == null) {
                this.recoderBattle.mountSilverTag = 0;
            }
            return this.recoderBattle.mountSilverTag;
        }
        static set SilverTag(v) {
            this.recoderBattle.mountSilverTag = v;
            MsgManager.getInst().event(BattleEvent.UserTagChange);
        }
        static get GoldTag() {
            if (this.recoderBattle.mountGoldTag == null) {
                this.recoderBattle.mountGoldTag = 0;
            }
            return this.recoderBattle.mountGoldTag;
        }
        static set GoldTag(v) {
            this.recoderBattle.mountGoldTag = v;
            MsgManager.getInst().event(BattleEvent.UserTagChange);
        }
        static InitSilverTagInBegin(x, y, count) {
            ArrowTips.inst.Hide();
            if (PlayerData.NewScene) {
                ArrowTips.inst.Show();
                for (var i = 0; i < count; i++) {
                    var p = new PosState();
                    p.x = x;
                    p.y = i * 0.1;
                    p.z = y;
                    p.r = 0;
                    this.silverTagInGroud.push(p);
                }
            }
        }
        static InitGoldTagInBegin(x, y, count) {
            if (PlayerData.NewScene) {
                for (var i = 0; i < count; i++) {
                    var p = new PosState();
                    p.x = x;
                    p.y = i * 0.1;
                    p.z = y;
                    p.r = 0;
                    this.goldTagInGroud.push(p);
                }
            }
        }
        static get silverTagInGroud() {
            if (this.recoderBattle.silverTagInGroud == null) {
                this.recoderBattle.silverTagInGroud = [];
            }
            return this.recoderBattle.silverTagInGroud;
        }
        static set silverTagInGroud(v) {
            this.recoderBattle.silverTagInGroud = v;
        }
        static get goldTagInGroud() {
            if (this.recoderBattle.goldTagInGroud == null) {
                this.recoderBattle.goldTagInGroud = [];
            }
            return this.recoderBattle.goldTagInGroud;
        }
        static set goldTagInGroud(v) {
            this.recoderBattle.goldTagInGroud = v;
        }
        static AddTag(x, y, z, r, isEnemy) {
            var pos = new PosState();
            pos.x = x;
            pos.y = y;
            pos.z = z;
            pos.r = r;
            if (isEnemy) {
                this.goldTagInGroud.push(pos);
            }
            else {
                this.silverTagInGroud.push(pos);
            }
            return pos;
        }
        static DeleteTag(state, isEnemy = false) {
            if (isEnemy) {
                var i = this.goldTagInGroud.indexOf(state);
                this.goldTagInGroud.splice(i, 1);
            }
            else {
                var i = this.silverTagInGroud.indexOf(state);
                this.silverTagInGroud.splice(i, 1);
            }
        }
        static get CurrentPlayerWeapon() {
            if (this.recoderBattle.ammoFactory != null) {
                var bi = this.recoderBattle.ammoFactory.id;
                var wi = ConfigHelper.BuildingMap[bi].spwan_id;
                var w = ConfigHelper.WeaponConfigMap[wi];
                return w;
            }
            else {
                return ConfigHelper.InitWeaponConfig;
            }
        }
        static get WeaponFactory() {
            if (this.recoderBattle.ammoFactory == null) {
                var config = ConfigHelper.WeaponFactorys[0];
                var state = new BuildingState();
                state.id = config.id;
                state.tag = 0;
                state.slotName = 'AmmoFactory';
                state.type = config.type;
                state.level = config.level;
                state.unlocked = config.show_unlock <= PlayerData.CurrentRankLevel;
                this.recoderBattle.ammoFactory = state;
            }
            return this.recoderBattle.ammoFactory;
        }
        static get Airport() {
            if (this.recoderBattle.airport == null) {
                var config = ConfigHelper.Airport[0];
                var state = new BuildingState();
                state.id = config.id;
                state.tag = 0;
                state.slotName = 'AirportBuilding';
                state.type = config.type;
                state.level = config.level;
                state.unlocked = config.show_unlock <= PlayerData.CurrentRankLevel;
                this.recoderBattle.airport = state;
            }
            return this.recoderBattle.airport;
        }
        static get WeaponCountInFactory() {
            if (this.recoderBattle.ammoWeaponCount == null) {
                this.recoderBattle.ammoWeaponCount = 0;
            }
            return this.recoderBattle.ammoWeaponCount;
        }
        static set WeaponCountInFactory(v) {
            this.recoderBattle.ammoWeaponCount;
            this.recoderBattle.ammoWeaponCount = Math.max(0, v);
        }
        static get WeaponTagToCreateWeapon() {
            if (this.recoderBattle.ammonTagToCreateWeapon == null) {
                this.recoderBattle.ammonTagToCreateWeapon = 0;
            }
            return this.recoderBattle.ammonTagToCreateWeapon;
        }
        static set WeaponTagToCreateWeapon(v) {
            this.recoderBattle.ammonTagToCreateWeapon = v;
        }
        static get EnemyIndex() {
            if (this.recoderBattle.enemyDeadIndex == null) {
                this.recoderBattle.enemyDeadIndex = [];
            }
            return this.recoderBattle.enemyDeadIndex;
        }
        static ResetLittleBattleData() {
            this.goldTagInGroud = [];
            this.silverTagInGroud = [];
            this.recoderBattle.enemyDeadIndex = [];
        }
        static CleasrBigSceneData() {
            DataManager.inst.Clear('battleInfo');
            PlayerData.NewScene = true;
            this._recorderBattle = DataManager.inst.GetDataRecorder('battleInfo');
        }
        static get StorageBox() {
            if (this._recorderBattle.storageBox == null) {
                var state = new BuildingState();
                var config = ConfigHelper.CurrentLevelBuildingMap[BuilidngType.StorageBox]['StorageBox'][0];
                state.id = config.id;
                state.level = config.level;
                state.slotName = 'StorageBox';
                state.tag = 0;
                state.type = BuilidngType.StorageBox;
                this._recorderBattle.storageBox = state;
            }
            return this._recorderBattle.storageBox;
        }
        static get Mortars() {
            if (this._recorderBattle.mortars == null) {
                this._recorderBattle.mortars = [];
                var mortarsInSlot = ConfigHelper.CurrentLevelBuildingMap[BuilidngType.MortarBuilding];
                var slots = Object.keys(mortarsInSlot);
                slots.forEach(slot => {
                    var config = mortarsInSlot[slot][0];
                    var state = new BuildingState();
                    state.id = config.id;
                    state.level = config.level;
                    state.slotName = slot;
                    state.tag = 0;
                    state.type = BuilidngType.MortarBuilding;
                    this.recoderBattle.mortars.push(state);
                });
            }
            return this._recorderBattle.mortars;
        }
        static get TagAllCollect() {
            return this._recorderBattle.tagAllCollect || 0;
        }
        static set TagAllCollect(v) {
            this._recorderBattle.tagAllCollect = v;
        }
        static get EnemyTagAllCollect() {
            return this._recorderBattle.enemyTagAllCollect || 0;
        }
        static set EnemyTagAllCollect(v) {
            this._recorderBattle.enemyTagAllCollect = v;
        }
    }

    class ReportMgr {
        constructor() {
            this._reportNoSave = [
                ReportType.LEVEL_UP,
                ReportType.ENTER_GAME,
                ReportType.GAME_END,
                ReportType.COLLECT_TAG,
                ReportType.WIN_REBORN,
                ReportType.WIN_REBORN_VIDEO,
                ReportType.WIN_TAG,
                ReportType.WIN_TAG_VIDEO,
                ReportType.WIN_AIRTAG,
                ReportType.WIN_AIRTAG_VIDEO,
                ReportType.BUILD_ENEMY1,
                ReportType.BUILD_ENEMY2,
                ReportType.BUILD_ENEMY3,
                ReportType.BUILD_ENEMY4,
                ReportType.BUILD_TANK,
                ReportType.WIN_SHARE,
                ReportType.WIN_SHARE_CLOSE,
                ReportType.WIN_SHARE_OK,
                ReportType.BUILD_WEAPON,
                ReportType.CLICK_ROBOT_BTN,
                ReportType.CLICK_ROBOT,
                ReportType.SHARE_ROBOT,
                ReportType.SHARE_OPENDATA_ROBOT,
            ];
            this._reportNoSuf = [ReportType.LOADING_START, ReportType.LOADING_END];
            this._sp = {};
        }
        static get inst() {
            if (!!!this._inst) {
                this._inst = new ReportMgr();
            }
            return this._inst;
        }
        sendReport(type, addSuf = true, args) {
            let copy = type;
            try {
                if (type == ReportType.GUIDE_COLLECT_TAG) {
                    if (!!this._sp[type]) {
                        return;
                    }
                    this._sp[type] = 1;
                }
                if (addSuf && (type == ReportType.GUIDE_LEVEL_UP || type.indexOf(ReportType.LEVEL_UP) < 0)) {
                    if (this._reportNoSave.indexOf(type) < 0) {
                        if (this._hasLocal(type)) {
                            return;
                        }
                        else {
                            this._addLocal(type);
                        }
                    }
                }
                if (GameSDK$1.isFirstPlay) {
                    type = "新用户-" + type;
                }
                else {
                    type = "老用户-" + type;
                }
                type = this._getVersionStr(GameSDK$1.sdk.version) + "-" + type + ((this._reportNoSuf.indexOf(copy) >= 0 || !addSuf) ? "" : ("-关卡" + PlayerData.CurrentLevelScene));
                GameSDK$1.sdk.sendEvent(type + (!!args ? JSON.stringify(args) : ""));
                console.log("【打点上报】" + type);
            }
            catch (e) {
                GameSDK$1.sdk.sendEvent("上报事件失败-" + copy + (!!e ? ("-" + JSON.stringify(e)) : ""));
            }
        }
        get _local() {
            return LocalCache$1.get("LogickData._reportGuide", "");
        }
        _hasLocal(report) {
            return this._local.indexOf(report) >= 0;
        }
        _addLocal(report) {
            let str = this._local + report + ",";
            LocalCache$1.set("LogickData._reportGuide", str);
        }
        _getVersionStr(version) {
            if (!!!version) {
                version = "";
            }
            return version.replace(/\./g, "");
        }
    }

    class _AllPrefabsNames {
    }
    _AllPrefabsNames.TestEff = 'TestEff';
    _AllPrefabsNames.zhg_01 = 'zhg_01';
    _AllPrefabsNames.zhg_02 = 'zhg_02';
    _AllPrefabsNames.zhg_03 = 'zhg_03';
    _AllPrefabsNames.zhg_04 = 'zhg_04';
    _AllPrefabsNames.zhg_05 = 'zhg_05';
    _AllPrefabsNames.yan = 'yan';
    _AllPrefabsNames.yan02 = 'yan02';
    _AllPrefabsNames.GroundLevel1 = 'GroundLevel1';
    _AllPrefabsNames.GroundLevel2 = 'GroundLevel2';
    _AllPrefabsNames.GroundLevel3 = 'GroundLevel3';
    _AllPrefabsNames.GroundLevel4 = 'GroundLevel4';
    _AllPrefabsNames.GroundLevel5 = 'GroundLevel5';
    _AllPrefabsNames.GroundLevel6 = 'GroundLevel6';
    _AllPrefabsNames.GroundLevel7 = 'GroundLevel7';
    _AllPrefabsNames.GroundLevel8 = 'GroundLevel8';
    _AllPrefabsNames.FootSoliderTentLevel0 = 'FootSoliderTentLevel0';
    _AllPrefabsNames.FootSoliderTentLevel1 = 'FootSoliderTentLevel1';
    _AllPrefabsNames.FootSoliderTentLevel2 = 'FootSoliderTentLevel2';
    _AllPrefabsNames.FootSoliderTentLevel3 = 'FootSoliderTentLevel3';
    _AllPrefabsNames.TankHangarLevel0 = 'TankHangarLevel0';
    _AllPrefabsNames.TankHangarLevel1 = 'TankHangarLevel1';
    _AllPrefabsNames.TankHangarLevel2 = 'TankHangarLevel2';
    _AllPrefabsNames.TankHangarLevel3 = 'TankHangarLevel3';
    _AllPrefabsNames.AmmoFactory0 = 'AmmoFactory0';
    _AllPrefabsNames.AmmoFactory1 = 'AmmoFactory1';
    _AllPrefabsNames.AmmoFactory2 = 'AmmoFactory2';
    _AllPrefabsNames.AmmoFactory3 = 'AmmoFactory3';
    _AllPrefabsNames.AmmoFactory4 = 'AmmoFactory4';
    _AllPrefabsNames.playerSet = 'playerSet';
    _AllPrefabsNames.DogTag = 'DogTag';
    _AllPrefabsNames.DogTagEnemy = 'DogTagEnemy';
    _AllPrefabsNames.PlayerLoacation = 'PlayerLoacation';
    _AllPrefabsNames.RankIcons = 'RankIcons';
    _AllPrefabsNames.kaiqiang = 'kaiqiang';
    _AllPrefabsNames.penhuo = 'penhuo';
    _AllPrefabsNames.Mine = 'Mine';
    _AllPrefabsNames.NukeBuilding = 'NukeBuilding';
    _AllPrefabsNames.Nuke05 = 'Nuke05';
    _AllPrefabsNames.chuan1 = 'chuan1';
    _AllPrefabsNames.baozha03 = 'baozha03';
    _AllPrefabsNames.MortarBuilding0 = 'MortarBuilding0';
    _AllPrefabsNames.MortarBuilding1 = 'MortarBuilding1';
    _AllPrefabsNames.ExchageBuilding = 'ExchageBuilding';
    _AllPrefabsNames.StorageBox0 = 'StorageBox0';
    _AllPrefabsNames.StorageBox1 = 'StorageBox1';
    _AllPrefabsNames.BombBullet = 'BombBullet';
    _AllPrefabsNames.AirportBuilding0 = 'AirportBuilding0';
    _AllPrefabsNames.AirportBuilding1 = 'AirportBuilding1';
    _AllPrefabsNames.AirportBuilding2 = 'AirportBuilding2';
    _AllPrefabsNames.AirportBuilding3 = 'AirportBuilding3';
    _AllPrefabsNames.WarPlane = 'WarPlane';
    _AllPrefabsNames.DirectionalLight2 = 'DirectionalLight2';
    _AllPrefabsNames.AimTarget = 'AimTarget';
    _AllPrefabsNames.Helicopter_Friendly = 'Helicopter_Friendly';
    _AllPrefabsNames.PlaneBomb = 'PlaneBomb';
    _AllPrefabsNames.jluoshan = 'jluoshan';
    _AllPrefabsNames.FX_lihua = 'FX_lihua';
    _AllPrefabsNames.ma_skin = 'ma_skin';
    _AllPrefabsNames.Minigun_E = 'Minigun_E';
    _AllPrefabsNames.HangMuView = 'HangMuView';
    _AllPrefabsNames.Jijia = 'Jijia';
    _AllPrefabsNames.DropPod = 'DropPod';
    _AllPrefabsNames.Cube = 'Cube';
    _AllPrefabsNames.TestRole = 'TestRole';
    _AllPrefabsNames.TestRoleBig = 'TestRoleBig';
    _AllPrefabsNames.WeaponFlamethrower = 'WeaponFlamethrower';
    _AllPrefabsNames.WeaponBazooka = 'WeaponBazooka';
    _AllPrefabsNames.WeaponAk = 'WeaponAk';
    _AllPrefabsNames.WeaponMauser = 'WeaponMauser';
    _AllPrefabsNames.WeaponShotgun = 'WeaponShotgun';
    _AllPrefabsNames.wuqihe1 = 'wuqihe1';
    _AllPrefabsNames.wuqihe2 = 'wuqihe2';
    _AllPrefabsNames.wuqihe3 = 'wuqihe3';
    _AllPrefabsNames.wuqihe4 = 'wuqihe4';
    _AllPrefabsNames.wuqihe5 = 'wuqihe5';
    _AllPrefabsNames.CameraNode = 'CameraNode';
    _AllPrefabsNames.DirectionalLight = 'DirectionalLight';
    _AllPrefabsNames.Bullet = 'Bullet';
    _AllPrefabsNames.Water = 'Water';
    _AllPrefabsNames.SoliderStackingArea = 'SoliderStackingArea';
    _AllPrefabsNames.EnemyFlag = 'EnemyFlag';
    _AllPrefabsNames.Arrow = 'Arrow';
    _AllPrefabsNames.db_01 = 'db_01';
    _AllPrefabsNames.db_02 = 'db_02';
    _AllPrefabsNames.db_03 = 'db_03';
    _AllPrefabsNames.db_01_blue = 'db_01_blue';
    _AllPrefabsNames.db_02_blue = 'db_02_blue';
    _AllPrefabsNames.db_03_blue = 'db_03_blue';
    _AllPrefabsNames.commandeMat = 'commandeMat';
    _AllPrefabsNames.Tank_Enemy = 'Tank_Enemy';
    _AllPrefabsNames.Tank_Friendly = 'Tank_Friendly';
    _AllPrefabsNames.Wall = 'Wall';
    _AllPrefabsNames.SoldierOnTower = 'SoldierOnTower';
    _AllPrefabsNames.Gate = 'Gate';
    _AllPrefabsNames.WoodenBarrier = 'WoodenBarrier';
    _AllPrefabsNames.MagicBuffBlue = 'MagicBuffBlue';
    _AllPrefabsNames.MagicBuffYellow = 'MagicBuffYellow';
    _AllPrefabsNames.CameraAim = 'CameraAim';

    let AdItemsCDRecorder = class AdItemsCDRecorder extends DataProxy {
    };
    AdItemsCDRecorder = __decorate([
        data_bind('ad_cd')
    ], AdItemsCDRecorder);
    class ADData {
        static get recoder() {
            if (this._recorder == null) {
                this._recorder = DataManager.inst.GetDataRecorder('ad_cd');
                Laya.timer.frameLoop(1, this, this.frameLoop);
            }
            return this._recorder;
        }
        static frameLoop() {
            var dt = LayaUtils.deltaTime;
            this.nukeCD -= dt;
            this.GoldDropBox -= dt;
            this.Ship -= dt;
            this.Bomber -= dt;
            this.Helicopter -= dt;
            this.HorseCD -= dt;
            this.MechaCD -= dt;
            this.HangMuCD -= dt;
        }
        static get nukeCD() {
            if (this.recoder.hd == null) {
                this.recoder.hd = this.initCD;
            }
            return this.recoder.hd;
        }
        static set nukeCD(v) {
            this.recoder.hd = Math.max(0, v);
        }
        static get GoldDropBox() {
            if (this.recoder.jkt == null) {
                this.recoder.jkt = this.initCD;
            }
            return this.recoder.jkt;
        }
        static set GoldDropBox(v) {
            this.recoder.jkt = Math.max(0, v);
        }
        static get Ship() {
            if (this.recoder.xyj == null) {
                this.recoder.xyj = this.initCD;
            }
            return this.recoder.xyj;
        }
        static set Ship(v) {
            this.recoder.xyj = Math.max(0, v);
        }
        static get Bomber() {
            if (this.recoder.hzj == null) {
                this.recoder.hzj = this.initCD;
            }
            return this.recoder.hzj;
        }
        static set Bomber(v) {
            this.recoder.hzj = Math.max(0, v);
        }
        static get Helicopter() {
            if (this.recoder.zsj == null) {
                this.recoder.zsj = this.initCD;
            }
            return this.recoder.zsj;
        }
        static set Helicopter(v) {
            this.recoder.zsj = Math.max(0, v);
        }
        static get HorseCD() {
            if (this.recoder.horseCD == null) {
                this.recoder.horseCD = 0;
            }
            return this.recoder.horseCD;
        }
        static set HorseCD(v) {
            this.recoder.horseCD = v;
        }
        static get MechaCD() {
            if (this.recoder.mechaCD == null) {
                this.recoder.mechaCD = 0;
            }
            return this.recoder.mechaCD;
        }
        static set MechaCD(v) {
            this.recoder.mechaCD = v;
        }
        static get HangMuCD() {
            if (this.recoder.hm == null) {
                this.recoder.hm = 0;
            }
            return this.recoder.hm;
        }
        static set HangMuCD(v) {
            this.recoder.hm = v;
        }
    }
    ADData.initCD = 300000;

    var _EAllScenePrefabsNames;
    (function (_EAllScenePrefabsNames) {
        _EAllScenePrefabsNames["prefab2"] = "@TestEff@@zhg_01@@zhg_02@@zhg_03@@zhg_04@@zhg_05@@yan@@yan02@@GroundLevel1@@GroundLevel2@@GroundLevel3@@GroundLevel4@@GroundLevel5@@GroundLevel6@@GroundLevel7@@GroundLevel8@@FootSoliderTentLevel0@@FootSoliderTentLevel1@@FootSoliderTentLevel2@@FootSoliderTentLevel3@@TankHangarLevel0@@TankHangarLevel1@@TankHangarLevel2@@TankHangarLevel3@@AmmoFactory0@@AmmoFactory1@@AmmoFactory2@@AmmoFactory3@@AmmoFactory4@@playerSet@@DogTag@@DogTagEnemy@@PlayerLoacation@@RankIcons@@kaiqiang@@penhuo@@Mine@@NukeBuilding@@Nuke05@@chuan1@@baozha03@@MortarBuilding0@@MortarBuilding1@@ExchageBuilding@@StorageBox0@@StorageBox1@@BombBullet@@AirportBuilding0@@AirportBuilding1@@AirportBuilding2@@AirportBuilding3@@WarPlane@@DirectionalLight2@@AimTarget@@Helicopter_Friendly@@PlaneBomb@@jluoshan@@FX_lihua@@ma_skin@@Minigun_E@@HangMuView@@Jijia@@DropPod@";
        _EAllScenePrefabsNames["prefab"] = "@Cube@@TestRole@@TestRoleBig@@WeaponFlamethrower@@WeaponBazooka@@WeaponAk@@WeaponMauser@@WeaponShotgun@@wuqihe1@@wuqihe2@@wuqihe3@@wuqihe4@@wuqihe5@@CameraNode@@DirectionalLight@@Bullet@@Water@@SoliderStackingArea@@EnemyFlag@@Arrow@@db_01@@db_02@@db_03@@db_01_blue@@db_02_blue@@db_03_blue@@commandeMat@@Tank_Enemy@@Tank_Friendly@@Wall@@SoldierOnTower@@Gate@@WoodenBarrier@@MagicBuffBlue@@MagicBuffYellow@@CameraAim@";
    })(_EAllScenePrefabsNames || (_EAllScenePrefabsNames = {}));

    class PrefabManager extends Singleton {
        constructor() {
            super();
            this.prefabMap = {};
            this.PrepareData();
        }
        PrepareData() {
            for (var sceneName in _EAllScenePrefabsNames) {
                var scene = _EAllScenePrefabsNames[sceneName];
                if (scene == '' || scene == null) {
                    continue;
                }
                var prefabs = scene.split('@');
                prefabs.forEach(element => {
                    if (element != '') {
                        var item = this.prefabMap[element] = new PrefabItem();
                        item.isCache = false;
                        item.name = element;
                        item.pkg = sceneName;
                    }
                });
            }
        }
        LoadPrefab(map, completed, progress, caller) {
            var urls = [];
            for (var key in map) {
                if (map[key] == null) {
                    console.error(key, ' 没有导出到prefab表');
                }
                urls.push(map[key].GetUrl());
            }
            ResLoadManager.LoadPrefabRes(urls, completed, progress, caller);
        }
        LoadPrefabByName(name) {
            return new ProgressPromise((r, p, j) => {
                var item = this.prefabMap[name];
                if (item != null) {
                    var prefab = this.GetPrefab(name);
                    if (prefab != null) {
                        p(1);
                        r(prefab);
                        return;
                    }
                    ResLoadManager.LoadPrefabRes([item.GetUrl()], b => {
                        if (b) {
                            var ins = Laya.loader.getRes(item.GetUrl());
                            if (ins) {
                                r(ins);
                            }
                            else {
                                j();
                            }
                        }
                        else {
                            j();
                        }
                    }, p, this);
                }
                else {
                    j();
                }
            });
        }
        GetPrefab(name) {
            if (this.prefabMap[name] != null) {
                return Laya.loader.getRes(this.prefabMap[name].GetUrl());
            }
            else {
                console.error('not find ', name);
            }
        }
        UnLoadPrefab(list) {
            var urls = [];
            list.forEach(key => {
                urls.push(key.GetUrl());
            });
            ResLoadManager.ClearRes(urls);
        }
    }
    class PrefabItem {
        GetUrl() {
            return `unity/LayaScene_${this.pkg}/Conventional/${this.name}.lh`;
        }
    }

    class VertexUVMaterial extends Laya.Material {
        constructor() {
            super();
            this.setShaderName("VertexUV");
            this.renderQueue = Laya.Material.RENDERQUEUE_TRANSPARENT - 1;
            this.alphaTest = true;
            this.depthWrite = true;
            this.cull = Laya.RenderState.CULL_BACK;
            this.blend = Laya.RenderState.BLEND_ENABLE_ALL;
            this.blendSrc = Laya.RenderState.BLENDPARAM_SRC_ALPHA;
            this.blendDst = Laya.RenderState.BLENDPARAM_ONE_MINUS_SRC_ALPHA;
            this.depthTest = Laya.RenderState.DEPTHTEST_LESS;
        }
        set mainColor(value) {
            this._shaderValues.setVector(VertexUVMaterial.MainColor, value);
        }
        get mainColor() {
            return this._shaderValues.getVector(VertexUVMaterial.MainColor);
        }
        set progress(v) {
            this._shaderValues.setNumber(VertexUVMaterial.Progress, v);
        }
        set radius(v) {
            this._shaderValues.setNumber(VertexUVMaterial.Radius, v);
        }
    }
    VertexUVMaterial.MainColor = Laya.Shader3D.propertyNameToID("u_MainColor");
    VertexUVMaterial.Progress = Laya.Shader3D.propertyNameToID("u_progress");
    VertexUVMaterial.Radius = Laya.Shader3D.propertyNameToID("u_radius");

    class DelayDestory extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.delay = 1000;
        }
        onUpdate() {
            this.delay -= LayaUtils.deltaTime;
            if (this.delay <= 0) {
                this.sp.destroy(true);
            }
        }
    }

    class RotateYYoYo extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.oldEulerY = 0;
            this.d = 0;
        }
        onStart() {
            this.oldEulerY = this.transform.localRotationEulerY;
        }
        onUpdate() {
            this.d += LayaUtils.deltaTimeSec;
            this.transform.localRotationEulerY = this.oldEulerY + Math.sin(this.d * 2) * 0.8 * 20;
        }
    }

    class SceneShow3D {
        constructor() {
            this._RenderTexture_sprite = null;
            this.initScene();
            this.initCamera();
            this.initLight();
        }
        initScene() {
            this.scene = new Laya.Scene3D();
            Laya.stage.addChildAt(this.scene, 0);
        }
        initCamera() {
            this.camera = new Laya.Camera(0, 0.01, 200);
            this.scene.addChild(this.camera);
            this.camera.transform.position = new Laya.Vector3(18, 26, -20.89);
            this.camera.transform.lookAt(new Laya.Vector3(-3.7, 0, 4), LayaUtils.UP);
            this.camera.orthographic = true;
            this.scene.ambientColor = new Laya.Vector3(1, 1, 1);
            this.camera.orthographicVerticalSize = 28;
        }
        initLight() {
            var directionLight = new Laya.DirectionLight();
            this.scene.addChild(directionLight);
            directionLight.color = new Laya.Vector3(1, 1, 1);
            directionLight.transform.position = new Laya.Vector3(0, 3, 0);
            directionLight.transform.rotation = new Laya.Quaternion(0.1093816, 0.8754261, 0.4082179, -0.2345697);
            directionLight.intensity = 1;
            {
                var directionLight1 = new Laya.DirectionLight();
                directionLight.addChild(directionLight1);
                directionLight1.color = new Laya.Vector3(0.7688679, 0.8638439, 1);
                directionLight1.transform.position = new Laya.Vector3(0, 0, 0);
                directionLight1.transform.rotation = new Laya.Quaternion(-0.05103943, 0.693257, -0.4439691, 0.5654036);
                directionLight1.intensity = 0.3;
            }
        }
        clearRenderTexture() {
            if (this._RenderTexture_sprite)
                this._RenderTexture_sprite.graphics.clear();
            this._RenderTexture_sprite = null;
        }
        setRenderTexture(sprite) {
            this._RenderTexture_sprite = sprite;
            this.camera.clearColor.setValue(0, 0, 0, 0);
            this.camera.clearFlag = Laya.BaseCamera.CLEARFLAG_SKY;
            this.camera.renderingOrder = -1;
            this.camera.renderTarget = new Laya.RenderTexture(sprite.width, sprite.height, Laya.BaseTexture.FORMAT_R8G8B8A8);
            sprite.graphics.drawImage(new Laya.Texture(this.camera.renderTarget, Laya.Texture.DEF_UV));
        }
        addChild(node, zIndex = null) {
            if (zIndex == null)
                zIndex = this.scene.numChildren;
            this.scene.addChildAt(node, zIndex);
        }
        get playerNode() {
            if (!this._playerNode) {
                this._playerNode = new Laya.Sprite3D;
                if (this.scene)
                    this.scene.addChild(this._playerNode);
                this._playerNode.transform.localPosition = new Laya.Vector3(0, 0, 0);
                this._playerNode.transform.rotationEuler = v3_zeor;
            }
            return this._playerNode;
        }
        destroy() {
            this.camera.renderTarget.destroy();
            this.clearRenderTexture();
            if (this.scene) {
                this.scene.destroy();
                this.scene = null;
            }
            this.camera = null;
        }
    }

    class UI_AttackUp extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "AttackUp"));
        }
        onConstruct() {
            this.m_img = (this.getChildAt(0));
        }
    }
    UI_AttackUp.URL = "ui://ixnra5xgomf0xtr";

    class UI_FreeNuke extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "FreeNuke"));
        }
        onConstruct() {
            this.m_img = (this.getChildAt(0));
        }
    }
    UI_FreeNuke.URL = "ui://ixnra5xgl6znxtk";

    class UI_ui_main extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "ui_main"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m_guide_step = this.getControllerAt(1);
            this.m_bg = (this.getChildAt(0));
            this.m_mechaTimer = (this.getChildAt(1));
            this.m_hp = (this.getChildAt(2));
            this.m_hpHorse = (this.getChildAt(3));
            this.m_joyStick = (this.getChildAt(4));
            this.m_comp_zhiyin = (this.getChildAt(5));
            this.m_btn_fight = (this.getChildAt(6));
            this.m_comp_zidan = (this.getChildAt(7));
            this.m_guide = (this.getChildAt(8));
            this.m_bottomVideo = (this.getChildAt(9));
            this.m_btn_setting = (this.getChildAt(10));
            this.m_btn_hongzha = (this.getChildAt(11));
            this.m_btn_hedan = (this.getChildAt(12));
            this.m_btn_chuan = (this.getChildAt(13));
            this.m_btn_kongtou = (this.getChildAt(14));
            this.m_btn_jijia = (this.getChildAt(15));
            this.m_btn_zhanjian = (this.getChildAt(16));
            this.m_ad = (this.getChildAt(17));
            this.m_left = (this.getChildAt(18));
            this.m_comp_chaopiao2 = (this.getChildAt(19));
            this.m_comp_chaopiao1 = (this.getChildAt(20));
            this.m_btn_feiji = (this.getChildAt(21));
            this.m_mingri = (this.getChildAt(22));
            this.m_toupiao = (this.getChildAt(23));
            this.m_btn_try = (this.getChildAt(24));
            this.m_btn_gdyx = (this.getChildAt(25));
            this.m_right = (this.getChildAt(27));
            this.m_comp_guanqiajindu = (this.getChildAt(28));
            this.m_top = (this.getChildAt(29));
            this.m_rankTips = (this.getChildAt(30));
            this.m_t0 = this.getTransitionAt(0);
        }
    }
    UI_ui_main.URL = "ui://ixnra5xgyyz1u";

    class AdMgr {
        constructor() {
            this._isInit = false;
            this._bottomChangeTime = 5 * 60 * 1000;
            this._showBanner = true;
        }
        static get inst() {
            if (!!!this._inst) {
                this._inst = new AdMgr();
            }
            return this._inst;
        }
        init() {
            if (this._isInit) {
                return;
            }
            this._isInit = true;
        }
        showBottomAd(show, hideOnShort = true) {
            return;
        }
        showBanner(show) {
            return;
        }
        showBottomCustom(show) {
            return;
        }
        showLeftAd(show) {
            return;
        }
        showRightAd(show) {
            return;
        }
        showInsertAd() {
            return;
        }
        showRectAd() {
            return;
        }
    }

    var CtrlEvent;
    (function (CtrlEvent) {
        CtrlEvent["PlayerSteerMove"] = "PlayerSteerMove";
        CtrlEvent["PlayerSteerEnd"] = "PlayerSteerEnd";
        CtrlEvent["StartDriver"] = "StartSrive";
        CtrlEvent["EndDriver"] = "EndDriver";
    })(CtrlEvent || (CtrlEvent = {}));

    var CampType;
    (function (CampType) {
        CampType["Red"] = "red";
        CampType["Blue"] = "blue";
    })(CampType || (CampType = {}));

    class ThrowDestroy extends Laya.Script3D {
        constructor() {
            super(...arguments);
            this.v = 10;
            this.vx = 0;
            this.vz = 0;
            this.g = -18;
            this.rx = 0;
            this.rz = 0;
            this.cd = 0;
        }
        onAwake() {
            this.rx = (Math.random() - 0.5) * 360;
            this.rz = (Math.random() - 0.5) * 360;
            this.vx = (Math.random() - 0.5) * 10;
            this.vz = (Math.random() - 0.5) * 10;
            this.transfrom = this.owner.transform;
        }
        onUpdate() {
            var dt = LayaUtils.deltaTimeSec;
            this.cd += dt;
            if (this.cd > 0.8) {
                this.owner.destroy(true);
                return;
            }
            this.v += this.g * dt;
            this.transfrom.localPositionY += this.v * dt;
            this.transfrom.localPositionX += this.vx * dt;
            this.transfrom.localPositionZ += this.vz * dt;
            this.transfrom.localRotationEulerX += this.rx * dt;
            this.transfrom.localRotationEulerZ += this.rz * dt;
        }
    }

    var tagH = 0.1;
    class BattleEntityComponent extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.mr = [];
        }
        onAwake() {
            this.rotateNode = this.Find("RotateNode") || this.sp;
        }
        onStart() {
            if (this.ani) {
                this.ani.speed = 1;
            }
            this.ChangeWeapon();
            Utils.GetAllSKinMeshRender(this.sp, this.mr);
            Utils.GetAllMeshRenderer(this.sp, this.mr);
            this.mr.forEach(r => {
                r['_old_i'] = r.material.albedoIntensity;
            });
            if (this.entity.isPlayer) {
                MsgManager.getInst().on(BattleEvent.AddTagToPlayer, this, this.onAddTag);
                MsgManager.getInst().on(BattleEvent.DeleteTagFromPlayer, this, this.onRemoveTag);
                this.tagNode = this.Find('DogTagPoint');
                var t = this.tagNodeEnemy = new Laya.Sprite3D('DogTagEnemy');
                this.tagNode.parent.addChild(t);
                t.transform.localPosition = this.tagNode.transform.localPosition;
                t.transform.localScale = this.tagNode.transform.localScale;
                t.transform.localRotation = this.tagNode.transform.localRotation;
                this.initTag();
            }
            var polygon = this.Find('polygon');
            if (polygon) {
                var ps = [];
                for (var i = 0; i < polygon.numChildren; i++) {
                    var c = polygon.getChildAt(i);
                    ps.push(new Laya.Vector2(c.transform.position.x, c.transform.position.z));
                }
                this.polygon = new PolygonWrap(ps);
                MapComponent.inst.collisionPolygon.push(this.polygon);
            }
        }
        ChangeWeapon() {
            if (this.entity.config.spwan_weapon == 0)
                return;
            if (this.entity.weaponConfig.prefab != null) {
                this.weaponPos = this.Find(GlobalConstPoxy.Inst.data.weaponNode) || this.sp;
                var weaponPrefab = this.entity.weaponConfig.prefab;
                if (this.entity.isPlayer && BattleData.WeaponFactory != null && BattleData.WeaponFactory.level > 0) {
                    var factory = ConfigHelper.BuildingMap[BattleData.WeaponFactory.id];
                    weaponPrefab = ConfigHelper.WeaponConfigMap[factory.spwan_id].prefab;
                }
                var weapon = this.weapon = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(weaponPrefab));
                this.weaponPos.addChild(weapon);
                weapon.transform.localPosition = new Laya.Vector3();
            }
            if (this.entity.weaponConfig.fx_shoot != null) {
                PrefabManager.getInst().LoadPrefabByName(this.entity.weaponConfig.fx_shoot).then(fxp => {
                    this.shootFx = Laya.Sprite3D.instantiate(fxp);
                    var n = Utils.GetNode(this.sp, 'Muzzle') || this.weapon;
                    n.addChild(this.shootFx);
                    this.shootFx.transform.localPosition = LayaUtils.Vector3Zero;
                    this.shootFx.transform.localRotationEuler = LayaUtils.Vector3Zero;
                    this.shootFx.transform.localScale = new Laya.Vector3(0.5, 0.5, 0.5);
                    Utils.WaitFrame(2).then(() => {
                        this.shootFx.active = false;
                    });
                });
            }
            var emiter = this.Find('MuzzleRoot');
            if (emiter) {
                this.entity.emiter = emiter;
            }
        }
        initTag() {
            this.tagNodeEnemy.removeChildren();
            this.tagNode.removeChildren();
            if (BattleData.GoldTag > 0) {
                for (var i = 0; i < BattleData.GoldTag; i++) {
                    var tag = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(_AllPrefabsNames.DogTagEnemy));
                    tag.getChildAt(1).active = false;
                    this.tagNodeEnemy.addChild(tag);
                    tag.transform.localPosition = new Laya.Vector3(0, 0, BattleData.SilverTag * tagH + i * tagH);
                    tag.transform.localRotationEulerX = 90;
                }
            }
            if (BattleData.SilverTag > 0) {
                for (var i = 0; i < BattleData.SilverTag; i++) {
                    var tag = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(_AllPrefabsNames.DogTag));
                    tag.getChildAt(1).active = false;
                    this.tagNode.addChild(tag);
                    tag.transform.localPosition = new Laya.Vector3(0, 0, i * tagH);
                    tag.transform.localRotationEulerX = 90;
                }
            }
        }
        onDestroy() {
            super.onDestroy();
            this.hpBar = null;
            MsgManager.getInst().offAllCaller(this);
            if (this.polygon != null && MapComponent.inst != null && MapComponent.inst.sp != null && !MapComponent.inst.sp.destroyed) {
                var i = MapComponent.inst.collisionPolygon.indexOf(this.polygon);
                MapComponent.inst.collisionPolygon.splice(i, 1);
            }
            MsgManager.getInst().event(BattleEvent.HideAttackUPBar, this.entity);
            if (this.attackUpBar != null) {
                this.attackUpBar.dispose();
                this.attackUpBar = null;
            }
        }
        onAddTag(isEnemy, sp) {
            var oldP = sp.transform.position;
            var oldE = sp.transform.rotationEuler;
            SoundManager.inst.playSound('TagCollect2');
            ReportMgr.inst.sendReport(ReportType.GUIDE_COLLECT_TAG);
            if (GuideData.Step == 0 && BattleData.SilverTag >= 10) {
                GuideData.Step = GuideStepType.建立兵营;
            }
            if (GuideData.Step == GuideStepType.第二关收集) {
                GuideData.Step = GuideStepType.None;
            }
            sp.getChildAt(1).active = false;
            if (isEnemy) {
                this.tagNodeEnemy.addChild(sp);
                sp.transform.position = oldP;
                var z = this.tagNode.numChildren * tagH - tagH + this.tagNodeEnemy.numChildren * tagH;
                sp['tween'] = Laya.Tween.to(sp.transform, { localPositionX: 0, localPositionY: 0, localPositionZ: z }, 80);
            }
            else {
                this.tagNode.addChild(sp);
                for (var i = 0; i < this.tagNodeEnemy.numChildren; i++) {
                    var t = this.tagNodeEnemy.getChildAt(i);
                    if (t['tween'] != null) {
                        var r = t['tween'];
                        r.clear();
                        t['tween'] = null;
                    }
                    t.transform.localPositionX = 0;
                    t.transform.localPositionZ = this.tagNode.numChildren * tagH - tagH + (i + 1) * tagH;
                    t.transform.localPositionY = 0;
                }
                sp.transform.position = oldP;
                var z = this.tagNode.numChildren * tagH - tagH;
                Laya.Tween.to(sp.transform, { localPositionX: 0, localPositionY: 0, localPositionZ: z }, 80);
            }
            sp.transform.rotationEuler = oldE;
            Laya.Tween.to(sp.transform, { localRotationEulerX: 90, localRotationEulerY: 0, localRotationEulerZ: 0 }, 80);
        }
        onRemoveTag(isEnemyTag = false, flyToTarget) {
            SoundManager.inst.playSound('TagCollect1');
            var tail;
            if (isEnemyTag) {
                tail = this.tagNodeEnemy.getChildAt(this.tagNodeEnemy.numChildren - 1);
            }
            else {
                tail = this.tagNode.getChildAt(this.tagNode.numChildren - 1);
                for (var i = 0; i < this.tagNodeEnemy.numChildren; i++) {
                    var t = this.tagNodeEnemy.getChildAt(i);
                    t.transform.localPositionZ -= tagH;
                }
            }
            if (flyToTarget != null) {
                var oldP = tail.transform.position;
                var oldR = tail.transform.rotationEuler;
                flyToTarget.addChild(tail);
                tail.transform.position = oldP;
                tail.transform.rotationEuler = oldR;
                Laya.Tween.to(tail.transform, { localPositionX: 0, localPositionY: 0, localPositionZ: 0 }, 120);
                Laya.Tween.to(tail.transform, { localRotationEulerX: 0, localRotationEulerY: 0, localRotationEulerZ: 0 }, 120);
                tail.addComponent(DelayDestory).delay = 130;
            }
            else {
                tail.destroy(true);
                console.error('fly to tartget is null', flyToTarget);
            }
        }
        PlayAni(name, duration, forceReplay = false, cross = false) {
            if (this._ani) {
                var state = this.ani.getControllerLayer(0).getAnimatorState(name);
                if (state == null) {
                    return;
                }
                this.ani.getControllerLayer(0).getCurrentPlayState();
                if (!forceReplay) {
                    if (this.currentAni == name) {
                        return;
                    }
                }
                if (duration == -1) {
                    this.ani.speed = 1;
                }
                else {
                    this.ani.speed = state.clip.duration() / duration;
                }
                if (cross) {
                    this.ani.crossFade(name, 0.01);
                }
                else {
                    this.ani.play(name, 0, forceReplay ? 0 : null);
                }
                this.currentAni = name;
            }
        }
        BlinkShow() {
            try {
                var scale = this.entity.attackUP ? 1.5 : 1;
                this.mr.forEach(r => {
                    var mat = r.material;
                    if (mat) {
                        mat.albedoIntensity = r['_old_i'] * scale;
                    }
                });
            }
            catch (e) {
            }
        }
        BlinkFade() {
            if (this.entity.attackUP)
                return;
            try {
                this.mr.forEach(r => {
                    var mat = r.material;
                    if (mat) {
                        mat.albedoIntensity = r['_old_i'] * 1.5;
                    }
                });
                Laya.timer.once(100, this, this.Reset, null, true);
            }
            catch (e) {
            }
        }
        Reset() {
            this.mr.forEach(r => {
                var mat = r.material;
                if (mat) {
                    mat.albedoIntensity = r['_old_i'];
                }
            });
        }
    }

    class QuadraticCurve {
        constructor(p1, p2, p3) {
            this.p1 = p1;
            this.p2 = p2;
            this.p3 = p3;
        }
        GetPos(t, out = null) {
            let arg1 = (1 - t) * (1 - t);
            let arg2 = t * (1 - t) * 2;
            let arg3 = t * t;
            var x = arg1 * this.p1.x + this.p2.x * arg2 + this.p3.x * arg3;
            var y = arg1 * this.p1.y + this.p2.y * arg2 + this.p3.y * arg3;
            var z = arg1 * this.p1.z + this.p2.z * arg2 + this.p3.z * arg3;
            out = out || new Laya.Vector3();
            out.setValue(x, y, z);
            return out;
        }
    }

    class DogTagComponent extends BaseScript3D {
    }
    class DogTagComponentAni extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.RY = Math.random() * 360;
            this.randomScale = 2;
            this.t = 0;
            this.all = 400;
        }
        onStart() {
            this.newPos = this.transform.position.clone();
            var d = 0.5 - Math.random();
            d *= 2;
            this.newPos.x += d;
            d = 0.5 - Math.random();
            d *= this.randomScale;
            this.newPos.z += d;
            this.entity.state.x = this.newPos.x;
            this.entity.state.y = this.newPos.y;
            this.entity.state.z = this.newPos.z;
            this.entity.state.r = this.RY;
            var mid = new Laya.Vector3();
            Laya.Vector3.subtract(this.newPos, this.transform.position, mid);
            Laya.Vector3.scale(mid, 0.5, mid);
            Laya.Vector3.add(this.transform.position, mid, mid);
            mid.y += 4;
            this.curve = new QuadraticCurve(this.transform.position.clone(), mid, this.newPos);
        }
        onUpdate() {
            this.t += LayaUtils.deltaTime;
            this.t = Math.min(this.all, this.t);
            this.transform.localRotationEulerZ = 360 * (this.t / this.all);
            this.transform.localRotationEulerY = this.RY * (this.t / this.all);
            if (this.curve == null) {
                return;
            }
            this.curve.GetPos(this.t / this.all, this.transform.position);
            this.UpdatePosition();
            if (this.t == this.all) {
                this.destroy();
            }
        }
    }

    class LineRener {
        constructor() {
            this.up = new Laya.Vector3(0, 1, 0);
            this.pos = [];
            this.halfWidth = 0.15;
        }
        Update() {
            if (this.mesh) {
                this.UpdatePos();
                this.mesh.setPositions(this.pos);
            }
        }
        createMesh(linePos, width = 0.15) {
            var linePos = this.linePos = linePos;
            var halfWidth = this.halfWidth = width / 2;
            var vertexCount = linePos.length * 2;
            var indexCount = (linePos.length - 1) * 6;
            var vertexDeclaration = Laya.VertexMesh.getVertexDeclaration("POSITION,NORMAL");
            var vertexFloatStride = vertexDeclaration.vertexStride / 4;
            var vertices = new Float32Array(vertexCount * vertexFloatStride);
            var vc = 0;
            var indices = new Uint16Array(indexCount);
            var ic = 0;
            var p = new Laya.Vector3();
            var up = this.up;
            for (var i = 0; i < linePos.length; i++) {
                var cur = linePos[i];
                if (i == 0) {
                    let next = linePos[i + 1];
                    Laya.Vector3.subtract(next, cur, p);
                    Laya.Vector3.cross(up, p, p);
                    Laya.Vector3.normalize(p, p);
                    Laya.Vector3.scale(p, halfWidth, p);
                    Laya.Vector3.add(cur, p, p);
                    this.pos.push(p.clone());
                    Laya.Vector3.subtract(next, cur, p);
                    Laya.Vector3.cross(p, up, p);
                    Laya.Vector3.normalize(p, p);
                    Laya.Vector3.scale(p, halfWidth, p);
                    Laya.Vector3.add(cur, p, p);
                    this.pos.push(p.clone());
                }
                else {
                    let pre = linePos[i - 1];
                    Laya.Vector3.subtract(cur, pre, p);
                    Laya.Vector3.cross(up, p, p);
                    Laya.Vector3.normalize(p, p);
                    Laya.Vector3.scale(p, halfWidth, p);
                    Laya.Vector3.add(cur, p, p);
                    this.pos.push(p.clone());
                    Laya.Vector3.subtract(cur, pre, p);
                    Laya.Vector3.cross(p, up, p);
                    Laya.Vector3.normalize(p, p);
                    Laya.Vector3.scale(p, halfWidth, p);
                    Laya.Vector3.add(cur, p, p);
                    this.pos.push(p.clone());
                    indices[ic++] = i * 2 - 1;
                    indices[ic++] = i * 2 - 2;
                    indices[ic++] = i * 2;
                    indices[ic++] = i * 2;
                    indices[ic++] = i * 2 + 1;
                    indices[ic++] = i * 2 - 1;
                }
            }
            for (var i = 0; i < this.pos.length; i++) {
                var p = this.pos[i];
                vertices[vc++] = p.x;
                vertices[vc++] = p.y;
                vertices[vc++] = p.z;
                vertices[vc++] = 0;
                vertices[vc++] = 1;
                vertices[vc++] = 0;
            }
            this.mesh = Laya.PrimitiveMesh['_createMesh'](vertexDeclaration, vertices, indices);
        }
        UpdatePos() {
            var linePos = this.linePos;
            var halfWidth = this.halfWidth;
            var p = new Laya.Vector3();
            var up = this.up;
            for (var i = 0; i < linePos.length; i++) {
                var cur = linePos[i];
                if (i == 0) {
                    let next = linePos[i + 1];
                    Laya.Vector3.subtract(next, cur, p);
                    Laya.Vector3.cross(up, p, p);
                    Laya.Vector3.normalize(p, p);
                    Laya.Vector3.scale(p, halfWidth, p);
                    Laya.Vector3.add(cur, p, p);
                    p.cloneTo(this.pos[i * 2]);
                    Laya.Vector3.subtract(next, cur, p);
                    Laya.Vector3.cross(p, up, p);
                    Laya.Vector3.normalize(p, p);
                    Laya.Vector3.scale(p, halfWidth, p);
                    Laya.Vector3.add(cur, p, p);
                    p.cloneTo(this.pos[i * 2 + 1]);
                }
                else {
                    let pre = linePos[i - 1];
                    Laya.Vector3.subtract(cur, pre, p);
                    Laya.Vector3.cross(up, p, p);
                    Laya.Vector3.normalize(p, p);
                    Laya.Vector3.scale(p, halfWidth, p);
                    Laya.Vector3.add(cur, p, p);
                    p.cloneTo(this.pos[i * 2]);
                    Laya.Vector3.subtract(cur, pre, p);
                    Laya.Vector3.cross(p, up, p);
                    Laya.Vector3.normalize(p, p);
                    Laya.Vector3.scale(p, halfWidth, p);
                    Laya.Vector3.add(cur, p, p);
                    p.cloneTo(this.pos[i * 2 + 1]);
                }
            }
        }
    }

    class MathUtils {
        static PointInPolygon(p, polygon, segmentLine) {
            if (polygon.length < 3) {
                return true;
            }
            var insecCount = 0;
            for (var i = 0; i < polygon.length; i++) {
                var seg1 = polygon[i];
                var seg2 = polygon[(i == (polygon.length - 1) ? 0 : (i + 1))];
                var minY = Math.min(seg1.y, seg2.y);
                var maxY = Math.max(seg1.y, seg2.y);
                var maxX = Math.max(seg1.x, seg2.x);
                var minX = Math.min(seg1.x, seg2.x);
                if ((minX - 1) <= p.x && (maxX + 1) >= p.x && (minY - 1) <= p.y && (maxY + 1) >= p.y) {
                    segmentLine.push(seg1, seg2);
                }
                if (minY <= p.y && maxY >= p.y) {
                    if (maxX >= p.x) {
                        var p1x = p.x - seg1.x;
                        var p1y = p.y - seg1.y;
                        var p2x = 1000 - seg2.x;
                        var p2y = 0 - p.y;
                        var ppx = seg2.x - seg1.x;
                        var ppy = seg2.y - seg1.y;
                        var tmp1 = p1x * ppy - p1y * ppx;
                        var tmp2 = ppx * p2y - ppy * p2x;
                        if (tmp1 * tmp2 >= 0) {
                            insecCount++;
                        }
                    }
                }
            }
            return (insecCount % 2 == 1);
        }
        static PointInPolygon2(p, polygon) {
            if (polygon.length < 3) {
                return true;
            }
            var insecCount = 0;
            for (var i = 0; i < polygon.length; i++) {
                var seg1 = polygon[i];
                var seg2 = polygon[(i == (polygon.length - 1) ? 0 : (i + 1))];
                var minY = Math.min(seg1.y, seg2.y);
                var maxY = Math.max(seg1.y, seg2.y);
                var maxX = Math.max(seg1.x, seg2.x);
                if (minY <= p.y && maxY >= p.y) {
                    if (maxX >= p.x) {
                        var p1x = p.x - seg1.x;
                        var p1y = p.y - seg1.y;
                        var p2x = 1000 - seg2.x;
                        var p2y = 0 - p.y;
                        var ppx = seg2.x - seg1.x;
                        var ppy = seg2.y - seg1.y;
                        var tmp1 = p1x * ppy - p1y * ppx;
                        var tmp2 = ppx * p2y - ppy * p2x;
                        if (tmp1 * tmp2 >= 0) {
                            insecCount++;
                        }
                    }
                }
            }
            return (insecCount % 2 == 1);
        }
        static multiply(p0, p1, p2) {
            return ((p1.x - p0.x) * (p2.y - p0.y) - (p2.x - p0.x) * (p1.y - p0.y));
        }
        static ScalePath(ps, scale) {
            var center = new Laya.Vector2();
            ps.forEach(p => {
                LayaUtils.addVec2(p, center, center);
            });
            LayaUtils.scaleVec2(center, 1 / ps.length, center);
            var dis = new Laya.Vector2();
            ps.forEach(p => {
                LayaUtils.subtractVec2(p, center, dis);
                var len = LayaUtils.getLengthVec2(dis);
                LayaUtils.setLengthVec2(dis, len + scale, dis);
                LayaUtils.addVec2(dis, center, p);
            });
        }
        static Graham_scan(pointSet, ch, n) {
            var i, j, k = 0, top = 2;
            for (i = 1; i < n; i++) {
                if ((pointSet[i].y < pointSet[k].y) ||
                    ((pointSet[i].y == pointSet[k].y) && (pointSet[i].x < pointSet[k].x))) {
                    k = i;
                }
            }
            var tmp = pointSet[0];
            pointSet[0] = pointSet[k];
            pointSet[k] = tmp;
            var use = n;
            for (i = 1; i < use - 1; i++) {
                k = i;
                for (j = i + 1; j < use; j++) {
                    var direct = this.multiply(pointSet[0], pointSet[k], pointSet[j]);
                    if (direct > 0) {
                        k = j;
                    }
                    else if (direct == 0) {
                        var dis = LayaUtils.distanceSquaredVec2(pointSet[0], pointSet[j]) - LayaUtils.distanceSquaredVec2(pointSet[0], pointSet[k]);
                        use--;
                        if (dis > 0) {
                            pointSet[k] = pointSet[j];
                            pointSet[j] = pointSet[use];
                            j--;
                        }
                        else {
                            tmp = pointSet[use];
                            pointSet[use] = pointSet[j];
                            pointSet[j] = tmp;
                        }
                    }
                }
                tmp = pointSet[i];
                pointSet[i] = pointSet[k];
                pointSet[k] = tmp;
            }
            ch.push(pointSet[0]);
            ch.push(pointSet[1]);
            ch.push(pointSet[2]);
            for (i = 3; i < use; i++) {
                while (!(this.multiply(pointSet[i], ch[top - 1], ch[top]) < 0)) {
                    top--;
                    ch.pop();
                }
                top++;
                ch.push(pointSet[i]);
            }
        }
        static GetPathLen(edge) {
            let totalLen = 0;
            for (var i = 0; i < edge.length; i++) {
                let l;
                let r;
                if (i == 0) {
                    l = edge[edge.length - 1];
                    r = edge[i];
                }
                else if (i == edge.length - 1) {
                    l = edge[i];
                    r = edge[0];
                }
                else {
                    l = edge[i];
                    r = edge[i + 1];
                }
                totalLen += LayaUtils.distanceVec2(l, r);
            }
            return totalLen;
        }
    }

    class CatmullRomCurve {
        constructor(p1, p2, p3, p4) {
            this.p1 = p1;
            this.p2 = p2;
            this.p3 = p3;
            this.p4 = p4;
        }
        GetPos(f, out = null) {
            const T0 = this.p1;
            const P0 = this.p2;
            const P1 = this.p3;
            const T1 = this.p4;
            const Ft1 = -0.5;
            const Ft2 = 1.5;
            const Ft3 = -1.5;
            const Ft4 = 0.5;
            const Fu2 = -2.5;
            const Fu3 = 2;
            const Fu4 = -0.5;
            const Fv1 = -0.5;
            const Fv3 = 0.5;
            const FAX = Ft1 * T0.x + Ft2 * P0.x + Ft3 * P1.x + Ft4 * T1.x;
            const FBX = T0.x + Fu2 * P0.x + Fu3 * P1.x + Fu4 * T1.x;
            const FCX = Fv1 * T0.x + Fv3 * P1.x;
            const FDX = P0.x;
            const FAY = Ft1 * T0.y + Ft2 * P0.y + Ft3 * P1.y + Ft4 * T1.y;
            const FBY = T0.y + Fu2 * P0.y + Fu3 * P1.y + Fu4 * T1.y;
            const FCY = Fv1 * T0.y + Fv3 * P1.y;
            const FDY = P0.y;
            const FAZ = Ft1 * T0.z + Ft2 * P0.z + Ft3 * P1.z + Ft4 * T1.z;
            const FBZ = T0.z + Fu2 * P0.z + Fu3 * P1.z + Fu4 * T1.z;
            const FCZ = Fv1 * T0.z + Fv3 * P1.z;
            const FDZ = P0.z;
            let FX = (((FAX * f + FBX) * f + FCX) * f + FDX);
            let FY = (((FAY * f + FBY) * f + FCY) * f + FDY);
            let FZ = (((FAZ * f + FBZ) * f + FCZ) * f + FDZ);
            out = out || new Laya.Vector3();
            out.x = FX;
            out.y = FY;
            out.z = FZ;
            return out;
        }
    }

    class MapCampManager extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.smoothPaths = [];
            this.smoothLen = 100;
        }
        onAwake() {
            MapCampManager.inst = this;
        }
        onDestroy() {
            MapCampManager.inst = null;
            MsgManager.getInst().offAllCaller(this);
        }
        SetPoints(pos) {
            this.smoothPaths = [];
            if (pos.length < 4) {
                if (this.lastLine == null) {
                    this.lastLine.destroy(true);
                }
            }
            var edge = [];
            MathUtils.Graham_scan(pos, edge, pos.length);
            MathUtils.ScalePath(edge, 1);
            var edgeLen = edge.length;
            let p13 = new Laya.Vector3();
            let p23 = new Laya.Vector3();
            let p33 = new Laya.Vector3();
            let p43 = new Laya.Vector3();
            for (var i = 0; i < edgeLen; i++) {
                let fragLen = 10;
                let p1;
                let p2;
                let p3;
                let p4;
                if (i == 0) {
                    p1 = edge[edgeLen - 1].clone();
                    p2 = edge[i].clone();
                    p3 = edge[i + 1].clone();
                    p4 = edge[i + 2].clone();
                }
                else if (i == edgeLen - 2) {
                    p1 = edge[i - 1].clone();
                    p2 = edge[i].clone();
                    p3 = edge[i + 1].clone();
                    p4 = edge[0].clone();
                }
                else if (i == edgeLen - 1) {
                    p1 = edge[i - 1].clone();
                    p2 = edge[i].clone();
                    p3 = edge[0].clone();
                    p4 = edge[1].clone();
                }
                else {
                    p1 = edge[i - 1].clone();
                    p2 = edge[i].clone();
                    p3 = edge[i + 1].clone();
                    p4 = edge[i + 2].clone();
                }
                var t = new Laya.Vector2();
                LayaUtils.subtractVec2(p1, p2, t);
                LayaUtils.setLengthVec2(t, 5, t);
                LayaUtils.addVec2(p2, t, p1);
                LayaUtils.subtractVec2(p4, p3, t);
                LayaUtils.setLengthVec2(t, 5, t);
                LayaUtils.addVec2(p3, t, p4);
                p13.setValue(p1.x, 0.055, p1.y);
                p23.setValue(p2.x, 0.055, p2.y);
                p33.setValue(p3.x, 0.055, p3.y);
                p43.setValue(p4.x, 0.055, p4.y);
                let curve = new CatmullRomCurve(p13, p23, p33, p43);
                for (var j = 0; j < fragLen; j++) {
                    var p = new Laya.Vector3();
                    curve.GetPos(j / fragLen, p);
                    this.smoothPaths.push(p);
                }
            }
            this.smoothPaths.push(this.smoothPaths[0].clone());
            var line = new LineRener();
            line.createMesh(this.smoothPaths, 0.4);
            if (this.lastLine != null) {
                this.lastLine.meshFilter.sharedMesh.destroy();
                this.lastLine.destroy(true);
            }
            var b = new Laya.MeshSprite3D(line.mesh);
            this.lastLine = b;
            var mat = b.meshRenderer.material = new Laya.UnlitMaterial();
            mat.renderMode = 2;
            mat.albedoColor = new Laya.Vector4(1, 1, 1, 0.7);
            this.sp.addChild(b);
        }
    }

    class SimpleMoveCtrl extends Laya.Script3D {
        constructor() {
            super(...arguments);
            this.lastPos = new Laya.Vector2;
            this.Angle = 0;
            this.IsHolding = false;
            this.IsEnding = false;
        }
        onEnable() {
            Laya.stage.on(Laya.Event.MOUSE_DOWN, this, this._onMouseDown);
            Laya.stage.on(Laya.Event.MOUSE_MOVE, this, this._onMouseDrag);
            Laya.stage.on(Laya.Event.MOUSE_UP, this, this._onMouseUp);
        }
        onDisable() {
            Laya.stage.off(Laya.Event.MOUSE_DOWN, this, this._onMouseDown);
            Laya.stage.off(Laya.Event.MOUSE_MOVE, this, this._onMouseDrag);
            Laya.stage.off(Laya.Event.MOUSE_UP, this, this._onMouseUp);
        }
        _onMouseDown() {
            this.lastPos.setValue(Laya.stage.mouseX, Laya.stage.mouseY);
            this.IsHolding = true;
            this.owner.event(SimpleMoveCtrl.StartDrag, [Laya.stage.mouseX, Laya.stage.mouseY]);
        }
        _onMouseDrag() {
            if (this.IsHolding) {
                this.owner.event(SimpleMoveCtrl.DragMove, [Laya.stage.mouseX, Laya.stage.mouseY, this.lastPos.x, this.lastPos.y]);
                this.lastPos.setValue(Laya.stage.mouseX, Laya.stage.mouseY);
            }
        }
        _onMouseUp() {
            this.IsHolding = false;
            this.IsEnding = true;
            this.owner.event(SimpleMoveCtrl.EndDrag);
        }
        onLateUpdate() {
            if (this.IsEnding) {
                this.IsEnding = false;
            }
        }
    }
    SimpleMoveCtrl.StartDrag = 'StartDrag';
    SimpleMoveCtrl.EndDrag = 'EndDrag';
    SimpleMoveCtrl.DragMove = 'DragMove';

    var AniClip;
    (function (AniClip) {
        AniClip["Idle"] = "Rifle Idle_2";
        AniClip["Walk"] = "WalkAndAim";
        AniClip["Attack"] = "Firing Rifle";
        AniClip["Dead"] = "Falling Back Death_1";
        AniClip["Mount"] = "Male Sitting Pose";
    })(AniClip || (AniClip = {}));

    class PlayerRun extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.currentSpeed = 7;
        }
        get entity() {
            return this.entityComponent.entity;
        }
        onAwake() {
            this.moveCtrl = this.sp.addComponent(SimpleMoveCtrl);
            LevelManager.getInst().player = this.sp;
            MsgManager.getInst().on(BattleEvent.RankLevelUp, this, this.onLevelUP);
        }
        onStart() {
            this.entityComponent = this.owner.getComponent(BattleEntityComponent);
            this.entityComponent.PlayAni(AniClip.Attack, 1);
            this.quan = this.Find('quan');
            this.buffView = this.Find('ZHG_quan');
            this.buffView.transform.localScale = new Laya.Vector3(10, 1, 10);
            this.buffView.transform.localPositionY = 0.07;
        }
        onDestroy() {
            super.onDestroy();
            MsgManager.getInst().offAllCaller(this);
        }
        onLevelUP() {
            var pos = this.transform.position;
            PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.MagicBuffYellow).then(p => {
                try {
                    var f = Laya.Sprite3D.instantiate(p);
                    var dd = f.addComponent(DelayDestory);
                    dd.delay = 1000;
                    MapComponent.inst.sp.addChild(f);
                    f.transform.position = pos;
                }
                catch (e) {
                    console.error(e);
                }
            });
        }
        onLateUpdate() {
            this.quan.transform.position.y = 0.06;
            this.quan.transform.position = this.quan.transform.position;
            this.quan.transform.rotationEuler = LayaUtils.Vector3Zero;
            if (this.entity.battleManager.isBattleState) {
                this.buffView.active = true;
                this.buffView.transform.position.y = 0.1;
                this.buffView.transform.position = this.buffView.transform.position;
                this.buffView.transform.rotationEuler = LayaUtils.Vector3Zero;
            }
            else {
                this.buffView.active = false;
            }
        }
        onUpdate() {
            if (this.entity.mount) {
                this.entityComponent.PlayAni('Male Sitting Pose', -1, true, false);
                return;
            }
            if (this.entity.aimTarget != null && this.entity.CanAttackAimEnemy()) {
                if (!LayaUtils.IsZeroVec2(this.entity.steerBehaviorUnit.velocity)) {
                    this.entityComponent.PlayAni(this.entity.weaponConfig.aim_run, 0.5, false, true);
                }
            }
            else {
                if (!LayaUtils.IsZeroVec2(this.entity.steerBehaviorUnit.velocity)) {
                    this.entityComponent.PlayAni(this.entity.weaponConfig.run, 0.5, false, true);
                }
                else {
                    this.entityComponent.PlayAni(this.entity.weaponConfig.idle, -1, false, true);
                }
            }
        }
    }

    class EntitySprite3D extends Laya.Sprite3D {
    }
    class PolygonWrap {
        constructor(ps) {
            this.maxX = -1000000;
            this.minX = 100000;
            this.maxY = -1000000;
            this.minY = 1000000;
            this.debug = false;
            this.points = ps;
            for (var i = 0; i < ps.length; i++) {
                let p = ps[i];
                this.maxX = Math.max(p.x, this.maxX);
                this.minX = Math.min(p.x, this.minX);
                this.maxY = Math.max(p.y, this.maxY);
                this.minY = Math.min(p.y, this.minY);
            }
        }
        FastTest(p) {
            var test = p.x >= this.minX && p.x <= this.maxX && p.y >= this.minY && p.y <= this.maxY;
            if (this.debug) {
                console.log(this.minX, p.x, this.maxX, '', this.minY, p.y, this.maxY, test);
            }
            return test;
        }
    }
    class MapComponent extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.enemyIndex = 0;
            this.entitys = [];
            this.mapPolygon = [];
            this.collisionPolygon = [];
            this.enemyGroup = {};
            this.deDebug = true;
            this.allEnemyCount = 0;
            this.spwanEntitiesTmp = [];
        }
        onAwake() {
            MapComponent.inst = this;
            this.mapNode = this.sp.parent;
            this.initSlotMap();
            MsgManager.getInst().on(BattleEvent.SpwanEntityToView, this, this.onSpwanEntity);
            MsgManager.getInst().on(BattleEvent.ChangeWeapon, this, this.onChangeWeapon);
            MsgManager.getInst().on(BattleEvent.ChangeSoliderSkin, this, this.onChangeModel);
            MsgManager.getInst().on(BattleEvent.SpwanDogTag, this, this.onSpwanDogTag);
            this.sp.addComponent(MapCampManager);
            this.campManager = this.sp.addComponent(MapCampManagerComponent);
            var polygonNode = this.Find('polygon');
            for (var i = 0; i < polygonNode.numChildren; i++) {
                var p = polygonNode.getChildAt(i);
                this.mapPolygon.push(new Laya.Vector2(p.transform.position.x, p.transform.position.z));
            }
            this.enemyConfig = LevelManager.getInst().breakLoadConfig['enemySet'];
            Utils.WaitFrame(2).then(() => {
                this.enemyConfig.break = false;
                MsgManager.getInst().on(EventType.BuildPrefabNodeCompletedNotice, this, this.onBuildPrefabCompleted);
                var old = this.Find('enemySet');
                var p = old.parent;
                old.destroy(true);
                LevelManager.getInst().BuildNode(this.enemyConfig, p, false);
                MsgManager.getInst().off(EventType.BuildPrefabNodeCompletedNotice, this, this.onBuildPrefabCompleted);
                MsgManager.getInst().event(BattleEvent.EnemyLoadEnd);
                this.AttachCovers();
            });
        }
        onDestroy() {
            MapComponent.inst = null;
            MsgManager.getInst().offAllCaller(this);
        }
        initSlotMap() {
        }
        onBuildPrefabCompleted(name, node) {
            switch (name) {
                case _AllPrefabsNames.EnemyFlag:
                    break;
                case _AllPrefabsNames.Mine:
                    break;
                default:
                    this.spwanEnemyEntity(name, node);
                    break;
            }
        }
        spwanEnemyEntity(preafbName, role) {
            var index = this.enemyIndex++;
            this.allEnemyCount++;
            if (BattleData.EnemyIndex.indexOf(index) > -1) {
                role.destroy(true);
                return;
            }
            var groupName = role.parent.name;
            if (this.enemyGroup[groupName] == null) {
                this.enemyGroup[groupName] = [];
            }
            var debug = false;
            if (this.deDebug) {
                debug = true;
                this.deDebug = false;
            }
            MsgManager.getInst().event(BattleEvent.SpwanEntityFromView, preafbName, role.name, 1, CampType.Red, role, debug);
            var entity = role.entity;
            entity.enemyIdnex = index;
            entity.PosV3 = role.transform.position.clone();
            entity.formationPos = role.transform.position.clone();
            entity.formationAngle = role.transform.rotationEuler.y - 180;
            var c = role.addComponent(BattleEntityComponent);
            role.entityCom = c;
            c.entity = entity;
            entity.aniView = c;
            this.enemyGroup[groupName].push(entity);
            entity.team = this.enemyGroup[groupName];
            this.entitys.push(role);
            this.spwanEntitiesTmp.push(role);
        }
        AttachCovers() {
            this.spwanEntitiesTmp.forEach(role => {
                var coverEntity = role.parent.getComponent(BattleEntityComponent);
                if (coverEntity != null) {
                    role.entity.cover = coverEntity.entity;
                    role.entity.seekRadiusSq = role.entity.attackRadiusSq;
                    coverEntity.entity.becovered = coverEntity.entity.becovered || [];
                    coverEntity.entity.becovered.push(role.entity);
                }
            });
        }
        onSpwanEntity(entity, config) {
            var pos;
            var angle = 0;
            if (config == null) {
                var slot = this.campManager.GetSpwanSlot(entity.config.type);
                if (slot == null) {
                    throw new Error('out of solider count!');
                }
                pos = slot.transform.position.clone();
                angle = slot.transform.rotationEuler.y = -180;
            }
            else {
                pos = entity.PosV3;
            }
            var prefab = PrefabManager.getInst().GetPrefab(entity.config.prefab);
            if (prefab == null) {
                console.error('not load entity prefab ', entity.config.prefab);
            }
            var role = Laya.Sprite3D.instantiate(prefab);
            this.mapNode.addChild(role);
            role.transform.position = pos;
            role.transform.localRotationEulerY = angle + 180;
            if (config != null) {
                var matName = config.material;
                var p = PrefabManager.getInst().GetPrefab(_AllPrefabsNames.commandeMat);
                var sp = p.getChildByName(matName);
                var mat = sp.meshRenderer.sharedMaterial;
                var mrs = [];
                Utils.GetAllSKinMeshRender(role, mrs);
                mrs.forEach(mr => {
                    mr.material = mat;
                });
            }
            entity.formationPos = pos.clone();
            entity.formationAngle = angle;
            var c = role.addComponent(BattleEntityComponent);
            c.entity = entity;
            entity.aniView = c;
            role.entity = entity;
            role.entityCom = c;
            this.entitys.push(role);
            if (entity.isPlayer) {
                role.addComponent(PlayerRun);
                this.player = role;
            }
        }
        onChangeWeapon(entity) {
            var view = this.entitys.find(e => { return e.entity == entity; });
            var bec = view.getComponent(BattleEntityComponent);
            bec.ChangeWeapon();
        }
        onChangeModel(entity, prefabName, material = null) {
            var oldview = this.entitys.find(e => { return e.entity == entity; });
            if (oldview != null) {
                var i = this.entitys.indexOf(oldview);
                this.entitys.splice(i, 1);
            }
            else {
                return;
            }
            var prefab = PrefabManager.getInst().GetPrefab(prefabName);
            if (prefab == null) {
                console.error('not load entity prefab ', prefabName);
            }
            var role = Laya.Sprite3D.instantiate(prefab);
            role.entity = entity;
            var parent = oldview.parent || this.sp;
            parent.addChild(role);
            role.transform.position = oldview.transform.position;
            role.transform.localRotationEulerY = oldview.transform.localRotationEulerY;
            oldview.destroy(true);
            var c = role.addComponent(BattleEntityComponent);
            c.entity = entity;
            entity.aniView = c;
            role.entityCom = c;
            role.entity = entity;
            this.entitys.push(role);
            if (entity.isPlayer) {
                role.addComponent(PlayerRun);
                this.player = role;
            }
            if (material != null) {
                var matName = material;
                var p = PrefabManager.getInst().GetPrefab(_AllPrefabsNames.commandeMat);
                var sp = p.getChildByName(matName);
                var mat = sp.meshRenderer.sharedMaterial;
                var mrs = [];
                Utils.GetAllSKinMeshRender(role, mrs);
                mrs.forEach(mr => {
                    mr.material = mat;
                });
            }
        }
        onSpwanDogTag(tag, ani = false) {
            var preafbName = tag.isEnemy ? _AllPrefabsNames.DogTagEnemy : _AllPrefabsNames.DogTag;
            var sp = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(preafbName));
            this.sp.addChild(sp);
            sp.transform.localPositionX = tag.pos.x;
            sp.transform.localPositionY = tag.y;
            sp.transform.localPositionZ = tag.pos.y;
            sp.transform.localRotationEulerY = tag.r;
            if (tag.y > 0) {
                sp.getChildAt(1).active = false;
            }
            var c = sp.addComponent(DogTagComponent);
            tag.view = c;
            if (ani) {
                tag.ani = sp.addComponent(DogTagComponentAni);
                tag.ani.entity = tag;
            }
        }
        CreatePlayerNode() {
            var node = this.Find("PlayerLocation");
            var p = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(_AllPrefabsNames.PlayerLoacation));
            node.addChild(p);
            p.transform.localPosition = LayaUtils.Vector3Zero;
        }
        onUpdate() {
            for (var i = this.entitys.length - 1; i >= 0; i--) {
                var e = this.entitys[i];
                if (e == null) {
                    this.entitys.splice(i, 1);
                    continue;
                }
                var entity = e.entity;
                if (!e.destroyed) {
                    if (entity.fsm.CurrentState) {
                        if (entity.fsm.CurrentState.name != 'AttackState' && !entity.isPlayer) {
                            if (!LayaUtils.IsZeroVec2(entity.steerBehaviorUnit.velocity)) {
                                entity.aniView.PlayAni(entity.weaponConfig.run, 0.4, false, false);
                            }
                            else {
                                entity.aniView.PlayAni(entity.weaponConfig.idle, -1, false, false);
                            }
                        }
                    }
                    if (entity.isDeath) {
                        this.entitys.splice(i, 1);
                        e.entityCom.PlayAni(entity.weaponConfig.death, -1, true, false);
                        var viewNode = Utils.GetNode(e.entityCom.sp, 'View');
                        if (viewNode) {
                            for (var i = 0; i < viewNode.numChildren; i++) {
                                viewNode.getChildAt(i).addComponent(ThrowDestroy);
                            }
                        }
                        e.addComponent(DelayDestory);
                        e.entityCom.destroy();
                        continue;
                    }
                    if (e.entity.isPlayer && e.entity.mount) {
                        continue;
                    }
                    if (e.entity.battleManager.isBattleState) {
                        e.entity.aniView.BlinkShow();
                    }
                    if (!this.isSame(e.transform.position, e.entity.PosV3)) {
                        e.transform.position = e.entity.PosV3;
                    }
                    if (entity.aniView != null && entity.aniView.ani != null && !entity.isPlayer) {
                        if (!GameManager.getInst().battleManager.aimMode && LayaUtils.distanceSquaredVec2(entity.Pos, SimpleCameraFollow.CenterPos) > 625) {
                            if (entity.aniView.sp.active) {
                                entity.aniView.sp.active = false;
                            }
                        }
                        else {
                            if (entity.aniView.sp.active == false) {
                                entity.aniView.currentAni = null;
                                entity.aniView.sp.active = true;
                            }
                        }
                    }
                    if (entity.steerBehaviorUnit.maxSpeed == 0 || entity.mount) {
                        continue;
                    }
                    var rTran = e.entityCom.rotateNode;
                    if (entity.aimTarget == null && rTran == e) {
                        if (e.entity.config.type == SoilderType.Tank) {
                            rTran = e.entityCom.sp;
                        }
                        var ey = e.entity.steerBehaviorUnit.angle + 180;
                        rTran.transform.rotationEuler.y = Utils.AngleLerp(rTran.transform.rotationEuler.y, ey, LayaUtils.deltaTimeSec * 10);
                        rTran.transform.rotationEuler = rTran.transform.rotationEuler;
                    }
                    else if (entity.aimTarget != null) {
                        if (e.entity.config.type == SoilderType.Tank) {
                            if (!e.entity.CanAttackAimEnemy()) {
                                rTran = e.entityCom.sp;
                            }
                        }
                        var oldY = rTran.transform.localRotationEulerY;
                        rTran.transform.lookAt(entity.aimTarget.PosV3, LayaUtils.UP);
                        rTran.transform.localRotationEulerX = rTran.transform.localRotationEulerZ = 0;
                        var newY = rTran.transform.localRotationEulerY + 180;
                        rTran.transform.localRotationEulerY = Utils.AngleLerp(oldY, newY, LayaUtils.deltaTimeSec * 10);
                        e.entity.steerBehaviorUnit.angle = e.transform.localRotationEulerY + 180;
                    }
                }
            }
        }
        isSame(a, b) {
            var dx = a.x - b.x;
            var dz = a.z - b.z;
            return Laya.MathUtils3D.isZero(dx) && Laya.MathUtils3D.isZero(dz);
        }
    }

    var Shader3D = Laya.Shader3D;
    var SubShader = Laya.SubShader;
    class FlagMaterial extends Laya.Material {
        constructor() {
            super();
            this.MAIN_Color = Shader3D.propertyNameToID("u_MainColor");
            this.setShaderName("Flag");
            this.cull = Laya.RenderState.CULL_NONE;
        }
        static initShader() {
            var vs = `
        #include "Lighting.glsl";
        
        attribute vec4 a_Position;
        attribute vec2 a_Textcoord;
        uniform mat4 u_WorldMat;
        uniform mat4 u_MvpMatrix;
        uniform float u_Time;
        
        
        void main()
        {
            vec4 p = vec4(a_Position);
            float a = min(1.0,abs(p.x)*0.3);

            p.z += a*a*sin(u_Time*8.0 + (1.5 - p.x) * 21.5) * 0.5;
            p.y += a*cos(u_Time*2.0 + p.x * 21.2) * 0.3;
            // p.x -= a*a*sin(u_Time*8.0 + (1.5 - p.x) * 1.2) * 0.3;
            
            gl_Position = u_MvpMatrix * p;
            gl_Position = remapGLPositionZ(gl_Position);
        }`;
            var ps = `
        #ifdef HIGHPRECISION
        precision highp float;
        #else
        precision mediump float;
        #endif

        #include "Lighting.glsl";
        
        uniform sampler2D u_MainTex;
        uniform sampler2D u_NoiseTex;
        uniform vec4 u_MainColor;

        void main()
        {
            gl_FragColor = u_MainColor;
        }
        `;
            var attributeMap = {
                "a_Position": Laya.VertexMesh.MESH_POSITION0,
                "a_Normal": Laya.VertexMesh.MESH_NORMAL0,
            };
            var uniformMap = {
                "u_WorldMat": Shader3D.PERIOD_SPRITE,
                'u_MvpMatrix': Shader3D.PERIOD_SPRITE,
                'u_MainColor': Shader3D.PERIOD_MATERIAL,
                'u_Time': Shader3D.PERIOD_SCENE,
            };
            var shader = Shader3D.add("Flag", null, null, true);
            var subShader = new SubShader(attributeMap, uniformMap);
            shader.addSubShader(subShader);
            subShader.addShaderPass(vs, ps);
        }
        set mainColor(value) {
            this._shaderValues.setVector(this.MAIN_Color, value);
        }
    }

    class Flag extends BaseScript3D {
        onAwake() {
            FlagMaterial.initShader();
            var flag = this.Find('RedFlag');
            var mat = flag.getChildAt(0).meshRenderer.material;
            var oldColor = mat.albedoColor;
            var flagMat = flag.getChildAt(0).meshRenderer.material = new FlagMaterial();
            flagMat.mainColor = oldColor;
            flag = this.Find('BlueFlag');
            mat = flag.getChildAt(0).meshRenderer.material;
            oldColor = mat.albedoColor;
            flagMat = flag.getChildAt(0).meshRenderer.material = new FlagMaterial();
            flagMat.mainColor = oldColor;
        }
        onEnable() {
            this.destroy();
        }
    }

    var WaterWaveVS = `
#include 'Lighting.glsl';

attribute vec4 a_Position;

attribute vec2 a_Textcoord;

attribute vec3 a_Normal;

uniform mat4 u_WorldMat;

uniform mat4 u_MvpMatrix;
        
varying vec2 v_textcoord;

varying vec3 v_normal;

varying vec4 v_wpos;
        
void main()
{
    v_textcoord = a_Textcoord;

    v_normal = a_Normal;

    v_wpos = u_WorldMat * a_Position;

    gl_Position = u_MvpMatrix * a_Position;

    gl_Position = remapGLPositionZ(gl_Position);
}
`;
    var WaterWaveFS = `
#if defined(GL_FRAGMENT_PRECISION_HIGH)
    precision highp float;
#else
    precision mediump float;
#endif
#include "Lighting.glsl";
uniform sampler2D u_MainTex;
uniform sampler2D u_NoiseTex;
uniform vec3 u_CameraPos;
uniform vec3 u_AmbientColor;
uniform sampler2D u_LightBuffer;
uniform float u_Time; 
uniform vec4 u_Ting;
varying vec2 v_textcoord;
varying vec3 v_normal;
varying vec4 v_wpos;

uniform float u_WaterFogStart;
uniform float u_WaterFogRange;
uniform vec3 u_WaterFogColor;

void main()
{
    DirectionLight directionLight = getDirectionLight(u_LightBuffer , 0);

    vec2 uv = v_textcoord * vec2(u_Ting.x , u_Ting.y);
    uv.x -= floor(uv.x);
    uv.y -= floor(uv.y);

    float time = u_Time / 8.0;

    float _NoiseSpeedX = 1.0;
    float _NoiseSpeedY = 1.0;
    float _NoiseScaleX = 0.2;
    float _NoiseScaleY = 0.2;
    // float _SpecularGlossy = 0.16;
    // float _SpecularIntensity = 0.5;
    float _NoiseBrightOffset = 0.25;
    vec2 tuv1 = uv + vec2(time * _NoiseSpeedX , 0);
    vec2 tuv2 = uv + vec2(0 , time * _NoiseSpeedY); 
    vec2 ouvxy = vec2(texture2D(u_NoiseTex,tuv1).r , texture2D(u_NoiseTex,tuv2).r);
    ouvxy -= _NoiseBrightOffset;
    ouvxy *= vec2(_NoiseScaleX ,  _NoiseScaleY); 
    vec4 col = texture2D(u_MainTex , uv + ouvxy);
    vec3 normal = v_normal;
    normal.xy += ouvxy.xy; 
    vec3 L = normalize(directionLight.direction.xyz);
    vec3 N = normalize(normal);
    // float LdotN = dot(L, N);
    vec3 diffuse = col.rgb; 
    
    // vec3 specular = vec3(0.0);
    vec3 V = normalize(u_CameraPos.xyz - v_wpos.xyz);
    // vec3 H = normalize(L + V);
    // if (LdotN > 0.0)
    // {
    //     float HdotN = max(0.0, dot(H, N)); // blinn-phone
    //     specular = directionLight.color.rgb * pow(HdotN, _SpecularGlossy * 100.0) * _SpecularIntensity;
    // } 
     vec3 ambient = u_AmbientColor.rgb; 
    // gl_FragColor = vec4(diffuse + specular + ambient, 1.0); 

    gl_FragColor = vec4(diffuse , 1.0); 
    
    // gl_FragColor = texture2D(u_MainTex , uv);

    #ifdef WATERFOG
		// vec3 vv = u_CameraPos - v_PositionTT;
		// float vvl = length(vv);

		float lerpFact=clamp(( 1.0/gl_FragCoord.w - u_WaterFogStart ) / u_WaterFogRange , 0.0 , 1.0);

		if(lerpFact > 0.1)
		{
			gl_FragColor.rgb=mix(gl_FragColor.rgb , u_WaterFogColor , lerpFact / 2.0 + 0.5);
		}else
		{
			gl_FragColor.rgb=mix(gl_FragColor.rgb , u_WaterFogColor , smoothstep(0.0 , 0.1 , lerpFact) * 0.5);
		}
	#endif
}
`;
    class WaterWaveMaterial extends Laya.Material {
        constructor() {
            super();
            this.MAIN_TEX = Laya.Shader3D.propertyNameToID("u_MainTex");
            this.NOISE_TEX = Laya.Shader3D.propertyNameToID("u_NoiseTex");
            this.u_Ting = Laya.Shader3D.propertyNameToID("u_Ting");
            this.setShaderName("Water");
            this.tingOffset = new Laya.Vector4(50, 50, 0, 0);
        }
        set mainTexture(value) {
            this._shaderValues.setTexture(this.MAIN_TEX, value);
        }
        get mainTexture() {
            return this._shaderValues.getTexture(this.MAIN_TEX);
        }
        set normalTexture(value) {
            this._shaderValues.setTexture(this.NOISE_TEX, value);
        }
        get normalTexture() {
            return this._shaderValues.getTexture(this.NOISE_TEX);
        }
        set tingOffset(value) {
            this._shaderValues.setVector(this.u_Ting, value);
        }
        set fogStart(value) {
            this._shaderValues.setNumber(WaterWaveMaterial.u_WaterFogStart, value);
        }
        set fogRange(value) {
            this._shaderValues.setNumber(WaterWaveMaterial.u_WaterFogRange, value);
        }
        set fogColor(value) {
            if (value) {
                this._shaderValues.addDefine(WaterWaveMaterial.WATERFOG);
            }
            else {
                this._shaderValues.removeDefine(WaterWaveMaterial.WATERFOG);
            }
            this._shaderValues.setVector3(WaterWaveMaterial.u_WaterFogColor, value);
        }
        static init() {
            var attributeMap = {
                "a_Position": Laya.VertexMesh.MESH_POSITION0,
                "a_Normal": Laya.VertexMesh.MESH_NORMAL0,
                "a_Textcoord": Laya.VertexMesh.MESH_TEXTURECOORDINATE0
            };
            var uniformMap = {
                "u_WorldMat": Laya.Shader3D.PERIOD_SPRITE,
                'u_MvpMatrix': Laya.Shader3D.PERIOD_SPRITE,
                'u_LightBuffer': Laya.Shader3D.PERIOD_SCENE,
                'u_AmbientColor': Laya.Shader3D.PERIOD_SCENE,
                'u_CameraPos': Laya.Shader3D.PERIOD_CAMERA,
                'u_Time': Laya.Shader3D.PERIOD_SCENE,
                'u_MainTex': Laya.Shader3D.PERIOD_MATERIAL,
                'u_NoiseTex': Laya.Shader3D.PERIOD_MATERIAL,
                'u_Ting': Laya.Shader3D.PERIOD_MATERIAL,
                'u_WaterFogStart': Laya.Shader3D.PERIOD_MATERIAL,
                'u_WaterFogRange': Laya.Shader3D.PERIOD_MATERIAL,
                'u_WaterFogColor': Laya.Shader3D.PERIOD_MATERIAL,
            };
            var shader = Laya.Shader3D.add("Water", true);
            var subShader = new Laya.SubShader(attributeMap, uniformMap);
            shader.addSubShader(subShader);
            subShader.addShaderPass(WaterWaveVS, WaterWaveFS);
        }
    }
    WaterWaveMaterial.u_WaterFogStart = Laya.Shader3D.propertyNameToID("u_WaterFogStart");
    WaterWaveMaterial.u_WaterFogRange = Laya.Shader3D.propertyNameToID("u_WaterFogRange");
    WaterWaveMaterial.u_WaterFogColor = Laya.Shader3D.propertyNameToID("u_WaterFogColor");
    WaterWaveMaterial.WATERFOG = Laya.Shader3D.getDefineByName("WATERFOG");

    let WaterMatHelper = class WaterMatHelper extends Singleton {
    };
    __decorate([
        res_assign('res/tex/waves_n.png')
    ], WaterMatHelper.prototype, "noise", void 0);
    __decorate([
        res_assign('res/tex/water.png')
    ], WaterMatHelper.prototype, "water", void 0);
    WaterMatHelper = __decorate([
        res_auto_bind()
    ], WaterMatHelper);

    class GameItemAttachMgr extends Singleton {
        Register() {
            MsgManager.getInst().on(EventType.BuildNodeNotice, this, this.onBuildNode);
            MsgManager.getInst().on(EventType.BuildPrefabNodeNotice, this, this.onBuildPrefabNode);
            MsgManager.getInst().on(EventType.BuildNodeCompletedNotice, this, this.onBuildNodeCompleted);
            MsgManager.getInst().on(EventType.BuildPrefabNodeCompletedNotice, this, this.onBuildPrefabCompleted);
        }
        Reset() {
        }
        onPreCreatePrefab(prefab, nodeName) {
            return prefab;
        }
        onBuildNode(name, node, nodeConfig) {
            switch (name) {
                case 'enemySet':
                    nodeConfig.break = true;
                    break;
            }
        }
        onBuildPrefabNode(name, node) {
        }
        onBuildNodeCompleted(name, node) {
            switch (name) {
                case 'Map':
                    node.addComponent(MapComponent);
                    break;
            }
        }
        onBuildPrefabCompleted(name, node) {
            switch (name) {
                case _AllPrefabsNames.CameraNode:
                    var cam = Utils.GetNode(node, 'Main Camera');
                    cam.enableHDR = false;
                    cam.clearFlag = Laya.CameraClearFlags.DepthOnly;
                    cam.orthographic = true;
                    cam.orthographicVerticalSize = 14.68 * 2;
                    cam.farPlane = 100;
                    cam.nearPlane = 5;
                    LevelManager.getInst().mainCamera = cam;
                    LevelManager.getInst().mainCameraRotRoot = node;
                    node.addComponent(SimpleCameraFollow);
                    break;
                case _AllPrefabsNames.Water:
                    var r = node.meshRenderer;
                    WaterWaveMaterial.init();
                    var mat = r.sharedMaterial = new WaterWaveMaterial();
                    mat.mainTexture = WaterMatHelper.getInst().water;
                    mat.normalTexture = WaterMatHelper.getInst().noise;
                    mat.tingOffset = new Laya.Vector4(60, 60, 0, 0);
                    break;
                case _AllPrefabsNames.EnemyFlag:
                    node.addComponent(Flag);
                    break;
                case _AllPrefabsNames.yan:
                    break;
                case _AllPrefabsNames.Mine:
                    node.addComponent(DelayDestory);
                    break;
            }
        }
    }

    let LevelManager = class LevelManager extends Singleton {
        constructor() {
            super();
            this.isSkinModel = false;
            this.lastLevelPrefabs = {};
            this.currentLevelPrefabs = {};
            this.breakLoadConfig = {};
        }
        _init_() {
            this.ResetScene();
        }
        ResetScene() {
            this.sceneConfig = {};
            this.sceneRawConfig.forEach(item => {
                this.sceneConfig[item.n] = item;
            });
            if (this.scene == null) {
                this.scene = new Laya.Scene3D();
                Laya.stage.addChild(this.scene);
                this.scene.zOrder = -1;
                var a = this.sceneConfig['ambientColor'];
                if (a != null && a.ad != null) {
                    this.scene.ambientColor = this.parseColor3(a.ad.c);
                }
                var fog = this.sceneConfig['Fog'];
                if (fog != null && fog.fd != null) {
                    this.scene.enableFog = true;
                    this.scene.fogColor = this.parseColor3(fog.fd.c);
                    this.scene.fogStart = fog.fd.start;
                    this.scene.fogRange = fog.fd.end - fog.fd.start;
                }
            }
        }
        LoadCachePrefab(p) {
            return new Promise((r) => {
                var node = this.sceneConfig['cache'];
                if (node == null) {
                    r();
                    return;
                }
                var ps = {};
                this.CollectPrefabs(node, ps);
                var keys = Object.keys(ps);
                keys.forEach(key => {
                    if (ps[key] == null) {
                        console.log(key, ' is null');
                    }
                    ps[key].isCache = true;
                });
                PrefabManager.getInst().LoadPrefab(ps, () => {
                    r();
                }, p, this);
            });
        }
        LoadStatic() {
            return new ProgressPromise((r, progress) => {
                var node = this.sceneConfig['static'];
                if (node == null) {
                    r();
                    return;
                }
                if (this.staticEnvNode != null) {
                    this.staticEnvNode.destroy();
                }
                this.staticEnvNode = new Laya.Sprite3D("staticEnvNode");
                this.scene.addChild(this.staticEnvNode);
                var p = {};
                this.CollectPrefabs(node, p);
                try {
                    PrefabManager.getInst().LoadPrefab(p, () => {
                        try {
                            this.BuildNode(node, this.staticEnvNode);
                            var items = Object.keys(p).map(e => {
                                return p[e];
                            });
                            PrefabManager.getInst().UnLoadPrefab(items);
                            r();
                        }
                        catch (e) {
                        }
                    }, progress, this);
                }
                catch (e) {
                    console.error(e);
                }
            });
        }
        LoadLevel(name) {
            if (this.sceneConfig[name] == null) {
                throw new Error(`not has level{${name}} config!`);
            }
            console.log('load level:', name);
            this.lastLevelPrefabs = this.currentLevelPrefabs;
            this.currentLevelPrefabs = {};
            var scene = this.currentSceneConfg = this.sceneConfig[name];
            this.CollectPrefabs(scene, this.currentLevelPrefabs);
            this.UnLoadLastScene();
            this.unLoadNotUsedPrafab();
            GameItemAttachMgr.getInst().Reset();
            return this.LoadDynamicScene();
        }
        GameStartLoad() {
            return new ProgressPromise((r, p) => {
                this.LoadCachePrefab(p1 => {
                    p(p1 * 0.5);
                }).then(() => {
                    console.log('动态资源加载完成');
                    this.LoadStatic().progress(p2 => {
                        p(0.5 + p2 * 0.5);
                    }).then(() => {
                        p(1);
                        console.log('静态资源加载完成');
                        r();
                    }).catch(e => {
                        console.error(e);
                    });
                }).catch(e => {
                    console.error(e);
                });
            });
        }
        CollectPrefab(ps, name) {
            ps[name] = PrefabManager.getInst().prefabMap[name];
        }
        CollectPrefabs(node, ps) {
            if (node == null)
                return;
            var prefabName = node.pn;
            if (prefabName != null && ps[prefabName] == null) {
                ps[node.pn] = PrefabManager.getInst().prefabMap[prefabName];
                this.CollectPrefab(ps, prefabName);
            }
            if (node.c != null && node.c.length != 0) {
                for (var i = 0; i < node.c.length; i++) {
                    this.CollectPrefabs(node.c[i], ps);
                }
            }
        }
        UnLoadLastScene() {
            if (this.currentSceneNode != null) {
                this.currentSceneNode.destroy(true);
                this.currentSceneNode = null;
            }
        }
        unLoadNotUsedPrafab() {
            var diffP = [];
            var keys = Object.keys(this.lastLevelPrefabs);
            for (var i = 0; i < keys.length; i++) {
                var key = keys[i];
                if (this.currentLevelPrefabs[key] == null && !this.lastLevelPrefabs[key].isCache) {
                    diffP.push(this.lastLevelPrefabs[key]);
                }
            }
            PrefabManager.getInst().UnLoadPrefab(diffP);
        }
        LoadDynamicScene() {
            return new ProgressPromise((r, p) => {
                if (this.currentLevelPrefabs == null) {
                    throw new Error('场景配置为空,是否导出场景?');
                }
                this.LoadPrefabs(this.currentLevelPrefabs, (v) => { p(v); }).then(() => {
                    this.currentSceneNode = new Laya.Sprite3D('currentSceneNode');
                    this.scene.addChild(this.currentSceneNode);
                    this.BuildNode(this.currentSceneConfg, this.currentSceneNode);
                    r();
                    this.breakLoadConfig = {};
                });
            });
        }
        LoadPrefabs(collect, progress = null) {
            return new Promise(r => {
                PrefabManager.getInst().LoadPrefab(collect, () => { r(); }, progress, this);
            });
        }
        BuildNode(node, parent, checkBreak = true) {
            var sp;
            let prefabName;
            if (node.pn != null) {
                prefabName = node.pn;
                if (this.onPreCreatePrefab != null) {
                    prefabName = this.onPreCreatePrefab(prefabName, node.n);
                }
                var prefab = PrefabManager.getInst().GetPrefab(prefabName);
                if (!prefab) {
                    console.error(prefabName, 'not load!');
                }
                sp = Laya.Sprite3D.instantiate(prefab);
                sp.name = node.n;
            }
            if (sp == null) {
                sp = new Laya.Sprite3D(node.n);
                sp.name = node.n;
            }
            sp.transform.localPosition = new Laya.Vector3();
            sp.transform.localRotationEulerY = sp.transform.localRotationEulerY;
            sp.transform.localScale = sp.transform.localScale;
            parent.addChild(sp);
            if (node.t != null) {
                if (node.t.p != null) {
                    sp.transform.localPosition = this.parseVector3(node.t.p);
                }
                if (node.t.e != null) {
                    sp.transform.localRotationEuler = this.parseVector3(node.t.e);
                }
                if (node.t.s != null) {
                    sp.transform.localScale = this.parseVector3(node.t.s);
                }
                if (sp instanceof Laya.Camera || node.n.toLocaleLowerCase().includes('camera')) ;
                if (node.n == 'light') {
                    if (sp instanceof Laya.DirectionLight) {
                        sp.shadowMode = Laya.ShadowMode.None;
                    }
                }
            }
            if (prefabName != null) {
                MsgManager.getInst().event(EventType.BuildPrefabNodeNotice, prefabName, sp, node.n);
            }
            MsgManager.getInst().event(EventType.BuildNodeNotice, node.n, sp, node);
            if (node.pd != null) {
                this.SetPrefabDiff(node.pd, sp);
            }
            if (node.c != null) {
                if (checkBreak && node.break == true) {
                    this.breakLoadConfig[node.n] = node;
                }
                else {
                    for (var i = 0; i < node.c.length; i++) {
                        this.BuildNode(node.c[i], sp);
                    }
                }
            }
            if (prefabName != null) {
                MsgManager.getInst().event(EventType.BuildPrefabNodeCompletedNotice, prefabName, sp, node.n);
            }
            MsgManager.getInst().event(EventType.BuildNodeCompletedNotice, node.n, sp);
        }
        SetPrefabDiff(pd, node) {
            if (pd.child != null) {
                for (var i = 0; i < pd.child.length; i++) {
                    var child = pd.child[i];
                    var sp = node.getChildAt(child.index);
                    if (child.transform) {
                        if (child.transform.p != null) {
                            let old = this.parseVector3(child._transform.p);
                            Laya.Vector3.add(old, this.parseVector3(child.transform.p), old);
                            sp.transform.localPosition = old;
                        }
                        if (child.transform.e != null) {
                            let old = this.parseVector3(child._transform.e);
                            Laya.Vector3.add(old, this.parseVector3(child.transform.e), old);
                            sp.transform.localRotationEuler = old;
                        }
                        if (child.transform.s != null) {
                            let old = this.parseVector3(child._transform.s);
                            Laya.Vector3.add(old, this.parseVector3(child.transform.s), old);
                            sp.transform.localScale = old;
                        }
                    }
                    if (child.child != null) {
                        this.SetPrefabDiff(child, sp);
                    }
                }
            }
            if (pd.unknownChild != null) {
                for (var i = 0; i < pd.unknownChild.length; i++) {
                    var uk = pd.unknownChild[i];
                    this.BuildNode(uk, node);
                }
            }
        }
        parseVector3(str, _outV3) {
            var strs = str.split(',');
            _outV3 = _outV3 || new Laya.Vector3();
            _outV3.setValue(Number(strs[0]), Number(strs[1]), Number(strs[2]));
            return _outV3;
        }
        parseColor(str) {
            var strs = str.split(',');
            return new Laya.Vector4(Number(strs[0]), Number(strs[1]), Number(strs[2]), Number(strs[3]));
        }
        parseColor3(str) {
            var strs = str.split(',');
            return new Laya.Vector3(Number(strs[0]), Number(strs[1]), Number(strs[2]));
        }
    };
    __decorate([
        res_assign("res/scene/level.json")
    ], LevelManager.prototype, "sceneRawConfig", void 0);
    LevelManager = __decorate([
        res_auto_bind()
    ], LevelManager);

    class SimpleCameraFollow extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.ray = new Laya.Ray(new Laya.Vector3, new Laya.Vector3);
            this.point = new Laya.Vector2(Laya.stage.width / 2, Laya.stage.height / 2);
            this.deltaR = 3;
            this.lookTarget = new Laya.Vector3();
            this.tmp = new Laya.Vector3();
        }
        static get TargetPos() {
            var player = SimpleCameraFollow.inst.target || LevelManager.getInst().player;
            this.targetPos.setValue(player.transform.position.x, player.transform.position.z);
            return this.targetPos;
        }
        static get HasTarget() {
            return this.inst.target != null;
        }
        static get CenterPos() {
            if (this.inst != null) {
                this.centerPos.setValue(this.inst.tmp.x, this.inst.tmp.z);
            }
            return this.centerPos;
        }
        onAwake() {
            SimpleCameraFollow.inst = this;
        }
        onStart() {
            this.camera = this.Find('Main Camera');
            this.cameraStaticLocalPos = this.camera.transform.localPosition.clone();
        }
        onUpdate() {
            try {
                this.camera.viewportPointToRay(this.point, this.ray);
                var scale = -this.ray.origin.y / this.ray.direction.y;
                Laya.Vector3.scale(this.ray.direction, scale, this.tmp);
                Laya.Vector3.add(this.ray.origin, this.tmp, this.tmp);
            }
            catch (err) {
                ReportMgr.inst.sendReport("Code_Error_SimpleCameraFollow " + JSON.stringify(err), false);
            }
        }
        onLateUpdate() {
            var player = this.target || LevelManager.getInst().player;
            if (player == null || player.transform == null)
                return;
            var r = Math.min(1, this.target ? (LayaUtils.deltaTimeSec * this.deltaR) : 0.7);
            this.transform.position.z = Laya.MathUtil.lerp(this.transform.position.z, player.transform.position.z, r);
            this.transform.position.y = Math.min(4, player.transform.position.y);
            this.transform.position.x = Laya.MathUtil.lerp(this.transform.position.x, player.transform.position.x, r);
            this.UpdatePosition();
        }
        onDestroy() {
            SimpleCameraFollow.inst = null;
        }
        SmallShake() {
            return;
        }
    }
    SimpleCameraFollow.targetPos = new Laya.Vector2;
    SimpleCameraFollow.centerPos = new Laya.Vector2;

    var GameProcedureEvent;
    (function (GameProcedureEvent) {
        GameProcedureEvent["EnterBattleGame"] = "EnterBattleGame";
        GameProcedureEvent["StartSteeringMove"] = "StartSteeringMove";
        GameProcedureEvent["StartBattle"] = "StartBattle";
        GameProcedureEvent["StopBattle"] = "StopBattle";
    })(GameProcedureEvent || (GameProcedureEvent = {}));

    class UI_ui_aim extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "ui_aim"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m_guide_step = this.getControllerAt(1);
            this.m_bg = (this.getChildAt(0));
            this.m_comp_zhiyin = (this.getChildAt(1));
            this.m_t0 = this.getTransitionAt(0);
        }
    }
    UI_ui_aim.URL = "ui://ixnra5xgeperxta";

    let AimMenu = class AimMenu extends BaseUI {
        constructor() {
            super(...arguments);
            this.left = 0;
            this.right = 0;
            this.top = 0;
            this.bottom = 0;
            this.point = new Laya.Vector2();
            this.ray = new Laya.Ray(new Laya.Vector3(), new Laya.Vector3());
            this.stepMove = 0;
            this.oldTarget = new Laya.Vector3();
        }
        get URL() {
            return UI_ui_aim.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_bg];
        }
        OnCreate() {
            this.left = Laya.stage.width / 4;
            this.right = Laya.stage.width * 3 / 4;
            this.top = Laya.stage.height / 4;
            this.bottom = Laya.stage.height * 3 / 4;
        }
        OnShow(event, radius) {
            this.eventType = event;
            this.ui.m_comp_zhiyin.visible = true;
            this.ui.m_bg.on(Laya.Event.MOUSE_DOWN, this, this.onStartMove);
            this.aimNode = MapCampManagerComponent.inst.SpwanAimCircle();
            this.camera = this.aimNode.getChildAt(0);
            this.camera.orthographic = true;
            this.camera.orthographicVerticalSize = 50;
            this.aim = this.aimNode.getChildAt(1);
            radius *= 2;
            this.aim.transform.localScale = new Laya.Vector3(radius, radius, radius);
            GameManager.getInst().battleManager.aimMode = true;
        }
        OnHide() {
            MsgManager.getInst().offAllCaller(this);
            this.ui.m_bg.off(Laya.Event.MOUSE_DOWN, this, this.onStartMove);
            GameManager.getInst().battleManager.aimMode = false;
        }
        OnDispose() {
        }
        onStartMove(e) {
            if (this.touchID != null)
                return;
            try {
                this.touchID = e.touchId;
                Laya.timer.frameLoop(1, this, this.onMove);
                this.ui.on(Laya.Event.MOUSE_UP, this, this.onEndMove);
                this.ui.on(Laya.Event.MOUSE_OUT, this, this.onEndMove);
                this.point.setValue(Laya.stage.mouseX, Laya.stage.mouseY);
                this.camera.viewportPointToRay(this.point, this.ray);
                var oldx = this.ray.origin.x + this.ray.direction.x * 50;
                this.point.x += 1;
                this.camera.viewportPointToRay(this.point, this.ray);
                this.stepMove = this.ray.origin.x + this.ray.direction.x * 50 - oldx;
            }
            catch (err) {
                ReportMgr.inst.sendReport("Code_Error_驱逐舰 " + JSON.stringify(err), false);
            }
        }
        onMove() {
            try {
                var x = this.ray.origin.x + this.ray.direction.x * 50;
                var y = this.ray.origin.z + this.ray.direction.z * 50;
                this.point.setValue(Laya.stage.mouseX, Laya.stage.mouseY);
                this.camera.viewportPointToRay(this.point, this.ray);
                var dx = this.ray.origin.x + this.ray.direction.x * 50 - x;
                var dy = this.ray.origin.z + this.ray.direction.z * 50 - y;
                this.aim.transform.position.x += dx;
                this.aim.transform.position.z += dy;
                this.aim.transform.position = this.aim.transform.position;
                this.aim.transform.position.cloneTo(this.oldTarget);
                if (Laya.stage.mouseX < this.left) {
                    this.aimNode.transform.position.x += 0.1;
                }
                if (Laya.stage.mouseX > this.right) {
                    this.aimNode.transform.position.x -= 0.1;
                }
                if (Laya.stage.mouseY < this.top) {
                    this.aimNode.transform.position.z += 0.1;
                }
                if (Laya.stage.mouseY > this.bottom) {
                    this.aimNode.transform.position.z -= 0.1;
                }
                this.aimNode.transform.position = this.aimNode.transform.position;
                this.aim.transform.position = this.oldTarget;
            }
            catch (err) {
                ReportMgr.inst.sendReport("Code_Error_驱逐舰 " + JSON.stringify(err), false);
            }
        }
        onEndMove() {
            try {
                Laya.timer.clear(this, this.onMove);
                this.ui.off(Laya.Event.MOUSE_UP, this, this.onEndMove);
                this.ui.off(Laya.Event.MOUSE_OUT, this, this.onEndMove);
                this.touchID = null;
                UIManager.inst.PopUI(this.layer, false);
                var p = this.aim.transform.position.clone();
                p.y = 0;
                this.aimNode.destroy();
                MsgManager.getInst().event(this.eventType, p);
            }
            catch (err) {
                ReportMgr.inst.sendReport("Code_Error_驱逐舰 " + JSON.stringify(err), false);
            }
        }
    };
    AimMenu = __decorate([
        ui_register()
    ], AimMenu);

    class UI_ui_aim_hm extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "ui_aim_hm"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m_guide_step = this.getControllerAt(1);
            this.m_bg = (this.getChildAt(0));
            this.m_t0 = this.getTransitionAt(0);
        }
    }
    UI_ui_aim_hm.URL = "ui://ixnra5xgkdf4xu6";

    let HangMuAimUI = class HangMuAimUI extends BaseUI {
        constructor() {
            super(...arguments);
            this.left = 0;
            this.right = 0;
            this.top = 0;
            this.bottom = 0;
            this.targetPos = new Laya.Vector3;
            this.revers = false;
            this.point = new Laya.Vector2();
            this.ray = new Laya.Ray(new Laya.Vector3(), new Laya.Vector3());
            this.oldTarget = new Laya.Vector3();
        }
        get URL() {
            return UI_ui_aim_hm.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_bg];
        }
        OnCreate() {
            this.left = Laya.stage.width / 4;
            this.right = Laya.stage.width * 3 / 4;
            this.top = Laya.stage.height / 4;
            this.bottom = Laya.stage.height * 3 / 4;
        }
        OnShow() {
            this.ui.m_bg.on(Laya.Event.MOUSE_DOWN, this, this.onStartMove);
            this.aimNode = MapCampManagerComponent.inst.SpwanAimCircle();
            this.camera = this.aimNode.getChildAt(0);
            this.camera.orthographic = true;
            this.camera.orthographicVerticalSize = 50;
            this.aimNode.getChildAt(1).active = false;
            GameManager.getInst().battleManager.aimMode = true;
            this.aimTips = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(_AllPrefabsNames.Arrow));
            MapComponent.inst.sp.addChild(this.aimTips);
            var ms = this.aimTips.getChildAt(0);
            this.mat = ms.meshRenderer.material;
            this.mat.albedoColor = Utils.hexToRgba01('#3579BC');
        }
        OnHide() {
            MsgManager.getInst().offAllCaller(this);
            this.aimNode.destroy();
            this.ui.m_bg.off(Laya.Event.MOUSE_DOWN, this, this.onStartMove);
            GameManager.getInst().battleManager.aimMode = false;
        }
        OnDispose() {
        }
        onStartMove(e) {
            if (this.touchID != null)
                return;
            try {
                this.touchID = e.touchId;
                Laya.timer.frameLoop(1, this, this.onMove);
                this.ui.on(Laya.Event.MOUSE_UP, this, this.onEndMove);
                this.ui.on(Laya.Event.MOUSE_OUT, this, this.onEndMove);
                this.point.setValue(Laya.stage.mouseX, Laya.stage.mouseY);
                this.camera.viewportPointToRay(this.point, this.ray);
                var oldx = this.ray.origin.x + this.ray.direction.x * 50;
                var oldZ = this.ray.origin.z + this.ray.direction.z * 50;
                this.aimTips.transform.position = new Laya.Vector3(oldx, 1, oldZ);
            }
            catch (err) {
                ReportMgr.inst.sendReport("Code_Error_航母 " + JSON.stringify(err), false);
            }
        }
        onMove() {
            try {
                var x = this.ray.origin.x + this.ray.direction.x * 50;
                var y = this.ray.origin.z + this.ray.direction.z * 50;
                this.point.setValue(Laya.stage.mouseX, Laya.stage.mouseY);
                this.camera.viewportPointToRay(this.point, this.ray);
                var dx = this.ray.origin.x + this.ray.direction.x * 50 - x;
                var dy = this.ray.origin.z + this.ray.direction.z * 50 - y;
                var d = 6 * LayaUtils.deltaTimeSec;
                if (Laya.stage.mouseX < this.left) {
                    this.aimNode.transform.position.x += d;
                }
                if (Laya.stage.mouseX > this.right) {
                    this.aimNode.transform.position.x -= d;
                }
                if (Laya.stage.mouseY < this.top) {
                    this.aimNode.transform.position.z += d;
                }
                if (Laya.stage.mouseY > this.bottom) {
                    this.aimNode.transform.position.z -= d;
                }
                this.aimNode.transform.position = this.aimNode.transform.position;
                var oldx = this.ray.origin.x + this.ray.direction.x * 50;
                var oldZ = this.ray.origin.z + this.ray.direction.z * 50;
                this.targetPos.setValue(oldx, 1, oldZ);
                var dis = Laya.Vector3.distance(this.aimTips.transform.position, this.targetPos);
                dis = Math.min(20, dis);
                this.aimTips.transform.localScaleZ = dis;
                this.aimTips.transform.lookAt(this.targetPos, LayaUtils.UP);
                this.aimTips.transform.localRotationEulerY += 180;
                this.consCheckRevers();
                if (this.revers) {
                    this.mat.tilingOffsetY = -dis;
                }
                else {
                    this.mat.tilingOffsetY = dis;
                }
            }
            catch (err) {
                ReportMgr.inst.sendReport("Code_Error_航母 " + JSON.stringify(err), false);
            }
        }
        consCheckRevers() {
            if (this.aimTips.transform.position.x > this.targetPos.x) {
                this.revers = true;
            }
            else {
                this.revers = false;
            }
        }
        onEndMove() {
            try {
                Laya.timer.clear(this, this.onMove);
                this.ui.off(Laya.Event.MOUSE_UP, this, this.onEndMove);
                this.ui.off(Laya.Event.MOUSE_OUT, this, this.onEndMove);
                this.touchID = null;
                UIManager.inst.PopUI(this.layer, false);
                var poses = [this.aimTips.transform.position.clone(), this.targetPos.clone()];
                poses.forEach(p => {
                    p.y = 10;
                });
                MsgManager.getInst().event(BattleEvent.FireHangMu, poses);
                ADData.HangMuCD = ADData.initCD;
                this.aimTips.destroy(true);
            }
            catch (err) {
                ReportMgr.inst.sendReport("Code_Error_航母 " + JSON.stringify(err), false);
            }
        }
    };
    HangMuAimUI = __decorate([
        ui_register()
    ], HangMuAimUI);

    class HangMuAni extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.fejis = [];
            this.time = 0;
        }
        onStart() {
            var paotai = this.sp.getChildAt(1);
            paotai.getChildAt(0).addComponent(RotateYYoYo);
            paotai.getChildAt(1).addComponent(RotateYYoYo);
            paotai.getChildAt(2).addComponent(RotateYYoYo);
            this.fejis.push(this.Find('jianzaiji2'));
            this.fejis.push(this.Find('jianzaiji1'));
            this.fejis.push(this.Find('jianzaiji3'));
            this.fejis.forEach(f => {
                f.active = false;
            });
        }
        onUpdate() {
            this.time -= LayaUtils.deltaTimeSec;
            if (this.time <= 0) {
                this.time = 4;
                for (var i = 0; i < this.fejis.length; i++) {
                    var f = this.fejis[i];
                    var plane = Laya.Sprite3D.instantiate(f);
                    plane.active = true;
                    this.sp.addChild(plane);
                    plane.transform.localRotation = f.transform.localRotation;
                    plane.transform.localPosition = f.transform.localPosition;
                    plane.transform.localScale = f.transform.localScale;
                    plane.addComponent(JianZaiJiAni).delay = i * 0.4;
                }
            }
        }
    }
    var tmp$7 = new Laya.Vector3();
    class JianZaiJiAni extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.startY = 0;
            this.upTime = 1;
            this.flyTime = 0;
            this.force = new Laya.Vector3(0, 0, 20);
            this.speed = new Laya.Vector3();
            this.delay = 0;
            this.oldDelay = 0;
        }
        onStart() {
            this.oldDelay = this.delay;
            this.oldPos = this.transform.localPosition.clone();
            this.oldRotate = this.transform.localRotationEuler.clone();
            this.startY = this.transform.localPositionY;
            this.init();
        }
        init() {
            this.upTime = 1;
            this.flyTime = 0;
            this.delay = this.oldDelay;
            this.transform.localPosition = this.oldPos;
            this.transform.localRotationEuler = this.oldRotate;
            this.transform.getForward(this.force);
            Laya.Vector3.normalize(this.force, this.force);
            Laya.Vector3.scale(this.force, 30, this.force);
            this.force.y += 6;
        }
        onUpdate() {
            this.delay -= LayaUtils.deltaTimeSec;
            if (this.delay > 0)
                return;
            this.upTime -= LayaUtils.deltaTimeSec;
            if (this.upTime > 0) {
                this.transform.localPositionY += LayaUtils.deltaTimeSec;
            }
            else {
                Laya.Vector3.scale(this.force, LayaUtils.deltaTimeSec, tmp$7);
                Laya.Vector3.add(tmp$7, this.speed, this.speed);
                Laya.Vector3.scale(this.speed, LayaUtils.deltaTimeSec, this.speed);
                Laya.Vector3.add(this.transform.position, tmp$7, tmp$7);
                this.transform.lookAt(tmp$7, LayaUtils.UP);
                this.transform.position = tmp$7;
                this.flyTime += LayaUtils.deltaTimeSec;
                if (this.flyTime > 2) {
                    this.sp.destroy(true);
                }
            }
        }
    }

    let HangMuJinDuUI = class HangMuJinDuUI extends BaseUI {
        get URL() {
            wxBinder.bindAll();
            return UI_Comp_hangmv2.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_loader, this.ui.m_zhezhao];
        }
        OnCreate() {
            this.ui.m_btn_fanhui.onClick(this, this.onClickClose);
            this.ui.m_btn_duanzao.onClick(this, this.onClickDuanZao);
        }
        OnShow(...args) {
            this.refresh();
            SoundManager.inst.playSound('Splash');
            this._scene3D = new SceneShow3D();
            this._scene3D.setRenderTexture(this.ui.m_loader.displayObject);
            PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.HangMuView).then(p => {
                var node = Laya.Sprite3D.instantiate(p);
                node.transform.localPositionY += 5;
                this._scene3D.playerNode.addChild(node);
                node.transform.localRotationEuler = new Laya.Vector3(9, -60, -2);
                var water = Utils.GetNode(node, 'Water');
                var mat = water.meshRenderer.material = new WaterWaveMaterial();
                mat.mainTexture = WaterMatHelper.getInst().water;
                mat.normalTexture = WaterMatHelper.getInst().noise;
                mat.tingOffset = new Laya.Vector4(80, 80, 0, 0);
                var hangmu = node.getChildAt(0);
                hangmu.addComponent(HangMuAni);
                hangmu.transform.localScale = new Laya.Vector3(0.6, 0.6, 0.6);
            });
            this.ui.m_add1.visible = false;
        }
        OnHide() {
            this._scene3D.destroy();
        }
        OnDispose() {
        }
        refresh() {
            this.ui.m_progressBar.value = PlayerData.HangMuNengLiang;
            this.ui.m_progressBar.max = 300;
            this.ui.m_progressV.text = PlayerData.HangMuNengLiang + '/300';
            this.ui.m_power.text = PlayerData.HangMuNengLiangNotUse + '';
        }
        onClickDuanZao() {
            if (PlayerData.HangMuNengLiangNotUse > 0) {
                this.ui.m_addOne.play();
                ReportMgr.inst.sendReport(ReportType.CLICK_DUANZAO);
                PlayerData.HangMuNengLiang++;
                PlayerData.HangMuNengLiangNotUse--;
                this.refresh();
                if (PlayerData.HangMuNengLiang >= 300) {
                    PlayerData.UnlockHangMu = true;
                    ReportMgr.inst.sendReport(ReportType.GET_HANGMU);
                    TipsMgr.Tips('恭喜解锁航母!');
                    this.onClickClose();
                }
            }
            else {
                TipsMgr.Tips('能量不足!');
            }
        }
        onClickClose() {
            UIManager.inst.PopUI(this.layer, false);
        }
    };
    __decorate([
        throttle(300)
    ], HangMuJinDuUI.prototype, "onClickDuanZao", null);
    HangMuJinDuUI = __decorate([
        ui_register()
    ], HangMuJinDuUI);

    let HangMuNengLiangUI = class HangMuNengLiangUI extends BaseUI {
        get URL() {
            wxBinder.bindAll();
            return UI_HangMuNengLiang.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_bg];
        }
        OnCreate() {
            this.ui.m_btn_fanhui.onClick(this, this.onClickClose);
            this.ui.m_comp_fengxiang.onClick(this, this.onClickClose);
        }
        OnShow(...args) {
            GameManager.getInst().adShowed = false;
        }
        OnHide() {
        }
        OnDispose() {
        }
        onClickClose() {
            UIManager.inst.PopUI(this.layer, true);
        }
    };
    HangMuNengLiangUI = __decorate([
        ui_register()
    ], HangMuNengLiangUI);

    class UI_ui_setting extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("setting", "ui_setting"));
        }
        onConstruct() {
            this.m_comp_mengban = (this.getChildAt(0));
            this.m_comp_shezhineirong = (this.getChildAt(1));
            this.m_btn_fanhui = (this.getChildAt(2));
            this.m_t0 = this.getTransitionAt(0);
        }
    }
    UI_ui_setting.URL = "ui://lzkirqqhc75x0";

    class UI_Native$1 extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("setting", "Native"));
        }
        onConstruct() {
            this.m_content = (this.getChildAt(0));
        }
    }
    UI_Native$1.URL = "ui://lzkirqqhn1ov1a";

    class UI_NativeAdBanner$1 extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("setting", "NativeAdBanner"));
        }
        onConstruct() {
            this.m_icon = (this.getChildAt(1));
            this.m_describle = (this.getChildAt(4));
            this.m_logo = (this.getChildAt(6));
            this.m_newLogo = (this.getChildAt(7));
        }
    }
    UI_NativeAdBanner$1.URL = "ui://lzkirqqhn1ov1b";

    class UI_comp_shezhi extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("setting", "comp_shezhi"));
        }
        onConstruct() {
            this.m_img_bg = (this.getChildAt(0));
            this.m_btn_yinxiao = (this.getChildAt(1));
            this.m_btn_yinyue = (this.getChildAt(2));
            this.m_btn_shock = (this.getChildAt(3));
            this.m_img_biaoti = (this.getChildAt(4));
            this.m_btn_yinsi = (this.getChildAt(5));
        }
    }
    UI_comp_shezhi.URL = "ui://lzkirqqhn8glj";

    class UI_btn_kongzhianniu extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("setting", "btn_kongzhianniu"));
        }
        onConstruct() {
            this.m_loader_kai = (this.getChildAt(0));
            this.m_loader_guan = (this.getChildAt(1));
        }
    }
    UI_btn_kongzhianniu.URL = "ui://lzkirqqhn8glk";

    class settingBinder {
        static bindAll() {
            fgui.UIObjectFactory.setExtension(UI_ui_setting.URL, UI_ui_setting);
            fgui.UIObjectFactory.setExtension(UI_Native$1.URL, UI_Native$1);
            fgui.UIObjectFactory.setExtension(UI_NativeAdBanner$1.URL, UI_NativeAdBanner$1);
            fgui.UIObjectFactory.setExtension(UI_comp_shezhi.URL, UI_comp_shezhi);
            fgui.UIObjectFactory.setExtension(UI_btn_kongzhianniu.URL, UI_btn_kongzhianniu);
        }
    }

    let SettingUI = class SettingUI extends BaseUI {
        constructor() {
            super(...arguments);
            this.c_comp_native = null;
        }
        get URL() {
            settingBinder.bindAll();
            return UI_ui_setting.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_comp_mengban];
        }
        OnCreate() {
            var p = this.ui.localToGlobal(this.ui.m_btn_fanhui.x, this.ui.m_btn_fanhui.y);
            p.x = 0;
            p = this.ui.globalToLocal(p.x, p.y, p);
            this.ui.m_btn_fanhui.x = p.x;
            this.ui.m_btn_fanhui.onClick(this, () => {
                UIManager.inst.PopUI(this.layer, false);
            });
            this.ui.m_comp_shezhineirong.m_btn_shock.onClick(this, this.onSettingChange);
            this.ui.m_comp_shezhineirong.m_btn_yinxiao.onClick(this, this.onSettingChange);
            this.ui.m_comp_shezhineirong.m_btn_yinyue.onClick(this, this.onMusiceChange);
        }
        clickNativeBannerView() {
            console.log("点击广告！");
            this.reportAdClick();
        }
        reportAdClick() {
        }
        onLoadSucced(res) {
            this.c_comp_native.visible = true;
            this.renderView(res);
        }
        onLoadErr() {
            console.log("原生广告加载失败了");
            this.c_comp_native.visible = false;
        }
        renderView(res) {
            console.log("原生数据==========================");
            console.log(JSON.stringify(res));
            let info = JSON.parse(JSON.stringify(res));
            let nativeInfo = info.adList[0];
            console.log(JSON.stringify(info));
            this.initNatveView(nativeInfo);
        }
        initNatveView(nativeInfo) {
            console.log(JSON.stringify(nativeInfo) + "<======================================>");
            nativeInfo = JSON.parse(JSON.stringify(nativeInfo));
            if (nativeInfo.imgUrlList && nativeInfo.imgUrlList.length > 0) {
                if (this.icon.url) {
                    console.log("存在=================》");
                }
                else {
                    console.log("不存在===============>");
                }
                this.icon.visible = true;
                console.log(nativeInfo.imgUrlList[0] + "网址");
                this.icon = new fairygui.GLoader();
                this.icon.setSize(300, 300);
                this.icon.url = nativeInfo.imgUrlList[0];
                this.newLog.url = "http://wxoss.jpsdk.com/ads/201906/10031560162475.png";
                console.log("777777777");
            }
            else if (nativeInfo.icon) {
                this.logo.visible = true;
                this.logo.url = nativeInfo.icon;
                console.log("kkkkk");
            }
            else ;
            this.describle.text = nativeInfo.title;
            this._useNtAdId = nativeInfo.adId;
        }
        OnShow(...args) {
            this.ui.m_comp_shezhineirong.m_btn_shock.selected = !SoundManager.inst.vibrate;
            this.ui.m_comp_shezhineirong.m_btn_yinxiao.selected = !SoundManager.inst.sound;
            this.ui.m_comp_shezhineirong.m_btn_yinyue.selected = !SoundManager.inst.music;
            this.ui.m_comp_shezhineirong.m_btn_yinsi.onClick(this, this.YinSIFun);
            let data = { closePos: 1, panelPos: 0 };
            GameSDK$1.GameNativeInsert && GameSDK$1.GameNativeInsert.load(data);
            GameSDK$1.hideBanner();
        }
        YinSIFun() {
            UIManager.inst.PushUI(UIYinSi, UILayer.mid, 0);
        }
        onSettingChange() {
            SoundManager.inst.vibrate = !this.ui.m_comp_shezhineirong.m_btn_shock.selected;
            SoundManager.inst.sound = !this.ui.m_comp_shezhineirong.m_btn_yinxiao.selected;
            MsgManager.getInst().event(EventType.Setting);
        }
        onMusiceChange() {
            SoundManager.inst.music = !this.ui.m_comp_shezhineirong.m_btn_yinyue.selected;
            MsgManager.getInst().event(EventType.Setting);
        }
        onLoadSuccesHandler() {
            this.c_comp_native.visible = true;
        }
        onLoadFailHandler() {
            console.log("test visiable==>>", this.c_comp_native);
            if (this.c_comp_native) ;
        }
        comNative() {
            console.log("test visiable==>>", this.c_comp_native);
            if (this.c_comp_native) {
                this.c_comp_native.visible = false;
            }
        }
        OnHide() {
            GameSDK$1.GameNativeInsert && GameSDK$1.GameNativeInsert.setIsShow(false);
            GameSDK$1.showBanner();
        }
        OnDispose() {
        }
    };
    SettingUI = __decorate([
        ui_register()
    ], SettingUI);

    let TouPiaoUI = class TouPiaoUI extends BaseUI {
        get URL() {
            wxBinder.bindAll();
            return UI_TouPiao.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_bg];
        }
        OnCreate() {
            var p = new Laya.Point();
            p = this.ui.globalToLocal(p.x, p.y, p);
            this.ui.m_btn_fanhui.x = p.x + this.ui.m_btn_fanhui.width / 2;
            this.ui.m_btn_fanhui.onClick(this, this.onClose);
            this.ui.m_view.m_haizhan1.onClick(this, this.onHaiZhan);
            this.ui.m_view.m_kongzhan1.onClick(this, this.onKongZhan);
            this.ui.m_view.m_duizhan1.onClick(this, this.onDuiZhan);
        }
        OnShow(...args) {
            AdMgr.inst.showBottomAd(true, false);
            GameManager.getInst().Pause();
        }
        OnHide() {
            AdMgr.inst.showBottomAd(false);
            GameManager.getInst().Resume();
        }
        OnDispose() {
        }
        onHaiZhan() {
            GameSDK.clickADVideo(this.ui.m_view.m_haizhan1, this, b => {
                if (b) {
                    this.onClose();
                    Laya.timer.once(500, this, () => {
                        TipsMgr.Tips('投票成功!');
                    });
                }
            }, '投票_海战');
        }
        onKongZhan() {
            GameSDK.clickADVideo(this.ui.m_view.m_kongzhan1, this, b => {
                if (b) {
                    this.onClose();
                    Laya.timer.once(500, this, () => {
                        TipsMgr.Tips('投票成功!');
                    });
                }
            }, '投票_空战');
        }
        onDuiZhan() {
            GameSDK.clickADVideo(this.ui.m_view.m_kongzhan1, this, b => {
                if (b) {
                    this.onClose();
                    Laya.timer.once(500, this, () => {
                        TipsMgr.Tips('投票成功!');
                    });
                }
            }, '投票_对战');
        }
        onClose() {
            UIManager.inst.PopUI(this.layer, true);
        }
    };
    TouPiaoUI = __decorate([
        ui_register()
    ], TouPiaoUI);

    let MainUI = class MainUI extends BaseUI {
        constructor() {
            super(...arguments);
            this.showCount = 0;
            this.attackBuffView = [];
            this.headPos = new Laya.Vector3();
            this.hpPos = new Laya.Vector4();
            this.hpPoint = new Laya.Point();
            this.tmpPoint = new Laya.Point();
            this.centerPos = new Laya.Point();
            this.currentPos = new Laya.Point();
            this.deltaPos = new Laya.Vector2();
            this.p = new Laya.Vector3();
            this.q = new Laya.Quaternion();
        }
        get URL() {
            return UI_ui_main.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_bg];
        }
        get LoaderFillType() {
            return fgui.LoaderFillType.Scale;
        }
        get VerticalAlign() {
            return LoaderVerticalAlign.middle;
        }
        OnCreate() {
            this.pool = new fgui.GObjectPool();
            this.ui.m_toupiao.m_tips.selectedIndex = 1;
            this.ui.m_btn_fight.onClick(this, () => {
                WxMgr.inst.battleStart = true;
                ReportMgr.inst.sendReport(ReportType.GUIDE_BATTLE_START);
                MsgManager.getInst().event(GameProcedureEvent.StartBattle);
            });
            this.ui.m_btn_gdyx.visible = false;
            this.ui.m_comp_zhiyin.visible = true;
            this.ui.m_bottomVideo.onClick(this, this._onClickBottomVideo);
            this.ui.m_toupiao.onClick(this, () => {
                this.ui.m_toupiao.m_tips.selectedIndex = 0;
                UIManager.inst.PushUI(TouPiaoUI, this.layer + 1);
            });
            this.ui.m_mingri.onClick(this, () => {
                AlertMgr.inst.showYuyue();
            });
            this.ui.m_btn_jijia.onClick(this, this.onClickMechaButton);
            this.ui.m_btn_try.m_tips.selectedIndex = 1;
            this.ui.m_btn_try.onClick(this, this.onClickHourseButton);
            this.ui.m_btn_hongzha.onClick(this, () => {
                if (ADData.Bomber > 0) {
                    GameSDK$1.clickADVideo(this.ui.m_btn_hongzha, this, b => {
                        if (b) {
                            ADData.Bomber = 0;
                        }
                        else {
                            TipsMgr.Tips('看完整视频得奖励');
                        }
                    }, '主页轰炸机');
                }
                else {
                    MsgManager.getInst().event(BattleEvent.CallBomber);
                    ADData.Bomber = ADData.initCD;
                }
            });
            this.ui.m_btn_feiji.onClick(this, () => {
                GameSDK$1.clickADVideo(this.ui.m_btn_feiji, this, b => {
                    if (b) {
                        ADData.Helicopter = ADData.initCD;
                        SoundManager.inst.playVibrate(true);
                        UIManager.inst.PushUI(AimMenu, this.layer, BattleEvent.CallHelicopter, GameManager.HelicopterRadius);
                    }
                    else {
                        TipsMgr.Tips('看完整视频得奖励');
                    }
                }, '直升机');
            });
            this.ui.m_btn_kongtou.onClick(this, () => {
                if (ADData.GoldDropBox > 0) {
                    GameSDK$1.clickADVideo(this.ui.m_btn_kongtou, this, b => {
                        if (b) {
                            ADData.GoldDropBox = 0;
                        }
                        else {
                            TipsMgr.Tips('看完整视频得奖励');
                        }
                    }, '金空投');
                }
                else {
                    ADData.GoldDropBox = ADData.initCD;
                    SoundManager.inst.playVibrate(true);
                    MsgManager.getInst().event(BattleEvent.CallDropBox);
                }
            });
            this.ui.m_btn_chuan.onClick(this, () => {
                if (ADData.Ship > 0) {
                    GameSDK$1.clickADVideo(this.ui.m_btn_chuan, this, b => {
                        if (b) {
                            ADData.Ship = 0;
                        }
                        else {
                            TipsMgr.Tips('看完整视频得奖励');
                        }
                    }, '巡洋舰');
                }
                else {
                    ADData.Ship = ADData.initCD;
                    SoundManager.inst.playVibrate(true);
                    UIManager.inst.PushUI(AimMenu, this.layer, BattleEvent.FireShip, GameManager.ShipRadius);
                }
            });
            this.ui.m_btn_gdyx.onClick(this, () => {
                AdMgr.inst.showRectAd();
            });
            this.ui.m_btn_hedan.onClick(this, () => {
                if (PlayerData.FreeNuke > 0) {
                    if (GuideData.Step == GuideStepType.使用免费核弹) {
                        GuideData.Step = GuideStepType.战斗中_隐藏Tips;
                    }
                    ReportMgr.inst.sendReport(ReportType.NUKE_USE_FREE);
                    PlayerData.FreeNuke--;
                    ADData.nukeCD = ADData.initCD;
                    SoundManager.inst.playVibrate(true);
                    UIManager.inst.PushUI(AimMenu, this.layer, BattleEvent.FireBomb, GameManager.NukeRadius);
                    this._refreshNukeGuide();
                    return;
                }
                if (PlayerData.NormalNuke) {
                    if (ADData.nukeCD > 0) {
                        GameSDK$1.clickADVideo(this.ui.m_btn_hedan, this, b => {
                            if (b) {
                                ADData.nukeCD = 0;
                            }
                            else {
                                TipsMgr.Tips('看完整视频得奖励');
                            }
                        }, '主页核弹');
                    }
                    else {
                        ADData.nukeCD = ADData.initCD;
                        SoundManager.inst.playVibrate(true);
                        UIManager.inst.PushUI(AimMenu, this.layer, BattleEvent.FireBomb, GameManager.NukeRadius);
                    }
                }
                else {
                    GameSDK$1.clickADVideo(this.ui.m_btn_hedan, this, b => {
                        if (b) {
                            ADData.nukeCD = ADData.initCD;
                            SoundManager.inst.playVibrate(true);
                            UIManager.inst.PushUI(AimMenu, this.layer, BattleEvent.FireBomb, GameManager.NukeRadius);
                        }
                        else {
                            TipsMgr.Tips('看完整视频得奖励');
                        }
                    }, '主页核弹');
                }
            });
            this.ui.m_btn_setting.onClick(this, () => {
                UIManager.inst.PushUI(SettingUI, this.layer);
            });
            this.ui.m_btn_zhanjian.onClick(this, () => {
                ReportMgr.inst.sendReport(ReportType.CLICK_HANGMU);
                if (PlayerData.UnlockHangMu) {
                    if (ADData.HangMuCD > 0) {
                        GameSDK$1.clickADVideo(this.ui.m_btn_zhanjian, this, b => {
                            if (b) {
                                UIManager.inst.PushUI(HangMuAimUI, this.layer);
                            }
                        }, '航母');
                    }
                    else {
                        UIManager.inst.PushUI(HangMuAimUI, this.layer);
                    }
                }
                else {
                    if (PlayerData.CurrentLevelID < GameManager.HangMuUnlockScene) {
                        TipsMgr.Tips('战斗航母第三关解锁');
                    }
                    else {
                        UIManager.inst.PushUI(HangMuJinDuUI, this.layer);
                    }
                }
            });
            var hScale = Laya.stage.height / Utils.safeArea.height;
            var rect = this.ui.localToGlobalRect(0, 0, this.ui.width, this.ui.height);
            var p = this.ui.globalToLocal(0, Utils.safeArea.top * hScale * rect.height / this.ui.height);
            this.ui.m_top.y = p.y + 40;
            var p = this.ui.globalToLocal(0, 0, p);
            this.ui.m_left.x = p.x;
            var p = this.ui.globalToLocal(Laya.stage.width, 0, p);
            var delat = this.ui.width - this.ui.m_right.x;
            this.ui.m_right.x = p.x - delat;
        }
        startBottomVideoLoop() {
            this.ui.m_bottomVideo.visible = false;
        }
        _refreshBottomVideo() {
            if (PlayerData.HorseAdCount < 5) {
                let random = Math.floor(Math.random() * 2);
                this.ui.m_bottomVideo.m_comp_video.m_c1.selectedIndex = random;
            }
            else {
                this.ui.m_bottomVideo.m_comp_video.m_c1.selectedIndex = 1;
            }
            Laya.timer.loop(30 * 1000, this, this._refreshBottomVideo);
        }
        _onClickBottomVideo() {
            GameSDK$1.clickADVideo(this.ui, this, (s) => {
                if (s) {
                    switch (this.ui.m_bottomVideo.m_comp_video.m_c1.selectedIndex) {
                        case 0:
                            MsgManager.getInst().event(BattleEvent.MountHorse);
                            PlayerData.HorseAdCount++;
                            if (!GameSDK$1.isLimit()) {
                                if (PlayerData.HorseAdCount >= 5) {
                                    ReportMgr.inst.sendReport(ReportType.GET_CAERRIER);
                                    TipsMgr.Tips('已永久解锁战马!');
                                    PlayerData.mountsId = 2;
                                }
                                else {
                                    TipsMgr.Tips(`再使用${5 - PlayerData.HorseAdCount}次战马后,将永久解锁战马!`);
                                    PlayerData.mountsId = 1;
                                }
                            }
                            else {
                                PlayerData.mountsId = 1;
                            }
                            break;
                        case 1:
                            SoundManager.inst.playVibrate(true);
                            UIManager.inst.PushUI(AimMenu, this.layer, BattleEvent.FireBomb, GameManager.NukeRadius);
                            break;
                    }
                    Laya.timer.clear(this, this._refreshBottomVideo);
                    this._refreshBottomVideo();
                }
            }, this.ui.m_bottomVideo.m_comp_video.m_c1.selectedIndex == 0 ? "bottom_战马视频" : "bottom_核弹视频");
        }
        showBottomVideo() {
            try {
                if (GameSDK$1.isLimit()) {
                    this.ui.m_bottomVideo.visible = false;
                    return;
                }
                ;
                this.ui.m_bottomVideo.visible = WxMgr.inst.battleStart && this.getShowWinLen() == 1;
            }
            catch (e) {
            }
        }
        getShowWinLen() {
            let count = 0;
            let temp;
            for (let i in UILayer) {
                temp = UIManager.inst.uiLayarMap[UILayer[i]];
                if (!!temp) {
                    for (let j in temp) {
                        if (temp[j] instanceof TipsUI) {
                            continue;
                        }
                        if (temp[j].show == true) {
                            count++;
                        }
                    }
                }
            }
            return count;
        }
        _refreshNukeGuide() {
            this.ui.m_btn_hedan.m_c2.selectedIndex = PlayerData.FreeNuke > 0 ? 1 : 0;
        }
        onFeijiShowEnd() {
            this.ui.m_btn_feiji.visible = false;
        }
        updateTimer() {
            var cd = ADData.GoldDropBox;
            this.ui.m_btn_kongtou.m_c1.selectedIndex = cd <= 0 ? 0 : 1;
            this.ui.m_btn_kongtou.m_cd.text = this.TimeToString(cd);
            cd = ADData.MechaCD;
            this.ui.m_btn_jijia.m_c1.selectedIndex = cd <= 0 ? 0 : 1;
            this.ui.m_btn_jijia.m_cd.text = this.TimeToString(cd);
            cd = ADData.Bomber;
            this.ui.m_btn_hongzha.m_c1.selectedIndex = cd <= 0 ? 0 : 1;
            this.ui.m_btn_hongzha.m_cd.text = this.TimeToString(cd);
            cd = ADData.Ship;
            this.ui.m_btn_chuan.m_c1.selectedIndex = cd <= 0 ? 0 : 1;
            this.ui.m_btn_chuan.m_cd.text = this.TimeToString(cd);
            cd = ADData.HorseCD;
            this.ui.m_btn_try.m_cd.selectedIndex = (PlayerData.HorseAdCount >= 5 && cd > 0) ? 1 : 0;
            this.ui.m_btn_try.m_cd_2.text = this.TimeToString(cd);
            if (PlayerData.UnlockHangMu) {
                cd = ADData.HangMuCD;
                this.ui.m_btn_zhanjian.m_cd.selectedIndex = cd > 0 ? 1 : 0;
                this.ui.m_btn_zhanjian.m_tips.selectedIndex = cd > 0 ? 0 : 1;
                this.ui.m_btn_zhanjian.m_c1.selectedIndex = 0;
                this.ui.m_btn_zhanjian.m_cd_2.text = this.TimeToString(cd);
            }
            else {
                this.ui.m_btn_zhanjian.m_tips.selectedIndex = PlayerData.HangMuNengLiangNotUse > 0 ? 1 : 0;
                this.ui.m_btn_zhanjian.m_c1.selectedIndex = 1;
            }
            this.ui.m_btn_zhanjian.m_c1.selectedIndex = PlayerData.CurrentLevelID < GameManager.HangMuUnlockScene ? 1 : 0;
            if (PlayerData.FreeNuke > 0) {
                this.ui.m_btn_hedan.m_c1.selectedIndex = 0;
            }
            else {
                if (PlayerData.NormalNuke) {
                    cd = ADData.nukeCD;
                    this.ui.m_btn_hedan.m_c1.selectedIndex = cd <= 0 ? 0 : 1;
                    this.ui.m_btn_hedan.m_cd.text = this.TimeToString(cd);
                }
                else {
                    this.ui.m_btn_hedan.m_c1.selectedIndex = 3;
                }
            }
        }
        TimeToString(mil) {
            var seconds = Math.floor(mil / 1000);
            var s = seconds % 60;
            var min = Math.floor(seconds / 60) % 60;
            return Utils.formatNumber(min) + ':' + Utils.formatNumber(s);
        }
        updateAdBtns(forceHide = true) {
            this.ui.m_btn_kongtou.visible = true;
            var rank = PlayerData.CurrentRankLevel;
            this.ui.m_btn_hongzha.visible = !forceHide && (PlayerData.CurrentLevelID > 1) && (BattleData.Airport.level > 0);
            this.ui.m_btn_chuan.visible = !forceHide && rank >= 6;
            this.ui.m_btn_hedan.visible = !forceHide && PlayerData.UnlockNuke;
            this.ui.m_btn_try.visible = !forceHide && PlayerData.CurrentLevelID > 1;
            this.ui.m_toupiao.visible = !forceHide && PlayerData.CurrentLevelID > 1;
            this.ui.m_btn_jijia.visible = !forceHide && PlayerData.CurrentLevelID > 1;
            this.ui.m_btn_jijia.visible = false;
        }
        OnShow(...args) {
            this.refreshHorseButton();
            Laya.timer.frameLoop(1, this, this.refreshHorseButton);
            Laya.timer.frameOnce(2, this, this.refreshHangMu);
            this.ui.m_bg.on(Laya.Event.MOUSE_DOWN, this, this.onMoveStart);
            this.ui.m_joyStick.visible = false;
            this.ui.m_c1.selectedPage = 'main';
            this.ui.m_mechaTimer.visible = false;
            if (!PlayerData.NormalNuke && PlayerData.YuyueNuke) {
                this.ui.m_mingri.visible = true;
                if (Date.now() > PlayerData.YuyueDate) {
                    this.ui.m_mingri.m_tips.selectedIndex = 1;
                }
                else {
                    this.ui.m_mingri.m_tips.selectedIndex = 0;
                }
            }
            else {
                this.ui.m_mingri.visible = false;
            }
            Laya.timer.loop(1000, this, this.updateNukeYuyue);
            MsgManager.getInst().on(BattleEvent.ShowAttackUPBar, this, this.onShowAttackUPBar);
            MsgManager.getInst().on(BattleEvent.HideAttackUPBar, this, this.onHideAttackUPBar);
            MsgManager.getInst().on(BattleEvent.EnemyCountInfo, this, this.onEnemyCountChange);
            MsgManager.getInst().on(BattleEvent.EnemyLoadEnd, this, this.onEnemyLoadEnd);
            MsgManager.getInst().on(BattleEvent.UserTagChange, this, this.onUpdateTag);
            MsgManager.getInst().on(BattleEvent.SetGuideStep, this, this.onGuideStep);
            MsgManager.getInst().on(BattleEvent.RankLevelUp, this, this.onRankUP);
            MsgManager.getInst().on(BattleEvent.UpdatePlayerPos, this, this.onPlayerPos);
            MsgManager.getInst().on(BattleEvent.UpdateMechaFreeTime, this, this.onMechaFreeTime);
            MsgManager.getInst().on(BattleEvent.TriggerFreeNuke, this, this.onAddFreeNuke);
            MsgManager.getInst().on(BattleEvent.TriggerFeiJi, this, () => {
                if (PlayerData.CurrentLevelID == 1)
                    return;
                console.log('打电话');
                this.ui.m_btn_feiji.visible = true;
                this.ui.m_btn_feiji.m_show.play();
                this.ui.m_btn_feiji.m_show.setHook('end', Laya.Handler.create(this, this.onFeijiShowEnd));
                Utils.WaitTime(250).then(() => {
                    var cd = 10;
                    Utils.WaitWithBlock(() => {
                        cd -= LayaUtils.deltaTimeSec;
                        this.ui.m_btn_feiji.m_progress.value = Math.round(cd / 10 * 100);
                        return cd <= 0;
                    });
                });
            });
            MsgManager.getInst().on(BattleEvent.UnlockFirstSoilderTent, this, () => {
                this.updateAdBtns(false);
            });
            this.ui.m_rankTips.visible = false;
            GameSDK$1.GameNativeInsert && GameSDK$1.GameNativeInsert.setIsShow(false);
            MsgManager.getInst().on(BattleEvent.HideMainUI, this, () => {
                this.ui.visible = false;
            });
            MsgManager.getInst().on(BattleEvent.ShowMainUI, this, () => {
                this.ui.visible = true;
            });
            MsgManager.getInst().on(BattleEvent.TurnOffBattle, this, () => {
                this.ui.m_c1.selectedPage = 'main';
            });
            MsgManager.getInst().on(BattleEvent.TurnOnBattle, this, () => {
                this.ui.m_c1.selectedPage = 'fight';
            });
            MsgManager.getInst().on(BattleEvent.PlaneStart, this, () => {
                this.ui.m_c1.selectedPage = 'plane';
            });
            MsgManager.getInst().on(BattleEvent.PlaneEnd, this, () => {
                this.ui.m_c1.selectedPage = 'main';
            });
            MsgManager.getInst().on(BattleEvent.FireBombStart, this, () => {
                this.ui.m_c1.selectedPage = 'nuke';
                this.ui.touchable = false;
            });
            MsgManager.getInst().on(BattleEvent.FireBombEnd, this, () => {
                this.ui.m_c1.selectedPage = 'main';
                this.ui.touchable = true;
            });
            MsgManager.getInst().on(BattleEvent.UpdatePlaneBullet, this, (b) => {
                this.ui.m_comp_zidan.m_text_shuliang.text = b + '';
            });
            MsgManager.getInst().on(BattleEvent.MountHorse, this, () => {
            });
            MsgManager.getInst().on(BattleEvent.MountHorseEnd, this, () => {
            });
            MsgManager.getInst().on(BattleEvent.MountMecha, this, () => {
            });
            MsgManager.getInst().on(BattleEvent.MountMechaEnd, this, () => {
            });
            this.onUpdateTag();
            var allSceneCount = PlayerData.CurrentLevelBattleCount;
            var current = PlayerData.CurrentLevelBattleIndex;
            var flags = [this.ui.m_comp_guanqiajindu.m_comp_qizhi1, this.ui.m_comp_guanqiajindu.m_comp_qizhi2, this.ui.m_comp_guanqiajindu.m_comp_qizhi3, this.ui.m_comp_guanqiajindu.m_comp_qizhi4];
            flags.forEach(flag => {
                flag.m_c1.selectedIndex = 0;
            });
            if (allSceneCount > 1) {
                for (var i = 1; i <= current; i++) {
                    flags[i - 1].m_c1.selectedIndex = 1;
                }
            }
            var tent = BattleData.GetTent('TentLocation1');
            if (tent == null || tent.level == 0) {
                this.updateAdBtns(true);
            }
            else {
                this.updateAdBtns(false);
            }
            this.ui.m_hp.m_show.selectedIndex = 1;
            this.ui.m_hpHorse.m_show.selectedIndex = 1;
            Laya.timer.frameLoop(1, this, this.updateTimer);
            this.ui.m_btn_feiji.visible = false;
            this.ui.m_bottomVideo.visible = false;
            this.startBottomVideoLoop();
            Laya.timer.once(1000 * 60, this, () => {
                GameSDK$1.showBanner();
            });
        }
        updateBanner() {
            GameSDK$1.showBanner();
        }
        updateNukeYuyue() {
            if (!PlayerData.NormalNuke && PlayerData.YuyueNuke) {
                this.ui.m_mingri.visible = true;
                this.ui.m_mingri.m_tips.selectedIndex = 1;
            }
            else {
                this.ui.m_mingri.visible = false;
            }
            this._refreshNukeGuide();
        }
        showNativeInsertAd() {
            this.showCount++;
            console.log("人数升级===============》");
            let closePos = this.showCount % 2;
            let data = { closePos: closePos, panelPos: 1 };
            GameSDK$1.GameNativeInsert && GameSDK$1.GameNativeInsert.load(data, true);
        }
        onRankUP() {
            this.updateAdBtns(false);
            this.showNativeInsertAd();
            Utils.WaitTime(3000).then(() => {
                this.ui.m_rankTips.visible = true;
            });
            AdMgr.inst.showLeftAd(true);
            AdMgr.inst.showRightAd(true);
            AdMgr.inst.showBottomAd(true, false);
            Utils.WaitTime(5000).then(() => {
                AdMgr.inst.showLeftAd(false);
                AdMgr.inst.showRightAd(false);
                AdMgr.inst.showBottomAd(false);
                this.ui.m_rankTips.visible = false;
            });
            var icons = PrefabManager.getInst().GetPrefab(_AllPrefabsNames.RankIcons);
            var iconPre = icons.getChildByName((PlayerData.CurrentCommanderConfig.level - 1) + '').meshRenderer.material.albedoTexture.url;
            var icon = icons.getChildByName(PlayerData.CurrentCommanderConfig.level + '').meshRenderer.material.albedoTexture.url;
            this.ui.m_rankTips.m_aniView.m_rank.m_current.url = icon;
            this.ui.m_rankTips.m_aniView.m_rank.m_pre.url = iconPre;
            this.ui.m_rankTips.m_aniView.m_rank.m_show.play();
            this.ui.m_rankTips.m_aniView.m_last.text = ConfigHelper.CommanderConfigLevelMap[PlayerData.CurrentCommanderConfig.level - 1].name;
            this.ui.m_rankTips.m_aniView.m_current.text = PlayerData.CurrentCommanderConfig.name;
            this.ui.m_rankTips.m_aniView.m_extSlot.text = PlayerData.CurrentCommanderConfig.extSlot + '';
        }
        OnHide() {
            this.ui.m_bg.off(Laya.Event.MOUSE_MOVE, this, this.onMove);
            this.ui.m_bg.off(Laya.Event.MOUSE_UP, this, this.onMoveEnd);
            this.ui.off(Laya.Event.MOUSE_OUT, this, this.onMoveEnd);
            this.onMoveEnd();
            MsgManager.getInst().offAllCaller(this);
            Laya.timer.clear(this, this.updateNukeYuyue);
            Laya.timer.clear(this, this.refreshHorseButton);
            Laya.timer.clear(this, this.refreshHangMu);
            Laya.timer.clear(this, this.showBottomVideo);
            Laya.timer.clear(this, this._refreshBottomVideo);
            if (!GameSDK$1.isCurrentVersion) {
                Laya.timer.clear(this, this.updateBanner);
            }
        }
        OnDispose() {
        }
        onGuideStep(step, open = true) {
            this.ui.m_guide_step.selectedIndex = open ? 1 : 0;
            this.ui.m_guide.m_c1.selectedIndex = step;
        }
        onEnemyLoadEnd() {
            this.ui.m_comp_guanqiajindu.visible = true;
        }
        onUpdateTag() {
            this.ui.m_comp_chaopiao2.title = BattleData.SilverTag + '';
            this.ui.m_comp_chaopiao1.title = BattleData.GoldTag + '';
            if (GuideData.Step == GuideStepType.升级提示) {
                if (ArrowTips.inst && BattleData.GoldTag >= PlayerData.CurrentCommanderConfig.cost_gold) {
                    ArrowTips.inst.target = ArrowTips.inst.target = MapCampManagerComponent.inst.GetSlotNode('PlayerLocation');
                }
            }
        }
        onEnemyCountChange(count, allCount) {
            var allSceneCount = PlayerData.CurrentLevelBattleCount;
            var current = PlayerData.CurrentLevelBattleIndex;
            this.ui.m_comp_guanqiajindu.m_levelCount.selectedPage = allSceneCount + '';
            var step = 1 / allSceneCount;
            this.ui.m_comp_guanqiajindu.max = 100;
            var pv = Math.floor(((allCount - count) / allCount * step + current * step) * 100);
            this.ui.m_comp_guanqiajindu.tweenValue(pv, 0.1);
        }
        onPlayerPos(pos, entity) {
            this.headPos.setValue(pos.x, pos.y + 3, pos.z);
            LevelManager.getInst().mainCamera.worldToViewportPoint(this.headPos, this.hpPos);
            var p = this.ui.globalToLocal(this.hpPos.x, this.hpPos.y, this.tmpPoint);
            this.ui.m_hp.x = Laya.MathUtil.lerp(this.ui.m_hp.x, p.x, 0.8);
            this.ui.m_hp.y = Laya.MathUtil.lerp(this.ui.m_hp.y, p.y, 0.8);
            this.ui.m_mechaTimer.x = Laya.MathUtil.lerp(this.ui.m_mechaTimer.x, p.x, 0.8);
            this.ui.m_mechaTimer.y = Laya.MathUtil.lerp(this.ui.m_mechaTimer.y, p.y, 0.8);
            this.ui.m_hpHorse.x = Laya.MathUtil.lerp(this.ui.m_hpHorse.x, p.x, 0.8);
            this.ui.m_hpHorse.y = Laya.MathUtil.lerp(this.ui.m_hpHorse.y, p.y - this.ui.m_hp.height, 0.8);
            this.ui.m_hpHorse.visible = entity.mount;
            var hp = entity.shieldHP;
            var fullHP = entity.shieldFullHP;
            if (this.ui.m_hpHorse.value > hp) {
                this.showHP();
            }
            else if (hp == fullHP) {
                this.hideHP();
            }
            this.ui.m_hpHorse.max = fullHP;
            this.ui.m_hpHorse.value = hp;
            hp = entity.hp;
            fullHP = entity.fullHP;
            if (this.ui.m_hp.value > hp) {
                this.showHP();
            }
            else if (hp == fullHP && entity.shieldHP == entity.shieldFullHP) {
                this.hideHP();
            }
            this.ui.m_hp.max = fullHP;
            this.ui.m_hp.value = hp;
        }
        onMechaFreeTime(pos, t) {
            this.ui.m_mechaTimer.visible = t > 0;
            this.ui.m_mechaTimer.title = t.toFixed();
        }
        onAddFreeNuke(pos) {
            var p = new Laya.Vector4();
            SimpleCameraFollow.inst.camera.worldToViewportPoint(pos, p);
            var point = this.ui.globalToLocal(p.x, p.y);
            var icon = UI_FreeNuke.createInstance();
            this.ui.addChild(icon);
            icon.x = point.x;
            icon.y = point.y;
            Laya.Tween.to(icon, { x: this.ui.m_btn_hedan.x + this.ui.m_btn_hedan.width / 2, y: this.ui.m_btn_hedan.y + this.ui.m_btn_hedan.height / 2 }, 1000, Laya.Ease.linearIn, Laya.Handler.create(this, () => {
                icon.dispose();
            }));
        }
        onShowAttackUPBar(entity) {
            var bar = entity.aniView.attackUpBar;
            if (bar == null) {
                bar = this.pool.getObject(UI_AttackUp.URL).asCom;
            }
            entity.aniView.attackUpBar = bar;
            this.ui.addChild(bar);
            bar.sortingOrder = 0;
            entity.aniView.transform.position.cloneTo(this.p);
            this.p.y += 2.5;
            LevelManager.getInst().mainCamera.worldToViewportPoint(this.p, this.hpPos);
            var p = this.ui.globalToLocal(this.hpPos.x, this.hpPos.y, this.tmpPoint);
            bar.x = p.x;
            bar.y = p.y;
        }
        onHideAttackUPBar(entity) {
            var bar = entity.aniView.attackUpBar;
            if (bar == null)
                return;
            this.ui.removeChild(bar);
            this.pool.returnObject(bar);
            entity.aniView.attackUpBar = null;
        }
        showHP() {
            this.ui.m_hp.m_show.selectedIndex = 1;
            this.ui.m_hpHorse.m_show.selectedIndex = 1;
            this.ui.m_hp.alpha = 1;
            this.ui.m_hpHorse.alpha = 1;
        }
        hideHP() {
            if (this.ui.m_hp.m_show.selectedIndex == 1) {
                this.ui.m_hp.m_show.selectedIndex = 0;
                this.ui.m_hpHorse.m_show.selectedIndex = 0;
            }
        }
        onClickMechaButton() {
            this._onClickMecha();
        }
        _onClickMecha() {
            if (PlayerData.mountsId == 3 || PlayerData.mountsId == 4) {
                TipsMgr.Tips("当前已有机甲");
                return;
            }
            if (PlayerData.UnlockMecha) {
                GameSDK$1.showADVideo(this, s => {
                    if (s) {
                        PlayerData.mountsId = 4;
                        MsgManager.getInst().event(BattleEvent.MountMecha);
                        if (!!GameManager.getInst().battleManager.mechaBox) {
                            GameManager.getInst().battleManager.mechaBox.onClickJijia();
                        }
                    }
                }, "");
            }
        }
        onClickHourseButton() {
            if (PlayerData.mountsId == 1 || PlayerData.mountsId == 2) {
                TipsMgr.Tips("当前已有战马");
                return;
            }
            if (PlayerData.HorseAdCount < 5) {
                this.ui.m_btn_try.m_tips.selectedIndex = 0;
                GameSDK$1.clickADVideo(this.ui.m_btn_try, this, b => {
                    if (b) {
                        MsgManager.getInst().event(BattleEvent.MountHorse);
                        PlayerData.HorseAdCount++;
                        ReportMgr.inst.sendReport(ReportType.TRY_CAERRIER_VIDEO + PlayerData.HorseAdCount);
                        if (!GameSDK$1.isLimit()) {
                            if (PlayerData.HorseAdCount >= 5) {
                                ReportMgr.inst.sendReport(ReportType.GET_CAERRIER);
                                TipsMgr.Tips('已永久解锁战马!');
                                PlayerData.mountsId = 2;
                            }
                            else {
                                TipsMgr.Tips(`再使用${5 - PlayerData.HorseAdCount}次战马后,将永久解锁战马!`);
                                PlayerData.mountsId = 1;
                            }
                        }
                        else {
                            PlayerData.mountsId = 1;
                        }
                    }
                }, '骑马');
            }
            else {
                if (ADData.HorseCD > 0) {
                    TipsMgr.Tips('战马冷却中');
                    return;
                }
                PlayerData.mountsId = 2;
                MsgManager.getInst().event(BattleEvent.MountHorse);
            }
        }
        refreshHorseButton() {
            if (PlayerData.HorseAdCount >= 5) {
                this.ui.m_btn_try.m_ad.selectedIndex = 0;
                this.ui.m_btn_try.m_proogress_zhanma.value = ADData.HorseCD / 60000 * 100;
            }
            else {
                this.ui.m_btn_try.m_ad.selectedIndex = 1;
                this.ui.m_btn_try.m_proogress_zhanma.value = 0;
            }
            if (GameSDK$1.isLimit() || PlayerData.HorseAdCount == 0 || PlayerData.HorseAdCount == 5) {
                this.ui.m_btn_try.m_adProgress.text = '';
                this.ui.m_btn_try.m_yongjiu.selectedIndex = 1;
            }
            else {
                this.ui.m_btn_try.m_yongjiu.selectedIndex = 0;
            }
            this.ui.m_btn_try.m_adProgress.text = `${5 - PlayerData.HorseAdCount}`;
        }
        refreshHangMu() {
            if (!PlayerData.UnlockHangMu && this.ui.visible) {
                if (GameManager.getInst().adShowed && PlayerData.CurrentLevelID >= GameManager.HangMuUnlockScene && SimpleCameraFollow.inst.target == null && PlayerData.HangMuNengLiangToday < 100) {
                    if (PlayerData.HangMuNengLiang == 0 && PlayerData.HangMuNengLiangNotUse == 0 && PlayerData.HangMuNengLiangToday == 0) {
                        UIManager.inst.PushUI(HangMuNengLiangUI, UILayer.mid);
                    }
                    else {
                        this.ui.m_btn_zhanjian.m_fly.selectedIndex = 1;
                        Utils.WaitTime(2500).then(() => {
                            if (!this.ui.isDisposed) {
                                this.ui.m_btn_zhanjian.m_fly.selectedIndex = 0;
                            }
                        });
                    }
                    GameManager.getInst().adShowed = false;
                    PlayerData.HangMuNengLiangToday += 10;
                    PlayerData.HangMuNengLiangNotUse += 10;
                    TipsMgr.Tips('航母能量+10');
                }
            }
            Laya.timer.frameOnce(1, this, this.refreshHangMu);
        }
        onMoveStart() {
            if (this.ui.touchable == false || GameManager.getInst().battleManager.aimMode || SimpleCameraFollow.inst.target != null)
                return;
            this.centerX = Laya.stage.mouseX;
            this.centerY = Laya.stage.mouseY;
            this.ui.globalToLocal(Laya.stage.mouseX, Laya.stage.mouseY, this.centerPos);
            this.ui.m_joyStick.x = this.centerPos.x;
            this.ui.m_joyStick.y = this.centerPos.y;
            this.ui.m_joyStick.m_holder.x = this.ui.m_joyStick.width / 2;
            this.ui.m_joyStick.m_holder.y = this.ui.m_joyStick.height / 2;
            this.ui.m_bg.on(Laya.Event.MOUSE_MOVE, this, this.onMove);
            this.ui.m_bg.on(Laya.Event.MOUSE_UP, this, this.onMoveEnd);
            this.ui.on(Laya.Event.MOUSE_OUT, this, this.onMoveEnd);
        }
        onMove() {
            if (this.ui.touchable == false || GameManager.getInst().battleManager.aimMode || SimpleCameraFollow.inst.target != null)
                return;
            if (this.ui.m_comp_zhiyin.visible) {
                this.ui.m_comp_zhiyin.visible = false;
            }
            if (this.ui.m_joyStick.visible == false) {
                this.ui.m_joyStick.visible = true;
            }
            this.ui.globalToLocal(Laya.stage.mouseX, Laya.stage.mouseY, this.currentPos);
            this.deltaPos.x = this.currentPos.x - this.centerPos.x;
            this.deltaPos.y = this.currentPos.y - this.centerPos.y;
            LayaUtils.truncateVec2(this.deltaPos, 100, this.deltaPos);
            this.ui.m_joyStick.m_holder.x = this.deltaPos.x + this.ui.m_joyStick.width / 2;
            this.ui.m_joyStick.m_holder.y = this.deltaPos.y + this.ui.m_joyStick.height / 2;
            if (LevelManager.getInst().mainCameraRotRoot != null) {
                var r = LevelManager.getInst().mainCameraRotRoot.transform.rotationEuler.y * Laya.MathUtils3D.Deg2Rad;
                var p = this.p;
                p.setValue(this.deltaPos.x, 0, this.deltaPos.y);
                var q = this.q;
                Laya.Quaternion.createFromAxisAngle(LayaUtils.UP, r, q);
                Laya.Vector3.transformQuat(p, q, p);
                this.deltaPos.setValue(p.x, p.z);
            }
            LayaUtils.normalizeVec2(this.deltaPos, this.deltaPos);
            MsgManager.getInst().event(CtrlEvent.PlayerSteerMove, this.deltaPos);
        }
        onMoveEnd() {
            this.ui.m_joyStick.visible = false;
            MsgManager.getInst().event(CtrlEvent.PlayerSteerEnd);
            this.ui.m_bg.off(Laya.Event.MOUSE_MOVE, this, this.onMove);
            this.ui.m_bg.off(Laya.Event.MOUSE_UP, this, this.onMoveEnd);
            this.ui.off(Laya.Event.MOUSE_OUT, this, this.onMoveEnd);
        }
    };
    __decorate([
        throttle(3000)
    ], MainUI.prototype, "hideHP", null);
    MainUI = __decorate([
        ui_register()
    ], MainUI);

    class BombFlyToTargetAni extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.mid = new Laya.Vector3();
            this.allTime = 2;
            this.time = 0;
            this.delay = 0;
        }
        onStart() {
            Laya.Vector3.subtract(this.targetPos, this.transform.position, this.mid);
            Laya.Vector3.scale(this.mid, 0.5, this.mid);
            Laya.Vector3.add(this.transform.position, this.mid, this.mid);
            this.mid.y += 30;
            this.curve = new QuadraticCurve(this.transform.position.clone(), this.mid.clone(), this.targetPos);
            SoundManager.inst.playSound('NucleerAlarm');
        }
        onUpdate() {
            this.delay -= LayaUtils.deltaTimeSec;
            if (this.delay > 0)
                return;
            this.time += LayaUtils.deltaTimeSec;
            this.time = Math.min(this.time, this.allTime);
            this.curve.GetPos(this.time / this.allTime, this.mid);
            this.transform.lookAt(this.mid, LayaUtils.UP);
            this.transform.localRotationEulerY += 90 * LayaUtils.deltaTimeSec;
            this.transform.position = this.mid;
            if (this.time == this.allTime) {
                var sp = this.sp;
                SoundManager.inst.playSound('NukeExp1');
                PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.Nuke05).then((p) => {
                    var fx = Laya.Sprite3D.instantiate(p);
                    this.node.addChild(fx);
                    fx.transform.position = sp.transform.position;
                    var dd = fx.addComponent(DelayDestory);
                    dd.delay = 4000;
                    sp.destroy(true);
                });
                this.destroy();
            }
        }
    }

    class FlyUpAndDownComponent extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.d = 0;
            this.originY = 0;
            this.delay = 1;
            this.oldZRotation = 0;
        }
        onStart() {
            this.originY = this.transform.localPositionY;
            this.delay += (0.5 - Math.random()) * 0.2;
            this.oldZRotation = this.transform.localRotationEulerZ;
        }
        onLateUpdate() {
            this.d += LayaUtils.deltaTimeSec;
            this.transform.localPositionY = this.originY + Math.sin(this.d * 5 * this.delay) * 0.4;
            this.transform.localRotationEulerZ = this.oldZRotation + Math.sin(this.d * 5 * this.delay) * 0.4 * 20;
        }
    }

    class BomFallToY0 extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.speed = 0;
        }
        onStart() {
            SoundManager.inst.playSound('Falling');
        }
        onUpdate() {
            this.speed += LayaUtils.deltaTimeSec * -50;
            this.transform.localPositionY += this.speed * LayaUtils.deltaTimeSec;
            if (this.transform.localPositionY <= 0) {
                this.onFallEnd();
                this.enabled = false;
            }
        }
        onFallEnd() {
            SoundManager.inst.playSound('Bomb_Explosions');
            PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.baozha03).then((p) => {
                if (this.node == null || this.node.destroyed)
                    return;
                if (p) {
                    var fx = Laya.Sprite3D.instantiate(p);
                    if (this.node != null) {
                        this.node.addChild(fx);
                    }
                    else {
                        MapComponent.inst.sp.addChild(fx);
                    }
                    fx.transform.position = this.sp.transform.position;
                    var dd = fx.addComponent(DelayDestory);
                    dd.delay = 700;
                }
                this.sp.destroy(true);
                SoundManager.inst.playVibrate(true);
                if (this.endCB) {
                    this.endCB();
                }
            });
        }
    }

    class RotateSelf extends Laya.Script3D {
        constructor() {
            super(...arguments);
            this.angleSpeed = new Laya.Vector3();
        }
        onUpdate() {
            let t = this.owner.transform;
            t.localRotationEulerY += this.angleSpeed.y * LayaUtils.deltaTimeSec;
            t.localRotationEulerZ += this.angleSpeed.z * LayaUtils.deltaTimeSec;
            t.localRotationEulerX += this.angleSpeed.x * LayaUtils.deltaTimeSec;
        }
    }

    let ControlUI = class ControlUI extends BaseUI {
        constructor() {
            super(...arguments);
            this.closeRight = 1;
            this._yuyueKey = "LogicData.yuyue";
        }
        get URL() {
            wxBinder.bindAll();
            return UI_ui_jm.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_hedan, this.ui.m_hedan_mask, this.ui.m_hongzha3D];
        }
        OnCreate() {
            var p = new Laya.Point();
            p = this.ui.globalToLocal(p.x, p.y, p);
            this.ui.m_btn_fh.onClick(this, this._onClickReborn);
            this.ui.m_btn_2bei.onClick(this, this._onClickDouble);
            this.ui.m_btn_rtfq.onClick(this, this._onClickClose, [false]);
            this.ui.m_btn_close1.onClick(this, this._onClickClose, [false]);
            this.ui.m_btn_close3.onClick(this, this._onClickClose, [false]);
            this.ui.m_btn_fashe.onClick(this, this._onClickFashe);
            this.ui.m_btn_yueyue.onClick(this, this._onClickYuyue, [1]);
            this.ui.m_btn_quxioa.onClick(this, this._onClickYuyue, [2]);
            this.ui.m_btn_qifei.onClick(this, this._onClickQifei);
            this.ui.m_btn_ljhd.onClick(this, this._onClickLjhd);
            this.ui.m_btn_fanhui.onClick(this, () => {
                UIManager.inst.ShowUI(MainUI, UILayer.base);
            });
            this.ui.m_btn_xhuanze.onClick(this, this._onClickToggle);
            this.ui.m_c1.on(fgui.Events.STATE_CHANGED, this, this.onStateChange);
        }
        onStateChange() {
            console.log(this.ui.m_c1.selectedPage);
            console.log(this.ui.m_icon.icon);
            switch (this.ui.m_c1.selectedIndex) {
                case 2:
                case 3:
                case 6:
                case 7:
                    this.ui.m_icon.icon = 'ui://s88k2hl2eper2j';
            }
        }
        OnShow(page, ...data) {
            console.log("===========>1");
            if (Laya.Browser.onTTMiniGame) {
                console.log("===========>2");
                this.ui.m_c2.selectedPage = 'close1';
            }
            else if (Laya.Browser.onMiniGame) {
                console.log("===========>3");
                this.ui.m_c2.selectedPage = 'close2';
            }
            else {
                console.log("===========>4");
                this.ui.m_c2.selectedPage = 'close1';
            }
            console.log("===========>5");
            this._refreshBtn();
            console.log("===========>6");
            switch (page) {
                case TipsType.ADD_DOGTAG:
                    console.log("===========>7");
                    this._refreshAddTag(data[0], data[1]);
                    this.ui.m_c1.selectedIndex = page;
                    this.showInsertAd();
                    break;
                case TipsType.HONGZHA:
                    console.log("===========>8");
                    this.playerHongZhaAni();
                    this._refreshAirPlane();
                    this.ui.m_c1.selectedIndex = page;
                    this.ui.m_btn_qifei.getControllerAt(0).selectedIndex = 0;
                    break;
                case TipsType.YUYUE:
                    console.log("===========>9");
                    this._refreshYuyue();
                    Laya.timer.frameLoop(1, this, this.updateYuYueTimer);
                    break;
                case TipsType.NUKE:
                    this.playeNukeAni();
                    this._refreshNuke();
                    this.ui.m_c1.selectedIndex = page;
                    this.ui.m_btn_fashe.getControllerAt(0).selectedIndex = 0;
                    break;
                case TipsType.FUJIANG:
                    this._refreshAddTag(data[0], data[1]);
                    this.ui.m_c1.selectedIndex = page;
                    this.showInsertAd();
                    break;
                case TipsType.REBORN:
                    this._refreshReborn(data[0]);
                    this.ui.m_c1.selectedIndex = page;
                    break;
            }
            GameManager.getInst().Pause();
            console.log("===========>10");
        }
        showInsertAd() {
            let data1 = { closePos: this.closeRight, panelPos: 1 };
            this.closeRight = this.closeRight == 0 ? 1 : 0;
            GameSDK.GameNativeInsert && GameSDK.GameNativeInsert.load(data1, true);
        }
        _refreshAddTag(val, gold) {
            this.ui.m_img_diban2.getChildAt(1).text = "+" + val;
            this.ui.m_img_diban2.m_type.selectedIndex = gold ? 1 : 0;
            this._laterShow(this.ui.m_btn_rtfq, 800);
        }
        _refreshReborn(showVideo) {
            this.ui.m_btn_fh.m_c1.selectedIndex = showVideo ? 1 : 0;
            this._laterShow(this.ui.m_btn_rtfq, 800);
        }
        _refreshAirPlane() {
            this._laterShow(this.ui.m_btn_qifei, 1500);
            this._laterShow(this.ui.m_btn_rtfq, 2300);
            this._laterShow(this.ui.m_img_lingqujl, 1500);
            this._laterShow(this.ui.m_btn_close3, 2300);
        }
        _refreshYuyue() {
            if (PlayerData.YuyueNuke) {
                if (new Date().getTime() < PlayerData.YuyueDate) {
                    this.ui.m_c1.selectedPage = '预约成功';
                    this.ui.m_btn_yueyue.m_c1.selectedIndex = 2;
                }
                else {
                    this.ui.m_c1.selectedPage = '预约领取';
                    this.ui.m_btn_yueyue.m_c1.selectedIndex = 3;
                }
            }
            else {
                this.ui.m_c1.selectedPage = '明日奖励';
                this.ui.m_btn_yueyue.m_c1.selectedIndex = 4;
            }
        }
        _refreshNuke() {
            this._laterShow(this.ui.m_btn_fashe, 1500);
            this._laterShow(this.ui.m_img_lingqujl, 1500);
            this._laterShow(this.ui.m_btn_close3, 2300);
            this._laterShow(this.ui.m_btn_rtfq, 2300);
        }
        _refreshYuyueState(isYuyue) {
            this.ui.m_img_wz3.visible = false;
            this.ui.m_img_yycg.visible = false;
            this.ui.m_img_tishi.visible = false;
            if (isYuyue) {
                if (this._yuyueTime > 0) {
                    this.ui.m_btn_yueyue.m_c1.selectedIndex = 2;
                    this.ui.m_btn_yueyue.m_label_time.text = this._getTimeStr(Math.max(this._yuyueTime, 0));
                    this.ui.m_img_tishi.visible = true;
                    this.ui.m_img_wz3.visible = true;
                    this.ui.m_img_yycg.visible = true;
                }
                else {
                    this.ui.m_btn_yueyue.m_c1.selectedIndex = 3;
                }
            }
            else {
                this.ui.m_btn_yueyue.m_c1.selectedIndex = 1;
                this.ui.m_img_tishi.visible = true;
            }
        }
        _getTimeStr(time) {
            time = Math.floor(time / 1000);
            let s = time % 60;
            let m = Math.floor(time / 60) % 60;
            let h = Math.floor(time / 3600);
            return this._replaceNum(h) + ":" + this._replaceNum(m) + ":" + this._replaceNum(s);
        }
        _replaceNum(val) {
            return (val < 10 ? "0" : "") + val;
        }
        get _yuyueTime() {
            let time = LocalCache$1.get(this._yuyueKey, Date.now());
            return time - new Date(new Date().toLocaleDateString()).getTime();
        }
        get _isYuyue() {
            return LocalCache$1.get(this._yuyueKey, 0) != 0;
        }
        set _yuyueTime(val) {
            LocalCache$1.set(this._yuyueKey, val);
        }
        _load3D(loader, url) {
        }
        playeNukeAni() {
            if (this._scene3D == null) {
                this._scene3D = new SceneShow3D();
                this._scene3D.setRenderTexture(this.ui.m_hedan.displayObject);
            }
            this._scene3D.playerNode.removeChildren();
            var nuke = MapComponent.inst.Find('Nuke');
            var bomb = nuke.clone();
            bomb.active = true;
            this._scene3D.playerNode.addChild(bomb);
            bomb.transform.position = new Laya.Vector3(6, 0, -15);
            bomb.transform.rotation = nuke.transform.rotation;
            var c = bomb.addComponent(BombFlyToTargetAni);
            c.targetPos = new Laya.Vector3;
            c.node = this._scene3D.playerNode;
            Laya.timer.once(6000, this, this.playeNukeAni);
        }
        playerHongZhaAni() {
            if (this._scene3D == null) {
                this._scene3D = new SceneShow3D();
                this._scene3D.setRenderTexture(this.ui.m_hongzha3D.displayObject);
            }
            this._scene3D.playerNode.removeChildren();
            PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.WarPlane).then(p => {
                var root = new Laya.Sprite3D();
                var rs = root.addComponent(RotateSelf);
                rs.angleSpeed = new Laya.Vector3(0, -120, 0);
                this._scene3D.playerNode.addChild(root);
                root.transform.localPosition = new Laya.Vector3(0, 10, 0);
                var plane = Laya.Sprite3D.instantiate(p);
                plane.transform.localPosition = new Laya.Vector3(3, 0, 0);
                plane.transform.localRotationEuler = LayaUtils.Vector3Zero;
                var smoke = Utils.GetNode(plane, 'SmokeDark');
                if (smoke) {
                    smoke.active = false;
                }
                plane.addComponent(FlyUpAndDownComponent);
                root.addChild(plane);
            });
            PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.BombBullet).then(() => {
                PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.baozha03).then(() => {
                    Laya.timer.once(500, this, this.spwanPlaneBomb);
                });
            });
        }
        spwanPlaneBomb() {
            var bombPrefab = PrefabManager.getInst().GetPrefab(_AllPrefabsNames.BombBullet);
            if (this._scene3D == null || this._scene3D.playerNode == null || this._scene3D.playerNode.destroyed)
                return;
            var b = Laya.Sprite3D.instantiate(bombPrefab);
            var cc = b.addComponent(BomFallToY0);
            cc.node = this._scene3D.playerNode;
            this._scene3D.playerNode.addChild(b);
            b.transform.position = new Laya.Vector3(0, 9, 0);
            b.transform.position.x += (0.5 - Math.random()) * 8;
            b.transform.position.z += (0.5 - Math.random()) * 8;
            b.transform.position = b.transform.position;
            if (this._scene3D == null || this._scene3D.playerNode == null || this._scene3D.playerNode.destroyed)
                return;
            Laya.timer.once(500, this, this.spwanPlaneBomb);
        }
        OnHide() {
            if (!!this._scene3D) {
                this._scene3D.destroy();
            }
            AdMgr.inst.showLeftAd(false);
            AdMgr.inst.showRightAd(false);
            AdMgr.inst.showBottomAd(false);
            this._scene3D = null;
            Laya.timer.clearAll(this);
            GameManager.getInst().Resume();
            Laya.timer.clear(this, this.updateYuYueTimer);
            Laya.timer.clear(this, this.playeNukeAni);
            Laya.timer.clear(this, this.spwanPlaneBomb);
            GameSDK.GameNativeInsert && GameSDK.GameNativeInsert.setIsShow(false);
        }
        OnDispose() {
            MsgManager.getInst().offAllCaller(this);
        }
        _laterShow(cpt, time) {
            if (Laya.Browser.onTTMiniGame)
                return;
            cpt.visible = false;
            Laya.timer.once(time, this, () => {
                cpt.visible = true;
            });
        }
        _onClickClose(success) {
            switch (this.ui.m_c1.selectedPage) {
                        }
            UIManager.inst.ShowUI(MainUI, UILayer.base);
            AlertMgr.inst.runCloseCB(success);
        }
        _onClickToggle() {
            this._refreshBtn();
        }
        _refreshBtn() {
        }
        _onClickLjhd() {
            ReportMgr.inst.sendReport(ReportType.WIN_TAG_VIDEO);
            GameSDK.clickADVideo(this.ui.m_btn_ljhd, this, s => {
                if (s) {
                    this._onClickClose(true);
                }
            }, "额外铭牌");
        }
        _onClickQifei() {
            ReportMgr.inst.sendReport(ReportType.WIN_PLANE_VIDEO);
            GameSDK.clickADVideo(this.ui.m_btn_qifei, this, s => {
                if (s) {
                    this._onClickClose(true);
                }
            }, "第一局轰炸机视频使用");
        }
        _onClickYuyue(type) {
            switch (this.ui.m_c1.selectedPage) {
                case "明日奖励":
                    this.ui.m_c1.selectedPage = '预约成功';
                    this.ui.m_btn_yueyue.m_c1.selectedIndex = 2;
                    PlayerData.YuyueNuke = true;
                    var day_date = new Date(new Date(new Date().getTime() + 24 * 60 * 60 * 1000).setHours(1, 0, 0, 0)).getTime();
                    PlayerData.YuyueDate = day_date;
                    TipsMgr.Tips("预约成功!");
                    this.OnShow(TipsType.YUYUE);
                    if (type == 1) {
                        ReportMgr.inst.sendReport(ReportType.NUKE_CLICK_YUYUE);
                    }
                    else if (type == 2) {
                        ReportMgr.inst.sendReport(ReportType.NUKE_CLICK_YUYUE_N);
                    }
                    UIManager.inst.ShowUI(MainUI, UILayer.base);
                    break;
                case '预约成功':
                    return;
                case '预约领取':
                    PlayerData.YuyueNuke = false;
                    PlayerData.NormalNuke = true;
                    ADData.nukeCD = 0;
                    UIManager.inst.ShowUI(MainUI, UILayer.base);
                    break;
            }
        }
        updateYuYueTimer() {
            this.ui.m_btn_yueyue.touchable = true;
            var now_time = new Date().getTime();
            if (now_time >= PlayerData.YuyueDate) {
                this.ui.m_btn_yueyue.m_c1.selectedIndex = 3;
                this.ui.m_c1.selectedPage = '预约领取';
            }
            else if (this.ui.m_btn_yueyue.m_c1.selectedIndex == 2) {
                this.ui.m_btn_yueyue.touchable = false;
                this.ui.m_btn_yueyue.m_label_time.text = Utils.TimeToStringMS2(PlayerData.YuyueDate - now_time, 0);
            }
        }
        _onClickFashe() {
            if (this._scene3D != null) {
                this._scene3D.playerNode.active = false;
            }
            ReportMgr.inst.sendReport(ReportType.WIN_NUKEALERT_VIDEO);
            GameSDK.clickADVideo(this.ui.m_btn_fashe, this, s => {
                if (s) {
                    this._onClickClose(true);
                }
                else {
                    if (this._scene3D != null) {
                        this._scene3D.playerNode.active = true;
                    }
                }
            }, "核弹试用视频");
        }
        _onClickDouble() {
            ReportMgr.inst.sendReport(ReportType.WIN_AIRTAG_VIDEO);
            GameSDK.clickADVideo(this.ui.m_btn_2bei, this, s => {
                if (s) {
                    this._onClickClose(true);
                    MsgManager.getInst().event(BattleEvent.SpwanDogTagDouble);
                }
            }, "空投双倍铭牌");
        }
        _onClickReborn() {
            if (this.ui.m_btn_fh.m_c1.selectedIndex == 1) {
                ReportMgr.inst.sendReport(ReportType.WIN_REBORN_VIDEO);
                GameSDK.clickADVideo(this.ui.m_btn_fh, this, s => {
                    if (s) {
                        AlertMgr.inst.addRebornTime();
                        this._onClickClose(true);
                    }
                }, "复活");
            }
            else {
                AlertMgr.inst.addRebornTime();
                this._onClickClose(true);
            }
        }
        _onClickGiveUp() {
            this._onClickClose(false);
        }
    };
    ControlUI = __decorate([
        ui_register()
    ], ControlUI);

    class AlertMgr {
        constructor() {
            this._rebornKey = "LogicData.rebornTime";
            this._showNukeKey = "LogicData._showNukeKey";
            this._playingKey = "LogicData._playingKey";
            this._playerRankKey = "LogicData._playerRankKey";
            this._playerRankCountKey = "LogicData._playerRankCountKey";
            this._selfEnemyCleanKey = "LogicData._selfEnemyCleanKey";
        }
        static get inst() {
            if (!!!this._inst) {
                this._inst = new AlertMgr();
            }
            return this._inst;
        }
        resetRebornTime() {
            LocalCache$1.set(this._rebornKey, 0);
        }
        addRebornTime() {
            LocalCache$1.set(this._rebornKey, LocalCache$1.get(this._rebornKey, 0) + 1);
        }
        showReborn() {
            ReportMgr.inst.sendReport(ReportType.WIN_REBORN);
            let showVideo = LocalCache$1.get(this._rebornKey, 0) > 0;
            UIManager.inst.ShowUI(ControlUI, UILayer.base, TipsType.REBORN, showVideo);
        }
        showAddDogTag(val, gold) {
            ReportMgr.inst.sendReport(ReportType.WIN_TAG);
            UIManager.inst.ShowUI(ControlUI, UILayer.base, TipsType.ADD_DOGTAG, val, gold);
        }
        showHongzha() {
            UIManager.inst.ShowUI(ControlUI, UILayer.base, TipsType.HONGZHA);
        }
        showYuyue() {
            ReportMgr.inst.sendReport(ReportType.NUKE_YUYUE_WIN);
            UIManager.inst.ShowUI(ControlUI, UILayer.base, TipsType.YUYUE);
        }
        showNuke() {
            if (PlayerData.CurrentLevelID == 1 && !LocalCache$1.get(this._showNukeKey, false) && this.getSelfEnemyCleanCount() == 1) {
                LocalCache$1.set(this._showNukeKey, true);
                ReportMgr.inst.sendReport(ReportType.WIN_NUKEALERT);
                UIManager.inst.ShowUI(ControlUI, UILayer.base, TipsType.NUKE);
                return true;
            }
            return false;
        }
        showFujiang(val, isGold) {
            ReportMgr.inst.sendReport(ReportType.WIN_AIRTAG);
            UIManager.inst.ShowUI(ControlUI, UILayer.base, TipsType.FUJIANG, val, isGold);
        }
        setCloseCB(success, fail) {
            if (!!this._successCB) {
                this._successCB = null;
            }
            if (!!this._failCB) {
                this._failCB = null;
            }
            this._successCB = success;
            this._failCB = fail;
        }
        runCloseCB(success) {
            if (success && !!this._successCB) {
                this._successCB.run();
            }
            else if (!success && !!this._failCB) {
                this._failCB.run();
            }
            this._successCB = null;
            this._failCB = null;
        }
        savePlayerLevel(bigLevel, smallLevel) {
            let level = bigLevel + smallLevel;
            if (LocalCache$1.get(this._playingKey, "") != level) {
                LocalCache$1.set(this._playingKey, level);
                LocalCache$1.set(this._playerRankKey, PlayerData.CurrentRankLevel);
                LocalCache$1.set(this._playerRankCountKey, BattleData.PlayerNode.tag);
            }
        }
        resetPlayerLevel() {
            PlayerData.CurrentRankLevel = LocalCache$1.get(this._playerRankKey, PlayerData.CurrentRankLevel);
            BattleData.PlayerNode.tag = LocalCache$1.get(this._playerRankCountKey, BattleData.PlayerNode.tag);
        }
        addSelfEnemyCleanCount() {
            LocalCache$1.set(this._selfEnemyCleanKey, this.getSelfEnemyCleanCount() + 1);
        }
        getSelfEnemyCleanCount() {
            return LocalCache$1.get(this._selfEnemyCleanKey, 0);
        }
    }
    var TipsType;
    (function (TipsType) {
        TipsType[TipsType["ADD_DOGTAG"] = 0] = "ADD_DOGTAG";
        TipsType[TipsType["HONGZHA"] = 1] = "HONGZHA";
        TipsType[TipsType["YUYUE"] = 2] = "YUYUE";
        TipsType[TipsType["NUKE"] = 3] = "NUKE";
        TipsType[TipsType["FUJIANG"] = 4] = "FUJIANG";
        TipsType[TipsType["REBORN"] = 5] = "REBORN";
    })(TipsType || (TipsType = {}));

    class BaseTriggerEntity {
        constructor(pos) {
            this.inTriggerTime = 0;
            this.isTrigger = false;
            this.needRestTriggerState = false;
            this.triggerArea = new PolygonWrap(pos);
        }
        checkIn(p) {
            if (this.triggerArea.FastTest(p)) {
                return MathUtils.PointInPolygon2(p, this.triggerArea.points);
            }
            return false;
        }
        Update(deltaTime, playerPos) {
            if (this.onUpdate) {
                this.onUpdate();
            }
            if (this.checkIn(playerPos)) {
                this.inTriggerTime += deltaTime;
                if (this.inTriggerTime >= 0.4) {
                    this.inTriggerTime = 0;
                    if (this.needRestTriggerState || this.isTrigger == false) {
                        this.isTrigger = true;
                        this.EnterTrigger();
                    }
                }
            }
            else {
                this.inTriggerTime = 0;
                if (this.needRestTriggerState || this.isTrigger) {
                    this.isTrigger = false;
                    this.ExitTrigger();
                }
            }
        }
    }

    class DropBoxTriggerEntity extends BaseTriggerEntity {
        constructor(node, gold = true, showVideo = false) {
            super([new Laya.Vector2(node.transform.position.x, node.transform.position.z)]);
            this.gold = true;
            this.showVideo = false;
            this.pos = new Laya.Vector2(node.transform.position.x, node.transform.position.z);
            this.gold = gold;
            this.model = node;
            this.showVideo = showVideo;
        }
        checkIn(p) {
            return LayaUtils.distanceSquaredVec2(this.pos, p) < 9;
        }
        ExitTrigger() {
        }
        EnterTrigger() {
            for (var i = 0; i < 10; i++) {
                var p = new PosState();
                p.x = this.model.transform.position.x;
                p.y = i * 0.1;
                p.z = this.model.transform.position.z;
                p.r = 0;
                BattleData.goldTagInGroud.push(p);
                MsgManager.getInst().event(BattleEvent.CreateDogTag, p, this.gold);
            }
            var ani = this.model.getComponent(Laya.Animator);
            ani.play('Open');
            Utils.GetNode(this.model, 'Extras').destroy(true);
            Utils.GetNode(this.model, 'jls (1)').destroy(true);
            SoundManager.inst.playSound('DropBoxOpen');
            var dd = this.model.addComponent(DelayDestory);
            dd.delay = this.gold ? 3000 : 30000;
            if (!Laya.Browser.onQGMiniGame) {
                if (this.showVideo) {
                    GameSDK.clickADVideo(null, this, s => {
                        if (s) {
                            MsgManager.getInst().event(BattleEvent.SpwanDogTagDouble);
                        }
                    }, "3D空投视频");
                }
                else {
                    Utils.WaitTime(300).then(() => {
                        MsgManager.getInst().on(BattleEvent.SpwanDogTagDouble, this, this.onDouble);
                        AlertMgr.inst.showFujiang(10, this.gold);
                    });
                }
            }
        }
        onDouble() {
            MsgManager.getInst().offAllCaller(this);
            if (this.model.destroyed)
                return;
            for (var i = 0; i < 10; i++) {
                var p = new PosState();
                p.x = this.model.transform.position.x;
                p.y = i * 0.1 + 0.9;
                p.z = this.model.transform.position.z;
                p.r = 0;
                BattleData.goldTagInGroud.push(p);
                MsgManager.getInst().event(BattleEvent.CreateDogTag, p, this.gold);
            }
        }
    }

    class ScaleDestroy extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.scaleTime = 0.7;
            this.currentScale = new Laya.Vector3();
        }
        onAwake() {
            this.currentScale = this.sp.transform.localScale.clone();
        }
        onStart() {
            this.sp.transform.localScale.cloneTo(this.currentScale);
            this.t = fgui.GTween.to4(this.currentScale.x, this.currentScale.y, this.currentScale.z, this.transform.localPositionY, 0, 0, 0, 4, this.scaleTime);
        }
        onUpdate() {
            var g = this.t;
            if (g == null)
                return;
            this.transform.localScaleX = g.value.x;
            this.transform.localScaleY = g.value.y;
            this.transform.localScaleZ = g.value.z;
            this.transform.localPositionY = g.value.w;
        }
        onDestroy() {
            this.t.kill();
        }
    }

    class DropBoxBuilding extends BaseScript3D {
        constructor() {
            super();
            this.isVideo = false;
            this.isEneyTag = true;
            this.forceLockCam = true;
            this.speed = 0;
            this.parachuteShow = true;
        }
        onStart() {
            if (this.forceLockCam) {
                SimpleCameraFollow.inst.target = this.sp;
            }
            this.fly = this.sp.getChildAt(0).addComponent(FlyUpAndDownComponent);
            SoundManager.inst.playSound('Falling');
            if (!!this._trigger) {
                this._trigger = null;
            }
            this._trigger = new DropBoxTriggerEntity(this.sp, this.isEneyTag, this.isVideo);
        }
        onDestroy() {
            if (this.forceLockCam && SimpleCameraFollow.inst.target == this.sp) {
                SimpleCameraFollow.inst.target = null;
            }
        }
        onUpdate() {
            this.speed += LayaUtils.deltaTimeSec * -10;
            this.transform.localPositionY += this.speed * LayaUtils.deltaTimeSec;
            if (this.transform.localPositionY < 6) {
                if (this.parachuteShow) {
                    this.parachuteShow = false;
                    var p = this.Find('jls');
                    p.addComponent(ThrowDestroy);
                    p.addComponent(ScaleDestroy);
                    this.fly.destroy();
                }
            }
            if (this.transform.localPositionY <= 0) {
                this.onFallEnd();
                var boxTrigger = new DropBoxTriggerEntity(this.sp, true, this.isVideo);
                MsgManager.getInst().event(BattleEvent.SpwanDropBoxTrigger, boxTrigger);
                this.destroy();
            }
        }
        onFallEnd() {
            if (this.isVideo) ;
            var sp = this.sp;
            this.transform.localPositionY = 0;
            sp.transform.localRotationEulerZ = sp.transform.localRotationEulerX = sp.transform.localRotationEulerY = 0;
            PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.yan02).then(p => {
                SoundManager.inst.playSound('DropBoxOpen');
            });
        }
    }

    class FreeNukeTrigger extends BaseTriggerEntity {
        constructor(node, gold = true) {
            super([new Laya.Vector2(node.transform.position.x, node.transform.position.z)]);
            this.pos = new Laya.Vector2(node.transform.position.x, node.transform.position.z);
            this.model = node;
        }
        checkIn(p) {
            return LayaUtils.distanceSquaredVec2(this.pos, p) < 4;
        }
        ExitTrigger() {
        }
        EnterTrigger() {
            MsgManager.getInst().event(BattleEvent.TriggerFreeNuke, this.model.transform.position);
            ArrowTips.inst.Hide();
            this.model.destroy(true);
            ADData.nukeCD = 0;
            PlayerData.FreeNuke++;
        }
    }

    class FreeNukeFall extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.speed = 0;
            this.parachuteShow = true;
        }
        onStart() {
            ReportMgr.inst.sendReport(ReportType.NUKE_FALL_FREE_START);
            SimpleCameraFollow.inst.target = this.sp;
            SoundManager.inst.playSound('Falling');
            this.Find('tuowei').active = false;
            this.speed = 0;
            var view = this.Find('NuclearBomb');
            view.transform.localPositionZ = -1;
        }
        onDestroy() {
            if (SimpleCameraFollow.inst.target == this.sp) {
                SimpleCameraFollow.inst.target = null;
            }
        }
        onUpdate() {
            this.speed += LayaUtils.deltaTimeSec * -40;
            this.transform.localPositionY += this.speed * LayaUtils.deltaTimeSec;
            if (this.transform.localPositionY <= 0) {
                this.onFallEnd();
                this.destroy();
            }
        }
        onFallEnd() {
            this.transform.localPositionY = 0;
            var boxTrigger = new FreeNukeTrigger(this.sp);
            ReportMgr.inst.sendReport(ReportType.NUKE_FALL_FREE_END);
            MsgManager.getInst().event(BattleEvent.SpwanFreeNukeTrigger, boxTrigger);
            SoundManager.inst.playSound('DropBoxOpen');
            ArrowTips.inst.target = this.sp;
            GuideData.Step = GuideStepType.拾取免费核弹;
        }
    }

    var tmp$6 = new Laya.Vector3();
    var invertTmp$1 = new Laya.Vector3();
    class Helicopter extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.selfPos = new Laya.Vector2();
            this.velocity = new Laya.Vector3();
            this.move = true;
            this.bulletCount = 3;
            this.attackCD = 0;
        }
        onAwake() {
            var rs = this.Find('Propeller').addComponent(RotateSelf);
            rs.angleSpeed = new Laya.Vector3(0, 7200, 0);
            this.sp.getChildAt(0).addComponent(FlyUpAndDownComponent);
            PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.PlaneBomb).then(p => {
                this.bombPrefab = p;
            });
            MsgManager.getInst().event(BattleEvent.FireBombStart);
        }
        onStart() {
            Laya.Vector3.subtract(this.targetPos, this.transform.position, this.velocity);
            Laya.Vector3.normalize(this.velocity, this.velocity);
            Laya.Vector3.scale(this.velocity, 12, this.velocity);
            SimpleCameraFollow.inst.target = this.sp;
            this.playSound();
        }
        playSound() {
            SoundManager.inst.playSound('Helicopter');
            Laya.timer.once(1645, this, this.playSound);
        }
        onDestroy() {
            if (SimpleCameraFollow.inst.target == this.sp) {
                SimpleCameraFollow.inst.target = null;
            }
            MsgManager.getInst().event(BattleEvent.FireBombEnd);
            Laya.timer.clearAll(this);
        }
        onUpdate() {
            if (this.move) {
                Laya.Vector3.subtract(this.targetPos, this.transform.position, this.velocity);
                Laya.Vector3.normalize(this.velocity, this.velocity);
                Laya.Vector3.scale(this.velocity, 12, this.velocity);
                if (Laya.Vector3.distanceSquared(this.transform.position, this.targetPos) < 0.01) {
                    this.move = false;
                    return;
                }
                Laya.Vector3.scale(this.velocity, LayaUtils.deltaTimeSec, tmp$6);
                Laya.Vector3.scale(this.velocity, -LayaUtils.deltaTimeSec, invertTmp$1);
                Laya.Vector3.add(this.transform.position, tmp$6, tmp$6);
                this.transform.position = tmp$6;
                Laya.Vector3.add(this.transform.position, invertTmp$1, invertTmp$1);
                this.transform.lookAt(invertTmp$1, LayaUtils.UP);
            }
            if (this.bulletCount > 0) {
                this.selfPos.setValue(this.transform.position.x, this.transform.position.z);
                if (LayaUtils.distanceVec2(this.selfPos, this.targetEnemyPos) < 4) {
                    this.attackCD -= LayaUtils.deltaTimeSec;
                    if (this.attackCD <= 0) {
                        this.attackCD += 0.6;
                        var b = Laya.Sprite3D.instantiate(this.bombPrefab);
                        MapComponent.inst.sp.addChild(b);
                        var cc = b.addComponent(BomFallToY0);
                        cc.node = MapComponent.inst.sp;
                        b.transform.position = this.transform.position;
                        b.transform.position.x += (0.5 - Math.random()) * 3;
                        b.transform.position.z += (0.5 - Math.random()) * 3;
                        cc.endCB = () => {
                            MsgManager.getInst().event(BattleEvent.KillEnemy, this.targetEnemyPos, 4, 200);
                        };
                        this.bulletCount--;
                        if (this.bulletCount <= 0) {
                            this.sp.addComponent(DelayDestory);
                        }
                    }
                }
            }
        }
    }

    class MechaBoxComponent extends BaseScript3D {
        onAwake() {
            this.button = this.Find('anniu');
            this.buttonAD = this.Find('anniu_03');
            this.box = this.Find('RotaionRoot');
            this.openAni = this.Find('DropPod2').getComponent(Laya.Animator);
            this.jijiaNode = this.Find('DropPodInteractable');
            this.flamethrower = this.Find('FlamethrowerSpikyRed');
            this.flamethrower.active = false;
            this.onDrop = this.Find('OnDropParticles');
            this.onDrop.active = false;
            this.smoke = this.Find('SmokeExplosionWhite');
            this.smoke.active = false;
            this.repaireSmoke = this.Find('repaireSmoke');
            this.repaireSmoke.active = false;
        }
        onUpdate() {
            if (GameManager.getInst().battleManager.mountHorse) {
                this.button.active = this.buttonAD.active = false;
            }
            else {
                this.button.active = PlayerData.FreeMecha > 0 || (PlayerData.UnlockMecha);
                this.buttonAD.active = PlayerData.FreeMecha <= 0 && PlayerData.UnlockMecha;
            }
        }
        onDestroy() {
        }
        DropBox() {
            MsgManager.getInst().event(BattleEvent.HideMainUI);
            this.button.active = false;
            SimpleCameraFollow.inst.target = this.box;
            this.flamethrower.active = true;
            this.box.transform.localPositionY = 20;
            Laya.Tween.to(this.box.transform, { localPositionY: 0.2 }, 5000, Laya.Ease.linearInOut, Laya.Handler.create(this, this.onDropEnd));
        }
        onDropEnd() {
            this.flamethrower.active = false;
            this.onDrop.active = true;
            this.button.active = true;
            Utils.WaitTime(1000).then(() => {
                this.onDrop.active = false;
                MsgManager.getInst().event(BattleEvent.ShowMainUI);
                if (SimpleCameraFollow.inst.target == this.box) {
                    SimpleCameraFollow.inst.target = null;
                }
            });
        }
        OpenBox() {
            this.smoke.active = true;
            this.openAni.play('open');
            this.repaireSmoke.active = false;
            Utils.WaitTime(1000).then(() => {
                this.repaireSmoke.active = false;
                this.smoke.active = false;
            });
        }
        CloseBox() {
            this.openAni.play('Drop Pod Closed');
        }
    }

    class MapCampManagerComponent extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.slots = [];
            this.tankSlos = [];
            this.currentSlotIndex = 0;
            this.currentTankSlotIndex = 0;
        }
        onAwake() {
            MapCampManagerComponent.inst = this;
            this.playerLocation = this.Find(GlobalConstPoxy.Inst.data.PlayerLocationNode) || this.sp;
            var slots = this.Find('Slot');
            for (var i = 0; i < slots.numChildren; i++) {
                var slot = slots.getChildAt(i);
                this.slots.push(slot);
            }
            var tankSlots = this.Find('TankSlot');
            for (var i = 0; i < tankSlots.numChildren; i++) {
                var slot = tankSlots.getChildAt(i);
                this.tankSlos.push(slot);
            }
            this.yuanjiao = this.Find('YuanJiao');
            var p = this.Find('YuanJiao').getChildAt(0);
            if (p) {
                var mat = p.meshRenderer.material = new VertexUVMaterial();
                var c = Utils.hexToRgba01("#545454");
                c.w = 0.6;
                mat.mainColor = c;
                mat.progress = 1;
                mat.radius = 0.1;
            }
            this.updateSlot();
            MsgManager.getInst().on(BattleEvent.ResetFootSoliderSlot, this, this.onResetFootSoliderSlot);
            MsgManager.getInst().on(BattleEvent.RankLevelUp, this, this.updateSlot);
        }
        onDestroy() {
            MsgManager.getInst().offAllCaller(this);
        }
        updateSlot() {
            var show = 0;
            for (var i = 0; i < this.slots.length; i++) {
                this.slots[i].active = (i < PlayerData.CurrentCommanderConfig.soldiers);
                if (this.slots[i].active) {
                    show++;
                }
            }
            if (this.yuanjiao) {
                this.yuanjiao.transform.localScaleZ = Math.ceil(show / 5) + 0.6;
            }
            for (var i = 0; i < this.tankSlos.length; i++) {
                this.tankSlos[i].active = (i < PlayerData.CurrentCommanderConfig.tanks);
            }
        }
        GetSpwanSlot(type) {
            switch (type) {
                case SoilderType.Soilder:
                    return this.GetSoilderSpawnSlot();
                case SoilderType.Tank:
                    return this.GetTankSpawnSlot();
            }
            return null;
        }
        GetSoilderSpawnSlot() {
            if (this.currentSlotIndex < PlayerData.CurrentCommanderConfig.soldiers) {
                return this.slots[this.currentSlotIndex++];
            }
            return null;
        }
        GetTankSpawnSlot() {
            if (this.currentTankSlotIndex < PlayerData.CurrentCommanderConfig.tanks) {
                return this.tankSlos[this.currentTankSlotIndex++];
            }
            return null;
        }
        onResetFootSoliderSlot() {
            this.currentSlotIndex = 0;
            this.currentTankSlotIndex = 0;
        }
        SpwanBuilding(slotName, prefabName, fx = false) {
            SoundManager.inst.playCreateBuilding();
            var slot = this.GetSlotNode(slotName);
            slot.removeChildren();
            var building = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(prefabName));
            slot.addChild(building);
            building.transform.localPosition = LayaUtils.Vector3Zero;
            var info = new BuildingInfo();
            info.node = building;
            info.spwanPos = (Utils.GetNode(building, 'SpawnPos') || building).transform.position;
            var triggerNode = Utils.GetNode(building, 'TriggerPolygon');
            if (triggerNode) {
                for (var i = 0; i < triggerNode.numChildren; i++) {
                    var tn = triggerNode.getChildAt(i);
                    var p2 = new Laya.Vector2(tn.transform.position.x, tn.transform.position.z);
                    info.trigerPolygon.push(p2);
                }
            }
            var polygonNode = Utils.GetNode(building, "polygon");
            if (polygonNode != null) {
                for (var i = 0; i < polygonNode.numChildren; i++) {
                    let child = polygonNode.getChildAt(i);
                    info.polygon.push(new Laya.Vector2(child.transform.position.x, child.transform.position.z));
                }
            }
            if (fx) {
                var view = Utils.GetNode(building, 'View');
                if (view != null) {
                    Utils.WaitFrame(2).then(() => {
                        var x = view.transform.localRotationEulerX;
                        view.transform.localRotationEulerX = -30;
                        view.transform.localScale = LayaUtils.Vector3Zero;
                        Laya.Tween.to(view.transform, { localPositionY: 4, localScaleX: 1, localScaleY: 1, localScaleZ: 1 }, 100, Laya.Ease.expoIn);
                        Laya.Tween.to(view.transform, { localPositionY: 0 }, 100, Laya.Ease.expoIn, null, 200);
                        Laya.Tween.to(view.transform, { localRotationEulerX: -30 }, 100, Laya.Ease.expoOut);
                        Laya.Tween.to(view.transform, { localRotationEulerX: x }, 100, Laya.Ease.expoOut, null, 200);
                    });
                }
                var yp = PrefabManager.getInst().GetPrefab(_AllPrefabsNames.yan);
                var yanwu = Laya.Sprite3D.instantiate(yp);
                this.sp.addChild(yanwu);
                yanwu.transform.position = building.transform.position;
                yanwu.addComponent(DelayDestory);
            }
            return info;
        }
        GetSlotNode(name) {
            return this.Find(name);
        }
        SpwanWeaponFactory(slotName, config, fx = false) {
            SoundManager.inst.playCreateBuilding();
            var slot = this.GetSlotNode(slotName);
            slot.removeChildren();
            var prefab = PrefabManager.getInst().GetPrefab(config.prefab);
            var building = Laya.Sprite3D.instantiate(prefab);
            slot.addChild(building);
            building.transform.localPosition = LayaUtils.Vector3Zero;
            building.transform.localRotationEuler = new Laya.Vector3(0, 0, 0);
            if (fx) {
                var view = Utils.GetNode(building, 'View');
                if (view != null) {
                    Utils.WaitFrame(2).then(() => {
                        var x = view.transform.localRotationEulerX;
                        view.transform.localRotationEulerX = -30;
                        view.transform.localScale = LayaUtils.Vector3Zero;
                        Laya.Tween.to(view.transform, { localPositionY: 4, localScaleX: 1, localScaleY: 1, localScaleZ: 1 }, 100, Laya.Ease.expoIn);
                        Laya.Tween.to(view.transform, { localPositionY: 0 }, 100, Laya.Ease.expoIn, null, 200);
                        Laya.Tween.to(view.transform, { localRotationEulerX: -30 }, 100, Laya.Ease.expoOut);
                        Laya.Tween.to(view.transform, { localRotationEulerX: x }, 100, Laya.Ease.expoOut, null, 200);
                    });
                }
                var yanwu = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(_AllPrefabsNames.yan));
                this.sp.addChild(yanwu);
                yanwu.transform.position = building.transform.position;
                yanwu.addComponent(DelayDestory);
            }
            return building;
        }
        SpawnNuke(slotName, config) {
            var slot = this.GetSlotNode(slotName);
            slot.removeChildren();
            var prefab = PrefabManager.getInst().GetPrefab(config.prefab);
            var building = Laya.Sprite3D.instantiate(prefab);
            slot.addChild(building);
            building.transform.localPosition = LayaUtils.Vector3Zero;
            building.transform.localRotationEuler = new Laya.Vector3(0, 0, 0);
            return building;
        }
        SpawnShip() {
            var slot = this.GetSlotNode('chuanLocation');
            slot.removeChildren();
            var prefab = PrefabManager.getInst().GetPrefab(_AllPrefabsNames.chuan1);
            var building = Laya.Sprite3D.instantiate(prefab);
            slot.addChild(building);
            building.transform.localPosition = LayaUtils.Vector3Zero;
            building.transform.localRotationEuler = new Laya.Vector3(0, 0, 0);
            return building;
        }
        SpwanExchange(fx) {
            SoundManager.inst.playCreateBuilding();
            var slot = this.GetSlotNode('ExchangeBuilding');
            slot.removeChildren();
            var prefab = PrefabManager.getInst().GetPrefab(_AllPrefabsNames.ExchageBuilding);
            var building = Laya.Sprite3D.instantiate(prefab);
            slot.addChild(building);
            building.transform.localPosition = LayaUtils.Vector3Zero;
            building.transform.localRotationEuler = new Laya.Vector3(0, 0, 0);
            if (fx) {
                var view = Utils.GetNode(building, 'View');
                if (view != null) {
                    Utils.WaitFrame(2).then(() => {
                        var x = view.transform.localRotationEulerX;
                        view.transform.localRotationEulerX = -30;
                        view.transform.localScale = LayaUtils.Vector3Zero;
                        Laya.Tween.to(view.transform, { localPositionY: 4, localScaleX: 1, localScaleY: 1, localScaleZ: 1 }, 100, Laya.Ease.expoIn);
                        Laya.Tween.to(view.transform, { localPositionY: 0 }, 100, Laya.Ease.expoIn, null, 200);
                        Laya.Tween.to(view.transform, { localRotationEulerX: -30 }, 100, Laya.Ease.expoOut);
                        Laya.Tween.to(view.transform, { localRotationEulerX: x }, 100, Laya.Ease.expoOut, null, 200);
                    });
                }
                var yanwu = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(_AllPrefabsNames.yan));
                this.sp.addChild(yanwu);
                yanwu.transform.position = building.transform.position;
                yanwu.addComponent(DelayDestory);
            }
            return building;
        }
        SpawnStorageBox(config, fx = true) {
            SoundManager.inst.playCreateBuilding();
            var slot = this.GetSlotNode('StorageBox');
            slot.removeChildren();
            var prefab = PrefabManager.getInst().GetPrefab(config.prefab);
            var building = Laya.Sprite3D.instantiate(prefab);
            slot.addChild(building);
            building.transform.localPosition = LayaUtils.Vector3Zero;
            building.transform.localRotationEuler = new Laya.Vector3(0, 0, 0);
            if (fx) {
                var view = Utils.GetNode(building, 'View');
                if (view != null) {
                    Utils.WaitFrame(2).then(() => {
                        var x = view.transform.localRotationEulerX;
                        view.transform.localRotationEulerX = -30;
                        view.transform.localScale = LayaUtils.Vector3Zero;
                        Laya.Tween.to(view.transform, { localPositionY: 4, localScaleX: 1, localScaleY: 1, localScaleZ: 1 }, 100, Laya.Ease.expoIn);
                        Laya.Tween.to(view.transform, { localPositionY: 0 }, 100, Laya.Ease.expoIn, null, 200);
                        Laya.Tween.to(view.transform, { localRotationEulerX: -30 }, 100, Laya.Ease.expoOut);
                        Laya.Tween.to(view.transform, { localRotationEulerX: x }, 100, Laya.Ease.expoOut, null, 200);
                    });
                }
                var yanwu = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(_AllPrefabsNames.yan));
                this.sp.addChild(yanwu);
                yanwu.transform.position = building.transform.position;
                yanwu.addComponent(DelayDestory);
            }
            return building;
        }
        SpawnMotarBuilding(slotName, config, fx = false) {
            SoundManager.inst.playCreateBuilding();
            var slot = this.GetSlotNode(slotName);
            slot.removeChildren();
            var prefab = PrefabManager.getInst().GetPrefab(config.prefab);
            var building = Laya.Sprite3D.instantiate(prefab);
            slot.addChild(building);
            building.transform.localPosition = LayaUtils.Vector3Zero;
            building.transform.localRotationEuler = new Laya.Vector3(0, 0, 0);
            if (fx) {
                var view = Utils.GetNode(building, 'View');
                if (view != null) {
                    Utils.WaitFrame(2).then(() => {
                        var x = view.transform.localRotationEulerX;
                        view.transform.localRotationEulerX = -30;
                        view.transform.localScale = LayaUtils.Vector3Zero;
                        Laya.Tween.to(view.transform, { localPositionY: 4, localScaleX: 1, localScaleY: 1, localScaleZ: 1 }, 100, Laya.Ease.expoIn);
                        Laya.Tween.to(view.transform, { localPositionY: 0 }, 100, Laya.Ease.expoIn, null, 200);
                        Laya.Tween.to(view.transform, { localRotationEulerX: -30 }, 100, Laya.Ease.expoOut);
                        Laya.Tween.to(view.transform, { localRotationEulerX: x }, 100, Laya.Ease.expoOut, null, 200);
                    });
                }
                var yanwu = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(_AllPrefabsNames.yan));
                this.sp.addChild(yanwu);
                yanwu.transform.position = building.transform.position;
                yanwu.addComponent(DelayDestory);
            }
            return building;
        }
        SpawnAirportBuilding(config) {
            var slot = this.GetSlotNode('AirportBuilding');
            slot.removeChildren();
            var prefab = PrefabManager.getInst().GetPrefab(config.prefab);
            var building = Laya.Sprite3D.instantiate(prefab);
            slot.addChild(building);
            building.transform.localPosition = LayaUtils.Vector3Zero;
            building.transform.localRotationEuler = new Laya.Vector3(0, 0, 0);
            return building;
        }
        SpawnHelicopter(target) {
            PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.Helicopter_Friendly).then(p => {
                var h = Laya.Sprite3D.instantiate(p);
                this.sp.addChild(h);
                var spawnPoint = this.Find('zsjLocation');
                h.transform.position = spawnPoint.transform.position;
                h.transform.rotation = spawnPoint.transform.rotation;
                var fj = h.addComponent(Helicopter);
                fj.targetEnemyPos = target;
                fj.targetPos = new Laya.Vector3(target.x, 8, target.y);
            });
        }
        SpwanDropBox() {
            PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.jluoshan).then(p => {
                try {
                    var box = Laya.Sprite3D.instantiate(p);
                    this.sp.addChild(box);
                    var spwanP = this.Find('DropBox').transform.position;
                    spwanP.y = 30;
                    spwanP.x += (0.5 - Math.random()) * 4;
                    spwanP.z += (0.5 - Math.random()) * 4;
                    box.transform.position = spwanP;
                    box.addComponent(DropBoxBuilding);
                }
                catch (e) {
                    console.error(e);
                }
            });
        }
        SpawnDropSilverBox(pos) {
            PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.jluoshan).then(p => {
                var box = Laya.Sprite3D.instantiate(p);
                box.getChildByName("anniu").active = false;
                this.sp.addChild(box);
                var spwanP = new Laya.Vector3(pos.x + (0.5 - Math.random()) * 4, 30, pos.y + (0.5 - Math.random()) * 4);
                box.transform.position = spwanP;
                var dbb = box.addComponent(DropBoxBuilding);
                dbb.isEneyTag = false;
                dbb.isVideo = !GameSDK.isLimit() && Math.random() * 100 <= +GameSDK.data.DropBoxVideo;
            });
        }
        SpwanFreeNuke() {
            var nuke = this.Find('Nuke');
            var bomb = nuke.clone();
            this.sp.addChild(bomb);
            var spwanP = this.Find('DropBox').transform.position.clone();
            spwanP.y = 30;
            spwanP.x = (0.5 - Math.random()) * 4;
            spwanP.z = (0.5 - Math.random()) * 4;
            bomb.transform.position = spwanP;
            bomb.addComponent(FreeNukeFall);
            bomb.active = true;
        }
        SpwanAimCircle() {
            var p = this.Find('EnemyFlag').transform.position;
            var aim = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(_AllPrefabsNames.CameraAim));
            this.sp.addChild(aim);
            aim.transform.localPositionX = p.x;
            aim.transform.localPositionZ = p.z;
            aim.transform.localPositionY = 50;
            return aim;
        }
        SpwanMechaBox() {
            var node = this.Find('MechaBox');
            return new Promise(r => {
                PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.DropPod).then(prefab => {
                    var pod = Laya.Sprite3D.instantiate(prefab);
                    node.addChild(pod);
                    pod.transform.localPosition = LayaUtils.Vector3Zero;
                    var mb = pod.addComponent(MechaBoxComponent);
                    r(mb);
                });
            });
        }
        SpwanHangMu() {
            var nodeP = this.Find('HangMu');
            return PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.HangMuView).then(prefab => {
                var node = Laya.Sprite3D.instantiate(prefab);
                node.transform.localPosition = LayaUtils.Vector3Zero;
                nodeP.addChild(node);
                node.transform.localRotationEuler = new Laya.Vector3(0, 180, 0);
                var water = Utils.GetNode(node, 'Water');
                water.destroy(true);
                var hangmu = node.getChildAt(0);
                hangmu.transform.localRotationEulerY = 0;
                var paotai = hangmu.getChildAt(1);
                paotai.getChildAt(0).addComponent(RotateYYoYo);
                paotai.getChildAt(1).addComponent(RotateYYoYo);
                paotai.getChildAt(2).addComponent(RotateYYoYo);
                return node;
            });
        }
    }
    class BuildingInfo {
        constructor() {
            this.trigerPolygon = [];
            this.polygon = [];
        }
    }

    let GuideRecorder = class GuideRecorder extends DataProxy {
    };
    GuideRecorder = __decorate([
        data_bind('guide')
    ], GuideRecorder);
    class GuideData {
        static get recoder() {
            if (this._recorder == null) {
                this._recorder = DataManager.inst.GetDataRecorder('guide');
            }
            return this._recorder;
        }
        static get Step() {
            return this.recoder.step || 0;
        }
        static set Step(v) {
            this.recoder.step = v;
            this.changeStep();
        }
        static changeStep() {
            var open = true;
            switch (this.Step) {
                case GuideStepType.收集银牌:
                    ArrowTips.inst.Show();
                    break;
                case GuideStepType.建立兵营:
                    ArrowTips.inst.Show();
                    ArrowTips.inst.target = MapCampManagerComponent.inst.GetSlotNode('TentLocation1');
                    break;
                case GuideStepType.升级提示:
                    ArrowTips.inst.Show();
                    var tagName = 'PlayerLocation';
                    if (BattleData.GoldTag < PlayerData.CurrentCommanderConfig.cost_gold) {
                        tagName = 'TagLocation';
                    }
                    ArrowTips.inst.target = MapCampManagerComponent.inst.GetSlotNode(tagName);
                    break;
                case GuideStepType.开始战斗:
                    ArrowTips.inst.Show();
                    ArrowTips.inst.target = MapCampManagerComponent.inst.GetSlotNode('ChargeArea');
                    break;
                case GuideStepType.战斗中_隐藏Tips:
                    open = false;
                    ArrowTips.inst.Hide();
                    break;
                case GuideStepType.银牌升级:
                    ArrowTips.inst.Show();
                    var tent = MapCampManagerComponent.inst.GetSlotNode('TentLocation1');
                    var target = Utils.GetNode(tent, 'Unlocked');
                    if (target == null) {
                        ArrowTips.inst.Hide();
                    }
                    ArrowTips.inst.target = target;
                    break;
                case GuideStepType.夺取旗帜:
                    ArrowTips.inst.Show();
                    ArrowTips.inst.target = MapCampManagerComponent.inst.GetSlotNode('EnemyFlag');
                    break;
                case GuideStepType.第二关收集:
                    ArrowTips.inst.Show();
                    ArrowTips.inst.target = MapCampManagerComponent.inst.GetSlotNode('TagLocation');
                    break;
                case GuideStepType.拾取免费核弹:
                    ArrowTips.inst.Show();
                    break;
                case GuideStepType.使用免费核弹:
                    ArrowTips.inst.Hide();
                    break;
                default:
                    open = false;
                    ArrowTips.inst.Hide();
                    break;
            }
            MsgManager.getInst().event(BattleEvent.SetGuideStep, this.Step, open);
        }
    }
    var GuideStepType;
    (function (GuideStepType) {
        GuideStepType[GuideStepType["\u6536\u96C6\u94F6\u724C"] = 0] = "\u6536\u96C6\u94F6\u724C";
        GuideStepType[GuideStepType["\u5EFA\u7ACB\u5175\u8425"] = 1] = "\u5EFA\u7ACB\u5175\u8425";
        GuideStepType[GuideStepType["\u5347\u7EA7\u63D0\u793A"] = 2] = "\u5347\u7EA7\u63D0\u793A";
        GuideStepType[GuideStepType["\u5F00\u59CB\u6218\u6597"] = 3] = "\u5F00\u59CB\u6218\u6597";
        GuideStepType[GuideStepType["\u6218\u6597\u4E2D_\u9690\u85CFTips"] = 4] = "\u6218\u6597\u4E2D_\u9690\u85CFTips";
        GuideStepType[GuideStepType["\u94F6\u724C\u5347\u7EA7"] = 5] = "\u94F6\u724C\u5347\u7EA7";
        GuideStepType[GuideStepType["\u62FE\u53D6\u514D\u8D39\u6838\u5F39"] = 6] = "\u62FE\u53D6\u514D\u8D39\u6838\u5F39";
        GuideStepType[GuideStepType["\u4F7F\u7528\u514D\u8D39\u6838\u5F39"] = 7] = "\u4F7F\u7528\u514D\u8D39\u6838\u5F39";
        GuideStepType[GuideStepType["\u593A\u53D6\u65D7\u5E1C"] = 8] = "\u593A\u53D6\u65D7\u5E1C";
        GuideStepType[GuideStepType["\u7B2C\u4E00\u5173\u5B8C\u6210"] = 9] = "\u7B2C\u4E00\u5173\u5B8C\u6210";
        GuideStepType[GuideStepType["\u7B2C\u4E8C\u5173\u6536\u96C6"] = 10] = "\u7B2C\u4E8C\u5173\u6536\u96C6";
        GuideStepType[GuideStepType["None"] = 11] = "None";
    })(GuideStepType || (GuideStepType = {}));

    let FontDataHelper = class FontDataHelper extends Singleton {
        constructor() {
            super(...arguments);
            this.fontData = {};
        }
        _init_() {
            this.rawData.Font3D.values.forEach(v => {
                this.fontData[v[0]] = {
                    x: v[1],
                    y: v[2],
                    xoffset: v[3],
                    yoffset: v[4],
                    xadvance: v[5],
                    width: v[6],
                    height: v[7],
                    chnl: v[8]
                };
            });
        }
        getFontData(name) {
            return this.fontData[name];
        }
    };
    __decorate([
        res_assign('res/font3D/font.png')
    ], FontDataHelper.prototype, "font", void 0);
    __decorate([
        res_assign('res/font3D/font_red.png')
    ], FontDataHelper.prototype, "font_red", void 0);
    __decorate([
        res_assign('res/ConfigJson/game_data.json')
    ], FontDataHelper.prototype, "rawData", void 0);
    FontDataHelper = __decorate([
        res_auto_bind()
    ], FontDataHelper);

    class Vertexo {
        constructor() {
            this.x = 0;
            this.y = 0;
            this.z = 0;
            this.nx = 0;
            this.ny = 0;
            this.nz = 0;
            this.u = 0;
            this.v = 0;
        }
    }
    class MeshData {
        constructor() {
            this.vbArray = [];
            this.width = 0;
            this.vertexs = [];
            this.triangle = [];
        }
        addVertex(x, y, z, nx, ny, nz, u, v) {
            var vertex = new Vertexo();
            vertex.x = x;
            vertex.y = y;
            vertex.z = z;
            vertex.nx = nx;
            vertex.ny = ny;
            vertex.nz = nz;
            vertex.u = u;
            vertex.v = v;
            this.vertexs.push(vertex);
        }
        combineMesh(list) {
            for (var i = 0; i < list.length; i++) {
                var data = list[i];
                var begin = this.vertexs.length;
                for (var j = 0; j < data.vertexs.length; j++) {
                    this.vertexs.push(data.vertexs[j]);
                }
                for (var j = 0; j < data.triangle.length; j++) {
                    this.triangle.push(data.triangle[j] + begin);
                }
            }
            return this;
        }
        createMesh() {
            var vertexDeclaration = Laya.VertexMesh.getVertexDeclaration("POSITION,NORMAL,UV");
            for (var i = 0; i < this.vertexs.length; i++) {
                var ver = this.vertexs[i];
                this.vbArray.push(ver.x, ver.y, ver.z, ver.nx, ver.ny, ver.nz, ver.u, ver.v);
            }
            var vertices = new Float32Array(this.vbArray);
            var indices = new Uint16Array(this.triangle);
            return Laya.PrimitiveMesh._createMesh(vertexDeclaration, vertices, indices);
        }
        get mesh() {
            return this.createMesh();
        }
        destroy() {
            this.vbArray = null;
            this.vertexs = null;
            this.triangle = null;
        }
    }
    class TextFieldScript3D2 extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.isStart = false;
            this.isLoadComplete = true;
            this.fontName = "";
            this.align = "left";
            this.localScale = null;
            this.localPosition = null;
            this.mat = null;
            this._mesh = null;
            this._text = "";
            this.textWidth = 1024;
            this.textHeight = 113;
        }
        onAwake() {
            this.localScale = this.transform.localScale.clone();
            this.localPosition = this.transform.localPosition.clone();
            this.mat = this.owner.meshRenderer.material;
            this.originX = this.transform.localPositionX;
            this.originY = this.transform.localPositionY;
            this.originZ = this.transform.localPositionZ;
        }
        initFonts(fontName, align = "center") {
            this.fontName = fontName;
            this.align = align;
            [
                "res/font3D/" + this.fontName + ".png"
            ];
            var text = FontDataHelper.getInst()[fontName];
            this.mat = this.owner.meshRenderer.material;
            this.loadComplete(text);
        }
        onStart() {
            if (this.isLoadComplete) {
                this.showFonts();
            }
            this.isStart = true;
        }
        set text(value) {
            if (this._text === value)
                return;
            this._text = value;
            if (this.isLoadComplete && this.isStart) {
                this.showFonts();
            }
        }
        get text() {
            return this._text;
        }
        showFonts() {
            if (this._text == "")
                return;
            let fonts = this._text;
            let lists = [];
            this.getFontDatasAndWidth(fonts, lists) + 2;
            let meshs = [];
            for (var i = 0; i < lists.length; i++) {
                meshs.push(this.createMesh(lists[i]));
            }
            let b = 0;
            for (var i = 0; i < meshs.length; i++) {
                for (var j = 0; j < meshs[i].vertexs.length; j++) {
                    meshs[i].vertexs[j].x += b;
                }
                b -= (meshs[i].width + 0.02);
            }
            meshs[0].combineMesh(meshs.slice(1, meshs.length));
            this._mesh = meshs[0].mesh;
            var bounds = this._mesh.bounds;
            this.transform.localPositionX = this.originX - bounds.getCenter().x;
            this.transform.localPositionZ = this.originZ - bounds.getCenter().z;
            this.transform.localPositionY = this.originY - bounds.getCenter().y;
            this.owner.meshFilter.sharedMesh = this._mesh;
            if (this.oldMesh != null) {
                this.oldMesh.destroy();
            }
            this.oldMesh = this._mesh;
        }
        createMesh(data) {
            var mesh = new MeshData();
            let L = 0.5;
            let R = 0.5 - (data.width / 60);
            let U = ((data.height + data.yoffset) / 60) - 0.5;
            let D = -0.5 + (data.yoffset) / 60;
            let step = 1 / this.textWidth * 0.5;
            mesh.addVertex(L, U, 0, 0, 0, 0, data.x / this.textWidth + step, data.y / this.textHeight + step);
            mesh.addVertex(R, U, 0, 0, 0, 0, (data.x + data.width) / this.textWidth + step, data.y / this.textHeight + step);
            mesh.addVertex(R, D, 0, 0, 0, 0, (data.x + data.width) / this.textWidth + step, (data.y + data.height) / this.textHeight + step);
            mesh.addVertex(L, D, 0, 0, 0, 0, data.x / this.textWidth + step, (data.y + data.height) / this.textHeight + step);
            mesh.triangle.push(0, 1, 2);
            mesh.triangle.push(2, 3, 0);
            mesh.width = Math.abs(L - R);
            return mesh;
        }
        loadComplete(tex) {
            this.isLoadComplete = true;
            this.mat.albedoTexture = tex;
            this.textHeight = tex.height;
            this.textWidth = tex.width;
            if (this.isStart) {
                this.showFonts();
            }
        }
        getBLine(allWidth, texWidth) {
            let bLine = 1;
            switch (this.align) {
                case "left":
                    break;
                case "center":
                    bLine = Math.floor((texWidth - allWidth) / 2);
                    break;
                case "right":
                    bLine = texWidth - allWidth;
                    break;
            }
            return bLine;
        }
        getTextPosition() {
            let pos = this.localPosition.clone();
            switch (this.align) {
                case "left":
                    pos.x = (this.transform.localScale.x - this.localScale.x) / 2 + pos.x;
                    break;
                case "center":
                    break;
                case "right":
                    pos.x = pos.x - (this.transform.localScale.x - this.localScale.x) / 2;
                    break;
            }
            return pos;
        }
        getFontDatasAndWidth(fonts, list) {
            let width = 0;
            for (var i = 0; i < fonts.length; i++) {
                let data = FontDataHelper.getInst().getFontData(this.fontName + "_" + fonts.charAt(i).charCodeAt(0));
                if (data) {
                    list.push(data);
                    width += data.width;
                }
            }
            return width;
        }
        onDestroy() {
            if (this._mesh) {
                this._mesh.destroy();
                this._mesh = null;
            }
        }
    }

    class AirportBuilding {
        constructor() {
            this.inTriggerTime = 0;
            this.isTrigger = false;
            this.inTagTriggerTime = 0;
            this.isTagTrigger = false;
            this.planePos = new Laya.Vector2();
            this.triggerPosAD = new Laya.Vector2();
            this.triggerPos = new Laya.Vector2();
            this.isReady = false;
            this.slotName = 'AirportBuilding';
            this.state = BattleData.Airport;
            this.init();
        }
        init() {
            this.config = ConfigHelper.BuildingMap[BattleData.Airport.id];
            var node = this.node = MapCampManagerComponent.inst.SpawnAirportBuilding(this.config);
            this.tag = Utils.GetNode(node, 'Tag');
            this.initYuanJian();
            this.initProgress();
            this.WeaponCrates = Utils.GetNode(node, 'WeaponCrates');
            var triggerNode = Utils.GetNode(node, 'triggerArea');
            this.adTips = Utils.GetNode(this.node, 'anniu_03');
            if (triggerNode) {
                var ps = [];
                for (var i = 0; i < triggerNode.numChildren; i++) {
                    var c = triggerNode.getChildAt(i);
                    ps.push(new Laya.Vector2(c.transform.position.x, c.transform.position.z));
                }
                this.triggerArea = new PolygonWrap(ps);
            }
            var wenzi = Utils.GetNode(node, 'WenZi');
            if (wenzi) {
                this.wenziP = wenzi.addComponent(TextFieldScript3D2);
                this.wenziP.initFonts('font');
            }
            wenzi = Utils.GetNode(node, 'Time');
            if (wenzi) {
                this.wenzi = wenzi.addComponent(TextFieldScript3D2);
                this.wenzi.initFonts('font_red');
            }
            this.udpateState();
            this.SetReady(ADData.Bomber > 0);
            var button = Utils.GetNode(this.node, 'anniu');
            if (button) {
                this.triggerPosAD.setValue(button.transform.position.x, button.transform.position.z);
            }
            var planePos = Utils.GetNode(this.node, 'WarPlanePos');
            if (planePos != null) {
                this.planePos.setValue(planePos.transform.position.x, planePos.transform.position.z);
                this.planePos.cloneTo(this.triggerPos);
                PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.WarPlane).then(p => {
                    var plane = Laya.Sprite3D.instantiate(p);
                    planePos.addChild(plane);
                    plane.transform.localPosition = LayaUtils.Vector3Zero;
                    plane.transform.localRotationEuler = LayaUtils.Vector3Zero;
                    var smoke = Utils.GetNode(this.node, 'SmokeDark');
                    if (smoke) {
                        smoke.active = false;
                    }
                });
            }
            this.udpateShow();
            this.CheckUpdate();
        }
        DeltaWeapon() {
            this.WeaponCrates.numChildren;
            var wn = this.WeaponCrates.getChildAt(BattleData.WeaponCountInFactory);
            wn.removeChildren();
        }
        checkInArea(p, useFast = false) {
            if (this.triggerArea) {
                return MathUtils.PointInPolygon2(p, this.triggerArea.points);
            }
            return false;
        }
        UpdateTrigger(deltaTime, playerPos) {
            if (this.checkInArea(playerPos)) {
                this.inTriggerTime += deltaTime;
                if (this.inTriggerTime >= 0.2) {
                    this.inTriggerTime = 0;
                    if (this.isTrigger == false) {
                        this.isTrigger = true;
                        this.EnterTrigger();
                    }
                }
            }
            else {
                this.inTriggerTime = 0;
                if (this.isTrigger) {
                    this.isTrigger = false;
                    this.ExitTrigger();
                }
            }
            if (this.state.level > 0 && this.node.active) {
                if (ADData.Bomber <= 0) {
                    if (LayaUtils.distanceSquaredVec2(playerPos, this.planePos) < 4) {
                        this.StartEngine();
                        ADData.Bomber = ADData.initCD;
                        return;
                    }
                }
                if (LayaUtils.distanceSquaredVec2(playerPos, this.triggerPosAD) < 1) {
                    if (ADData.Bomber <= 0) {
                        this.StartEngine();
                        ADData.Bomber = ADData.initCD;
                    }
                    else {
                        GameSDK.showADVideo(this, b => {
                            GameManager.getInst().battleManager.entiytManager.player.Pos.x -= 2;
                            if (b) {
                                this.StartEngine();
                                ADData.Bomber = ADData.initCD;
                                this.isReady = true;
                            }
                            else {
                                TipsMgr.Tips('看完整视频得奖励');
                            }
                        }, '3D按钮轰炸机');
                    }
                }
            }
            if (ADData.Bomber > 0) {
                this.adTips.active = true;
                if (this.isReady) {
                    this.isReady = false;
                    this.SetReady(false);
                }
            }
            else {
                this.adTips.active = false;
                if (!this.isReady) {
                    this.isReady = true;
                    this.SetReady(true);
                }
            }
            if (this.wenzi) {
                this.wenzi.text = Utils.TimeToString(ADData.Bomber);
            }
        }
        SetReady(v) {
            if (v) {
                this.ResetEngine();
            }
        }
        EnterTrigger() {
            if (this.state.unlocked) {
                this.GetTagFromPlayer();
            }
        }
        ExitTrigger() {
        }
        GetTagFromPlayer() {
            if (BattleData.SilverTag > 0) {
                this.state.tag++;
                BattleData.SilverTag--;
                MsgManager.getInst().event(BattleEvent.DeleteTagFromPlayer, false, this.tag);
                this.updateProgress();
                if (this.CheckUpdate()) {
                    SoundManager.inst.playVibrate(false);
                    this.udpateState();
                    return;
                }
                if (this.isTrigger) {
                    Laya.timer.once(80, this, this.GetTagFromPlayer);
                }
            }
        }
        CheckUpdate() {
            if (this.state.tag >= this.config.cost_coin) {
                this.GetNextState();
                return true;
            }
        }
        GetNextState() {
            var level = this.config.level + 1;
            var bs = ConfigHelper.CurrentLevelBuildingMap[this.state.type][this.slotName];
            var building;
            var maxLevel = level;
            bs.forEach(b => {
                if (level == b.level) {
                    building = b;
                }
                maxLevel = Math.max(maxLevel, b.level);
            });
            if (building != null) {
                this.config = building;
                this.state.level = level;
                this.state.id = building.id;
                this.state.tag = 0;
                this.state.unlocked = (level == maxLevel);
                this.init();
                return true;
            }
        }
        udpateState() {
            this.updateLocked();
            this.updateProgress();
            this.updateRandk();
        }
        updateRandk() {
            var ranks = Utils.GetNode(this.node, 'Ranks');
            if (ranks == null)
                return;
            var icons = PrefabManager.getInst().GetPrefab(_AllPrefabsNames.RankIcons);
            var icon = icons.getChildByName(ConfigHelper.CommanderConfigLevelMap[this.config.show_unlock].level + '');
            if (icon != null) {
                ranks.meshRenderer.material = icon.meshRenderer.sharedMaterial;
            }
        }
        initYuanJian() {
            var p = Utils.GetNode(this.node, 'YuanJiao');
            if (p == null)
                return;
            var mat = p.meshRenderer.material = new VertexUVMaterial();
            var c = Utils.hexToRgba01("#545454");
            c.w = 0.6;
            mat.mainColor = c;
            mat.progress = 1;
            mat.radius = 0.1;
        }
        initProgress() {
            var p = Utils.GetNode(this.node, 'JinDu');
            if (p == null)
                return;
            var mat = this.progress = p.meshRenderer.material = new VertexUVMaterial();
            var c = Utils.hexToRgba01("#24AAF8");
            c.w = 1;
            mat.mainColor = c;
            mat.progress = 0;
            mat.radius = 0.1;
            this.updateProgress();
        }
        updateProgress() {
            var max = this.config.cost_coin;
            var value = this.state.tag;
            if (this.progress) {
                this.progress.progress = value / max;
            }
            if (this.wenziP) {
                this.wenziP.text = `${value}/${max}`;
            }
        }
        updateLocked() {
            this.state.unlocked = this.config.show_unlock <= PlayerData.CurrentRankLevel;
            var showUnlcoked = this.config.show_unlock <= PlayerData.CurrentRankLevel;
            var showLocked = this.config.show_lock <= PlayerData.CurrentRankLevel;
            if (showUnlcoked) {
                showLocked = false;
            }
            var locked = Utils.GetNode(this.node, 'Locked');
            if (locked) {
                locked.active = showLocked;
            }
            var unlocked = Utils.GetNode(this.node, 'Unlocked');
            if (unlocked) {
                unlocked.active = showUnlcoked;
            }
            if (showLocked) {
                this.CheckUpdate();
            }
        }
        udpateShow() {
            this.config.show_unlock <= PlayerData.CurrentRankLevel;
            this.config.show_lock <= PlayerData.CurrentRankLevel;
            if (PlayerData.CurrentLevelID > 1) {
                this.node.active = true;
            }
            else {
                this.node.active = false;
            }
            return;
        }
        ResetEngine() {
            var smoke = Utils.GetNode(this.node, 'SmokeDark');
            if (smoke) {
                smoke.active = false;
            }
            var BombIcons = Utils.GetNode(this.node, 'BombIcons');
            if (BombIcons) {
                BombIcons.transform.localPositionY = 0.1;
                for (var i = 0; i < BombIcons.numChildren; i++) {
                    var c = BombIcons.getChildAt(i).getChildAt(0);
                    c.active = true;
                }
            }
        }
        StartEngine() {
            ArrowTips.inst.Hide();
            var BombIcons = Utils.GetNode(this.node, 'BombIcons');
            var p = Utils.GetNode(this.node, 'AirportPlaneEnterInteractable');
            var tp = p.transform.position;
            for (var i = 0; i < BombIcons.numChildren; i++) {
                var c = BombIcons.getChildAt(i).getChildAt(0);
                var t = Laya.Tween.to(c.transform.position, { x: tp.x, y: tp.y, z: tp.z }, 200, Laya.Ease.linearIn, Laya.Handler.create(this, this._endBullet, [c]), i * 30, false);
                t.update = Laya.Handler.create(this, this._updateBullet, [c], false);
            }
            MsgManager.getInst().event(CtrlEvent.StartDriver, Utils.GetNode(this.node, 'WarPlane'), this.node);
        }
        _endBullet(bullet) {
            bullet.transform.localPosition = LayaUtils.Vector3Zero;
            bullet.active = false;
        }
        _updateBullet(bullet) {
            bullet.transform.position = bullet.transform.position;
        }
    }

    var tmp$5 = new Laya.Vector2();
    var tmp2$1 = new Laya.Vector2();
    class SteeringBahaviorUnit {
        constructor() {
            this.closeCollision = false;
            this.moveable = true;
            this.lazyMove = false;
            this.radius = 0.5;
            this.maxSpeed = 2;
            this.maxForce = 30;
            this.pos = new Laya.Vector2();
            this.velocity = new Laya.Vector2();
            this.angle = 0;
            this.formationAngle = 0;
            this.steeringForce = new Laya.Vector2();
            this.forceMove = false;
            this.head = new Laya.Vector2;
            this.attachRadius = 0;
            this.deltaPos = new Laya.Vector2();
            this.needMoveTimer = 1;
            this.needMove = true;
        }
        update(deltaTime, others) {
            if (this.forceMove) {
                this.updateDesiredVelocity();
            }
            else {
                if (this.moveable) {
                    this.updateDesiredForce();
                }
            }
            this.updateSteeringBeta(deltaTime, others);
        }
        updateDesiredForce() {
            if (this.target == null)
                return;
            var distance = LayaUtils.distanceVec2(this.pos, this.target);
            if (distance <= this.attachRadius || this.lazyMove || (this.target == null)) ;
            else {
                LayaUtils.subtractVec2(this.target, this.pos, this.steeringForce);
                LayaUtils.truncateVec2(this.steeringForce, this.maxForce, this.steeringForce);
            }
        }
        updateDesiredVelocity() {
            if (this.target == null)
                return;
            var distance = LayaUtils.distanceSquaredVec2(this.pos, this.target);
            if (distance < 0.03) {
                this.needMoveTimer = 0.5;
                this.needMove = false;
                this.mass = this.originalMass;
                this.velocity.setValue(0, 0);
                this.pos.setValue(this.target.x, this.target.y);
                this.angle = this.formationAngle;
            }
            else {
                if (this.needMove == false) {
                    this.needMoveTimer -= LayaUtils.deltaTimeSec;
                    if (this.needMoveTimer <= 0) {
                        this.needMove = true;
                        this.mass = 100;
                    }
                }
                if (this.needMove == false)
                    return;
                this.mass = 100;
                LayaUtils.subtractVec2(this.target, this.pos, this.velocity);
                LayaUtils.setLengthVec2(this.velocity, this.maxSpeed, this.velocity);
            }
        }
        updateSteeringBeta(deltaTime, others) {
            var steering = this.steeringForce;
            var head = this.head;
            LayaUtils.setLengthVec2(this.velocity, this.radius * 0.1, head);
            LayaUtils.addVec2(this.pos, head, head);
            if (!this.closeCollision) {
                var obstacles = [];
                for (var i = 0; i < others.length; i++) {
                    var other = others[i];
                    if (other.id == this.id || other.closeCollision)
                        continue;
                    let dis = LayaUtils.distanceVec2(other.pos, head);
                    let sq = this.radius + other.radius;
                    if (dis < sq) {
                        obstacles.push(new ObstacleHitInfo(other, sq, dis));
                    }
                }
                obstacles.forEach(obstacle => {
                    let pos = tmp2$1;
                    LayaUtils.subtractVec2(head, obstacle.unit.pos, pos);
                    LayaUtils.normalizeVec2(pos, pos);
                    var deltaDis = obstacle.limitDis - obstacle.realDis;
                    LayaUtils.setLengthVec2(pos, deltaDis, pos);
                    var selftP = obstacle.unit.mass / (obstacle.unit.mass + this.mass);
                    var otherP = this.mass / (obstacle.unit.mass + this.mass);
                    obstacle.unit.pos.x -= otherP * pos.x;
                    obstacle.unit.pos.y -= otherP * pos.y;
                    this.pos.x += selftP * pos.x;
                    this.pos.y += selftP * pos.y;
                });
            }
            if (!LayaUtils.IsZeroVec2(this.velocity)) {
                if (LayaUtils.IsZeroVec2(steering)) {
                    LayaUtils.scaleVec2(this.velocity, -1, tmp$5);
                    LayaUtils.setLengthVec2(tmp$5, 4, tmp$5);
                    LayaUtils.addVec2(tmp$5, steering, steering);
                }
            }
        }
        UpdatePoseBeta(deltaTime) {
            if (!this.forceMove) {
                var steering = this.steeringForce;
                LayaUtils.scaleVec2(steering, deltaTime, steering);
                LayaUtils.addVec2(this.velocity, steering, this.velocity);
            }
            LayaUtils.truncateVec2(this.velocity, this.maxSpeed, this.velocity);
            var deltaPosition = this.deltaPos;
            LayaUtils.scaleVec2(this.velocity, deltaTime, deltaPosition);
            if (!LayaUtils.IsZeroVec2(this.velocity)) {
                var angle = Laya.MathUtil.getRotation(0, 0, -this.velocity.x, this.velocity.y);
                if (angle < 0) {
                    angle += 360;
                }
                angle += 90;
                if (angle > 360) {
                    angle -= 360;
                }
                this.angle = angle;
            }
            this.steeringForce.setValue(0, 0);
        }
    }
    class ObstacleHitInfo {
        constructor(unit, limitDis, realDis) {
            this.unit = unit;
            this.limitDis = limitDis;
            this.realDis = realDis;
        }
    }

    new Laya.Vector3();
    class SteeringBehaviorManager extends Singleton {
        constructor() {
            super();
            this.unitMap = {};
        }
        updateUnits(elapseSeconds) {
            var vs = Object.values(this.unitMap);
            for (let key in this.unitMap) {
                this.unitMap[key].update(elapseSeconds, vs);
            }
            for (let key in this.unitMap) {
                this.unitMap[key].UpdatePoseBeta(elapseSeconds);
            }
        }
        CreateUnit(id, radius, speed, mass, x = 0, y = 0) {
            var unit = new SteeringBahaviorUnit();
            unit.id = id;
            unit.mass = mass;
            unit.originalMass = mass;
            unit.pos.setValue(x, y);
            unit.radius = radius;
            unit.maxSpeed = speed;
            this.AddUnit(unit);
            return unit;
        }
        DeleteUnit(unit) {
            delete this.unitMap[unit.id];
        }
        AddUnit(unit) {
            this.unitMap[unit.id] = unit;
        }
    }

    class MoveAndDestroy extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.veclocity = new Laya.Vector3();
            this.deltaPos = new Laya.Vector3();
        }
        onStart() {
            this.sp.transform.lookAt(this.target, LayaUtils.UP);
            this.maxLen = Laya.Vector3.distance(this.transform.position, this.target);
            Laya.Vector3.subtract(this.target, this.transform.position, this.veclocity);
            Laya.Vector3.normalize(this.veclocity, this.veclocity);
            Laya.Vector3.scale(this.veclocity, this.speed, this.veclocity);
        }
        onUpdate() {
            Laya.Vector3.scale(this.veclocity, LayaUtils.deltaTimeSec, this.deltaPos);
            this.maxLen -= Laya.Vector3.scalarLength(this.deltaPos);
            if (this.maxLen <= 0) {
                this.sp.destroy(true);
                return;
            }
            Laya.Vector3.add(this.transform.position, this.deltaPos, this.deltaPos);
            this.transform.position = this.deltaPos;
        }
    }

    class IFsmState {
    }

    class FsmState extends IFsmState {
        _OnInit(fsm) {
            this.fsm = fsm;
            this.OnInit(fsm);
        }
        ChangeState(fsm, state) {
            fsm.ChangeState(state);
        }
    }

    const maxSeekDuration = 1;
    class WanderState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'WanderState';
            this.seekCD = 0;
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            fsm.Owner.setMoveFlag(true);
            fsm.Owner.PlayAni(fsm.Owner.weaponConfig.run, 2 / fsm.Owner.config.speed);
            fsm.Owner.steerBehaviorUnit.mass = 500;
            this.seekCD = 0;
        }
        OnUpdate(fsm, elapseSeconds) {
            this.seekCD -= elapseSeconds;
            if (this.seekCD <= 0) {
                this.seekCD = maxSeekDuration;
                if (fsm.Owner.battleManager.isBattleState) {
                    fsm.Owner.SeekEnemy();
                }
                else {
                    fsm.Owner.EnemySeekPlayer();
                }
            }
            if (fsm.Owner.CanAttackAimEnemy()) {
                fsm.ChangeState(AttackState);
            }
            else if (fsm.Owner.aimTarget == null && !fsm.Owner.battleManager.isBattleState) {
                fsm.ChangeState(FormationAndSeekState);
            }
        }
        OnLeave(fsm) {
            fsm.Owner.steerBehaviorUnit.mass = fsm.Owner.steerBehaviorUnit.originalMass;
        }
        OnDestroy(fsm) {
        }
    }

    class FormationAndSeekState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'FormationAndSeekState';
            this.cd = 0.3;
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            this.cd = 0;
            this.originSeekRadiusSq = fsm.Owner.seekRadiusSq;
            fsm.Owner.seekRadiusSq = fsm.Owner.attackRadiusSq;
            fsm.Owner.SetFormationPos();
            fsm.Owner.setMoveFlag(true);
            fsm.Owner.steerBehaviorUnit.forceMove = true;
        }
        OnUpdate(fsm, elapseSeconds) {
            if (fsm.Owner.config.damage == 0)
                return;
            this.cd -= elapseSeconds;
            if (this.cd < 0) {
                this.cd = 0.3;
                if (fsm.Owner.battleManager.isBattleState) {
                    fsm.Owner.SeekEnemy();
                }
                else {
                    fsm.Owner.EnemySeekPlayer();
                }
                if (fsm.Owner.HasAminTarget()) {
                    fsm.ChangeState(WanderState);
                }
                else {
                    fsm.Owner.SetFormationPos();
                    fsm.Owner.setMoveFlag(true);
                }
            }
        }
        OnLeave(fsm) {
            MsgManager.getInst().offAllCaller(this);
            fsm.Owner.steerBehaviorUnit.forceMove = false;
            fsm.Owner.seekRadiusSq = this.originSeekRadiusSq;
        }
        OnDestroy(fsm) {
        }
    }

    var tmp$4 = new Laya.Vector3();
    class AttackState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'AttackState';
            this.attDuration = 0;
            this.seekDuration = 0;
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            this.attDuration = 0;
            this.seekDuration = 0.2;
            fsm.Owner.steerBehaviorUnit.lazyMove = true;
            fsm.Owner.steerBehaviorUnit.velocity.setValue(0, 0);
        }
        OnUpdate(fsm, elapseSeconds) {
            if (fsm.Owner.aniView.sp == null) {
                if (fsm.Owner.aniView.sp && fsm.Owner.aniView.sp.active && fsm.Owner.aniView.shootFx != null) {
                    fsm.Owner.aniView.shootFx.active = false;
                }
                return;
            }
            fsm.Owner.steerBehaviorUnit.velocity;
            this.attDuration -= elapseSeconds;
            this.seekDuration -= elapseSeconds;
            if (this.seekDuration <= 0) {
                this.seekDuration += 0.3;
                if (fsm.Owner.CanAttackAimEnemy() == false) {
                    switch (fsm.Owner.camp) {
                        case CampType.Blue:
                            fsm.ChangeState(WanderState);
                            break;
                        case CampType.Red:
                            fsm.ChangeState(FormationAndSeekState);
                            break;
                    }
                    return;
                }
            }
            if (fsm.Owner.aimTarget != null && fsm.Owner.aimTarget.camp == CampType.Red) {
                fsm.Owner.aimTarget.AlertEnemy(fsm.Owner);
            }
            if (this.attDuration < 0) {
                fsm.Owner.PlayAni(fsm.Owner.weaponConfig.shoot, fsm.Owner.attackDuration, true, fsm.Owner.isPlayer);
                SoundManager.inst.playShoot(fsm.Owner.weaponConfig.sound);
                if (fsm.Owner.aniView.sp && fsm.Owner.aniView.sp.active && fsm.Owner.aniView.shootFx != null) {
                    fsm.Owner.aniView.shootFx.active = true;
                    if (fsm.Owner.weaponConfig.id != 4) {
                        Utils.WaitTime(fsm.Owner.attackDuration * 100).then(() => {
                            fsm.Owner.aniView.shootFx.active = false;
                        });
                    }
                    else {
                        Utils.WaitTime(fsm.Owner.attackDuration * 1000).then(() => {
                            fsm.Owner.aniView.shootFx.active = false;
                        });
                    }
                }
                this.attDuration = fsm.Owner.attackDuration;
                if (fsm.Owner.attackRadius > 2.5) {
                    var miss = Math.random() < 0.5;
                    var tartgetPos;
                    if (miss) {
                        Laya.Vector3.subtract(fsm.Owner.aimTarget.PosV3, fsm.Owner.PosV3, tmp$4);
                        Laya.Vector3.scale(tmp$4, 0.5 * Math.random() + 1.1, tmp$4);
                        Laya.Vector3.add(fsm.Owner.aimTarget.PosV3, tmp$4, tmp$4);
                        tmp$4.y = 0;
                        tartgetPos = tmp$4.clone();
                    }
                    else {
                        tartgetPos = fsm.Owner.aimTarget.PosV3.clone();
                        tartgetPos.y += 1.2;
                        fsm.Owner.Attack(fsm.Owner.aimTarget, 0.2);
                    }
                    if (fsm.Owner.aniView.sp && fsm.Owner.aniView.sp.active) {
                        var bullet = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(_AllPrefabsNames.Bullet));
                        LevelManager.getInst().scene.addChild(bullet);
                        var ft = bullet.addComponent(MoveAndDestroy);
                        ft.target = tartgetPos;
                        ft.speed = 30;
                        bullet.transform.position = fsm.Owner.emiterPos;
                    }
                }
                else {
                    fsm.Owner.Attack(fsm.Owner.aimTarget);
                }
                if (fsm.Owner.battleManager.isBattleState) {
                    fsm.Owner.SeekEnemy();
                }
                else {
                    fsm.Owner.EnemySeekPlayer();
                }
            }
        }
        OnLeave(fsm) {
            if (fsm.Owner.aniView.sp && fsm.Owner.aniView.sp.active && fsm.Owner.aniView.shootFx != null) {
                fsm.Owner.aniView.shootFx.active = false;
            }
            fsm.Owner.steerBehaviorUnit.lazyMove = false;
        }
        OnDestroy(fsm) {
        }
    }

    class CastSkillState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'CastSkillState';
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
        }
        OnUpdate(fsm, elapseSeconds) {
        }
        OnLeave(fsm) {
        }
        OnDestroy(fsm) {
        }
    }

    class DogTagEntity {
        constructor(state, isEnemy = false) {
            this.pos = new Laya.Vector2(state.x, state.z);
            this.y = state.y;
            this.r = state.r;
            this.isEnemy = isEnemy;
            this.state = state;
        }
    }

    class DeathState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'DeathState';
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            fsm.Owner.isDeath = true;
            if (fsm.Owner.becovered == null) {
                var state = BattleData.AddTag(fsm.Owner.Pos.x + 0.5 - Math.random(), 0, fsm.Owner.Pos.y + 0.5 - Math.random(), Math.random() * 360, fsm.Owner.camp == CampType.Red);
                var tag = new DogTagEntity(state, fsm.Owner.camp == CampType.Red);
                fsm.Owner.battleManager.dogTagManager.AddDogTag(tag, true);
                fsm.Owner.Dispose();
            }
            else {
                fsm.Owner.becovered.forEach(e => {
                    e.fsm.ChangeState(DeathState);
                });
            }
            if (fsm.Owner.enemyIdnex > -1) {
                BattleData.EnemyIndex.push(fsm.Owner.enemyIdnex);
            }
            if (fsm.Owner.camp == CampType.Red) {
                SoundManager.inst.EnemyDeathVibrate();
            }
        }
        OnUpdate(fsm, elapseSeconds) {
        }
        OnLeave(fsm) {
        }
        OnDestroy(fsm) {
        }
    }

    class FormationState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'FormationState';
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            fsm.Owner.SetFormationPos();
            fsm.Owner.setMoveFlag(true);
            fsm.Owner.steerBehaviorUnit.forceMove = true;
        }
        OnUpdate(fsm, elapseSeconds) {
        }
        OnLeave(fsm) {
            MsgManager.getInst().offAllCaller(this);
            fsm.Owner.steerBehaviorUnit.forceMove = false;
        }
        OnDestroy(fsm) {
        }
    }

    class IdleState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'IdleState';
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            fsm.Owner.setMoveFlag(false);
            fsm.Owner.PlayAni(fsm.Owner.weaponConfig.idle, -1);
            fsm.Owner.steerBehaviorUnit.velocity.setValue(0, 0);
        }
        OnUpdate(fsm, elapseSeconds) {
        }
        OnLeave(fsm) {
        }
        OnDestroy(fsm) {
        }
    }

    class AimCircle extends BaseScript3D {
        onAwake() {
            MsgManager.getInst().on(CtrlEvent.EndDriver, this, this.onEndDrive);
        }
        onUpdate() {
            if (this.target == null || this.target.transform == null)
                return;
            var p = this.target.transform.position;
            this.transform.position.setValue(p.x, p.y - 8, p.z);
            this.UpdatePosition();
        }
        onDestroy() {
            MsgManager.getInst().offAllCaller(this);
        }
        onEndDrive() {
            this.sp.destroy(true);
        }
    }

    var tmp$3 = new Laya.Vector3();
    class PlayerCtrlState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'PlayerCtrlState';
            this.steer = new Laya.Vector2();
            this.isSteer = false;
            this.attackCD = 3;
            this.seekCD = 1;
            this.openSeek = false;
            this.isMount = false;
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            MsgManager.getInst().on(CtrlEvent.PlayerSteerMove, this, this.onSteer);
            MsgManager.getInst().on(CtrlEvent.PlayerSteerEnd, this, this.onSteerEnd);
            MsgManager.getInst().on(CtrlEvent.StartDriver, this, this.onStartDrive);
            MsgManager.getInst().on(GameProcedureEvent.StartBattle, this, this.onStartBattle);
            MsgManager.getInst().on(GameProcedureEvent.StopBattle, this, this.onStopBattle);
            fsm.Owner.steerBehaviorUnit.angle = 180;
            fsm.Owner.PlayAni(fsm.Owner.weaponConfig.idle, -1, true, false);
            this.attackCD = 3;
        }
        OnUpdate(fsm, elapseSeconds) {
            if (this.isSteer) {
                fsm.Owner.steerBehaviorUnit.steeringForce.setValue(this.steer.x, this.steer.y);
            }
            else {
                this.fsm.Owner.steerBehaviorUnit.velocity.setValue(0, 0);
            }
            if (this.openSeek) {
                this.seekCD -= elapseSeconds;
                if (this.seekCD <= 0) {
                    this.seekCD += 0.2;
                    if (fsm.Owner.aimTarget == null || !fsm.Owner.CanAttackAimEnemy()) {
                        fsm.Owner.SeekEnemy();
                    }
                }
                else {
                    if (!fsm.Owner.CanAttackAimEnemy()) {
                        fsm.Owner.aimTarget = null;
                    }
                }
                this.attackCD += elapseSeconds;
                if (this.attackCD >= BattleData.CurrentPlayerWeapon.attack_speed) {
                    this.attackCD -= BattleData.CurrentPlayerWeapon.attack_speed;
                    if (fsm.Owner.CanAttackAimEnemy()) {
                        fsm.Owner.PlayAni(fsm.Owner.weaponConfig.shoot, BattleData.CurrentPlayerWeapon.attack_speed, true);
                        SoundManager.inst.playShoot(fsm.Owner.weaponConfig.sound);
                        if (fsm.Owner.aniView.shootFx != null) {
                            fsm.Owner.aniView.shootFx.active = true;
                            Utils.WaitTime(fsm.Owner.attackDuration * 500).then(() => {
                                fsm.Owner.aniView.shootFx.active = false;
                            });
                        }
                        var miss = Math.random() < 0.5;
                        var tartgetPos;
                        if (miss) {
                            Laya.Vector3.subtract(fsm.Owner.aimTarget.PosV3, fsm.Owner.PosV3, tmp$3);
                            Laya.Vector3.scale(tmp$3, 0.5 * Math.random() + 1.1, tmp$3);
                            Laya.Vector3.add(fsm.Owner.aimTarget.PosV3, tmp$3, tmp$3);
                            tmp$3.y = 0;
                            tartgetPos = tmp$3.clone();
                        }
                        else {
                            tartgetPos = fsm.Owner.aimTarget.PosV3.clone();
                            tartgetPos.y += 1.2;
                            fsm.Owner.Attack(fsm.Owner.aimTarget, 0.2);
                        }
                        var bullet = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(_AllPrefabsNames.Bullet));
                        LevelManager.getInst().scene.addChild(bullet);
                        var ft = bullet.addComponent(MoveAndDestroy);
                        ft.target = tartgetPos;
                        ft.speed = 30;
                        bullet.transform.position = fsm.Owner.emiterPos;
                    }
                }
            }
            else {
                fsm.Owner.aimTarget = null;
            }
        }
        OnLeave(fsm) {
            MsgManager.getInst().offAllCaller(this);
        }
        OnDestroy(fsm) {
        }
        onSteer(steer) {
            this.isSteer = true;
            this.steer.setValue(-steer.x * 200, -steer.y * 200);
        }
        onSteerEnd() {
            this.isSteer = false;
            this.steer.setValue(0, 0);
            this.fsm.Owner.steerBehaviorUnit.velocity.setValue(0, 0);
        }
        onStartBattle() {
            this.openSeek = true;
        }
        onStopBattle() {
            this.openSeek = false;
        }
        onStartDrive(mount, node) {
            this.fsm.SetData('airport', node);
            this.fsm.SetData('mount', mount);
            this.fsm.SetData('bullet_count', 15 * BattleData.Airport.level);
            var oldP = mount.transform.position;
            mount.transform.rotation;
            var e = this.fsm.Owner;
            e.PosV3 = oldP;
            e.aniView.transform.position = oldP;
            e.steerBehaviorUnit.angle = 180;
            this.fsm.ChangeState(PlayerCtrlPlaneState);
        }
    }

    var tmp$2 = new Laya.Vector3();
    var invertTmp = new Laya.Vector3();
    class PlayerCtrlPlaneState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'PlayerCtrlPlaneState';
            this.steer = new Laya.Vector2();
            this.isSteer = false;
            this.attackCD = 0;
            this.isMount = false;
            this.bulletCount = 0;
            this.ready = true;
            this.isGoHome = false;
            this.planeForce = new Laya.Vector2();
            this.planePos = new Laya.Vector2();
            this.planeSpeed = new Laya.Vector2();
            this.path = [];
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            this.bulletCount = fsm.GetData('bullet_count');
            MsgManager.getInst().event(BattleEvent.UpdatePlaneBullet, this.bulletCount);
            MsgManager.getInst().on(BattleEvent.BattleWin, this, this.goHome);
            this.mount = fsm.GetData('mount');
            var rs = Utils.GetNode(this.mount, 'Plane_Propeller').addComponent(RotateSelf);
            rs.angleSpeed = new Laya.Vector3(0, 0, 1500);
            fsm.Owner.steerBehaviorUnit.angle = 180;
            fsm.Owner.PlayAni(AniClip.Mount, -1, true, false);
            fsm.Owner.aniView.sp.active = false;
            PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.BombBullet).then(p => {
                this.bombPrefab = p;
            });
            PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.AimTarget).then(p => {
                var aimC = this.aimCircle = Laya.Sprite3D.instantiate(p);
                MapComponent.inst.sp.addChild(aimC);
                var ac = aimC.addComponent(AimCircle);
                ac.target = this.fsm.Owner.aniView.sp;
            });
            this.isGoHome = false;
            this.isSteer = false;
            this.GoFight();
            this.fsm.Owner.steerBehaviorUnit.velocity.setValue(0, 0);
            fsm.Owner.steerBehaviorUnit.angle = 180;
            fsm.Owner.steerBehaviorUnit.closeCollision = true;
            MsgManager.getInst().event(BattleEvent.PlaneStart);
            this.addModel();
        }
        addModel() {
            var commanderConfig = ConfigHelper.CommanderConfigLevelMap[PlayerData.CurrentRankLevel];
            var heroID = commanderConfig.solider_id;
            var soliderConfig = ConfigHelper.SoliderConfigMap[heroID];
            var role = this.model = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(soliderConfig.prefab));
            this.mount.getChildAt(0).addChild(role);
            this.mount.getChildAt(0).addComponent(FlyUpAndDownComponent);
            role.transform.localPosition = LayaUtils.Vector3Zero;
            role.transform.localRotationEuler = LayaUtils.Vector3Zero;
            var matName = commanderConfig.material;
            var p = PrefabManager.getInst().GetPrefab(_AllPrefabsNames.commandeMat);
            var sp = p.getChildByName(matName);
            var mat = sp.meshRenderer.sharedMaterial;
            var mrs = [];
            Utils.GetAllSKinMeshRender(role, mrs);
            mrs.forEach(mr => {
                mr.material = mat;
            });
            var ani = role.getComponent(Laya.Animator);
            if (ani) {
                ani.play(AniClip.Mount);
            }
        }
        GoFight() {
            this.ready = true;
            var p = this.fsm.Owner.steerBehaviorUnit.pos;
            Laya.Tween.to(p, { y: p.y + 15 }, 2000, Laya.Ease.linearIn, Laya.Handler.create(this, () => {
                MsgManager.getInst().on(CtrlEvent.PlayerSteerMove, this, this.onSteer);
                MsgManager.getInst().on(CtrlEvent.PlayerSteerEnd, this, this.onSteerEnd);
                this.ready = false;
            }), 1000).update = Laya.Handler.create(this, this.resetAngle, null, false);
            Laya.Tween.to(this.fsm.Owner, { fieldY: 8 }, 2000, Laya.Ease.linearIn, Laya.Handler.create(this, () => {
            }), 1000);
        }
        resetAngle() {
            this.fsm.Owner.steerBehaviorUnit.angle = 180;
        }
        updateTran() {
            var plane = this.mount;
            if (this.fsm.Owner.aniView.sp == null || this.fsm.Owner.aniView.transform == null)
                return;
            plane.transform.position = this.fsm.Owner.aniView.transform.position;
            if (this.isGoHome)
                return;
            plane.transform.localRotationEulerY = this.fsm.Owner.steerBehaviorUnit.angle + 180;
        }
        OnUpdate(fsm, elapseSeconds) {
            if (this.ready || this.isGoHome) {
                this.fsm.Owner.steerBehaviorUnit.angle = 180;
            }
            this.updateTran();
            if (this.isGoHome) {
                this.onPathFly();
                return;
            }
            if (!this.ready && !this.isGoHome) {
                if (this.isSteer) {
                    fsm.Owner.steerBehaviorUnit.steeringForce.setValue(this.steer.x, this.steer.y);
                }
                else {
                    this.fsm.Owner.steerBehaviorUnit.velocity.setValue(0, 0);
                }
            }
            if (this.bulletCount > 0) {
                this.attackCD += elapseSeconds;
                if (this.attackCD >= 0.7) {
                    this.attackCD -= 0.7;
                    var enemies = fsm.Owner.GetEnemies(fsm.Owner.Pos, 6);
                    if (enemies.length > 0) {
                        var b = Laya.Sprite3D.instantiate(this.bombPrefab);
                        MapComponent.inst.sp.addChild(b);
                        var cc = b.addComponent(BomFallToY0);
                        cc.node = MapComponent.inst.sp;
                        b.transform.position = fsm.Owner.PosV3;
                        b.transform.position.x += (0.5 - Math.random()) * 2;
                        b.transform.position.z += (0.5 - Math.random()) * 2;
                        cc.endCB = () => {
                            enemies.forEach(e => {
                                fsm.Owner.battleManager.damageManager.AddDamge(e, 10, fsm.Owner);
                            });
                        };
                        this.bulletCount--;
                        MsgManager.getInst().event(BattleEvent.UpdatePlaneBullet, this.bulletCount);
                    }
                }
            }
            else {
                if (!this.isGoHome) {
                    this.goHome();
                }
            }
        }
        get Pos() {
            return this.fsm.Owner.PosV3;
        }
        set Pos(v) {
            this.fsm.Owner.PosV3 = v;
        }
        onPathFly() {
            var target = this.path[0];
            var dis = LayaUtils.deltaTimeSec * 8;
            Laya.Vector3.subtract(target, this.Pos, tmp$2);
            while (dis * dis > Laya.Vector3.scalarLengthSquared(tmp$2)) {
                this.path.shift();
                if (this.path.length <= 0) {
                    this.Exit();
                    return;
                }
                this.Pos = target;
                dis -= Laya.Vector3.scalarLength(tmp$2);
                target = this.path[0];
                Laya.Vector3.subtract(target, this.Pos, tmp$2);
            }
            Laya.Vector3.normalize(tmp$2, tmp$2);
            Laya.Vector3.scale(tmp$2, -dis * 100, invertTmp);
            Laya.Vector3.scale(tmp$2, dis, tmp$2);
            Laya.Vector3.add(this.Pos, invertTmp, invertTmp);
            Laya.Vector3.add(tmp$2, this.Pos, tmp$2);
            var oldR = this.mount.transform.rotation.clone();
            this.mount.transform.lookAt(invertTmp, LayaUtils.UP);
            Laya.Quaternion.lerp(oldR, this.mount.transform.rotation, 0.1, oldR);
            this.mount.transform.rotation = oldR;
            this.Pos = tmp$2;
        }
        Exit() {
            this.fsm.ChangeState(PlayerCtrlState);
        }
        OnLeave(fsm) {
            MsgManager.getInst().offAllCaller(this);
            Laya.timer.clearAll(this);
            var rs = Utils.GetNode(this.mount, 'Plane_Propeller').getComponent(RotateSelf);
            if (rs) {
                rs.destroy();
            }
            fsm.Owner.aniView.sp.active = true;
            fsm.Owner.fieldY = 0;
            fsm.Owner.steerBehaviorUnit.closeCollision = false;
            MsgManager.getInst().event(BattleEvent.PlaneEnd);
            this.model.destroy(true);
            this.mount.transform.localPosition = LayaUtils.Vector3Zero;
            this.mount.transform.localRotationEuler = LayaUtils.Vector3Zero;
            var c = this.mount.getChildAt(0).getComponent(FlyUpAndDownComponent);
            if (c) {
                c.destroy();
            }
        }
        OnDestroy(fsm) {
        }
        goHome() {
            MsgManager.getInst().off(CtrlEvent.PlayerSteerMove, this, this.onSteer);
            MsgManager.getInst().off(CtrlEvent.PlayerSteerEnd, this, this.onSteerEnd);
            this.isGoHome = true;
            if (this.aimCircle != null) {
                this.aimCircle.removeSelf();
                this.aimCircle.destroy(true);
            }
            var airport = this.fsm.GetData('airport');
            var planePos = this.Pos;
            var flyPos = Utils.GetNode(airport, 'flyPos').transform.position;
            var startPos = Utils.GetNode(airport, 'startPos').transform.position;
            var smoke = Utils.GetNode(this.mount, 'SmokeDark');
            smoke.active = true;
            var p1 = new Laya.Vector3();
            Laya.Vector3.scale(planePos, 2, p1);
            Laya.Vector3.subtract(p1, flyPos, p1);
            var p5 = new Laya.Vector3();
            Laya.Vector3.scale(startPos, 2, p5);
            Laya.Vector3.subtract(p5, startPos, p5);
            var tmp = new Laya.Vector3();
            Laya.Vector3.subtract(planePos, flyPos, tmp);
            var len1 = Laya.Vector3.scalarLength(tmp);
            Laya.Vector3.subtract(startPos, flyPos, tmp);
            var len2 = Laya.Vector3.scalarLength(tmp);
            var total = len1 + len2;
            var count1 = Math.floor(len1 / total * 50);
            var count2 = Math.floor(len2 / total * 50);
            var curve = new CatmullRomCurve(p1, planePos, flyPos, startPos);
            for (var i = 0; i <= count1; i++) {
                curve.GetPos(1 / count1, tmp);
                this.path.push(tmp.clone());
            }
            curve = new CatmullRomCurve(planePos, flyPos, startPos, p5);
            for (var i = 1; i <= count2; i++) {
                curve.GetPos(i / count2, tmp);
                this.path.push(tmp.clone());
            }
        }
        onSteer(steer) {
            this.isSteer = true;
            this.steer.setValue(-steer.x * 200, -steer.y * 200);
        }
        onSteerEnd() {
            this.isSteer = false;
            this.steer.setValue(0, 0);
            this.fsm.Owner.steerBehaviorUnit.velocity.setValue(0, 0);
        }
        FireBomb() {
        }
    }

    var tmp$1 = new Laya.Vector3();
    class PlayerCtrlMountHorseState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'PlayerCtrlMountHorseState';
            this.steer = new Laya.Vector2();
            this.isSteer = false;
            this.attackCD = 3;
            this.seekCD = 1;
            this.openSeek = false;
            this.oldSpeed = 5;
            this.currentMountAni = '';
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            MsgManager.getInst().on(CtrlEvent.PlayerSteerMove, this, this.onSteer);
            MsgManager.getInst().on(CtrlEvent.PlayerSteerEnd, this, this.onSteerEnd);
            MsgManager.getInst().on(CtrlEvent.StartDriver, this, this.onStartDrive);
            MsgManager.getInst().on(GameProcedureEvent.StartBattle, this, this.onStartBattle);
            MsgManager.getInst().on(GameProcedureEvent.StopBattle, this, this.onStopBattle);
            var config = ConfigHelper.SoliderConfigMap[ConfigHelper.MaxrankLevel.solider_id];
            MsgManager.getInst().event(BattleEvent.ChangeSoliderSkin, this.fsm.Owner, config.prefab, ConfigHelper.MaxrankLevel.material);
            fsm.Owner.steerBehaviorUnit.angle = 180;
            this.attackCD = 3;
            var sp = fsm.Owner.aniView.sp;
            fsm.Owner.aniView.PlayAni('Male Sitting Pose', -1, true, false);
            PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.ma_skin).then((p) => {
                this.mount = Laya.Sprite3D.instantiate(p);
                MapComponent.inst.sp.addChild(this.mount);
                this.mount.transform.position = sp.transform.position;
                this.mount.transform.rotation = sp.transform.rotation;
                this.mountAni = this.mount.getComponent(Laya.Animator);
                if (this.mountAni) {
                    this.mountAni.play('ma_idle');
                }
                var node = Utils.GetNode(this.mount, 'Player_point') || this.mount;
                node.addChild(sp);
                fsm.Owner.mount = true;
                fsm.Owner.shieldHP = fsm.Owner.shieldFullHP;
                sp.transform.localRotationEuler = LayaUtils.Vector3Zero;
                sp.transform.localPosition = new Laya.Vector3(0, 0, 0);
            });
            SoundManager.inst.playSound('Neigh1');
        }
        OnUpdate(fsm, elapseSeconds) {
            fsm.Owner.steerBehaviorUnit.maxSpeed = PlayerData.CurrentCommanderConfig.speed + 4;
            fsm.Owner.steerBehaviorUnit.mass = 1000;
            if (this.isSteer) {
                fsm.Owner.steerBehaviorUnit.steeringForce.setValue(this.steer.x, this.steer.y);
            }
            else {
                this.fsm.Owner.steerBehaviorUnit.velocity.setValue(0, 0);
            }
            if (this.openSeek) {
                this.seekCD -= elapseSeconds;
                if (this.seekCD <= 0) {
                    this.seekCD += 0.2;
                    if (fsm.Owner.aimTarget == null || !fsm.Owner.CanAttackAimEnemy()) {
                        fsm.Owner.SeekEnemy();
                    }
                }
                else {
                    if (!fsm.Owner.CanAttackAimEnemy()) {
                        fsm.Owner.aimTarget = null;
                    }
                }
                this.attackCD += elapseSeconds;
                if (this.attackCD >= BattleData.CurrentPlayerWeapon.attack_speed) {
                    this.attackCD -= BattleData.CurrentPlayerWeapon.attack_speed;
                    if (fsm.Owner.CanAttackAimEnemy()) {
                        fsm.Owner.PlayAni(fsm.Owner.weaponConfig.shoot, BattleData.CurrentPlayerWeapon.attack_speed, true);
                        SoundManager.inst.playShoot(fsm.Owner.weaponConfig.sound);
                        if (fsm.Owner.aniView.shootFx != null) {
                            fsm.Owner.aniView.shootFx.active = true;
                            Utils.WaitTime(fsm.Owner.attackDuration * 500).then(() => {
                                fsm.Owner.aniView.shootFx.active = false;
                            });
                        }
                        var miss = Math.random() < 0.5;
                        var tartgetPos;
                        if (miss) {
                            Laya.Vector3.subtract(fsm.Owner.aimTarget.PosV3, fsm.Owner.PosV3, tmp$1);
                            Laya.Vector3.scale(tmp$1, 0.5 * Math.random() + 1.1, tmp$1);
                            Laya.Vector3.add(fsm.Owner.aimTarget.PosV3, tmp$1, tmp$1);
                            tmp$1.y = 0;
                            tartgetPos = tmp$1.clone();
                        }
                        else {
                            tartgetPos = fsm.Owner.aimTarget.PosV3.clone();
                            tartgetPos.y += 1.2;
                            fsm.Owner.Attack(fsm.Owner.aimTarget, 0.2);
                        }
                        var bullet = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(_AllPrefabsNames.Bullet));
                        LevelManager.getInst().scene.addChild(bullet);
                        var ft = bullet.addComponent(MoveAndDestroy);
                        ft.target = tartgetPos;
                        ft.speed = 30;
                        bullet.transform.position = fsm.Owner.emiterPos;
                    }
                }
            }
            else {
                fsm.Owner.aimTarget = null;
            }
            this.updateTran();
        }
        updateTran() {
            if (this.mount == null)
                return;
            var plane = this.mount;
            if (this.fsm.Owner.aniView.sp == null || this.fsm.Owner.aniView.transform == null)
                return;
            plane.transform.position = this.fsm.Owner.PosV3;
            plane.transform.localRotationEulerY = Utils.AngleLerp(plane.transform.localRotationEulerY, this.fsm.Owner.steerBehaviorUnit.angle + 180, 0.4);
            if (!LayaUtils.IsZeroVec2(this.fsm.Owner.steerBehaviorUnit.velocity)) {
                if (this.currentMountAni !== MountAniType$1.run) {
                    if (this.mountAni) {
                        this.mountAni.crossFade(MountAniType$1.run, 0.1);
                    }
                    this.currentMountAni = MountAniType$1.run;
                }
            }
            else {
                if (this.currentMountAni !== MountAniType$1.idle) {
                    if (this.mountAni) {
                        this.mountAni.crossFade(MountAniType$1.idle, 0.1);
                    }
                    this.currentMountAni = MountAniType$1.idle;
                }
            }
        }
        OnLeave(fsm) {
            if (PlayerData.mountsId == 1 || PlayerData.mountsId == 2) {
                PlayerData.mountsId = 0;
            }
            MsgManager.getInst().offAllCaller(this);
            fsm.Owner.mount = false;
            fsm.Owner.mountY = 0;
            MapComponent.inst.sp.addChild(fsm.Owner.aniView.sp);
            var config = ConfigHelper.SoliderConfigMap[PlayerData.CurrentCommanderConfig.solider_id];
            MsgManager.getInst().event(BattleEvent.ChangeSoliderSkin, this.fsm.Owner, config.prefab, PlayerData.CurrentCommanderConfig.material);
            if (this.mountAni && fsm.Owner.shieldFullHP <= 0) {
                this.mountAni.play(MountAniType$1.death);
                SoundManager.inst.playSound('Neigh2');
            }
            MsgManager.getInst().event(BattleEvent.MountHorseEnd);
            ADData.HorseCD = 300000;
            this.mount.addComponent(DelayDestory).delay = 5000;
            GameManager.getInst().battleManager.mountHorse = false;
        }
        OnDestroy(fsm) {
        }
        onSteer(steer) {
            this.isSteer = true;
            this.steer.setValue(-steer.x * 200, -steer.y * 200);
        }
        onSteerEnd() {
            this.isSteer = false;
            this.steer.setValue(0, 0);
            this.fsm.Owner.steerBehaviorUnit.velocity.setValue(0, 0);
        }
        onStartBattle() {
            this.openSeek = true;
        }
        onStopBattle() {
            this.openSeek = false;
        }
        onStartDrive(mount, node) {
            this.fsm.SetData('airport', node);
            this.fsm.SetData('mount', mount);
            this.fsm.SetData('bullet_count', 15 * BattleData.Airport.level);
            var oldP = mount.transform.position;
            mount.transform.rotation;
            var e = this.fsm.Owner;
            e.PosV3 = oldP;
            e.aniView.transform.position = oldP;
            e.steerBehaviorUnit.angle = 180;
            this.fsm.ChangeState(PlayerCtrlPlaneState);
        }
    }
    var MountAniType$1;
    (function (MountAniType) {
        MountAniType["idle"] = "ma_idle";
        MountAniType["run"] = "ma_run";
        MountAniType["death"] = "ma_death";
    })(MountAniType$1 || (MountAniType$1 = {}));

    class MechaAniStep extends BaseScript3D {
        stepl() {
            SoundManager.inst.playSound('Robots_Movement_Transformer_Footsteps_Left');
            SimpleCameraFollow.inst.SmallShake();
        }
        stepr() {
            SoundManager.inst.playSound('Robots_Movement_Transformer_Footsteps_Right');
            SimpleCameraFollow.inst.SmallShake();
        }
    }
    __decorate([
        throttle(100)
    ], MechaAniStep.prototype, "stepl", null);
    __decorate([
        throttle(100)
    ], MechaAniStep.prototype, "stepr", null);

    new Laya.Vector3();
    class PlayerCtrlMountMechaState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'PlayerCtrlMountMechaState';
            this.steer = new Laya.Vector2();
            this.isSteer = false;
            this.attackCD = 4;
            this.seekCD = 1;
            this.openSeek = true;
            this.oldSpeed = 5;
            this.isFree = false;
            this.currentMountAni = '';
        }
        OnInit(fsm) {
        }
        OnEnter(fsm, isfree = false, isInit = false) {
            if (isfree) {
                ReportMgr.inst.sendReport(ReportType.TRY_ROBOT);
            }
            this.isFree = isfree;
            fsm.Owner.shieldFullHP = 400;
            MsgManager.getInst().on(CtrlEvent.PlayerSteerMove, this, this.onSteer);
            MsgManager.getInst().on(CtrlEvent.PlayerSteerEnd, this, this.onSteerEnd);
            MsgManager.getInst().on(CtrlEvent.StartDriver, this, this.onStartDrive);
            MsgManager.getInst().on(GameProcedureEvent.StartBattle, this, this.onStartBattle);
            MsgManager.getInst().on(GameProcedureEvent.StopBattle, this, this.onStopBattle);
            ConfigHelper.SoliderConfigMap[ConfigHelper.MaxrankLevel.solider_id];
            fsm.Owner.steerBehaviorUnit.angle = 180;
            this.attackCD = 3;
            fsm.Owner.mountY = 0;
            var sp = fsm.Owner.aniView.sp;
            fsm.Owner.aniView.PlayAni('Male Sitting Pose', -1, true, false);
            fsm.Owner.aniView.Find('WeaponPos').active = false;
            this.mount = fsm.GetData('mount');
            this.rotateNode = Utils.GetNode(this.mount, 'RotateNode') || this.mount;
            let pos;
            if (isInit) {
                pos = MapCampManagerComponent.inst.playerLocation.transform.position.clone();
                pos.z -= 3.5;
            }
            else {
                pos = MapCampManagerComponent.inst.Find('MechaBox').transform.position.clone();
            }
            fsm.Owner.PosV3 = pos;
            MapComponent.inst.sp.addChild(this.mount);
            this.mountAni = this.mount.getComponent(Laya.Animator);
            if (this.mountAni) {
                this.mount.addComponent(MechaAniStep);
                this.mountAni.play(MountAniType.idle);
            }
            var node = Utils.GetNode(this.mount, 'Player_point') || this.mount;
            node.addChild(sp);
            fsm.Owner.mount = true;
            fsm.Owner.shieldHP = fsm.Owner.shieldFullHP;
            sp.transform.localRotationEuler = LayaUtils.Vector3Zero;
            sp.transform.localPosition = new Laya.Vector3(0, 0, 0);
            this.emiterNode = Utils.GetNode(this.mount, 'Muzzle');
            PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.penhuo).then(fxp => {
                this.shootFx = Laya.Sprite3D.instantiate(fxp);
                var n = Utils.GetNode(this.mount, 'Muzzle');
                n.addChild(this.shootFx);
                this.shootFx.transform.localPosition = LayaUtils.Vector3Zero;
                this.shootFx.transform.localRotationEuler = LayaUtils.Vector3Zero;
                this.shootFx.transform.localScale = new Laya.Vector3(0.5, 0.5, 0.5);
                Utils.WaitFrame(2).then(() => {
                    this.shootFx.active = false;
                });
            });
            SoundManager.inst.playSound('Robots_Movement_Transformer_Footsteps_Left');
            Utils.WaitTime(300).then(() => {
                SoundManager.inst.playSound('Robots_Movement_Transformer_Footsteps_Right');
            });
            fsm.Owner.steerBehaviorUnit.radius = 1;
            PlayerData.UnlockMecha = true;
        }
        OnUpdate(fsm, elapseSeconds) {
            fsm.Owner.steerBehaviorUnit.maxSpeed = PlayerData.CurrentCommanderConfig.speed + 3;
            fsm.Owner.steerBehaviorUnit.mass = 1000;
            if (this.isSteer) {
                fsm.Owner.steerBehaviorUnit.steeringForce.setValue(this.steer.x, this.steer.y);
            }
            else {
                this.fsm.Owner.steerBehaviorUnit.velocity.setValue(0, 0);
            }
            if (this.openSeek) {
                this.seekCD -= elapseSeconds;
                if (this.seekCD <= 0) {
                    this.seekCD += 0.2;
                    if (fsm.Owner.aimTarget == null || !fsm.Owner.CanAttackAimEnemy()) {
                        fsm.Owner.SeekEnemy();
                    }
                }
                else {
                    if (!fsm.Owner.CanAttackAimEnemy()) {
                        fsm.Owner.aimTarget = null;
                        this.attackCD = 4;
                    }
                    else {
                        this.playSound();
                        this.attackCD = 4;
                    }
                }
                this.attackCD += elapseSeconds;
                if (this.attackCD >= 1.8) {
                    this.attackCD -= 1.8;
                    if (fsm.Owner.CanAttackAimEnemy()) {
                        if (this.shootFx != null) {
                            this.shootFx.active = true;
                            Laya.timer.once(1000, this, this.hideShootFX, null, true);
                        }
                        var tartgetPos;
                        tartgetPos = fsm.Owner.aimTarget.PosV3.clone();
                        tartgetPos.y += 1.2;
                        fsm.Owner.Attack(fsm.Owner.aimTarget, 0.2, 40);
                    }
                }
            }
            else {
                fsm.Owner.aimTarget = null;
            }
            this.updateTran();
            if (this.isFree) {
                var t = Math.max(0, 30 - this.fsm.CurrentStateTime);
                MsgManager.getInst().event(BattleEvent.UpdateMechaFreeTime, this.fsm.Owner.aniView.transform.position, t);
                if (t == 0) {
                    MsgManager.getInst().event(BattleEvent.MechFreeEnd);
                    this.fsm.ChangeState(PlayerCtrlState);
                }
            }
        }
        playSound() {
            SoundManager.inst.playShoot('Flamethrower');
        }
        hideShootFX() {
            this.shootFx.active = false;
        }
        updateTran() {
            if (this.mount == null)
                return;
            var mount = this.mount;
            if (this.fsm.Owner.aniView.sp == null || this.fsm.Owner.aniView.transform == null)
                return;
            mount.transform.position = this.fsm.Owner.PosV3;
            mount.transform.localRotationEulerY = Utils.AngleLerp(mount.transform.localRotationEulerY, this.fsm.Owner.steerBehaviorUnit.angle + 180, LayaUtils.deltaTimeSec * 10);
            if (!LayaUtils.IsZeroVec2(this.fsm.Owner.steerBehaviorUnit.velocity)) {
                if (this.currentMountAni !== MountAniType.run) {
                    if (this.mountAni) {
                        this.mountAni.crossFade(MountAniType.run, 0.1);
                    }
                    this.currentMountAni = MountAniType.run;
                }
            }
            else {
                if (this.currentMountAni !== MountAniType.idle) {
                    if (this.mountAni) {
                        this.mountAni.crossFade(MountAniType.idle, 0.1);
                    }
                    this.currentMountAni = MountAniType.idle;
                }
            }
            if (this.fsm.Owner.aimTarget != null) {
                var oldY = this.rotateNode.transform.localRotationEulerY;
                this.rotateNode.transform.lookAt(this.fsm.Owner.aimTarget.PosV3, LayaUtils.UP);
                this.rotateNode.transform.localRotationEulerX = this.rotateNode.transform.localRotationEulerZ = 0;
                var newY = this.rotateNode.transform.localRotationEulerY + 180;
                this.rotateNode.transform.localRotationEulerY = Utils.AngleLerp(oldY, newY, LayaUtils.deltaTimeSec * 10);
            }
            else {
                this.rotateNode.transform.localRotationEulerY = Utils.AngleLerp(this.rotateNode.transform.localRotationEulerY, 0, LayaUtils.deltaTimeSec);
            }
        }
        OnLeave(fsm) {
            if (PlayerData.mountsId == 3 || PlayerData.mountsId == 4) {
                PlayerData.mountsId = 0;
            }
            MsgManager.getInst().event(BattleEvent.MountMechaEnd);
            MsgManager.getInst().offAllCaller(this);
            fsm.Owner.mount = false;
            fsm.Owner.mountY = 0;
            fsm.Owner.shieldFullHP = 300;
            MapComponent.inst.sp.addChild(fsm.Owner.aniView.sp);
            ConfigHelper.SoliderConfigMap[PlayerData.CurrentCommanderConfig.solider_id];
            if (this.mountAni && fsm.Owner.shieldHP <= 0) {
                this.mountAni.play(MountAniType.idle);
                PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.baozha03).then((p) => {
                    if (p) {
                        var fx = Laya.Sprite3D.instantiate(p);
                        MapComponent.inst.sp.addChild(fx);
                        fx.transform.position = fxPos;
                        var dd = fx.addComponent(DelayDestory);
                        dd.delay = 700;
                    }
                    SoundManager.inst.playVibrate(true);
                });
            }
            if (PlayerData.UnlockMecha) {
                ADData.MechaCD = 300000;
            }
            if (this.isFree) {
                MsgManager.getInst().event(BattleEvent.UpdateMechaFreeTime, this.fsm.Owner.aniView.transform.position, 0);
            }
            this.mount.addComponent(DelayDestory).delay = 400;
            var fxPos = this.mount.transform.position;
            GameManager.getInst().battleManager.mountHorse = false;
            fsm.Owner.aniView.Find('WeaponPos').active = true;
            fsm.Owner.steerBehaviorUnit.radius = 0.5;
            PlayerData.UnlockMecha = true;
        }
        OnDestroy(fsm) {
        }
        onSteer(steer) {
            this.isSteer = true;
            this.steer.setValue(-steer.x * 200, -steer.y * 200);
        }
        onSteerEnd() {
            this.isSteer = false;
            this.steer.setValue(0, 0);
            this.fsm.Owner.steerBehaviorUnit.velocity.setValue(0, 0);
        }
        onStartBattle() {
            this.openSeek = true;
        }
        onStopBattle() {
            this.openSeek = false;
        }
        onStartDrive(mount, node) {
            this.fsm.SetData('airport', node);
            this.fsm.SetData('mount', mount);
            this.fsm.SetData('bullet_count', 15 * BattleData.Airport.level);
            var oldP = mount.transform.position;
            mount.transform.rotation;
            var e = this.fsm.Owner;
            e.PosV3 = oldP;
            e.aniView.transform.position = oldP;
            e.steerBehaviorUnit.angle = 180;
            this.fsm.ChangeState(PlayerCtrlPlaneState);
        }
    }
    __decorate([
        throttle(2000)
    ], PlayerCtrlMountMechaState.prototype, "playSound", null);
    var MountAniType;
    (function (MountAniType) {
        MountAniType["idle"] = "Idle";
        MountAniType["run"] = "Run";
    })(MountAniType || (MountAniType = {}));

    class PlayerDeathState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'PlayerDeathState';
            this.time = 0;
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            fsm.Owner.isDeath = true;
            fsm.Owner.PlayAni(fsm.Owner.weaponConfig.death, -1, true);
            var pr = fsm.Owner.aniView.sp.getComponent(PlayerRun);
            if (pr) {
                pr.destroy();
            }
            ReportMgr.inst.sendReport('死亡_' + PlayerData.CurrentLevelScene, false);
            fsm.Owner.aniView.Find('DogTagPoint').destroy(true);
            fsm.Owner.aniView.Find('DogTagEnemy').destroy(true);
            for (var i = 0; i < BattleData.SilverTag; i++) {
                var state = BattleData.AddTag(fsm.Owner.Pos.x + 0.5 - Math.random(), Math.random() * 0.3, fsm.Owner.Pos.y + 0.5 - Math.random(), Math.random() * 360, false);
                var tag = new DogTagEntity(state, false);
                fsm.Owner.battleManager.dogTagManager.AddDogTag(tag, true);
            }
            BattleData.SilverTag = 0;
            for (var i = 0; i < BattleData.GoldTag; i++) {
                var state = BattleData.AddTag(fsm.Owner.Pos.x + 0.5 - Math.random(), Math.random() * 0.3, fsm.Owner.Pos.y + 0.5 - Math.random(), Math.random() * 360, true);
                var tag = new DogTagEntity(state, true);
                fsm.Owner.battleManager.dogTagManager.AddDogTag(tag, true);
            }
            BattleData.GoldTag = 0;
            Utils.WaitTime(2000).then(() => {
                AlertMgr.inst.setCloseCB(Laya.Handler.create(this, () => {
                    MsgManager.getInst().event(BattleEvent.SpwanPlayer);
                    fsm.Owner.Dispose();
                }), Laya.Handler.create(this, () => {
                    GameManager.getInst().Reset();
                }));
                AlertMgr.inst.showReborn();
            });
        }
        OnUpdate(fsm, elapseSeconds) {
        }
        OnLeave(fsm) {
        }
        OnDestroy(fsm) {
        }
    }

    var tmp1 = new Laya.Vector2();
    var tmp2 = new Laya.Vector2();
    class BattleEntity {
        constructor(battleManager, id, level, camp, formationPos, config, weaponConfig, isPlayer = false, isDebug = false) {
            this.isPlayer = false;
            this.isDeath = false;
            this.enemyIdnex = -1;
            this.mount = false;
            this.mountY = 0;
            this.fieldY = 0;
            this.formationAngle = 0;
            this.shieldHP = 0;
            this.shieldFullHP = 300;
            this.attackUP = false;
            this._pos = new Laya.Vector3();
            this.rcoverHP = false;
            this.battleManager = battleManager;
            this.isPlayer = isPlayer;
            this.id = id;
            this.isDebug = isDebug;
            this.config = config;
            this.weaponConfig = weaponConfig;
            this.level = level;
            this.seekRadiusSq = config.seek_radius * config.seek_radius;
            this.attackRadiusSq = config.attack_radius * config.attack_radius;
            this.hp = this.fullHP = config.hp;
            var states = [
                new FormationAndSeekState(),
                new FormationState(),
                new IdleState(),
                new WanderState(),
                new AttackState(),
                new CastSkillState(),
                new DeathState()
            ];
            if (this.isPlayer) {
                states.push(new PlayerCtrlState(), new PlayerCtrlMountHorseState(), new PlayerCtrlMountMechaState(), new PlayerCtrlPlaneState(), new PlayerDeathState());
            }
            this.fsm = Fsm.Create(config.id + '_' + id, this, states, isDebug);
            this.formationPos = formationPos;
            this.steerBehaviorUnit = SteeringBehaviorManager.getInst().CreateUnit(id, config.radius, config.speed, config.mass);
            this.steerBehaviorUnit.moveable = false;
            if (this.config.type == SoilderType.Builidng) {
                SteeringBehaviorManager.getInst().DeleteUnit(this.steerBehaviorUnit);
            }
            this.camp = camp;
            if (isPlayer) {
                this.fsm.Start(PlayerCtrlState);
                this.hp = 10000;
            }
            else {
                Utils.WaitFrame(2).then(() => {
                    switch (this.camp) {
                        case CampType.Blue:
                            this.fsm.Start(FormationState);
                            break;
                        case CampType.Red:
                            this.fsm.Start(FormationAndSeekState);
                            break;
                    }
                });
            }
        }
        get emiterPos() {
            try {
                if (this.emiter == null) {
                    return this.PosV3;
                }
                else {
                    return this.emiter.transform.position;
                }
            }
            catch (err) {
                ReportMgr.inst.sendReport("Code_Error_BattleEntity " + JSON.stringify(err), false);
                return new Laya.Vector3();
            }
        }
        hasAni() {
            return this.weaponConfig != null;
        }
        get seekRadius() {
            return this.config.seek_radius;
        }
        get attackRadius() {
            return this.config.attack_radius;
        }
        get damage() {
            var dmg = this.config.damage + this.weaponConfig.damage + (this.attackUP ? 3 : 0);
            if (this.camp == CampType.Red) {
                if (PlayerData.CurrentLevelID == 1) {
                    return dmg * 0.6;
                }
                if (PlayerData.CurrentLevelID == 2) {
                    return dmg * 0.8;
                }
            }
            return dmg;
        }
        get attackDuration() {
            return this.weaponConfig.attack_speed;
        }
        get PosV3() {
            this._pos.setValue(this.steerBehaviorUnit.pos.x, this.fieldY + this.mountY, this.steerBehaviorUnit.pos.y);
            return this._pos;
        }
        set PosV3(v) {
            this.fieldY = v.y;
            this.steerBehaviorUnit.pos.setValue(v.x, v.z);
        }
        get IsMove() {
            return !LayaUtils.IsZeroVec2(this.steerBehaviorUnit.velocity);
        }
        get Pos() {
            return this.steerBehaviorUnit.pos;
        }
        set Pos(v) {
            this.steerBehaviorUnit.pos.setValue(v.x, v.y);
        }
        Update(deltaTime) {
            this.fsm.OnUpdate(deltaTime);
            if (this.lastAngle == null) {
                this.lastAngle = this.steerBehaviorUnit.angle;
            }
            else {
                let d = Math.abs(this.steerBehaviorUnit.angle - this.lastAngle);
                if (d > 1) {
                    var r = Math.max(0.5, (20 - d) / 20);
                    LayaUtils.scaleVec2(this.steerBehaviorUnit.deltaPos, r, this.steerBehaviorUnit.deltaPos);
                }
                this.lastAngle = this.steerBehaviorUnit.angle;
            }
            LayaUtils.addVec2(this.steerBehaviorUnit.pos, this.steerBehaviorUnit.deltaPos, this.steerBehaviorUnit.pos);
            if (this.isPlayer) {
                this.crossSegmentLine = [];
                var p = this.steerBehaviorUnit.pos;
                if (!MathUtils.PointInPolygon(this.steerBehaviorUnit.pos, MapComponent.inst.mapPolygon, this.crossSegmentLine)) {
                    let minDis = Number.MAX_VALUE;
                    let lineP1;
                    let lineP2;
                    for (let i = 0; i < this.crossSegmentLine.length; i += 2) {
                        let p1 = this.crossSegmentLine[i];
                        let p2 = this.crossSegmentLine[i + 1];
                        let dis = Utils.PointToSegmentDist2(p.x, p.y, p1.x, p1.y, p2.x, p2.y);
                        if (dis < minDis) {
                            minDis = dis;
                            lineP1 = p1;
                            lineP2 = p2;
                        }
                    }
                    if (lineP1 != null && lineP2 != null) {
                        var d = tmp1;
                        LayaUtils.subtractVec2(lineP1, lineP2, d);
                        LayaUtils.normalizeVec2(d, d);
                        var e = tmp2;
                        LayaUtils.subtractVec2(p, lineP2, e);
                        var l = LayaUtils.DotVec2(e, d);
                        d.x = lineP2.x + d.x * l;
                        d.y = lineP2.y + d.y * l;
                        LayaUtils.subtractVec2(d, p, d);
                        LayaUtils.setLengthVec2(d, minDis, d);
                        LayaUtils.addVec2(p, d, p);
                    }
                }
                if (this.fieldY == 0) {
                    MapComponent.inst.collisionPolygon.forEach(wrap => {
                        if (wrap.FastTest(p)) {
                            if (MathUtils.PointInPolygon(p, wrap.points, this.crossSegmentLine)) {
                                let minDis = Number.MAX_VALUE;
                                let lineP1;
                                let lineP2;
                                for (let i = 0; i < wrap.points.length; i++) {
                                    let p1 = wrap.points[i];
                                    let p2 = wrap.points[(i == (wrap.points.length - 1)) ? 0 : (i + 1)];
                                    let dis = Utils.PointToSegmentDist2(p.x, p.y, p1.x, p1.y, p2.x, p2.y);
                                    if (dis < minDis) {
                                        minDis = dis;
                                        lineP1 = p1;
                                        lineP2 = p2;
                                    }
                                }
                                if (lineP1 != null && lineP2 != null) {
                                    var d = tmp1;
                                    LayaUtils.subtractVec2(lineP1, lineP2, d);
                                    LayaUtils.normalizeVec2(d, d);
                                    var e = tmp2;
                                    LayaUtils.subtractVec2(p, lineP2, e);
                                    var l = LayaUtils.DotVec2(e, d);
                                    d.x = lineP2.x + d.x * l;
                                    d.y = lineP2.y + d.y * l;
                                    LayaUtils.subtractVec2(d, p, d);
                                    LayaUtils.setLengthVec2(d, minDis, d);
                                    LayaUtils.addVec2(p, d, p);
                                }
                            }
                        }
                    });
                }
                if (this.rcoverHP) {
                    if (this.mount) {
                        if (this.shieldHP < this.shieldFullHP) {
                            this.shieldHP = Math.min(this.shieldFullHP, this.shieldHP + 100 * LayaUtils.deltaTimeSec);
                        }
                    }
                    if (this.hp < this.fullHP) {
                        this.hp = Math.min(this.fullHP, this.hp + 50 * LayaUtils.deltaTimeSec);
                    }
                }
                MsgManager.getInst().event(BattleEvent.UpdatePlayerPos, this.aniView.transform.position, this);
            }
        }
        AlertEnemy(target) {
            this.targetHitMe = target;
            this.team.forEach(e => {
                e.targetHitMe = target;
                if (e.aimTarget == null) {
                    e.aimTarget = target;
                }
            });
        }
        setMoveFlag(moveable) {
            this.steerBehaviorUnit.moveable = moveable;
        }
        SetFormationPos() {
            this.steerBehaviorUnit.target = new Laya.Vector2(this.formationPos.x, this.formationPos.z);
            this.steerBehaviorUnit.formationAngle = this.formationAngle;
        }
        GetEnemies(pos, radius) {
            var radiusSq = radius * radius;
            this.battleManager.entiytManager.RefreshCampGroup();
            var enemyGroup = this.battleManager.entiytManager.GetEnemyGroup(this.camp);
            var target = [];
            enemyGroup.forEach(enemy => {
                if (enemy.isDeath || enemy.cover != null)
                    return;
                let disSq = LayaUtils.distanceSquaredVec2(this.Pos, enemy.Pos);
                if (disSq < radiusSq) {
                    target.push(enemy);
                }
            });
            return target;
        }
        GetNearstEnemy() {
            this.battleManager.entiytManager.RefreshCampGroup();
            var enemyGroup = this.battleManager.entiytManager.GetEnemyGroup(this.camp);
            var target;
            var minDisSq = Number.MAX_VALUE;
            enemyGroup.forEach(enemy => {
                if (enemy.isDeath || enemy.cover != null)
                    return;
                let disSq = LayaUtils.distanceSquaredVec2(this.Pos, enemy.Pos);
                if (disSq < minDisSq) {
                    minDisSq = disSq;
                    target = enemy;
                }
            });
            return target;
        }
        SeekEnemy() {
            if (this.forceAimTarget != null) {
                if (!this.forceAimTarget.isDeath) {
                    this.aimTarget = this.forceAimTarget;
                    this.steerBehaviorUnit.target = this.aimTarget.Pos;
                    this.steerBehaviorUnit.attachRadius = this.config.attack_radius + this.aimTarget.config.radius;
                    return;
                }
                else {
                    this.forceAimTarget = null;
                }
            }
            var enemyGroup = this.battleManager.entiytManager.GetEnemyGroup(this.camp);
            var target;
            var minDisSq = Number.MAX_VALUE;
            enemyGroup.forEach(enemy => {
                if (enemy.isDeath || enemy.cover != null || !this.CanAttack(enemy))
                    return;
                let disSq = LayaUtils.distanceSquaredVec2(this.Pos, enemy.Pos);
                if (disSq < this.seekRadiusSq && disSq < minDisSq) {
                    minDisSq = disSq;
                    target = enemy;
                }
            });
            if (target == null && this.targetHitMe) {
                if (this.targetHitMe.isDeath) {
                    this.targetHitMe = null;
                }
                else {
                    target = this.targetHitMe;
                }
            }
            if (target == null && this.camp == CampType.Red) {
                this.EnemySeekPlayer();
                target = this.aimTarget;
            }
            this.aimTarget = target;
            if (target != null) {
                this.steerBehaviorUnit.target = target.Pos;
                this.steerBehaviorUnit.attachRadius = this.config.attack_radius + target.config.radius;
            }
            else {
                this.steerBehaviorUnit.moveable = false;
                this.steerBehaviorUnit.target = null;
            }
        }
        EnemySeekPlayer() {
            this.aimTarget = null;
            var player = this.battleManager.entiytManager.player;
            if (player == null || player.isDeath)
                return;
            let disSq = LayaUtils.distanceSquaredVec2(this.Pos, player.Pos);
            if (disSq <= this.seekRadiusSq) {
                this.aimTarget = player;
                this.steerBehaviorUnit.target = player.Pos;
                this.steerBehaviorUnit.attachRadius = this.config.attack_radius + player.config.radius;
            }
            else {
                this.aimTarget = null;
            }
        }
        CanAttack(other) {
            return true;
        }
        HasAminTarget() {
            return this.aimTarget != null;
        }
        CanAttackAimEnemy() {
            if (this.aimTarget == null)
                return false;
            let disSq = LayaUtils.distanceSquaredVec2(this.Pos, this.aimTarget.Pos);
            var limitDis = (this.attackRadius + this.aimTarget.config.radius);
            limitDis *= limitDis;
            var r = (disSq <= limitDis);
            return r;
        }
        Attack(defender, delayTime = 0, damge = 0) {
            this.battleManager.damageManager.AddDamge(defender, this.damage + damge, defender, delayTime);
        }
        OnHit(dmg) {
            if (this.isDeath)
                return;
            if (this.isPlayer) {
                this.rcoverHP = false;
                Laya.timer.once(2500, this, this.AddHP, null, true);
                if (GuideData.Step < GuideStepType.开始战斗) {
                    dmg.damage = 0.1;
                }
                if (this.fieldY > 0) {
                    dmg.damage *= 0.2;
                }
            }
            if (this.mount && this.shieldHP > 0) {
                this.shieldHP = Math.max(0, this.shieldHP - dmg.damage);
                if (this.shieldHP <= 0) {
                    this.MountDeath();
                    GameManager.getInst().battleManager.mountHorse = false;
                }
            }
            else {
                this.hp -= dmg.damage;
                if (this.hp <= 0) {
                    this.hp = 0;
                    this.Death();
                }
            }
            this.aniView.BlinkFade();
        }
        AddHP() {
            this.rcoverHP = true;
        }
        Death() {
            if (this.isDeath)
                return;
            this.isDeath = true;
            if (this.isPlayer) {
                this.fsm.ChangeState(PlayerDeathState);
            }
            else {
                this.fsm.ChangeState(DeathState);
            }
        }
        MountDeath() {
            this.fsm.ChangeState(PlayerCtrlState);
        }
        Dispose() {
            SteeringBehaviorManager.getInst().DeleteUnit(this.steerBehaviorUnit);
        }
        PlayAni(name, duration, forceReplay = false, cross = false) {
            if (this.aniView) {
                this.aniView.PlayAni(name, duration, forceReplay, cross);
            }
        }
        AddArrow(target) {
            var pr = PrefabManager.getInst().GetPrefab(_AllPrefabsNames.Arrow);
            var p = MapComponent.inst.sp.addChild(Laya.Sprite3D.instantiate(pr));
            var a = p.addComponent(ArrowTips);
            if (GuideData.Step == GuideStepType.拾取免费核弹) {
                GuideData.Step = GuideStepType.战斗中_隐藏Tips;
            }
            if (GuideData.Step != GuideStepType.使用免费核弹) {
                a.target = target;
            }
            a.hostNode = this.aniView.sp;
            a.hostEntity = this;
        }
    }
    var AniType;
    (function (AniType) {
        AniType["Idle"] = "Rifle Idle_1";
        AniType["Walk"] = "Rifle Run_1";
        AniType["Attack"] = "Firing Rifle";
        AniType["Dead"] = "Falling Back Death_1";
    })(AniType || (AniType = {}));

    class BattleEntityManager {
        constructor(battleManager) {
            this._idIndex = 0;
            this.entities = [];
            this.campGroup = {};
            this.battleManager = battleManager;
            MsgManager.getInst().on(BattleEvent.SpwanEntity, this, this.AddBattleEntity);
            MsgManager.getInst().on(BattleEvent.SpwanEntityFromView, this, this.AddBattleEnityByView);
        }
        get idIndex() {
            return this._idIndex++;
        }
        AddBattleEnityByView(heroPrefab, weaponPrefabName, level, camp, viewSprite, isDebug = false) {
            var config = ConfigHelper.GetSoliderByPrefab(heroPrefab);
            if (config == null) {
                throw new Error(`heroID:${heroPrefab} not find config`);
            }
            var weaponConfig = ConfigHelper.GetWeaponByPrefab(weaponPrefabName);
            if (weaponConfig == null) {
                weaponConfig = ConfigHelper.WeaponConfigMap[config.spwan_weapon];
            }
            var entity = new BattleEntity(this.battleManager, this.idIndex, level, camp, viewSprite.transform.position, config, weaponConfig, false, isDebug);
            this.entities.push(entity);
            viewSprite.entity = entity;
        }
        AddPlayerEntity(commandLevel, pos, isDebug = false) {
            var commanderConfig = ConfigHelper.CommanderConfigLevelMap[commandLevel];
            var heroID = commanderConfig.solider_id;
            var soliderConfig = ConfigHelper.SoliderConfigMap[heroID];
            var weaponConfig = ConfigHelper.InitWeaponConfig;
            if (BattleData.WeaponFactory != null && BattleData.WeaponFactory.level > 0) {
                var factory = ConfigHelper.BuildingMap[BattleData.WeaponFactory.id];
                weaponConfig = ConfigHelper.WeaponConfigMap[factory.spwan_id];
            }
            var entity = new BattleEntity(this.battleManager, this.idIndex, 1, CampType.Blue, new Laya.Vector3(), soliderConfig, weaponConfig, true, isDebug);
            entity.PosV3 = pos;
            this.entities.push(entity);
            this.player = entity;
            entity.hp = entity.fullHP = commanderConfig.hp;
            entity.steerBehaviorUnit.maxSpeed = commanderConfig.speed;
            entity.steerBehaviorUnit.mass = commanderConfig.mass;
            MsgManager.getInst().event(BattleEvent.SpwanEntityToView, entity, commanderConfig);
            entity.AddArrow(MapCampManagerComponent.inst.GetSlotNode('TagLocation'));
        }
        AddBattleEntity(heroID, level, camp, pos, isPlyer = false, isDebug = false) {
            var config = ConfigHelper.SoliderConfigMap[heroID];
            if (config == null) {
                throw new Error(`heroID:${heroID} not find config`);
            }
            var weaponConfig = ConfigHelper.InitWeaponConfig;
            if (BattleData.WeaponFactory != null && BattleData.WeaponCountInFactory > 0) {
                var factory = ConfigHelper.BuildingMap[BattleData.WeaponFactory.id];
                weaponConfig = ConfigHelper.WeaponConfigMap[factory.spwan_id];
                MsgManager.getInst().event(BattleEvent.DeleteWeaponFromfactory, 1);
                BattleData.WeaponCountInFactory--;
            }
            var entity = new BattleEntity(this.battleManager, this.idIndex, level, camp, new Laya.Vector3(), config, weaponConfig, isPlyer, isDebug);
            entity.PosV3 = pos;
            this.entities.push(entity);
            MsgManager.getInst().event(BattleEvent.SpwanEntityToView, entity);
            if (entity.battleManager.isBattleState) {
                entity.seekRadiusSq = 1000000;
                entity.steerBehaviorUnit.radius = 0.7;
                entity.fsm.ChangeState(WanderState);
            }
            return entity;
        }
        Dispose() {
            this.entities.forEach(entity => {
                entity.fsm.ExiteCurrentState();
                entity.Dispose();
            });
            this.entities = null;
            MsgManager.getInst().offAllCaller(this);
        }
        RemoveBattleEntity(entity) {
            var index = this.entities.indexOf(entity);
            entity.Dispose();
            this.entities.splice(index, 1);
        }
        RefreshCampGroup() {
            this.campGroup = {};
            this.entities.forEach(entity => {
                if (!entity.isPlayer) {
                    var camp = entity.camp;
                    this.campGroup[camp] = this.campGroup[camp] || [];
                    this.campGroup[camp].push(entity);
                    if (entity.battleManager.isBattleState && entity.camp === CampType.Blue && entity.fsm.CurrentState instanceof FormationState) {
                        entity.fsm.ChangeState(WanderState);
                    }
                    if (entity.camp === CampType.Blue && entity.battleManager.isBattleState) {
                        if (LayaUtils.distanceSquaredVec2(entity.Pos, this.player.Pos) <= 25) {
                            entity.attackUP = true;
                            MsgManager.getInst().event(BattleEvent.ShowAttackUPBar, entity);
                        }
                        else {
                            entity.attackUP = false;
                            MsgManager.getInst().event(BattleEvent.HideAttackUPBar, entity);
                        }
                    }
                    else {
                        entity.attackUP = false;
                        MsgManager.getInst().event(BattleEvent.HideAttackUPBar, entity);
                    }
                }
            });
        }
        GetEnemyGroup(camp) {
            switch (camp) {
                case CampType.Blue:
                    return this.campGroup[CampType.Red] || [];
                case CampType.Red:
                    return this.campGroup[CampType.Blue] || [];
            }
        }
        GetSelfEntityAll() {
            return this.campGroup[CampType.Blue] || [];
        }
        GetNearstEnemy() {
            var enemyGroup = this.GetEnemyGroup(CampType.Blue);
            var target;
            var minDisSq = Number.MAX_VALUE;
            var p = this.player.Pos;
            enemyGroup.forEach(enemy => {
                if (enemy.isDeath || enemy.cover != null)
                    return;
                let disSq = LayaUtils.distanceSquaredVec2(p, enemy.Pos);
                if (disSq < minDisSq) {
                    minDisSq = disSq;
                    target = enemy;
                }
            });
            return target;
        }
        Update(elapseSeconds) {
            let len = this.entities.length;
            var newEntities = [];
            for (var i = 0; i < len; i++) {
                var entity = this.entities[i];
                entity.Update(elapseSeconds);
                if (entity.isDeath == false) {
                    newEntities.push(entity);
                }
                else {
                    entity.Death();
                }
            }
            this.entities = newEntities;
        }
        Clear() {
            let len = this.entities.length;
            for (var i = 0; i < len; i++) {
                var entity = this.entities[i];
                entity.Death;
            }
        }
        Attack() {
        }
    }

    class DamageInfo {
        constructor() {
            this.delayTime = 0;
        }
        Reset(attacker = null, defender = null, damage = 0, delayTime = 0) {
            this.attacker = attacker;
            this.defender = defender;
            this.damage = damage;
            this.delayTime = delayTime;
        }
    }

    class DamageManager {
        constructor() {
            this.pool = [];
            this.damages = {};
            this.damgeIndex = 0;
        }
        AddDamge(defender, damge, attacker, delayTime = 0) {
            if (defender == null) {
                console.log('defender is null');
                return;
            }
            var dmg;
            if (this.pool.length > 0) {
                dmg = this.pool.pop();
            }
            else {
                dmg = new DamageInfo();
            }
            dmg.id = this.damgeIndex++;
            this.damages[dmg.id] = dmg;
            dmg.Reset(attacker, defender, damge, delayTime);
            return dmg;
        }
        Return(dmg) {
            this.pool.push(dmg);
        }
        Update(deltaTime) {
            var keys = Object.keys(this.damages);
            keys.forEach(key => {
                var damge = this.damages[key];
                damge.delayTime -= deltaTime;
                if (damge.defender == null) {
                    console.error('damgeinfo defender is null', damge.id);
                    return;
                }
                if (damge.delayTime <= 0) {
                    damge.defender.OnHit(damge);
                    delete this.damages[key];
                    this.Return(damge);
                }
            });
        }
    }

    class DogTagManager {
        constructor() {
            this.tags = [];
            this.needPikupTags = [];
            this.pickTime = 0;
        }
        Udpate(deltaTime, pos) {
            var picks;
            this.pickTime += deltaTime;
            if (this.pickTime > 0.085) {
                this.pickTime -= 0.085;
                this.tags.sort((l, r) => {
                    if (l.y > r.y)
                        return -1;
                    if (l.y < r.y)
                        return 1;
                    return 0;
                });
                for (var i = 0; i < this.tags.length; i++) {
                    let tag = this.tags[i];
                    if (tag.ani != null && !tag.ani.destroyed) {
                        continue;
                    }
                    if (LayaUtils.distanceSquaredVec2(tag.pos, pos) < 4) {
                        picks = tag;
                        this.tags.splice(i, 1);
                        break;
                    }
                }
            }
            if (picks) {
                if (picks.isEnemy) {
                    BattleData.GoldTag++;
                }
                else {
                    BattleData.SilverTag++;
                }
                BattleData.DeleteTag(picks.state, picks.isEnemy);
                MsgManager.getInst().event(BattleEvent.AddTagToPlayer, picks.isEnemy, picks.view.sp);
                if (picks.isEnemy) {
                    BattleData.EnemyTagAllCollect++;
                }
                else {
                    BattleData.TagAllCollect++;
                }
            }
        }
        AddDogTag(entity, ani) {
            this.tags.push(entity);
            MsgManager.getInst().event(BattleEvent.SpwanDogTag, entity, ani);
        }
    }

    class ExchageBuilding {
        constructor(node) {
            this.triggerTime = 0;
            this.isTrigger = false;
            this.unlock = false;
            this.node = node;
            this.tagSilver = Utils.GetNode(node, 'DogTag');
            this.tagGold = Utils.GetNode(node, 'DogTagEnemy');
            var trigger = Utils.GetNode(node, 'triggerPolygon');
            var ps = [];
            for (var i = 0; i < trigger.numChildren; i++) {
                var c = trigger.getChildAt(i);
                ps.push(new Laya.Vector2(c.transform.position.x, c.transform.position.z));
            }
            this.triggerArea = new PolygonWrap(ps);
            if (PlayerData.CurrentRankLevel >= PlayerData.CurrentLevel.max_rank) {
                this.Unlock(false);
            }
            else {
                this.node.active = false;
            }
        }
        Unlock(showAni = true) {
            this.unlock = true;
            this.node.active = true;
            if (!showAni)
                return;
            SoundManager.inst.playVibrate(false);
            var view = Utils.GetNode(this.node, 'Desk');
            if (view != null) {
                Utils.WaitFrame(2).then(() => {
                    var x = view.transform.localRotationEulerX;
                    view.transform.localRotationEulerX = -30;
                    view.transform.localScale = LayaUtils.Vector3Zero;
                    Laya.Tween.to(view.transform, { localPositionY: 4, localScaleX: 1, localScaleY: 1, localScaleZ: 1 }, 100, Laya.Ease.expoIn);
                    Laya.Tween.to(view.transform, { localPositionY: 0 }, 100, Laya.Ease.expoIn, null, 200);
                    Laya.Tween.to(view.transform, { localRotationEulerX: -30 }, 100, Laya.Ease.expoOut);
                    Laya.Tween.to(view.transform, { localRotationEulerX: x }, 100, Laya.Ease.expoOut, null, 200);
                });
            }
        }
        UpdateTrigger(deltaTime, playerPos) {
            if (!this.node.active)
                return;
            if (!this.unlock)
                return;
            if (this.checkInArea(playerPos)) {
                this.triggerTime += deltaTime;
                if (this.triggerTime > 0.3) {
                    if (this.isTrigger == false) {
                        this.isTrigger = true;
                        this.GetTagFromPlayer();
                    }
                }
            }
            else {
                this.isTrigger = false;
            }
        }
        checkInArea(p, useFast = false) {
            if (this.triggerArea) {
                return MathUtils.PointInPolygon2(p, this.triggerArea.points);
            }
            return false;
        }
        GetTagFromPlayer() {
            if (BattleData.GoldTag > 0) {
                BattleData.GoldTag--;
                BattleData.SilverTag++;
                MsgManager.getInst().event(BattleEvent.DeleteTagFromPlayer, true, this.tagGold);
                var tag = this.tagSilver.clone();
                MapComponent.inst.sp.addChild(tag);
                tag.transform.position = this.tagSilver.transform.position;
                MsgManager.getInst().event(BattleEvent.AddTagToPlayer, false, tag);
                if (this.isTrigger) {
                    Laya.timer.once(80, this, this.GetTagFromPlayer);
                }
            }
        }
    }

    class HangMuBuilidng {
        constructor(node) {
            this.fejis = [];
            this.node = node;
            this.outPos = Utils.GetNode(node, 'outPos');
            this.fejis.push(Utils.GetNode(node, 'jianzaiji2'));
            this.fejis.push(Utils.GetNode(node, 'jianzaiji1'));
            this.fejis.push(Utils.GetNode(node, 'jianzaiji3'));
            this.fejis.forEach(f => {
                f.active = false;
            });
        }
        Fire(line) {
            line.sort((l, r) => {
                var dis1 = Laya.Vector3.distanceSquared(this.outPos.transform.position, l);
                var dis2 = Laya.Vector3.distanceSquared(this.outPos.transform.position, r);
                if (dis1 < dis2) {
                    return -1;
                }
                if (dis2 > dis1) {
                    return 1;
                }
                return 0;
            });
            line.unshift(this.outPos.transform.position);
            this.line = line;
            MsgManager.getInst().event(BattleEvent.HideMainUI);
            for (var i = 0; i < this.fejis.length; i++) {
                var f = this.fejis[i];
                var plane = Laya.Sprite3D.instantiate(f);
                plane.active = true;
                MapComponent.inst.sp.addChild(plane);
                plane.transform.rotation = f.transform.rotation;
                plane.transform.position = f.transform.position;
                plane.transform.setWorldLossyScale(f.transform.getWorldLossyScale());
                var jzjf = plane.addComponent(JianZaiJiFire);
                jzjf.delay = i * 0.1 + (0.5 - Math.random()) * 0.1;
                jzjf.line = line;
                if (i == this.fejis.length - 2) {
                    jzjf.fake = false;
                    SimpleCameraFollow.inst.target = plane;
                    SimpleCameraFollow.inst.deltaR = 1000;
                }
            }
        }
    }
    var tmp = new Laya.Vector3();
    class JianZaiJiFire extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.fireStep = 4;
            this.maxSpeed = 20;
            this.upTime = 1;
            this.speed = new Laya.Vector3();
            this.delay = 0;
            this.line = [];
            this.forward = new Laya.Vector3();
            this.currentLineIndex = 0;
            this.bombDis = 0;
            this.fake = true;
            this.tmpR = new Laya.Quaternion();
        }
        onStart() {
            this.init();
        }
        init() {
            this.upTime = 1;
            Laya.Vector3.subtract(this.line[this.currentLineIndex], this.transform.position, this.speed);
            Laya.Vector3.normalize(this.speed, this.speed);
            Laya.Vector3.scale(this.speed, this.maxSpeed, this.speed);
        }
        onDestroy() {
        }
        onUpdate() {
            this.delay -= LayaUtils.deltaTimeSec;
            if (this.delay > 0)
                return;
            this.upTime -= LayaUtils.deltaTimeSec;
            if (this.upTime > 0) {
                this.transform.localPositionY += LayaUtils.deltaTimeSec;
            }
            else {
                Laya.Vector3.scale(this.speed, LayaUtils.deltaTimeSec, tmp);
                var dis = Laya.Vector3.scalarLength(tmp);
                Laya.Vector3.add(this.transform.position, tmp, tmp);
                this.transform.rotation.cloneTo(this.tmpR);
                this.transform.lookAt(tmp, LayaUtils.UP);
                Laya.Quaternion.slerp(this.tmpR, this.transform.rotation, 0.1, this.tmpR);
                this.transform.rotation = this.tmpR;
                this.transform.position = tmp;
                if (this.currentLineIndex >= this.line.length) {
                    this.bombDis += dis;
                    if (this.bombDis > 8) {
                        if (SimpleCameraFollow.inst.target == this.sp) {
                            MsgManager.getInst().event(BattleEvent.ShowMainUI);
                            SimpleCameraFollow.inst.target = null;
                            SimpleCameraFollow.inst.deltaR = 3;
                        }
                        this.sp.addComponent(DelayDestory).delay = 600;
                    }
                    return;
                }
                Laya.Vector3.subtract(this.line[this.currentLineIndex], tmp, this.forward);
                if (Laya.Vector3.dot(this.forward, this.speed) < 0) {
                    this.currentLineIndex++;
                    if (this.currentLineIndex < this.line.length) {
                        Laya.Vector3.subtract(this.line[this.currentLineIndex], this.transform.position, this.speed);
                        Laya.Vector3.normalize(this.speed, this.speed);
                        Laya.Vector3.scale(this.speed, this.maxSpeed, this.speed);
                    }
                }
                if (this.currentLineIndex > 1 && this.currentLineIndex < this.line.length) {
                    this.bombDis += dis;
                    if (this.bombDis > this.fireStep) {
                        this.bombDis -= this.fireStep;
                        PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.PlaneBomb).then(p => {
                            var b = Laya.Sprite3D.instantiate(p);
                            MapComponent.inst.sp.addChild(b);
                            var cc = b.addComponent(BomFallToY0);
                            cc.node = MapComponent.inst.sp;
                            b.transform.position = this.transform.position;
                            var pos = new Laya.Vector2(b.transform.position.x, b.transform.position.z);
                            cc.endCB = () => {
                                MsgManager.getInst().event(BattleEvent.KillEnemy, pos, 4, 20);
                            };
                        });
                    }
                }
            }
        }
    }

    class MechaBuilding extends BaseTriggerEntity {
        constructor(mecha) {
            super([new Laya.Vector2(mecha.sp.transform.position.x, mecha.sp.transform.position.z)]);
            this.isReady = true;
            this.isAD = false;
            this.isFree = false;
            var node = mecha.sp;
            this.com = mecha;
            this.pos = new Laya.Vector2(mecha.button.transform.position.x, mecha.button.transform.position.z);
            this.pos2 = new Laya.Vector2(mecha.jijiaNode.transform.position.x, mecha.jijiaNode.transform.position.z);
            this.model = node;
        }
        checkIn(p) {
            this.isReady = false;
            this.isFree = false;
            this.isAD = false;
            if (!PlayerData.UnlockMecha) {
                this.com.repaireSmoke.active = false;
                this.com.buttonAD.active = false;
                if (PlayerData.FreeMecha > 0) {
                    this.com.button.active = true;
                    this.isReady = true;
                    this.isFree = true;
                }
                else {
                    this.com.button.active = false;
                    return false;
                }
            }
            else {
                this.isAD = true;
                this.com.repaireSmoke.active = true;
                this.com.buttonAD.active = true;
                this.isReady = true;
            }
            return LayaUtils.distanceSquaredVec2(this.pos, p) < 4 || LayaUtils.distanceSquaredVec2(this.pos2, p) < 6.25;
        }
        EnterTrigger() {
            if (this.isReady) {
                if (this.isFree) {
                    this.mountMecha(this.isFree);
                    if (ArrowTips.inst != null) {
                        ArrowTips.inst.Hide();
                    }
                    PlayerData.FreeMecha--;
                    return;
                }
                if (this.isAD) {
                    if (PlayerData.UnlockMecha) {
                        GameSDK.showADVideo(this, b => {
                            if (b) {
                                this.mountMecha();
                            }
                        }, '3D机甲按钮');
                    }
                }
                else {
                    this.mountMecha();
                }
            }
        }
        mountMecha(free = false) {
            this.com.OpenBox();
            PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.Jijia).then(p => {
                var jijia = Laya.Sprite3D.instantiate(p);
                MapComponent.inst.sp.addChild(jijia);
                jijia.transform.position = this.com.jijiaNode.transform.position;
                Utils.WaitTime(1000).then(() => {
                    PlayerData.mountsId = free ? 3 : 4;
                    MsgManager.getInst().event(BattleEvent.MountMecha, jijia, free);
                });
                Utils.WaitTime(10000).then(() => {
                    this.com.CloseBox();
                });
            });
        }
        onClickJijia() {
            if (!!this.com) {
                this.com.OpenBox();
                Utils.WaitTime(10000).then(() => {
                    this.com.CloseBox();
                });
            }
        }
        ExitTrigger() {
        }
    }

    class MortarBulletFlyToTarget extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.mid = new Laya.Vector3();
            this.allTime = 2;
            this.time = 0;
        }
        onStart() {
            Laya.Vector3.subtract(this.targetPos, this.transform.position, this.mid);
            Laya.Vector3.scale(this.mid, 0.5, this.mid);
            Laya.Vector3.add(this.transform.position, this.mid, this.mid);
            this.mid.y += 20;
            this.curve = new QuadraticCurve(this.transform.position.clone(), this.mid.clone(), this.targetPos);
            SoundManager.inst.playSound('Bomb_Explosions');
            Utils.WaitTime(1000).then(() => {
                SoundManager.inst.playSound('Falling');
            });
        }
        onUpdate() {
            this.time += LayaUtils.deltaTimeSec;
            this.time = Math.min(this.time, this.allTime);
            this.curve.GetPos(this.time / this.allTime, this.mid);
            this.transform.lookAt(this.mid, LayaUtils.UP);
            this.transform.position = this.mid;
            if (this.time == this.allTime) {
                SoundManager.inst.playSound('Bomb_Explosions');
                PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.baozha03).then((p) => {
                    var fx = Laya.Sprite3D.instantiate(p);
                    MapComponent.inst.sp.addChild(fx);
                    fx.transform.position = this.sp.transform.position;
                    var dd = fx.addComponent(DelayDestory);
                    dd.delay = 700;
                    this.sp.destroy(true);
                    SoundManager.inst.playVibrate(true);
                });
            }
        }
    }

    class MortarBuilding {
        constructor(player, state) {
            this.triggerTime = 0;
            this.cd = 1;
            this.state = state;
            this.player = player;
            this.init(true);
        }
        get show_locked() {
            if (this.show_unlocked) {
                return false;
            }
            return this.config.show_lock <= PlayerData.CurrentRankLevel;
        }
        get show_unlocked() {
            return this.config.show_unlock <= PlayerData.CurrentRankLevel;
        }
        init(first = false) {
            this.config = ConfigHelper.BuildingMap[this.state.id];
            var node = MapCampManagerComponent.inst.SpawnMotarBuilding(this.state.slotName, this.config, !first);
            this.node = node;
            this.tag = Utils.GetNode(node, 'Tag');
            var wenzi = Utils.GetNode(node, 'WenZi');
            if (wenzi) {
                this.wenzi = wenzi.addComponent(TextFieldScript3D2);
                this.wenzi.initFonts('font');
            }
            else {
                this.wenzi = null;
            }
            var polygon = Utils.GetNode(node, 'TriggerPolygon');
            if (polygon) {
                var ps = [];
                for (var i = 0; i < polygon.numChildren; i++) {
                    var c = polygon.getChildAt(i);
                    ps.push(new Laya.Vector2(c.transform.position.x, c.transform.position.z));
                }
                this.triggerArea = new PolygonWrap(ps);
            }
            else {
                this.triggerArea = null;
            }
            this.initYuanJian();
            this.initProgress();
            this.UpdateLock();
        }
        initYuanJian() {
            var p = Utils.GetNode(this.node, 'YuanJiao');
            if (p == null)
                return;
            var mat = p.meshRenderer.material = new VertexUVMaterial();
            var c = Utils.hexToRgba01("#545454");
            c.w = 0.6;
            mat.mainColor = c;
            mat.progress = 1;
            mat.radius = 0.1;
            var p = Utils.GetNode(this.node, 'WeaponCrateArea');
            if (p == null)
                return;
            var mat = p.meshRenderer.material = new VertexUVMaterial();
            var c = Utils.hexToRgba01("#545454");
            c.w = 0.6;
            mat.mainColor = c;
            mat.progress = 1;
            mat.radius = 0.1;
        }
        initProgress() {
            var p = Utils.GetNode(this.node, 'JinDu');
            if (p == null)
                return;
            var mat = this.progress = p.meshRenderer.material = new VertexUVMaterial();
            var c = Utils.hexToRgba01("#24AAF8");
            c.w = 1;
            mat.mainColor = c;
            mat.progress = 0;
            mat.radius = 0.1;
            this.updateProgress();
        }
        updateProgress() {
            var max = this.config.cost_coin;
            var value = this.state.tag;
            if (this.progress) {
                this.progress.progress = value / max;
            }
            if (this.wenzi) {
                this.wenzi.text = `${value}/${max}`;
            }
        }
        UpdateLock() {
            var node = this.node;
            var showlock = this.config.show_lock <= PlayerData.CurrentRankLevel;
            var showUnlock = this.config.show_unlock <= PlayerData.CurrentRankLevel;
            if (showUnlock) {
                showlock = false;
            }
            var locked = Utils.GetNode(node, 'Locked');
            if (locked) {
                locked.active = showlock;
            }
            var unlocked = Utils.GetNode(node, 'Unlocked');
            if (unlocked) {
                unlocked.active = showUnlock;
            }
            var p = Utils.GetNode(this.node, 'YuanJiao');
            if (p == null)
                return;
            if (!showUnlock && !showlock) {
                p.active = false;
            }
            else {
                p.active = true;
            }
        }
        UpdateTrigger(deltaTime, ppos) {
            if (!this.node.active)
                return;
            if (this.triggerArea) {
                if (MathUtils.PointInPolygon2(ppos, this.triggerArea.points)) {
                    this.triggerTime += LayaUtils.deltaTimeSec;
                    if (this.triggerTime >= 0.1) {
                        this.triggerTime -= 0.1;
                        if (BattleData.SilverTag > 0) {
                            BattleData.SilverTag--;
                            this.state.tag++;
                            MsgManager.getInst().event(BattleEvent.DeleteTagFromPlayer, false, this.tag);
                            this.updateProgress();
                            this.CheckUpdate();
                        }
                        else {
                            if (this.config.cost_coin - this.state.tag < 5) {
                                AlertMgr.inst.setCloseCB(Laya.Handler.create(this, () => {
                                    BattleData.SilverTag += 20;
                                    MsgManager.getInst().event(BattleEvent.InitUserTag);
                                }));
                                AlertMgr.inst.showAddDogTag(20, false);
                            }
                        }
                    }
                }
                else {
                    this.triggerTime = 0;
                }
            }
        }
        CheckUpdate() {
            if (this.state.tag >= this.config.cost_coin) {
                this.state.tag = 0;
                var level = this.state.level + 1;
                var bs = ConfigHelper.CurrentLevelBuildingMap[this.state.type][this.state.slotName];
                var building;
                var maxLevel = level;
                bs.forEach(b => {
                    if (level == b.level) {
                        building = b;
                    }
                    maxLevel = Math.max(maxLevel, b.level);
                });
                if (building != null) {
                    this.state.level++;
                    this.state.id = building.id;
                    this.config = building;
                    this.init(false);
                    SoundManager.inst.playVibrate(false);
                    console.log('update mortar');
                    return true;
                }
                console.log('not update mortar', maxLevel);
            }
            return false;
        }
        initFireCD(cd) {
            this.cd = cd;
        }
        udpateFire(deltaTime) {
            this.cd -= deltaTime;
            if (this.cd <= 0) {
                this.cd += 5;
                this.FireBomb();
            }
        }
        FireBomb() {
            try {
                if (this.state.level == 0)
                    return;
                var nuke = Utils.GetNode(this.node, 'MortarBullet');
                var bomb = nuke.clone();
                MapComponent.inst.sp.addChild(bomb);
                bomb.transform.position = nuke.transform.position;
                bomb.transform.rotation = nuke.transform.rotation;
                var target = this.player.GetNearstEnemy();
                var p;
                if (target == null) {
                    var p2 = this.player.battleManager.triggerManager.flag.pos;
                    p = new Laya.Vector3(p2.x, 0, p2.y);
                }
                else {
                    p = target.PosV3.clone();
                }
                var c = bomb.addComponent(MortarBulletFlyToTarget);
                c.targetPos = p;
                Utils.WaitTime(200).then(() => {
                    MsgManager.getInst().event(BattleEvent.KillEnemy, new Laya.Vector2(p.x, p.z), 3, GlobalConstPoxy.Inst.data.MissleDamage);
                });
            }
            catch (err) {
                ReportMgr.inst.sendReport("Code_Error_MortarBuilding " + JSON.stringify(err), false);
            }
        }
    }

    class BombFlyToTaget extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.mid = new Laya.Vector3();
            this.allTime = 4;
            this.time = 0;
            this.delay = 1;
        }
        onDestroy() {
            if (this.ac && !this.ac.isStopped) {
                this.ac.stop();
            }
        }
        onStart() {
            Laya.Vector3.subtract(this.targetPos, this.transform.position, this.mid);
            Laya.Vector3.scale(this.mid, 0.5, this.mid);
            Laya.Vector3.add(this.transform.position, this.mid, this.mid);
            this.mid.y += 40;
            this.curve = new QuadraticCurve(this.transform.position.clone(), this.mid.clone(), this.targetPos);
            this.ac = SoundManager.inst.playSound('NucleerAlarm');
        }
        onUpdate() {
            this.delay -= LayaUtils.deltaTimeSec;
            if (this.delay > 0)
                return;
            this.time += LayaUtils.deltaTimeSec;
            this.time = Math.min(this.time, this.allTime);
            this.curve.GetPos(this.time / this.allTime, this.mid);
            this.transform.lookAt(this.mid, LayaUtils.UP);
            this.transform.localRotationEulerY += 90 * LayaUtils.deltaTimeSec;
            this.transform.position = this.mid;
            if (this.time == this.allTime) {
                var sp = this.sp;
                SoundManager.inst.playSound('NukeExp1');
                fgui.GTween.shake(SimpleCameraFollow.inst.camera.transform.localPositionX, SimpleCameraFollow.inst.camera.transform.localPositionZ, 1, 2).onUpdate((t) => {
                    SimpleCameraFollow.inst.camera.transform.localPositionX = t.value.x;
                    SimpleCameraFollow.inst.camera.transform.localPositionZ = t.value.y;
                }).onComplete(() => {
                    SimpleCameraFollow.inst.camera.transform.localPosition = SimpleCameraFollow.inst.cameraStaticLocalPos;
                });
                PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.Nuke05).then((p) => {
                    if (MapComponent.inst == null || MapComponent.inst.sp.destroyed)
                        return;
                    var fx = Laya.Sprite3D.instantiate(p);
                    MapComponent.inst.sp.addChild(fx);
                    fx.transform.position = sp.transform.position;
                    var dd = fx.addComponent(DelayDestory);
                    dd.delay = 4000;
                    sp.destroy(true);
                    SoundManager.inst.playVibrate(true);
                });
                this.destroy();
            }
        }
    }

    class NukeBuilding {
        constructor(node, player) {
            this.hideWaitTent = false;
            this.glassTrigger = false;
            this.buttonTrigger = false;
            this.player = player;
            this.node = node;
            this.button = Utils.GetNode(node, 'anniu');
            this.adTips = Utils.GetNode(node, 'anniu_03');
            this.buttonPos = new Laya.Vector2(this.button.transform.position.x, this.button.transform.position.z);
            this.glass = Utils.GetNode(node, 'Glass');
            this.glassPos = new Laya.Vector2(this.glass.transform.position.x, this.glass.transform.position.z);
            var wenzi = Utils.GetNode(node, 'Time');
            this.wenzi = wenzi.addComponent(TextFieldScript3D2);
            this.wenzi.initFonts('font_red');
            this.updateShow();
        }
        get config() {
            return ConfigHelper.CurrentLevelBuildingMap[BuilidngType.NukeBuilding]['NukeBuilding'][0];
        }
        get unlocked() {
            return PlayerData.UnlockNuke;
        }
        updateShow() {
            if (this.hideWaitTent) {
                this.node.active = false;
                return;
            }
            if (this.unlocked) {
                this.node.active = true;
            }
            else {
                this.node.active = false;
            }
        }
        UpdateTrigger(deltaTime, ppos) {
            if (!this.node.active)
                return;
            if (!this.unlocked) {
                return;
            }
            if (LayaUtils.distanceSquaredVec2(ppos, this.glassPos) < 9) {
                if (this.glassTrigger == false) {
                    this.glassTrigger = true;
                    this.glass.transform.localRotationEulerX = 90;
                }
            }
            else {
                if (this.glassTrigger) {
                    this.glassTrigger = false;
                    this.glass.transform.localRotationEulerX = 0;
                }
            }
            if (PlayerData.NormalNuke && ADData.nukeCD <= 0) {
                this.adTips.active = false;
            }
            else {
                ADData.nukeCD = ADData.initCD;
                this.adTips.active = true;
            }
            if (LayaUtils.distanceSquaredVec2(ppos, this.buttonPos) < 2.25) {
                if (PlayerData.FreeNuke > 0) {
                    if (GuideData.Step == GuideStepType.使用免费核弹) {
                        GuideData.Step = GuideStepType.战斗中_隐藏Tips;
                    }
                    ADData.nukeCD = ADData.initCD;
                    UIManager.inst.PushUI(AimMenu, UILayer.base, BattleEvent.FireBomb, GameManager.NukeRadius);
                    ReportMgr.inst.sendReport("3D按钮" + ReportType.NUKE_USE_FREE);
                    PlayerData.FreeNuke--;
                    GameManager.getInst().battleManager.entiytManager.player.Pos.x -= 2;
                }
                else {
                    GameSDK.showADVideo(this, (b) => {
                        GameManager.getInst().battleManager.entiytManager.player.Pos.x -= 2;
                        if (b) {
                            ADData.nukeCD = ADData.initCD;
                            UIManager.inst.PushUI(AimMenu, UILayer.base, BattleEvent.FireBomb, GameManager.NukeRadius);
                        }
                        else {
                            TipsMgr.Tips('看完整视频得奖励');
                        }
                    }, '3D按钮核弹');
                }
            }
            if (PlayerData.NormalNuke && ADData.nukeCD <= 0 || PlayerData.FreeNuke > 0) {
                this.adTips.active = false;
            }
            if (PlayerData.NormalNuke) {
                this.wenzi.text = Utils.TimeToString(ADData.nukeCD);
            }
            else {
                this.wenzi.text = '8888';
            }
        }
        FireBomb(p) {
            SoundManager.inst.playVibrate(true);
            ADData.nukeCD = ADData.initCD;
            var nuke = Utils.GetNode(this.node, 'Nuke');
            var bomb = nuke.clone();
            bomb.active = true;
            MapComponent.inst.sp.addChild(bomb);
            bomb.transform.position = nuke.transform.position;
            bomb.transform.rotation = nuke.transform.rotation;
            nuke.active = false;
            SimpleCameraFollow.inst.target = bomb;
            this.player.GetNearstEnemy();
            var c = bomb.addComponent(BombFlyToTaget);
            c.targetPos = p;
            Utils.WaitTime(5500).then(() => {
                MsgManager.getInst().event(BattleEvent.KillEnemy, new Laya.Vector2(p.x, p.z), 10.8);
            });
            Laya.timer.once(9000, this, this.onBombEnd, [bomb], false);
            MsgManager.getInst().event(BattleEvent.HideMainUI);
        }
        onBombEnd(bomb) {
            MsgManager.getInst().event(BattleEvent.ShowMainUI);
            if (SimpleCameraFollow.inst.target == bomb) {
                SimpleCameraFollow.inst.target = null;
                bomb = null;
            }
            if (this.node.destroyed)
                return;
            var nuke = Utils.GetNode(this.node, 'Nuke');
            nuke.active = true;
            if (GameManager.getInst().battleManager.needYuyue) {
                GameManager.getInst().battleManager.needYuyue = false;
                PlayerData.YuyueNuke = true;
            }
        }
    }

    class CubicCurve {
        constructor(p1, p2, p3, p4) {
            this.p1 = p1;
            this.p2 = p2;
            this.p3 = p3;
            this.p4 = p4;
        }
        GetPos(t, out = null) {
            let arg1 = Math.pow((1 - t), 3);
            let arg2 = t * (1 - t) * (1 - t) * 3;
            let arg3 = (1 - t) * t * t * 3;
            let arg4 = Math.pow(t, 3);
            var x = arg1 * this.p1.x + this.p2.x * arg2 + this.p3.x * arg3 + this.p4.x * arg4;
            var y = arg1 * this.p1.y + this.p2.y * arg2 + this.p3.y * arg3 + this.p4.y * arg4;
            var z = arg1 * this.p1.z + this.p2.z * arg2 + this.p3.z * arg3 + this.p4.z * arg4;
            out = out || new Laya.Vector3();
            out.setValue(x, y, z);
            return out;
        }
    }

    class BombSmallFlyToTarget extends BaseScript3D {
        constructor() {
            super(...arguments);
            this.mid = new Laya.Vector3();
            this.allTime = 3;
            this.time = 0;
            this.delay = 0.4;
        }
        onStart() {
            Laya.Vector3.subtract(this.targetPos, this.transform.position, this.mid);
            Laya.Vector3.scale(this.mid, 0.5, this.mid);
            Laya.Vector3.add(this.transform.position, this.mid, this.mid);
            this.mid.y += 20;
            var p1 = this.transform.position.clone();
            var p2 = p1.clone();
            p2.y += 5;
            var p3 = this.mid.clone();
            this.curve = new CubicCurve(p1, p2, p3, this.targetPos);
            Utils.WaitTime(this.delay * 1000).then(() => {
                SoundManager.inst.playSound('Bomb_Explosions');
            });
            Utils.WaitTime(1000).then(() => {
                SoundManager.inst.playSound('Falling');
            });
        }
        onUpdate() {
            this.delay -= LayaUtils.deltaTimeSec;
            if (this.delay > 0) {
                this.transform.position = this.oldTarget.transform.position;
                this.transform.rotation = this.oldTarget.transform.rotation;
                return;
            }
            this.time += LayaUtils.deltaTimeSec;
            this.time = Math.min(this.time, this.allTime);
            this.curve.GetPos(this.time / this.allTime, this.mid);
            this.transform.lookAt(this.mid, LayaUtils.UP);
            this.transform.position = this.mid;
            if (this.time == this.allTime) {
                SoundManager.inst.playSound('Bomb_Explosions');
                PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.baozha03).then((p) => {
                    var fx = Laya.Sprite3D.instantiate(p);
                    MapComponent.inst.sp.addChild(fx);
                    fx.transform.position = this.sp.transform.position;
                    var dd = fx.addComponent(DelayDestory);
                    dd.delay = 700;
                    this.sp.destroy(true);
                    SoundManager.inst.playVibrate(true);
                });
            }
        }
    }

    class ShipBuilding {
        constructor(node, player) {
            this.node = node;
            this.node.addComponent(FlyUpAndDownComponent);
        }
        get config() {
            return ConfigHelper.CurrentLevelBuildingMap[BuilidngType.chuan1]['chuanLocation'][0];
        }
        get unlocked() {
            return this.config.show_unlock <= PlayerData.CurrentRankLevel;
        }
        FireBomb(p) {
            SoundManager.inst.playSound('Splash');
            var nuke = Utils.GetNode(this.node, 'paotai');
            for (var i = 0; i < 3; i++) {
                var bombP = nuke.getChildAt(i);
                var bomb = bombP.clone();
                MapComponent.inst.sp.addChild(bomb);
                bomb.transform.position = bombP.transform.position;
                bomb.transform.rotation = bombP.transform.rotation;
                if (i == 2) {
                    SimpleCameraFollow.inst.deltaR = 50;
                    SimpleCameraFollow.inst.target = bomb;
                }
                var c = bomb.addComponent(BombSmallFlyToTarget);
                c.oldTarget = bombP;
                c.targetPos = p;
                c.delay += 0.1 * i;
                Utils.WaitTime(3400 + 100 * i).then(() => {
                    MsgManager.getInst().event(BattleEvent.KillEnemy, new Laya.Vector2(p.x, p.z), 4);
                });
                bombP.active = false;
                Laya.timer.once(2000, this, this.activeBombTarget, [bombP], false);
            }
            Laya.timer.once(4500, this, this.clearTarget, [bomb], true);
            MsgManager.getInst().event(BattleEvent.HideMainUI);
        }
        activeBombTarget(n) {
            n.active = true;
        }
        clearTarget(bomb) {
            MsgManager.getInst().event(BattleEvent.ShowMainUI);
            if (SimpleCameraFollow.inst.target == bomb) {
                SimpleCameraFollow.inst.target = null;
                SimpleCameraFollow.inst.deltaR = 3;
            }
        }
    }

    class SoliderTent {
        constructor(slotName, state, spawnPos, triggerArea, polygonArea, node) {
            this.timer = 0;
            this.cd = 3;
            this.inTriggerTime = 0;
            this.isTrigger = false;
            this.state = state;
            this.node = node;
            this.Tag = Utils.GetNode(node, 'Tag');
            this.slotName = slotName;
            this.config = ConfigHelper.BuildingMap[state.id];
            this.cd = this.config.spwan_cd;
            this.spwanCount = this.config.spawn_count;
            this.spwanSoliderType = this.config.spwan_type;
            this.spwanSoliderID = this.config.spwan_id;
            this.spawnPos = spawnPos;
            this.triggerArea = new PolygonWrap(triggerArea);
            this.polygonArea = new PolygonWrap(polygonArea);
            this.initYuanJian();
            this.initProgress();
            this.UpdateState();
        }
        checkIn(p) {
            return MathUtils.PointInPolygon2(p, this.triggerArea.points);
        }
        Update(deltaTime, playerPos) {
            if (this.config.show_lock > PlayerData.CurrentRankLevel)
                return;
            if (this.checkIn(playerPos)) {
                this.inTriggerTime += deltaTime;
                if (this.inTriggerTime >= 0.4) {
                    if (this.isTrigger == false) {
                        this.isTrigger = true;
                        this.GetTagFromPlayer();
                    }
                }
            }
            else {
                this.inTriggerTime = 0;
                this.isTrigger = false;
            }
        }
        GetTagFromPlayer() {
            if (GuideData.Step == GuideStepType.收集银牌)
                return;
            if (BattleData.SilverTag > 0 && this.state.unlocked) {
                this.state.tag++;
                BattleData.SilverTag--;
                MsgManager.getInst().event(BattleEvent.DeleteTagFromPlayer, false, this.Tag || this.node);
                this.updateProgress();
                if (this.CheckUpdate()) {
                    SoundManager.inst.playVibrate(false);
                    return;
                }
                if (this.isTrigger) {
                    Laya.timer.once(80, this, this.GetTagFromPlayer);
                }
                if (BattleData.SilverTag <= 0) {
                    if (this.config.cost_coin - this.state.tag < 5) {
                        AlertMgr.inst.setCloseCB(Laya.Handler.create(this, () => {
                            BattleData.SilverTag += 20;
                            MsgManager.getInst().event(BattleEvent.InitUserTag);
                        }));
                        AlertMgr.inst.showAddDogTag(20, false);
                    }
                }
            }
        }
        CheckUpdate() {
            if (this.state.tag >= this.config.cost_coin) {
                if (this.slotName == 'TentLocation1' && this.state.level == 0 && GuideData.Step == 1) {
                    ReportMgr.inst.sendReport(ReportType.GUIDE_BUILD_ENEMY);
                    GuideData.Step = GuideStepType.升级提示;
                }
                this.GetNextState();
                return true;
            }
        }
        GetNextState() {
            var level = this.config.level + 1;
            var bs = ConfigHelper.CurrentLevelBuildingMap[this.state.type][this.slotName];
            var building;
            var maxLevel = level;
            bs.forEach(b => {
                if (level == b.level) {
                    building = b;
                }
                maxLevel = Math.max(maxLevel, b.level);
            });
            if (building != null) {
                this.config = building;
                this.state.level = level;
                this.state.id = building.id;
                this.state.tag = 0;
                this.state.unlocked = (level == maxLevel);
                if (this.state.slotName == 'TentLocation1' && level == 1) {
                    MsgManager.getInst().event(BattleEvent.UnlockFirstSoilderTent);
                }
                if (+this.config.type == 1 && this.config.level == 1) {
                    let type;
                    switch (this.config.slot_name) {
                        case "TentLocation1":
                            type = ReportType.BUILD_ENEMY1;
                            break;
                        case "TentLocation2":
                            type = ReportType.BUILD_ENEMY2;
                            break;
                        case "TentLocation3":
                            type = ReportType.BUILD_ENEMY3;
                            break;
                        case "TentLocation4":
                            type = ReportType.BUILD_ENEMY4;
                            break;
                    }
                    if (!!type) {
                        ReportMgr.inst.sendReport(type);
                    }
                }
                MsgManager.getInst().event(BattleEvent.SpwanSoilderTent, this.slotName, building, this.state);
                return true;
            }
        }
        UpdateState() {
            this.updateLocked();
            this.updateRandk();
            this.updateProgress();
        }
        updateRandk() {
            var ranks = Utils.GetNode(this.node, 'Ranks');
            if (ranks == null)
                return;
            var icons = PrefabManager.getInst().GetPrefab(_AllPrefabsNames.RankIcons);
            var icon = icons.getChildByName(ConfigHelper.CommanderConfigLevelMap[this.config.show_unlock].level + '');
            if (icon != null) {
                ranks.meshRenderer.material = icon.meshRenderer.sharedMaterial;
            }
        }
        initYuanJian() {
            var p = Utils.GetNode(this.node, 'YuanJiao');
            if (p == null)
                return;
            this.yuanJiao = p;
            var mat = p.meshRenderer.material = new VertexUVMaterial();
            var c = Utils.hexToRgba01("#545454");
            c.w = 0.6;
            mat.mainColor = c;
            mat.progress = 1;
            mat.radius = 0.1;
        }
        initProgress() {
            var p = Utils.GetNode(this.node, 'JinDu');
            if (p == null)
                return;
            var mat = this.progress = p.meshRenderer.material = new VertexUVMaterial();
            var c = Utils.hexToRgba01("#24AAF8");
            c.w = 1;
            mat.mainColor = c;
            mat.progress = 0;
            mat.radius = 0.1;
            var wenzi = Utils.GetNode(this.node, 'WenZi');
            if (wenzi) {
                this.wenzi = wenzi.addComponent(TextFieldScript3D2);
                this.wenzi.initFonts('font');
            }
            this.updateProgress();
            var p = Utils.GetNode(this.node, 'AreaProgress');
            if (p == null)
                return;
            var mat = this.spwanProgress = p.meshRenderer.material = new VertexUVMaterial();
            var c = Utils.hexToRgba01("#639516");
            c.w = 1;
            mat.mainColor = c;
            mat.progress = 0;
            mat.radius = 0.1;
        }
        updateProgress() {
            if (this.progress) {
                var max = this.config.cost_coin;
                var value = this.state.tag;
                this.progress.progress = value / max;
            }
            if (this.wenzi) {
                this.wenzi.text = `${value}/${max}`;
            }
        }
        updateLocked() {
            this.state.unlocked = this.config.show_unlock <= PlayerData.CurrentRankLevel;
            var showUnlcoked = this.config.show_unlock <= PlayerData.CurrentRankLevel;
            var showLocked = this.config.show_lock <= PlayerData.CurrentRankLevel;
            if (showUnlcoked) {
                showLocked = false;
            }
            var locked = Utils.GetNode(this.node, 'Locked');
            if (locked) {
                locked.active = showLocked;
            }
            var unlocked = Utils.GetNode(this.node, 'Unlocked');
            if (unlocked) {
                unlocked.active = showUnlcoked;
            }
            if (this.yuanJiao) {
                this.yuanJiao.active = showLocked || showUnlcoked;
            }
            if (this.slotName == 'TentLocation1' && this.state.level == 1 && GuideData.Step == GuideStepType.战斗中_隐藏Tips && showLocked) {
                GuideData.Step = GuideStepType.银牌升级;
            }
        }
    }

    class SoliderTentManager {
        constructor(battleManager) {
            this.tents = [];
            this.battleManager = battleManager;
        }
        UpdateSpwan(deltsTime) {
            let cd;
            for (var i = 0; i < this.tents.length; i++) {
                var tent = this.tents[i];
                if (tent.spwanCount == 0)
                    continue;
                if (tent.spwanSoliderType == 1) {
                    if (this.battleManager.IsEnoughFootSolider)
                        continue;
                }
                else if (tent.spwanSoliderType == 3) {
                    if (this.battleManager.IsEnoughTank)
                        continue;
                }
                tent.timer += deltsTime;
                cd = tent.cd;
                if (GuideData.Step <= GuideStepType.开始战斗 && this.battleManager.currentSoliderSlotCount < 3) {
                    cd = 0.1;
                }
                if (tent.spwanProgress != null) {
                    tent.spwanProgress.progress = tent.timer / cd;
                }
                if (tent.timer > cd) {
                    tent.timer -= cd;
                    SoundManager.inst.playSound('SoldierSpawn');
                    MsgManager.getInst().event(BattleEvent.SpwanEntity, tent.spwanSoliderID, 1, CampType.Blue, tent.spawnPos);
                    if (tent.spwanSoliderType == 1) {
                        this.battleManager.currentSoliderSlotCount++;
                    }
                    if (tent.spwanSoliderType == 3) {
                        this.battleManager.currentTankSlotCount++;
                    }
                }
            }
        }
        UpdateTrigger(deltaTime, pos) {
            for (var i = 0; i < this.tents.length; i++) {
                var tent = this.tents[i];
                tent.Update(deltaTime, pos);
            }
        }
        AddTent(slotName, config, state, isInit = false) {
            var oldTent = this.FindTentBySlot(slotName, config.type);
            if (oldTent != null) {
                var oldIndex = this.tents.indexOf(oldTent);
                if (oldIndex >= 0) {
                    this.tents.splice(oldIndex, 1);
                }
                var i = MapComponent.inst.collisionPolygon.indexOf(oldTent.polygonArea);
                MapComponent.inst.collisionPolygon.splice(i, 1);
                MsgManager.getInst().event(BattleEvent.DestroySoilderTentToView, t);
            }
            var info = MapCampManagerComponent.inst.SpwanBuilding(slotName, config.prefab, !isInit);
            var t = new SoliderTent(slotName, state, info.spwanPos, info.trigerPolygon, info.polygon, info.node);
            t.UpdateState();
            switch (config.type) {
                case BuilidngType.SoliderTent:
                case BuilidngType.Tankhangat:
                    this.tents.push(t);
                    break;
            }
            MapComponent.inst.collisionPolygon.push(t.polygonArea);
            MsgManager.getInst().event(BattleEvent.SpwanSoilderTentToView, t);
        }
        FindTentBySlot(slotName, type) {
            switch (type) {
                case BuilidngType.SoliderTent:
                    for (var i = 0; i < this.tents.length; i++) {
                        if (this.tents[i].slotName === slotName) {
                            return this.tents[i];
                        }
                    }
                    break;
                case BuilidngType.Tankhangat:
                    return this.tankTent;
            }
            return null;
        }
        UpdateTents() {
            this.tents.forEach(tent => {
                tent.UpdateState();
            });
        }
    }

    class StorageBox {
        constructor() {
            this.blueTriggerTime = 0;
            this.bluePos = new Laya.Vector2();
            this.yellowTriggerTime = 0;
            this.yellowPos = new Laya.Vector2();
            this.mat = [];
            this.triggerTime = 0;
            this.config = ConfigHelper.BuildingMap[BattleData.StorageBox.id];
            this.init(true);
        }
        init(first = false) {
            var node = this.node = MapCampManagerComponent.inst.SpawnStorageBox(this.config, !first);
            this.blueNode = Utils.GetNode(node, 'ArrowBlue');
            if (this.blueNode) {
                this.bluePos = new Laya.Vector2(this.blueNode.transform.position.x, this.blueNode.transform.position.z);
                this.mat.push(this.blueNode.meshRenderer.material);
            }
            this.yellowNode = Utils.GetNode(node, 'ArrowYellow');
            if (this.yellowNode) {
                this.yellowPos = new Laya.Vector2(this.yellowNode.transform.position.x, this.yellowNode.transform.position.z);
                this.mat.push(this.yellowNode.meshRenderer.material);
            }
            this.tagStack = Utils.GetNode(node, 'DogTagStack');
            this.tag = Utils.GetNode(node, 'DogTag');
            this.state = BattleData.StorageBox;
            this.state.unlocked = this.config.show_unlock <= PlayerData.CurrentRankLevel;
            var polygon = Utils.GetNode(node, 'TriggerPolygon');
            if (polygon) {
                var ps = [];
                for (var i = 0; i < polygon.numChildren; i++) {
                    var c = polygon.getChildAt(i);
                    ps.push(new Laya.Vector2(c.transform.position.x, c.transform.position.z));
                }
                this.triggerArea = new PolygonWrap(ps);
            }
            var wenzi = Utils.GetNode(node, 'WenZi');
            if (wenzi) {
                this.wenzi = wenzi.addComponent(TextFieldScript3D2);
                this.wenzi.initFonts('font');
            }
            var wenziStack = Utils.GetNode(node, 'WenZiStack');
            if (wenziStack) {
                this.wenziStack = wenziStack.addComponent(TextFieldScript3D2);
                this.wenziStack.initFonts('font');
            }
            this.UpdateLock();
            this.initYuanJian();
            this.initProgress();
            this.updateProgressStack();
            this.CheckUpdate();
            if (this.state.level > 0) {
                for (var i = 0; i < PlayerData.StorageTag; i++) {
                    this.addDogtagToStack();
                }
            }
        }
        initYuanJian() {
            var p = Utils.GetNode(this.node, 'YuanJiao');
            if (p) {
                var mat = p.meshRenderer.material = new VertexUVMaterial();
                var c = Utils.hexToRgba01("#545454");
                c.w = 0.6;
                mat.mainColor = c;
                mat.progress = 1;
                mat.radius = 0.1;
            }
            var p = Utils.GetNode(this.node, 'YuanJiao1');
            if (p) {
                var mat = p.meshRenderer.material = new VertexUVMaterial();
                var c = Utils.hexToRgba01("#545454");
                c.w = 0.6;
                mat.mainColor = c;
                mat.progress = 1;
                mat.radius = 0.1;
            }
            var p = Utils.GetNode(this.node, 'YuanJiao2');
            if (p) {
                var mat = p.meshRenderer.material = new VertexUVMaterial();
                var c = Utils.hexToRgba01("#545454");
                c.w = 0.6;
                mat.mainColor = c;
                mat.progress = 1;
                mat.radius = 0.1;
            }
        }
        initProgress() {
            var p = Utils.GetNode(this.node, 'JinDu');
            if (p == null)
                return;
            var mat = this.progress = p.meshRenderer.material = new VertexUVMaterial();
            var c = Utils.hexToRgba01("#24AAF8");
            c.w = 1;
            mat.mainColor = c;
            mat.progress = 0;
            mat.radius = 0.1;
            this.updateProgress();
        }
        updateProgress() {
            var max = this.config.cost_coin;
            var value = this.state.tag;
            if (this.progress) {
                this.progress.progress = value / max;
            }
            if (this.wenzi) {
                this.wenzi.text = `${value}/${max}`;
            }
        }
        updateProgressStack() {
            var max = this.config.spawn_count;
            var value = PlayerData.StorageTag;
            if (this.wenziStack) {
                this.wenziStack.text = `${value}/${max}`;
            }
        }
        get unlocked() {
            return this.config.show_lock <= PlayerData.CurrentRankLevel || PlayerData.UnlockStorage;
        }
        UpdateLock() {
            var node = this.node;
            var showlock = this.config.show_lock <= PlayerData.CurrentRankLevel;
            var showUnlock = this.config.show_unlock <= PlayerData.CurrentRankLevel;
            if (showUnlock) {
                showlock = false;
            }
            var locked = Utils.GetNode(node, 'Locked');
            if (locked) {
                locked.active = showlock;
            }
            var unlocked = Utils.GetNode(node, 'Unlocked');
            if (unlocked) {
                unlocked.active = showUnlock;
            }
        }
        UpdateTrigger(deltaTime, playerPos) {
            if (!this.node.active)
                return;
            if (this.state.level == 0) {
                this.state.unlocked = this.config.show_unlock <= PlayerData.CurrentRankLevel;
                if (this.state.unlocked == false)
                    return;
                if (MathUtils.PointInPolygon2(playerPos, this.triggerArea.points)) {
                    this.triggerTime += LayaUtils.deltaTimeSec;
                    if (this.triggerTime > 0.1) {
                        this.triggerTime -= 0.1;
                        if (BattleData.SilverTag > 0) {
                            BattleData.SilverTag--;
                            this.state.tag++;
                            MsgManager.getInst().event(BattleEvent.DeleteTagFromPlayer, false, this.tag);
                            this.updateProgress();
                            this.CheckUpdate();
                        }
                    }
                }
                else {
                    this.triggerTime = 0;
                }
            }
            else {
                if (PlayerData.StorageTag < this.config.spawn_count && LayaUtils.distanceSquaredVec2(playerPos, this.bluePos) < 1) {
                    this.blueTriggerTime += LayaUtils.deltaTimeSec;
                    if (this.blueTriggerTime > 0.1) {
                        this.blueTriggerTime -= 0.1;
                        if (BattleData.SilverTag > 0) {
                            BattleData.SilverTag--;
                            PlayerData.StorageTag++;
                            MsgManager.getInst().event(BattleEvent.DeleteTagFromPlayer, false, null);
                            this.addDogtagToStack();
                            this.updateProgressStack();
                        }
                    }
                }
                else {
                    this.blueTriggerTime = 0;
                }
                if (PlayerData.StorageTag > 0 && LayaUtils.distanceSquaredVec2(playerPos, this.yellowPos) < 1) {
                    this.yellowTriggerTime += LayaUtils.deltaTimeSec;
                    if (this.yellowTriggerTime > 0.1) {
                        this.yellowTriggerTime -= 0.1;
                        BattleData.SilverTag++;
                        PlayerData.StorageTag--;
                        MsgManager.getInst().event(BattleEvent.AddTagToPlayer, false, this.removeDogTagFromStack());
                        this.updateProgressStack();
                    }
                }
                else {
                    this.yellowTriggerTime = 0;
                }
                this.mat.forEach(m => {
                    m.tilingOffsetZ -= 1 * deltaTime;
                });
            }
        }
        addDogtagToStack() {
            if (this.tagStack.numChildren >= 20)
                return;
            var tag = Laya.Sprite3D.instantiate(PrefabManager.getInst().GetPrefab(_AllPrefabsNames.DogTag));
            this.tagStack.addChild(tag);
            tag.transform.localRotationEuler = LayaUtils.Vector3Zero;
            tag.transform.localPosition = new Laya.Vector3(0, this.tagStack.numChildren * 0.1, 0);
            console.log('add', this.tagStack.numChildren);
        }
        removeDogTagFromStack() {
            console.log(this.tagStack.numChildren);
            var c = this.tagStack.getChildAt(this.tagStack.numChildren - 1);
            var oldP = c.transform.position;
            var oldE = c.transform.rotation;
            MapCampManagerComponent.inst.sp.addChild(c);
            c.transform.position = oldP;
            c.transform.rotation = oldE;
            return c;
        }
        CheckUpdate() {
            if (this.state.tag >= this.config.cost_coin || PlayerData.UnlockStorage) {
                PlayerData.UnlockStorage = true;
                this.state.tag = 0;
                var level = this.state.level + 1;
                var bs = ConfigHelper.CurrentLevelBuildingMap[this.state.type][this.state.slotName];
                var building;
                var maxLevel = level;
                bs.forEach(b => {
                    if (level == b.level) {
                        building = b;
                    }
                    maxLevel = Math.max(maxLevel, b.level);
                });
                if (building != null) {
                    this.state.level++;
                    this.state.id = building.id;
                    this.config = building;
                    this.init();
                    SoundManager.inst.playVibrate(false);
                    return true;
                }
            }
            return false;
        }
    }

    class BattleTriggerEntity extends BaseTriggerEntity {
        constructor(node) {
            super([]);
            this.pos = new Laya.Vector2(node.transform.position.x, node.transform.position.z);
            var p = Utils.GetNode(node, 'YuanJiao');
            if (p) {
                var mat = p.meshRenderer.material = new VertexUVMaterial();
                var c = Utils.hexToRgba01("#545454");
                c.w = 0.6;
                mat.mainColor = c;
                mat.progress = 1;
                mat.radius = 0.1;
            }
            this.needRestTriggerState = true;
        }
        checkIn(p) {
            var dx = Math.abs(this.pos.x - p.x);
            var dy = Math.abs(this.pos.y - p.y);
            return (dx <= 3 && dy <= 4);
        }
        ExitTrigger() {
            MsgManager.getInst().event(BattleEvent.TurnOffBattle);
        }
        EnterTrigger() {
            if (GuideData.Step < GuideStepType.开始战斗)
                return;
            if (GuideData.Step == GuideStepType.开始战斗) {
                GuideData.Step = GuideStepType.战斗中_隐藏Tips;
            }
            MsgManager.getInst().event(BattleEvent.TurnOnBattle);
        }
    }

    class EnemyFlagEntity {
        constructor(node) {
            this.isTriggered = false;
            this.pos = new Laya.Vector2(node.transform.position.x, node.transform.position.z);
            this.node = node;
            this.redFlag = Utils.GetNode(node, 'RedFlag');
            this.blueFlag = Utils.GetNode(node, 'BlueFlag');
        }
        Update(ppos) {
            if (!this.isTriggered && LayaUtils.distanceSquaredVec2(this.pos, ppos) < 4) {
                this.Trigger();
            }
        }
        Trigger() {
            this.isTriggered = true;
            SoundManager.inst.playSound('Flag');
            Utils.WaitTime(1000).then(() => {
                SoundManager.inst.playSound('Flag');
            });
            Laya.Tween.to(this.redFlag.transform, { localPositionY: -1.66 }, 1000);
            Laya.Tween.to(this.blueFlag.transform, { localPositionY: 7 }, 1000, Laya.Ease.linearInOut, null, 1000);
            Utils.WaitTime(2000).then(() => {
                SoundManager.inst.playSound('Base_Capture');
            });
            Utils.WaitTime(3000).then(() => {
                Utils.WaitTime(100).then(() => {
                    MsgManager.getInst().event(BattleEvent.TakeEnemyFlag);
                });
            });
        }
    }

    class PlayerLocationTrigger extends BaseTriggerEntity {
        constructor(state, pos, node) {
            super(pos);
            this.node = node;
            this.state = state;
            this.tag = Utils.GetNode(node, 'Tag');
            this.oldLocalScale = this.tag.transform.localScale.clone();
            var wenzi = Utils.GetNode(node, 'WenZi');
            this.wenzi = wenzi.addComponent(TextFieldScript3D2);
            this.wenzi.initFonts('font');
            var title = Utils.GetNode(node, 'Title');
            this.wenziTitle = title.addComponent(TextFieldScript3D2);
            this.wenziTitle.initFonts('font');
            this.initYuanJian();
            this.initProgress();
            this.updateRandk();
            this.updateLocked();
        }
        ExitTrigger() {
        }
        updateRandk() {
            var ranks = Utils.GetNode(this.node, 'Ranks');
            if (ranks == null)
                return;
            var icons = PrefabManager.getInst().GetPrefab(_AllPrefabsNames.RankIcons);
            var icon = icons.getChildByName(PlayerData.CurrentCommanderConfig.level + '');
            if (icon != null) {
                ranks.meshRenderer.material = icon.meshRenderer.sharedMaterial;
            }
            this.wenziTitle.text = PlayerData.CurrentCommanderConfig.name;
        }
        initYuanJian() {
            var p = Utils.GetNode(this.node, 'YuanJiao');
            var mat = p.meshRenderer.material = new VertexUVMaterial();
            var c = Utils.hexToRgba01("#545454");
            c.w = 0.6;
            mat.mainColor = c;
            mat.progress = 1;
            mat.radius = 0.1;
        }
        initProgress() {
            var p = Utils.GetNode(this.node, 'JinDu');
            var mat = this.progress = p.meshRenderer.material = new VertexUVMaterial();
            var c = Utils.hexToRgba01("#24AAF8");
            c.w = 1;
            mat.mainColor = c;
            mat.progress = 0;
            mat.radius = 0.1;
            this.updateProgress();
        }
        updateProgress() {
            var max = PlayerData.CurrentCommanderConfig.cost_gold;
            var value = this.state.tag;
            this.progress.progress = value / max;
            this.wenzi.text = `${value}/${max}`;
        }
        EnterTrigger() {
            if (!this.state.unlocked)
                return;
            if (GuideData.Step < GuideStepType.升级提示)
                return;
            if (GuideData.Step == GuideStepType.升级提示) {
                if (BattleData.GoldTag < (PlayerData.CurrentCommanderConfig.cost_gold - this.state.tag))
                    return;
            }
            if (BattleData.GoldTag > 0) {
                this.state.tag++;
                BattleData.GoldTag--;
                Laya.Tween.to(this.tag.transform, { localScaleX: 0.9, localScaleY: 0.9, localScaleZ: 0.9 }, 100);
                Laya.Tween.to(this.tag.transform, { localScaleX: this.oldLocalScale.x, localScaleY: this.oldLocalScale.y, localScaleZ: this.oldLocalScale.z }, 100, Laya.Ease.expoInOut, null, 100);
                MsgManager.getInst().event(BattleEvent.DeleteTagFromPlayer, true, this.tag);
                if (this.CheckUpdate()) {
                    if (GuideData.Step == GuideStepType.升级提示) {
                        GuideData.Step = GuideStepType.战斗中_隐藏Tips;
                    }
                    ReportMgr.inst.sendReport(ReportType.GUIDE_LEVEL_UP);
                    ReportMgr.inst.sendReport(ReportType.LEVEL_UP + PlayerData.CurrentRankLevel);
                    this.isTrigger = false;
                    return;
                }
                if (this.isTrigger) {
                    Laya.timer.once(80, this, this.EnterTrigger);
                }
                this.updateProgress();
            }
            else if (GuideData.Step > GuideStepType.升级提示) {
                if (PlayerData.CurrentCommanderConfig.cost_gold - this.state.tag < 5) {
                    AlertMgr.inst.setCloseCB(Laya.Handler.create(this, () => {
                        BattleData.GoldTag += 20;
                        MsgManager.getInst().event(BattleEvent.InitUserTag);
                    }));
                    AlertMgr.inst.showAddDogTag(20, true);
                }
            }
        }
        CheckUpdate() {
            if (this.state.tag >= PlayerData.CurrentCommanderConfig.cost_gold) {
                PlayerData.CurrentRankLevel++;
                this.state.tag = 0;
                this.state.level = PlayerData.CurrentRankLevel;
                this.updateRandk();
                this.updateProgress();
                this.updateLocked();
                return true;
            }
        }
        updateLocked() {
            var level = PlayerData.CurrentLevel;
            this.state.unlocked = level.max_rank > PlayerData.CurrentRankLevel;
            var locked = Utils.GetNode(this.node, 'Locked');
            if (locked != null) {
                locked.active = !this.state.unlocked;
            }
            var unlocked = Utils.GetNode(this.node, 'Unlocked');
            if (unlocked != null) {
                unlocked.active = this.state.unlocked;
            }
        }
    }

    class TriggerManger {
        constructor() {
            this.boxes = [];
        }
        AddPlayer(node) {
            var p = Utils.GetNode(node, 'TriggerPolygon');
            var ps = [];
            if (p) {
                for (var i = 0; i < p.numChildren; i++) {
                    var c = p.getChildAt(i);
                    var p2 = new Laya.Vector2(c.transform.position.x, c.transform.position.z);
                    ps.push(p2);
                }
            }
            var t = new PlayerLocationTrigger(BattleData.PlayerNode, ps, node);
            this.player = t;
        }
        Update(deltaTime, playerPos) {
            this.player.Update(deltaTime, playerPos);
            this.updateBoxTrigger(playerPos);
            this.udpateFreeNuke(playerPos);
        }
        AddFlag(node) {
            this.flag = new EnemyFlagEntity(node);
        }
        UpdateWinState(deltaTime, ppos) {
            this.flag.Update(ppos);
        }
        AddBattleTrigger(node) {
            this.battleTrigger = new BattleTriggerEntity(node);
        }
        UpdateBattleTrigger(deltaTime, ppos) {
            this.battleTrigger.Update(deltaTime, ppos);
        }
        addBoxTrigger(t) {
            this.boxes.push(t);
        }
        updateBoxTrigger(ppos) {
            for (var i = this.boxes.length - 1; i > -1; i--) {
                var box = this.boxes[i];
                if (box.checkIn(ppos)) {
                    box.EnterTrigger();
                    this.boxes.splice(i, 1);
                }
            }
        }
        addNukeTrigger(t) {
            this.freeNuke = t;
        }
        udpateFreeNuke(ppos) {
            if (this.freeNuke == null)
                return;
            if (this.freeNuke.checkIn(ppos)) {
                ReportMgr.inst.sendReport(ReportType.NUKE_GET_FREE);
                GuideData.Step = GuideStepType.使用免费核弹;
                this.freeNuke.EnterTrigger();
                this.freeNuke = null;
            }
        }
    }

    class WeaponFactory {
        constructor(node) {
            this.inTriggerTime = 0;
            this.isTrigger = false;
            this.inTagTriggerTime = 0;
            this.isTagTrigger = false;
            this.spwanTimer = 0;
            this.maxStack = 40;
            this.slotName = 'AmmoFactory';
            this.node = node;
            this.tag = Utils.GetNode(node, 'Tag');
            this.tag2 = Utils.GetNode(node, 'Tag2');
            this.config = ConfigHelper.BuildingMap[BattleData.WeaponFactory.id];
            this.state = BattleData.WeaponFactory;
            this.initYuanJian();
            this.initProgress();
            this.WeaponCrates = Utils.GetNode(node, 'WeaponCrates');
            var triggerNode = Utils.GetNode(node, 'TriggerPolygon');
            if (triggerNode) {
                var ps = [];
                for (var i = 0; i < triggerNode.numChildren; i++) {
                    var c = triggerNode.getChildAt(i);
                    ps.push(new Laya.Vector2(c.transform.position.x, c.transform.position.z));
                }
                this.triggerArea = new PolygonWrap(ps);
            }
            var triggerNode = Utils.GetNode(node, 'TriggerPolygonTag');
            if (triggerNode) {
                var ps = [];
                for (var i = 0; i < triggerNode.numChildren; i++) {
                    var c = triggerNode.getChildAt(i);
                    ps.push(new Laya.Vector2(c.transform.position.x, c.transform.position.z));
                }
                this.tagTriggderArea = new PolygonWrap(ps);
            }
            var polygonNode = Utils.GetNode(node, 'polygon');
            if (polygonNode) {
                var ps = [];
                for (var i = 0; i < polygonNode.numChildren; i++) {
                    var c = polygonNode.getChildAt(i);
                    ps.push(new Laya.Vector2(c.transform.position.x, c.transform.position.z));
                }
                this.polygonArea = new PolygonWrap(ps);
            }
            var wenzi = Utils.GetNode(node, 'WenZi');
            if (wenzi) {
                this.wenzi = wenzi.addComponent(TextFieldScript3D2);
                this.wenzi.initFonts('font');
            }
            var wenziTag = Utils.GetNode(node, 'TagWenZi');
            if (wenziTag) {
                this.wenziTag = wenziTag.addComponent(TextFieldScript3D2);
                this.wenziTag.initFonts('font');
            }
            this.udpateState();
            this.initWeapon();
            this.udpateTagsToCreateWeapon();
        }
        initWeapon() {
            BattleData.WeaponCountInFactory = Math.max(0, BattleData.WeaponCountInFactory);
            BattleData.WeaponCountInFactory = Math.min(this.maxStack, BattleData.WeaponCountInFactory);
            for (var i = 0; i < BattleData.WeaponCountInFactory; i++) {
                var wn = this.WeaponCrates.getChildAt(i);
                var weaponConfig = ConfigHelper.WeaponConfigMap[this.config.spwan_id];
                var weaponPrefab = PrefabManager.getInst().GetPrefab(weaponConfig.box_prefa);
                var weapon = Laya.Sprite3D.instantiate(weaponPrefab);
                wn.addChild(weapon);
                weapon.transform.localPosition = LayaUtils.Vector3Zero;
                weapon.transform.localPositionY += 0.5;
                weapon.transform.localRotationEuler = LayaUtils.Vector3Zero;
                weapon.transform.localScale = new Laya.Vector3(1.2, 1.15, 3.16);
            }
        }
        DeltaWeapon() {
            this.WeaponCrates.numChildren;
            var wn = this.WeaponCrates.getChildAt(BattleData.WeaponCountInFactory - 1);
            wn.removeChildren();
        }
        checkInArea(p, useFast = false) {
            if (this.triggerArea) {
                return MathUtils.PointInPolygon2(p, this.triggerArea.points);
            }
            return false;
        }
        checkInTagArea(p) {
            if (this.tagTriggderArea) {
                return MathUtils.PointInPolygon2(p, this.tagTriggderArea.points);
            }
            return false;
        }
        UpdateSpwanWeapon(deltaTime) {
            if (this.config.spawn_count == 0)
                return;
            if (BattleData.WeaponTagToCreateWeapon <= 0 || BattleData.WeaponCountInFactory >= this.maxStack)
                return;
            this.spwanTimer += deltaTime;
            if (this.spwanTimer > this.config.spwan_cd) {
                this.spwanTimer -= this.config.spwan_cd;
                for (var i = 0; i < 2; i++) {
                    BattleData.WeaponCountInFactory++;
                    if (BattleData.WeaponCountInFactory > this.maxStack) {
                        BattleData.WeaponCountInFactory = this.maxStack;
                        continue;
                    }
                    var wn = this.WeaponCrates.getChildAt(BattleData.WeaponCountInFactory - 1);
                    var weaponConfig = ConfigHelper.WeaponConfigMap[this.config.spwan_id];
                    var weaponPrefab = PrefabManager.getInst().GetPrefab(weaponConfig.box_prefa);
                    var weapon = Laya.Sprite3D.instantiate(weaponPrefab);
                    wn.addChild(weapon);
                    weapon.transform.localPosition = LayaUtils.Vector3Zero;
                    weapon.transform.localPositionY += 0.5;
                    weapon.transform.localRotationEuler = LayaUtils.Vector3Zero;
                    weapon.transform.localScale = new Laya.Vector3(1.2, 1.15, 3.16);
                }
                BattleData.WeaponTagToCreateWeapon--;
                this.udpateTagsToCreateWeapon();
            }
        }
        UpdateTrigger(deltaTime, playerPos) {
            if (this.checkInArea(playerPos)) {
                this.inTriggerTime += deltaTime;
                if (this.inTriggerTime >= 0.4) {
                    this.inTriggerTime = 0;
                    if (this.isTrigger == false) {
                        this.isTrigger = true;
                        this.EnterTrigger();
                    }
                }
            }
            else {
                this.inTriggerTime = 0;
                if (this.isTrigger) {
                    this.isTrigger = false;
                    this.ExitTrigger();
                }
            }
            if (this.checkInTagArea(playerPos)) {
                this.inTagTriggerTime += deltaTime;
                if (this.inTagTriggerTime >= 0.4) {
                    this.inTagTriggerTime = 0;
                    if (this.isTagTrigger == false) {
                        this.isTagTrigger = true;
                        this.EnterTagTrigger();
                    }
                }
            }
            else {
                this.inTagTriggerTime = 0;
                if (this.isTagTrigger) {
                    this.isTagTrigger = false;
                    this.ExitTagTrigger();
                }
            }
        }
        EnterTrigger() {
            if (this.state.unlocked) {
                this.GetTagFromPlayer();
            }
        }
        ExitTrigger() {
        }
        GetTagFromPlayer() {
            if (BattleData.SilverTag > 0) {
                this.state.tag++;
                BattleData.SilverTag--;
                MsgManager.getInst().event(BattleEvent.DeleteTagFromPlayer, false, this.tag);
                this.updateProgress();
                if (this.CheckUpdate()) {
                    SoundManager.inst.playVibrate(false);
                    this.udpateState();
                    this.initWeapon();
                    return;
                }
                if (this.isTrigger) {
                    Laya.timer.once(120, this, this.GetTagFromPlayer);
                }
            }
            else {
                if (this.config.cost_coin - this.state.tag < 5) {
                    AlertMgr.inst.setCloseCB(Laya.Handler.create(this, () => {
                        BattleData.SilverTag += 20;
                        MsgManager.getInst().event(BattleEvent.InitUserTag);
                    }));
                    AlertMgr.inst.showAddDogTag(20, false);
                }
            }
        }
        CheckUpdate() {
            if (this.state.tag >= this.config.cost_coin) {
                this.GetNextState();
                return true;
            }
        }
        GetNextState() {
            var level = this.config.level + 1;
            var bs = ConfigHelper.CurrentLevelBuildingMap[this.state.type][this.slotName];
            var building;
            var maxLevel = level;
            bs.forEach(b => {
                if (level == b.level) {
                    building = b;
                }
                maxLevel = Math.max(maxLevel, b.level);
            });
            if (building != null) {
                this.config = building;
                this.state.level = level;
                this.state.id = building.id;
                this.state.tag = 0;
                this.state.unlocked = (level == maxLevel);
                MsgManager.getInst().event(BattleEvent.SpwanWeaponFactoryTent, this.slotName, building, this.state);
                return true;
            }
        }
        udpateState() {
            this.updateLocked();
            this.updateProgress();
            this.updateRandk();
        }
        updateRandk() {
            var ranks = Utils.GetNode(this.node, 'Ranks');
            if (ranks == null)
                return;
            var icons = PrefabManager.getInst().GetPrefab(_AllPrefabsNames.RankIcons);
            var icon = icons.getChildByName(ConfigHelper.CommanderConfigLevelMap[this.config.show_unlock].level + '');
            if (icon != null) {
                ranks.meshRenderer.material = icon.meshRenderer.sharedMaterial;
            }
        }
        initYuanJian() {
            var p = Utils.GetNode(this.node, 'YuanJiao');
            if (p == null)
                return;
            var mat = p.meshRenderer.material = new VertexUVMaterial();
            var c = Utils.hexToRgba01("#545454");
            c.w = 0.6;
            mat.mainColor = c;
            mat.progress = 1;
            mat.radius = 0.1;
            var p = Utils.GetNode(this.node, 'WeaponCrateArea');
            if (p == null)
                return;
            var mat = p.meshRenderer.material = new VertexUVMaterial();
            var c = Utils.hexToRgba01("#545454");
            c.w = 0.6;
            mat.mainColor = c;
            mat.progress = 1;
            mat.radius = 0.1;
        }
        initProgress() {
            var p = Utils.GetNode(this.node, 'JinDu');
            if (p == null)
                return;
            var mat = this.progress = p.meshRenderer.material = new VertexUVMaterial();
            var c = Utils.hexToRgba01("#24AAF8");
            c.w = 1;
            mat.mainColor = c;
            mat.progress = 0;
            mat.radius = 0.1;
            this.updateProgress();
        }
        updateProgress() {
            var max = this.config.cost_coin;
            var value = this.state.tag;
            if (this.progress) {
                this.progress.progress = value / max;
            }
            if (this.wenzi) {
                console.log(value, max);
                this.wenzi.text = `${value}/${max}`;
            }
        }
        updateLocked() {
            this.state.unlocked = this.config.show_unlock <= PlayerData.CurrentRankLevel;
            var showUnlcoked = this.config.show_unlock <= PlayerData.CurrentRankLevel;
            var showLocked = this.config.show_lock <= PlayerData.CurrentRankLevel;
            if (showUnlcoked) {
                showLocked = false;
            }
            var locked = Utils.GetNode(this.node, 'Locked');
            if (locked) {
                locked.active = showLocked;
            }
            var unlocked = Utils.GetNode(this.node, 'Unlocked');
            if (unlocked) {
                unlocked.active = showUnlcoked;
            }
            var p = Utils.GetNode(this.node, 'YuanJiao');
            if (p) {
                if (!showUnlcoked && !showLocked) {
                    p.active = false;
                }
                else {
                    p.active = true;
                }
            }
        }
        EnterTagTrigger() {
            this.GetTagFromPlayer2();
        }
        ExitTagTrigger() {
        }
        GetTagFromPlayer2() {
            if (BattleData.SilverTag > 0 && BattleData.WeaponTagToCreateWeapon < 20) {
                BattleData.WeaponTagToCreateWeapon++;
                BattleData.SilverTag--;
                MsgManager.getInst().event(BattleEvent.DeleteTagFromPlayer, false, this.tag2);
                if (this.isTagTrigger) {
                    Laya.timer.once(80, this, this.GetTagFromPlayer2);
                }
                this.udpateTagsToCreateWeapon();
            }
        }
        udpateTagsToCreateWeapon() {
            if (this.wenziTag) {
                this.wenziTag.text = `${BattleData.WeaponTagToCreateWeapon}/${20}`;
            }
        }
    }

    class BattleManager {
        constructor() {
            this.isBattleState = false;
            this.currentSoliderSlotCount = 0;
            this.enemyLoaded = false;
            this.mountHorse = false;
            this.aimMode = false;
            this.needYuyue = false;
            this.spwanFreeNuke = false;
            this.freeMechaTime = 30;
            this.currentTankSlotCount = 0;
        }
        get MaxSoliderCount() {
            return PlayerData.CurrentCommanderConfig.soldiers;
        }
        get MaxTankSoliderCount() {
            return PlayerData.CurrentCommanderConfig.tanks;
        }
        get IsEnoughFootSolider() {
            return this.currentSoliderSlotCount >= this.MaxSoliderCount;
        }
        get IsEnoughTank() {
            return this.currentTankSlotCount >= this.MaxTankSoliderCount;
        }
        StartNewGame() {
            this.entiytManager = new BattleEntityManager(this);
            this.damageManager = new DamageManager();
            this.soliderTentManager = new SoliderTentManager(this);
            this.dogTagManager = new DogTagManager();
            this.triggerManager = new TriggerManger();
            MsgManager.getInst().on(BattleEvent.SpwanWeaponFactoryTent, this, this.createWeaponFactory);
            MsgManager.getInst().on(BattleEvent.EnemyLoadEnd, this, this.onEnemyLoadEnd);
            MsgManager.getInst().on(BattleEvent.DeleteWeaponFromfactory, this, this.GetWeaponFromFactory);
            MsgManager.getInst().on(BattleEvent.KillEnemy, this, this.KillEnemy);
            MsgManager.getInst().on(BattleEvent.FireBomb, this, this.FireBomb);
            MsgManager.getInst().on(BattleEvent.FireShip, this, this.FireShip);
            MsgManager.getInst().on(BattleEvent.FireHangMu, this, this.FireHangMu);
            MsgManager.getInst().on(BattleEvent.RankLevelUp, this, this.onRankUP);
            MsgManager.getInst().on(BattleEvent.CallHelicopter, this, this.onSpawnHelicopter);
            MsgManager.getInst().on(BattleEvent.CreateDogTag, this, this.onCreateDogTag);
            MsgManager.getInst().on(BattleEvent.CallDropBox, this, this.onCallDropBox);
            MsgManager.getInst().on(BattleEvent.CallBomber, this, this.onCallBomber);
            MsgManager.getInst().on(BattleEvent.SpwanPlayer, this, this.onCreatePlayer);
            MsgManager.getInst().on(BattleEvent.SpwanDropBoxTrigger, this, this.onSpwanBoxTrigger);
            MsgManager.getInst().on(BattleEvent.SpwanFreeNukeTrigger, this, this.onSpwanFreeNuke);
            MsgManager.getInst().on(BattleEvent.MountHorse, this, this.onMountHorse);
            MsgManager.getInst().on(BattleEvent.InitUserTag, this, this.onInitUserTag);
            MsgManager.getInst().on(BattleEvent.MountMecha, this, this.onMountJijia);
            this.mountHorse = false;
        }
        ClearGame() {
            let mountsId = PlayerData.mountsId;
            this.entiytManager.Dispose();
            if (mountsId == 2) {
                PlayerData.mountsId = mountsId;
                ADData.HorseCD = 0;
            }
            else if (mountsId == 4) {
                PlayerData.mountsId = mountsId;
                ADData.MechaCD = 0;
            }
            MsgManager.getInst().offAllCaller(this);
        }
        onEnemyLoadEnd() {
            this.entiytManager.RefreshCampGroup();
            var enemeyCount = this.entiytManager.GetEnemyGroup(CampType.Blue).length;
            MsgManager.getInst().event(BattleEvent.EnemyCountInfo, enemeyCount, MapComponent.inst.allEnemyCount);
            this.enemyLoaded = true;
            this.createLine();
        }
        createLine() {
            var pos = [];
            var flagPos = MapComponent.inst.Find('EnemyFlag').transform.position;
            var center = new Laya.Vector2(flagPos.x, flagPos.z);
            pos.push(new Laya.Vector2(center.x + 2, center.y));
            pos.push(new Laya.Vector2(center.x - 2, center.y));
            pos.push(new Laya.Vector2(center.x, center.y + 2));
            pos.push(new Laya.Vector2(center.x, center.y - 2));
            this.entiytManager.GetEnemyGroup(CampType.Blue).forEach(p => {
                pos.push(p.Pos.clone());
            });
            MapCampManager.inst.SetPoints(pos);
        }
        createWeaponFactory(fx = true) {
            var state = BattleData.WeaponFactory;
            var config = ConfigHelper.BuildingMap[state.id];
            var b = MapCampManagerComponent.inst.SpwanWeaponFactory('AmmoFactory', config, fx);
            if (this.weaponfacory) {
                var i = MapComponent.inst.collisionPolygon.indexOf(this.weaponfacory.polygonArea);
                MapComponent.inst.collisionPolygon.splice(i, 1);
            }
            this.weaponfacory = new WeaponFactory(b);
            if (this.weaponfacory.polygonArea) {
                MapComponent.inst.collisionPolygon.push(this.weaponfacory.polygonArea);
            }
        }
        GetWeaponFromFactory() {
            if (this.weaponfacory == null)
                return;
            this.weaponfacory.DeltaWeapon();
        }
        createNuke() {
            var config = ConfigHelper.CurrentLevelBuildingMap[BuilidngType.NukeBuilding]['NukeBuilding'][0];
            var b = MapCampManagerComponent.inst.SpawnNuke('NukeBuilding', config);
            this.nuke = new NukeBuilding(b, this.entiytManager.player);
        }
        createShip() {
            var b = MapCampManagerComponent.inst.SpawnShip();
            this.ship = new ShipBuilding(b, this.entiytManager.player);
        }
        createExchange() {
            var b = MapCampManagerComponent.inst.SpwanExchange(false);
            this.exchange = new ExchageBuilding(b);
        }
        createStorageBox() {
            this.storage = new StorageBox();
        }
        createMortars() {
            this.mortars = [];
            BattleData.Mortars.forEach(m => {
                var mb = new MortarBuilding(this.entiytManager.player, m);
                this.mortars.push(mb);
            });
        }
        createAirport() {
            this.airport = new AirportBuilding();
            this.airport.udpateShow();
        }
        createMechaBox() {
            MapCampManagerComponent.inst.SpwanMechaBox().then(mb => {
                this.mechaBox = new MechaBuilding(mb);
                if (this.soliderTentManager.FindTentBySlot('TentLocation1', BuilidngType.SoliderTent).state.level == 0) {
                    this.mechaBox.model.active = false;
                }
            });
        }
        createHangMu() {
            MapCampManagerComponent.inst.SpwanHangMu().then(node => {
                this.hangMu = new HangMuBuilidng(node);
            });
        }
        udpateTrigger(deltaTime, ppos) {
            this.mortars.forEach(m => {
                m.UpdateTrigger(deltaTime, ppos);
            });
            if (!this.entiytManager.player.mount) {
                if (this.mechaBox) {
                    this.mechaBox.Update(deltaTime, ppos);
                }
            }
        }
        initMortarsCD() {
            for (var i = 0; i < this.mortars.length; i++) {
                this.mortars[i].initFireCD(i * 0.5 + 0.4);
            }
        }
        updateMortarsFire(deltaTime) {
            this.mortars.forEach(m => {
                m.udpateFire(deltaTime);
            });
        }
        FireBomb(p) {
            this.nuke.FireBomb(p);
        }
        FireShip(p) {
            this.ship.FireBomb(p);
        }
        FireHangMu(line) {
            if (this.hangMu == null)
                return;
            this.hangMu.Fire(line);
        }
        KillEnemy(center, radius, damage) {
            this.entiytManager.RefreshCampGroup();
            var group = this.entiytManager.GetEnemyGroup(CampType.Blue);
            var sq = radius * radius;
            group.forEach(e => {
                if (LayaUtils.distanceSquaredVec2(e.Pos, center) < sq) {
                    if (damage == null) {
                        e.Death();
                    }
                    else {
                        var di = new DamageInfo();
                        di.damage = damage;
                        e.OnHit(di);
                    }
                }
            });
        }
        HideOtherBuildingExceptFirtSoilderTent() {
            if (this.soliderTentManager.FindTentBySlot('TentLocation1', BuilidngType.SoliderTent).state.level > 0) {
                return;
            }
            MsgManager.getInst().on(BattleEvent.UnlockFirstSoilderTent, this, this.onFirstSoliderTentUnlocked);
            this.storage.node.active = false;
            this.exchange.node.active = false;
            this.mortars.forEach(m => {
                m.node.active = false;
            });
            this.soliderTentManager.tents.forEach(t => {
                if (t.slotName != 'TentLocation1') {
                    t.node.active = false;
                }
            });
            this.weaponfacory.node.active = false;
            this.nuke.hideWaitTent = true;
            this.nuke.node.active = false;
            this.airport.node.active = false;
            if (this.mechaBox) {
                this.mechaBox.model.active = false;
            }
        }
        onRankUP() {
            ReportMgr.inst.sendReport('升级 ' + PlayerData.CurrentRankLevel + '_' + PlayerData.CurrentLevelScene, false);
            if (PlayerData.CurrentRankLevel == PlayerData.CurrentLevel.max_rank) {
                this.exchange.Unlock();
            }
            this.soliderTentManager.UpdateTents();
            this.storage.UpdateLock();
            this.mortars.forEach(m => {
                if (m.show_locked) {
                    m.node.active = true;
                }
                m.UpdateLock();
            });
            this.nuke.updateShow();
            this.airport.udpateState();
            this.airport.udpateShow();
            if (GuideData.Step == GuideStepType.升级提示) {
                GuideData.Step = GuideStepType.开始战斗;
            }
        }
        onFirstSoliderTentUnlocked() {
            if (this.storage.unlocked) {
                this.storage.node.active = true;
            }
            if (PlayerData.CurrentRankLevel >= PlayerData.CurrentLevel.max_rank) {
                this.exchange.Unlock(false);
            }
            if (PlayerData.CurrentRankLevel >= PlayerData.CurrentLevel.max_rank) {
                this.exchange.Unlock(false);
            }
            this.mortars.forEach(m => {
                if (m.show_locked) {
                    m.node.active = true;
                }
            });
            this.soliderTentManager.tents.forEach(t => {
                if (t.slotName != 'TentLocation1') {
                    t.node.active = true;
                }
            });
            this.weaponfacory.node.active = true;
            this.nuke.hideWaitTent = false;
            this.nuke.updateShow();
            this.airport.udpateShow();
            this.nuke.updateShow();
            if (PlayerData.CurrentLevelID == 2 && ADData.Bomber == 0) {
                var node = MapCampManagerComponent.inst.GetSlotNode('WarPlane');
                if (node == null)
                    return;
                var p = node.transform.position;
                p.y = 0.1;
                var s = new Laya.Sprite3D();
                MapComponent.inst.sp.addChild(s);
                s.transform.position = p;
                ArrowTips.inst.target = s;
                ArrowTips.inst.Show();
            }
            if (this.mechaBox != null) {
                this.mechaBox.model.active = true;
                if (PlayerData.CurrentLevelID == 2) {
                    this.mechaBox.com.DropBox();
                }
            }
        }
        onSpawnHelicopter(p) {
            if (p == null)
                return;
            this.entiytManager.GetEnemyGroup(CampType.Blue);
            var pos = new Laya.Vector2(p.x, p.z);
            MapCampManagerComponent.inst.SpawnHelicopter(pos);
        }
        onCreateDogTag(pos, isEnemy) {
            var entity = new DogTagEntity(pos, isEnemy);
            this.dogTagManager.AddDogTag(entity, false);
        }
        onCallDropBox() {
            MapCampManagerComponent.inst.SpwanDropBox();
        }
        onCallBomber() {
            this.airport.StartEngine();
        }
        DropSilverBox() {
            var enemies = this.entiytManager.GetEnemyGroup(CampType.Blue);
            if (enemies == null || enemies.length == 0) {
                return;
            }
            var e = Utils.rand(enemies);
            MapCampManagerComponent.inst.SpawnDropSilverBox(e.Pos);
        }
        onCreatePlayer() {
            this.entiytManager.player;
            var pos = MapCampManagerComponent.inst.playerLocation.transform.position.clone();
            pos.z -= 3.5;
            this.entiytManager.AddPlayerEntity(PlayerData.CurrentRankLevel, pos);
            console.log('creat player');
            GuideData.Step = GuideData.Step;
            if (PlayerData.mountsId != 0) {
                this.mountHorse = true;
                if (PlayerData.mountsId == 1 || PlayerData.mountsId == 2) {
                    MsgManager.getInst().event(BattleEvent.MountHorse);
                }
                else if (PlayerData.mountsId == 3 || PlayerData.mountsId == 4) {
                    MsgManager.getInst().event(BattleEvent.MountMecha, null, false, true);
                }
            }
            else {
                this.mountHorse = false;
            }
        }
        onSpwanBoxTrigger(t) {
            this.triggerManager.addBoxTrigger(t);
        }
        onSpwanFreeNuke(t) {
            this.triggerManager.freeNuke = t;
        }
        onMountJijia(jijia, free = false, isInit = false) {
            this.mountHorse = true;
            if (jijia == null) {
                PrefabManager.getInst().LoadPrefabByName(_AllPrefabsNames.Jijia).then(p => {
                    jijia = Laya.Sprite3D.instantiate(p);
                    this.entiytManager.player.fsm.SetData('mount', jijia);
                    this.entiytManager.player.fsm.ChangeState(PlayerCtrlMountMechaState, free, isInit);
                });
            }
            else {
                this.entiytManager.player.fsm.SetData('mount', jijia);
                this.entiytManager.player.fsm.ChangeState(PlayerCtrlMountMechaState, free, isInit);
            }
        }
        onMountHorse() {
            this.mountHorse = true;
            this.entiytManager.player.fsm.ChangeState(PlayerCtrlMountHorseState);
        }
        onInitUserTag() {
            this.entiytManager.player.aniView.initTag();
        }
    }

    class UI_ui_dituy extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("ditu", "ui_dituy"));
        }
        onConstruct() {
            this.m_bg = (this.getChildAt(0));
            this.m_comp_map = (this.getChildAt(1));
        }
    }
    UI_ui_dituy.URL = "ui://iuqvtd3filsk6";

    class UI_Comp_zhuangtai extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("ditu", "Comp_zhuangtai"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
        }
    }
    UI_Comp_zhuangtai.URL = "ui://iuqvtd3filsk7";

    class UI_comp_map1 extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("ditu", "comp_map1"));
        }
        onConstruct() {
            this.m_item7 = (this.getChildAt(1));
            this.m_item6 = (this.getChildAt(2));
            this.m_item5 = (this.getChildAt(3));
            this.m_item4 = (this.getChildAt(4));
            this.m_item3 = (this.getChildAt(5));
            this.m_item2 = (this.getChildAt(6));
            this.m_item1 = (this.getChildAt(7));
            this.m_group_item = (this.getChildAt(8));
        }
    }
    UI_comp_map1.URL = "ui://iuqvtd3fqdlo8";

    class UI_comp_map2 extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("ditu", "comp_map2"));
        }
        onConstruct() {
            this.m_item6 = (this.getChildAt(1));
            this.m_item5 = (this.getChildAt(2));
            this.m_item4 = (this.getChildAt(3));
            this.m_item3 = (this.getChildAt(4));
            this.m_item2 = (this.getChildAt(5));
            this.m_item1 = (this.getChildAt(6));
            this.m_group_item = (this.getChildAt(7));
        }
    }
    UI_comp_map2.URL = "ui://iuqvtd3fqdlo9";

    class UI_comp_map extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("ditu", "comp_map"));
        }
        onConstruct() {
            this.m_next2 = (this.getChildAt(0));
            this.m_next1 = (this.getChildAt(1));
            this.m_cur = (this.getChildAt(2));
            this.m_last = (this.getChildAt(3));
            this.m_group_map = (this.getChildAt(4));
        }
    }
    UI_comp_map.URL = "ui://iuqvtd3fqdlod";

    class dituBinder {
        static bindAll() {
            fgui.UIObjectFactory.setExtension(UI_ui_dituy.URL, UI_ui_dituy);
            fgui.UIObjectFactory.setExtension(UI_Comp_zhuangtai.URL, UI_Comp_zhuangtai);
            fgui.UIObjectFactory.setExtension(UI_comp_map1.URL, UI_comp_map1);
            fgui.UIObjectFactory.setExtension(UI_comp_map2.URL, UI_comp_map2);
            fgui.UIObjectFactory.setExtension(UI_comp_map.URL, UI_comp_map);
        }
    }

    let DituUI = class DituUI extends BaseUI {
        constructor() {
            super(...arguments);
            this._itemArr = [7, 6, 7, 6];
            this._total = 26;
            this._noMoveIndex = 2;
            this._returnIndex = 15;
        }
        get URL() {
            dituBinder.bindAll();
            return UI_ui_dituy.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_bg];
        }
        get VerticalAlign() {
            return LoaderVerticalAlign.bottom;
        }
        OnCreate() {
            this._itemComp = [this.ui.m_comp_map.m_last, this.ui.m_comp_map.m_cur, this.ui.m_comp_map.m_next1, this.ui.m_comp_map.m_next2];
        }
        OnShow() {
            this._refreshView();
            this._tweenStart().then(() => {
                UIManager.inst.PopUI(this.layer, false);
                if (GuideData.Step == GuideStepType.第一关完成) {
                    GuideData.Step = GuideStepType.第二关收集;
                }
            });
        }
        OnHide() {
        }
        OnDispose() {
        }
        _getFlag(idx) {
            for (let i = 0; i < this._itemArr.length; ++i) {
                if (idx < this._itemArr[i]) {
                    return this._itemComp[i]["m_item" + (idx + 1)];
                }
                idx -= this._itemArr[i];
            }
            return null;
        }
        get _startIdx() {
            return LocalCache$1.get("LogicData._mapStartIndex", 0);
        }
        set _startIdx(val) {
            LocalCache$1.set("LogicData._mapStartIndex", val);
        }
        _refreshView() {
            let flag;
            let idx = this._startIdx;
            for (let i = 0; i < this._total; ++i) {
                flag = this._getFlag(i);
                if (i < idx) {
                    flag.getControllerAt(0).selectedIndex = 0;
                }
                else if (i == idx) {
                    this.ui.m_comp_map.y = this.ui.height - this.ui.m_comp_map.height + (idx > this._noMoveIndex ? this._getBehand(this._noMoveIndex, i) : 0);
                    flag.getControllerAt(0).selectedIndex = 1;
                }
                else {
                    flag.getControllerAt(0).selectedIndex = 3;
                }
            }
        }
        _getBehand(start, end) {
            let cur = this._getFlag(start);
            let next = this._getFlag(end);
            let val = (cur.parent.y + cur.y) - (next.parent.y + next.y);
            return val;
        }
        _tweenStart() {
            return new Promise((resolve, reject) => {
                let func = () => {
                    let idx = this._startIdx;
                    if (idx < this._noMoveIndex) {
                        idx += 1;
                        if (idx >= this._returnIndex) {
                            idx = this._noMoveIndex;
                        }
                        this._startIdx = idx;
                        this._refreshView();
                        Laya.timer.once(1000, this, () => {
                            resolve();
                        });
                    }
                    else {
                        Laya.Tween.to(this.ui.m_comp_map, { y: this.ui.m_comp_map.y + this._getBehand(this._startIdx, this._startIdx + 1) }, 1000, Laya.Ease.backOut, Laya.Handler.create(this, () => {
                            idx += 1;
                            if (idx >= this._returnIndex) {
                                idx = this._noMoveIndex;
                            }
                            this._startIdx = idx;
                            this._refreshView();
                            Laya.timer.once(800, this, () => {
                                resolve();
                            });
                        }));
                    }
                };
                Laya.timer.once(500, this, func);
            });
        }
    };
    DituUI = __decorate([
        ui_register()
    ], DituUI);

    class UI_ui_jiesuan extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("jiesuan", "ui_jiesuan"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m_c2 = this.getControllerAt(1);
            this.m_new = this.getControllerAt(2);
            this.m_comp_mengban = (this.getChildAt(0));
            this.m_btn_close1 = (this.getChildAt(1));
            this.m_btn_ad = (this.getChildAt(2));
            this.m_group_guanggaoanniu = (this.getChildAt(3));
            this.m_btn_close12 = (this.getChildAt(4));
            this.m_btn_ad2 = (this.getChildAt(5));
            this.m_group_guanggaoanniu2 = (this.getChildAt(6));
            this.m_btn_close2 = (this.getChildAt(7));
            this.m_btn_close3 = (this.getChildAt(8));
            this.m_comp_chaopiao2 = (this.getChildAt(10));
            this.m_comp_chaopiao1 = (this.getChildAt(11));
            this.m_comp_tishi = (this.getChildAt(13));
            this.m_comp_xianshi = (this.getChildAt(14));
            this.m_comp_minzi = (this.getChildAt(15));
        }
    }
    UI_ui_jiesuan.URL = "ui://cefo0910tiqfd";

    class UI_Comp_xianshi extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("jiesuan", "Comp_xianshi"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
        }
    }
    UI_Comp_xianshi.URL = "ui://cefo0910vpo9g";

    class UI_Comp_minzi extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("jiesuan", "Comp_minzi"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
        }
    }
    UI_Comp_minzi.URL = "ui://cefo0910vpo9h";

    class jiesuanBinder {
        static bindAll() {
            fgui.UIObjectFactory.setExtension(UI_ui_jiesuan.URL, UI_ui_jiesuan);
            fgui.UIObjectFactory.setExtension(UI_Comp_xianshi.URL, UI_Comp_xianshi);
            fgui.UIObjectFactory.setExtension(UI_Comp_minzi.URL, UI_Comp_minzi);
        }
    }

    class UI_btn_xuanyao extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("share", "btn_xuanyao"));
        }
        onConstruct() {
            this.m_img_anniutubiao = (this.getChildAt(0));
        }
    }
    UI_btn_xuanyao.URL = "ui://9gh5nhg3figx8t";

    class UI_comp_zhanshi extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("share", "comp_zhanshi"));
        }
        onConstruct() {
            this.m_loader_img = (this.getChildAt(0));
            this.m_t0 = this.getTransitionAt(0);
        }
    }
    UI_comp_zhanshi.URL = "ui://9gh5nhg3n8glf";

    class UI_ui_share extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("share", "ui_share"));
        }
        onConstruct() {
            this.m_comp_mengban = (this.getChildAt(0));
            this.m_comp_share = (this.getChildAt(1));
            this.m_btn_share = (this.getChildAt(2));
            this.m_vedio = (this.getChildAt(3));
            this.m_btn_fanhui = (this.getChildAt(4));
            this.m_openPlay = this.getTransitionAt(0);
            this.m_t1 = this.getTransitionAt(1);
        }
    }
    UI_ui_share.URL = "ui://9gh5nhg3nzi80";

    class shareBinder {
        static bindAll() {
            fgui.UIObjectFactory.setExtension(UI_btn_xuanyao.URL, UI_btn_xuanyao);
            fgui.UIObjectFactory.setExtension(UI_comp_zhanshi.URL, UI_comp_zhanshi);
            fgui.UIObjectFactory.setExtension(UI_ui_share.URL, UI_ui_share);
        }
    }

    let ShareUI = class ShareUI extends BaseUI {
        get URL() {
            shareBinder.bindAll();
            return UI_ui_share.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_comp_mengban];
        }
        OnCreate() {
            var p = this.ui.localToGlobal(this.ui.m_btn_fanhui.x, this.ui.m_btn_fanhui.y);
            p.x = 0;
            p = this.ui.globalToLocal(p.x, p.y, p);
            this.ui.m_btn_fanhui.x = p.x;
            this.ui.m_btn_fanhui.onClick(this, () => {
                UIManager.inst.PopUI(this.layer, false);
            });
            this.ui.m_btn_share.onClick(this, this.onClickShare);
            this.ui.m_vedio.onClick(this, this.onClickShare);
        }
        OnShow(...args) {
            var cam = LevelManager.getInst().mainCamera;
            var g = this.ui.m_comp_share.m_loader_img;
            cam.renderTarget = new Laya.RenderTexture(g.width, g.height);
            var rtex = new Laya.Texture(cam.renderTarget, Laya.Texture.DEF_UV);
            g.displayObject.graphics.clear();
            g.displayObject.graphics.drawTexture(rtex);
            Utils.WaitFrame(2).then(() => {
                cam.renderTarget = null;
            });
        }
        OnHide() {
        }
        OnDispose() {
        }
        onClickShare() {
            GameSDK.shareVideoRecord(b => {
                if (b) {
                    UIManager.inst.PopUI(this.layer, false);
                }
            });
        }
    };
    ShareUI = __decorate([
        ui_register()
    ], ShareUI);

    let SettleUI = class SettleUI extends BaseUI {
        get URL() {
            jiesuanBinder.bindAll();
            return UI_ui_jiesuan.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_comp_mengban];
        }
        OnCreate() {
            this.ui.m_btn_close1.onClick(this, () => {
                UIManager.inst.PopUI(this.layer, false);
                MsgManager.getInst().event(BattleEvent.GoToNextScene);
            });
        }
        OnShow(...args) {
            GameSDK$1.isRight = !GameSDK$1.isRight;
            let closePos = GameSDK$1.isRight ? 1 : 0;
            let data = { closePos: closePos, panelPos: 1 };
            console.log("结算加载广告");
            GameSDK$1.GameNativeInsert && GameSDK$1.GameNativeInsert.load(data, true);
            console.log("结算加载广告结束");
            AlertMgr.inst.resetRebornTime();
            this.ui.m_comp_chaopiao2.title = BattleData.TagAllCollect + '';
            this.ui.m_comp_chaopiao1.title = BattleData.EnemyTagAllCollect + '';
            if (Laya.Browser.onTTMiniGame || Laya.Browser.onPC) {
                UIManager.inst.PushUI(ShareUI, UILayer.top);
            }
            switch (PlayerData.CurrentLevelID) {
                case 1:
                    this.ui.m_new.selectedIndex = 0;
                    break;
                case 2:
                    this.ui.m_new.selectedIndex = 1;
                    break;
                default:
                    this.ui.m_new.selectedIndex = 3;
                    break;
            }
        }
        OnHide() {
            GameSDK$1.GameNativeInsert && GameSDK$1.GameNativeInsert.setIsShow(false);
            AdMgr.inst.showLeftAd(false);
            AdMgr.inst.showRightAd(false);
            AdMgr.inst.showBottomCustom(false);
        }
        OnDispose() {
        }
    };
    SettleUI = __decorate([
        ui_register()
    ], SettleUI);

    class UI_ui_loading extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Loading", "ui_loading"));
        }
        onConstruct() {
            this.m_graph_bg = (this.getChildAt(0));
            this.m_img_jiazaizhong = (this.getChildAt(1));
            this.m_progress_ratio2 = (this.getChildAt(2));
            this.m_progress_ratio = (this.getChildAt(3));
            this.m_bottom = (this.getChildAt(8));
            this.m_img_logo = (this.getChildAt(9));
            this.m_t1 = this.getTransitionAt(0);
        }
    }
    UI_ui_loading.URL = "ui://ibgwcr5zfof24";

    let LoadingUI = class LoadingUI extends BaseUI {
        get URL() {
            return UI_ui_loading.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_graph_bg];
        }
        set subProgressValue(v) {
            if (v == 0) {
                this.ui.m_progress_ratio.value = v * 100;
            }
            else {
                this.ui.m_progress_ratio.tweenValue(v * 100, 0.1);
            }
        }
        set mainProgressValue(v) {
            if (v == 0) {
                this.ui.m_progress_ratio2.value = v * 100;
            }
            else {
                this.ui.m_progress_ratio2.tweenValue(v * 100, 0.1);
            }
        }
        OnCreate() {
            this.ui.m_progress_ratio.max = 100;
            this.ui.m_progress_ratio2.max = 100;
            var rect = this.ui.localToGlobalRect(0, this.ui.m_bottom.y, this.ui.width, this.ui.height);
            var p = this.ui.globalToLocal(0, Laya.stage.height - this.ui.m_bottom.height * rect.height / this.ui.height);
            this.ui.m_bottom.y = p.y;
        }
        OnShow(...args) {
        }
        OnHide() {
        }
        OnDispose() {
            MsgManager.getInst().offAllCaller(this);
        }
        onClose() {
            UIManager.inst.PopUI(this.layer, true);
        }
    };
    LoadingUI = __decorate([
        ui_register()
    ], LoadingUI);

    class BattlePreparetionState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'BattlePreparetionState';
            this.lineCd = 2;
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            MsgManager.getInst().event(BattleEvent.ResetFootSoliderSlot);
            MsgManager.getInst().on(GameProcedureEvent.StartBattle, this, this.onStartBattle);
            MsgManager.getInst().on(BattleEvent.SpwanSoilderTent, this, this.onAddSoliderTent);
            MsgManager.getInst().on(BattleEvent.RankLevelUp, this, this.onUpdateSoliderTent);
            fsm.Owner.entiytManager.player.steerBehaviorUnit.mass += 2000;
        }
        OnUpdate(fsm, elapseSeconds) {
            var battleManager = fsm.Owner;
            fsm.Owner.entiytManager.player.steerBehaviorUnit.maxSpeed = PlayerData.CurrentCommanderConfig.speed + 4;
            fsm.Owner.entiytManager.player.steerBehaviorUnit.mass = 1000;
            battleManager.soliderTentManager.UpdateSpwan(elapseSeconds);
            battleManager.entiytManager.RefreshCampGroup();
            battleManager.damageManager.Update(elapseSeconds);
            battleManager.entiytManager.Update(elapseSeconds);
            battleManager.triggerManager.Update(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
            battleManager.weaponfacory.UpdateSpwanWeapon(elapseSeconds);
            if (fsm.Owner.entiytManager.player.fieldY == 0 && !fsm.Owner.entiytManager.player.isDeath) {
                battleManager.dogTagManager.Udpate(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                battleManager.triggerManager.UpdateBattleTrigger(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                battleManager.nuke.UpdateTrigger(elapseSeconds, battleManager.entiytManager.player.Pos);
                if (!fsm.Owner.entiytManager.player.IsMove) {
                    battleManager.soliderTentManager.UpdateTrigger(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                    battleManager.weaponfacory.UpdateTrigger(elapseSeconds, battleManager.entiytManager.player.Pos);
                    fsm.Owner.exchange.UpdateTrigger(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                    fsm.Owner.storage.UpdateTrigger(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                    fsm.Owner.airport.UpdateTrigger(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                    fsm.Owner.udpateTrigger(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                }
            }
            SteeringBehaviorManager.getInst().updateUnits(elapseSeconds);
            var currentEnemyCount = fsm.Owner.entiytManager.GetEnemyGroup(CampType.Blue).length;
            MsgManager.getInst().event(BattleEvent.EnemyCountInfo, currentEnemyCount, MapComponent.inst.allEnemyCount);
            if (battleManager.enemyLoaded) {
                if (battleManager.entiytManager.GetEnemyGroup(CampType.Blue).length == 0) {
                    fsm.ChangeState(BattleWinState);
                }
            }
            this.lineCd -= elapseSeconds;
            if (this.lineCd <= 0) {
                this.lineCd += 5;
                if (fsm.Owner.enemyLoaded) {
                    fsm.Owner.createLine();
                }
            }
        }
        OnLeave(fsm) {
            MsgManager.getInst().offAllCaller(this);
            fsm.Owner.currentSoliderSlotCount = 0;
            fsm.Owner.currentTankSlotCount = 0;
            MsgManager.getInst().event(BattleEvent.ResetFootSoliderSlot);
            MsgManager.getInst().event(BattleEvent.TurnOffBattle);
            fsm.Owner.entiytManager.player.steerBehaviorUnit.mass -= 2000;
        }
        OnDestroy(fsm) {
        }
        onStartBattle() {
            this.fsm.ChangeState(BattleState);
        }
        SpwanPlayer() {
        }
        onAddSoliderTent(slotName, config, state) {
            this.fsm.Owner.soliderTentManager.AddTent(slotName, config, state);
        }
        onUpdateSoliderTent() {
            var config = ConfigHelper.SoliderConfigMap[PlayerData.CurrentCommanderConfig.solider_id];
            MsgManager.getInst().event(BattleEvent.ChangeSoliderSkin, this.fsm.Owner.entiytManager.player, config.prefab, PlayerData.CurrentCommanderConfig.material);
            this.fsm.Owner.soliderTentManager.UpdateTents();
            this.fsm.Owner.weaponfacory.udpateState();
        }
    }

    class BattleInitCampState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'BattleInitCampState';
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            if (fsm.Owner.entiytManager.player == null) {
                fsm.Owner.onCreatePlayer();
            }
            this.createPlayerLacation();
            this.createSoliderTent();
            this.createTankTent();
            this.createInitTag();
            this.createBattleTrigger();
            this.createWeaponFactory();
            this.createNuke();
            this.createShip();
            this.createExchange();
            this.createStorageBox();
            this.createMortars();
            this.createAirport();
            this.createMechaBox();
            this.createHangMu();
            fsm.Owner.HideOtherBuildingExceptFirtSoilderTent();
            GuideData.Step = GuideData.Step;
        }
        OnUpdate(fsm, elapseSeconds) {
            fsm.ChangeState(BattlePreparetionState);
        }
        OnLeave(fsm) {
        }
        OnDestroy(fsm) {
        }
        createSoliderTent() {
            var tents = BattleData.SoliderTents;
            for (var i = 0; i < tents.length; i++) {
                var tent = tents[i];
                this.fsm.Owner.soliderTentManager.AddTent(tent.slotName, ConfigHelper.BuildingMap[tent.id], tent, true);
            }
        }
        createTankTent() {
            var tent = BattleData.TankTent;
            this.fsm.Owner.soliderTentManager.AddTent(tent.slotName, ConfigHelper.BuildingMap[tent.id], tent, true);
        }
        updateSoliderTent() {
        }
        createInitTag() {
            var node = MapComponent.inst.Find('TagLocation');
            var pos = node.transform.position;
            BattleData.InitSilverTagInBegin(pos.x, pos.z, 10);
            BattleData.InitGoldTagInBegin(pos.x - 1, pos.z, 10);
            PlayerData.NewScene = false;
            for (var i = 0; i < BattleData.silverTagInGroud.length; i++) {
                var tag = new DogTagEntity(BattleData.silverTagInGroud[i], false);
                this.fsm.Owner.dogTagManager.AddDogTag(tag, false);
            }
            for (var i = 0; i < BattleData.goldTagInGroud.length; i++) {
                var tag = new DogTagEntity(BattleData.goldTagInGroud[i], true);
                this.fsm.Owner.dogTagManager.AddDogTag(tag, false);
            }
        }
        createPlayerLacation() {
            MapComponent.inst.CreatePlayerNode();
            var node = MapComponent.inst.Find('PlayerLocation');
            this.fsm.Owner.triggerManager.AddPlayer(node);
        }
        createBattleTrigger() {
            this.fsm.Owner.triggerManager.AddBattleTrigger(MapComponent.inst.Find('ChargeArea'));
        }
        createWeaponFactory() {
            this.fsm.Owner.createWeaponFactory(false);
        }
        createNuke() {
            this.fsm.Owner.createNuke();
        }
        createShip() {
            this.fsm.Owner.createShip();
        }
        createExchange() {
            this.fsm.Owner.createExchange();
        }
        createStorageBox() {
            this.fsm.Owner.createStorageBox();
        }
        createMortars() {
            this.fsm.Owner.createMortars();
        }
        createAirport() {
            this.fsm.Owner.createAirport();
        }
        createMechaBox() {
            if (PlayerData.CurrentLevelID >= GameManager.MechBoxUnlockScene) {
                this.fsm.Owner.createMechaBox();
            }
        }
        createHangMu() {
            if (PlayerData.CurrentLevelID >= GameManager.HangMuUnlockScene) {
                this.fsm.Owner.createHangMu();
            }
        }
    }

    class BattleSceneLoadState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'BattleSceneLoadState';
            this.loadTag = false;
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            this.ui = UIManager.inst.PushUI(LoadingUI, UILayer.base);
            this.loadTag = false;
            this.ui.subProgressValue = 0;
            LevelManager.getInst().LoadLevel(PlayerData.CurrentLevelScene).progress(p => {
                this.ui.subProgressValue = p;
                this.ui.mainProgressValue = 1;
            }).then(() => {
                this.loadTag = true;
            });
        }
        OnUpdate(fsm, elapseSeconds) {
            if (this.loadTag) {
                fsm.ChangeState(BattleInitCampState);
            }
        }
        OnLeave(fsm) {
            if (PlayerData.mountsId == 1 || PlayerData.mountsId == 3) {
                PlayerData.mountsId = 0;
            }
            AlertMgr.inst.savePlayerLevel(PlayerData.CurrentLevelID, PlayerData.CurrentLevelScene);
            fsm.Owner.StartNewGame();
            UIManager.inst.PopUI(this.ui.layer, false);
            UIManager.inst.PushUI(MainUI, UILayer.base);
            AdMgr.inst.init();
        }
        OnDestroy(fsm) {
        }
    }

    class BattleWinState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'BattleWinState';
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            this.createEnemyFlag();
            var entities = this.fsm.Owner.entiytManager.GetSelfEntityAll();
            entities.forEach(e => {
                e.fsm.ChangeState(IdleState);
            });
            MsgManager.getInst().event(BattleEvent.BattleWin);
            MsgManager.getInst().on(BattleEvent.TakeEnemyFlag, this, this.onTakeEnemyFlag);
            MsgManager.getInst().on(BattleEvent.GoToNextScene, this, this.onGoToNextScene);
            MsgManager.getInst().on(BattleEvent.RankLevelUp, this, this.onUpdateSoliderTent);
            MsgManager.getInst().on(BattleEvent.SpwanSoilderTent, this, this.onAddSoliderTent);
            if (GuideData.Step < GuideStepType.夺取旗帜) {
                GuideData.Step = GuideStepType.夺取旗帜;
            }
            else {
                ArrowTips.inst.target = MapCampManagerComponent.inst.GetSlotNode('EnemyFlag');
                ArrowTips.inst.Show();
            }
            fsm.Owner.createLine();
            fsm.Owner.entiytManager.RefreshCampGroup();
        }
        OnUpdate(fsm, elapseSeconds) {
            SteeringBehaviorManager.getInst().updateUnits(elapseSeconds);
            fsm.Owner.entiytManager.Update(elapseSeconds);
            fsm.Owner.entiytManager.player.steerBehaviorUnit.maxSpeed = PlayerData.CurrentCommanderConfig.speed + 4;
            fsm.Owner.entiytManager.player.steerBehaviorUnit.mass = 1000;
            if (fsm.Owner.entiytManager.player.fieldY == 0 && !fsm.Owner.entiytManager.player.isDeath) {
                fsm.Owner.dogTagManager.Udpate(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                fsm.Owner.triggerManager.UpdateWinState(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                if (fsm.Owner.entiytManager.player.fieldY == 0) {
                    fsm.Owner.soliderTentManager.UpdateTrigger(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                    fsm.Owner.triggerManager.Update(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                    fsm.Owner.weaponfacory.UpdateTrigger(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                    fsm.Owner.exchange.UpdateTrigger(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                    fsm.Owner.storage.UpdateTrigger(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                }
            }
        }
        OnLeave(fsm) {
            fsm.Owner.ClearGame();
            PlayerData.GoToNextLevelScene();
            UIManager.inst.PushUI(DituUI, UILayer.lowMid);
        }
        OnDestroy(fsm) {
        }
        createEnemyFlag() {
            this.fsm.Owner.triggerManager.AddFlag(MapComponent.inst.Find('EnemyFlag'));
        }
        onTakeEnemyFlag() {
            console.log('夺旗回调');
            if (GuideData.Step == GuideStepType.夺取旗帜) {
                GuideData.Step = GuideStepType.第一关完成;
            }
            if (PlayerData.CurrentLevelID == 1) {
                AlertMgr.inst.setCloseCB(Laya.Handler.create(this, () => {
                    ADData.Bomber = 0;
                    UIManager.inst.PushUI(SettleUI, UILayer.base);
                    console.log('看视频成功');
                }), Laya.Handler.create(this, () => {
                    console.log('看视频失败');
                    ADData.Bomber = ADData.initCD;
                    UIManager.inst.PushUI(SettleUI, UILayer.base);
                }));
                console.log('显示轰炸机');
                try {
                    AlertMgr.inst.showHongzha();
                }
                catch (e) {
                    console.log('显示轰炸机error', JSON.stringify(e));
                }
            }
            else {
                console.log('显示结算界面');
                UIManager.inst.PushUI(SettleUI, UILayer.base);
            }
            GameSDK$1.stageOnEnd({ level: PlayerData.CurrentLevelScene });
        }
        onGoToNextScene() {
            this.fsm.ChangeState(BattleSceneLoadState);
        }
        onUpdateSoliderTent() {
            var config = ConfigHelper.SoliderConfigMap[PlayerData.CurrentCommanderConfig.solider_id];
            MsgManager.getInst().event(BattleEvent.ChangeSoliderSkin, this.fsm.Owner.entiytManager.player, config.prefab, PlayerData.CurrentCommanderConfig.material);
            this.fsm.Owner.soliderTentManager.UpdateTents();
        }
        onAddSoliderTent(slotName, config, state) {
            SoundManager.inst.playSound('TentBuild');
            this.fsm.Owner.soliderTentManager.AddTent(slotName, config, state);
        }
    }

    class BattleState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'BattleState';
            this.dropBoxCD = 15;
            this.feijiCD = 20;
            this.lineCd = 5;
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            fsm.Owner.isBattleState = true;
            var aimTarget = fsm.Owner.entiytManager.GetNearstEnemy();
            var entities = this.fsm.Owner.entiytManager.GetSelfEntityAll();
            entities.forEach(e => {
                if (e.isPlayer)
                    return;
                e.forceAimTarget = aimTarget;
                e.seekRadiusSq = 1000000;
                e.steerBehaviorUnit.radius = 0.7;
                e.fsm.ChangeState(WanderState);
            });
            MsgManager.getInst().on(BattleEvent.RankLevelUp, this, this.onUdpatePlayerSkin);
            MsgManager.getInst().on(BattleEvent.SpwanSoilderTent, this, this.onAddSoliderTent);
            SoundManager.inst.playSound('quanjunchuji');
            fsm.Owner.initMortarsCD();
        }
        OnUpdate(fsm, elapseSeconds) {
            fsm.Owner.entiytManager.player.steerBehaviorUnit.maxSpeed = PlayerData.CurrentCommanderConfig.speed + 3;
            fsm.Owner.entiytManager.player.steerBehaviorUnit.mass = 1000;
            fsm.Owner.entiytManager.RefreshCampGroup();
            fsm.Owner.damageManager.Update(elapseSeconds);
            fsm.Owner.entiytManager.Update(elapseSeconds);
            SteeringBehaviorManager.getInst().updateUnits(elapseSeconds);
            this.lineCd -= elapseSeconds;
            if (this.lineCd <= 0) {
                this.lineCd += 1;
                if (fsm.Owner.enemyLoaded) {
                    fsm.Owner.createLine();
                }
            }
            if (fsm.Owner.entiytManager.player.fieldY == 0 && !fsm.Owner.entiytManager.player.isDeath) {
                fsm.Owner.dogTagManager.Udpate(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                fsm.Owner.nuke.UpdateTrigger(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                if (fsm.Owner.entiytManager.player.IsMove == false) {
                    fsm.Owner.soliderTentManager.UpdateTrigger(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                    fsm.Owner.weaponfacory.UpdateSpwanWeapon(elapseSeconds);
                    fsm.Owner.weaponfacory.UpdateTrigger(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                    fsm.Owner.exchange.UpdateTrigger(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                    fsm.Owner.storage.UpdateTrigger(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
                }
                fsm.Owner.triggerManager.Update(elapseSeconds, fsm.Owner.entiytManager.player.Pos);
            }
            fsm.Owner.updateMortarsFire(elapseSeconds);
            if (GameManager.getInst().battleManager.triggerManager.boxes.length < 2) {
                this.dropBoxCD -= elapseSeconds;
                if (this.dropBoxCD <= 0) {
                    this.dropBoxCD += 45;
                    fsm.Owner.DropSilverBox();
                }
            }
            this.feijiCD -= elapseSeconds;
            if (this.feijiCD <= 0) {
                this.feijiCD += 45;
                MsgManager.getInst().event(BattleEvent.TriggerFeiJi);
            }
            var currentEnemyCount = fsm.Owner.entiytManager.GetEnemyGroup(CampType.Blue).length;
            MsgManager.getInst().event(BattleEvent.EnemyCountInfo, currentEnemyCount, MapComponent.inst.allEnemyCount);
            if (currentEnemyCount == 0) {
                fsm.ChangeState(BattleWinState);
            }
            else if (fsm.Owner.entiytManager.GetSelfEntityAll().length == 0) {
                fsm.ChangeState(BattlePreparetionState);
                AlertMgr.inst.addSelfEnemyCleanCount();
                if (AlertMgr.inst.showNuke()) {
                    PlayerData.UnlockNuke = true;
                    fsm.Owner.onRankUP();
                    AlertMgr.inst.setCloseCB(Laya.Handler.create(this, () => {
                        SoundManager.inst.playVibrate(true);
                        UIManager.inst.PushUI(AimMenu, UILayer.base, BattleEvent.FireBomb, GameManager.NukeRadius);
                        fsm.Owner.needYuyue = true;
                    }), Laya.Handler.create(this, () => {
                        fsm.Owner.spwanFreeNuke = true;
                        PlayerData.YuyueNuke = true;
                        Utils.WaitTime(2000).then(() => {
                            AlertMgr.inst.setCloseCB(Laya.Handler.create(this, () => {
                            }));
                        });
                    }));
                }
                console.log(fsm.Owner.spwanFreeNuke, PlayerData.CurrentLevelID, AlertMgr.inst.getSelfEnemyCleanCount());
                if (fsm.Owner.spwanFreeNuke && PlayerData.CurrentLevelID == 1 && AlertMgr.inst.getSelfEnemyCleanCount() == 2) {
                    MapCampManagerComponent.inst.SpwanFreeNuke();
                }
            }
        }
        OnLeave(fsm) {
            MsgManager.getInst().offAllCaller(this);
            fsm.Owner.isBattleState = false;
            MsgManager.getInst().event(GameProcedureEvent.StopBattle);
        }
        OnDestroy(fsm) {
        }
        onUdpatePlayerSkin() {
            var config = ConfigHelper.SoliderConfigMap[PlayerData.CurrentCommanderConfig.solider_id];
            MsgManager.getInst().event(BattleEvent.ChangeSoliderSkin, this.fsm.Owner.entiytManager.player, config.prefab, PlayerData.CurrentCommanderConfig.material);
            this.fsm.Owner.soliderTentManager.UpdateTents();
        }
        onAddSoliderTent(slotName, config, state) {
            SoundManager.inst.playSound('TentBuild');
            this.fsm.Owner.soliderTentManager.AddTent(slotName, config, state);
        }
    }

    class MainUIState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'MainUIState';
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            MsgManager.getInst().on(GameProcedureEvent.EnterBattleGame, this, this.onEnterBattleGame);
            SoundManager.inst.PlayBGM('Bgm', 0.5);
            GameSDK.initConfig(() => { });
        }
        OnUpdate(fsm, elapseSeconds) {
            this.onEnterBattleGame();
        }
        OnLeave(fsm) {
            MsgManager.getInst().off(GameProcedureEvent.EnterBattleGame, this, this.onEnterBattleGame);
        }
        OnDestroy(fsm) {
        }
        onEnterBattleGame() {
            this.fsm.ChangeState(BattleSceneLoadState);
            Laya.timer.once(1000, this, () => {
                if (+GameSDK.data.ShowMainInsert == 1) {
                    GameSDK.showADInsert();
                }
            });
        }
    }

    class DebugDataState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'DebugDataState';
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            fsm.ChangeState(MainUIState);
        }
        OnUpdate(fsm, elapseSeconds) {
        }
        OnLeave(fsm) {
        }
        OnDestroy(fsm) {
        }
    }

    class UI_Battle extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Battle", "Battle"));
        }
        onConstruct() {
            this.m_hpNode = (this.getChildAt(0));
            this.m_composePanel = (this.getChildAt(2));
            this.m_blueTowerHP = (this.getChildAt(3));
            this.m_redTowerHP = (this.getChildAt(4));
            this.m_spCD = (this.getChildAt(5));
            this.m_random = (this.getChildAt(6));
        }
    }
    UI_Battle.URL = "ui://dg65isdealea0";

    class UI_HeroItem extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Battle", "HeroItem"));
        }
        onConstruct() {
            this.m_dianshu = this.getControllerAt(0);
            this.m_rarity = this.getControllerAt(1);
            this.m_icon = (this.getChildAt(0));
        }
    }
    UI_HeroItem.URL = "ui://dg65isdeop9bgc";

    class UI_HpNode extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Battle", "HpNode"));
        }
        onConstruct() {
            this.m_bg = (this.getChildAt(0));
        }
    }
    UI_HpNode.URL = "ui://dg65isdeqxgd7";

    class UI_hpBarTower extends fgui.GProgressBar {
        static createInstance() {
            return (fgui.UIPackage.createObject("Battle", "hpBarTower"));
        }
        onConstruct() {
            this.m_camp = this.getControllerAt(0);
            this.m_bg = (this.getChildAt(0));
            this.m_barbg = (this.getChildAt(6));
        }
    }
    UI_hpBarTower.URL = "ui://dg65isdeqxgd8";

    class UI_spProgress extends fgui.GProgressBar {
        static createInstance() {
            return (fgui.UIPackage.createObject("Battle", "spProgress"));
        }
        onConstruct() {
            this.m_tag0 = (this.getChildAt(2));
            this.m_tag1 = (this.getChildAt(3));
            this.m_tag2 = (this.getChildAt(4));
            this.m_tag3 = (this.getChildAt(5));
            this.m_tag4 = (this.getChildAt(6));
        }
    }
    UI_spProgress.URL = "ui://dg65isdeqxgdb";

    class UI_spProgressTag extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Battle", "spProgressTag"));
        }
        onConstruct() {
            this.m_touch = this.getControllerAt(0);
        }
    }
    UI_spProgressTag.URL = "ui://dg65isdeqxgdc";

    class UI_hpBar extends fgui.GProgressBar {
        static createInstance() {
            return (fgui.UIPackage.createObject("Battle", "hpBar"));
        }
        onConstruct() {
            this.m_camp = this.getControllerAt(0);
            this.m_fade = this.getTransitionAt(0);
        }
    }
    UI_hpBar.URL = "ui://dg65isdeqxgdgb";

    class BattleBinder {
        static bindAll() {
            fgui.UIObjectFactory.setExtension(UI_Battle.URL, UI_Battle);
            fgui.UIObjectFactory.setExtension(UI_HeroItem.URL, UI_HeroItem);
            fgui.UIObjectFactory.setExtension(UI_HpNode.URL, UI_HpNode);
            fgui.UIObjectFactory.setExtension(UI_hpBarTower.URL, UI_hpBarTower);
            fgui.UIObjectFactory.setExtension(UI_spProgress.URL, UI_spProgress);
            fgui.UIObjectFactory.setExtension(UI_spProgressTag.URL, UI_spProgressTag);
            fgui.UIObjectFactory.setExtension(UI_hpBar.URL, UI_hpBar);
        }
    }

    class LoadingBinder {
        static bindAll() {
            fgui.UIObjectFactory.setExtension(UI_ui_loading.URL, UI_ui_loading);
            fgui.UIObjectFactory.setExtension(UI_Tips.URL, UI_Tips);
            fgui.UIObjectFactory.setExtension(UI_TipsItem.URL, UI_TipsItem);
        }
    }

    class UI_comp_tishi extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "comp_tishi"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m_img_diban = (this.getChildAt(0));
            this.m_img_wz1 = (this.getChildAt(1));
            this.m_img_wz2 = (this.getChildAt(2));
            this.m_img_wz3 = (this.getChildAt(3));
            this.m_img_wz4 = (this.getChildAt(4));
            this.m_img_wz5 = (this.getChildAt(5));
            this.m_img_wz6 = (this.getChildAt(6));
            this.m_img_wz7 = (this.getChildAt(7));
        }
    }
    UI_comp_tishi.URL = "ui://ixnra5xg110sxrx";

    class UI_HpBarHorse extends fgui.GProgressBar {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "HpBarHorse"));
        }
        onConstruct() {
            this.m_show = this.getControllerAt(0);
            this.m_fade = this.getTransitionAt(0);
        }
    }
    UI_HpBarHorse.URL = "ui://ixnra5xgeperxt6";

    class UI_btn_tongyong extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "btn_tongyong"));
        }
        onConstruct() {
            this.m_tips = this.getControllerAt(0);
            this.m_ad = this.getControllerAt(1);
            this.m_yongjiu = this.getControllerAt(2);
            this.m_cd = this.getControllerAt(3);
            this.m_proogress_zhanma = (this.getChildAt(1));
            this.m_adProgress = (this.getChildAt(7));
            this.m_cd_2 = (this.getChildAt(8));
        }
    }
    UI_btn_tongyong.URL = "ui://ixnra5xgik6mh7";

    class UI_comp_mengban extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "comp_mengban"));
        }
        onConstruct() {
            this.m_graph_mengban = (this.getChildAt(0));
        }
    }
    UI_comp_mengban.URL = "ui://ixnra5xgjycsxu8";

    class UI_comp_login extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "comp_login"));
        }
        onConstruct() {
            this.m_btn_login = (this.getChildAt(3));
            this.m_descripe = (this.getChildAt(4));
            this.m_btn_tuichu = (this.getChildAt(5));
        }
    }
    UI_comp_login.URL = "ui://ixnra5xgjycsxu9";

    class UI_JoyStick extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "JoyStick"));
        }
        onConstruct() {
            this.m_holder = (this.getChildAt(1));
        }
    }
    UI_JoyStick.URL = "ui://ixnra5xgljybxry";

    class UI_progressbar extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "progressbar"));
        }
        onConstruct() {
            this.m_bar = (this.getChildAt(0));
        }
    }
    UI_progressbar.URL = "ui://ixnra5xgljybxs0";

    class UI_Progress_zhanma extends fgui.GProgressBar {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "Progress_zhanma"));
        }
        onConstruct() {
            this.m_bar_ = (this.getChildAt(0));
        }
    }
    UI_Progress_zhanma.URL = "ui://ixnra5xglpe8xtp";

    class UI_HpBar extends fgui.GProgressBar {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "HpBar"));
        }
        onConstruct() {
            this.m_show = this.getControllerAt(0);
            this.m_fade = this.getTransitionAt(0);
        }
    }
    UI_HpBar.URL = "ui://ixnra5xgm049xsr";

    class UI_Native extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "Native"));
        }
        onConstruct() {
            this.m_content = (this.getChildAt(0));
        }
    }
    UI_Native.URL = "ui://ixnra5xgn1ovxui";

    class UI_NativeAdBanner extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "NativeAdBanner"));
        }
        onConstruct() {
            this.m_icon = (this.getChildAt(1));
            this.m_describle = (this.getChildAt(4));
            this.m_logo = (this.getChildAt(6));
        }
    }
    UI_NativeAdBanner.URL = "ui://ixnra5xgn1ovxuj";

    class UI_ui_bottomvideo extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "ui_bottomvideo"));
        }
        onConstruct() {
            this.m_comp_video = (this.getChildAt(0));
        }
    }
    UI_ui_bottomvideo.URL = "ui://ixnra5xgqf3k5n";

    class UI_btn_zhanshu extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "btn_zhanshu"));
        }
        onConstruct() {
            this.m_tips = this.getControllerAt(1);
        }
    }
    UI_btn_zhanshu.URL = "ui://ixnra5xgqk3dxt5";

    class UI_btn_mrlq extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "btn_mrlq"));
        }
        onConstruct() {
            this.m_tips = this.getControllerAt(1);
        }
    }
    UI_btn_mrlq.URL = "ui://ixnra5xgql6xxte";

    class UI_Air extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "Air"));
        }
        onConstruct() {
            this.m_progress = (this.getChildAt(2));
            this.m_show = this.getTransitionAt(0);
        }
    }
    UI_Air.URL = "ui://ixnra5xgql6xxti";

    class UI_LevelUpTip extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "LevelUpTip"));
        }
        onConstruct() {
            this.m_aniView = (this.getChildAt(1));
        }
    }
    UI_LevelUpTip.URL = "ui://ixnra5xgu27xxst";

    class UI_LevelUp extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "LevelUp"));
        }
        onConstruct() {
            this.m__mask = (this.getChildAt(0));
            this.m_pre = (this.getChildAt(2));
            this.m_current = (this.getChildAt(3));
            this.m_show = this.getTransitionAt(0);
        }
    }
    UI_LevelUp.URL = "ui://ixnra5xgu27xxsw";

    class UI_RankAni extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "RankAni"));
        }
        onConstruct() {
            this.m_rank = (this.getChildAt(6));
            this.m_last = (this.getChildAt(8));
            this.m_current = (this.getChildAt(9));
            this.m_extSlot = (this.getChildAt(11));
        }
    }
    UI_RankAni.URL = "ui://ixnra5xgu27xxsy";

    class UI_ArrowFlow extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "ArrowFlow"));
        }
        onConstruct() {
            this.m_t1 = this.getTransitionAt(0);
        }
    }
    UI_ArrowFlow.URL = "ui://ixnra5xgu27xxt0";

    class UI_Comp_video extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "Comp_video"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m_t0 = this.getTransitionAt(0);
        }
    }
    UI_Comp_video.URL = "ui://ixnra5xgul0c5u";

    class UI_comp_zhiyin extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "comp_zhiyin"));
        }
        onConstruct() {
            this.m_t0 = this.getTransitionAt(0);
        }
    }
    UI_comp_zhiyin.URL = "ui://ixnra5xgulu1ri";

    class UI_btn_fight extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "btn_fight"));
        }
        onConstruct() {
            this.m_ani = (this.getChildAt(2));
            this.m_t0 = this.getTransitionAt(0);
            this.m_t1 = this.getTransitionAt(1);
        }
    }
    UI_btn_fight.URL = "ui://ixnra5xgulu1rj";

    class UI_btn_jineng extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "btn_jineng"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(1);
            this.m_c2 = this.getControllerAt(2);
            this.m_loader_tubiao = (this.getChildAt(1));
            this.m_loader_mingzi = (this.getChildAt(4));
            this.m_cd = (this.getChildAt(6));
        }
    }
    UI_btn_jineng.URL = "ui://ixnra5xgulu1rk";

    class UI_comp_zidan extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "comp_zidan"));
        }
        onConstruct() {
            this.m_text_shuliang = (this.getChildAt(2));
        }
    }
    UI_comp_zidan.URL = "ui://ixnra5xgulu1rl";

    class UI_comp_qizi extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "comp_qizi"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
        }
    }
    UI_comp_qizi.URL = "ui://ixnra5xgulu1rn";

    class UI_Progress_guanqiajindu extends fgui.GProgressBar {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "Progress_guanqiajindu"));
        }
        onConstruct() {
            this.m_levelCount = this.getControllerAt(0);
            this.m_comp_qizhi1 = (this.getChildAt(4));
            this.m_comp_qizhi2 = (this.getChildAt(5));
            this.m_comp_qizhi3 = (this.getChildAt(6));
            this.m_comp_qizhi4 = (this.getChildAt(7));
        }
    }
    UI_Progress_guanqiajindu.URL = "ui://ixnra5xgulu1ro";

    class UI_btn_jijia extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "btn_jijia"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(1);
            this.m_cd = (this.getChildAt(1));
        }
    }
    UI_btn_jijia.URL = "ui://ixnra5xgvpo9xtz";

    class UI_btn_zhanjain extends fgui.GButton {
        static createInstance() {
            return (fgui.UIPackage.createObject("Main", "btn_zhanjain"));
        }
        onConstruct() {
            this.m_c1 = this.getControllerAt(0);
            this.m_cd = this.getControllerAt(1);
            this.m_fly = this.getControllerAt(2);
            this.m_tips = this.getControllerAt(3);
            this.m_cd_2 = (this.getChildAt(3));
            this.m_ani = this.getTransitionAt(0);
        }
    }
    UI_btn_zhanjain.URL = "ui://ixnra5xgvpo9xu0";

    class MainBinder {
        static bindAll() {
            fgui.UIObjectFactory.setExtension(UI_comp_tishi.URL, UI_comp_tishi);
            fgui.UIObjectFactory.setExtension(UI_HpBarHorse.URL, UI_HpBarHorse);
            fgui.UIObjectFactory.setExtension(UI_ui_aim.URL, UI_ui_aim);
            fgui.UIObjectFactory.setExtension(UI_btn_tongyong.URL, UI_btn_tongyong);
            fgui.UIObjectFactory.setExtension(UI_ui_login.URL, UI_ui_login);
            fgui.UIObjectFactory.setExtension(UI_comp_mengban.URL, UI_comp_mengban);
            fgui.UIObjectFactory.setExtension(UI_comp_login.URL, UI_comp_login);
            fgui.UIObjectFactory.setExtension(UI_ui_aim_hm.URL, UI_ui_aim_hm);
            fgui.UIObjectFactory.setExtension(UI_FreeNuke.URL, UI_FreeNuke);
            fgui.UIObjectFactory.setExtension(UI_JoyStick.URL, UI_JoyStick);
            fgui.UIObjectFactory.setExtension(UI_progressbar.URL, UI_progressbar);
            fgui.UIObjectFactory.setExtension(UI_Progress_zhanma.URL, UI_Progress_zhanma);
            fgui.UIObjectFactory.setExtension(UI_HpBar.URL, UI_HpBar);
            fgui.UIObjectFactory.setExtension(UI_Native.URL, UI_Native);
            fgui.UIObjectFactory.setExtension(UI_NativeAdBanner.URL, UI_NativeAdBanner);
            fgui.UIObjectFactory.setExtension(UI_AttackUp.URL, UI_AttackUp);
            fgui.UIObjectFactory.setExtension(UI_ui_bottomvideo.URL, UI_ui_bottomvideo);
            fgui.UIObjectFactory.setExtension(UI_btn_zhanshu.URL, UI_btn_zhanshu);
            fgui.UIObjectFactory.setExtension(UI_btn_mrlq.URL, UI_btn_mrlq);
            fgui.UIObjectFactory.setExtension(UI_Air.URL, UI_Air);
            fgui.UIObjectFactory.setExtension(UI_LevelUpTip.URL, UI_LevelUpTip);
            fgui.UIObjectFactory.setExtension(UI_LevelUp.URL, UI_LevelUp);
            fgui.UIObjectFactory.setExtension(UI_RankAni.URL, UI_RankAni);
            fgui.UIObjectFactory.setExtension(UI_ArrowFlow.URL, UI_ArrowFlow);
            fgui.UIObjectFactory.setExtension(UI_Comp_video.URL, UI_Comp_video);
            fgui.UIObjectFactory.setExtension(UI_comp_zhiyin.URL, UI_comp_zhiyin);
            fgui.UIObjectFactory.setExtension(UI_btn_fight.URL, UI_btn_fight);
            fgui.UIObjectFactory.setExtension(UI_btn_jineng.URL, UI_btn_jineng);
            fgui.UIObjectFactory.setExtension(UI_comp_zidan.URL, UI_comp_zidan);
            fgui.UIObjectFactory.setExtension(UI_comp_qizi.URL, UI_comp_qizi);
            fgui.UIObjectFactory.setExtension(UI_Progress_guanqiajindu.URL, UI_Progress_guanqiajindu);
            fgui.UIObjectFactory.setExtension(UI_btn_jijia.URL, UI_btn_jijia);
            fgui.UIObjectFactory.setExtension(UI_btn_zhanjain.URL, UI_btn_zhanjain);
            fgui.UIObjectFactory.setExtension(UI_ui_main.URL, UI_ui_main);
        }
    }

    class PlatformData {
        constructor() {
            this.appId = '';
            this.appKey = '';
            this.bannerId = '';
            this.rewardVideoId = '';
            this.interstitialId = '';
            this.nativeId = '';
            this.nativeBannerIds = [];
            this.nativeIconIds = [];
            this.nativeinterstitialIds = [];
            this.nativeinpageIds = [];
            this.shareId = '';
        }
    }

    class BDData extends PlatformData {
        constructor() {
            super(...arguments);
            this.appId = '';
            this.appKey = '';
            this.bannerId = '';
            this.rewardVideoId = '';
            this.interstitialId = '';
            this.nativeId = '';
            this.nativeBannerIds = [];
            this.nativeIconIds = [];
            this.nativeinterstitialIds = [];
            this.nativeinpageIds = [];
            this.shareId = '';
        }
    }

    class OPPOData extends PlatformData {
        constructor() {
            super(...arguments);
            this.appId = '';
            this.appKey = '';
            this.bannerId = '';
            this.rewardVideoId = '';
            this.interstitialId = '';
            this.nativeId = '';
            this.nativeBannerIds = [];
            this.nativeIconIds = [];
            this.nativeinterstitialIds = [];
            this.nativeinpageIds = [];
            this.shareId = '';
        }
    }

    class QQData extends PlatformData {
        constructor() {
            super(...arguments);
            this.appId = '1112114764';
            this.appKey = 'Mh2Ay8TbDytZUDEo';
            this.bannerId = '5f680c6263264f3060412de3c3d6b3c9';
            this.rewardVideoId = '';
            this.interstitialId = '';
            this.nativeId = '';
            this.nativeBannerIds = [];
            this.nativeIconIds = [];
            this.nativeinterstitialIds = [];
            this.nativeinpageIds = [];
            this.shareId = '';
        }
    }

    class QTTData extends PlatformData {
        constructor() {
            super(...arguments);
            this.appId = '';
            this.appKey = '';
            this.bannerId = '';
            this.rewardVideoId = '';
            this.interstitialId = '';
            this.nativeId = '';
            this.nativeBannerIds = [];
            this.nativeIconIds = [];
            this.nativeinterstitialIds = [];
            this.nativeinpageIds = [];
            this.shareId = '';
        }
    }

    class TTData extends PlatformData {
        constructor() {
            super(...arguments);
            this.appId = 'tt11ac67118e95953002';
            this.appKey = '802d1e87d87276465424ffe732392767480efdc3';
            this.bannerId = '';
            this.rewardVideoId = '7gl632822ht7e7crqe';
            this.interstitialId = '1akai8l8978lh3l2ih';
            this.nativeId = '';
            this.nativeBannerIds = [];
            this.nativeIconIds = [];
            this.nativeinterstitialIds = [];
            this.nativeinpageIds = [];
            this.shareId = '1eqqj25f7gqf2p0cv4';
        }
    }

    class VIVOData extends PlatformData {
        constructor() {
            super(...arguments);
            this.appId = '105511615';
            this.appKey = '25f61e57709b2092bd382953c3593835';
            this.bannerId = '';
            this.rewardVideoId = '14d280a01a354bf7ae8c12b74069a87c';
            this.interstitialId = '';
            this.nativeId = '';
            this.nativeBannerIds = [];
            this.nativeIconIds = [];
            this.nativeinterstitialIds = [];
            this.nativeinpageIds = [];
            this.shareId = '';
        }
    }

    class WXData extends PlatformData {
        constructor() {
            super(...arguments);
            this.appId = 'wx835ee25d7f37dc6c';
            this.appKey = '';
            this.bannerId = '';
            this.rewardVideoId = 'adunit';
            this.interstitialId = '';
            this.nativeId = '';
            this.nativeBannerIds = [];
            this.nativeIconIds = [];
            this.nativeinterstitialIds = [];
            this.nativeinpageIds = [];
            this.shareId = '';
        }
    }

    var PlatformCommonEvent;
    (function (PlatformCommonEvent) {
        PlatformCommonEvent["PAUSE_AUDIO"] = "PAUSE_AUDIO";
        PlatformCommonEvent["RESUM_AUDIO"] = "RESUM_AUDIO";
        PlatformCommonEvent["AD_CONFIG_GETTED"] = "AD_CONFIG_GETTED";
        PlatformCommonEvent["SELF_AD_INITED"] = "SELF_AD_INITED";
    })(PlatformCommonEvent || (PlatformCommonEvent = {}));

    var EPlatformType;
    (function (EPlatformType) {
        EPlatformType["None"] = "EPlatformType_None";
        EPlatformType["Web"] = "EPlatformType_Web";
        EPlatformType["WX"] = "EPlatformType_WX";
        EPlatformType["TT"] = "EPlatformType_TT";
        EPlatformType["QQ"] = "EPlatformType_QQ";
        EPlatformType["VIVO"] = "EPlatformType_VIVO";
        EPlatformType["OPPO"] = "EPlatformType_OPPO";
        EPlatformType["BD"] = "EPlatformType_BD";
        EPlatformType["KG"] = "EPlatformType_KG";
        EPlatformType["Alipay"] = "EPlatformType_Alipay";
        EPlatformType["HW"] = "EPlatformType_HW";
        EPlatformType["QTT"] = "EPlatformType_QTT";
    })(EPlatformType || (EPlatformType = {}));

    class DefaultDevice {
        Vibrate(isLong) {
            console.log("调用震动", isLong);
            if (!navigator.vibrate) {
                console.log("不支持设备震动！");
                return;
            }
            if (isLong) {
                navigator.vibrate(400);
            }
            else {
                navigator.vibrate(15);
            }
        }
    }

    class WXDevice extends DefaultDevice {
        Vibrate(isLong) {
            console.log("调用震动", isLong);
            if (window['wx']) {
                if (isLong) {
                    Laya.timer.callLater(wx, wx.vibrateLong, [null]);
                }
                else {
                    Laya.timer.callLater(wx, wx.vibrateShort, [null]);
                }
            }
        }
    }

    class PlatformManagerProxy {
        static get instance() {
            if (this._instance == null) {
                this._instance = new PlatformManagerProxy();
            }
            return this._instance;
        }
        get PlatformInstance() {
            if (!this.m_platformInstance) {
                console.log('还没有设置过平台实例代理！');
            }
            return this.m_platformInstance;
        }
        set PlatformInstance(_instance) {
            this.m_platformInstance = _instance;
        }
        static get platformStr() {
            return PlatformManagerProxy.GetPlatformStr(this._instance.m_platformInstance.platform);
        }
        static GetPlatformStr(platformEnum) {
            switch (platformEnum) {
                case EPlatformType.None:
                    return "未识别";
                case EPlatformType.Web:
                    return "网页";
                case EPlatformType.BD:
                    return "百度";
                case EPlatformType.OPPO:
                    return "Oppo";
                case EPlatformType.QQ:
                    return "QQ";
                case EPlatformType.TT:
                    return "头条";
                case EPlatformType.VIVO:
                    return "Vivo";
                case EPlatformType.WX:
                    return "微信";
                case EPlatformType.QTT:
                    return "趣头条";
                default:
                    return "未定义" + platformEnum;
            }
        }
    }

    class DefaultRecordManager {
        constructor() {
            this.supportRecord = false;
            this.isRecording = false;
            this.isPausing = false;
            this.isRecordSuccess = false;
        }
        StartRecord(onStart, onOverTime) {
            console.log("该平台" + PlatformManagerProxy.platformStr + "不支持录制视频");
        }
        StopRecord(onStop) {
            console.log("该平台" + PlatformManagerProxy.platformStr + "不支持录制视频");
        }
        Pause(onPause) {
            console.log("该平台" + PlatformManagerProxy.platformStr + "不支持录制视频");
        }
        Resume(onReume) {
            console.log("该平台" + PlatformManagerProxy.platformStr + "不支持录制视频");
        }
        RecordClip(timeRange) {
            console.log("该平台" + PlatformManagerProxy.platformStr + "不支持录制视频");
        }
        ShareVideo(onSuccess, onCancel, onFailed) {
            if (onFailed) {
                onFailed.run();
            }
        }
    }

    class ShareInfo {
    }

    class ShareManagerAKhgfawgsda {
        constructor() {
            this._shareInfoList = [];
        }
        static get instance() {
            if (this._instance == null) {
                this._instance = new ShareManagerAKhgfawgsda();
            }
            return this._instance;
        }
        AddShareInfo(shareInfo) {
            for (let getInfo of this._shareInfoList) {
                if (getInfo.shareId == shareInfo.shareId)
                    return;
            }
            this._shareInfoList.push(shareInfo);
        }
        GetShareInfo(id = null) {
            if (this._shareInfoList.length == 0) {
                let fakeShareInfo = new ShareInfo();
                return fakeShareInfo;
            }
            if (id != null) {
                for (let shareInfo of this._shareInfoList) {
                    if (shareInfo.shareId == id)
                        return shareInfo;
                }
            }
            let randomShare = this._shareInfoList[Math.floor(this._shareInfoList.length * Math.random())];
            return randomShare;
        }
        ShareAppMessage(shareInfo) {
            PlatformManagerProxy.instance.PlatformInstance.ShareAppMessage(shareInfo, Laya.Handler.create(this, () => {
                console.log("分享成功");
            }), null);
        }
    }

    class WXPlatform {
        constructor() {
            this.platform = EPlatformType.WX;
            this.safeArea = null;
            this.recordManager = new DefaultRecordManager();
            this.device = new WXDevice();
            this.loginCode = null;
            this.isSupportJumpOther = true;
            this._isBannerLoaded = false;
            this._isVideoLoaded = false;
            this._isInterstitialLoaded = false;
            this._cacheVideoAD = false;
            this.NavigateToAppSuccess = null;
        }
        Init(platformData) {
            this._base = window["wx"];
            if (this._base == null) {
                console.error("平台初始化错误", PlatformManagerProxy.platformStr);
                return;
            }
            this.platformData = platformData;
            this.recordManager.Platform = this;
            this._InitLauchOption();
            this._InitShareInfo();
            this._InitSystemInfo();
            this._CreateBannerAd();
            this._CreateVideoAd();
            this._CreateInterstitalAd();
            window["iplatform"] = this;
        }
        _CheckUpdate() {
            let updateManager = this._base.getUpdateManager();
            if (updateManager == null)
                return;
            updateManager.onCheckForUpdate(function (res) {
                console.log("onCheckForUpdate", res.hasUpdate);
                if (res.hasUpdate) {
                    this._base.showToast({
                        title: "即将有更新请留意"
                    });
                }
            });
            updateManager.onUpdateReady(() => {
                this._base.showModal({
                    title: "更新提示",
                    content: "新版本已经准备好，是否立即使用？",
                    success: function (res) {
                        if (res.confirm) {
                            updateManager.applyUpdate();
                        }
                        else {
                            this._base.showToast({
                                icon: "none",
                                title: "小程序下一次「冷启动」时会使用新版本"
                            });
                        }
                    }
                });
            });
            updateManager.onUpdateFailed(() => {
                this._base.showToast({
                    title: "更新失败，下次启动继续..."
                });
            });
        }
        _Login() {
        }
        GetStorage(key) {
            if (this.base && this.base.getStorageSync && key) {
                try {
                    return this.base.getStorageSync(key);
                }
                catch (error) {
                    console.log('getStorageSync error: ', JSON.stringify(error));
                    return null;
                }
            }
        }
        SetStorage(key, data) {
            if (this.base && this.base.getStorageSync && key) {
                try {
                    return this.base.setStorageSync(key, data);
                }
                catch (error) {
                    console.log('setStorageSync error: ', JSON.stringify(error));
                }
            }
        }
        _OnLoginSuccess(res) {
            console.log(PlatformManagerProxy.platformStr, "登录成功", res);
            this.loginState.isLogin = true;
            this.loginState.code = res.code;
        }
        _InitLauchOption() {
            this._base.onShow(this._OnShow);
            this._base.onHide(this._OnHide);
            let res = this._base.getLaunchOptionsSync();
            this._OnShow(res);
        }
        _InitShareInfo() {
            this._base.showShareMenu({
                withShareTicket: true,
                success: (res) => {
                    console.log("InitShareSuccess", res);
                },
                fail: (res) => {
                    console.log("InitShareFailed", res);
                },
                complete: (res) => {
                    console.log("InitShareComplete", res);
                }
            });
            this._base.onShareAppMessage(() => {
                let shareInfo = ShareManagerAKhgfawgsda.instance.GetShareInfo();
                return WXPlatform._WrapShareInfo(shareInfo);
            });
        }
        static _WrapShareInfo(shareInfo) {
            let shareObj = {};
            if (shareInfo.shareTitle) {
                shareObj["title"] = shareInfo.shareTitle;
            }
            if (shareInfo.shareImg) {
                shareObj["imageUrl"] = shareInfo.shareImg;
            }
            if (shareInfo.sharePath) {
                shareObj["query"] = {};
                let pathSplit = shareInfo.sharePath.split("?");
                let params = pathSplit[1].split("&");
                for (let getParam of params) {
                    let splitParam = getParam.split("=");
                    shareObj["query"][splitParam[0]] = splitParam[1];
                }
            }
            return shareObj;
        }
        _InitSystemInfo() {
            this.base = this._base;
            try {
                if (this._base.getSystemInfoSync != null) {
                    this.systemInfo = this._base.getSystemInfoSync();
                    console.log("系统信息已获取", this.systemInfo);
                    this.safeArea = this.systemInfo.safeArea;
                    this._cacheScreenScale = this.systemInfo.screenWidth / Laya.stage.width;
                }
            }
            catch (e) {
                console.error(e);
                console.error("获取设备信息失败,执行默认初始化");
                this.safeArea = null;
            }
        }
        _CreateInterstitalAd() {
            if (!this.platformData.interstitialId) {
                console.log("无有效的插页广告ID,取消加载");
                return;
            }
            this._interstitalFailedCount = 0;
            let intAdObj = {};
            intAdObj["adUnitId"] = this.platformData.interstitialId;
            this._intersitialAd = this._base.createInterstitialAd(intAdObj);
            if (!this._intersitialAd)
                return;
            this._intersitialAd.onLoad(() => {
                console.log("插页广告加载成功");
                this._isInterstitialLoaded = true;
            });
            this._intersitialAd.onError((err) => {
                this._interstitalFailedCount++;
                console.error("插页广告加载失败", err);
                if (this._interstitalFailedCount > 10) {
                    console.log("第", this._interstitalFailedCount, "次重新加载插页广告");
                    this._intersitialAd.load();
                }
            });
        }
        _CreateVideoAd() {
            console.log('vedio ad id', this.platformData.rewardVideoId);
            if (!this._cacheVideoAD) {
                console.log("当前策略为不缓存视频广告");
                return;
            }
            let createRewardedVideoAd = this._base["createRewardedVideoAd"];
            if (createRewardedVideoAd == null) {
                console.error("无createRewardedVideoAd方法,跳过初始化");
                return;
            }
            if (!this.platformData.rewardVideoId) {
                console.log("无有效的视频广告ID,取消加载");
                return;
            }
            this._videoFailedCount = 0;
            let videoObj = {};
            videoObj["adUnitId"] = this.platformData.rewardVideoId;
            videoObj['multiton'] = true;
            this._rewardVideo = createRewardedVideoAd(videoObj);
            this._rewardVideo.onLoad(() => {
                console.log("视频广告加载成功");
                this._isVideoLoaded = true;
            });
            this._rewardVideo.onError((res) => {
                this._videoFailedCount++;
                console.error("视频广告加载失败", res);
                console.log("重新创建视频广告");
                this._rewardVideo.destroy();
            });
            this._rewardVideo.onClose((res) => {
                Laya.stage.event(PlatformCommonEvent.RESUM_AUDIO);
                console.log("视频回调", res);
                let isEnd = res["isEnded"];
                if (isEnd) {
                    if (this._rewardSuccessed)
                        this._rewardSuccessed.run();
                }
                else {
                    if (this._rewardSkipped)
                        this._rewardSkipped.run();
                }
            });
            this._rewardVideo.load();
        }
        _CreateBannerAd() {
            if (!this.platformData.bannerId) {
                console.log("无有效的banner广告ID,取消加载");
                return;
            }
            let windowWidth = this._base.getSystemInfoSync().windowWidth;
            let windowHeight = this._base.getSystemInfoSync().windowHeight;
            let bannerObj = {};
            bannerObj["adUnitId"] = this.platformData.bannerId;
            bannerObj["adIntervals"] = 30;
            let styleObj = {};
            styleObj["left"] = 0;
            styleObj["top"] = 0;
            styleObj["width"] = 300;
            bannerObj["style"] = styleObj;
            this._bannerAd = this._base.createBannerAd(bannerObj);
            this._isBannerLoaded = false;
            this._bannerAd.onLoad(() => {
                console.log("banner加载成功");
                this._isBannerLoaded = true;
                this._bannerAd.style.top = windowHeight - this._bannerAd.style.realHeight;
                this._bannerAd.style.left = (windowWidth - this._bannerAd.style.realWidth) / 2;
            });
            this._bannerAd.onError((res) => {
                console.error("banner广告加载失败", res);
            });
        }
        IsBannerAvaliable() {
            return this._isBannerLoaded;
        }
        IsVideoAvaliable() {
            return this._isVideoLoaded;
        }
        IsInterstitalAvaliable() {
            return this._isInterstitialLoaded;
        }
        ShowBannerAd() {
            if (!this.IsBannerAvaliable()) {
                return;
            }
            this._bannerAd.show();
        }
        HideBannerAd() {
            this._bannerAd.hide();
        }
        _DoCacheShowVideo(onSuccess, onSkipped) {
            if (!this._isVideoLoaded) {
                console.error("视频广告尚未加载好");
                return;
            }
            this._rewardSuccessed = onSuccess;
            this._rewardSkipped = onSkipped;
            this._isVideoLoaded = false;
            Laya.stage.event(PlatformCommonEvent.PAUSE_AUDIO);
            this._rewardVideo.show();
        }
        _DoNoCacheShowVideo(onSuccess, onSkipped) {
            this._rewardSuccessed = onSuccess;
            this._rewardSkipped = onSkipped;
            if (!this.platformData.rewardVideoId) {
                console.log("无有效的视频广告ID,取消加载");
                onSuccess.run();
                return;
            }
            let createRewardedVideoAd = this._base["createRewardedVideoAd"];
            if (createRewardedVideoAd == null) {
                console.error("无createRewardedVideoAd方法,跳过初始化");
                onSuccess.run();
                return;
            }
            this._videoFailedCount = 0;
            let videoObj = {};
            videoObj["adUnitId"] = this.platformData.rewardVideoId;
            if (this._rewardVideo) {
                this._rewardVideo.offClose(this.onVideoClose);
            }
            this._rewardVideo = createRewardedVideoAd(videoObj);
            this._rewardVideo.onLoad(() => {
                console.log("视频广告加载成功");
                this._isVideoLoaded = true;
            });
            this._rewardVideo.onError((res) => {
                this._videoFailedCount++;
                console.error("视频广告加载失败", res, this._videoFailedCount);
            });
            this._rewardVideo.onClose((res) => {
                Laya.stage.event(PlatformCommonEvent.RESUM_AUDIO);
                console.log("视频回调", res);
                let isEnd = res["isEnded"];
                new Promise((r) => {
                    Laya.timer.frameOnce(1, this, r);
                }).then(() => {
                    if (isEnd) {
                        if (this._rewardSuccessed)
                            this._rewardSuccessed.run();
                    }
                    else {
                        if (this._rewardSkipped)
                            this._rewardSkipped.run();
                    }
                });
            });
            this._rewardVideo.load().then(() => {
                console.log("激励视频 加载成功");
                return this._rewardVideo.show().then(() => {
                }).catch((err) => {
                    console.error(err);
                });
            });
        }
        onVideoClose(res) {
            Laya.stage.event(PlatformCommonEvent.RESUM_AUDIO);
            console.log("视频回调", res);
            let isEnd = res["isEnded"];
            new Promise((r) => {
                Laya.timer.frameOnce(1, this, r);
            }).then(() => {
                if (isEnd) {
                    if (this._rewardSuccessed)
                        this._rewardSuccessed.run();
                }
                else {
                    if (this._rewardSkipped)
                        this._rewardSkipped.run();
                }
            });
        }
        ShowRewardVideoAd(onSuccess, onSkipped) {
            if (this._cacheVideoAD) {
                this._DoCacheShowVideo(onSuccess, onSkipped);
            }
            else {
                this._DoNoCacheShowVideo(onSuccess, onSkipped);
            }
        }
        ShowRewardVideoAdAsync() {
            return new Promise(function (resolve) {
                PlatformManagerProxy.instance.PlatformInstance.ShowRewardVideoAd(Laya.Handler.create(this, () => {
                    resolve(true);
                }), Laya.Handler.create(this, () => {
                    resolve(false);
                }));
            });
        }
        ShowInterstitalAd() {
            if (!this._isInterstitialLoaded) {
                console.error("插页广告尚未加载好");
                return;
            }
            this._intersitialAd.show();
        }
        GetFromAppId() {
            if (this.lauchOption.referrerInfo == null) {
                return null;
            }
            if (!this.lauchOption.referrerInfo.appId) {
                return null;
            }
            return this.lauchOption.referrerInfo.appId;
        }
        _OnShow(res) {
            console.log(PlatformManagerProxy.platformStr, "OnShow", res);
            PlatformManagerProxy.instance.PlatformInstance.lauchOption = res;
            PlatformManagerProxy.instance.PlatformInstance._CheckUpdate();
            this.NavigateToAppSuccess = null;
            if (PlatformManagerProxy.instance.PlatformInstance.onResume) {
                PlatformManagerProxy.instance.PlatformInstance.onResume.runWith(res);
            }
        }
        _OnHide(res) {
            console.log(PlatformManagerProxy.platformStr, "OnHide", res);
            if (PlatformManagerProxy.instance.PlatformInstance.onPause) {
                PlatformManagerProxy.instance.PlatformInstance.onPause.runWith(res);
            }
            if (this.NavigateToAppSuccess) {
                this.NavigateToAppSuccess();
            }
        }
        ShareAppMessage(shareInfo, onSuccess, onFailed) {
            console.log("分享消息", shareInfo);
            let shareObj = WXPlatform._WrapShareInfo(shareInfo);
            this._base.shareAppMessage(shareObj);
            if (onSuccess) {
                onSuccess.run();
            }
        }
        LoadSubpackage(name, onSuccess, onFailed, onProgress) {
            if (this._base['loadSubpackage'] == null) {
                console.log("无加载子包方法,跳过加载子包", name);
                if (onSuccess) {
                    onSuccess.run();
                }
                return;
            }
            let loadObj = {};
            loadObj["name"] = name;
            loadObj["success"] = () => {
                console.log("分包加载成功", name);
                if (onSuccess) {
                    onSuccess.run();
                }
            };
            loadObj["fail"] = () => {
                console.error("分包加载失败", name);
                if (onFailed) {
                    onFailed.run();
                }
            };
            let loadTask = this._base.loadSubpackage(loadObj);
            loadTask.onProgressUpdate((res) => {
                if (Laya.Browser.onMobile) {
                    console.log("分包加载进度", res);
                }
                if (onProgress) {
                    onProgress.runWith(res.progress / 100);
                }
            });
        }
        RecordEvent(eventId, param) {
            console.log("记录事件", eventId, param);
            let aldSendEvent = this._base["aldSendEvent"];
            if (aldSendEvent == null) {
                console.error("阿拉丁sdk尚未接入,请检查配置");
                return;
            }
            if (param != null) {
                aldSendEvent(eventId, param);
            }
            else {
                aldSendEvent(eventId);
            }
        }
        CreateShareVideoBtn(x, y, width, height) {
            let btnObj = {};
            btnObj.style = {
                left: x * this._cacheScreenScale,
                top: y * this._cacheScreenScale,
                height: height * this._cacheScreenScale,
                width: width * this._cacheScreenScale
            };
            btnObj.share = {
                query: {
                    tick: 1
                },
                bgm: "",
                timeRange: [0, 60 * 1000]
            };
            if (this._shareVideoBtn == null) {
                this._shareVideoBtn = this._base.createGameRecorderShareButton(btnObj);
            }
            else {
                this._shareVideoBtn.show();
            }
        }
        HideShareVideoBtn() {
            if (this._shareVideoBtn != null) {
                this._shareVideoBtn.hide();
            }
        }
        ShowToast(str) {
            this._base.showToast({
                title: str,
                duration: 2000
            });
        }
        OpenGameBox(appIds) {
            console.error("当前平台", PlatformManagerProxy.platformStr, "暂不支持互推游戏盒子");
        }
        NavigateToApp(appid, path, extra, showGC, isbanner, adid) {
            return new Promise((resolve, reject) => {
                wx.navigateToMiniProgram({
                    appId: appid,
                    path: path,
                    extraData: extra,
                    envVersion: '',
                    success: (res) => {
                        console.log('小游戏跳转成功', res);
                        resolve(true);
                    },
                    fail: () => {
                        console.log('小游戏跳转失败：');
                        reject(false);
                    },
                    complete: () => { }
                });
            });
        }
        createShortcut() {
            console.log('暂未实现');
        }
    }

    class BDPlatform extends WXPlatform {
        constructor() {
            super(...arguments);
            this.platform = EPlatformType.BD;
            this._showVideoLoad = false;
        }
        Init(platformData) {
            this._base = window["swan"];
            if (this._base == null) {
                console.error("平台初始化错误");
                return;
            }
            this.platformData = platformData;
            this.recordManager.Platform = this;
            this._InitLauchOption();
            this._InitShareInfo();
            this._InitSystemInfo();
            this._isBannerLoaded = false;
            this._isBannerShowed = false;
            this._CreateVideoAd();
            this._CreateInterstitalAd();
            window["iplatform"] = this;
        }
        _CreateBannerAd() {
            if (!this.platformData.bannerId) {
                console.log("无有效的banner广告ID,取消加载");
                return;
            }
            let windowWidth = this._base.getSystemInfoSync().windowWidth;
            let windowHeight = this._base.getSystemInfoSync().windowHeight;
            let bannerObj = {};
            bannerObj["adUnitId"] = this.platformData.bannerId;
            bannerObj["appSid"] = this.platformData.sid;
            let styleObj = {};
            styleObj["left"] = 0;
            styleObj["top"] = 0;
            styleObj["width"] = windowWidth;
            bannerObj["style"] = styleObj;
            this._bannerAd = this._base.createBannerAd(bannerObj);
            this._bannerAd.onLoad(() => {
                console.log("banner加载成功");
                this._isBannerLoaded = true;
                this._bannerAd.style.top = windowHeight - this._bannerAd.style.height;
                this._bannerAd.show();
            });
            this._bannerAd.onError((res) => {
                console.error("banner广告加载失败", res);
            });
        }
        _CreateVideoAd() {
            if (!this.platformData.rewardVideoId) {
                console.log("无有效的视频广告ID,取消加载");
                return;
            }
            this._videoFailedCount = 0;
            let videoObj = {};
            videoObj["adUnitId"] = this.platformData.rewardVideoId;
            videoObj["appSid"] = this.platformData.sid;
            this._rewardVideo = this._base.createRewardedVideoAd(videoObj);
            this._rewardVideo.onLoad(() => {
                console.log("视频广告加载成功");
                this._isVideoLoaded = true;
            });
            this._rewardVideo.onError((res) => {
                this._videoFailedCount++;
                console.error("视频广告加载失败", res);
                if (this._videoFailedCount > 10) {
                    console.log("第", this._videoFailedCount, "次重新加载视频广告");
                    this._rewardVideo.load();
                }
            });
            this._rewardVideo.onClose((res) => {
                this._base.hideLoading();
                Laya.stage.event(PlatformCommonEvent.RESUM_AUDIO);
                console.log("视频回调", res);
                let isEnd = res["isEnded"];
                if (isEnd) {
                    if (this._rewardSuccessed)
                        this._rewardSuccessed.run();
                }
                else {
                    if (this._rewardSkipped)
                        this._rewardSkipped.run();
                }
                this._rewardVideo.load();
            });
        }
        _CreateInterstitalAd() {
        }
        LoadSubpackage(name, onSuccess, onFailed, onProgress) {
            let loadObj = {};
            loadObj["name"] = name;
            loadObj["success"] = () => {
                console.log("分包加载成功", name);
                if (onSuccess) {
                    onSuccess.run();
                }
            };
            loadObj["fail"] = () => {
                console.error("分包加载失败", name);
                if (onFailed) {
                    onFailed.run();
                }
            };
            let loadTask = this._base.loadSubpackage(loadObj);
            loadTask.onProgressUpdate((res) => {
                if (onProgress) {
                    let value = res.progress / 100;
                    if (isNaN(value)) {
                        value = res.loaded / res.total;
                    }
                    onProgress.runWith(value);
                }
            });
        }
        RecordEvent(eventId, param) {
            this._base.reportAnalytics(eventId, param);
        }
        ShowBannerAd() {
            if (this._isBannerLoaded) {
                return;
            }
            this._CreateBannerAd();
        }
        HideBannerAd() {
            if (!this._isBannerLoaded)
                return;
            this._isBannerLoaded = false;
            if (this._bannerAd) {
                this._bannerAd.destroy();
            }
        }
    }

    class WebRecordManager extends DefaultRecordManager {
        constructor() {
            super(...arguments);
            this.supportRecord = false;
        }
        ShareVideo(onSuccess, onCancel, onFailed) {
            if (this.supportRecord) {
                console.log("强制模拟成功");
                if (onSuccess) {
                    onSuccess.run();
                }
            }
            else {
                console.log("强制模拟失败");
                if (onFailed) {
                    onFailed.run();
                }
            }
        }
    }

    class DefaultPlatform {
        constructor() {
            this.platform = EPlatformType.Web;
            this.safeArea = null;
            this.recordManager = new WebRecordManager();
            this.device = new DefaultDevice();
            this.systemInfo = null;
            this.isSupportJumpOther = true;
        }
        Init(platformData) {
            this.loginState = {
                isLogin: false,
                code: null
            };
            this.recordManager.Platform = this;
            Laya.timer.once(500, this, this._FakeLoginEnd);
        }
        _FakeLoginEnd() {
            if (this.onLoginEnd)
                this.onLoginEnd.run();
        }
        IsBannerAvaliable() {
            return false;
        }
        IsVideoAvaliable() {
            return true;
        }
        IsInterstitalAvaliable() {
            return false;
        }
        ShowBannerAd() {
            console.log("调用ShowBannerAd");
        }
        HideBannerAd() {
            console.log("调用HideBannerAd");
        }
        ShowRewardVideoAd(onSuccess, onSkipped) {
            console.log("调用ShowRewardVideoAd");
            onSuccess.run();
        }
        ShowRewardVideoAdAsync() {
            return new Promise(function (resolve) {
                PlatformManagerProxy.instance.PlatformInstance.ShowRewardVideoAd(Laya.Handler.create(this, () => {
                    resolve(true);
                }), Laya.Handler.create(this, () => {
                    resolve(false);
                }));
            });
        }
        ShowInterstitalAd() {
            console.log("调用ShowInterstitalAd");
        }
        GetFromAppId() {
            return null;
        }
        ShareAppMessage(obj, onSuccess = null, onFailed = null) {
            console.log("分享消息", obj);
            if (onSuccess) {
                onSuccess.run();
            }
        }
        LoadSubpackage(name, onSuccess, onFailed) {
            if (onSuccess) {
                onSuccess.run();
            }
        }
        RecordEvent(eventId, param) {
            console.log("记录事件", eventId, param);
        }
        ShareVideoInfo() {
            console.log(PlatformManagerProxy.platformStr, "暂未实现录屏功能");
        }
        _CheckUpdate() {
        }
        ShowToast(str) {
            console.log('显示消息：', str);
        }
        OpenGameBox() {
            console.error("当前平台", PlatformManagerProxy.platformStr, "暂不支持互推游戏盒子");
        }
        NavigateToApp(appid, path, extra) {
            return new Promise((resolve, reject) => {
                console.error("当前平台", PlatformManagerProxy.platformStr, `暂不支持小程序跳转appid:${appid}`);
                resolve(false);
            });
        }
        createShortcut() {
            console.log('创建桌面图标');
        }
        GetStorage(key) {
            console.log('读本地存储');
            return Laya.LocalStorage.getItem(key);
        }
        SetStorage(key, data) {
            console.log('写本地存储');
            Laya.LocalStorage.setItem(key, data);
        }
    }

    class OppoPlatform extends WXPlatform {
        constructor() {
            super(...arguments);
            this.platform = EPlatformType.OPPO;
            this.safeArea = null;
            this.recordManager = new DefaultRecordManager();
            this.device = new DefaultDevice();
            this.isSupportJumpOther = true;
            this._isBannerLoaded = false;
            this._isVideoLoaded = false;
            this._isInterstitialLoaded = false;
            this._isInterstitialCanShow = true;
            this._nativeAdLoaded = false;
            this._cacheVideoAD = false;
        }
        Init(platformData) {
            this._base = window["qg"];
            if (this._base == null) {
                console.error("平台初始化错误", PlatformManagerProxy.platformStr);
                return;
            }
            this.platformData = platformData;
            this.recordManager.Platform = this;
            this._InitLauchOption();
            this._Login();
            this._InitSystemInfo();
            this.getSystemInfo();
            if (this.systemInfo.platformVersion >= 1051) ;
            else {
                this._base.initAdService({
                    appId: platformData.appId,
                    isDebug: true,
                    success: () => {
                        console.log("oppo广告", "初始化广告服务成功", platformData);
                        this._CreateVideoAd();
                    },
                    fail: () => {
                        console.error("oppo广告", "初始化广告服务失败");
                    }
                });
            }
            window["iplatform"] = this;
        }
        getSystemInfo() {
            if (this._base.getSystemInfo != null) {
                this._base.getSystemInfo({
                    success: (res) => {
                        this.systemInfo = res;
                        console.log(this.systemInfo);
                    },
                    fail: () => { },
                    complete: () => { }
                });
            }
        }
        reportMonitor() {
            console.log('oppo上报数据', this.systemInfo);
            if (this.systemInfo && this.systemInfo.platformVersion >= 1060 && this._base.reportMonitor != null) {
                this._base.reportMonitor('game_scene', 0);
            }
        }
        _CheckUpdate() {
        }
        _Login() {
            this.loginState = {
                isLogin: false,
                code: ""
            };
            let loginData = {};
            loginData.success = (res) => {
                this._OnLoginSuccess(res);
            };
            loginData.fail = (res) => {
                console.error(PlatformManagerProxy.platformStr, "登录失败", res);
                this.loginState.isLogin = false;
                this.loginState.code = "";
            };
            loginData.complete = (res) => {
                if (this.onLoginEnd != null) {
                    this.onLoginEnd.run();
                }
            };
            if (this._base.login != null) {
                this._base.login(loginData);
            }
        }
        _OnLoginSuccess(res) {
            console.log(PlatformManagerProxy.platformStr, "登录成功", res);
            this.loginState.isLogin = true;
            this.loginState.code = res.token;
        }
        ShareAppMessage(obj, onSuccess, onFailed) {
        }
        _InitLauchOption() {
            if (this._base.onShow != null) {
                this._base.onShow(this._OnShow);
            }
            if (this._base.onHide != null) {
                this._base.onHide(this._OnHide);
            }
            if (this._base.getLaunchOptionsSync != null) {
                let res = this._base.getLaunchOptionsSync();
                this._OnShow(res);
            }
        }
        canCreateShortcut() {
            return new Promise((resolve, reject) => {
                qg['hasShortcutInstalled']({
                    success: function (res) {
                        resolve(!res);
                    },
                    fail: function (err) {
                        reject();
                    },
                    complete: function () {
                    }
                });
            });
        }
        createShortcut() {
            return new Promise((resolve, reject) => {
                qg['hasShortcutInstalled']({
                    success: function (res) {
                        if (res == false) {
                            qg['installShortcut']({
                                success: function () {
                                    resolve();
                                },
                                fail: function (err) {
                                    reject();
                                },
                                complete: function () { }
                            });
                        }
                        else {
                            resolve();
                        }
                    },
                    fail: function (err) {
                        reject();
                    },
                    complete: function () { }
                });
            });
        }
        _CreateInterstitalAd() {
        }
        _CreateVideoAd() {
            if (!this._cacheVideoAD) {
                console.log("当前策略为不缓存视频广告");
                return;
            }
            let createRewardedVideoAd = this._base["createRewardedVideoAd"];
            if (createRewardedVideoAd == null) {
                console.error("无createRewardedVideoAd方法,跳过初始化");
                return;
            }
            if (!this.platformData.rewardVideoId) {
                console.log("无有效的视频广告ID,取消加载");
                return;
            }
            this._videoFailedCount = 0;
            let videoObj = {};
            videoObj["adUnitId"] = this.platformData.rewardVideoId;
            this._rewardVideo = createRewardedVideoAd(videoObj);
            this._rewardVideo.onLoad(() => {
                console.log("视频广告加载成功");
                this._isVideoLoaded = true;
            });
            this._rewardVideo.onError((res) => {
                this._videoFailedCount++;
                console.error("视频广告加载失败", res);
                if (this._videoFailedCount > 10) {
                    console.log("第", this._videoFailedCount, "次重新加载视频广告");
                    this._rewardVideo.load();
                }
            });
            this._rewardVideo.onClose((res) => {
                Laya.stage.event(PlatformCommonEvent.RESUM_AUDIO);
                console.log("视频回调", res);
                let isEnd = res["isEnded"];
                new Promise((r) => {
                    Laya.timer.frameOnce(1, this, r);
                }).then(() => {
                    if (isEnd) {
                        if (this._rewardSuccessed)
                            this._rewardSuccessed.run();
                    }
                    else {
                        if (this._rewardSkipped)
                            this._rewardSkipped.run();
                    }
                });
            });
        }
        IsBannerAvaliable() {
            return this._isBannerLoaded;
        }
        IsVideoAvaliable() {
            return this._isVideoLoaded;
        }
        IsInterstitalAvaliable() {
            return false;
        }
        IsNativeAvaliable() {
            return this._nativeAdLoaded;
        }
        ShowBannerAd() {
            return __awaiter(this, void 0, void 0, function* () {
                if (!this.platformData.bannerId) {
                    console.log("无有效的banner广告ID,取消加载");
                    return;
                }
                if (this._bannerAd) {
                    this._bannerAd.show();
                    console.log('展示已有banner');
                    return;
                }
                this.HideBannerAd();
                this._bannerAd = this._base.createBannerAd({
                    adUnitId: this.platformData.bannerId
                });
                let isBannerLoading = true;
                let loadSuccess = false;
                this._bannerAd.show().then((res) => {
                    console.log("banner加载成功", res);
                    if (res['code'] == 0) {
                        loadSuccess = true;
                    }
                    isBannerLoading = false;
                }).catch((res) => {
                    console.error("banner加载失败", res);
                    isBannerLoading = false;
                });
                while (isBannerLoading) {
                    yield new Promise((r) => {
                        Laya.timer.frameOnce(1, this, r);
                    });
                }
                if (loadSuccess)
                    return;
                console.log("banner展示失败,展示native广告");
                if (this._bannerAd) {
                    this._bannerAd.destroy();
                    this._bannerAd = null;
                }
                for (let i = 0; i < this.platformData.nativeIconIds.length; ++i) {
                    let ret = yield this._ShowNativeBanner(i);
                    if (ret) {
                        break;
                    }
                    this._bannerAd.destroy();
                }
            });
        }
        _ShowNativeBanner(index) {
            return __awaiter(this, void 0, void 0, function* () {
                let nativeBanner = this.base.createNativeAd({
                    adUnitId: this.platformData.nativeBannerIds[index]
                });
                this._bannerAd = nativeBanner;
                let loadRet = yield nativeBanner.load();
                if (loadRet["code"] == 0) {
                    let adList = loadRet['adList'];
                    if (adList == null || adList.length == 0) {
                        console.error("native banner加载失败", loadRet);
                        return false;
                    }
                    let adData = adList[0];
                    if (adData == null) {
                        console.error("native banner加载失败", loadRet);
                        return false;
                    }
                    return true;
                }
                else {
                    console.error("native banner加载失败", loadRet);
                    return false;
                }
            });
        }
        HideBannerAd() {
            if (this._bannerAd) {
                this._bannerAd.destroy();
                this._bannerAd = null;
            }
        }
        ShowNativeAd() {
            return __awaiter(this, void 0, void 0, function* () {
                if (!this.IsNativeAvaliable()) {
                    return;
                }
            });
        }
        HideNativeAd() {
            if (!this.IsNativeAvaliable()) {
                return;
            }
        }
        _DoCacheShowVideo(onSuccess, onSkipped) {
            if (!this._isVideoLoaded) {
                console.error("视频广告尚未加载好");
                return;
            }
            this._rewardSuccessed = onSuccess;
            this._rewardSkipped = onSkipped;
            this._isVideoLoaded = false;
            Laya.stage.event(PlatformCommonEvent.PAUSE_AUDIO);
            this._rewardVideo.show();
        }
        _DoNoCacheShowVideo(onSuccess, onSkipped) {
            this._rewardSuccessed = onSuccess;
            this._rewardSkipped = onSkipped;
            if (!this.platformData.rewardVideoId) {
                console.log("无有效的视频广告ID,取消加载");
                this._rewardSkipped && this._rewardSkipped.run();
                return;
            }
            let createRewardedVideoAd = this._base["createRewardedVideoAd"];
            if (createRewardedVideoAd == null) {
                console.error("无createRewardedVideoAd方法,跳过初始化");
                this._rewardSkipped && this._rewardSkipped.run();
                return;
            }
            if (this._rewardVideo) {
                this._rewardVideo.destroy();
            }
            let videoObj = {};
            videoObj["adUnitId"] = this.platformData.rewardVideoId;
            this._rewardVideo = createRewardedVideoAd(videoObj);
            console.log("广告创建完成", videoObj);
            this._rewardVideo.onClose((res) => {
                Laya.stage.event(PlatformCommonEvent.RESUM_AUDIO);
                console.log("视频回调", res);
                let isEnd = res["isEnded"];
                new Promise((r) => {
                    Laya.timer.frameOnce(1, this, r);
                }).then(() => {
                    if (isEnd) {
                        if (this._rewardSuccessed)
                            this._rewardSuccessed.run();
                    }
                    else {
                        if (this._rewardSkipped)
                            this._rewardSkipped.run();
                    }
                });
            });
            this._rewardVideo.onError((err) => {
                console.log("广告组件出现问题", err);
                if (this._rewardSkipped)
                    this._rewardSkipped.run();
            });
            this._rewardVideo.onLoad((res) => {
                console.log("广告加载成功", res);
            });
            this._rewardVideo.load().then(() => {
                this._rewardVideo.show();
            });
        }
        ShowRewardVideoAd(onSuccess, onSkipped) {
            if (this._cacheVideoAD) {
                this._DoCacheShowVideo(onSuccess, onSkipped);
            }
            else {
                this._DoNoCacheShowVideo(onSuccess, onSkipped);
            }
        }
        ShowRewardVideoAdAsync() {
            return new Promise(function (resolve) {
                PlatformManagerProxy.instance.PlatformInstance.ShowRewardVideoAd(Laya.Handler.create(this, () => {
                    resolve(true);
                }), Laya.Handler.create(this, () => {
                    resolve(false);
                }));
            });
        }
        _DisableInterstitalAd() {
            return __awaiter(this, void 0, void 0, function* () {
                this._isInterstitialCanShow = false;
                yield new Promise((r) => {
                    Laya.timer.once(60 * 1000, this, r);
                });
                this._isInterstitialCanShow = true;
            });
        }
        GetFromAppId() {
            if (this.lauchOption.referrerInfo == null) {
                return null;
            }
            if (!this.lauchOption.referrerInfo.appId) {
                return null;
            }
            return this.lauchOption.referrerInfo.appId;
        }
        CreatShortcut() {
            return new Promise((resolve, reject) => {
                qg['hasShortcutInstalled']({
                    success: function (res) {
                        if (res == false) {
                            qg['installShortcut']({
                                success: function () {
                                    resolve();
                                },
                                fail: function (err) {
                                    reject();
                                },
                                complete: function () { }
                            });
                        }
                        else {
                            resolve();
                        }
                    },
                    fail: function (err) {
                        reject();
                    },
                    complete: function () { }
                });
            });
        }
        LoadSubpackage(name, onSuccess, onFailed, onProgress) {
            let loadObj = {};
            loadObj["name"] = name;
            loadObj["success"] = () => {
                console.log("分包加载成功", name);
                if (onSuccess) {
                    onSuccess.run();
                }
            };
            loadObj["fail"] = () => {
                console.error("分包加载失败", name);
                if (onFailed) {
                    onFailed.run();
                }
            };
            let loadTask = this._base.loadSubpackage(loadObj);
            loadTask.onProgressUpdate((res) => {
                console.log("分包加载进度", res);
                if (onProgress) {
                    onProgress.runWith(res.progress / 100);
                }
            });
        }
        RecordEvent(eventId, param) {
            console.log("[记录事件]", eventId, param);
        }
        CreateShareVideoBtn(x, y, width, height) {
        }
        HideShareVideoBtn() {
            if (this._shareVideoBtn != null) {
                this._shareVideoBtn.hide();
            }
        }
        ShowToast(str) {
            this._base.showToast({
                title: str,
                duration: 2000
            });
        }
        OpenGameBox(appIds) {
            console.error("当前平台", PlatformManagerProxy.platformStr, "暂不支持互推游戏盒子");
        }
        NavigateToApp(appId, path, extra) {
            return new Promise((resolve, reject) => {
                Laya.Browser.window.qg.navigateToMiniGame({
                    pkgName: appId,
                    path: path,
                    extraData: extra,
                    success: function () {
                        resolve(true);
                        console.log('oppo小游戏跳转成功');
                    },
                    fail: function (res) {
                        reject(false);
                        console.log('oppo小游戏跳转失败：', JSON.stringify(res));
                    }
                });
            });
        }
    }

    class QQPlatform extends WXPlatform {
        constructor() {
            super(...arguments);
            this.platform = EPlatformType.QQ;
            this.isBannerShowing = false;
        }
        Init(platformData) {
            this._base = window["qq"];
            if (this._base == null) {
                console.error("平台初始化错误", PlatformManagerProxy.platformStr);
                return;
            }
            this.platformData = platformData;
            this.recordManager.Platform = this;
            this._InitLauchOption();
            this._Login();
            this._InitShareInfo();
            this._InitSystemInfo();
            this._CreateBannerAd();
            this._CreateVideoAd();
            this._CreateInterstitalAd();
            window["iplatform"] = this;
            console.error("平台初始化完成", PlatformManagerProxy.platformStr);
        }
        _InitSystemInfo() {
            try {
                let systemInfo = this._base.getSystemInfoSync();
                this._cacheScreenScale = systemInfo.screenWidth / Laya.stage.width;
                this.safeArea = {};
                this.safeArea.width = systemInfo.windowWidth;
                this.safeArea.height = systemInfo.windowHeight;
                this.safeArea.top = systemInfo.statusBarHeight;
                this.safeArea.bottom = 0;
                console.log("QQ覆写_InitSystemInfo", this.safeArea);
            }
            catch (e) {
                console.error(e);
                console.error("获取设备信息失败,执行默认初始化");
                this.safeArea = null;
            }
        }
        _CreateBannerAd(show, top) {
            if (!this.platformData.bannerId) {
                console.log("无有效的banner广告ID,取消加载");
                return;
            }
            let windowWidth = this._base.getSystemInfoSync().windowWidth;
            let windowHeight = this._base.getSystemInfoSync().windowHeight;
            let bannerObj = {};
            bannerObj["adUnitId"] = this.platformData.bannerId;
            let styleObj = {};
            styleObj["top"] = ((top == null) ? (windowHeight - 80) : top);
            styleObj["width"] = 300;
            styleObj["left"] = (windowWidth - styleObj["width"]) / 2;
            bannerObj["style"] = styleObj;
            bannerObj['adIntervals'] = 30;
            this._bannerAd = this._base.createBannerAd(bannerObj);
            this._isBannerLoaded = false;
            console.log('banner', this._bannerAd);
            this._bannerAd.onLoad(() => {
                this._isBannerLoaded = true;
                if (show && this.isBannerShowing) {
                    this._bannerAd.show();
                    console.log('---------banner show');
                }
                else {
                    console.log('++++', show, this.isBannerShowing);
                }
            });
            this._bannerAd.onError((res) => {
                console.error("banner广告加载失败", res);
            });
            this._bannerAd.onResize((size) => {
                this._bannerAd.style.left = (windowWidth - 300) / 2;
            });
        }
        IsBannerAvaliable() {
            return this._isBannerLoaded;
        }
        IsVideoAvaliable() {
            return this._isVideoLoaded;
        }
        IsInterstitalAvaliable() {
            return this._isInterstitialLoaded;
        }
        ShowBannerAd(top) {
            if (!this.IsBannerAvaliable()) {
                return;
            }
            top && (this._bannerAd.style.top = top);
            this._bannerAd.show();
            this.isBannerShowing = true;
            Laya.timer.loop(15 * 1000, this, this.refreshBanner);
        }
        ShowBannerNew(top) {
            this._bannerAd && this._bannerAd.hide();
            Laya.timer.clear(this, this.refreshBanner);
            this.isBannerShowing = true;
            console.log('--------------ShowBannerNew');
            this._CreateBannerAd(true, top);
        }
        refreshBanner() {
            if (this.isBannerShowing) {
                this._bannerAd.hide();
                this._CreateBannerAd(true);
            }
        }
        HideBannerAd() {
            if (!this.IsBannerAvaliable())
                return;
            console.log('--------------HideBannerAd');
            if (this._bannerAd) {
                this._bannerAd.destroy();
                Laya.timer.clear(this, this.refreshBanner);
                this.isBannerShowing = false;
            }
        }
        _DoCacheShowVideo(onSuccess, onSkipped) {
            if (!this._isVideoLoaded) {
                console.error("视频广告尚未加载好");
                return;
            }
            this._rewardSuccessed = onSuccess;
            this._rewardSkipped = onSkipped;
            this._isVideoLoaded = false;
            Laya.stage.event(PlatformCommonEvent.PAUSE_AUDIO);
            this._rewardVideo.show();
        }
        _DoNoCacheShowVideo(onSuccess, onSkipped) {
            this._rewardSuccessed = onSuccess;
            this._rewardSkipped = onSkipped;
            if (!this._isVideoLoaded || !this._rewardVideo) {
                if (!this.platformData.rewardVideoId) {
                    console.log("无有效的视频广告ID,取消加载");
                    onSkipped.run();
                    return;
                }
                let createRewardedVideoAd = this._base["createRewardedVideoAd"];
                if (createRewardedVideoAd == null) {
                    console.error("无createRewardedVideoAd方法,跳过初始化");
                    onSkipped.run();
                    return;
                }
                this._videoFailedCount = 0;
                let videoObj = {};
                videoObj["adUnitId"] = this.platformData.rewardVideoId;
                this._rewardVideo = createRewardedVideoAd(videoObj);
                this._rewardVideo.onLoad(() => {
                    console.log("视频广告加载成功");
                    this._isVideoLoaded = true;
                });
                this._rewardVideo.onError((res) => {
                    this._videoFailedCount++;
                    console.error("视频广告加载失败", res, this._videoFailedCount);
                });
                this._rewardVideo.onClose((res) => {
                    Laya.stage.event(PlatformCommonEvent.RESUM_AUDIO);
                    console.log(" NoCache - 视频回调", res);
                    let isEnd = res["isEnded"];
                    console.log("noCache---", isEnd, "----", !!this._rewardSuccessed, "-----", !!this._rewardSkipped);
                    if (isEnd) {
                        if (this._rewardSuccessed)
                            this._rewardSuccessed.run();
                    }
                    else {
                        if (this._rewardSkipped)
                            this._rewardSkipped.run();
                    }
                });
            }
            this._rewardVideo.show().then(() => {
            }).catch(err => {
                console.log("广告组件出现问题", err);
                this._rewardVideo.load().then(() => {
                    console.log("手动加载成功");
                    return this._rewardVideo.show().then(() => {
                    });
                });
            });
        }
        ShowRewardVideoAd(onSuccess, onSkipped) {
            if (this._cacheVideoAD) {
                this._DoCacheShowVideo(onSuccess, onSkipped);
            }
            else {
                this._DoNoCacheShowVideo(onSuccess, onSkipped);
            }
        }
        ShowRewardVideoAdAsync() {
            return new Promise(function (resolve) {
                PlatformManagerProxy.instance.PlatformInstance.ShowRewardVideoAd(Laya.Handler.create(this, () => {
                    resolve(true);
                }, null, true), Laya.Handler.create(this, () => {
                    resolve(false);
                }, null, true));
            });
        }
        ShowInterstitalAd() {
            if (!this._isInterstitialLoaded) {
                console.error("插页广告尚未加载好");
                return;
            }
            this._intersitialAd.show();
        }
        OpenGameBox(appIds = []) {
            this.showAppBox();
        }
        showAppBox() {
            if (this.appBox) {
                this.appBox.show();
            }
        }
        createAppBox(show) {
            if (!this.appBox) {
                this.appBox = this._base.createAppBox({
                    adUnitId: 'a99333d9b9be445d09d5a41997aae4d9'
                });
            }
            this.appBox.load().then(() => {
                if (show) {
                    this.appBox.show();
                }
            });
            this.appBox.onClose(() => {
                console.log('关闭盒子');
            });
        }
        hideAppBox() {
            if (this.appBox) {
                this.appBox.destroy();
            }
        }
        showBlockAd(count = 4) {
            console.log('创建积木广告');
            this.hideBlockAd();
            let obj = {
                adUnitId: "b39860c2f4633a6031ac01ca82531279",
                style: { left: 0, top: 0 },
                size: count,
                orientation: 'vertical'
            };
            this.blockAd = this._base.createBlockAd(obj);
            this.blockAd.onLoad(() => {
                console.log('积木广告加载完成');
                this.blockAd.show().then(() => { console.log('积木展示成功'); }).catch(e => {
                    console.error('积木展示失败', e);
                });
            });
            this.blockAd.onError((err) => {
                console.error('积木广告加载错误', err);
            });
            this.blockAd.onResize((res) => {
                console.log('积木resize', res);
            });
            console.log(this.blockAd);
        }
        hideBlockAd() {
            if (this.blockAd) {
                this.blockAd.hide();
            }
        }
    }

    class QTTPlatform extends WXPlatform {
        constructor() {
            super(...arguments);
            this.platform = EPlatformType.QTT;
        }
        Init(platformData) {
            this._base = window["qttGame"];
            if (this._base == null) {
                console.error("平台初始化错误");
                return;
            }
            this.platformData = platformData;
            this.recordManager.Platform = this;
            window["iplatform"] = this;
        }
        IsBannerAvaliable() {
            return true;
        }
        ShowBannerAd() {
            this._base.showBanner({ index: 1 });
        }
        HideBannerAd() {
            this._base.hideBanner();
        }
        IsVideoAvaliable() {
            return true;
        }
        ShowRewardVideoAd(onSuccess, onSkipped) {
            let options = {};
            options.index = 1;
            options.gametype = 1;
            options.rewardtype = 1;
            options.data = {};
            options.data.title = "获得奖励";
            Laya.stage.event(PlatformCommonEvent.PAUSE_AUDIO);
            this._base.showVideo((res) => {
                Laya.stage.event(PlatformCommonEvent.RESUM_AUDIO);
                if (res == 1) {
                    if (onSuccess) {
                        onSuccess.run();
                    }
                }
                else {
                    if (onSkipped) {
                        onSkipped.run();
                    }
                }
            }, options);
        }
        ShowInterstitalAd() {
            this.ShowHDReward();
        }
        ShowHDReward() {
            let options = {};
            options.index = 1;
            options.rewardtype = 1;
            this._base.showHDReward(options);
        }
        RecordEvent(eventId, param) {
            console.log("记录事件", eventId, param);
        }
    }

    class TTDevice extends DefaultDevice {
        constructor(base) {
            super();
            this._base = base;
        }
        Vibrate(isLong) {
            console.log("调用震动", isLong);
            if (isLong) {
                this._base.vibrateLong({
                    success(res) { },
                    fail(res) {
                        console.error("调用震动失败", res);
                    },
                    complete(res) { }
                });
            }
            else {
                this._base.vibrateShort({
                    success(res) { },
                    fail(res) {
                        console.error("调用震动失败", res);
                    },
                    complete(res) { }
                });
            }
        }
    }

    class TTRecordManager extends DefaultRecordManager {
        constructor(base) {
            super();
            this.supportRecord = true;
            this._base = base;
            this.isRecording = false;
            this.isRecordSuccess = false;
            this.isPausing = false;
            this._nativeManager = this._base.getGameRecorderManager();
            this._nativeManager.onStart((res) => {
                console.log("平台开始录制", res);
                this.isRecording = true;
                this.isRecordSuccess = false;
                this._cacheStartHandle && this._cacheStartHandle.run();
            });
            this._nativeManager.onStop((res) => {
                console.log("平台停止录制", res);
                this.videoSavePath = res.videoPath;
                this.isRecording = false;
                this.isRecordSuccess = true;
                if (this._cacheStopHandle) {
                    this._cacheStopHandle.run();
                }
                else if (this._cacheOverTimeHandle) {
                    this._cacheOverTimeHandle.run();
                }
            });
            this._nativeManager.onError((err) => {
                console.log("录制发生错误", err);
                this.isRecordSuccess = false;
                this.isRecording = false;
            });
            this._nativeManager.onPause((res) => {
                console.log("暂停录制视频", res);
                this.isPausing = true;
                this._cachePauseHandle && this._cachePauseHandle.run();
            });
            this._nativeManager.onResume((res) => {
                console.log("暂停录制视频", res);
                this.isPausing = false;
                this._cacheResumeHandle && this._cacheResumeHandle.run();
            });
        }
        StartRecord(onStart, onOverTime) {
            console.log("调用开始录屏");
            this._cacheStartHandle = onStart;
            this._cacheOverTimeHandle = onOverTime;
            this._cacheStopHandle = null;
            this._nativeManager.start({ duration: 300 });
        }
        Pause(onPause) {
            if (!this.isRecording) {
                console.error("当前未开始录制,无法暂停录制");
                return;
            }
            if (this.isPausing) {
                console.log("当前录制状态已暂停");
                return;
            }
            console.log("调用暂停录制");
            this._cachePauseHandle = onPause;
            this._nativeManager.pause();
        }
        Resume(onReume) {
            if (!this.isRecording) {
                console.error("当前未开始录制,无法恢复录制");
                return;
            }
            if (!this.isPausing) {
                console.log("当前录制状态正在进行中");
                return;
            }
            console.log("调用恢复录制");
            this._cacheResumeHandle = onReume;
            this._nativeManager.resume();
        }
        RecordClip(timeRange) {
            if (!this.isRecording) {
                console.error("当前未开始录制,无法记录精彩时刻");
                return;
            }
            if (this.isPausing) {
                console.log("当前录制状态已暂停,无法记录精彩时刻");
                return;
            }
            if (timeRange == null) {
                this._nativeManager.recordClip({});
            }
            else {
                this._nativeManager.recordClip({ timeRange: timeRange });
            }
        }
        StopRecord(onStop) {
            console.log("调用结束录屏");
            this._cacheStopHandle = onStop;
            this._nativeManager.stop();
        }
        ShareVideo(onSuccess, onCancel, onFailed) {
            if (this.isRecordSuccess) {
                let shareData = {
                    channel: "video",
                    title: "",
                    desc: "",
                    imageUrl: "",
                    templateId: this.Platform.platformData.shareId,
                    query: "",
                    extra: {
                        videoPath: this.videoSavePath,
                        videoTopics: ['抖音小游戏', '开间厕所当老板']
                    },
                    success() {
                        if (onSuccess) {
                            onSuccess.run();
                        }
                    },
                    fail(e) {
                        if (onCancel) {
                            onCancel.run();
                        }
                    }
                };
                this._base.shareAppMessage(shareData);
            }
            else {
                console.log("无视频可以分享");
                if (onFailed) {
                    onFailed.run();
                }
            }
        }
    }

    class TTPlatform extends WXPlatform {
        constructor() {
            super(...arguments);
            this.platform = EPlatformType.TT;
            this._showVideoLoad = false;
        }
        Init(platformData) {
            this._base = window["tt"];
            if (this._base == null) {
                console.error("平台初始化错误");
                return;
            }
            this.platformData = platformData;
            let tt = this._base;
            let systemInfo = tt.getSystemInfoSync();
            if (systemInfo.platform == "ios") {
                this.isSupportJumpOther = false;
            }
            let [major, minor] = systemInfo.SDKVersion.split(".");
            if (major >= 1 && minor >= 33) ;
            else {
                this.isSupportJumpOther = false;
            }
            this._InitLauchOption();
            this._InitShareInfo();
            this._InitSystemInfo();
            this._CreateBannerAd();
            this._CreateVideoAd();
            this._CreateInterstitalAd();
            this.recordManager = new TTRecordManager(this._base);
            this.recordManager.Platform = this;
            this.device = new TTDevice(this._base);
            window["iplatform"] = this;
        }
        _CreateBannerAd() {
            if (!this.platformData.bannerId) {
                console.log("无有效的banner广告ID,取消加载");
                return;
            }
            let windowWidth = this._base.getSystemInfoSync().windowWidth;
            let windowHeight = this._base.getSystemInfoSync().windowHeight;
            let bannerObj = {};
            bannerObj["adUnitId"] = this.platformData.bannerId;
            bannerObj["adIntervals"] = 30;
            let styleObj = {};
            styleObj["left"] = 0;
            styleObj["top"] = 0;
            styleObj["width"] = windowWidth;
            bannerObj["style"] = styleObj;
            this._bannerAd = this._base.createBannerAd(bannerObj);
            this._isBannerLoaded = false;
            if (this._bannerAd) {
                this._bannerAd.onLoad(() => {
                    console.log("banner加载成功", this._bannerAd);
                    this._isBannerLoaded = true;
                });
                this._bannerAd.onError((res) => {
                    console.error("banner广告加载失败", res);
                    this._bannerAd == null;
                });
                this._bannerAd.onResize((size) => {
                    this._bannerAd.style.top = windowHeight - size.height;
                    this._bannerAd.style.left = (windowWidth - size.width) / 2;
                });
            }
        }
        RecordEvent(eventId, param) {
            let reportAnalytics = this._base["reportAnalytics"];
            if (reportAnalytics) {
                if (param == null) {
                    param = {};
                }
                reportAnalytics(eventId, param);
            }
            else {
                console.error("reportAnalytics 方法不存在");
            }
        }
        ShowBannerAd() {
            if (!this.IsBannerAvaliable()) {
                return;
            }
            this._bannerAd.show();
        }
        ShareAppMessage(shareInfo, onSuccess, onFailed) {
            console.log("分享消息", shareInfo);
            let shareObj = WXPlatform._WrapShareInfo(shareInfo);
            shareObj["success"] = () => {
                if (onSuccess) {
                    onSuccess.run();
                }
            };
            shareObj["fail"] = () => {
                if (onFailed) {
                    onFailed.run();
                }
            };
            this._base.shareAppMessage(shareObj);
        }
        OpenGameBox(appIds) {
            let openData = [];
            for (let i = 0; i < appIds.length; ++i) {
                openData.push({
                    appId: appIds[i]
                });
            }
            this._base.showMoreGamesModal({
                appLaunchOptions: openData
            });
        }
        NavigateToApp(appid, path, extra) {
            return new Promise((resolve, reject) => {
                if (!this.isSupportJumpOther) {
                    reject(false);
                    console.log("当前平台不支持小游戏跳转", this);
                }
                else {
                    this._base.showMoreGamesModal({
                        appLaunchOptions: [
                            {
                                appId: this.platformData.appId,
                                query: "foo=bar&baz=qux",
                                extraData: {}
                            }
                        ],
                        success(res) {
                            resolve(true);
                            console.log("跳转小游戏成功", appid);
                        },
                        fail(err) {
                            reject(false);
                            console.log("跳转小游戏失败", appid);
                        }
                    });
                }
            });
        }
    }

    class VivoPlatform {
        constructor() {
            this.isSupportJumpOther = true;
            this.platform = EPlatformType.VIVO;
            this.recordManager = new DefaultRecordManager();
            this.safeArea = null;
            this.loginCode = null;
            this._isBannerLoaded = false;
            this._isVideoLoaded = false;
            this._isInterstitialLoaded = false;
            this._nativeAdLoaded = false;
            this._cacheVideoAD = false;
            this.NavigateToAppSuccess = null;
        }
        Init(platformData) {
            this._base = window["qg"];
            if (this._base == null) {
                console.error("平台初始化错误", PlatformManagerProxy.platformStr);
                return;
            }
            this.platformData = platformData;
            this.recordManager.Platform = this;
            this._InitLauchOption();
            this._Login();
            this._InitSystemInfo();
            this._CreateVideoAd();
            window["iplatform"] = this;
        }
        _CheckUpdate() {
            if (this._base.getUpdateManager == null)
                return;
            let updateManager = this._base.getUpdateManager();
            if (updateManager == null)
                return;
            updateManager.onCheckForUpdate(function (res) {
                console.log("onCheckForUpdate", res.hasUpdate);
                if (res.hasUpdate) {
                    this._base.showToast({
                        title: "即将有更新请留意"
                    });
                }
            });
            updateManager.onUpdateReady(() => {
                this._base.showModal({
                    title: "更新提示",
                    content: "新版本已经准备好，是否立即使用？",
                    success: function (res) {
                        if (res.confirm) {
                            updateManager.applyUpdate();
                        }
                        else {
                            this._base.showToast({
                                icon: "none",
                                title: "小程序下一次「冷启动」时会使用新版本"
                            });
                        }
                    }
                });
            });
            updateManager.onUpdateFailed(() => {
                this._base.showToast({
                    title: "更新失败，下次启动继续..."
                });
            });
        }
        _Login() {
            this.loginState = {
                isLogin: false,
                code: ""
            };
            let loginData = {};
            loginData.success = (res) => {
                this.loginCode = res.code;
                this._OnLoginSuccess(res);
                console.error(this.loginState);
            };
            loginData.fail = (res) => {
                console.error(PlatformManagerProxy.platformStr, "登录失败", res);
                this.loginState.isLogin = false;
                this.loginState.code = "";
            };
            loginData.complete = () => {
                if (this.onLoginEnd != null) {
                    this.onLoginEnd.run();
                }
            };
            this._base.login(loginData);
        }
        GetStorage(key) {
            if (this.base && this.base.getStorageSync && key) {
                try {
                    return this.base.getStorageSync(key);
                }
                catch (error) {
                    console.log('getStorageSync error: ', JSON.stringify(error));
                    return null;
                }
            }
        }
        SetStorage(key, data) {
            if (this.base && this.base.getStorageSync && key) {
                try {
                    return this.base.setStorageSync(key, data);
                }
                catch (error) {
                    console.log('setStorageSync error: ', JSON.stringify(error));
                }
            }
        }
        _OnLoginSuccess(res) {
            console.log(PlatformManagerProxy.platformStr, "登录成功", res);
            this.loginState.isLogin = true;
            this.loginState.code = res.code;
        }
        _InitLauchOption() {
            this._base.onShow(this._OnShow);
            this._base.onHide(this._OnHide);
            let res = this._base.getLaunchOptionsSync();
            this._OnShow(res);
        }
        _InitShareInfo() {
            this._base.showShareMenu({
                withShareTicket: true,
                success: (res) => {
                    console.log("InitShareSuccess", res);
                },
                fail: (res) => {
                    console.log("InitShareFailed", res);
                },
                complete: (res) => {
                    console.log("InitShareComplete", res);
                }
            });
            this._base.onShareAppMessage(() => {
                let shareInfo = ShareManagerAKhgfawgsda.instance.GetShareInfo();
                return VivoPlatform._WrapShareInfo(shareInfo);
            });
        }
        static _WrapShareInfo(shareInfo) {
            let shareObj = {};
            if (shareInfo.shareTitle) {
                shareObj["title"] = shareInfo.shareTitle;
            }
            if (shareInfo.shareImg) {
                shareObj["imageUrl"] = shareInfo.shareImg;
            }
            if (shareInfo.sharePath) {
                shareObj["query"] = {};
                let pathSplit = shareInfo.sharePath.split("?");
                let params = pathSplit[1].split("&");
                for (let getParam of params) {
                    let splitParam = getParam.split("=");
                    shareObj["query"][splitParam[0]] = splitParam[1];
                }
            }
            return shareObj;
        }
        _InitSystemInfo() {
            this.base = this._base;
            try {
                this.systemInfo = this._base.getSystemInfoSync();
                console.log("系统信息已获取", this.systemInfo);
                this.safeArea = this.systemInfo.safeArea;
                this._cacheScreenScale = this.systemInfo.screenWidth / Laya.stage.width;
            }
            catch (e) {
                console.error(e);
                console.error("获取设备信息失败,执行默认初始化");
                this.safeArea = null;
            }
        }
        _CreateInterstitalAd() {
            if (!this.platformData.interstitialId) {
                console.log("无有效的插页广告ID,取消加载");
                return;
            }
            this._interstitalFailedCount = 0;
            let intAdObj = {};
            intAdObj["adUnitId"] = this.platformData.interstitialId;
            this._intersitialAd = this._base.createInterstitialAd(intAdObj);
            if (!this._intersitialAd)
                return;
            this._intersitialAd.onLoad(() => {
                console.log("插页广告加载成功");
                this._isInterstitialLoaded = true;
            });
            this._intersitialAd.onError((err) => {
                this._interstitalFailedCount++;
                console.error("插页广告加载失败", err);
                if (this._interstitalFailedCount > 10) {
                    console.log("第", this._interstitalFailedCount, "次重新加载插页广告");
                    this._intersitialAd.load();
                }
            });
        }
        _CreateVideoAd() {
            console.log('vedio ad id', this.platformData.rewardVideoId);
            if (!this._cacheVideoAD) {
                console.log("当前策略为不缓存视频广告");
                return;
            }
            let createRewardedVideoAd = this._base["createRewardedVideoAd"];
            if (createRewardedVideoAd == null) {
                console.error("无createRewardedVideoAd方法,跳过初始化");
                return;
            }
            if (!this.platformData.rewardVideoId) {
                console.log("无有效的视频广告ID,取消加载");
                return;
            }
            this._videoFailedCount = 0;
            let videoObj = {};
            videoObj["adUnitId"] = this.platformData.rewardVideoId;
            this._rewardVideo = createRewardedVideoAd(videoObj);
            this._rewardVideo.onLoad(() => {
                console.log("视频广告加载成功");
                this._isVideoLoaded = true;
            });
            this._rewardVideo.onError((res) => {
                this._videoFailedCount++;
                console.error("视频广告加载失败", res);
                if (this._videoFailedCount > 10) {
                    console.log("第", this._videoFailedCount, "次重新加载视频广告");
                    this._rewardVideo.load();
                }
            });
            this._rewardVideo.onClose((res) => {
                Laya.stage.event(PlatformCommonEvent.RESUM_AUDIO);
                console.log("视频回调,_CreateVideoAd", res);
                let isEnd = res["isEnded"];
                console.log('is end', isEnd);
                new Promise((r) => {
                    Laya.timer.frameOnce(1, this, r);
                }).then(() => {
                    if (isEnd) {
                        if (this._rewardSuccessed)
                            this._rewardSuccessed.run();
                    }
                    else {
                        if (this._rewardSkipped)
                            this._rewardSkipped.run();
                    }
                });
            });
        }
        _CreateBannerAd() {
            if (!this.platformData.bannerId) {
                console.log("无有效的banner广告ID,取消加载");
                return;
            }
            this._base.getSystemInfoSync().windowWidth;
            let windowHeight = this._base.getSystemInfoSync().windowHeight;
            let bannerObj = {};
            bannerObj["adUnitId"] = this.platformData.bannerId;
            bannerObj["adIntervals"] = 30;
            let styleObj = {};
            styleObj["left"] = 0;
            styleObj["top"] = windowHeight - 80;
            styleObj["width"] = 300;
            bannerObj["style"] = styleObj;
            this._bannerAd = this._base.createBannerAd(bannerObj);
            this._isBannerLoaded = false;
            this._bannerAd.onLoad(() => {
                console.log("banner加载成功");
                this._isBannerLoaded = true;
            });
            this._bannerAd.onSize((size) => {
                console.log('banner onSize', size);
            });
            this._bannerAd.onError((res) => {
                console.error("banner广告加载失败", res);
            });
            console.log('创建banner');
        }
        IsBannerAvaliable() {
            return this._isBannerLoaded;
        }
        IsVideoAvaliable() {
            if (this._cacheVideoAD == false) {
                return true;
            }
            return this._isVideoLoaded;
        }
        IsInterstitalAvaliable() {
            return this._isInterstitialLoaded;
        }
        IsNativeAvaliable() {
            return this._nativeAdLoaded;
        }
        ShowBannerAd() {
            return __awaiter(this, void 0, void 0, function* () {
                if (!this.IsBannerAvaliable()) {
                    console.log('IsBannerAvaliable == false');
                    return;
                }
                if (!this.platformData.bannerId) {
                    console.log("无有效的banner广告ID,取消加载");
                    return;
                }
                if (this._bannerAd) {
                    this._bannerAd.show();
                    console.log('展示已有banner');
                    return;
                }
                this.HideBannerAd();
                this._bannerAd = this._base.createBannerAd({
                    adUnitId: this.platformData.bannerId
                });
                let isBannerLoading = true;
                let loadSuccess = false;
                this._bannerAd.show().then((res) => {
                    console.log("banner加载成功", res);
                    if (res['code'] == 0) {
                        loadSuccess = true;
                    }
                    isBannerLoading = false;
                }).catch((res) => {
                    console.error("banner加载失败", res);
                    isBannerLoading = false;
                });
                while (isBannerLoading) {
                    yield new Promise((r) => {
                        Laya.timer.frameOnce(1, this, r);
                    });
                }
                if (loadSuccess)
                    return;
                console.log("banner展示失败,展示native广告");
                if (this._bannerAd) {
                    this._bannerAd.destroy();
                    this._bannerAd = null;
                }
                for (let i = 0; i < this.platformData.nativeIconIds.length; ++i) {
                    let ret = yield this._ShowNativeBanner(i);
                    if (ret) {
                        break;
                    }
                    this._bannerAd.destroy();
                }
            });
        }
        HideBannerAd() {
            if (this._bannerAd != null) {
                this._bannerAd.hide();
            }
        }
        ShowNativeAd() {
            return __awaiter(this, void 0, void 0, function* () {
                if (!this.IsNativeAvaliable()) {
                    return;
                }
            });
        }
        HideNativeAd() {
            if (!this.IsNativeAvaliable()) {
                return;
            }
        }
        _ShowNativeBanner(index) {
            return __awaiter(this, void 0, void 0, function* () {
                let nativeBanner = this.base.createNativeAd({
                    adUnitId: this.platformData.nativeBannerIds[index]
                });
                this._bannerAd = nativeBanner;
                let loadRet = yield nativeBanner.load();
                if (loadRet["code"] == 0) {
                    let adList = loadRet['adList'];
                    if (adList == null || adList.length == 0) {
                        console.error("native banner加载失败", loadRet);
                        return false;
                    }
                    let adData = adList[0];
                    if (adData == null) {
                        console.error("native banner加载失败", loadRet);
                        return false;
                    }
                    return true;
                }
                else {
                    console.error("native banner加载失败", loadRet);
                    return false;
                }
            });
        }
        _DoCacheShowVideo(onSuccess, onSkipped) {
            if (!this._isVideoLoaded) {
                console.error("视频广告尚未加载好");
                return;
            }
            this._rewardSuccessed = onSuccess;
            this._rewardSkipped = onSkipped;
            this._isVideoLoaded = false;
            Laya.stage.event(PlatformCommonEvent.PAUSE_AUDIO);
            this._rewardVideo.show();
        }
        _DoNoCacheShowVideo(onSuccess, onSkipped) {
            this._rewardSuccessed = onSuccess;
            this._rewardSkipped = onSkipped;
            if (!this.platformData.rewardVideoId) {
                console.log("无有效的视频广告ID,取消加载");
                onSkipped.run();
                return;
            }
            let createRewardedVideoAd = this._base["createRewardedVideoAd"];
            if (createRewardedVideoAd == null) {
                console.error("无createRewardedVideoAd方法,跳过初始化");
                onSkipped.run();
                return;
            }
            this._videoFailedCount = 0;
            let videoObj = {};
            videoObj["adUnitId"] = this.platformData.rewardVideoId;
            if (this._rewardVideo) {
                this._rewardVideo.offClose(this.onVideoClose);
            }
            this._rewardVideo = createRewardedVideoAd(videoObj);
            this._rewardVideo.onLoad(() => {
                console.log("视频广告加载成功");
                this._isVideoLoaded = true;
            });
            this._rewardVideo.onError((res) => {
                this._videoFailedCount++;
                console.error("视频广告加载失败", res, this._videoFailedCount);
            });
            this._rewardVideo.onClose((res) => {
                Laya.stage.event(PlatformCommonEvent.RESUM_AUDIO);
                let isEnd = res["isEnded"];
                Utils.WaitFrame(1).then(() => {
                    if (isEnd) {
                        if (this._rewardSuccessed) {
                            console.log('this successed');
                            try {
                                this._rewardSuccessed.run();
                            }
                            catch (e) {
                                console.error(e);
                            }
                        }
                    }
                    else {
                        if (this._rewardSkipped)
                            this._rewardSkipped.run();
                    }
                });
            });
            this._rewardVideo.load().then(() => {
                console.log("激励视频 加载成功");
                return this._rewardVideo.show().then(() => {
                }).catch((err) => {
                    console.error(err);
                });
            });
        }
        onVideoClose(res) {
            Laya.stage.event(PlatformCommonEvent.RESUM_AUDIO);
            console.log("视频回调,onVideoClose", res);
            let isEnd = res["isEnded"];
            console.log('is end', isEnd);
            new Promise((r) => {
                Laya.timer.frameOnce(1, this, r);
            }).then(() => {
                if (isEnd) {
                    if (this._rewardSuccessed)
                        this._rewardSuccessed.run();
                }
                else {
                    if (this._rewardSkipped)
                        this._rewardSkipped.run();
                }
            });
        }
        ShowRewardVideoAd(onSuccess, onSkipped) {
            if (this._cacheVideoAD) {
                this._DoCacheShowVideo(onSuccess, onSkipped);
            }
            else {
                this._DoNoCacheShowVideo(onSuccess, onSkipped);
            }
        }
        ShowRewardVideoAdAsync() {
            SoundManager.inst.StopBGM();
            return new Promise((resolve) => {
                PlatformManagerProxy.instance.PlatformInstance.ShowRewardVideoAd(Laya.Handler.create(this, () => {
                    SoundManager.inst.PlayBGM();
                    console.log('rrr true');
                    resolve(true);
                }), Laya.Handler.create(this, () => {
                    SoundManager.inst.PlayBGM();
                    resolve(false);
                }));
            });
        }
        ShowInterstitalAd() {
            if (!this._isInterstitialLoaded) {
                console.error("插页广告尚未加载好");
                return;
            }
            this._intersitialAd.show();
        }
        GetFromAppId() {
            if (this.lauchOption.referrerInfo == null) {
                return null;
            }
            if (!this.lauchOption.referrerInfo.appId) {
                return null;
            }
            return this.lauchOption.referrerInfo.appId;
        }
        _OnShow(res) {
            console.log(PlatformManagerProxy.platformStr, "OnShow", res);
            PlatformManagerProxy.instance.PlatformInstance.lauchOption = res;
            PlatformManagerProxy.instance.PlatformInstance._CheckUpdate();
            this.NavigateToAppSuccess = null;
            new Promise((r) => {
                Laya.timer.frameOnce(1, this, r);
            }).then(() => {
                if (PlatformManagerProxy.instance.PlatformInstance.onResume) {
                    PlatformManagerProxy.instance.PlatformInstance.onResume.runWith(res);
                }
            });
        }
        _OnHide(res) {
            console.log(PlatformManagerProxy.platformStr, "OnHide", res);
            if (PlatformManagerProxy.instance.PlatformInstance.onPause) {
                PlatformManagerProxy.instance.PlatformInstance.onPause.runWith(res);
            }
            if (this.NavigateToAppSuccess) {
                this.NavigateToAppSuccess();
            }
        }
        ShareAppMessage(shareInfo, onSuccess, onFailed) {
            console.log("分享消息", shareInfo);
            let shareObj = VivoPlatform._WrapShareInfo(shareInfo);
            this._base.shareAppMessage(shareObj);
            if (onSuccess) {
                onSuccess.run();
            }
        }
        LoadSubpackage(name, onSuccess, onFailed, onProgress) {
            if (this._base['loadSubpackage'] == null) {
                console.log("无加载子包方法,跳过加载子包", name);
                if (onSuccess) {
                    onSuccess.run();
                }
                return;
            }
            let loadObj = {};
            loadObj["name"] = name;
            loadObj["success"] = () => {
                console.log("分包加载成功", name);
                if (onSuccess) {
                    onSuccess.run();
                }
            };
            loadObj["fail"] = () => {
                console.error("分包加载失败", name);
                if (onFailed) {
                    onFailed.run();
                }
            };
            let loadTask = this._base.loadSubpackage(loadObj);
            loadTask.onProgressUpdate((res) => {
                if (Laya.Browser.onMobile) {
                    console.log("分包加载进度", res);
                }
                if (onProgress) {
                    onProgress.runWith(res.progress / 100);
                }
            });
        }
        RecordEvent(eventId, param) {
            console.log("记录事件", eventId, param);
            let aldSendEvent = this._base["aldSendEvent"];
            if (aldSendEvent == null) {
                console.error("阿拉丁sdk尚未接入,请检查配置");
                return;
            }
            if (param != null) {
                aldSendEvent(eventId, param);
            }
            else {
                aldSendEvent(eventId);
            }
        }
        CreateShareVideoBtn(x, y, width, height) {
            let btnObj = {};
            btnObj.style = {
                left: x * this._cacheScreenScale,
                top: y * this._cacheScreenScale,
                height: height * this._cacheScreenScale,
                width: width * this._cacheScreenScale
            };
            btnObj.share = {
                query: {
                    tick: 1
                },
                bgm: "",
                timeRange: [0, 60 * 1000]
            };
            if (this._shareVideoBtn == null) {
                this._shareVideoBtn = this._base.createGameRecorderShareButton(btnObj);
            }
            else {
                this._shareVideoBtn.show();
            }
        }
        HideShareVideoBtn() {
            if (this._shareVideoBtn != null) {
                this._shareVideoBtn.hide();
            }
        }
        ShowToast(str) {
            this._base.showToast({
                message: str,
                duration: 0
            });
        }
        OpenGameBox(appIds) {
            console.error("当前平台", PlatformManagerProxy.platformStr, "暂不支持互推游戏盒子");
        }
        NavigateToApp(appid, path, extra, showGC, isbanner, adid) {
            return new Promise((resolve, reject) => {
                wx.navigateToMiniProgram({
                    appId: appid,
                    path: path,
                    extraData: extra,
                    envVersion: '',
                    success: (res) => {
                        console.log('小游戏跳转成功', res);
                        resolve(true);
                    },
                    fail: () => {
                        console.log('小游戏跳转失败：');
                        reject(false);
                    },
                    complete: () => { }
                });
            });
        }
        hasShortcut() {
            return new Promise(r => {
                qg['hasShortcutInstalled']({
                    success: function (status) {
                        r(status);
                    }, fail: function () {
                        r(false);
                    }
                });
            });
        }
        createShortcut() {
            return new Promise((resolve, reject) => {
                qg['hasShortcutInstalled']({
                    success: function (res) {
                        if (res == false) {
                            qg['installShortcut']({
                                success: function () {
                                    resolve();
                                },
                                fail: function (err) {
                                    reject(err);
                                },
                                complete: function () { }
                            });
                        }
                    },
                    fail: function (err) {
                        reject(err);
                    },
                    complete: function () { }
                });
            });
        }
    }

    class PlatformManager {
        static get instance() {
            if (this._instance == null) {
                this._instance = new PlatformManager();
            }
            return this._instance;
        }
        static get PlatformInstance() {
            if (!this.instance.m_platformInstance) {
                console.log('还没有设置过平台实例代理！');
            }
            return this.instance.m_platformInstance;
        }
        init() {
            if (this.m_platformInstance != null) {
                console.error("已调用过平台创建为", PlatformManagerProxy.GetPlatformStr(this.m_platformInstance.platform), "不能重复创建");
                return this.m_platformInstance;
            }
            let isQTT = window["qttGame"] != null;
            let isTT = window["tt"] != null;
            let result;
            if (isTT) {
                result = new TTPlatform();
                this.m_platformData = new TTData();
            }
            else if (Laya.Browser.onMiniGame) {
                result = new WXPlatform();
                this.m_platformData = new WXData();
            }
            else if (Laya.Browser.onBDMiniGame) {
                result = new BDPlatform();
                this.m_platformData = new BDData();
            }
            else if (isQTT) {
                result = new QTTPlatform();
                this.m_platformData = new QTTData();
            }
            else if (Laya.Browser.onQQMiniGame) {
                result = new QQPlatform();
                this.m_platformData = new QQData();
            }
            else if (Laya.Browser.onQGMiniGame) {
                result = new OppoPlatform();
                this.m_platformData = new OPPOData();
            }
            else if (Laya.Browser.onVVMiniGame) {
                result = new VivoPlatform();
                this.m_platformData = new VIVOData();
            }
            else {
                console.log("未识别平台,默认创建为web", Laya.Browser.userAgent);
                result = new DefaultPlatform();
            }
            this.m_platformInstance = result;
            PlatformManagerProxy.instance.PlatformInstance = result;
            window['$Platform'] = this.m_platformInstance;
            console.log("平台实例创建完成", PlatformManagerProxy.GetPlatformStr(this.m_platformInstance.platform));
        }
        initPlatform() {
            this.m_platformInstance.Init(this.m_platformData);
            console.log('平台初始化完成');
        }
    }

    class UI_ui_huaweicover extends fgui.GComponent {
        static createInstance() {
            return (fgui.UIPackage.createObject("huaweiLogin", "ui_huaweicover"));
        }
        onConstruct() {
            this.m_graph_bg = (this.getChildAt(0));
            this.m_bottom = (this.getChildAt(5));
            this.m_t1 = this.getTransitionAt(0);
        }
    }
    UI_ui_huaweicover.URL = "ui://gqywpino8p3z0";

    let LoginCoverUI = class LoginCoverUI extends BaseUI {
        get URL() {
            return UI_ui_huaweicover.URL;
        }
        get FullScreenItems() {
            return [this.ui.m_graph_bg];
        }
        OnCreate() {
            var rect = this.ui.localToGlobalRect(0, this.ui.m_bottom.y, this.ui.width, this.ui.height);
            var p = this.ui.globalToLocal(0, Laya.stage.height - this.ui.m_bottom.height * rect.height / this.ui.height);
            this.ui.m_bottom.y = p.y;
        }
        OnShow(...args) {
        }
        OnHide() {
        }
        OnDispose() {
            MsgManager.getInst().offAllCaller(this);
        }
        onClose() {
            UIManager.inst.PopUI(this.layer, true);
        }
    };
    LoginCoverUI = __decorate([
        ui_register()
    ], LoginCoverUI);

    class fguiUtils {
        static LoadPackage(resKey) {
            return new ProgressPromise((r, p) => {
                fgui.UIPackage.loadPackage(resKey, Laya.Handler.create(this, () => {
                    r();
                }), Laya.Handler.create(this, v => {
                    p(v);
                }, null, false));
            });
        }
    }

    var VertexUVVS = `
#include "Lighting.glsl";

uniform mat4 u_MvpMatrix;

attribute vec4 a_Position;

attribute vec2 a_TexCoord0;

attribute vec4 a_Color;

varying vec2 v_TexCoord0;

varying vec4 v_Color;

void main() {

    v_TexCoord0 = a_TexCoord0;

    v_Color = a_Color;

    gl_Position = u_MvpMatrix * a_Position;

    gl_Position = remapGLPositionZ(gl_Position);
}
`;
    var VertexUVFS = `
#ifdef GL_FRAMMENT_PRECISION_HIGH
precision highp float;
#else
precision mediump float;
#endif

varying vec2 v_TexCoord0;

varying vec4 v_Color;

uniform vec4 u_MainColor;
uniform float u_progress;
uniform float u_radius;

void main() {
    if(v_TexCoord0.y < 1.0 - u_progress) {
        discard;
    }

    vec2 uv = v_TexCoord0 - vec2(0.5, 0.5);
    float uvx = abs(uv.x);
    float uvy = abs(uv.y);
    uvx = max(0.0, uvx - (0.5 - u_radius));
    uvy = max(0.0, uvy - (0.5 - u_radius));

    if(uvx * uvx + uvy * uvy > u_radius * u_radius) {
        discard;
    }
    gl_FragColor = u_MainColor;
}
`;
    class MyShader {
        static initShader() {
            var attributeMap = {
                'a_Position': Laya.VertexMesh.MESH_POSITION0,
                'a_TexCoord0': Laya.VertexMesh.MESH_TEXTURECOORDINATE0
            };
            var uniformMap = {
                'u_MvpMatrix': Laya.Shader3D.PERIOD_SPRITE,
                'u_MainColor': Laya.Shader3D.PERIOD_MATERIAL,
                'u_progress': Laya.Shader3D.PERIOD_MATERIAL,
                'u_radius': Laya.Shader3D.PERIOD_MATERIAL,
            };
            let stateMap = {
                's_Cull': Laya.Shader3D.RENDER_STATE_CULL,
                's_Blend': Laya.Shader3D.RENDER_STATE_BLEND,
                's_BlendSrc': Laya.Shader3D.RENDER_STATE_BLEND_SRC,
                's_BlendDst': Laya.Shader3D.RENDER_STATE_BLEND_DST,
                's_DepthTest': Laya.Shader3D.RENDER_STATE_DEPTH_TEST,
                's_DepthWrite': Laya.Shader3D.RENDER_STATE_DEPTH_WRITE
            };
            var shader = Laya.Shader3D.add("VertexUV");
            var subShader = new Laya.SubShader(attributeMap, uniformMap);
            shader.addSubShader(subShader);
            subShader.addShaderPass(VertexUVVS, VertexUVFS, stateMap);
        }
    }

    class SubPKGLoader {
        constructor() {
            this.pkgProgress = {};
            this.pkgLoadTag = {};
            this._loadSub = [];
            this.inst = window['wx'];
            this.inst = (this.inst == null) ? this.inst = window['qq'] : this.inst;
            if (window['qg']) {
                this.inst = window['qg'];
            }
        }
        static get instance() {
            if (!!!this._inst) {
                this._inst = new SubPKGLoader();
            }
            return this._inst;
        }
        Load(progressCB, ...pkgs) {
            this.pkgs = pkgs;
            pkgs.forEach(pkg => {
                if (this._loadSub.indexOf(pkg) < 0) {
                    this.pkgProgress[pkg] = 0;
                    this.pkgLoadTag[pkg] = false;
                }
            });
            this.loadProgress = progressCB;
            return new Promise((r, j) => {
                this.loadResolve = r;
                this.loadFail = j;
                let find = false;
                for (var i = 0; i < this.pkgs.length; i++) {
                    var key = this.pkgs[i];
                    if (this.pkgProgress[key] == 0) {
                        this.loadSubPkg(key);
                        find = true;
                    }
                }
                if (!find) {
                    this.loadResolve();
                }
            });
        }
        loadSubPkg(pkg) {
            if (this.inst == null) {
                this.loadSubPkgProgress(pkg, 1);
                this.loadSubPkgSuccess(pkg);
                return;
            }
            let successFunc = () => { this.loadSubPkgSuccess(pkg); };
            let failFunc = (e) => { this.loadSubPkgFail(pkg, e); };
            successFunc.bind(this);
            failFunc.bind(this);
            var task = this.inst.loadSubpackage({
                name: pkg,
                success: successFunc,
                fail: failFunc
            });
            task.onProgressUpdate(res => {
                var progress = res.totalBytesWritten / res.totalBytesExpectedToWrite;
                this.loadSubPkgProgress(pkg, progress);
            });
        }
        loadSubPkgProgress(pkg, progress) {
            this.pkgProgress[pkg] = progress;
            var fragment = 1 / this.pkgs.length;
            var p = 0;
            this.pkgs.forEach(pkg => {
                p += this.pkgProgress[pkg] * fragment;
            });
            if (!!this.loadProgress) {
                this.loadProgress(p);
            }
        }
        loadSubPkgSuccess(pkg) {
            console.log('loadSubPkgSuccess', pkg);
            this.pkgProgress[pkg] = 1;
            this.pkgLoadTag[pkg] = true;
            var allcomplete = true;
            if (this._loadSub.indexOf(pkg) < 0) {
                this._loadSub.push(pkg);
            }
            for (var i = 0; i < this.pkgs.length; i++) {
                if (this.pkgLoadTag[this.pkgs[i]] == false) {
                    allcomplete = false;
                }
            }
            if (allcomplete) {
                this.loadResolve();
            }
        }
        loadSubPkgFail(pkg, e) {
            console.error(`load subpkg{${pkg}} error!, e`);
            this.loadFail(pkg);
        }
        checkSubLoaded(sub) {
            return this._loadSub.indexOf(sub) >= 0;
        }
    }

    var loadKey;
    (function (loadKey) {
        loadKey["subPkg"] = "subPkg";
        loadKey["uiPkg"] = "uiPkg";
        loadKey["createRes"] = "createRes";
        loadKey["level"] = "level";
    })(loadKey || (loadKey = {}));
    class LoadingState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'LoadingState';
            this.loadProgress = {};
            this.loadTag = false;
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            try {
                ReportMgr.inst.sendReport(ReportType.LOADING_START);
                this.loadingUI = UIManager.inst.ShowUI(LoadingUI, UILayer.base);
                this.loadingUI.mainProgressValue = 0;
                for (var key in loadKey) {
                    this.loadProgress[key] = 0;
                }
                Utils.PromiseQueue(null, () => {
                    var subpackages = [];
                    var sub = GameManager.getInst().gameJsonConfig.subpackages;
                    if (sub != null) {
                        sub.forEach(s => {
                            subpackages.push(s.name);
                        });
                        if (subpackages.indexOf("audio") >= 0) {
                            subpackages.splice(subpackages.indexOf("audio"), 1);
                        }
                    }
                    this.loadingUI.subProgressValue = 0;
                    return SubPKGLoader.instance.Load((p) => {
                        this.loadProgress[loadKey.subPkg] = p;
                        this.loadingUI.subProgressValue = p;
                        this.udpateProgress();
                    }, ...subpackages);
                }, () => {
                    return new Promise(r => {
                        Laya.loader.load([{ url: "assets/unity.zip", type: "plfb" }], Laya.Handler.create(this, () => {
                            r();
                        }));
                    });
                }, () => {
                    this.loadProgress[loadKey.subPkg] = 1;
                    this.loadingUI.subProgressValue = 0;
                    var pkgs = ['fui/jiesuan', 'fui/setting', 'fui/common', 'fui/Main', 'fui/wx', 'fui/ditu', 'fui/yinsizhengce', 'fui/gamesdk'];
                    if (Laya.Browser.onTTMiniGame || Laya.Browser.onPC) {
                        pkgs.push('fui/share');
                    }
                    return fguiUtils.LoadPackage(pkgs).progress(p => {
                        this.loadProgress[loadKey.uiPkg] = p;
                        this.loadingUI.subProgressValue = p;
                        this.udpateProgress();
                    });
                }, () => {
                    this.loadProgress[loadKey.uiPkg] = 1;
                    this.loadingUI.subProgressValue = 0;
                    return ResLoadManager.LoadCreateResP().progress(p => {
                        this.loadProgress[loadKey.createRes] = p;
                        this.loadingUI.subProgressValue = p;
                        this.udpateProgress();
                    });
                }, () => {
                    this.loadProgress[loadKey.createRes] = 1;
                    this.loadingUI.subProgressValue = 0;
                    LevelManager.getInst().onPreCreatePrefab = GameItemAttachMgr.getInst().onPreCreatePrefab.bind(GameItemAttachMgr.getInst());
                    return LevelManager.getInst().GameStartLoad().progress(p => {
                        this.loadProgress[loadKey.level] = p;
                        this.loadingUI.subProgressValue = p;
                        this.udpateProgress();
                    });
                }).then(() => {
                    this.loadTag = true;
                }).catch(e => {
                    console.error(e);
                });
            }
            catch (e) {
            }
        }
        OnUpdate(fsm, elapseSeconds) {
            if (this.loadTag) {
                MyShader.initShader();
                fsm.ChangeState(MainUIState);
            }
        }
        OnLeave(fsm) {
            var jsonData = Laya.LocalStorage.getItem("HadAgreeFirst");
            var d = Number(jsonData);
            if (d != 1) {
                UIManager.inst.PushUI(UIYinSi, UILayer.lowMid);
            }
        }
        OnDestroy(fsm) {
        }
        udpateProgress() {
            var p = 0;
            var fragment = 1 / Object.keys(this.loadProgress).length;
            for (var key in loadKey) {
                p += this.loadProgress[key] * fragment;
            }
            this.loadingUI.mainProgressValue = p;
        }
    }

    var loadFlag;
    (function (loadFlag) {
        loadFlag["gameJson"] = "gamejson";
        loadFlag["loadingUI"] = "loadingUI";
    })(loadFlag || (loadFlag = {}));
    class PreloadState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'PreloadState';
            this.loadFlag = {};
        }
        OnInit(fsm) {
            this.fsm = fsm;
        }
        OnEnter(fsm) {
            GameSDK$1.curentState = this;
            this.loadFlag[loadFlag.gameJson] = false;
            this.loadFlag[loadFlag.loadingUI] = false;
            Laya.loader.load('gamebake.json', Laya.Handler.create(this, this.onGameConfigLoad));
            fguiUtils.LoadPackage(['res/fui/Loading', 'fui/common', 'fui/Main', 'fui/huaweiLogin']).then(() => {
                this.onLoadingResLoad();
                UIManager.inst.PushUI(LoginCoverUI, UILayer.cover);
                GameSDK$1.huaWeiLogin();
            });
        }
        OnUpdate(fsm, elapseSeconds) {
        }
        OnLeave(fsm) {
        }
        OnDestroy(fsm) {
        }
        ChangeLoadingState() {
            UIManager.inst.PopUI(UILayer.cover, true);
            this.fsm.ChangeState(LoadingState);
        }
        onGameConfigLoad(json) {
            GameManager.getInst().gameJsonConfig = json;
            this.loadFlag[loadFlag.gameJson] = true;
            this.checkLoadFlag();
        }
        onLoadingResLoad() {
            this.loadFlag[loadFlag.loadingUI] = true;
            this.checkLoadFlag();
        }
        checkLoadFlag() {
            for (var key in loadFlag) {
                if (this.loadFlag[key] == false)
                    return;
            }
        }
    }

    class huaweiLoginBinder {
        static bindAll() {
            fgui.UIObjectFactory.setExtension(UI_ui_huaweicover.URL, UI_ui_huaweicover);
        }
    }

    class GameLauchState extends FsmState {
        constructor() {
            super(...arguments);
            this.name = 'GameLauchState';
        }
        OnInit(fsm) {
        }
        OnEnter(fsm) {
            PlatformManager.instance.init();
            PlatformManager.instance.initPlatform();
            GameItemAttachMgr.getInst().Register();
            fgui.UIConfig.packageFileExtension = 'bin';
            fgui.UIConfig.buttonSound = Global.soundDir + 'button.mp3';
            Laya.stage.addChild(fgui.GRoot.inst.displayObject);
            fgui.GRoot.inst.displayObject.zOrder = 100;
            MainBinder.bindAll();
            LoadingBinder.bindAll();
            BattleBinder.bindAll();
            huaweiLoginBinder.bindAll();
        }
        OnUpdate(fsm, elapseSeconds) {
            fsm.ChangeState(PreloadState);
        }
        OnLeave(fsm) {
        }
        OnDestroy(fsm) {
        }
    }

    class GameManager extends Singleton {
        constructor() {
            super();
            this.adShowed = false;
            this.battleManager = new BattleManager();
            MsgManager.getInst().on('AD_Showed', this, this.onADShowed);
        }
        onADShowed() {
            Utils.WaitFrame(3).then(() => {
                if (PlayerData.CurrentLevelID >= GameManager.HangMuUnlockScene) {
                    this.adShowed = true;
                }
            });
        }
        Run() {
            GameSDK$1.ctor();
            this.gameProcedureFsm = FsmManager.getInst().CreateFsm('game_manager', this.battleManager, [
                new GameLauchState(),
                new PreloadState(),
                new LoadingState(),
                new DebugDataState(),
                new MainUIState(),
                new BattleInitCampState(),
                new BattleSceneLoadState(),
                new BattlePreparetionState(),
                new BattleState(),
                new BattleWinState()
            ], true);
            this.gameProcedureFsm.Start(GameLauchState);
            this.gameProcedureFsm.SetData('debug', true);
            Laya.timer.frameLoop(1, this, this.OnUpdate);
        }
        Pause() {
            Laya.timer.clear(this, this.OnUpdate);
        }
        Resume() {
            Laya.timer.frameLoop(1, this, this.OnUpdate);
        }
        Stop() {
            Laya.timer.clear(this, this.OnUpdate);
        }
        Reset() {
            Laya.timer.callLater(this, () => {
                this.gameProcedureFsm.Owner.ClearGame();
                BattleData.CleasrBigSceneData();
                AlertMgr.inst.resetPlayerLevel();
                this.gameProcedureFsm.ChangeState(BattleSceneLoadState);
            });
        }
        OnUpdate() {
            FsmManager.getInst().Update(LayaUtils.deltaTimeSec * Laya.timer.scale);
        }
    }
    GameManager.NukeRadius = 5.2;
    GameManager.ShipRadius = 6;
    GameManager.HelicopterRadius = 4;
    GameManager.HangMuUnlockScene = 3;
    GameManager.MechBoxUnlockScene = 2;

    class Main {
        constructor() {
            var cfg3D = new Config3D();
            cfg3D.isAntialias = true;
            cfg3D.enableMultiLight = false;
            if (window["Laya3D"])
                Laya3D.init(GameConfig.width, GameConfig.height, cfg3D);
            else
                Laya.init(GameConfig.width, GameConfig.height, Laya["WebGL"]);
            Laya["Physics"] && Laya["Physics"].enable();
            Laya["DebugPanel"] && Laya["DebugPanel"].enable();
            Laya.stage.scaleMode = GameConfig.scaleMode;
            Laya.stage.screenMode = GameConfig.screenMode;
            Laya.stage.alignV = GameConfig.alignV;
            Laya.stage.alignH = GameConfig.alignH;
            Laya.URL.exportSceneToJson = GameConfig.exportSceneToJson;
            if (GameConfig.debug || Laya.Utils.getQueryString("debug") == "true")
                Laya.enableDebugPanel();
            if (GameConfig.physicsDebug && Laya["PhysicsDebugDraw"])
                Laya["PhysicsDebugDraw"].enable();
            if (GameConfig.stat)
                Laya.Stat.show(0, 100);
            if (!Laya.Browser._isMiniGame) {
                Laya.alertGlobalError(true);
            }
            Laya.ResourceVersion.enable("version.json", Laya.Handler.create(this, this.onVersionLoaded), Laya.ResourceVersion.FILENAME_VERSION);
            Laya.stage.frameRate = Laya.Stage.FRAME_FAST;
        }
        onVersionLoaded() {
            Laya.AtlasInfoManager.enable("fileconfig.json", Laya.Handler.create(this, this.onConfigLoaded));
        }
        onConfigLoaded() {
            GameManager.getInst().Run();
        }
    }
    new Main();

})();
//# sourceMappingURL=bundle.js.map
